"""
File        : GraphCluster.py
Author      : Pablo Boixeda & Ramon Aragues
Creation    : 4.2005
Contents    : Implements a Graph where the nodes are clusters composed of node attributes of another type of graph
Called from : Programs that implements graph clustering

=======================================================================================================

Implements a Graph where the nodes are clusters composed of node attributes of another type of graph

"""

# GraphCluster.py: Implements a Graph where the nodes are clusters composed of node attributes of another type of graph
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

from Graph import *
from GraphNode import *
from GraphClusterNodeAttribute import *

verbose = 0

verbose_detailed = 0

#-----------------------------------------------------------------------------------------------
class GraphCluster(Graph):
#-----------------------------------------------------------------------------------------------

    def __init__(self, graph_id=None):
        """
        Method used to initialize ClusterGraph object
        """
        
        self.old_node_correspondence={}

        self.node_id_counter = 0
        
        Graph.__init__(self, graphID= graph_id)

    def _get_new_node_id(self):
        """
        This method is created to give an id to new nodes

        Attention!!! Node Ids are normally controlled by class Clustering: this method is only used when we are initializing
        the GraphCluster from a Graph object (one cluster for each Node in Graph)
        """
        new_node_id = self.node_id_counter
        self.node_id_counter += 1
        return new_node_id


    def get_node_alternative_id(self, node_attribute_object):
        """
        returns the alternative id that will describe this node
        """
        alternative_id = "%s" %(node_attribute_object.get_node_id())

        # create string that will describe this node
        for one_element in node_attribute_object.get_list_elements():
            alternative_id += ".%s" %(one_element.get_node_id())

        return alternative_id

    
    def initialize_from_graph(self, graph):
        """
        initialize the GraphCluster from a Graph: places one GraphNodeAttribute in each cluster (ie. in each GraphClusterNodeAttribute)
        """
        if graph is None:    raise ValueError ("No Graph given to transform into GraphCluster")

        if verbose:   sys.stderr.write("Initializing GraphCluster from graph\n")

        dic_correspondences_nodes = {}   # dic that follows this structure:
                                        #
                                        #     { node_id in graph: node_object in GraphCluster,
                                        #       node_id in graph: node_object in GraphCluster,
                                        #       ................
                                        #     }

            
        for one_graph_edge in graph.get_edge_object_list():
            # for each edge, create a link between two Graph nodes (with GraphClusterNodeAttribute attributes) 
            # --
            # NODE a
            # --
            graph_node_id_a = one_graph_edge.get_start_node_id()

            if dic_correspondences_nodes.has_key(graph_node_id_a):
                # the GraphClusterNode was already created for this GraphNode... use it
                cluster_node_a = dic_correspondences_nodes[graph_node_id_a]
                cluster_id_a = cluster_node_a.get_node_id()
            else:
                # a new GraphClusterNode needs to be created.

                # get the attribute for graph_node_id_a
                graph_node_attribute_a = graph.get_node(identifier= graph_node_id_a, get_mode="error").get_node_attribute_object()

                cluster_id_a = self._get_new_node_id()

                # create nodes for IRGraph (standard GraphNode with a IRGraphNodeAttribute)
                cluster_attribute_a = GraphClusterNodeAttribute(node_id=cluster_id_a,
                                                                list_elements=[graph_node_attribute_a])

                cluster_node_a = GraphNode(nodeID= cluster_id_a , attribute=cluster_attribute_a,
                                           original = 1,
                                           alternative_id= self.get_node_alternative_id(node_attribute_object=cluster_attribute_a))

                
                self.add_node(cluster_node_a)
                dic_correspondences_nodes[graph_node_id_a] = cluster_node_a
            # END OF else: (if dic_correspondece_nodes.has_key(node_id_a):)
            
            # --
            # NODE b
            # --
            graph_node_id_b = one_graph_edge.get_end_node_id()

            if dic_correspondences_nodes.has_key(graph_node_id_b):
                # the GraphClusterNode was already created for this GraphNode... use it
                cluster_node_b = dic_correspondences_nodes[graph_node_id_b]
                cluster_id_b = cluster_node_b.get_node_id()
            else:
                # a new GraphClusterNode needs to be created.

                # get the attribute for graph_node_id_b
                graph_node_attribute_b = graph.get_node(identifier= graph_node_id_b, get_mode="error").get_node_attribute_object()

                cluster_id_b = self._get_new_node_id()

                # create nodes for IRGraph (standard GraphNode with a IRGraphNodeAttribute)
                cluster_attribute_b = GraphClusterNodeAttribute(node_id=cluster_id_b,
                                                                list_elements=[graph_node_attribute_b])

                cluster_node_b = GraphNode(nodeID= cluster_id_b , attribute=cluster_attribute_b ,
                                           original = 1,
                                           alternative_id= self.get_node_alternative_id(node_attribute_object=cluster_attribute_b) )
                
                self.add_node(cluster_node_b)
                dic_correspondences_nodes[graph_node_id_b] = cluster_node_b
            # END OF else: (if dic_correspondece_nodes.has_key(node_id_b):)

            # --
            # EDGE
            # --

            # create edge for GraphCluster (standard GraphEdge)
            new_edge_attribute = GraphEdgeAttribute() # empty attribute: nothing to do with it at the moment
            new_edge= GraphEdge(node1_id=cluster_id_a , node2_id= cluster_id_b,
                                attribute_object= new_edge_attribute)
            
            self.add_edge(new_edge)
            
        # END OF for one_graph_edge in graph.get_edge_object_list():


        

    def do_action(self):
        """
        this is a generic method that can be used by particular GraphCluster subclasses to do
        something after the clustering of this level (eg. in CIR, we fuse as well the clusters
        that have a common cluster partner).

        It is called from Clustering.cluster_graph after each step of the clustering

        In those GraphClusters that nothing has to be done are not affected, since this method won't exist
        and this generic method will be called (which doesn't do anything)
        """
        pass
    
    def create_grouped_node(self, node_id1= None, node_id2= None, new_node_id= None, old_graph= None):
        """
        This method joins two GraphNodes into one GraphNode. These two node attributtes are inserted
        into the new node attribute. This attribute is a list of attributes

        returns the resulting node

        "node_id1" is the id from one of the clusters to join
        
        "node_id2" is the id from one of the clusters to join
        
        "new_node_id" is the id of the new cluster node to create.
        
        "old_graph" is the graph that contains the nodes with node_id1 and node_id2
        
        """
        if node_id1 is None or node_id2 is None:
            raise ValueError("Error: how can I create a clustered node from a None id?")

        if verbose_detailed:
            sys.stderr.write("Joining nodes %s and %s to form a new node with id %s\n" %(node_id1, node_id2, new_node_id))

        node1=old_graph.get_node(node_id1)
        node2=old_graph.get_node(node_id2)
        
        if new_node_id is None:
            # if user didn't give us a node id, get it here from the old_graph
            new_node_id= self._get_new_node_id()

        NewClusterNodeAttribute= GraphClusterNodeAttribute(node_id= new_node_id)
        NewClusterNodeAttribute.add_element_list(list_node_attribute_object= node1.get_node_attribute_object().get_list_elements())
        NewClusterNodeAttribute.add_element_list(list_node_attribute_object= node2.get_node_attribute_object().get_list_elements())
            
        
        NewNode=GraphNode(nodeID= new_node_id, 
                          alternative_id= self.get_node_alternative_id(node_attribute_object=NewClusterNodeAttribute),
                          attribute = NewClusterNodeAttribute,
                          original = 0,
                          graph = self)

        return NewNode

    

    def print_cluster_composition(self, output_target= sys.stdout):
        """
        prints the node ids for each cluster in the graph

        This is very general: should be overwritten by a method specific to the clustering being performed
        """
        nodes=self.get_node_object_list()
        
        for node in nodes:
            # node is a cluster, which has an attribute that contains elements (each element is a GraphNodeAttribute of
            #                                                                   the graph we are clustering)
            output_target.write( "cluster %s: " %node.get_node_id() )
            elements_list=node.get_node_attribute_object().get_list_elements()
            for element in elements_list:
                output_target.write("%s --" %(element.get_node_id()) )
            output_target.write("\n")

    def print_pairs_same_cluster(self, output_target):
        """
        prints pairs of elements  that appear in the same cluster

        To be overwritten by the subclass of GraphCluster in case an element is not described by its identifier
        """
        nodes=self.get_node_object_list()
        
        for node in nodes:
            elements_list=node.get_node_attribute_object().get_list_elements()
            number_of_elements = len(elements_list)
            for i in range(number_of_elements):
                for j in range(i+1, number_of_elements):
                    output_target.write( "p;%s-%s\n" %(elements_list[i].get_node_id(), elements_list[j].get_node_id()) )
                # END OF for j in range(i+1, number_of_elements):
            # END OF for i in range(number_of_elements):
        # END OF for node in nodes:
        
    
    def print_cluster_interactions(self, output_target= sys.stdout):
        """
        prints the edge table of the graph

        This is very general: should be overwritten by a method specific to the clustering being performed
        """
        self.output_edges_table(output_target=output_target)
            

    def print_pairs_interactions(self, output_target):
        """
        prints pairs of elements whose clusters interact

        To be overwritten by the subclass of GraphCluster in case an element is not described by its identifier
        """
        for edge in self.get_edge_object_list():

            node_id_start = edge.get_start_node_id()
            node_start = self.get_node(identifier= node_id_start, get_mode="error") 
            list_elements_start = node_start.get_node_attribute_object().get_list_elements()
            
            node_id_end = edge.get_end_node_id()
            node_end = self.get_node(identifier= node_id_end, get_mode="error") 
            list_elements_end = node_end.get_node_attribute_object().get_list_elements()

            for one_element_start in list_elements_start:
                for one_element_end in list_elements_end:
                    output_target.write("i;%s-%s\n" %(one_element_start.get_node_id(), one_element_end.get_node_id()) )
                # END OF for one_element_end in list_elements_end:
            # END OF for one_element_start in list_elements_start:



    # TO DO!!! These three methods (print_proteins_* and print_cir_*) should not be here but on a higher class IPGGraphCluster

    def print_proteins_same_cluster(self, output_target, root_protein):
        """
        prints to "output_target" pairs of proteins (ie. proteins with elements in that cluster)  that appear in the same cluster
           -> it doesn't print out clusters with just one element

        "root_protein" is used to uniquely identify the clusters (clusters have unique ids if they were all obtained from the
                       same network... but since we are doing multiple calls from run_piana_protein_by_protein, clusters
                       need to be distinguished by their id in the network and the root protein that was used to generate them 
        """

        for node in self.get_node_object_list():
            elements_list= node.get_node_attribute_object().get_list_elements()
            
            #this_file_fd = file("/home/raragues/deleteme.txt", "w")
            #this_file_fd.write("elements in this cluster are: %s\n" %(elements_list))
            
            number_of_elements = len(elements_list)
            for i in range(number_of_elements):
                for j in range(i+1, number_of_elements):
		    # the way this loop is designed makes sure that only clusters with more than one element are printed out.
                    #  -> ie. if number of elements is 1, the process will never get into the inside loop

                    proteinPiana_i = elements_list[i].get_proteinPiana()
                    proteinPiana_j = elements_list[j].get_proteinPiana()
                    #this_file_fd.write("i:%s:%s j:%s:%s\n" %(i, proteinPiana_i, j, proteinPiana_j))
                    if proteinPiana_i != proteinPiana_j:
                        # not checking duplication: otherwise, we lose information on who's in each cluster

                        cluster_name = "%s.%s" %(root_protein, node.get_node_id())
                        
                        #this_file_fd.write("printing: share\t%s\t%s\t%s\n" %(cluster_name,proteinPiana_i ,proteinPiana_j ) )
                        output_target.write("share\t%s\t%s\t%s\n" %(cluster_name,proteinPiana_i , proteinPiana_j) )
                # END OF for j in range(i+1, number_of_elements):
            # END OF for i in range(number_of_elements):
        # END OF for node in self.get_node_object_list():
            

    def print_proteins_interactions(self, output_target, root_protein, similarity_threshold):
        """
        prints  to "output_target" pairs of proteins whose clusters interact, but only
        between clusters that have more than one element

        --> It is important to take into account that a cluster will have more than one element when 
            its proteins have a similarity score above the current threshold

        
        "root_protein" is used to uniquely identify the clusters (clusters have unique ids if they were all obtained from the
                       same network... but since we are doing multiple calls from run_multiple_pianas, clusters
                       need to be distinguished by their id in the network and the root protein that was used to generate them 
        """


        # ATTENTION!!! This, as it is implemented right now, won't work for metrics other than num_ints. Why? because
        #              in order to consider a CIR-CIR interaction as valid, the checking I do is based on the number
        #              of elements and not using ratios

        # in this method, all protein pairs are printed, even if they are repeated
        #  why? because it is important to know if the interaction comes from a cluster
        #       or from an original node, so that train_and_test_cir can detect which
        #       interactions correspond to clusters with more than one element.
        for edge in self.get_edge_object_list():

            node_id_start = edge.get_start_node_id()
            node_start = self.get_node(identifier= node_id_start, get_mode="error") 
            list_elements_start = node_start.get_node_attribute_object().get_list_elements()
            
            node_id_end = edge.get_end_node_id()
            node_end = self.get_node(identifier= node_id_end, get_mode="error") 
            list_elements_end = node_end.get_node_attribute_object().get_list_elements()

            # one of the two lists (list_elements_start or list_elements_end) is always an original cluster
            # (because, by definition of the method, there are no interactions between cluster). Moreover,
            # the elements inside the cluster with more than one element (ie. multiple cluster) do respect
            # the similarity threshold (otherwise, they wouldn't be inside the same cluster). We need to check
            # that the other region type of the interaction (the region type shared by all partners of the
            # multiple cluster) also respects the similarity threshold. How is that done? By checking how many
            # elements are there in the multiple cluster: the number of elements in the cluster is the number
            # of proteins that are common to the proteins that interact with the cluster, which is the same
            # as saying that the number of elements in the multiple cluster is the similarity among the partners of
            # the multiple cluster.
            #   Brief... just by checking that the number of elements in the multiple cluster is above the
            # current similarity score, we are imposing that both CIRs of the interaction are above the threshold
            num_elements_start = len(list_elements_start)
            num_elements_end = len(list_elements_end)

            if num_elements_start > 1:
                # comment J:
                # start is the cluster that has more than one element (ie. we are sure that its elements are above the similarity threshold)
                #  -> check that the number of elements is also above the similarity threshold (see explanation above: the number of elements
                #                                                                               is (by definition) the similatity for all
                #                                                                               clusters interacting with this cluster)
                if num_elements_start < similarity_threshold:
                    # comment K:
                    # if there are less than similarity threshold elements in the cluster, this is not a valid CIR-CIR interaction
                    # One of the CIRs is valid (and that's why those proteins are together in a cluster) but the other doesn't have
                    # enough interactions is common to be considered a CIR
                    continue
                
            elif num_elements_end > 1:
                # see comment J
                if num_elements_end < similarity_threshold:
                    # see comment K
                    continue
            else:
                # None of the clusters of this edge is a CIR (ie. they just contain one element). This is not a CIR-CIR
                # interaction we want to print
                continue

            
            for one_element_start in list_elements_start:
                for one_element_end in list_elements_end:

                    proteinPiana_start = one_element_start.get_proteinPiana()
                    proteinPiana_end = one_element_end.get_proteinPiana()
                    
                    cluster_1_name = "%s.%s" %(root_protein, node_start.get_node_id())
                    cluster_2_name = "%s.%s" %(root_protein, node_end.get_node_id())

                    output_target.write("int\t%s-%s\t%s\t%s\n" %(cluster_1_name, cluster_2_name, proteinPiana_start, proteinPiana_end) )
                # END OF for one_element_end in list_elements_end:
            # END OF for one_element_start in list_elements_start:
        # END OF for edge in self.get_edge_object_list():


    
    def _get_elements_key(self, elements_list):
        """
        returns an elements unique key based on the elements in "elements_list"

        attention! elements_list should not contain repetitions
        
        """

        elements_ids_list = []
        for one_element in elements_list:
            elements_ids_list.append(one_element.get_node_id())
        
        elements_ids_list.sort()

        elements_key = "%s" %elements_ids_list[0]
        
        for one_element_id in elements_ids_list[1:]:
            elements_key += ".%s" %(one_element_id)

        return elements_key


    def print_cir_cir_ints(self, output_target, root_protein):
        """
        prints  to "output_target" pairs of proteins whose clusters interact, but doing it in a special way:

                - only clusters with more than one element are considered for printing CIR-CIR interactions
                - However, in the way we have done the clustering for CIRs, two clusters with more
                  than one element never interact with each other (because to speed up things, a new cluster
                  will only interact with original_clusters, which are formed by just one protein)
                - therefore, we must 'infer' the CIR-CIR interactions by following the process detailed below in the comments
                  Basically, it consists in looking to the elements that interact with a cluster and inferring which
                  cluster is 'representing' those interacting elements in a single cluster.

        Attention! THis is a very strict way of determining a CIR-CIR interaction! This will only find CIR-CIR where all
                   the interactions are explained by the CIR. For a less restrictive interpretation, use print_
        
        "root_protein" is used to uniquely identify the clusters (clusters have unique ids if they were all obtained from the
                       same network... but since we are doing multiple calls from run_multiple_pianas, clusters
                       need to be distinguished by their id in the network and the root protein that was used to generate them 
        """
        already_printed_pairs = {}  # to avoid printing repetitions


        # the process to follow is this:
        #
        #    1. Create a dictionary that describes each cluster by its member elements
        #        
        #    2. Create a dictionary that describes for each cluster the elements that interact with it
        #        2.1 For each cluster, find elements that interact with it (by looking to cluster partners)
        #
        #    3. Infer CIR-CIR interactions from this information
        #       

        dic_cir_ids = {}    # used to describe a CIR by the elements that are in it
                            #
                            # this dictionary will follow the structure:
                            #  { elements_key: CIR id that contains them,
                            #    elements_key: CIR id that contains them,
                            #    .....................
                            #  }
                            #
                            # the elements key is formed by sorting the elements list and
                            # converting their ids to a string separated by dots

        dic_ints_cir = {}   # used to keep which elements interact with a given CIR
                            #
                            # this dictionary will follow the structure:
                            #  { elements_key: CIR id that interacts with them,
                            #    elements_key: CIR id that interacts with them,
                            #    .....................
                            #  }
                            #
                            # the elements key is formed as described in dic_cir_ids
                            

        
        nodes= self.get_node_object_list()
        
        for node in nodes:
            
            cluster_id = node.get_node_id() # this is the unique id for this cluster
            
            elements_list= node.get_node_attribute_object().get_list_elements()

            if len(elements_list)== 1:
                # we are not interested in clusters that only have one element
                continue
            
            # 2.
            
            # 2.1: get clusters that interact with node
            #        --> by definition, this partner clusters will only have one element each one
            cluster_partner_ids = node.get_neighbour_ids()


            if len(cluster_partner_ids)<= 1:
                # we are not interested in this cluster if it only interacts with one single cluster (ie. one element)
                continue

            # 1
            dic_cir_ids[self._get_elements_key(elements_list)] = cluster_id


            # 2.2: create a list with all elements in all clusters from 2.1
            partners_elements_list = []
            for one_cluster_id in cluster_partner_ids:
                partners_elements_list.extend( self.get_node(identifier=one_cluster_id,
                                                   get_mode="error").get_node_attribute_object().get_list_elements())


            dic_ints_cir[self._get_elements_key(partners_elements_list)] = cluster_id
        # END OF for node in nodes:


        # at this point, we can find any CIR through its elements (dic_cir_ids) and we can as well
        # find which CIR interacts with a certain list of elements (dic_ints_cir)
        # We can use this information to infer CIR-CIR interactions by:
        #  for each elements_key in dic_cir_ids (ie. for each cluster with more than one element in it)
        #      3.1 get the id for cluster  that has them (A)
        #      3.2 check if there is a cluster interacting with elements_key
        #          ie. a cluster B has interactions with a group of elements represented by cluster A?
        #      3.3 if there is a cluster B interacting with elements_key, create CIR-CIR interaction between A and B
        #          -> print interactions between all elements in A and all elements in B

        for one_elements_key in dic_cir_ids:
            cluster_a_id = dic_cir_ids[one_elements_key]

            if dic_ints_cir.has_key(one_elements_key):
                # if there is a cluster b that 'represents' all elements that interact with cluster_a_id,
                #    -> this is a CIR-CIR interaction: print interaction between proteins to output_target
                cluster_b_id = dic_ints_cir[one_elements_key]
                
                elements_in_a = self.get_node(identifier=cluster_a_id,
                                              get_mode="error").get_node_attribute_object().get_list_elements()

                elements_in_b = self.get_node(identifier=cluster_b_id,
                                              get_mode="error").get_node_attribute_object().get_list_elements()

                for one_element_start in elements_in_a:
                    for one_element_end in elements_in_b:
                        
                        proteinPiana_start = one_element_start.get_proteinPiana()
                        proteinPiana_end = one_element_end.get_proteinPiana()

                        # make sure that no duplication of pairs occurs...
                        if proteinPiana_start <= proteinPiana_end:
                            p1 = proteinPiana_start
                            cluster_1_name = "%s.%s" %(root_protein, cluster_a_id)
                            p2 = proteinPiana_end
                            cluster_2_name = "%s.%s" %(root_protein, cluster_b_id)
                        else:
                            p1 = proteinPiana_end
                            cluster_1_name = "%s.%s" %(root_protein, cluster_b_id)
                            p2 = proteinPiana_start
                            cluster_2_name = "%s.%s" %(root_protein, cluster_a_id)

                        key = "%s.%s" %(p1, p2)
                        if already_printed_pairs.has_key(key):
                            continue
                        already_printed_pairs[key] = None

                        output_target.write("cir_cir\t%s-%s\t%s\t%s\n" %(cluster_1_name, cluster_2_name, p1, p2) )
   
                    # END OF for one_element_b in elements_in_b:
                # END OF for one_element_a in elements_in_a:
                
            # END OF if dic_ints_cir.has_key(one_elements_key):
        # END OF for one_elements_key in dic_cir_ids:
    



 
