"""
File       : PianaGraphBuilder.py
Author     : R. Aragues & D. Jaeggi
Creation   : 31.07.2003
Contents   : class for building a piana graph from a database
Called from: PianaGraph and piana.py
Comments :
=========================================================================

This class is used to populate a PianaGraph object with interactions and proteins from a piana database

"""

# PianaGraphBuilder.py: implements a class for building protein-protien interaction networks from a piana database
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys

from GraphBuilder import *
from PianaDB import *
from PianaDBaccess import *
from sets import *
from PianaGraphNodeAttribute import *
from PianaGraphEdgeAttribute import *

import time

verbose = 0

class PianaGraphBuilder(GraphBuilder):
    """
    Class that implements graph building methods specific to Piana
    
    """
    def __init__(self, piana_access_object, depth=0,  hub_threshold= 0, use_self_ints='yes', list_protein_codes=None, id_type = None,
                 tax_id_value=0, list_source_dbs="all",  inverse_dbs="no", list_source_methods="all", inverse_methods="no", 
		 use_secondary_db="no", piana_secondary_access_object=None, force_secondary_db="no"):
        
        """
        Initilisation operations specific to Piana: mainly finding out the internal id (ie proteinPiana) for
        protein given as argument, and then calling the general __init__ of superclass GraphBuilder
        
        "depth" valid values are described in superclass GraphBuilder. It refers to the depth the graph should have.
        
        "list_protein_codes" are protein identifiers of id_type . These protein codes will be used to build the graph. All codes must be
        of the same type (e.g swissprot)
        
        "id_type" is an easy-to-remember  protein type name: establishes the type of protein identifiers that are being passed as argument.

           It is used to retrieve the internal proteinPiana from the database, which is as well the identifier used for node_ids of the graph. 
           Valid id_type are those defined in PianaGlobals.py. There is a direct correspondance between these types and the
           column names in the database

           valid "id_type" values are those described in the PIANA reference card (read README.piana_tutorial)



        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself

                      
        "tax_id_value" can be used to fix the species we want to work on: only proteinPianas from that species will be used to build the graph
        (this doesn't mean that 'a posteriori' added proteins will be limited as well to that species... it only concerns proteins in list_protein_codes)
        If the database has an interaction between a protein of this species and a protein of another species, both proteins will appear in the network
        regardless of this paramter: to control which proteins appear in your output you should use output-species-name in your piana configuration file

             --> valid values are 0 or a taxonomy id as defined by NCBI taxonomy
                    - 0: retrieves all proteinPianas for protein_codes, regardless of species

        "list_source_dbs" can be used to set which interaction databases will be used to build the network: only interactions from databases in the list
        will be added to the network

          -> "list_source_dbs" can be "all" or a list of interactions databases

           - "all": uses all interactions in the piana database
           - [db1, db2, ...]: uses only interactions from databases listed (for example, [dip, string, posas]
              -> valid database names are those defined in PianaGlobals.interaction_databases

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)
                       

         "list_source_methods" can be used to set which interaction detection methods will be used to build the network: only interactions
         that were detected through methods in the list will be added to the network

          -> "list_source_methods" can be "all" or a list of interactions methods

           - "all": uses all interaction detection methods in the piana database
           - [method1, method2, ...]: uses only interactions from methods listed (for example, [y2h, immuno]
              -> valid method names are those defined in PianaGlobals.method_names

        "inverse_methods" can be:
                       no (methods in list_source_methods will be used to build the network)
                       yes (all methods except those in list_source_methods will be used to build the network)
          
        """

        self.piana_access = piana_access_object
	self.use_secondary_db = use_secondary_db
        self.piana_secondary_access = piana_secondary_access_object
	self.force_secondary_db = force_secondary_db
        self.hub_threshold = hub_threshold
        self.list_source_dbs = list_source_dbs
        self.inverse_dbs = inverse_dbs
        self.list_source_methods = list_source_methods
        self.inverse_methods = inverse_methods
        self.use_self_ints = use_self_ints

        #javi:
        self.tax_id_value = tax_id_value

        # populate the list with node_ids that will be used to build the graph. node_ids are proteinPianas.
        #  since a protein_code can have several proteinPianas, the list of node_ids can be longer than the list of protein_codes
        list_node_id = []


        # Dictionary used to mantain all protein-protein interactions in memory in case it is necessary
        # It is used when there are a lot of nodes in the graph and consequently there are a lot of accessions to the database
        self.all_ppis = {}

        for protein_id in list_protein_codes:
            # Translating the identifier passed as argument into the piana internal identifier for a node
            # The identifier passed as argument must be translated because the id for graph nodes is proteinPiana
            list_node_id.extend( self.piana_access.get_list_proteinPiana(protein_id= protein_id,
                                                                         id_type= id_type,
                                                                         tax_id_value = tax_id_value,
                                                                         source_db_info="no") )
           
        # END OF for protein_code in list_protein_code:
        # remove duplicates

        list_node_id = list(Set(list_node_id))
        
        # OLD CODE (previous to optimization)
        # remove duplicates
        #dic_node_ids = {}
        #for node_id in list_node_id:
        #    dic_node_ids[node_id] = None

        #list_node_id= dic_node_ids.keys()
        
        if not list_node_id:
            sys.stderr.write("No protein found for your protein codes. Network could no be built\n")
            raise ValueError("No protein found for your protein codes. Network could no be built\n")
        else:
            # Call the superclass init method with pianaID identifiers for the nodes
            GraphBuilder.__init__(self, depth= depth, list_node_id = list_node_id, hub_threshold=hub_threshold)


    def _load_all_protein_protein_interactions(self):
        """
        Fills the attribute all_ppis with all protein-protein interactions of the database

        It speeds up the construction of networks when they are very large
        """
        initial_time=time.time()
        sys.stderr.write("Loading all protein-protein interactions in memory to speed up process...\n")
        temp_ppi_triplet_list = self.piana_access.get_all_protein_protein_interactions(use_self_ints=self.use_self_ints,
                                                                                       list_source_dbs=self.list_source_dbs,
                                                                                       inverse_dbs=self.inverse_dbs,
                                                                                       list_source_methods=self.list_source_methods,
                                                                                       inverse_methods=self.inverse_methods,
                                                                                       tax_id_value = self.tax_id_value)
        sys.stderr.write("Interactions loaded in %s seconds. Preparing them to an optimized search...\n" %(time.time()-initial_time))
        #self.all_ppis = dict([((proteinPiana_a,proteinPiana_b),interactionPiana) for proteinPiana_a,proteinPiana_b,interactionPiana in temp_ppi_triplet_list])
        initial_time=time.time()
        for triplet in temp_ppi_triplet_list:
            try:
                self.all_ppis[triplet[0]][triplet[1]]=triplet[2]
            except KeyError:
                self.all_ppis[triplet[0]]={}
                self.all_ppis[triplet[0]][triplet[1]]=triplet[2]
            try:
                self.all_ppis[triplet[1]][triplet[0]]=triplet[2]
            except KeyError:
                self.all_ppis[triplet[1]]={}
                self.all_ppis[triplet[1]][triplet[0]]=triplet[2]
        sys.stderr.write("done in %s seconds\n" %(time.time()-initial_time))
            
        temp_ppi_triplet_list = []


    def check_all_ppis_empty(self):
        """
        Check if the internal buffer (all ppis dictionary) is empty or not

        Returns None if it is empty. Otherwise, return the dictionary
        """

        if not self.all_ppis:
            return None
        else:
            return self.all_ppis

            
    def create_node_attribute(self, node_id):
        
        """
        Overrides method from GraphBuilder
        
        returns a PianaGraphNodeAttribute for node_id
        """

        if self.graph.get_memory_mode() == "speed":
            mem_mode="inFlight"
        else:
            mem_mode="onDemand"
            
        return PianaGraphNodeAttribute(proteinPiana_value= node_id,  piana_access = self.piana_access, mem_mode = mem_mode)

       
    def get_link_details(self, node_id, use_ppi_memory="no", mem_mode="onDemand"):
        """
        Overrides method from GraphBuilder: this method specifies how specific nodes and edges look like in PianaGraph
        
        Before calling this method we are only dealing with node_id. get_links will create the node objects from node_id and return
        linkDetail objects (container for two node_id, one edge and attributes of all of them). This linkDetails objects are specific
        to PianaGraph, but can be used by GraphBuilder without knowing their internal structure.
        
        This method uses self.list_source_dbs to limit the interactions found to those that were retrieved from databases that appear
        in the list of databases given by the user when initializing the builder
        
        This method uses self.list_source_methods to limit the interactions found to those that were found with methods that appear
        in the list of methods given by the user when initializing the builder
        
        returns empty list if no link_details are found or number of links is higher than hub_threshold
        
        """


        list_link_detail_objects = []

        if verbose:
            sys.stderr.write("in PianaGraphBuilder: list source dbs is %s and list source methods is %s\n" %(self.list_source_dbs, self.list_source_methods))

        # Getting all interaction_id where node_id is involved (respecting restrictions... ignore_unreliable is controlled by
        #                                                       piana.py, which will place on the list_source_dbs the appropiate source dbs)
        #
        # The behavior for retrieving the interactions is the following:
        #
        #    -> if force_secondary_db is yes, then use the secondary database (and that's all... nothing else to check)
        #           -> it means the user only wants to use the secondary database to retrieve interactions
        #    -> if force_secondary is no, then there are several possibilities:
        #         1. the user has set use_secondary_db to no, then get partners from primary
        #           -> he doesn't want to use a secondary DB even in cases where no interactions are found in the primary DB
        #         1. the user has set use_secondary_db to yes, then get partners from primary
        #                1.1 if partners are found, keep those and forget about the secondary database
        #                1.2 if no partners are found, look for partners in the secondary database
        if self.force_secondary_db == "yes":

            if not self.piana_secondary_access:
                raise ValueError("You are forcing to use a secondary PIANA database, but you didn't define one in piana_configuration_parameters.py")
	   
           ##  list_partners = self.piana_secondary_access.get_all_partners(proteinPiana_value=node_id,
##                                                                          use_self_ints= self.use_self_ints,
##                                                                          list_source_dbs= self.list_source_dbs,
##                                                                          inverse_dbs= self.inverse_dbs,
##                                                                          list_source_methods= self.list_source_methods,
##                                                                          inverse_methods= self.inverse_methods,
##                                                                          threshold= self.hub_threshold)
            list_partners = self.piana_secondary_access.get_all_partners_with_interactionPiana(proteinPiana_value=node_id,
                                                                                               use_self_ints= self.use_self_ints,
                                                                                               list_source_dbs= self.list_source_dbs,
                                                                                               inverse_dbs= self.inverse_dbs,
                                                                                               list_source_methods= self.list_source_methods,
                                                                                               inverse_methods= self.inverse_methods,
                                                                                               threshold= self.hub_threshold)
            piana_access_to_use = self.piana_secondary_access


        else:

            if use_ppi_memory == "yes":
                if not self.all_ppis:
                    # Load all protein-protein interactions
                    self._load_all_protein_protein_interactions()
                try:
                    list_partners = self.all_ppis[node_id].items()
                except KeyError:
                    list_partners = []
            
##             list_partners = self.piana_access.get_all_partners(proteinPiana_value=node_id,
##                                                                use_self_ints= self.use_self_ints,
##                                                                list_source_dbs= self.list_source_dbs,
##                                                                inverse_dbs= self.inverse_dbs,
##                                                                list_source_methods= self.list_source_methods,
##                                                                inverse_methods= self.inverse_methods,
##                                                                threshold= self.hub_threshold)

            else:
                #print "getting partners directly from database..."

                list_partners = self.piana_access.get_all_partners_with_interactionPiana(proteinPiana_value=node_id,
                                                                                         use_self_ints= self.use_self_ints,
                                                                                         list_source_dbs= self.list_source_dbs,
                                                                                         inverse_dbs= self.inverse_dbs,
                                                                                         list_source_methods= self.list_source_methods,
                                                                                         inverse_methods= self.inverse_methods,
                                                                                         threshold= self.hub_threshold)

                #print "FOUND: %s: %s" %(len(list_partners),list_partners)
                #print "FOUND: %s" %(len(list_partners))

            if not list_partners:
                # if no partners were found, check if user asked to use a secondary PIANA database
                if self.use_secondary_db == "yes":
                    # if there is a secondary DB to be used, get the partners in that database, 
                    # and set the secondary database as the piana_access to use for those interactions 
                    if not self.piana_secondary_access:
                        raise ValueError("You are asking to use a secondary PIANA database, but you didn't define one in piana_configuration_parameters.py")

##                     list_partners = self.piana_secondary_access.get_all_partners(proteinPiana_value=node_id,
##                                                                                  use_self_ints= self.use_self_ints,
##                                                                                  list_source_dbs= self.list_source_dbs,
##                                                                                  inverse_dbs= self.inverse_dbs,
##                                                                                  list_source_methods= self.list_source_methods,
##                                                                                  inverse_methods= self.inverse_methods,
##                                                                                  threshold= self.hub_threshold)
                    list_partners = self.piana_secondary_access.get_all_partners_with_interactionPiana(proteinPiana_value=node_id,
                                                                                                       use_self_ints= self.use_self_ints,
                                                                                                       list_source_dbs= self.list_source_dbs,
                                                                                                       inverse_dbs= self.inverse_dbs,
                                                                                                       list_source_methods= self.list_source_methods,
                                                                                                       inverse_methods= self.inverse_methods,
                                                                                                       threshold= self.hub_threshold)
                    
                    
                    piana_access_to_use = self.piana_secondary_access
                # END OF if self.use_secondary_db == "yes":
                else:
                    piana_access_to_use = self.piana_access
                # END OF else: (if self.use_secondary_db == "yes":)
            # END OF if not list_partners:
            else:
                # if there were partners for the primary database, set it as the one to use
                piana_access_to_use = self.piana_access
        # END OF else: (if self.force_secondary_db == "yes":)

        if verbose:
            sys.stderr.write("Number of interactions for proteinPiana %s found in database is: %s\n" %(node_id, len(list_interactions)) )


        # For each partner proteinPiana
        #  - find interactionPiana for that interaction
        #  - get attribute for the partner
        #  - get attribute for node_id (done only once before starting the loop: it doesn't change)
        #  - get attributes for interaction_id
        #  - create a  LinkDetail object
        #  - add all attributes previously obtained to LinkDetail
        #  - add the LinkDetail object to the list of LinkDetail objects

        if self.graph.get_memory_mode() == "memory":
            node_attr = "onDemand"
        else:
            node_attr = "inFlight"
        attribute_node_id = PianaGraphNodeAttribute(proteinPiana_value= node_id, piana_access = piana_access_to_use, mem_mode = node_attr)


        # for each interacion where node_id is involved, append on LinkDetail to the list of LinkDetail objects
        for partner_id in list_partners:

##             interaction_id = piana_access_to_use.get_interactionPiana(proteinPianaA_value= partner_id,
##                                                                       proteinPianaB_value= node_id,
##                                                                       list_source_dbs= "all", inverse_dbs="no",
##                                                                       list_source_methods= "all", inverse_methods="no") # no need to place restrictions
##                                                                                                                         # because they were placed before
##                                                                                                                         # when retrieving the partners

            interaction_id = partner_id[1]
            
            if verbose:
                sys.stderr.write( "----------------------------------------------------------------------\n")
                sys.stderr.write( "For interaction_id = %s the partner is %s\n" %(interaction_id, partner_id))
                sys.stderr.write( "----------------------------------------------------------------------\n")

            attribute_partner_id = PianaGraphNodeAttribute(partner_id[0], piana_access= self.piana_access, mem_mode = node_attr)  # the piana_access for the nodes is always
                                                                                                                               # the primary one. The DBs are synchronized
                                                                                                                               # so... no need to have secondary access for 
                                                                                                                               # the nodes
    
            attribute_interaction_id = PianaGraphEdgeAttribute(interaction_id, piana_access= piana_access_to_use, mem_mode = node_attr) # the piana_access that will be used
                                                                                                                                        # is the one that was used to get
                                                                                                                                        # the interaction, and the edge will
                                                                                                                                        # know where it comes from so it can
                                                                                                                                        # query the appropiate DB



            # link_detail_object is used to pass the attributes of two nodes and the edge connecting them
            link_detail_object = GraphBuilder.LinkDetail(node1_id= node_id,
                                                         node2_id= partner_id[0],
                                                         edge_id= interaction_id )

            # adding attribute of first node
            link_detail_object.set_attribute(attribute_object= attribute_node_id, node_id= node_id)
            
            # adding attribute of second node
            link_detail_object.set_attribute(attribute_object= attribute_partner_id,  node_id= partner_id[0])
            
            # adding attribute of edge
            link_detail_object.set_attribute(attribute_object= attribute_interaction_id)
            
            list_link_detail_objects.append(link_detail_object)

        # END OF for interaction_id in list_interactions:

        if verbose:
            sys.stderr.write( "returning the list of link detail objects\n")
               

        return list_link_detail_objects



    def empty_edges_buffer(self):

        self.all_ppis={}


    def get_link_details_two_nodes(self, node_id1, node_id2, use_ppi_memory="no"):
        """
        Overrides method from GraphBuilder: this method return a LinkDetail object between nodes "node_id1" and "node_id2"
        
        "use_ppi_memory" if this method has to be used repeatedly, it is much faster to store all protein-protein interactions in memory
                         and check them here than to find them in database each time
                         "yes" load all interactions in memory (only first time) and then use them
                         "no" use database connection    
        """

        # Check for correct order of node_ids
        if node_id1 > node_id2:
            node_id2,node_id1 = node_id1,node_id2

        if self.graph.get_memory_mode()=="speed":
            attr_mem_mode = "inFlight"
        else:
            attr_mem_mode = "onDemand"

        # As this method is used for interactions in the last level... (almost now), it is not necessary to create new attribute_node_id!!!
        #attribute_node_id1 = PianaGraphNodeAttribute(proteinPiana_value= node_id1, piana_access = self.piana_access, mem_mode = "inMemory")
        #attribute_node_id2 = PianaGraphNodeAttribute(proteinPiana_value= node_id2, piana_access = self.piana_access, mem_mode = "inMemory")

        interaction_id = None
        if use_ppi_memory == "yes":
            if not self.all_ppis:
                # Load all protein-protein interactions
                self._load_all_protein_protein_interactions()
            try:
                interaction_id = self.all_ppis[node_id1][node_id2]
            except KeyError:
                interaction_id = None

        else:
            #print "getting interaction (%s,%s) directly from database..." %(node_id1,node_id2)
            interaction_id = self.piana_access.get_interactionPiana(proteinPianaA_value = node_id1,
                                                                    proteinPianaB_value = node_id2,
                                                                    list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no")

        if interaction_id:
            attribute_interaction_id = PianaGraphEdgeAttribute(interaction_id, piana_access= self.piana_access, mem_mode = attr_mem_mode) # the piana_access that will be used
                                                                                                                                       # is the one that was used to get
                                                                                                                                       # the interaction, and the edge will
                                                                                                                                       # know where it comes from so it can
                                                                                                                                       # query the appropiate DB

            # link_detail_object is used to pass the attributes of two nodes and the edge connecting them
            link_detail_object = GraphBuilder.LinkDetail(node1_id= node_id1,
                                                         node2_id= node_id2,
                                                         edge_id= interaction_id )

            # adding attribute of first node
            #link_detail_object.set_attribute(attribute_object= attribute_node_id1, node_id= node_id1)
            
            # adding attribute of second node
            #link_detail_object.set_attribute(attribute_object= attribute_node_id2,  node_id= node_id2)
            
            # adding attribute of edge
            link_detail_object.set_attribute(attribute_object= attribute_interaction_id)

            return link_detail_object
       
        else:
            return None


    def build_graph(self,complete_taxID=None):
        """
        This is the specific building method for PianaGraphBuilder. If the graph is generated from a list of seed (root) nodes
        it calls the generic build_graph method. Otherwise, if complete specie is selected, it will construct the complete network
        for a specific specie

        "complete_taxID" has only to be introduced when a complete network for a specie must be builded
        
        """

        if self.graph.get_memory_mode() == "speed":
            attr_mem_mode = "inFlight"
        else:
            attr_mem_mode = "onDemand"

        if complete_taxID is None:
            GraphBuilder.build_graph(self)
        else:
            # First, obtain the list of all interactions for this specie
            all_interactions = self.get_all_interactions_for_taxID(taxID=complete_taxID)

            # Then, obtain all the proteins for this specie (to make sure all proteins will appear)
            proteins_with_given_species = self.piana_access.get_proteins_sharing_species(taxonomy_value=complete_taxID)

            # Insert all the nodes to the graph
            for actual_protein in proteins_with_given_species:
                node_attribute = self.create_node_attribute(node_id= actual_protein)
                node_object = self.graph.get_node(identifier= actual_protein,
                                                  attribute= node_attribute,
                                                  #attribute = None,
                                                  get_mode="new")
                node_object.set_as_root()
                node_object.set_level(0)
                self.graph.add_node(node_object= node_object)


            # Insert all the interactions to the graph
            number_inserted=0
            for actual_interaction in all_interactions:
                if number_inserted%1000 == 0:
                    print "%s inserted" %(number_inserted)
                interaction_attribute = PianaGraphEdgeAttribute(interactionPiana_value=actual_interaction[2],piana_access=self.piana_access,mem_mode=attr_mem_mode)
                edge_object = self.graph.get_edge(identifier1=actual_interaction[0],
                                                  identifier2=actual_interaction[1],
                                                  attribute_object= interaction_attribute,
                                                  #attribute_construction_values=(actual_interaction[2],piana_access)
                                                  #attribute_object=None,
                                                  get_mode="new")
                self.graph.add_edge(edge_object=edge_object)
                number_inserted += 1
                

    def get_all_interactions_for_taxID(self, taxID, mem_mode="onDemand"):
        """

        returns a list of all protein-protein interactions (proteinPiana,proteinPianaB,interactionPiana) for "taxID" and restrictions
        """


        list_link_detail_objects = []

        if verbose:
            sys.stderr.write("in PianaGraphBuilder: list source dbs is %s and list source methods is %s\n" %(self.list_source_dbs, self.list_source_methods))


        ## First, get all protein-protein interactions that satisfy the conditions
        # The behavior for retrieving the interactions is the following:
        #
        #    -> if force_secondary_db is yes, then use the secondary database (and that's all... nothing else to check)
        #           -> it means the user only wants to use the secondary database to retrieve interactions
        #    -> if force_secondary is no, then there are several possibilities:
        #         1. the user has set use_secondary_db to no, then get partners from primary
        #           -> he doesn't want to use a secondary DB even in cases where no interactions are found in the primary DB
        #         1. the user has set use_secondary_db to yes, then get partners from primary
        #                1.1 if partners are found, keep those and forget about the secondary database
        #                1.2 if no partners are found, look for partners in the secondary database
        if self.force_secondary_db == "yes":

            if not self.piana_secondary_access:
                raise ValueError("You are forcing to use a secondary PIANA database, but you didn't define one in piana_configuration_parameters.py")


            all_interactions = self.piana_secondary_access.get_all_protein_protein_interactions(use_self_ints=self.use_self_ints,
                                                                                                list_source_dbs=self.list_source_dbs,
                                                                                                inverse_dbs=self.inverse_dbs,
                                                                                                list_source_methods=self.list_source_methods,
                                                                                                inverse_methods=self.inverse_methods,
                                                                                                tax_id_value=taxID)
	   
            piana_access_to_use = self.piana_secondary_access

        else:

            all_interactions = self.piana_access.get_all_protein_protein_interactions(use_self_ints=self.use_self_ints,
                                                                                      list_source_dbs=self.list_source_dbs,
                                                                                      inverse_dbs=self.inverse_dbs,
                                                                                      list_source_methods=self.list_source_methods,
                                                                                      inverse_methods=self.inverse_methods,
                                                                                      tax_id_value=taxID)


            if not all_interactions:
                # if no partners were found, check if user asked to use a secondary PIANA database
                if self.use_secondary_db == "yes":
                    # if there is a secondary DB to be used, get the partners in that database, 
                    # and set the secondary database as the piana_access to use for those interactions 
                    if not self.piana_secondary_access:
                        raise ValueError("You are asking to use a secondary PIANA database, but you didn't define one in piana_configuration_parameters.py")


                    all_interactions = self.piana_secondary_access.get_all_protein_protein_interactions(use_self_ints=self.use_self_ints,
                                                                                                        list_source_dbs=self.list_source_dbs,
                                                                                                        inverse_dbs=self.inverse_dbs,
                                                                                                        list_source_methods=self.list_source_methods,
                                                                                                        inverse_methods=self.inverse_methods,
                                                                                                        tax_id_value=taxID)
                    piana_access_to_use = self.piana_secondary_access
                # END OF if self.use_secondary_db == "yes":
                else:
                    piana_access_to_use = self.piana_access
                # END OF else: (if self.use_secondary_db == "yes":)
            # END OF if not list_partners:
            else:
                # if there were partners for the primary database, set it as the one to use
                piana_access_to_use = self.piana_access
        # END OF else: (if self.force_secondary_db == "yes":)

        if verbose:
            sys.stderr.write("Number of interactions found is: %s\n" %(len(all_interactions)) )

        return all_interactions

