"""
 File       : PianaGraphEdgeAttribute.py
 Author     : R. Aragues & D. Jaeggi
 Creation   : 2003
 Contents   : class PianaGraphEdgeAttribute (attribute specific to Piana for an Edge)
 Called from: 

=======================================================================================================

This file implements class PianaGraphEdgeAttribute

It describes an interaction of the protein-protein interaction network


"""

# PianaGraphEdgeAttribute.py: class PianaGraphEdgeAttribute (attribute specific to Piana for an Edge)
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import copy

from GraphEdgeAttribute import *
from PianaDB import *
from PianaGraphEdgeTypes import *
from sets import *


verbose = 0
# --------------------------------
# Class PianaGraphEdgeAttribute
# --------------------------------   
    
class PianaGraphEdgeAttribute(GraphEdgeAttribute):
    """

    Class that implements attributes for Piana Edges: accessing edge values, setting edge values, ...
    
    """
    def __init__(self, interactionPiana_value= None, mem_mode = "onDemand", piana_access = None,
                 confidence_interaction_value= 0, propagated = None, extended = None, construction_parameters=()):

        """
        "construction_parameters" is the tuple with the minimum necessary data to generate the PianaGraphEdgeAttribute.
                                  This is used when the edges and EdgeAttributes are not stored in memory.
                                  The content of the tuple is the generated with the GraphEdgeAttribute method get_creation_information
                                  It has prevalence over all the rest of parameters
        
        "interactionPiana_value" is the database identifier for this edge

        "piana_access" is a PianaDBaccess object, used to retrieve information from the database
       
        "mem_mode" can be:
        --> "onDemand": values are retrieved from DB only when they are needed and stored then in memory
        --> "inMemory": values are retrieved from DB at initilisation and stored in memory
        --> "inFlight": values are retrieved from DB when needed and never are stored in memory

  
        "propagated" is an object of class PropagatedEdge
        
        "extended" is an object of class ExtendedEdge

        Attention!!! Right now, not using extended for anything: I do not add edges to the graph by extension (extension meant that
        edges from the database where added for nodes that were added to network during an expansion.

              --> There will just be propagated edges

        Attention!!! Currently, not using "confidence_interaction_value"
 
        """



        self.db_access = piana_access   # setting the connection to the piana DB
        self.mem_mode = mem_mode

        
        #self.dic_associated_interactionPiana = {}  # interactionPiana associated to this interaction (when unifying)
        self.set_associated_interactionPiana = Set()

        
        if interactionPiana_value is not None:
            self.interactionPiana = int(interactionPiana_value) # interactionPiana will be None if edge only comes from a propagation
        else:
            self.interactionPiana = None

        self.sourceDB_id = None # this will hold the integer that identifies which sourceDB_names have this interaction
                                #   -> this is the pianaDB way of identifying which sourceDBs are associated to an interaction
                                #      Here, we use this integer because it takes less memory than using the list

        #javi:
        if mem_mode=="inFlight":
            pass
        else:
            self.dic_methodID = {}   # this will hold (as keys) the methodIDs that are associated to this interaction
            self.dic_pubmedID = {}   # this will hold (as keys) the pubmed IDs that are associated to this interaction

        #if mem_mode=="inMemory":
        #    sys.stderr.write("Creating a PianaGraphEdgeAttribute Object with \"inMemory\"? How is it possible?")
        
        #if mem_mode=="onDemand":
        #    pass
        #elif mem_mode=="inMemory":
        #    self._load_edge_features()
        #elif mem_mode!=
        #    raise ValueError(""" mem_mode set to incorrect value """)
        if mem_mode=="inMemory":
            self._load_edge_features()

            
        self.confidenceInteraction = confidence_interaction_value

        # propagated is a list of objects containing interactions of another node are added to another node
        # because both nodes share a certain characteristic
        self.propagated = []
        if propagated is not None:
            self.set_propagation(propagated_edge_object= propagated)  # will contain objects PropagatedEdge
                              
        # extended is a list of objects containing interactions of a node that has been added to the graph as a result
        # of an expansion
        self.extended = []
        if extended is not None:
            self.set_extension(extended_edge_object= extended)  # will contain objects ExtendedEdge


        # Last, we get the values from the tuple construction_parameters. So, they will overwritte the other parameters
        if construction_parameters:
            self.interactionPiana = construction_parameters[0]
            self.db_access = construction_parameters[1]


    def __deepcopy__(self, memo):

        clone = copy.copy(self) # Make a shallow copy
        for name, value in vars(self).iteritems():
            if name != "db_access":
                setattr(clone, name, copy.deepcopy(value, memo))
            else:
                setattr(clone, name, value)

        return clone
            
    def __str__(self):
        """
        implementation of conversion to string for PianaGraphEdgeAttribute
        """

        string = "PianaGraphEdgeAttribute consisting of <interactionPiana=%s, confidenceInteraction=%s, sourceDB_names=%s, methodIDs=%s, associated_interactionPiana=%s, extended=%s, propagated=%s> \
                  " %(self.interactionPiana, self.confidenceInteraction,
                      self.get_list_sourceDB_names(),
                      self.get_list_methodID(),
                      self.set_associated_interactionPiana,
                      self.extended, self.propagated)
        
        return string

    # --------------------------------
    # PianaGraphEdgeAttribute "set" methods
    # --------------------------------
    #
    # Methods to set values of PianaGraphEdgeAttributes
    #
    # 

    #def _set_confidenceInteraction(self):
        
    #     self.confidenceInteraction = self.db.get_db_content(PianaGlobals.ConfidenceInteraction, self.interactionPiana, answer_mode="single")


        
#    def _get_sourceDB_methodID_key(self, sourceDBID=None, methodID=None):
#        """
#        returns a key for the pair sourceDB methodID
#        """
#        return sourceDBID + "." + methodID

    def _load_edge_features(self):
        """
        checks is the edge already has its features (i.e. sourceDB_id and methodID) and in case they were loaded, the
        corresponding object attributes are set
        """
        
        if not self.sourceDB_id:
            self._set_sourceDB_id()

        if not self.dic_methodID:
            self._set_dic_methodID()

        if not self.dic_pubmedID:
            self._set_dic_pubmedID()
    
    def _set_dic_methodID(self):
        """
        Updates by quering the database the dic_methodID
        """
        if self.interactionPiana is not None:
            # this edge comes from a database interaction: get methodID for it
            self.dic_methodID= dict(map(None, self.db_access.get_interaction_methodID_list(interactionPiana_value=self.interactionPiana), []))
        else:
            # this edge does not come from a database interaction: therefore, it has no methodID
            self.dic_methodID= {}

    # Javi:
    def _set_dic_pubmedID(self):
        """
        Updates by quering the database the dic_pubmedID
        """
        if self.interactionPiana is not None:
            # this edge comes from a database interaction: get pubmedID for it
            self.dic_pubmedID= dict(map(None, self.db_access.get_interaction_pubmedID_list(interactionPiana_value=self.interactionPiana), []))
        else:
            # this edge does not come from a database interaction: therefore, it has no pubmedID
            self.dic_methodID= {}
        
    def _set_sourceDB_id(self):
        """
        Updates by quering the database the sourceDB_id
        
        """
        if self.interactionPiana is not None:
            # this edge comes from a database interaction: get sourceDB_id  for it
            self.sourceDB_id = self.db_access.get_sourceDB_id_from_interactionPiana(interactionPiana_value=self.interactionPiana)
            if self.set_associated_interactionPiana is not None:
                for actual_interactionPiana in self.set_associated_interactionPiana:
                    self.sourceDB_id = self.sourceDB_id | self.db_access.get_sourceDB_id_from_interactionPiana(interactionPiana_value=actual_interactionPiana)
        else:
            # this edge does not come from a database interaction: therefore, it has no sourceDB or methodID
            #self.sourceDB_id = None
            self.sourceDB_id = 0


    #javi:
    def _set_associated_interactionPiana(self, list_interactionPiana):
        """
        Updates the associated interactionPiana with those in list_interactionPiana
        """

        self.set_associated_interactionPiana.update(list_interactionPiana)
        #(dict(map(None,list_interactionPiana,[])))

        
   
    # PianaGraphEdgeAttribute "get" methods
    # --------------------------------
    #
    # Methods to get values from PianaGraphEdgeAttributes
    #
    # If the value requested is not already in memory (i.e. mem_mode is onDemand), it is retrieved from the database and returned
    

    def get_db_edge_id(self):
        """
        Overrides method get_db_edge_id() of superclass GraphEdgeAttribute
        
        returns the interactionPiana identifier for this interaction

        It is the same as get_interactionPiana, but in this case it is overridden a method that all XXGraphEdgeAttribute must have
        This is the method that is called from Graph, since Graph doesn't know which kind of EdgeAttribute it is processing
        
        """
        
        return self.interactionPiana    

    def get_interactionPiana(self):
        """
        returns the interactionPiana identifier for this interaction
        """
        
        return self.interactionPiana    

    def get_sourceDB_id(self):
        """
        returns the integer that identifies in which sourceDBs this interaction was
        """
        if not self.sourceDB_id:
            self._set_sourceDB_id()
            
        return self.sourceDB_id
 
    def get_list_sourceDB_names(self):
        """
        returns a list with source databases for this interaction
        """
        if not self.sourceDB_id:
            self._set_sourceDB_id()

        return  self.db_access.get_list_sourceDB_names_from_sourceDB_id(sourceDB_id= self.sourceDB_id)
 
    def get_dic_methodID(self):
        """
        returns a list with methods for this interaction
        """
        if self.mem_mode!="inFlight":
            if not self.dic_methodID:
                self._set_dic_methodID()
      
            return self.dic_methodID
        else:
            print "getting method ID dictionary..."
            return_dict = dict(map(None,self.db_access.get_interaction_methodID_list(interactionPiana_value=self.interactionPiana),[]))
            for actual_interactionPiana in self.set_associated_interactionPiana:
                return_dict.update(dict(map(None, self.db_access.get_interaction_methodID_list(interactionPiana_value=actual_interactionPiana), [])))

            return return_dict


    def get_dic_pubmedID(self):
        """
        returns a dict with pubmeds for this interaction
        """
        if self.mem_mode!="inFlight":
            if not self.dic_pubmedID:
                self._set_dic_pubmedID()
      
            return self.dic_pubmedID
        else:
            return_dict = dict(map(None,self.db_access.get_interaction_pubmedID_list(interactionPiana_value=self.interactionPiana), []))
            for actual_interactionPiana in self.set_associated_interactionPiana:
                return_dict.update(dict(map(None,self.db_access.get_interaction_pubmedID_list(interactionPiana_value=actual_interactionPiana) , [])))

            return return_dict
        


    def get_list_pubmedID(self):
        """
        returns a list with methods for this interaction
        """
        if self.mem_mode!="inFlight":
            if not self.dic_pubmedID:
                self._set_dic_pubmedID()
      
            return self.dic_pubmedID.keys()
        else:
            return_dict = dict(map(None,self.db_access.get_interaction_pubmedID_list(interactionPiana_value=self.interactionPiana), []))
            for actual_interactionPiana in self.set_associated_interactionPiana:
                return_dict.update(dict(map(None,self.db_access.get_interaction_pubmedID_list(interactionPiana_value=actual_interactionPiana) , [])))

            return return_dict.keys()
        
    
    def get_list_methodID(self):
        """
        returns a list with methods for this interaction
        """
        if self.mem_mode != "inFlight":
            if not self.dic_methodID:
                self._set_dic_methodID()
      
            return self.dic_methodID.keys()
        else:
            
            return_dict = dict(map(None,self.db_access.get_interaction_methodID_list(interactionPiana_value=self.interactionPiana),[]))

            for actual_interactionPiana in self.set_associated_interactionPiana:
                return_dict.update(dict(map(None, self.db_access.get_interaction_methodID_list(interactionPiana_value=actual_interactionPiana), [])))

            return return_dict.keys()
        

    def merge_attribute(self, new_edge_attribute, ignore_ids= 0):
        """
        merges the attribute of the object with the new edge attribute passed as argument


        "ignore_ids" is used to avoid conflicts when unifying the network in PianaGraph
            -> set to 1 when unifying the network: it won't check that interactionPiana is the same in both attributes
        """

        if not isinstance(new_edge_attribute, PianaGraphEdgeAttribute):
            raise TypeError("attribute to be merged must be PianaGraphEdgeAttribute")

        #javi:
        if self.mem_mode!="inFlight":
            self._load_edge_features() # make sure the features of this attribute are loaded in the attribute

        new_edge_interactionPiana = new_edge_attribute.get_interactionPiana()
        
        if not ignore_ids:
            # 1. merging interactionPiana values
            #new_edge_interactionPiana = new_edge_attribute.get_interactionPiana()

            if self.interactionPiana is not None:
                if new_edge_interactionPiana is not None:
                    # if both interactionPiana are not None, make sure we are dealing with the same interaction. It wouldn't make sense
                    # to merge two attributes which dot not describe the same interaction.
                    if self.interactionPiana != new_edge_interactionPiana:
                        raise ValueError("Merge attribute of two different interactionPiana (%s and %s) interactions" %(self.interactionPiana,
                                                                                                                        new_edge_interactionPiana ))
                    # END OF if self.interactionPiana != new_edge_interactionPiana:

                    # if new_edge_interactionPiana is None, then the merging of the attributes will keep the one in self.interactionPiana

                # END OF if new_edge_interactionPiana is not None: ... 
            else:
                self.interactionPiana = new_edge_interactionPiana
        # END OF if not ignore_ids:

        # 2. merging sourceDB_id and list_methodID values
        new_edge_sourceDB_id = new_edge_attribute.get_sourceDB_id()

        self.sourceDB_id = self.get_sourceDB_id() | new_edge_sourceDB_id   # binary OR updates to which databases this edge belongs to

        if self.mem_mode!="inFlight":
            new_edge_dic_methodID = new_edge_attribute.get_dic_methodID()
            self.dic_methodID.update(new_edge_dic_methodID)

            # 2.1 Merging pubmed_id values (javi):
            new_edge_dic_pubmedID = new_edge_attribute.get_dic_pubmedID()
            self.dic_pubmedID.update(new_edge_dic_pubmedID)
        
        #list_sourceDB_methodID = new_edge_attribute.get_sourceDB_methodID()

        #for pair_sourceDBID_methodID in list_sourceDB_methodID:
        #    self.add_sourceDB_methodID(sourceDBID=pair_sourceDBID_methodID[0], methodID=pair_sourceDBID_methodID[1])

        # Merging associated interactionPianas:
        self.get_associated_interactionPiana().update(new_edge_attribute.get_associated_interactionPiana())
        self.get_associated_interactionPiana().update([new_edge_interactionPiana])
        
        # 3. merging propagated
        for propagation in new_edge_attribute.get_propagations():
            self.set_propagation(propagated_edge_object= propagation)

        # 3. merging extended
        for extension in new_edge_attribute.get_extensions():
            self.set_extension(extended_edge_object= extension)
        
        
    def distance(self):
        """
        To  do!!! Get distance that makes biological sense

        Returns a distance like value.

        Defaults to return 1
        """
        return 1.0
    # --------------------------------
    # PianaGraphEdgeAttribute  methods to handle origin of the edge
    # --------------------------------
    #
    # Methods to handle the origin of the edge
    #
    # And edge can be original, extended and propagated. It can be the three at the same time. It has to be at least one of them
    #
    # 
 
    def set_propagation(self, expansion_type= None, propagated_from_node= None, propagated_from_db_edge_id= None, propagated_from_edge_id= None,
                        propagated_edge_object=None):
        """

        Adds to the edge an object describing a propagation (interactions of another node are added to another node because both nodes share
                                                             a certain characteristic)

        if propagated_edge_object is not None, adds that object (avoiding adding an object that already exists)
        if not, creates a new object using expansion_type, propagated_from_node, propagated_from_db_edge_id and propagated_from_edge_id

        The propagation is described in the object propagated_edge_object of class PianaNodeExpansion PropagatedEdge
        """

        propagated_edge_object = PropagatedEdge(expansion_type=expansion_type,
                                                propagated_from_node= propagated_from_node,
                                                propagated_from_db_edge_id= propagated_from_db_edge_id, 
                                                propagated_from_edge_id= propagated_from_edge_id )

        if propagated_edge_object not in self.propagated:
            self.propagated.append(propagated_edge_object)


    def get_associated_interactionPiana(self):
        return self.set_associated_interactionPiana
  
    def get_propagations(self):
        """

        returns [] if it is not a edge coming from an propagation

        otherwise returns a list of objects PropagatedEdge

        """
         
        return self.propagated

   
    def set_extension(self, expansion_type=None, extended_from_node= None, extended_edge_object= None):
        """

        Adds to the edge an object describing an extension ( database interactions added to the graph because one of the nodes involved
                                                             in the interaction was added to the graph as a result of an expansion)


        if extended_edge_object is not None, adds that object (avoiding adding an object that already exists)
        if not, creates a new object using expansion_type and extended_from_node

        edge_extended_object is of class ExtendedEdge

        """
        if extended_edge_object is None:
            extended_edge_object = ExtendedEdge(expansion_type=expansion_type, extended_from_node= extended_from_node)
        
        if extended_edge_object not in self.extended:
            self.extended.append(extended_edge_object)
  
    def get_extensions(self):
        """

        returns [] if it is not an edge coming from an extension

        otherwise returns a list of objects ExtendedEdge

        """
         
        return self.extended

    def get_creation_information(self):
        """
        returns the tuple (interaction_id, piana_access)
        """

        return (self.get_interactionPiana(),self.db_access)

 
   #def get_confidenceInteraction(self):
   #
   #    if self.confidenceInteraction is None:
   #        
   #       self._setconfidenceInteraction()
   #       
   #    return self.confidenceInteraction
   
##     def get_sourceDB_methodID(self):
##         """
##         returns sourceDB_methodID attribute, which is a list of tuples [sourceDB, methodID]
##         """
##         if not self.sourceDB_methodID:
##             self._set_sourceDB_methodID()

##         return self.sourceDB_methodID

   
##     def _update_record(self):
##         """
##         updates lists with methods and source dbs
##         """
##         if not self.sourceDB_id:
##             self._set_sourceDB_id()

##         if not self.dic_methodID:
##             self._set_dic_methodID()

##         self.list_sourceDB = []
##         self.list_methodID = []
        
##         for sourceDB_methodID in self.sourceDB_methodID:
##             self.list_sourceDB.append(sourceDB_methodID[0])
##             self.list_methodID.append(sourceDB_methodID[1])
            
##         self.has_changed = 0

   
##     def add_sourceDB_methodID(self, sourceDBID =None, methodID= None):
##         """
##         adds (checking that there is no redundancy) the sourceDB and methodID to the current edge (used for merging edges)
##         """
##         key_to_use= self._get_sourceDB_methodID_key(sourceDBID= sourceDBID, methodID=methodID )

##         if not self.keys_sourceDB_methodID.has_key(key_to_use):
##             self.sourceDB_methodID.append([sourceDBID, methodID ])
##             self.keys_sourceDB_methodID[key_to_use] = None
##             self.has_changed = 1
