"""
 File       : PianaGraphNodeAttribute.py
 Author     : R. Aragues & D. Jaeggi
 Creation   : 2003
 Contents   : class PianaGraphNodeAttribute (attribute specific to Piana for a Node)
 Called from: 

=======================================================================================================

This file implements class PianaGraphNodeAttribute, describing the attribute of a GraphNode, specific to PianaGraph 

"""

# PianaGraphNodeAttribute.py: class PianaGraphNodeAttribute (attribute specific to Piana for a Node)
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues
import copy
from sets import *

from GraphNodeAttribute import *
from PianaDB import *

verbose = 0

# --------------------------------
# Class PianaGraphNodeAttribute
# --------------------------------
class PianaGraphNodeAttribute(GraphNodeAttribute):
    
    """
    Class that implements attributes for Piana Nodes: accessing node values, setting values, ...
    
    """
    def __init__(self, proteinPiana_value,  piana_access = None, mem_mode = "onDemand"):
        
        """
        "proteinPiana_value" is the proteinPiana of this node

        "piana_access" is a PianaDBaccess object used to access the database
        
        "mem_mode" can be:
        --> onDemand: values are retrieved from DB only when they are needed, and then are stored in memory all remaining time
        --> inMemory: values are retrieved from DB at initilisation and stored in memory all time
        --> inFlight: values are retrieved from DB when are needed, but never are stored in memory

        
        """

        self.db_access = piana_access    # setting the connection to the piana DB through class PianaDB

        # features of the node. All information about the characteristics particular to this type of node should be contained here
        #   -> characteristics typical of all nodes (root, expanded, ...) are kept at the GraphNode level

        self.proteinPiana = int(proteinPiana_value)
        self.molecularWeight = None 
        self.isoelectricPoint = None

        self.mem_mode = mem_mode


        # Attention: the following variables will only be used when 'unifiying' nodes!
        #            these variables are not used when building PianaGraph networks, only when printing results
        #            to fuse into a single node attribute all the information associated to proteinPianas that have
        #            the same external code
        self.set_associated_node_ids = Set([self.proteinPiana])   # Set of the proteinPianas of associated node obtained in unification

        # This attributes will be stored in memory only if mem_mode is not "inFlight"
        if self.mem_mode!="inFlight":
            self.dic_functions = None
            self.dic_descriptions = None
            self.dic_taxonomies = None
            self.dic_cell_location = None

        # This attributes are not influenced by memory usage:
        self.expression = None
	self.dic_special_labels = None
        self.dic_keywords_appearing = None

        self.union_neighbours = Set()


        if mem_mode == "onDemand":
            pass
            
        elif mem_mode == "inMemory":
            self._set_molecular_weight() 
            self._set_isoelectric_point() 
            
        elif mem_mode != "inFlight":
            raise ValueError(""" mem_mode set to incorrect value  %s(possible values are "onDemand" and "inMemory" """ %(mem_mode))

      
    def __deepcopy__(self, memo):
        
        clone = copy.copy(self) # Make a shallow copy
        for name, value in vars(self).iteritems():
            if name != "db_access":
                setattr(clone, name, copy.deepcopy(value, memo))
            else:
                setattr(clone, name, value)

        return clone
  
    
    def __getstate__(self):

        odict = self.__dict__.copy() # copy the dict since we change it
        return odict

 
    def __setstate__(self, dict):

        self.__dict__ = dict
        dict.__class__.__init__(dict)
        
    def __getnewargs__(self):

        return (self.proteinPiana,  self.db_access)

    # --------------------------------
    # PianaGraphNodeAttribute "set" methods
    # --------------------------------
    #
    # Methods to set values of PianaGraphNodeAttributes
    #
    #
    
    def merge_attribute(self, attribute_object, ignore_ids=0):
        """
        merges existing attribute with atribute_object passed as argument

        certain characteristics of the node will only be merged in ignore_ids is 1,
        since we are interested in having those characteristics filled
        only when we are unifying the nodes (ie. the operation by which PIANA
        decides which name will it use for the protein)  
        """
        if not ignore_ids:
            # 'ignore_ids' is used to avoid conflicts when unifying the network in PianaGraph, because we want to merge nodes with different proteinPianas
            if self.get_proteinPiana() != attribute_object.get_proteinPiana():
                raise ValueError("Trying to merge two node attributes with different proteinPiana: current is %s (%s) and merging is %s (%s)" %(
                    self.get_proteinPiana(),
                    type(self.get_proteinPiana()),
                    attribute_object.get_proteinPiana(),
                    type(attribute_object.get_proteinPiana()))  )

        else:
            # this is only executed when unifying a node

            # merge the list of proteinPianas associated to the unified node
            #self.get_associated_proteinPianas().update(attribute_object.get_associated_proteinPianas())
            #self.get_associated_nodes().update(attribute_object.get_associated_nodes())
            #self.get_associated_nodes().extend(attribute_object.get_associated_nodes())
            self.get_associated_node_ids_set().update(attribute_object.get_associated_node_ids_set())

            if self.mem_mode!="inFlight":
                # merge the list of descriptions
                self.get_dic_descriptions().update(attribute_object.get_dic_descriptions())
                
                # merge the list of functions
                self.get_dic_functions().update(attribute_object.get_dic_functions())
                
                # merge the list of taxonomies
                self.get_dic_taxonomies().update(attribute_object.get_dic_taxonomies())

                # merge the list of cell locations
                self.get_dic_cell_location().update(attribute_object.get_dic_cell_location())
            
            # merge the expressions
            if self.expression == attribute_object.expression or attribute_object.expression is None:
                pass
            
            elif self.expression is None:
                self.expression = attribute_object.expression
                
            else:
                self.expression = PianaGlobals.over_and_infra_expressed_protein
                if verbose:
                    sys.stderr.write("Weird... different expressions for nodes with the same ext_code: %s and %s\n" %(self.expression,
                                                                                                                      attribute_object.expression))

	    # merge the special labels
            self.get_dic_special_labels().update(attribute_object.get_dic_special_labels())
            
	    # merge keywords
            self.get_dic_keywords_appearing().update(attribute_object.get_dic_keywords_appearing())

            self.union_neighbours.update(attribute_object.union_neighbours)
        # END OF else: (if not ignore_ids:)


    def _set_union_neighbours(self, neighbours_set=Set()):
                                                                                
        self.union_neighbours.update(neighbours_set)

    
    def _set_molecular_weight(self):
        
        self.molecularWeight = self.db_access.get_protein_mw(proteinPiana= self.proteinPiana)
    
    def _set_isoelectric_point(self):
        
        self.isoelectricPoint =  self.db_access.get_protein_ip(proteinPiana= self.proteinPiana)

    #def _set_associated_proteinPianas(self, proteinPiana= None):
    def _set_associated_node_ids(self, proteinPiana = None):
        self.set_associated_node_ids.add(proteinPiana)
        #self.associated_proteinPianas[proteinPiana] = None

    def _set_dic_taxonomies(self):

        if self.mem_mode=="inFlight":
            sys.stderr.write("Trying to store in memory taxonomy attribute using inFlight mem mode. It is not possible!")
        else:
            list_taxonomies =  self.db_access.get_protein_taxonomy_ids(proteinPiana_value = self.proteinPiana )
            self.dic_taxonomies = dict(map(None, list_taxonomies, []))
            
    def _set_dic_cell_location(self):
        
        if self.mem_mode=="inFlight":
            sys.stderr.write("Trying to store in memory cell location attribute using inFlight mem mode. It is not possible!")
        else:
            list_cell_locations =  self.db_access.get_protein_subcellularLocation(proteinPiana_value = self.proteinPiana)
            self.dic_cell_location = dict(map(None, list_cell_locations, []))
            
    def _set_expression(self, expression = None):
        self.expression = expression

    def _set_special_labels(self, list_special_labels = None):
        
        self.dic_special_labels = dict(map(None, list_special_labels, []))
        
    def _set_keywords_appearing(self, user_keywords = []):
        # changed because it was repeating the query for description, function and keyword)
        #list_keywords_appearing = self.db_access.check_keywords_in_protein(list_proteinPiana=[self.proteinPiana], keywords=user_keywords)
        search_in = self.get_descriptions()
        search_in.extend(self.get_functions())
        search_in.extend(self.db_access.get_protein_keyword(proteinPiana_value=self.get_proteinPiana()))
        
        list_keywords_appearing = self.db_access.check_keywords_in_protein(list_proteinPiana=[self.proteinPiana], keywords=user_keywords, search_in = search_in)

        
        self.dic_keywords_appearing = dict(map(None, list_keywords_appearing, []))
        
    def _set_dic_descriptions(self):
        
        if self.mem_mode=="inFlight":
            sys.stderr.write("Trying to store in memory description attribute using inFlight mem mode. It is not possible!")
        else:
            list_descriptions =  self.db_access.get_protein_description(proteinPiana_value = self.proteinPiana)
            self.dic_descriptions = dict(map(None, list_descriptions, []))
            
    def _set_dic_functions(self):

        if self.mem_mode=="inFlight":
            sys.stderr.write("Trying to store in memory function attribute using inFlight mem mode. It is not possible!")
        else:
            list_functions =  self.db_access.get_protein_function(proteinPiana_value = self.proteinPiana)
            self.dic_functions = dict(map(None, list_functions, []))

    # --------------------------------
    # PianaGraphNodeAttribute "get" methods
    # --------------------------------
    #
    # Methods to get values from PianaGraphNodeAttributes
    #
    # If the value requested is not already in memory (i.e. mem_mode is onDemand), it is retrieved from the database


    def get_node_id(self):
        """
        overwrittes the method in GraphNodeAttribute, identifying this attribute with the proteinPiana
        """
        return self.proteinPiana 
    
    def get_proteinPiana(self):
                  
        return self.proteinPiana  # proteinPiana cannot be None


    def get_molecular_weight(self):
        """
        Returns the molecular weight of the protein from the database
    
        """
        if self.mem_mode!="inFlight":
            if self.molecularWeight is None:
                self._set_molecular_weight()
            
            return self.molecularWeight
        else:
            return self.db_access.get_protein_mw(proteinPiana= self.proteinPiana)
    
    def get_isoelectric_point(self):
        """
        Returns the molecular weight of the protein 

        """
        if self.mem_mode!="inFlight":
            if self.isoelectricPoint is None:
                self._set_isoelectric_point()
            
            return self.isoelectricPoint
        else:
            return self.db_access.get_protein_ip(proteinPiana= self.proteinPiana)

    # 
    # the following methods are used for unifying nodes, not for normal PianaGraph construction
    #

    def get_associated_node_ids_set(self):
        return self.get_associated_proteinPianas()

    def get_associated_proteinPianas(self):
        """
        Returns the set of associated proteinPianas. The original one too
        """

        return self.set_associated_node_ids
        #associated_proteinPianas = {self.proteinPiana: None}

        #for actual_node in self.associated_nodes:
        #    associated_proteinPianas[actual_node.get_node_id()] = None

        #for actual_node in self.set_associated_node_ids:
        #    associated_proteinPianas[actual_node] = None

        #return associated_proteinPianas

    #def get_associated_nodes(self):
    #    """
    #    Returns the 
    #
    #    return self.associated_nodes
    
    def get_dic_taxonomies(self):
        """
        Returns a dictionary {tax:None, tax:None, ...} with the taxonomies for this node
        """
        if self.mem_mode!="inFlight":
            if self.dic_taxonomies is None:
                self._set_dic_taxonomies()
            
            return self.dic_taxonomies
        else:
            return_dict = {}
            for actual_proteinPiana in self.set_associated_node_ids:
                return_dict.update(dict(map(None,self.db_access.get_protein_taxonomy_ids(proteinPiana_value = actual_proteinPiana),[])))

            return return_dict

    def get_dic_cell_location(self):
        """
        Returns a dictionary {location:None, location:None, ...} with the cellular locations for this node
        """

        if self.mem_mode!="inFlight":
            if self.dic_cell_location is None:
                self._set_dic_cell_location()
            return self.dic_cell_location
        else:        
            return_dict = {}
            
            for actual_proteinPiana in self.set_associated_node_ids:
                return_dict.update(dict(map(None,self.db_access.get_protein_subcellularLocation(proteinPiana_value = actual_proteinPiana),[])))
            
            return return_dict
    
    def get_cell_location(self):
        """
        Returns a dictionary {location:None, location:None, ...} with the cellular locations for this node
        """

        if self.mem_mode!="inFlight":
            if self.dic_cell_location is None:
                self._set_dic_cell_location()
            
            return self.dic_cell_location.keys()
        else:
            return_dict = {}
            for actual_node in self.set_associated_node_ids:
                return_dict.update(dict(map(None, self.db_access.get_protein_subcellularLocation(proteinPiana_value = actual_node), [])))

            return return_dict.keys()
        
    
    def get_keywords_appearing(self):
    #def get_keywords_appearing(self, user_keywords=[]):
        """
        returns the list of user keywords that appear in the name, description or function of this protein
        """
        if self.dic_keywords_appearing is None:
            self._set_keywords_appearing()

        return list(self.dic_keywords_appearing)
            
    
    def get_dic_keywords_appearing(self, user_keywords=[]):
        """
        returns the list of user keywords that appear in the name, description or function of this protein
        """
        if self.dic_keywords_appearing is None:
            self._set_keywords_appearing()

        return self.dic_keywords_appearing
            
    
    def get_expression(self):
        """
        returns the expression (over_and_infra_expressed, over_expressed, infra_expressed or None) for this node
        """
        return self.expression

    def get_special_labels(self):
        """
        returns the list of special labels for this node

	if no special labels are found, returns empty list
        """
        return self.dic_special_labels.keys()

    def get_dic_special_labels(self):
        """
        returns the list of special labels for this node

	if no special labels are found, returns empty list
        """
        return self.dic_special_labels
    
    def get_descriptions(self):
        """
        returns the list of descriptions for this node
        """
        if self.mem_mode!="inFlight":
            if self.dic_descriptions is None:
                self._set_dic_descriptions()
            
            return self.dic_descriptions.keys()
        else:
            return_dict = {}

            for actual_node in self.set_associated_node_ids:
                return_dict.update(dict(map(None, self.db_access.get_protein_description(proteinPiana_value = actual_node), [])))

            return return_dict.keys()
    
    def get_dic_descriptions(self):
        """
        returns the list of descriptions for this node
        """
        if self.mem_mode!="inFlight":
            if self.dic_descriptions is None:
                self._set_dic_descriptions()
                    
            return self.dic_descriptions
        else:
            return_dict = {}

            for actual_node in self.set_associated_node_ids:
                return_dict.update(dict(map(None,self.db_access.get_protein_description(proteinPiana_value = actual_node),[])))
            
            return return_dict
        
    
    def get_dic_functions(self):
        """
        returns the list of functions for this node
        """
        if self.mem_mode!="inFlight":
            if self.dic_functions is None:
                self._set_dic_functions()
            
            return self.dic_functions
        else:
            return_dict = {}
            for actual_node in self.set_associated_node_ids:
                return_dict.update(dict(map(None,self.db_access.get_protein_function(proteinPiana_value = actual_node),[])))

            return return_dict

    def get_functions(self):
        """
        returns the list of functions for this node
        """
        if self.mem_mode!="inFlight":
            if self.dic_functions is None:
                self._set_dic_functions()
            
            return self.dic_functions.keys()
        else:
            return_dict = {}
            for actual_node in self.set_associated_node_ids:
                return_dict.update(dict(map(None,self.db_access.get_protein_function(proteinPiana_value = actual_node),[])))

            return return_dict.keys()
        

    def get_union_neighbours(self):
        """
        returns the set of neighbours for the unified node
         (ie. list of neighbours for all nodes that have been 'unified' into this node attribute)
        """

        
        return self.union_neighbours

        
