"""
 File       : PianaApi.py
 Author     : R. Aragues
 Creation   : 17.03.2005
 Contents   : class that acts as an interface to all piana commands
 Called from: piana.py and web interface to PIANA

=======================================================================================================

If you simply want to use PIANA as a library for protein-protein interaction networks, this class
provides you with methods that execute a number of useful commands.

For example, you would something like this :

# 1. get from user strings describing the database, the host, mysql user and mysql password

piana_dbname= "pianaDB_limited"
piana_dbhost= "localhost"
piana_dbuser= None
piana_dbpass= None
piana_dbport= None

# 2. create the piana api object
piana_api = PianaApi(piana_dbname=piana_dbname, piana_dbhost= piana_dbhost, piana_dbuser=piana_dbuser, piana_dbpass= piana_dbpass)

# 3. create a network from a list of proteins

#  3.1 get from user the file name, type of code used, depth desired, tax id, hub threshold, source dbs to use and methods to use
file_name = "dummy_file.txt"
this_file_id_type = "geneName"  # any type of identifier accepted by PIANA (do python piana.py --help for info on how to know
                                     # types of identifiers accepted by your PIANA database
this_file_depth = 2
this_file_hub_threshold = 0
list_source_dbs= ["dip", "string"]
inverse_dbs= "no"
list_source_methods= ["y2h", "coip"]
inverse_methods= "no"

# 3.2 make the call to piana_api to add proteins in a file to the network
#    --> this will search for interactions in the database for those proteins
#        and create a network

piana_api.add_file_proteins_to_piana_graph(file_object= file(file_name,"r"),
                                           id_type= this_file_id_type,
                                           tax_id_value= this_file_tax_id,
                                           depth = this_file_depth,
                                           hub_threshold= this_file_hub_threshold,
                                           use_self_ints=use_self_ints,
                                           list_source_dbs= list_source_dbs,
                                           inverse_dbs= inverse_dbs,
                                           list_source_methods= list_source_methods,
                                           inverse_methods= inverse_methods )

# You could now continue calling commands from piana_api.*


"""

# PianaApi.py: class that acts as an interface to all piana commands
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


import sys
import random

from PianaGraph import *
from PianaGraphBuilder import *
from PianaGraphFilters import *
from PianaGraphExpansions import *
from PianaDBaccess import *
import PianaGlobals

from Clustering import *
from ClusteredGraphGenerator import *

from GoGraph import * # graph where each node is a GO term
from GoClusteringSimilarityFunction import *
from GoClusteringStopCondition import *

from IRGraph import * # graph where each node is an Interacting Region
from GraphCluster import * # graph where each node is a cluster with node elements

from CirClusteringSimilarityFunction import *
from CirClusteringStopCondition import *

import piana_configuration_parameters

if piana_configuration_parameters.piana_mode == "developer" or piana_configuration_parameters.piana_mode == "advanced_user":
    from PatchDecomposition import *
    from ScoringFunction import *
    from ScoringFunctionBenchmark import *



verbose = 0
verbose_species = 1

verbose_command_sequence = 1
verbose_prots_imotifs = 0   # prints info for each comparison of the clustering and every action taken
verbose_prots_imotifs_shallow = 1   # shows progress of clustering (very shallow)
verbose_prots_imotifs_iteration = 0 # shows how many nodes of the similarity matrix have been processed

class PianaApi(object):
    """
    class that acts as an interface to all piana commands
    """
    # PianaApi global variables
    #
    # these dictionaries are used to check the arguments received by PianaApi methods
    #

    info_values = {"all":None,
                   "scop":None,
                   "cath":None,
                   "no":None}

    valid_methods= PianaGlobals.method_names

    valid_output_formats= {"table":None,
                           "network":None}


    valid_output_modes= {"compact":None,
                         "extended":None}

    valid_print_modes= {"all":None,
                        "all_root":None,
                        "only_root":None,
                        "connecting":None}

    valid_format_modes= {"html":None,
                         "txt":None,
                         "dot":None,
                         "sif":None}


    valid_expansion_types = PianaGlobals.expansion_types

    valid_expansion_modes= {"all":None,
                            "root":None}

    valid_exp_output_modes = {"add":None,
                             "print":None}



    valid_term_types = {"molecular_function":None,
                        "biological_process":None,
                        "cellular_component":None}

    valid_sim_modes = {"min":None,
                       "max":None,
                       "random":None,
                       "average":None}


    yes_no_values = {"yes":None, "no":None}

    max_min_values = {"max":None, "min":None}
 
    def __init__(self, piana_dbname=None, piana_dbhost=None, piana_dbuser=None, piana_dbpass=None, piana_dbport=None,
                 use_secondary_db="no", network_name="name", memory_mode="memory" ):
	"""
	Creates your PIANA API object that you will then use to work with PIANA.

	"piana_dbname", "piana_dbhost", "piana_dbuser", "piana_dbpass" and "piana_dbport"are the parameters that are going to be used to create a
        connection to the MySQL PIANA database

	"use_secondary_db" must be set to 'yes' if you want PIANA to look for interactions in a secondary database in case it 
        does not find any in the primary database (the one that you have set with the above parameters).
             -> a detailed description of use_secondary_db can be read on piana/code/execs/conf_files/general_template.piana_conf
             -> if you set this parameter to 'yes', the behaviour described above will be applied to all proteins that you add
                to your network
                -> even if you set this parameter to 'no', you can still use the secondary PIANA database for specific proteins using
                argument force_secondary_db in PianaApi methods add_protein_to_piana_graph and add_proteins_file_to_piana_graph
             -> the secondary PIANA database that will be used is the one specified on piana/code/utilities/piana_configuration_parameters.py

         "memory_mode" is the flag that tells how to manage memory. It can be:
                      "memory" Edge objects and Edge Attributes will be stored in memory. It takes more memory, but it is faster. So, memory usage will be high
                      "speed" Edge Objects will not be stored in memory: they will be generated on the flight when are needed. So, speed will be decreased but memory usage will decrease considerably
	"""

        # creating connection to piana using PianaDBaccess class (piana_dbuser and piana_dbpassword can be None)
        self.piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser,
                                          dbpassword= piana_dbpass, dbport= piana_dbport)

        self.memory_mode = memory_mode

        #print "MEMORY MODE in PianaApi: %s" %(memory_mode)

	self.network_name= network_name

	if use_secondary_db != "yes" and use_secondary_db != "no":
	    raise ValueError("Wrong value given to parameter use_secondary_db: maybe you forgot to write it in your configuration file\n")

	self.use_secondary_db = use_secondary_db

	if piana_configuration_parameters.piana_secondary_dbname is not None:
	    self.piana_secondary_access = PianaDBaccess(dbname= piana_configuration_parameters.piana_secondary_dbname, 
							dbhost= piana_configuration_parameters.piana_secondary_dbhost, 
							dbuser=piana_configuration_parameters.piana_secondary_dbuser, 
							dbpassword= piana_configuration_parameters.piana_secondary_dbpass,
							dbport= piana_configuration_parameters.piana_secondary_dbport)
	else:
	    self.piana_secondary_access = None
	    
	    

        self.key_list_to_use = 1      # used for unifying protein names when printing information of protein lists
        
        self.user_root_proteins = {}  # keep a list of the names used by the user, in order to (in case of using the same protein code type) use those
                                      # names for the output   Contents of this dictionary will be None
        self.user_id_types = {}       # keep a list of the types of protein identifiers used by user. To be used in conjuction with user_root_proteins.
        


        self.valid_protein_ids = self.piana_access.get_valid_protein_ids()


        self.valid_source_dbs =  self.piana_access.get_valid_sources()

        # call to PianaGraph to create an empty graph object 
        self.piana_graph = PianaGraph(graphID=self.network_name, piana_access_object= self.piana_access, 
				      piana_secondary_access_object = self.piana_secondary_access,
                                      memory_mode = self.memory_mode )



       
    
    def _check_arguments(self,
                         method_name=None,
                         protein_code="def_val",
                         another_protein_code= "def_val",
                         protein_list= "def_val",
                         file_object= "def_val",
                         another_file_object= "def_val",
                         id_type=  "def_val",
                         another_id_type= "def_val",
                         alternative_id_types = "def_val",
                         protein_species_name= "def_val",
                         network_species_name= "def_val",
                         tax_id= "def_val",
                         depth="def_val",
                         distance="def_val",
                         info="def_val",
                         hub_threshold="def_val",
                         use_self_ints="def_val",
                         list_source_dbs="def_val",
                         inverse_dbs="def_val",
                         source_db="def_val",
                         list_source_methods="def_val",
                         inverse_methods="def_val",
                         method="def_val",
                         confidence="def_val",
                         expansion_type="def_val",
                         expansion_mode="def_val",
                         expansion_threshold="def_val",
                         exp_output_mode="def_val",
                         output_format="def_val",
                         output_mode="def_val",
                         print_mode="def_val",
                         format_mode="def_val",
                         term_type="def_val",
                         score_threshold="def_val",
                         sim_mode="def_val",
                         level_threshold="def_val",
                         distance_threshold="def_val",
                         rep_term="def_val",
                         print_id="def_val",
                         molecular_error_bounds="def_val",
                         isoelectric_error_bounds="def_val",
                         list_keywords="def_val", 
                         intersection_dbs="def_val",
                         file_over_expressed= "def_val",
                         file_infra_expressed= "def_val",
                         expression_id_type= "def_val",
			 force_secondary_db= "def_val",
			 directory="def_val",
			 force_expression="def_val",
			 dic_special_proteins_file= "def_val"):
        
        """
        checks arguments for PianaApi methods: arguments that are not to be checked must be left to their default value 'def_val' to make sure they
        are not checked...
        """
        # TO DO!!! THis is an horrendous way of checking arguments validity. Generalize it!!!

        if method_name is None:
            raise ValueError("check_arguments requires a method name")
        
        if protein_code != "def_val":
            if protein_code is None:
                # i cannot think of other invalid values for protein code...
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s does not accept %s as a protein code\
\n------------------------------------------------------------------------------------------\n" %(method_name, protein_code))
            
        if another_protein_code != "def_val":
            if another_protein_code is None:
                # i cannot think of other invalid values for another_protein code...
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s does not accept %s as a another_protein code\
\n------------------------------------------------------------------------------------------\n" %(method_name, another_protein_code))
            
        if protein_list != "def_val":
            if protein_list is None or not isinstance(protein_list, list):
                # i cannot think of other invalid values for protein list...
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s does not accept %s as a protein list\
\n------------------------------------------------------------------------------------------\n" %(method_name, protein_list))
            
        if file_object != "def_val":
            if exp_output_mode == "add":
                pass # in this mode, no file is required
            else:
                if not isinstance(file_object, file):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a file object (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, file_object))
        
        if another_file_object != "def_val":
            if exp_output_mode == "add":
                pass # in this mode, no file is required
            else:
                if not isinstance(another_file_object, file):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a file object (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, another_file_object))
            
        if another_file_object != "def_val":
            if not isinstance(another_file_object, file):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a file object (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, another_file_object))
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid id type (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, expression_id_type))
            
        if id_type != "def_val":
            if not self.valid_protein_ids.has_key(id_type):
                valid_string = "\n       --> valid id types are: "
                for tmp_id_type in self.valid_protein_ids.keys():
                    valid_string += "%s " %tmp_id_type
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid id type (your value was: %s)\n\
%s\
\n------------------------------------------------------------------------------------------\n" %(method_name, id_type, valid_string))
        
        if another_id_type != "def_val":
            if not self.valid_protein_ids.has_key(another_id_type):
                valid_string = "\n       --> valid id types are: "
                for tmp_id_type in self.valid_protein_ids.keys():
                    valid_string += "%s " %tmp_id_type
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid id type (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, another_id_type, valid_string))

        if alternative_id_types != "def_val":
            if not isinstance(alternative_id_types, list):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for alternative types (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, alternative_id_types))

            for type in alternative_id_types:
                if not self.valid_protein_ids.has_key(type):
                    valid_string = "\n       --> valid id types are: "
                    for tmp_id_type in self.valid_protein_ids.keys():
                        valid_string += "%s " %tmp_id_type
                    valid_string += "\n"
                
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires valid id types for alternative type names (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, type , valid_string))
                # END OF if not self.valid_protein_ids.has_key(type):
            # END OF for type in alternative_id_types:
                
        if protein_species_name != "def_val":
            if protein_species_name is None:
                # i cannot think of other invalid values for protein species...
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s does not accept %s as a protein species\
\n------------------------------------------------------------------------------------------\n" %(method_name, protein_species_name))
                 
        if network_species_name != "def_val":
            # this can be None only if tax_id is not None: it is the argument for create_species_piana_graph
            if tax_id != "def_val":
                if network_species_name is None and tax_id is None:
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s does not accept None for both protein species and tax_id\
\n------------------------------------------------------------------------------------------\n" %(method_name))
                    
            
        if depth != "def_val":
            if not isinstance(depth, int) or depth < 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid depth value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, depth))
            
        if distance != "def_val":
            if not isinstance(distance, int) or distance < 1:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid distance value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, distance))

        if info != "def_val":
            if not PianaApi.info_values.has_key(info):
                valid_string ="Invalid info value set. Valid info values are: "
                for info_val in PianaApi.info_values:
                    valid_string += " %s" %(info_val)
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid info value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, info, valid_string))
            
        if hub_threshold != "def_val":
            if not isinstance(hub_threshold, int) or hub_threshold < 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid hub_threshold value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, hub_threshold))
           
        if use_self_ints != "def_val":
            if not PianaApi.yes_no_values.has_key(use_self_ints):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid use_self_ints value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, use_self_ints ))

        if force_secondary_db != "def_val":
            if not PianaApi.yes_no_values.has_key(force_secondary_db):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid force_secondary_db value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, force_secondary_db ))

        if force_expression != "def_val":
            if not PianaApi.yes_no_values.has_key(force_expression):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid force_expression value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, force_secondary_db ))
 
        if list_source_dbs != "def_val":
            if list_source_dbs != "all":
                if not isinstance(list_source_dbs, list):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for source dbs (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, list_source_dbs))

                for tmp_source_db in list_source_dbs:
                    if not self.valid_source_dbs.has_key(tmp_source_db):
                        raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid source db %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, tmp_source_db ))
            
        if list_source_methods != "def_val":
            if list_source_methods != "all":
                if not isinstance(list_source_methods, list):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for source methods (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, list_source_methods))

                for tmp_method in list_source_methods:
                    if not PianaApi.valid_methods.has_key(tmp_method):
                        raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid source method %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, tmp_method))
            # END OF if list_source_methods != "all":

        if inverse_dbs != "def_val":
            if not PianaApi.yes_no_values.has_key(inverse_dbs):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid inverse_dbs value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, inverse_dbs ))
            
        if inverse_methods != "def_val":
            if not PianaApi.yes_no_values.has_key(inverse_methods):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid inverse_methods value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, inverse_methods ))

        if source_db != "def_val":
            if source_db is not None:
                if not self.valid_source_dbs.has_key(source_db):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid source db %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, source_db ))
            
        if method != "def_val":
            if method is not None:
                if not PianaApi.valid_methods.has_key(method):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid method %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, method ))
                
        if confidence != "def_val":
            if not isinstance(confidence, int):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid confidence value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, confidence))
            
        if expansion_type != "def_val":  
            if not PianaApi.valid_expansion_types.has_key(expansion_type):

                valid_string = "     --> valid expansion types are: "
                for tmp_exp_type in PianaApi.valid_expansion_types:
                    valid_string += " %s" %tmp_exp_type
                valid_string += "\n"

                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid expansion_type value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, expansion_type, valid_string))
            
        if expansion_mode != "def_val":  
            if not PianaApi.valid_expansion_modes.has_key(expansion_mode):

                valid_string = "     --> valid expansion modes are: "
                for tmp_exp_mode in PianaApi.valid_expansion_modes:
                    valid_string += " %s" %tmp_exp_mode
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid expansion_mode value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, expansion_mode, valid_string))
            
        if expansion_threshold != "def_val":
            if not isinstance(expansion_threshold, int) or expansion_threshold < 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid expansion_threshold value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, expansion_threshold))
            
        if exp_output_mode != "def_val":  
            if not PianaApi.valid_exp_output_modes.has_key(exp_output_mode):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid exp_output_mode value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, exp_output_mode))
            
        if output_format != "def_val":  
            if not PianaApi.valid_output_formats.has_key(output_format):

                valid_string = "     --> valid output formats are: "
                for tmp_out_format in PianaApi.valid_output_formats:
                    valid_string += " %s" %tmp_out_format
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid output_format value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, output_format, valid_string))
            
        if output_mode != "def_val":  
            if not PianaApi.valid_output_modes.has_key(output_mode):

                valid_string = "     --> valid output modes are: "
                for tmp_out_mode in PianaApi.valid_output_modes:
                    valid_string += " %s" %tmp_out_mode
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid output_mode value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, output_mode, valid_string))
            
        if print_mode != "def_val":  
            if not PianaApi.valid_print_modes.has_key(print_mode):

                valid_string = "     --> valid print modes are: "
                for tmp_print_mode in PianaApi.valid_print_modes:
                    valid_string += " %s" %tmp_print_mode
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid print_mode value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, print_mode, valid_string))
            
        if format_mode != "def_val":  
            if not PianaApi.valid_format_modes.has_key(format_mode):

                valid_string = "     --> valid format modes are: "
                for tmp_format_mode in PianaApi.valid_format_modes:
                    valid_string += " %s" %tmp_format_mode
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid format_mode value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, format_mode, valid_string))
            
        if term_type != "def_val":  
            if not PianaApi.valid_term_types.has_key(term_type):

                valid_string = "     --> valid term types are: "
                for tmp_term_type in PianaApi.valid_term_types:
                    valid_string += " %s" %tmp_term_type
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid term_type value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, term_type, valid_string))
            
        if score_threshold != "def_val":
            if not isinstance(score_threshold, float) or score_threshold <= 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid score_threshold value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, score_threshold))
            
        if sim_mode != "def_val":  
            if not PianaApi.valid_sim_modes.has_key(sim_mode):

                valid_string = "     --> valid sim modes are: "
                for tmp_sim_mode in PianaApi.valid_sim_modes:
                    valid_string += " %s" %tmp_sim_mode
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid sim_mode value (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, sim_mode, valid_string))
            
        if level_threshold != "def_val":
            if not isinstance(level_threshold, int) or level_threshold < 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid level_threshold value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, level_threshold))
            
        if distance_threshold != "def_val":
            if not isinstance(distance_threshold, int) or distance_threshold < 0:
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid distance_threshold value (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, distance_threshold))
            
        if rep_term != "def_val":  
            if not PianaApi.max_min_values.has_key(rep_term):

                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid rep_term value (min or max) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, rep_term))
            
        if print_id != "def_val":  
            if not PianaApi.yes_no_values.has_key(print_id):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid print_id value (yes or no) (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, print_id))

        if molecular_error_bounds != "def_val":
            if not isinstance(molecular_error_bounds, list):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for molecular weight error bounds (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, molecular_error_bounds))

            for error_bound in molecular_error_bounds:
                if not isinstance(error_bound, float):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid list of molecular weight error bounds %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, error_bound))

        if isoelectric_error_bounds != "def_val":
            if not isinstance(isoelectric_error_bounds, list):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for isoelectric point error bounds (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, isoelectric_error_bounds))

            for error_bound in isoelectric_error_bounds:
                if not isinstance(error_bound, float):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s found an invalid list of isoelectric point error bounds %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, error_bound))

        if list_keywords != "def_val":
            if not isinstance(list_keywords, list):
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list of keywords(your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, list_keywords))
            
        if intersection_dbs != "def_val":
            if intersection_dbs is not None:
                if not isinstance(intersection_dbs, list):
                    raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a list for intersection dbs (your value was: %s)\
\n------------------------------------------------------------------------------------------\n" %(method_name, intersection_dbs))

                for tmp_source_db in intersection_dbs:
                    if not self.valid_source_dbs.has_key(tmp_source_db):
                        raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s could not find the following third-party source database in your piana database: %s\
\n------------------------------------------------------------------------------------------\n" %(method_name, tmp_source_db ))
            # END OF  if intersection_dbs is not None:

        if (file_over_expressed != "def_val" and file_over_expressed is not None) or (
            file_infra_expressed != "def_val" and file_infra_expressed is not None):
            if not self.valid_protein_ids.has_key(expression_id_type):
                valid_string = "\n       --> valid id types are: "
                for tmp_id_type in self.valid_protein_ids.keys():
                    valid_string += "%s " %tmp_id_type
                valid_string += "\n"
                
                raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires valid id types for expression protein type (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, expression_id_type , valid_string))

        if file_over_expressed != "def_val":
            # nothing to check...
            pass

        if file_infra_expressed != "def_val":
            # nothing to check...
            pass

	if dic_special_proteins_file  != "def_val":
	    try:
		# this must be a dic { key:[id type, file name], key:[], ...}
		for one_label in dic_special_proteins_file:
		    if not self.valid_protein_ids.has_key(dic_special_proteins_file[one_label][0]):
			raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires valid id types for your special proteins (your value was: %s)\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name, dic_special_proteins_file[one_label][0] ))
	    except:
		raise ValueError("\
\n------------------------------------------------------------------------------------------\n \
Attention! Check your parameters and command arguments! Your configuration file contains incorrect values...\n \n \
PianaApi method %s requires a valid format for parameter dic_special_proteins_file\n \
%s \
\n------------------------------------------------------------------------------------------\n" %(method_name))
		
		    
		


        if directory != "def_val":
            # nothing to check... it will raise error if it doesn't exist
            pass



    def write_reference_card(self, file_object):
        """
        Writes the PIANA reference card to file object.

        It will (obviously) print the reference card with data from the PIANA database that is being used by this PianaApi object


        ATTENTION: this method can only be called from piana.py
                   Why? because I have hardcoded the path to general_template.piana_conf, and it is hardcoded as ./conf_files/general_template.piana_conf
                        and calling this method from any other directory will crash, as it will not find the template.
        
        """
        # print header for reference card
        #  -> number of interactions
        #  -> number of proteins
        #  -> database versions that were used to create this PIANA database
        # 
        sys.stdout.write("\n\t\t--------------------\t\t\n")
        sys.stdout.write("\t\tPIANA reference card\t\t\n")
        sys.stdout.write("\t\t--------------------\t\t\n\n")
        sys.stdout.write("Database:%s\n" %(self.piana_access.dbname))
        sys.stdout.write("Server:%s\n" %(self.piana_access.dbhost))
        #sys.stdout.write("Number of Proteins:\n")
        #sys.stdout.write("Number of Interactions:\n")
        sys.stdout.write("\n\n")
        sys.stdout.write("\nTHIRD-PARTY DATABASES THAT WERE USED TO CREATE THIS DATABASE:\n")
        these_source_dbs = self.piana_access.get_source_dbs_dict() # dictionary with info: key: databaseName, contents: 
        for one_source_db in these_source_dbs:
            # one_source_db is the name of the source db
            # these_source_dbs[one_source_db] is a dic with keys databaseVersion, parsed_file, date, database_description, database_type
            
            sys.stdout.write("\n Database %s version %s was parsed on %s (contains %s)\n" %(one_source_db,
                                                                                            these_source_dbs[one_source_db]["databaseVersion"],
                                                                                            these_source_dbs[one_source_db]["date"],
                                                                                            these_source_dbs[one_source_db]["database_type"]))


        # print protein id types in this database

        sys.stdout.write("\n\nVALID PIANA PROTEIN IDENTIFIER TYPES\n")
        sys.stdout.write("\n\n The following are identifier types accepted by PIANA\n")
        these_id_types = self.piana_access.get_valid_protein_ids() # dictionary with info: key: databaseName, contents: 
        for one_id_type in these_id_types:
            # one_id_type is the id type
            # these_id_types[one_id_type] is a dic with keys table, column, description
            sys.stdout.write("\n Use id type %s for %s\n" %(one_id_type, these_id_types[one_id_type]["description"]))


        # print interaction methods in this database

        sys.stdout.write("\n\nDETECTION METHODS ACCEPTED BY PIANA\n")
        sys.stdout.write("\n\n The following are the interaction methods accepted by PIANA\n")
        for one_method in PianaGlobals.method_names:
            # one_method is the PIANA name for this method
            # PianaGlobals.method_names[one_method] is a list of descriptions for this method
            sys.stdout.write(" - %s is the PIANA name for method descriptions: " %(one_method))
            for one_description in PianaGlobals.method_names[one_method]:
                sys.stdout.write("%s\t" %(one_description))
            sys.stdout.write("\n")
            


        # print colors used in .DOT file
        sys.stdout.write("\n\nCOLORS USED IN PIANA DOT FILES\n")
        sys.stdout.write("\n\n The following are the colors used by PIANA in DOT files \n(in Cytoscape you can choose your own colors)\n")
        sys.stdout.write("\n")
        sys.stdout.write("For interactions, edges are colored according to this table:\n")
        for one_id in PianaGlobals.interaction_source_databases_colors:
            sys.stdout.write(" - interactions coming from %s are printed in color %s\n" %(one_id, PianaGlobals.interaction_source_databases_colors[one_id] ))
        sys.stdout.write("\n")
        sys.stdout.write("For proteins, background of nodes are colored according to this table:\n")
        for one_id in PianaGlobals.node_fill_colors:
            sys.stdout.write(" - proteins that are %s are printed with background color %s\n" %(one_id, PianaGlobals.node_fill_colors[one_id] ))
        sys.stdout.write("\n")
        sys.stdout.write("For proteins, node borders are colored according to this table:\n")
        for one_id in PianaGlobals.node_border_colors:
            sys.stdout.write(" - proteins that are %s are printed with border color %s\n" %(one_id, PianaGlobals.node_border_colors[one_id] ))

        # print summary PIANA commands and their arguments


        sys.stdout.write("\n\nARGUMENTS YOU CAN GIVE TO piana.py THROUGH THE COMMAND LINE\n")
        sys.stdout.write("\nYou can always obtain the arguments by doing python piana.py --help\n")
        sys.stdout.write("\n")
        
        
        sys.stdout.write("\n\nSUMMARY OF PIANA PARAMETERS AND COMMANDS IN THE CONFIGURATION FILE\n")
        sys.stdout.write("\nA detailed description of PIANA commands and parameters can be\n read at piana/code/execs/conf_files/general_template.piana_conf\n")
        sys.stdout.write("\n")
        try:
            general_template_lines = utilities.return_non_commented_non_empty_lines("./conf_files/general_template.piana_conf")
        except:
            raise ValueError("general_template.piana_conf could not be found: you have either change the structure of PIANA directories, or you have deleted the file\n Please, make sure the file general_template.piana_conf is located under directory piana/code/execs/conf_files")

        section = 1
        sys.stdout.write(" ==> Input and Output parameters:\n\n")
        for one_template_line in general_template_lines:

            if one_template_line.startswith(">>>"):
                section = 2
                sys.stdout.write("\n\n ==> PIANA commands:\n\n")
                
                continue
                
            if section == 1:
                # if this is a parameter, print it as a parameter
                sys.stdout.write("    - %s \n" %one_template_line.split("=")[0])
            else:
                sys.stdout.write("    - %s \n" %one_template_line)
            
            
            
        sys.stdout.write("\n")

    def get_external_databases_in_pianaDB(self, database_type):
        """
        returns the names of external databases that have been integrated into the PIANA databases associated to this PianaApi

        "database_type" specifies which type of external databases will be returned
            -> valid values for database_type are:
                 - "all"
                 - "protein sequences"
                 - "protein attributes"
                 - "protein-protein interactions"
                 - "identifiers cross-references"
        """
        return self.piana_access.get_source_dbs_dict(database_type=database_type).keys()
    
    # ------------------------------------
    #
    #  Reset, Load, Save ... Piana Piana_graph
    # 
    # ------------------------------------

    def reset_piana_graph(self):
        """
        start a new piana_graph
        """
        self.piana_graph=  PianaGraph(graphID=self.network_name, piana_access_object= self.piana_access, 
				      piana_secondary_access_object = self.piana_secondary_access,
                                      memory_mode = self.memory_mode)

    def load_piana_graph(self, file_object):
        """
        loads a previously saved piana_graph into the current piana_graph (ie replaces current piana_graph by the one in file_object)

        file_object must have been opened using binary read (ie file_object= file(file_name, "rb") )
        
        """
        self._check_arguments(method_name= "load_piana_graph",
                              file_object= file_object)
            
        self.piana_graph = cPickle.load(file_object)

    def save_piana_graph(self, file_object):
        """
        saves current piana_graph in file file_object

        file_object must have been created using binary write (ir file_object = file(file_name, "wb")
        """
        self._check_arguments(method_name= "save_piana_graph",
                              file_object= file_object)
        
        cPickle.dump(self.piana_graph, file_object , 2)  # 2 is the pickle protocol to be applied


    

    # ------------------------------------
    #
    #  Miscelaneous 
    # 
    # ------------------------------------
    def get_one_tax_id_from_species_name(self, species_name=None):
        """
        returns the tax id associated to "species_name"

        If there are more than one tax id associated, raises an error (to avoid ambiguities)

        If there is no tax id associated to it, raises an error (to avoid using a name that doesn't exist)
        """

        if species_name == "all":
            # species_name all is codified in tax id with a 0
            return 0
        
        list_tax_ids = self.piana_access.get_taxonomies_from_species_name(species_name_value= species_name)

        num_tax_ids = len(list_tax_ids)

        if num_tax_ids == 1:
            return list_tax_ids[0]
        elif  num_tax_ids == 0:
            raise ValueError("No tax id associated to species name <%s>\n" %(species_name))
        else:
            raise ValueError("Ambiguous species name: <%s> is associated to tax ids %s \n" %(species_name, list_tax_ids))

    def get_random_protein_dic(self, tax_id_value, dic_size, 
                               id_type, alternative_id_types= [], 
                               force_ints="yes", use_self_ints="yes",
                               list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                               file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
                               dic_special_proteins_file= None):
        """
        returns a dic with random proteins (of size "dic_size") from species "tax_id_value"

             {  protein1:{'expression':expression_value, 'labels':[label1, label2, ...]},
                protein2: ....
             }

        the type of identifier used for the proteins will be id_type (or alternative_id_types)

        if "force_ints" is yes, it only returns proteins if they have known interactions.
            --> if set to "yes", then you must set all parameters associated to interactions


        expression_values will be inserted only if the following expression arguments are set:
    
           "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed
           "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed


        labels will be inserted only if argument dic_special_proteins_file is set
        
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        """
        if verbose:
            sys.stderr.write("-getting complete list of proteins from database that respects criteria required\n")
            
        # 1. get the complete list of potential proteins
        if force_ints == "yes":
            proteinPianas_dic = {}  # used to colect unique proteins of a given species that have interactions 
            interaction_triplets = self.piana_access.get_all_protein_protein_interactions( use_self_ints = use_self_ints,
                                                                                           list_source_dbs=list_source_dbs,
                                                                                           inverse_dbs= inverse_dbs,
                                                                                           list_source_methods=list_source_methods,
                                                                                           inverse_methods= inverse_methods)

            for one_triplet in interaction_triplets:
                # one_triplet[0] is proteinPianaA
                # one_triplet[1] is proteinPianaB
                if tax_id_value == 0:
                    proteinPianas_dic[one_triplet[0]] = None
                    proteinPianas_dic[one_triplet[1]] = None
                else:
                    # if a tax_id was set, make sure that the proteins are of that species
                    if tax_id_value in self.piana_access.get_protein_taxonomy_ids(proteinPiana_value=one_triplet[0]):
                        proteinPianas_dic[one_triplet[0]] = None
                        
                    if tax_id_value in self.piana_access.get_protein_taxonomy_ids(proteinPiana_value=one_triplet[1]):
                        proteinPianas_dic[one_triplet[1]] = None
            # END OF for one_triplet in interaction_triplets:
            
            all_proteinPianas = proteinPianas_dic.keys()
        else:
            all_proteinPianas= self.piana_access.get_all_proteinPiana(tax_id_value=tax_id_value)
        
            
        # END OF else: (if force_ints == "yes":)

        if verbose:         sys.stderr.write("-unifying names\n")
            
        self.key_list_to_use += 1

        # convert all_proteinPianas to a list of proteins in the type of identifier selected by the user
        all_proteins= self.piana_graph.get_list_unified_names(list_proteinPianas = all_proteinPianas,
                                                              output_id_type= id_type,
                                                              alternative_id_types=alternative_id_types,
                                                              key_list= self.key_list_to_use)
        

        if verbose:       sys.stderr.write("- selecting a random sample of the total list of size %s\n" %(len(all_proteinPianas)))
            
        random.seed()
        if dic_size == 0:
            # if list size desired has been set to Zero, it means we must return all proteinPianas that respect the criteria
            random_protein_list = all_proteins # it is easier to set random protein list to be all proteins than doing twice the
                                               # same process of checking for special info and printing it
        else:
            try:
                random_protein_list = random.sample( population=all_proteins, k=dic_size)
            except:
                raise ValueError("You are asking for a random set of size bigger than it possibly is")
        
        return self.piana_graph.get_proteins_special_labels(list_proteins=random_protein_list,
                                                            id_type=id_type,
                                                            alternative_id_types=alternative_id_types,
                                                            tax_id_value=tax_id_value,
                                                            dic_special_proteins_file=dic_special_proteins_file,
                                                            file_over_expressed= file_over_expressed,
                                                            file_infra_expressed= file_infra_expressed,
                                                            expression_id_type= expression_id_type)
        
    # ------------------------------------
    #
    #  Create, Add, ... to Piana Piana_graph
    # 
    # ------------------------------------
        
    def add_protein_to_piana_graph(self, protein_code, id_type, tax_id_value,
                                   depth, hub_threshold,
                                   use_self_ints = "yes", list_source_dbs="all",  inverse_dbs="no", list_source_methods="all", inverse_methods="no",
				   force_secondary_db="no"):
        """
        Adds one protein and its interactions to current piana_graph.
        
        
        "protein_code" is a string with the protein code (must be of type id_type)

        "id_type" is the type of code used for protein protein_code

        "tax_id_value" sets the species of the protein that is being added (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "depth" fixes the depth at which interactions will be added (eg. depth 2 will add parters of partners of protein protein_code)

        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself

        "hub_threshold" sets the maximum number of interactions a protein can have for it to be added to the piana_graph
           -> 0 is equivalent to not applying any hub threshold

        "list_source_dbs" sets the interaction databases that will be used to get interactions when building the piana_graph
           -> can be a list of dbs (eg ["dip", "string"]) or "all" (all source dbs used)

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)

        "list_source_methods" sets the methods that will be used to get interactions when building the piana_graph
           -> can be a list of methods (eg ["y2h", "tap"]) or "all" (all methods used)

        "inverse_methods" can be:
                no (methods in list_source_methods will be used to build the network)
                yes (all methods except those in list_source_methods will be used to build the network)

        "force_secondary_db" can be:
	        no (use standard piana database to retrieve interactions for this protein)
                yes (use the secondary piana database described in piana/code/utilities/piana_configuration_parameters.py)

        """

        self._check_arguments(method_name= "add_protein_to_piana_graph",
                              protein_code=protein_code,
                              id_type=id_type,
                              tax_id=tax_id_value,
                              depth=depth,
                              hub_threshold=hub_threshold,
                              use_self_ints= use_self_ints,
                              list_source_dbs=list_source_dbs,
                              inverse_dbs=inverse_dbs,
                              list_source_methods= list_source_methods,
                              inverse_methods=inverse_methods,
			      force_secondary_db=force_secondary_db)
        
        # builds a piana_graph for the protein passed as argument and then joins the current piana_graph (self.piana_graph) to the
        # piana_graph built from the new protein

        self.user_root_proteins[protein_code.upper()] = None # keeping the list of root proteins updated
        
        temp_piana_graph = PianaGraph("Temp Graph", piana_access_object= self.piana_access)
        
        #try:
        temp_piana_builder = PianaGraphBuilder(piana_access_object=self.piana_access,
                                               depth=depth,
                                               hub_threshold= hub_threshold,
                                               use_self_ints=use_self_ints,
                                               list_protein_codes= [protein_code],
                                               id_type= id_type,
                                               tax_id_value=tax_id_value,
                                               list_source_dbs= list_source_dbs,
                                               inverse_dbs= inverse_dbs,
                                               list_source_methods= list_source_methods,
                                               inverse_methods= inverse_methods,
                                               use_secondary_db= self.use_secondary_db,
                                               piana_secondary_access_object= self.piana_secondary_access,
                                               force_secondary_db= force_secondary_db)
        
        
        temp_piana_graph.build_graph(temp_piana_builder)
        
        self.piana_graph.join_graphs(temp_piana_graph)
        
        #except Exception, inst:
            # no information available for these proteins or error
       #     sys.stderr.write("---------------------------------------------------------------------------------------------------------------\n")
        #    sys.stderr.write("No proteins were added! PIANA error, database error, codes were unknown or you've set the wrong type for your proteins.\n")
        #    sys.stderr.write(" Error reported: %s -- %s)\n" %(inst, sys.exc_info()[0]))
        #    sys.stderr.write("---------------------------------------------------------------------------------------------------------------\n")

        
    def add_file_proteins_to_piana_graph(self, file_object, id_type, tax_id_value,
                                         depth, hub_threshold, 
                                         use_self_ints = "yes", list_source_dbs="all", inverse_dbs="no", list_source_methods="all", inverse_methods="no",
					 force_secondary_db="no"):
        """
        Adds proteins and their interactions to current piana_graph. Proteins to add are in file "file_object"
        
        "file_object" is a file object pointing to a file that contains one protein per line

        "id_type" is the type of code used for protein protein_code

        "tax_id_value" sets the species of the proteins that are being added (can be used for eliminating ambiguities between codes across species)
           --> valid species names are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "depth" fixes the depth at which interactions will be added (eg. depth 2 will add parters of partners of protein protein_code)

        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself


        "hub_threshold" sets the maximum number of interactions a protein can have for it to be added to the piana_graph
           -> 0 is equivalent to not applying any hub threshold

        "list_source_dbs" sets the interaction databases that will be used to get interactions when building the piana_graph
           -> can be a list of dbs (eg ["dip", "string"]) or "all" (all source dbs used)

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)

        "list_source_methods" sets the methods that will be used to get interactions when building the piana_graph
           -> can be a list of methods (eg ["y2h", "tap"]) or "all" (all methods used)

        "inverse_methods" can be:
                       no (methods in list_source_methods will be used to build the network)
                       yes (all methods except those in list_source_methods will be used to build the network)

        "force_secondary_db" can be:
	        no (use standard piana database to retrieve interactions for this protein)
                yes (use the secondary piana database described in piana/code/utilities/piana_configuration_parameters.py)
        """

        self._check_arguments(method_name= "add_file_proteins_to_piana_graph",
                              file_object=file_object,
                              id_type=id_type,
                              tax_id=tax_id_value,
                              depth=depth,
                              hub_threshold=hub_threshold,
                              use_self_ints= use_self_ints,
                              list_source_dbs=list_source_dbs,
                              inverse_dbs=inverse_dbs,
                              list_source_methods= list_source_methods,
                              inverse_methods=inverse_methods,
			      force_secondary_db=force_secondary_db)
        
        # builds a piana_graph for the proteins of the file and then joins the current piana_graph (self.piana_graph) to the
        # piana_graph built from these new proteins
        
        list_proteins = utilities.return_proteins_from_file(file_object= file_object,
                                                            id_type=id_type)


        for protein_name in list_proteins:
            self.user_root_proteins[protein_name.upper()] = None # keeping the list of root proteins updated

        temp_piana_graph = PianaGraph("Temp Graph", piana_access_object= self.piana_access, memory_mode = self.memory_mode)

        #try:
        temp_piana_builder = PianaGraphBuilder(piana_access_object=self.piana_access,
                                               depth=depth,
                                               hub_threshold= hub_threshold,
                                               use_self_ints=use_self_ints,
                                               list_protein_codes= list_proteins,
                                               id_type= id_type,
                                               tax_id_value=tax_id_value,
                                               list_source_dbs= list_source_dbs,
                                               inverse_dbs= inverse_dbs,
                                               list_source_methods= list_source_methods,
                                               inverse_methods= inverse_methods,
                                               use_secondary_db= self.use_secondary_db,
                                               piana_secondary_access_object= self.piana_secondary_access,
                                               force_secondary_db= force_secondary_db)
        
        temp_piana_graph.build_graph(temp_piana_builder)
        
        self.piana_graph.join_graphs(temp_piana_graph)

        #except Exception, inst:
            # no information available for these proteins or error
        #    sys.stderr.write("---------------------------------------------------------------------------------------------------------------\n")
        #    sys.stderr.write("No proteins were added! PIANA error, database error, codes were unknown or you've set the wrong type for your proteins.\n")
        #    sys.stderr.write(" Error reported: %s -- %s)\n" %(inst, sys.exc_info()[0]))
        #    sys.stderr.write("---------------------------------------------------------------------------------------------------------------\n")


    def add_interaction_to_piana_graph(self, protein_a, protein_b, id_type, source_db= None, method= None, confidence=1):
        """
        Adds one interaction between protein_a and protein_b to current piana_graph
         (adds only this interaction. This command doesn't search for interactions in the database for these proteins.)
         (this is not inserting the interaction into the piana database: it is just adding it to the current network)

        "source_db" is the database from which  you have obtained the interaction
            --> set it to None if this is not relevant for your analysis
            --> set it to 'user' if you want to label it as 'added by user' (will appear in yellow)
            
        "method" is the method you have used to detect the interaction
            --> set it to None if this is not relevant for your analysis
            
        "confidence" is not currently being used... set it to 1

        "id_type" is the type of code used for protein protein_codes protein_a and protein_b (it has to be the same for both)
        
        """
        
        self._check_arguments(method_name= "add_interaction_to_piana_graph",
                              protein_code=protein_a,
                              another_protein_code=protein_b,
                              id_type=id_type,
                              source_db=source_db,
                              method=method,
                              confidence=confidence)
        
        self.piana_graph.add_interaction(protein_a=protein_a, protein_b=protein_b, source_db=source_db, method=method,
                                         confidence=confidence, id_type= id_type)


    def add_file_interactions_to_piana_graph(self, file_object, id_type):
        """
        Adds interactions from a file to current piana_graph
         (adds only those interactions described in the file. This command doesn't search for interactions in the database for these proteins.)
         (this is not inserting the interactions into the piana database: it is just adding it to the current network)

        "file_object" is a file object describing one interaction per line

            --> The interactions file follows the format (set unknown values to None):

            protein_a<TAB>protein_b<TAB>source_database<TAB>detection_method<TAB>confidence_score

            for example, a line could be:

            HOG1    MOT1   None    y2h   None


            This format is described in detail in file piana/code/dbParsers/piana_text_intParser/README.piana_interaction_data_format


        "id_type" is the type of code used for protein protein_code
        
        """

        self._check_arguments(method_name= "add_file_interactions_to_piana_graph",
                              file_object=file_object,
                              id_type=id_type)
        
        for line in file_object:

            line_fields = line.split()

            protein_a  = line_fields[0]
            protein_b  = line_fields[1]

            # to allow the file having just the two proteins in a line, check for line length
            # depending on the line length we suppose that the user has given a different number of parameters
            
            if len(line_fields) > 2:
                if line_fields[2] == "None":
                    source_db = None
                else:
                    source_db  = line_fields[2]
            else:
                source_db = None

            if len(line_fields) > 3:
                if line_fields[3] == "None":
                    method     = None
                else:
                    method     = line_fields[3]
            else:
                method = None

            if len(line_fields) > 4:
                if line_fields[4] == "None":
                    confidence = None
                else:
                    confidence = float(line_fields[4])
            else:
                confidence = None

            self.piana_graph.add_interaction(protein_a=protein_a, protein_b=protein_b, source_db=source_db, method=method,
                                             confidence=confidence, id_type= id_type)

        # END OF for line in file_fd:


    def create_species_piana_graph(self,  species_name=None, tax_id=None, hub_threshold=0, use_self_ints='yes',
                                   list_source_dbs="all", inverse_dbs="no", list_source_methods="all", inverse_methods="no"):
        """
        Replaces current piana_graph with the protein-protein interaction piana_graph for a given species (using all proteins)

        The user can fix the species using a tax id or a species name (don't use both at the same time):
            "taxonomy_id" is the tax id for which the piana_graph will be built (set it to None if not using it)
               -> has to be a valid tax id (eg. 9606; ...)
        
            "species_name" is the species for which the piana_graph will be built (set it to None if not using it)
               -> has to be the name of a species (e.g "human", "yeast", ...)

          --> in case both species_name_value and taxonomy_value are set to something different from None, taxonomy_value is used

        "hub_threshold" sets the maximum number of interactions a protein can have for it to be added to the piana_graph
           -> 0 is equivalent to not applying any hub threshold


        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself
                      
        "list_source_dbs" sets the interaction databases that will be used to get interactions when building the piana_graph
           -> can be a list of dbs (eg ["dip", "string"]) or "all" (all source dbs used)

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)

        "list_source_methods" sets the methods that will be used to get interactions when building the piana_graph
           -> can be a list of methods (eg ["y2h", "tap"]) or "all" (all methods used)

        "inverse_methods" can be:
                       no (methods in list_source_methods will be used to build the network)
                       yes (all methods except those in list_source_methods will be used to build the network)


        Note: no depth argument is needed, since we are adding all interactions for all proteins of a given species
    
        """
        
        self._check_arguments(method_name= "create_species_piana_graph",
                              network_species_name=species_name,
                              tax_id=tax_id,
                              hub_threshold=hub_threshold,
                              use_self_ints=use_self_ints,
                              list_source_dbs=list_source_dbs,
                              inverse_dbs=inverse_dbs,
                              list_source_methods= list_source_methods,
                              inverse_methods=inverse_methods)

        # JAVI:
        piana_builder = PianaGraphBuilder(piana_access_object=self.piana_access,
                                          depth=0,
                                          hub_threshold= hub_threshold,
                                          use_self_ints=use_self_ints,
                                          list_protein_codes= [0],
                                          id_type= "proteinPiana",
                                          tax_id_value=0,
                                          list_source_dbs= list_source_dbs,
                                          inverse_dbs= inverse_dbs,
                                          list_source_methods= list_source_methods,
                                          inverse_methods= inverse_methods,
                                          use_secondary_db= self.use_secondary_db,
                                          piana_secondary_access_object= self.piana_secondary_access,
                                          force_secondary_db="no")  # it doesn't make sense to force a secondary db in this method
                                                                        #  -> if the user wants to use another DB to build the network
                                                                        #     he can set it to be the primary DB

        if species_name:
            list_taxonomy_value = self.piana_access.get_taxonomies_from_species_name(species_name_value = species_name_value)
            for actual_taxID in list_taxonomy_value:
                self.piana_graph.build_graph(piana_builder, complete_taxID=actual_taxID)
        elif tax_id:
            self.piana_graph.build_graph(piana_builder, complete_taxID=tax_id)
            
            
        
        
##         proteins_with_given_species = self.piana_access.get_proteins_sharing_species(species_name_value= species_name,
##                                                                                      taxonomy_value= tax_id)

	
## 	if verbose_species:
## 	    sys.stderr.write("Building network for <%s>-<%s>, which has %s proteins\n" %(species_name, tax_id, len(proteins_with_given_species) ) )

##         if proteins_with_given_species:
##             # Create piana_graph for all proteins in proteins_with_given_species
##             #             In the call to the builder we set tax_id_value to 0 (ie. ignore species) because the list of proteinPiana is 
##             #             already only of the species we wanted... and leaving it to 0 will speed up things in the sql queries
##             #             -> Moreover, we are looking at the network for yeast proteins: if there are known interactions with
##             #                proteins in other species, they belong to the network. If the user doesn't want any proteins from
##             #                other species in his output network, he should set output-proteins-species to his species of interest.
##             piana_builder = PianaGraphBuilder(piana_access_object=self.piana_access,
##                                               depth=0,
##                                               hub_threshold= hub_threshold,
##                                               use_self_ints=use_self_ints,
##                                               list_protein_codes= proteins_with_given_species,
##                                               id_type= "proteinPiana",
##                                               tax_id_value=0,
##                                               list_source_dbs= list_source_dbs,
##                                               inverse_dbs= inverse_dbs,
##                                               list_source_methods= list_source_methods,
##                                               inverse_methods= inverse_methods,
## 					      use_secondary_db= self.use_secondary_db,
## 					      piana_secondary_access_object= self.piana_secondary_access,
## 					      force_secondary_db="no")  # it doesn't make sense to force a secondary db in this method
##                                                                         #  -> if the user wants to use another DB to build the network
##                                                                         #     he can set it to be the primary DB

##             self.piana_graph.build_graph(piana_builder)
##         # END OF if proteins_with_given_species:

##         else:
##             sys.stderr.write("\nNo proteins found for species name %s or tax id %s !!!!\n" name, taxonomy_id))


    def create_database_method_piana_graph(self, tax_id_value=0, use_self_ints="yes",
                                           list_source_dbs="all", inverse_dbs="no", list_source_methods="all", inverse_methods="no"):
        """
        Replaces current piana_graph with the interaction piana_graph for a given list of databases and methods
        (ie. adds all interactions that appear in databases list_source_dbs that are of a method that appear in list_source_methods)

        "tax_id_value" is the species for which the piana_graph will be built
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy


        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself
                      
        "list_source_dbs" sets the interaction databases that will be used to get interactions when building the piana_graph
           -> can be a list of dbs (eg ["dip", "string"]) or "all" (all source dbs used)

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)

        "list_source_methods" sets the methods that will be used to get interactions when building the piana_graph
           -> can be a list of methods (eg ["y2h", "tap"]) or "all" (all methods used)

        "inverse_methods" can be:
                       no (methods in list_source_methods will be used to build the network)
                       yes (all methods except those in list_source_methods will be used to build the network)

        Note: no depth argument is needed, since we are adding all interactions that respect a certain criteria

    
        """
        
        self._check_arguments(method_name= "create_database_method_piana_graph",
                              tax_id=tax_id_value,
                              use_self_ints=use_self_ints,
                              list_source_dbs=list_source_dbs,
                              inverse_dbs=inverse_dbs,
                              list_source_methods= list_source_methods,
                              inverse_methods=inverse_methods)
        
        # TO DO!!! Add argument hub_threshold to this method

        if tax_id_value != 0:
            set_user_tax_ids = Set([tax_id_value])
            
        
        # this call returns the list of triplets (proteinPiana1, proteinPiana2, interactionPiana)
        #    -> use_self_ints is applied below inside the loop...
        all_database_method_interactions = self.piana_access.get_all_protein_protein_interactions(use_self_ints=use_self_ints,
                                                                                                  list_source_dbs= list_source_dbs,
                                                                                                  inverse_dbs= inverse_dbs,
                                                                                                  list_source_methods= list_source_methods,
                                                                                                  inverse_methods= inverse_methods)

        number_skipped_interactions = 0
        for interaction in all_database_method_interactions:

            # interaction[0] is proteinPianaA
            # interaction[1] is proteinPianaB
            # interaction[2] is interactionPiana


            if tax_id_value != 0:
                set_tax_id_a = Set(self.piana_access.get_protein_taxonomy_ids(proteinPiana_value = interaction[0]))
                set_tax_id_b = Set(self.piana_access.get_protein_taxonomy_ids(proteinPiana_value = interaction[1]))

                if not set_tax_id_a.intersection(set_user_tax_ids) or not set_tax_id_b.intersection(set_user_tax_ids):
                    # skip interaction if any of the two proteins is not of the species selected by user
                    number_skipped_interactions += 1
                    continue
            # END OF  if tax_id_value != 0:

            new_node_a_attribute = PianaGraphNodeAttribute(proteinPiana_value=interaction[0],
                                                           mem_mode = "onDemand",
                                                           piana_access = self.piana_access)
            new_node_b_attribute = PianaGraphNodeAttribute(proteinPiana_value=interaction[1],
                                                           mem_mode = "onDemand",
                                                           piana_access = self.piana_access)

            new_node_a = self.piana_graph.get_node(identifier=interaction[0], attribute=new_node_a_attribute)
            new_node_b = self.piana_graph.get_node(identifier=interaction[1], attribute=new_node_b_attribute)

            self.piana_graph.add_node(new_node_a)
            self.piana_graph.add_node(new_node_b)

            new_edge_attribute = PianaGraphEdgeAttribute(interactionPiana_value= interaction[2],
                                                         mem_mode = "onDemand",
                                                         piana_access = self.piana_access)

            new_edge = self.piana_graph.create_edge(node_id1= interaction[0],
                                                    node_id2= interaction[1],
                                                    attribute_object= new_edge_attribute)

            if new_edge[1] == 1:
                # edge is new
                self.piana_graph.add_edge(new_edge[0])
        # END OF for interaction in all_database_method_interactions:


    # -------------------------------------
    #
    #  Apply modifications/extensions to piana_graph
    # 
    # -------------------------------------
    
    def expand_piana_graph_interactions(self, expansion_type, expansion_mode, expansion_threshold, hub_threshold, exp_output_mode,
                                        output_file_object, id_type, alternative_id_types,
                                        output_tax_id,
                                        use_self_ints='yes', list_source_dbs="all", inverse_dbs="no", list_source_methods="all", inverse_methods="no"):
        """

        Expands interactions in the current piana_graph by propagating interactions to nodes "expansion_mode" from all nodes that
        have a common characteristic "expansion_type"

        "expansion_type" defines the characteristic that is used to propagate interactions between nodes
            -> valid expansion-type values are those defined in PianaGlobals.expansion_types (currently can be cog, scop (ie. scop family), interpro or ec)
            -> if two proteins share expansion-type, interactions are interpropagated

        "expansion_mode" defines to which nodes we will propagate the interactions to
             -> valid expansion-nodes values are: all (all proteins in piana_graph are expanded) or root (only root proteins are expanded)
             -> if you are looking for new interactions (predictions) for your input proteins, use root
             -> if you want to expand all the proteins in the piana_graph (partners of root proteins as well) use all
             -> root proteins are the source proteins used to build the piana_graph

        "expansion_threshold" is used to avoid propagating interactions when there are too many nodes that share the expansion type
             -> valid values are: 0 (no thresholds applied) and positive integers

        "hub_threshold" sets the maximum number of interactions a protein can have for it to be added to the piana_graph
           -> 0 is equivalent to not applying any hub threshold


        "use_self_ints" is used to avoid adding self interactions in the network
                      -> 'yes' adds all interactions
                      -> 'no' doesn't add interactions between a protein and itself
                      
        "list_source_dbs" sets the interaction databases that will be used to get interactions when building the piana_graph
           -> can be a list of dbs (eg ["dip", "string"]) or "all" (all source dbs used)

        "inverse_dbs" can be:
                       no (databases in list_source_dbs will be used to build the network)
                       yes (all databases except those in list_source_dbs will be used to build the network)

        "list_source_methods" sets the methods that will be used to get interactions when building the piana_graph
           -> can be a list of methods (eg ["y2h", "tap"]) or "all" (all methods used)

        "inverse_methods" can be:
                       no (methods in list_source_methods will be used to build the network)
                       yes (all methods except those in list_source_methods will be used to build the network)


        "exp_output_mode" sets whether new interactions are added to the piana_graph or printed to an output file
        
           -> valid exp-output-mode values are: add (add predictions to piana_graph) and print (print to output-target)
              -> 'add' will add to the piana_graph the predictions found by expansion
              -> 'print' will print to output-target (or to default results file) the list of predictions found by expansion
           -> for example, if you wanted to get predictions for root nodes using double cog expansion
              you would first use command expand-interactions with expansion-nodes=all and mode=add
              and then, another command expand-interactions with expansion-nodes=root and mode=print
              doing this "double expansion" you will be predicting interactions based on a previous expansion

           - if exp-output-mode is add, the following arguments can be ignored: leave them to blank:
           - if exp-output-mode is "print" then :

              -> "output_file_object" is the file object where the interactions will be printed


              -> "output_tax_id" restricts the species of proteins in the interactions that will be printed
                   --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
                   
              -> "id_type" is the type of code that should be used for printing proteins identifiers
                   -> valid protein-type values are those defined in PianaGlobals.valid_protein_ids

              -> the results will follow the following format (one interaction per line):

                 protein1<TAB>protein2<TAB>expansion_type<TAB>source_interactionPiana<TAB>source_proteinPiana

        """
        
        self._check_arguments(method_name= "expand_piana_graph_interactions",
                              expansion_type=expansion_type,
                              expansion_mode=expansion_mode,
                              expansion_threshold=expansion_threshold,
                              hub_threshold=hub_threshold,
                              use_self_ints=use_self_ints,
                              exp_output_mode=exp_output_mode,
                              file_object= output_file_object,                              
                              id_type=id_type,                             
                              alternative_id_types=alternative_id_types,
                              tax_id=output_tax_id,
                              list_source_dbs=list_source_dbs,
                              inverse_dbs=inverse_dbs,
                              list_source_methods= list_source_methods,
                              inverse_methods=inverse_methods)
        
        # TO DO!!! Introduce some kind of mechanism to create the expansion object without the if elif, just by
        #          using a dictionary where content are the class to be used to create the expansion object
        #          I guess I can use the expansion_types dictionary to say which class to use for each expansion
        #          type... can I write a class in a dictionary content?
        #          use apply() see email with python tips
        
        if expansion_type == "ec":
            expansion_object = ExpansionSameEC(piana_access_object = self.piana_access,
                                               use_self_ints = use_self_ints,
                                               list_source_dbs= list_source_dbs,
                                               inverse_dbs= inverse_dbs,
                                               list_source_methods= list_source_methods,
                                               inverse_methods= inverse_methods)

        elif expansion_type == "cog":
            expansion_object = ExpansionSameCog(piana_access_object = self.piana_access,
                                                use_self_ints = use_self_ints,
                                                list_source_dbs= list_source_dbs,
                                                inverse_dbs= inverse_dbs,
                                                list_source_methods= list_source_methods,
                                                inverse_methods= inverse_methods)

        elif expansion_type == "interpro":
            expansion_object = ExpansionSameInterpro(piana_access_object = self.piana_access,
                                                     use_self_ints = use_self_ints,
                                                     list_source_dbs= list_source_dbs,
                                                     inverse_dbs= inverse_dbs,
                                                     list_source_methods= list_source_methods,
                                                     inverse_methods= inverse_methods)

        elif expansion_type == "scop":
            expansion_object = ExpansionSameScop(piana_access_object = self.piana_access,
                                                 use_self_ints = use_self_ints,
                                                 list_source_dbs= list_source_dbs,
                                                 inverse_dbs= inverse_dbs,
                                                 list_source_methods= list_source_methods,
                                                 inverse_methods= inverse_methods)
            
        else:
            raise ValueError("Trying to apply expansion of unknown type %s: if you have created a new Expansion in PianaGraphExpansions, you must as well add the expansion key to PianaGlobals.expansion_types\n" %(expansion_type))
        # END OF  if expansion_type == ....

        if verbose:
            sys.stderr.write("\nApplying expansion to %s nodes using expansion type %s with output mode %s and class name %s\n" %(expansion_mode,
                                                                                                                                  expansion_type,
                                                                                                                                  exp_output_mode,
                                                                                                                                  output_tax_id))

        self.piana_graph.expand(expansion=expansion_object, expansion_mode=expansion_mode, exp_output_mode= exp_output_mode,
                                expansion_threshold= expansion_threshold, hub_threshold= hub_threshold,
                                output_target= output_file_object, id_type=id_type,
                                alternative_id_types= alternative_id_types,
                                class_name= output_tax_id)


    # -------------------------------------
    #
    #  Print Table, DOT, protein info... 
    # 
    # -------------------------------------

        
    def print_interactions(self, id_type, output_file_object, output_format, print_mode, format_mode, 
                           alternative_id_types, tax_id_value, use_self_ints="yes", list_keywords= [], intersection_dbs=None,
                           file_over_expressed= None, file_infra_expressed= None , expression_id_type= None,
			   dic_special_proteins_file= None, network_attributes_prefix= None):
        """
        Prints interactions from current piana_graph in the format chosen by the user.

        "id_type"  is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
             
        "output_file_object" is the file object (sys.stdout to print to screen) where interactions will be printed

        "output_format" is the format that will be followed for the output
           -'table': prints interactions in table format
           -'network': prints interactions in a format that can be visualized as a network

        "format_mode" sets the type of format that will be used for output

           -> valid formats for output_format 'table' are:
               - 'txt' will print flat text
               - 'html' will print html 

                  --> format_mode 'txt' for table will print a table in the format indicated 
                      in the description of command print-table-* in the template for piana 
                      configuration files: piana/code/execs/conf_files/general_template.piana_conf
 
                  --> format_mode 'html' produces an HTML table with the interactions and information 
                      associated to them
               
           -> valid formats for output_format 'network' are:
               - 'dot': uses DOT format as defined in www.graphviz.org
               - 'sif': uses SIF format as defined in www.cytoscape.org

                  --> format_dot will produce an output that can be then given to visualization programs
                      for example, neato from GraphViz, would work by:
                       $> cat output_in_dot_format | neato -Tgif -o network.gif

                  --> format_mode 'sif' will produce an output that can be imported from cytoscape
               


        "print_mode" sets which proteins will be printed
           -> "all" will print all interactions in the piana_graph
           -> "all_root" will print all interactions in the piana_graph where at least one partner is a root protein
           -> "only_root" will print only interactions between root proteins
           -> "connecting" will print only interactions between root proteins and those proteins that connect more than one root protein
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
               -> The list of values that this list can contain can be obtained by doing python piana.py or
                  looking to the PIANA reference card (see README.piana_tutorial)

             I suggest always placing md5 at the end of alternative types, so you never get a None in the output


        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "use_self_ints" determines if self interactions of proteins will be printed out
            --> valid values are 'yes' (i.e. print all interactions) and 'no' (i.e. do not print self interactions)
            (a self interaction is an interaction of a protein with itsel, as seen in homodimers)

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it highlights in red the proteins in the DOT file that contain at least one keyword in the function, description or name
           -> currently, it is not used when printing the interactions in a table
           -> If you are interested in highlighting proteins related to cancer, list_keywords could be: ['cancer', 'onco', 'carcinoma', 'tumor']
           
        "intersection_dbs" sets intersection mode, which only prints out interactions that appear in all dbs of the list being passed
            -> it can be None (no intersection mode applied) or a list of database names
            -> valid database names are those indicated in the PIANA reference card (see README.piana_tutorial)
            
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

	"network_attributes_prefix" is used to create files with information about the nodes and edges
            -> these files can then be used from cytoscape to visualize certain characteristics of the network, nodes and edges

        """

        self._check_arguments(method_name= "print_interactions",
                              file_object=output_file_object,
                              id_type=id_type,                            
                              alternative_id_types=alternative_id_types,
                              tax_id=tax_id_value,
                              use_self_ints=use_self_ints,
                              list_keywords=list_keywords,
                              intersection_dbs=intersection_dbs,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
                              dic_special_proteins_file=dic_special_proteins_file)
        
        
        self.piana_graph.output_interactions( id_type= id_type,
                                              output_target= output_file_object,
                                              output_format= output_format,
                                              filter_mode = "all",
                                              print_mode = print_mode,
                                              format_mode = format_mode,
                                              intersection_dbs = intersection_dbs,
                                              alternative_id_types= alternative_id_types,
                                              tax_id_value= tax_id_value,
                                              use_self_ints = use_self_ints,
                                              list_keywords= list_keywords,
                                              file_over_expressed= file_over_expressed,
                                              file_infra_expressed= file_infra_expressed,
                                              expression_id_type= expression_id_type,
					      dic_special_proteins_file=dic_special_proteins_file,
                                              user_protein_names = self.user_root_proteins,
					      network_attributes_prefix=network_attributes_prefix)


       
    def print_all_proteins_information(self,  id_type, output_file_object, output_mode= "compact", format_mode="txt",
                                       alternative_id_types= [], tax_id_value= 0, list_keywords= [],
                                       file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
				       dic_special_proteins_file= None):
        """
        Prints information about all the proteins in the current piana_graph
        
        "id_type" is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "output_file_object" is the file object (sys.stdout to print to screen) where interactions will be printed
        
        "output_mode" can be:
           - 'compact': all relevant information in one line
           - 'extended': all information in text paragraphs
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
           --> user must provide a list of valid easy-to-remember type names
           alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
           I suggest always placing md5 at the end of alternative types, so you never get a None in the output
           
        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
           

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it doesn't do anything
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

            
        """

        self._check_arguments(method_name= "print_all_proteins_information",
                              id_type=id_type,
                              file_object=output_file_object,
                              output_mode=output_mode,
                              format_mode=format_mode,
                              alternative_id_types=alternative_id_types,
                              tax_id=tax_id_value,
                              list_keywords=list_keywords,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file )

        self.piana_graph.output_all_proteins_information( id_type=id_type,
                                                          output_target= output_file_object,
                                                          output_mode= output_mode,
                                                          format_mode = format_mode,
                                                          filter_mode= "all",
                                                          alternative_id_types=alternative_id_types,
                                                          list_keywords= list_keywords,
                                                          tax_id_value= tax_id_value,
                                                          user_protein_names = self.user_root_proteins,
                                                          file_over_expressed= file_over_expressed,
                                                          file_infra_expressed= file_infra_expressed,
                                                          expression_id_type= expression_id_type,
							  dic_special_proteins_file=dic_special_proteins_file)
        
        
        
    def print_root_proteins_information(self,  id_type, output_file_object, output_mode= "compact", format_mode="txt",
                                        alternative_id_types= [], tax_id_value= 0, list_keywords= [],
                                        file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
					dic_special_proteins_file= None):
        """
        Prints information about root proteins in the current piana_graph
         -> a root protein is a protein that was given by the user as input (ie. protein of interest)
        
        "id_type" is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "output_file_object" is the file object (sys.stdout to print to screen) where interactions will be printed
        
        
        "output_mode" can be:
           - 'compact': all relevant information in one line
           - 'extended': all information in text paragraphs
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 
            
            
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
           --> user must provide a list of valid easy-to-remember type names
           alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
           I suggest always placing md5 at the end of alternative types, so you never get a None in the output
           
        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it doesn't do anything
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed
  
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }
          
        """

        self._check_arguments(method_name= "print_root_proteins_information",
                              id_type=id_type,
                              file_object=output_file_object,
                              output_mode=output_mode,
                              format_mode=format_mode,
                              alternative_id_types=alternative_id_types,
                              tax_id=tax_id_value,
                              list_keywords=list_keywords,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)
        
        self.piana_graph.output_root_proteins_information( id_type=id_type,
                                                           output_target= output_file_object,
                                                           output_mode= output_mode,
                                                           format_mode = format_mode,
                                                           filter_mode= "all",
                                                           alternative_id_types=alternative_id_types,
                                                           list_keywords= list_keywords,
                                                           tax_id_value=tax_id_value,
                                                           user_protein_names = self.user_root_proteins,
                                                           file_over_expressed= file_over_expressed,
                                                           file_infra_expressed= file_infra_expressed,
                                                           expression_id_type= expression_id_type,
							   dic_special_proteins_file=dic_special_proteins_file)
        
        
        
    def print_connecting_proteins_information(self,  id_type, output_file_object, output_mode= "compact", format_mode="txt",
                                              alternative_id_types= [], tax_id_value= 0, list_keywords= [],
                                              file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, 
					      dic_special_proteins_file= None):
        """
        Prints information about linkers in the current piana_graph
         -> a linker is a protein that connects two or more root proteins
        
        "id_type" is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "output_file_object" is the file object (sys.stdout to print to screen) where interactions will be printed
        
        
        "output_mode" can be:
           - 'compact': all relevant information in one line
           - 'extended': all information in text paragraphs
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 
            
            
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
           --> user must provide a list of valid easy-to-remember type names
           alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
           I suggest always placing md5 at the end of alternative types, so you never get a None in the output
               
        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it doesn't do anything
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed
     
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }
   
        """

        self._check_arguments(method_name= "print_connecting_proteins_information",
                              id_type=id_type,
                              file_object=output_file_object,
                              output_mode=output_mode,
                              format_mode=format_mode,
                              alternative_id_types=alternative_id_types,
                              tax_id=tax_id_value,
                              list_keywords=list_keywords,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)
        
        self.piana_graph.output_connecting_proteins_information(  id_type= id_type,
                                                                  output_target= output_file_object,
                                                                  output_mode= output_mode,
                                                                  format_mode = format_mode,
                                                                  filter_mode= "all",
                                                                  alternative_id_types= alternative_id_types,
                                                                  list_keywords= list_keywords,
                                                                  tax_id_value=tax_id_value,
                                                                  user_protein_names = self.user_root_proteins,
                                                                  file_over_expressed= file_over_expressed,
                                                                  file_infra_expressed= file_infra_expressed,
                                                                  expression_id_type= expression_id_type, 
								  dic_special_proteins_file=dic_special_proteins_file)
        
    def print_all_proteins_with_network_class(self,  id_type, output_file_object,
                                              alternative_id_types= [], tax_id_value= 0, 
                                              file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
                                              dic_special_proteins_file={}):
        """
        Prints all proteins in current piana_graph, indicating whether they are a root protein, a linker or a partner
         -> a root is a protein that was given as input by the user
         -> a linker is a protein that connects two or more root proteins
         -> a partner is a protein that is connected to just one root protein
        
        "id_type" is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "output_file_object" is the file object (sys.stdout to print to screen) where interactions will be printed
            
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
           --> user must provide a list of valid easy-to-remember type names
           alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
           I suggest always placing md5 at the end of alternative types, so you never get a None in the output
               
        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
   
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed
        
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        """

        self._check_arguments(method_name= "print_all_proteins_with_network_class",
                              id_type=id_type,
                              file_object=output_file_object,
                              alternative_id_types=alternative_id_types,
                              tax_id=tax_id_value,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)
        
        self.piana_graph.output_proteins_with_network_class(  id_type= id_type,
                                                              output_target= output_file_object,
                                                              alternative_id_types= alternative_id_types,
                                                              tax_id_value=tax_id_value,
                                                              user_protein_names = self.user_root_proteins,
                                                              file_over_expressed= file_over_expressed,
                                                              file_infra_expressed= file_infra_expressed,
                                                              expression_id_type= expression_id_type, 
                                                              dic_special_proteins_file=dic_special_proteins_file)
       
        
        
    def print_list_proteins_information(self, protein_list, input_id_type, output_file_object, output_id_type,
                                        alternative_id_types= [], output_mode="compact", format_mode="txt", list_keywords= [],
                                        tax_id_value=0,
                                        file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, dic_special_proteins_file= None ):
        """
        Prints information for all proteins in list "protein_list" (does not take into account the network, only these proteins)
         --> only works in compact mode to prevent the creation of enormous text files
        
        "protein_list" is the list of proteins for which you want to retrieve the information
        
        "input_id_type" is the type of code of proteins in the protein list
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "output_file_object" is the file object where the protein information will be written

        "output_id_type" is the type of code that will be used to identify proteins in the output file
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "tax_id_value" determines which species must be the proteins that will be printed
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
           
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
           --> user must provide a list of valid easy-to-remember type names
           alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
           I suggest always placing md5 at the end of alternative types, so you never get a None in the output
           
        "output_mode" can be:
           - 'extended' (multiple lines per protein to be shown directly to the screen)
           - 'compact' (one line per protein to be shown directly on the screen)
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it doesn't do anything
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed
          
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

     
        """

        self._check_arguments(method_name= "print_list_proteins_information",
                              protein_list= protein_list,
                              id_type=input_id_type,
                              file_object=output_file_object,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              output_mode=output_mode,
                              format_mode=format_mode,
                              list_keywords=list_keywords,
                              tax_id=tax_id_value,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)
        
        self.piana_graph.output_list_proteins_information(  list_proteins = protein_list,
                                                            input_id_type = input_id_type,
                                                            output_id_type= output_id_type,
                                                            output_target= output_file_object,
                                                            output_mode= output_mode,
                                                            format_mode = format_mode,
                                                            filter_mode= "all",
                                                            alternative_id_types= alternative_id_types,
                                                            list_keywords= list_keywords,
                                                            tax_id_value= tax_id_value,
                                                            key_list= self.key_list_to_use,
                                                            file_over_expressed= file_over_expressed,
                                                            file_infra_expressed= file_infra_expressed,
                                                            expression_id_type= expression_id_type,
							    dic_special_proteins_file=dic_special_proteins_file)
        self.key_list_to_use += 1


    def print_file_proteins_information(self, input_file_object, input_id_type, output_file_object, output_id_type,
                                        output_mode, format_mode="txt", list_keywords= [], alternative_id_types= [], tax_id_value=0,
                                        file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
					dic_special_proteins_file=None):
        """
        Prints information for all proteins in file "input_file_object"
         --> only works in compact mode to prevent the creation of enormous text files
        
        "input_file_object" is the file object with the protein codes for which you want to obtain the information
        
        "input_id_type" is the type of code that should be used for printing proteins identifiers
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "output_file_object" is the file object where the protein information will be written

        "output_id_type" is the type of code that will be used to identify proteins in the output file
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "tax_id_value" sets the species of the protein that will be printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
           
        "output_mode" can be:
           - 'extended' (multiple lines per protein to be shown directly to the screen)
           - 'compact' (one line per protein to be shown directly on the screen)
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it doesn't do anything
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }
        """
        
        self._check_arguments(method_name= "print_file_proteins_information",
                              another_file_object=input_file_object ,
                              id_type=input_id_type,
                              file_object=output_file_object,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              output_mode=output_mode,
                              format_mode=format_mode,
                              list_keywords=list_keywords,
                              tax_id=tax_id_value,
                              file_over_expressed=file_over_expressed,
                              file_infra_expressed=file_infra_expressed,
                              expression_id_type=expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)
        
        list_input_protein_codes = utilities.return_proteins_from_file(file_object=input_file_object, id_type= input_id_type)

        self.print_list_proteins_information(protein_list= list_input_protein_codes,
                                             input_id_type= input_id_type,
                                             output_file_object= output_file_object,
                                             output_id_type= output_id_type,
                                             alternative_id_types= alternative_id_types,
                                             output_mode= output_mode,
                                             format_mode= format_mode,
                                             list_keywords= list_keywords,
                                             tax_id_value= tax_id_value,
                                             file_over_expressed= file_over_expressed,
                                             file_infra_expressed= file_infra_expressed,
                                             expression_id_type= expression_id_type,
					     dic_special_proteins_file=dic_special_proteins_file)
        
    # -------------------------------------
    #
    #  Print specific information from piana_graph
    # 
    # -------------------------------------

    def print_shortest_route(self, protein_a_name, protein_b_name, input_id_type, output_id_type, alternative_id_types,
                             output_file_object, format_mode, tax_id_value = 0):
        """
        Prints to "output_file_object" the route from "protein_a_name" to "protein_b_name"
        

        "input_id_type" is the type of code used for protein query_protein
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "output_id_type" is the type of code that will be used to print proteins
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output
             
        "output_file_object"is the file object (sys.stdout to print to screen) where the route will be printed
        
        
        "format_mode" sets the type of format that will be used for output (Attention! Currently not working. Output is always TXT!)
               - 'txt' will print flat text
               - 'html' will print html 

        "tax_id_value" sets the species of the protein that will be printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        """

        
        self._check_arguments(method_name= "print_shortest_route",
                              protein_code=protein_a_name,
                              another_protein_code= protein_b_name,
                              id_type=input_id_type,
                              file_object=output_file_object,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              format_mode=format_mode,
                              tax_id= tax_id_value)
        
        self.piana_graph.output_shortest_route(protein_a_name= protein_a_name,
                                               protein_b_name= protein_b_name,
                                               input_id_type=input_id_type,
                                               output_id_type=output_id_type,
                                               alternative_id_types=alternative_id_types,
                                               output_file_object=output_file_object,
                                               format_mode=format_mode,
                                               tax_id_value = tax_id_value)
        

    def print_proteins_at_distance_x(self, query_protein, distance, input_id_type, output_id_type, alternative_id_types,
                                     output_file_object, format_mode, info, tax_id_value = 0):
        """
        Prints to "output_file_object" proteins from the network that are at distance "distance" from the protein "query protein"

        "query_protein" is the query protein: proteins returned will be at distance "distance" from this protein

        "distance" is the distance at which the proteins will be searched

        "input_id_type" is the type of code used for protein query_protein
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "output_id_type" is the type of code that will be used to print proteins
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output
             
        "output_file_object" is the file object (sys.stdout to print to screen) where proteins at distance X will be printed 
        
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "info" sets whether domain information associated to the proteins will be printed or not
           - 'all' will show all domain information
           - 'scop' will show SCOP domain information
           - 'cath' will show CATH domain information
           - 'no' shows no domain information
        

        "tax_id_value" sets the species of the protein that will be printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        
        """

        
        self._check_arguments(method_name= "print_proteins_at_distance_x",
                              protein_code=query_protein,
                              distance=distance,
                              id_type=input_id_type,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              file_object=output_file_object,
                              format_mode=format_mode,
                              info=info)
        
        # create the distance groups dic with keys
        distance_groups_dic = self.piana_graph.get_distance_groups( protein_ext_code= query_protein,
                                                                    input_id_type= input_id_type,
                                                                    output_id_type= output_id_type,
                                                                    alternative_id_types= alternative_id_types )


        # TO DO!!! This is not optimized at all! if the user asks for two distance groups, he will have to repeat all the process twice, which doesn't
        # make sense... all distance groups are calculated at once in get_distances()
        #  -> one possibility would be to modify get_distances so it only gets distances at the threshold fixed by the user... in a big piana_graph this
        #     would make a lot of sense.

        # TO DO!!! This is not taking into account the user_protein_names: user_protein_names = self.user_root_proteins
        #          It has to be taken into account, otherwise the protein names used might not be the ones that the user gave as root proteins

        # TO DO!!! This is a bit weird... I have to tell print_distance_group that its input type is the output type of get_distance_group...
        #          It would make more sense to first get proteins as proteinPiana and then output them in the code chosen by user
        if distance != "all":
            self.piana_graph.print_distance_group(distance_group= distance_groups_dic[distance],
                                                  input_id_type= output_id_type ,
                                                  format_mode= format_mode,
                                                  output_target= output_file_object,
                                                  info= info,
                                                  tax_id_value = tax_id_value)
        else:
            output_file_object = file(results_dir + output_name + ".1","w")
            self.piana_graph.print_distance_group(distance_group= distance_groups_dic[1],
                                                  input_id_type= output_id_type,
                                                  format_mode= format_mode,
                                                  output_target= output_file_object,
                                                  info= info,
                                                  tax_id_value = tax_id_value)
            
            output_file_object = file(results_dir + output_name + ".2","w")
            self.piana_graph.print_distance_group(distance_group= distance_groups_dic[2],
                                                  input_id_type= output_id_type,
                                                  format_mode= format_mode,
                                                  output_target= output_file_object,
                                                  info= info,
                                                  tax_id_value = tax_id_value)
            
            output_file_object = file(results_dir + output_name + ".3","w")
            self.piana_graph.print_distance_group(distance_group= distance_groups_dic[3],
                                                  input_id_type= output_id_type,
                                                  format_mode= format_mode,
                                                  output_target= output_file_object,
                                                  info= info,
                                                  tax_id_value = tax_id_value)



    def do_imotif_clustering(self, similarity_mode, minimum_score, imotif_mode, num_ints_thres, common_ints_thres, global_map, results_prefix,
                             results_dir, use_self_ints, list_source_dbs, inverse_dbs,  list_source_methods, inverse_methods, hub_threshold,
                             input_id_type, input_tax_id):
        """
        perform imotif clustering

        a detailed description of this arguments can be read on piana/code/execs/conf_files/general_template.piana_conf
        
        """
        
        def _get_similarity(similarity_mode, set_neighbours_1, set_neighbours_2 ):
            """
            returns similarity score between two nodes
            """

            if similarity_mode == "num_ints" or similarity_mode == "min_per" or similarity_mode == "max_per":

                if verbose_prots_imotifs_iteration:
                    sys.stderr.write("calc_%s." %(similarity_mode))

                num_partners_1= float(len(set_neighbours_1))
                num_partners_2= float(len(set_neighbours_2))

                if num_partners_1 < num_ints_thres or num_partners_2 < num_ints_thres:
                    # if any of the two clusters has less than the threshold interaction number, set similarity to 0
                    #        --> ie. has less than num_ints_thres known interactions
                    # 
                    similarity_value = 0
                else:
                    num_common_partners = len(set_neighbours_1.intersection(set_neighbours_2))

                    if num_common_partners < common_ints_thres:
                        # if there are less than 5 interactions in common, leave similarity to 0
                        similarity_value = 0
                    else:
                        try:
                            if similarity_mode == "min_per":
                                similarity_value = 100*num_common_partners / min(num_partners_1, num_partners_2)
                            elif similarity_mode == "max_per":
                                similarity_value = 100*num_common_partners / max(num_partners_1, num_partners_2)
                            elif similarity_mode == "num_ints":
                                similarity_value = num_common_partners

                        except:
                            similarity_value = 0
                    # END OF else: (if num_common_partners < 5:)
                # END OF else: (if num_partners_1 < 5 or num_partners_2 < 5:)
            else:
                raise ValueError("This similarity mode (%s) is not implemented!!" %(similarity_mode))

            return similarity_value

        # starts do_imotif_clustering
        
        # get the root protein
        root_proteins = self.piana_graph.get_root_node_ids()
        dic_root_proteins_g2 = {}

        if len(root_proteins) == 1:
            root_protein = root_proteins[0]
            # to speed up things afterwards, get now the list of proteins in G2
            #   -> we will only create multiple IRs for the root proteins and those in G2
            root_protein_g2 = self.piana_access.get_all_g2_partners(proteinPiana_value= root_protein,
                                                                    use_self_ints = use_self_ints,
                                                                    list_source_dbs=list_source_dbs , inverse_dbs=inverse_dbs,
                                                                    list_source_methods=list_source_methods , inverse_methods=inverse_methods,
                                                                    threshold = hub_threshold)
            for one_root_g2 in root_protein_g2:
                dic_root_proteins_g2[one_root_g2] = None
            
        else:
            root_protein = "multiple"

        if global_map:

            # if user set something different from None as global_map, then we are going to do the following
            #
            #   1. build the graph at depth 1 (we only want direct partners of the root protein)
            #   2. print headers of file *.results
            #        description\tprotein=root_protein\tlabel=0\tsimilarity=0\n
            #        share\tcluster_global\troot_protein\t0
            #
            #           -> share line is written so that cluster_global is considered as a cluster
            #              with more than one protein in it when doing the train_and_test_cir
            #
            #   3. For each partner of root (ie. for each interaction)
            #
            #      print line describing the interacion
            #         int\tcluster_global\tcluster_globalX\troot_protein\tpartner\n
            #
            #

            # 1.
            piana_graph_depth1 = PianaGraph("Graph depth1", piana_access_object= self.piana_access, memory_mode = self.memory_mode)
            piana_builder_depth1 = PianaGraphBuilder(piana_access_object=self.piana_access,
                                                     depth=1,
                                                     hub_threshold= hub_threshold,
                                                     use_self_ints=use_self_ints,
                                                     list_protein_codes=root_proteins,
                                                     id_type=input_id_type ,
                                                     tax_id_value=input_tax_id,
                                                     list_source_dbs= list_source_dbs,
                                                     inverse_dbs= inverse_dbs,
                                                     list_source_methods= list_source_methods,
                                                     inverse_methods= inverse_methods,
                                                     piana_secondary_access_object= None)

            piana_graph_depth1.build_graph(piana_builder_depth1)



            # 2.
            this_results_file_fd= file(global_map + "global_map." + str(root_protein) + "_0.results", "w")
            # write the description header
            this_results_file_fd.write("description\tprotein=%s\tlabel=0\tsimilarity=0\n" %(root_protein))

            # write a 'share' line to make sure that all clusters in this file 'appear to have' more than one protein
            this_results_file_fd.write("share\tcluster_name\t%s\t0\n" %(root_protein))



            # 3.
            root_node = piana_graph_depth1.get_node(identifier= root_protein, get_mode="error")

            i = 0
            for one_neighbour_id in root_node.get_neighbour_ids():
                this_results_file_fd.write("int\tcluster_name-cluster_name%s\t%s\t%s\n" %(i,
                                                                                          root_protein,
                                                                                          one_neighbour_id))
                i += 1

            # END OF for one_neighbour in root_node.get_neighbour_ids():


            this_results_file_fd.close()
            if verbose_command_sequence:
                sys.stderr.write(" Global mapping Done!!\n")

            return None

        # END OF if global_map:

        piana_file_fd = file(results_dir + "/" + results_prefix + "." + "ppi_network.dot", "w")
        self.piana_graph.output_dot_file(output_target=piana_file_fd, use_alternative_id="no")

        # --------------------------------------------------------------------------------------------------------
        # 1. Create the Interacting Protein Groups Graph (which is just a GraphCluster where clusters have proteins)
        # --------------------------------------------------------------------------------------------------------
        #
        #  The interesting thing is that since the similarity is calculated by looking to the proteins inside each
        #  cluster, the PPI graph needs to have depth 3 (to know all the partners for each protein being compared
        #  to the root protein) but the IPG graph only needs depth 2 (because no clusters at depth 3 are ever going
        #  to be clustered with clusters containing a root protein
        #
        #     --> this is only true when using max_per, min_per or combined_per
        #         When using num_ints, the detph can be 2 (change it in the configuration file create_imotif_results.piana_conf)
        #

        if verbose_prots_imotifs_shallow:
            sys.stderr.write("initializing iMotif clustered network -- ")


        # IPG Graph needs to be initialized from the depth 2 network of the root protein
        if verbose_prots_imotifs:
            sys.stderr.write("\nBuilding depth2 piana graph with parameters:")
            sys.stderr.write(" - hub_threshold=%s" %(hub_threshold))
            sys.stderr.write(" - use_self_ints=%s" %(use_self_ints))
            sys.stderr.write(" - list_protein_codes=%s" %(root_proteins))
            sys.stderr.write(" - id_type=%s" %(input_id_type))
            sys.stderr.write(" - tax_id_value=%s" %(input_tax_id))
            sys.stderr.write(" - list_source_dbs=%s" %(list_source_dbs))
            sys.stderr.write(" - inverse_dbs=%s" %(inverse_dbs))
            sys.stderr.write(" - list_source_methods=%s" %(list_source_methods))
            sys.stderr.write(" - inverse_methods=%s\n" %(inverse_methods))

        piana_graph_depth2 = PianaGraph("Graph depth2", piana_access_object= self.piana_access, memory_mode = self.memory_mode)
        piana_builder_depth2 = PianaGraphBuilder(piana_access_object=self.piana_access,
                                                 depth=2,
                                                 hub_threshold= hub_threshold,
                                                 use_self_ints=use_self_ints,
                                                 list_protein_codes=root_proteins,
                                                 id_type=input_id_type ,
                                                 tax_id_value=input_tax_id,
                                                 list_source_dbs= list_source_dbs,
                                                 inverse_dbs= inverse_dbs,
                                                 list_source_methods= list_source_methods,
                                                 inverse_methods= inverse_methods,
                                                 piana_secondary_access_object= None)

        piana_graph_depth2.build_graph(piana_builder_depth2)

        ipg_graph = GraphCluster(graph_id="G")
        ipg_graph.initialize_from_graph(graph=piana_graph_depth2 )

        set_original_node_ids = Set(ipg_graph.get_node_ids_list())


        file_alternative_fd = file(results_dir + "/" + results_prefix + "." + "ipg_network.initial.dot", "w")
        ipg_graph.output_dot_file(output_target=file_alternative_fd, use_alternative_id="yes")


        proteins_same_cluster = {}  # used to know which nodes coexist in the same cluster
                                    # so they are not used to create a new cluster
                                    #   -> only applies to original nodes (those that only contain 1 protein)
                                    # follows structure:
                                    #        { proteinPiana_1: {proteinPiana_sharing_cluster:None, proteinPiana_sharing_cluster, ..},
                                    #          proteinPiana_2: {proteinPiana_sharing_cluster:None, proteinPiana_sharing_cluster, ..},
                                    #          .......................
                                    #        }


        previous_nodes_similarities = {}  # used to keep those similarities that have already been calculated (between nodes of IPG graph)
                                          # follows structure:
                                          #       { key : similarity score,
                                          #         key : similarity score,
                                          #         .....................
                                          #       }
                                          # where key is a string lower_node_id.higher_node_id

        previous_prots_similarities = {}  # used to keep those similarities that have already been calculated (between proteins of PPI graph)
                                          # follows structure:
                                          #       { key : similarity score,
                                          #         key : similarity score,
                                          #         .....................
                                          #       }
                                          # where key is a string lower_proteinPiana.higher_proteinPiana

        node_pairs_to_skip = {}  # used to keep those node pairs that should be left with similarity 0, because
                                 # restrictions apply to them (eg. protein in original cluster already appears in the other cluster)
                                 # folows structure:
                                 #       { key : None,
                                 #         key : None,
                                 #         .....................
                                 #       }
                                 # where key is a string lower_node_id.higher_node_id


        clustering_level = 0

        if verbose_command_sequence:
            sys.stderr.write(" -- starting clustering! \n")

        while(1):
            # while the loop is not breaked by the stop condition, continue clustering
            if verbose_prots_imotifs:
                sys.stderr.write("\n ------------------------------------------------------------------------------\n")
                sys.stderr.write("               NEW CLUSTERING LEVEL %s\n" %(clustering_level))
                sys.stderr.write(" ------------------------------------------------------------------------------\n")
                sys.stderr.write(" proteins same cluster: %s\n" %(proteins_same_cluster))

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("lev_%s." %(clustering_level))


            # ------------------------------------------------------------
            # 2. obtain the similarity matrix at this clustering point
            # ------------------------------------------------------------
            if verbose_prots_imotifs_shallow:
                sys.stderr.write("s.")

            node_list= ipg_graph.get_node_object_list()

            number_nodes = len(node_list)

            similarity_matrix= numarray.zeros( ( number_nodes, number_nodes), numarray.Float32) 
            max_value=0
            node_id_vs_matrix_index= {} # keeps as keys the node ids and as content the matrix index for this node id (for this iteration)
            positions_to_cluster= []

            for i in range(number_nodes):
                # for each matrix position (i, j) calculate the similarity
                if verbose_prots_imotifs_iteration:
                    sys.stderr.write("%sof%s." %(i, number_nodes))

                node_1 = node_list[i]

                set_neighbours_1 = Set(node_1.get_neighbour_ids())

                node_id_1 = node_1.get_node_id()

                node_id_vs_matrix_index[node_id_1] = i

                node_attribute_1= node_1.get_node_attribute_object()
                protein_nodes_1 = node_attribute_1.get_list_elements()
                dic_protein_ids_1 = {}
                for one_protein_node in protein_nodes_1:
                    dic_protein_ids_1[one_protein_node.get_node_id()] = None

                protein_ids_1 = Set(dic_protein_ids_1.keys())

                if node_1.is_original():
                    protein_1= protein_ids_1.copy().pop() # there is only one id... it is an original node

                for j in range(i+1, number_nodes):
                    node_2 = node_list[j]

                    set_neighbours_2 = Set(node_2.get_neighbour_ids())

                    node_id_2 = node_2.get_node_id()

                    node_attribute_2= node_2.get_node_attribute_object()
                    protein_nodes_2 = node_attribute_2.get_list_elements()
                    dic_protein_ids_2 = {}
                    for one_protein_node in protein_nodes_2:
                        dic_protein_ids_2[one_protein_node.get_node_id()] = None

                    protein_ids_2 = Set(dic_protein_ids_2.keys())

                    if node_2.is_original():
                        protein_2 = protein_ids_2.copy().pop()  # there is only one id... it is an original node


                    #
                    # now, calculate the similarity between these two nodes
                    #


                    if node_id_1 < node_id_2:    nodes_key = "%s.%s" %(node_id_1, node_id_2)
                    else:                        nodes_key = "%s.%s" %(node_id_2, node_id_1)

                    if verbose_prots_imotifs:
                        sys.stderr.write("calculating similarity for node %s (proteins %s) and node %s (proteins %s)\n" %(node_1.get_node_id(),
                                                                                                                          protein_ids_1,
                                                                                                                          node_2.get_node_id(),
                                                                                                                          protein_ids_2,
                                                                                                                          ))

                    if node_pairs_to_skip.has_key(nodes_key):
                        continue

                    # do these checkings first, to make sure we don't consider for this clustering level
                    # things that have already been considered (ie. don't use clusters that already have
                    # been used to create a new cluster)
                    if node_1.is_original() and node_2.is_original():
                        # check if these proteins already coexist in one cluster
                        if proteins_same_cluster.has_key(protein_1):
                            if proteins_same_cluster[protein_1].has_key(protein_2):
                                # these two nodes already coexist in a cluster: do not process them again (ie. leave similarity to 0)
                                node_pairs_to_skip[nodes_key] = None
                                continue
                    # END OF if node_1.is_original() and node_2.is_original():

                    elif node_1.is_original() or node_2.is_original():
                        # if at least one of the two is original, check if the cluster already has the protein of the original node
                        if protein_ids_1.intersection(protein_ids_2):
                            # a cluster that already has the protein of the original node should not be compared (ie. leave similarity to 0)
                            node_pairs_to_skip[nodes_key] = None
                            continue
                    # END OF elif node_1.is_original() or node_2.is_original():

                    if previous_nodes_similarities.has_key(nodes_key):
                        # if this similarity was already calculated, use it
                        similarity_value = previous_nodes_similarities[nodes_key]
                    else:
                        # if it is a similarity between new nodes, calculate it

                        similarity_value = 0 # by default, similarity value is 0

                        #
                        # depending on the cir mode, decide whether we should consider clustering this pair or not
                        # 
                        do_clustering = False
                        if imotif_mode == "roots" and (dic_protein_ids_1.has_key(root_protein) or dic_protein_ids_2.has_key(root_protein)):
                            # in imotif_mode roots,only clusters that contain a root protein are used to form a new cluster
                            do_clustering = True

                        elif imotif_mode == "all":
                            # in imotif_mode all, all clusters can be fused, regardless of them having a root protein or not
                            # moreover, the fact that CIRs will have all proteins that should be there, we will be
                            # able to print CIR-CIR interactions afterwards
                            do_clustering = True


                        if do_clustering:
                            # we compare the neighbours for each cluster to calculate the similarity value between clusters
                            #
                            #
                            similarity_value = _get_similarity(similarity_mode=similarity_mode,
                                                               set_neighbours_1=set_neighbours_1,
                                                               set_neighbours_2=set_neighbours_2 )

                            if verbose_prots_imotifs:
                                sys.stderr.write("  -> score between node_id_1 %s (%s) and node_id_2 %s (%s) is %s\n" %(node_id_1, protein_ids_1,
                                                                                                                        node_id_2, protein_ids_2,
                                                                                                                        similarity_value))
                        # END OF if do_clustering:

                        previous_nodes_similarities[nodes_key] = similarity_value # keep this similarity to be used in the future

                    # END OF else: (if previous_nodes_similarities.has_key(nodes_key):)

                    if max_value < similarity_value:
                        # if this similarity is a new maximum, reset the list of positions to cluster and set this as new max_value
                        max_value = similarity_value
                        positions_to_cluster = [(i,j)]

                    elif max_value == similarity_value:
                        # if this similarity is the same as the max value, add the position to cluster to list
                        positions_to_cluster.append( (i,j) )
                    else:
                        pass

                    if similarity_value:
                        similarity_matrix[i][j] = similarity_value # i is always lower than j
                        #print "  -> sim between \nnode_id_1=%s (m_inx=%s) (%s)\nnode_id_2=%s (m_inx=%s) (%s)\nscore=%s (%s)" %(node_id_1,
                                                                                                                               #i,
                                                                                                                               #protein_ids_1,
                                                                                                                               #node_id_2,
                                                                                                                               #j,
                                                                                                                               #protein_ids_2,
                                                                                                                               #similarity_value,
                                                                                                                               #set_neighbours_1.intersection(set_neighbours_2))
                # END OF for j in range(i+1, number_nodes):
            # END OF for i in range(number_nodes):

            #print "----------------------------------------------------------\n"

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("m=%.0f." %(max_value))

            # ----------------------------------------------------------------------------------
            # 3. printing the network and clusters for this level (before creating new clusters)
            # ----------------------------------------------------------------------------------
            clustering_level_file_fd = file(results_dir + "/" + results_prefix + "." + similarity_mode + ".level_%s.dot" %clustering_level, "w")
            ipg_graph.output_dot_file(output_target=clustering_level_file_fd, use_alternative_id="yes")
            clustering_level_file_fd.close()

            this_level_composition_file_fd = file(results_dir + "/" + results_prefix + "." + similarity_mode + ".level_%s.cluster_composition" %clustering_level, "w")
            ipg_graph.print_cluster_composition(output_target= this_level_composition_file_fd)
            this_level_composition_file_fd.close()

            this_level_ints_file_fd = file(results_dir + "/" + results_prefix + "." + similarity_mode + ".level_%s.cluster_ints" %clustering_level , "w")
            ipg_graph.print_cluster_interactions(output_target= this_level_ints_file_fd)
            this_level_ints_file_fd.close()


            results_level_file_fd = file(results_dir + "/" + results_prefix + "." + similarity_mode + ".level_%s.results" %clustering_level, "w")
            results_level_file_fd.write("description\tprotein=%s\tlabel=%s\tsimilarity=%s\n" %(root_protein, clustering_level, max_value))
            ipg_graph.print_proteins_same_cluster(output_target=results_level_file_fd, root_protein=root_protein)
            ipg_graph.print_proteins_interactions(output_target=results_level_file_fd, root_protein=root_protein, similarity_threshold= max_value)

            if imotif_mode == "all":
                # CIR-CIR interactions can only be printed when we are in imotif_mode all
                # Why? Because in imotif_mode root, we will never have a case in which we can infer which
                #      CIR would be interacting with a given CIR by looking to the proteins that interact with it, because
                #      there would only be CIRs with root proteins. Read comments on print_cir_cir_ints to better understand
                #      this.
                #      However, in mode "all" and "root" we are also printing interactions between proteins. In a less strict
                #      definition of what are the interactions for a CIR, we can also say (and that's what we do in method
                #      print_proteins_interactions) that a CIR interacts with all original clusters partners that have been
                #      determined to be shared by all the CIR members
                ipg_graph.print_cir_cir_ints(output_target=results_level_file_fd, root_protein=root_protein, output_id_type=output_id_type)

            results_level_file_fd.close()


            clustering_level += 1


            # CHECK STOP CONDITION
            if max_value < minimum_score:
                # break the while if similarity score lower than threshold
                #print "max value %s is lower than minimum score %s" %(max_value, minimum_score)
                break


            # -------------------------------------------------------------
            # 4. use cluster with max similarity scores to create new clusters
            # -------------------------------------------------------------

            if verbose_prots_imotifs:
                sys.stderr.write( "positions to cluster are: %s\n" %(positions_to_cluster))

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("len=%s." %(number_nodes))

            used_nodes = {}   # node_ids of nodes that have already been used to generate a new cluster

            new_clusters = [] # list of new clusters that have been generated

            clustered_nodes_dic = {}  # dic that keeps as keys the current node ids and as values the
                                      # cluster object for next level graph
                                      #   -> used to add several clusters into the same one of the new level
                                      #      All clusters with the max_value that are transitive between
                                      #      them (ie. all of them have max_value in between) are clustered
                                      #      into the same cluster in just one clustering step


            correspondences_new_clusters = {}   # dic that keeps the node ids that were used to generate a new cluster
                                                #   follows structure:
                                                #       { new_cluster_id : [node id 1, node id 2, ...]
                                                #         new_cluster_id : [node id 3, node id 4, ...]
                                                #         ...............
                                                #       }

            for pair_i_j in positions_to_cluster:
                # for each position to cluster, create a new cluster from the clusters that are 'similar'
                #   -> unless restrictions apply

                i = pair_i_j[0]
                j = pair_i_j[1]

                i_node_id = node_list[i].get_node_id()
                j_node_id = node_list[j].get_node_id()

                #print "%s and %s are set to be clustered: checking" %(i_node_id, j_node_id)

                protein_nodes_i= node_list[i].get_node_attribute_object().get_list_elements()
                protein_ids_i = []
                for one_protein_node in protein_nodes_i:
                    protein_ids_i.append(one_protein_node.get_node_id())



                protein_nodes_j = node_list[j].get_node_attribute_object().get_list_elements()
                protein_ids_j = []
                for one_protein_node in protein_nodes_j:
                    protein_ids_j.append(one_protein_node.get_node_id())


                if used_nodes.has_key(j_node_id) and used_nodes.has_key(i_node_id):
                    # skip nodes when both have already been used to create a cluster
                    if verbose_prots_imotifs:
                        sys.stderr.write("skipping fusion cluster %s with %s because both were already fused somewhere else\n" %(i_node_id,
                                                                                                                                 j_node_id))

                    #print "skipping fusion %s with %s: both were already fused somewhere else" %(i_node_id,
                                                                                                 #j_node_id)
                    continue


                elif used_nodes.has_key(j_node_id) or used_nodes.has_key(i_node_id):
                    # one of the two nodes was already used to create a new cluster:
                    #   - if the new node also has max_score with the previous nodes assigned to that cluster, place it on that cluster
                    #   - otherwise, do not add it


                    if verbose_prots_imotifs:
                        sys.stderr.write("testing if a third cluster is added to a new cluster created in this iteration\n")


                    if used_nodes.has_key(j_node_id):
                        # i is the cluster that is a candidate for being added to the already created cluster
                        if verbose_prots_imotifs:
                            sys.stderr.write("i:(third) inserting %s in used_nodes\n" %(i_node_id))
                        to_add_node_id =  i_node_id
                        to_add_node = node_list[i]
                        to_add_index = i
                        to_add_proteins = protein_ids_i
                        already_added_index = j
                        already_added_node_id = j_node_id
                    else:
                        # j is the cluster that is a candidate for being added to the already created cluster
                        if verbose_prots_imotifs:
                            sys.stderr.write("j:(third) inserting %s in used_nodes\n" %(j_node_id))
                        to_add_node_id = j_node_id
                        to_add_node = node_list[j]
                        to_add_index = j
                        to_add_proteins = protein_ids_j
                        already_added_index = i
                        already_added_node_id = i_node_id
                    # END OF else: (if used_nodes.has_key(j_node_id):)

                    # Now, check if the similarity between to_add_node_id and the cluster previuosly created in this run
                    #   -> if the similarity is the same as the one between (i,j), then, proceed with joining the node
                    #   -> if the similarity is different, do not fuse the clusters: we are probably looking to a node
                    #      that will belong to two different clusters.
                    # How do I know the similarity between to_add_node_id and the other cluster? Te other cluster does not
                    # appear in the similarity matrix... because it was created in this iteration. But in correspondences_new_clusters
                    # I have a list of all the clusters that have been used to form this new cluster. Therefore, I can look
                    # for the similarity between to_add_node_id and all the clusters that were used to form the new cluster,
                    # and if any of those similarities is below the max_value, then I should avoid adding to_add_node_id to
                    # the cluster
                    keep_clustering = 1

                    # get the cluster where already_added_index was placed...
                    previous_cluster= clustered_nodes_dic[already_added_index]

                    #print "testing if %s can be placed in new cluster formed by %s" %(to_add_node_id,
                                                                                      #correspondences_new_clusters[previous_cluster.get_node_id()]
                                                                                      #)

                    for one_fusioned_node_id in correspondences_new_clusters[previous_cluster.get_node_id()]:

                        # get the matrix index corresponding to one_fusioned_node_id
                        one_fusioned_matrix_id = node_id_vs_matrix_index[one_fusioned_node_id]
                        one_fusioned_node = node_list[one_fusioned_matrix_id]
                        #print "           --> matrix id for %s is %s" %(one_fusioned_matrix_id, one_fusioned_node_id)

                        # check which one is lower one_fusioned_node_matrix_index or to_add_index
                        if to_add_index <= one_fusioned_matrix_id:
                            lower_index = to_add_index
                            higher_index= one_fusioned_matrix_id
                        else:
                            lower_index = one_fusioned_matrix_id
                            higher_index= to_add_index

                        if not similarity_matrix[lower_index][higher_index]:
                            # if the value in the similarity matrix for this pair has not been calculated
                            # (because no one of them was root, or because ...) then calculate it
                            #    -> it could be that the value was 0, and we would recalculate the value... but no big deal
                            #    -> since in the similarity matrix we do not calculate all values, but here we need
                            #       to know if two clusters are similar before putting them together, we need to
                            #       calculate this specific value

                            similarity_matrix[lower_index][higher_index] = _get_similarity(similarity_mode=similarity_mode,
                                                                                           set_neighbours_1= Set(one_fusioned_node.get_neighbour_ids()),
                                                                                           set_neighbours_2= Set(to_add_node.get_neighbour_ids()))
                        # END OF if not similarity_matrix[lower_index][higher_index]:

                        ##print "   --> checking sim between idx %s (id %s) and idx %s (id %s): %s" %(to_add_index, to_add_node_id,
                                                                                                    #one_fusioned_matrix_id,one_fusioned_node_id,
                                                                                                    #similarity_matrix[lower_index][higher_index])
                        if similarity_matrix[lower_index][higher_index] < max_value:
                            # if there is a previous node in the cluster with a similarity lower than the max_value
                            # with respect to the node we want to cluster, skip it: this node will create its own CIR
                            # instead of being clustered in this one
                            keep_clustering= 0
                            #print "   --> max_value (%s) not respected between the to_add_node and the new cluster (sim=%s)" %(max_value,
                                                                                                                               #similarity_matrix[lower_index][higher_index])
                            break
                    # END OF for one_fusioned_node_id in correspondences_new_clusters[previous_cluster.get_node_id()]:

                    if not keep_clustering:
                        continue

                    # the node is going to be added to the cluster already created in this iteration

                    #print "   --> adding the node"
                    # update nodes that have been 'used' to form a new cluster in this iteration
                    used_nodes[to_add_node_id] = None

                    # update the dic with cluster for this index_to_add
                    clustered_nodes_dic[to_add_index] = previous_cluster  

                    # add to previous_cluster the elements of to_add_node_id
                    to_add_node_list_elements = to_add_node.get_node_attribute_object().get_list_elements()
                    previous_cluster.get_node_attribute_object().add_element_list(to_add_node_list_elements)

                    # update correspondences and proteins_same_cluster

                    # -> get proteins that appear in the previous cluster
                    previous_elements = previous_cluster.get_node_attribute_object().get_list_elements()

                    protein_ids_previous = []
                    for one_protein_node in previous_elements:
                        protein_ids_previous.append(one_protein_node.get_node_id())

                    for one_previuos_protein in protein_ids_previous:
                        for one_to_add_protein in to_add_proteins:
                            if proteins_same_cluster.has_key(one_previuos_protein): proteins_same_cluster[one_previuos_protein][one_to_add_protein] = None
                            else:                                            proteins_same_cluster[one_previuos_protein] = {one_to_add_protein: None}

                            if proteins_same_cluster.has_key(one_to_add_protein): proteins_same_cluster[one_to_add_protein][one_previuos_protein] = None
                            else:                                            proteins_same_cluster[one_to_add_protein] = {one_previuos_protein: None}
                        # END OF for one_to_add_protein in protein_ids_j:
                    # END OF for one_previuos_protein in protein_ids_i:

                    correspondences_new_clusters[previous_cluster.get_node_id()].append(to_add_node_id)
                    continue
                else:
                    # none of the nodes has been previosly used to create a new cluster
                    # create a new cluster and add it to ipg_graph

                    # used_nodes keeps track of which nodes have already been used in this iteration of the clustering
                    if verbose_prots_imotifs:
                        sys.stderr.write("inserting %s and %s in used_nodes\n" %(i_node_id, j_node_id))

                    #print "both clusters are new: fusion them"

                    used_nodes[i_node_id]= None
                    used_nodes[j_node_id]= None

                    ##print "checking if node id %s (orig=%s) and node id %s (orig=%s) have to update proteins_same_cluster" %(i_node_id,
                    #                                                                                                         node_list[i].is_original(),
                    #                                                                                                         j_node_id,
                    #                                                                                                         node_list[j].is_original())

                    # proteins_same_cluster keeps track of which original nodes already co-exist in the same cluster
                    if verbose_prots_imotifs:
                        sys.stderr.write("updating proteins_same_cluster for %s and %s\n" %(protein_ids_i, protein_ids_j))

                    for one_i_protein in protein_ids_i:
                        for one_j_protein in protein_ids_j:
                            if proteins_same_cluster.has_key(one_i_protein): proteins_same_cluster[one_i_protein][one_j_protein] = None
                            else:                                        proteins_same_cluster[one_i_protein] = {one_j_protein: None}

                            if proteins_same_cluster.has_key(one_j_protein): proteins_same_cluster[one_j_protein][one_i_protein] = None
                            else:                                        proteins_same_cluster[one_j_protein] = {one_i_protein: None}
                    # END OF for one_i_protein in protein_ids_i:

                    clustered_node= ipg_graph.create_grouped_node(node_id1= node_list[i].get_node_id(),
                                                                  node_id2= node_list[j].get_node_id(),
                                                                  old_graph=ipg_graph)

                    clustered_nodes_dic[i] = clustered_node 
                    clustered_nodes_dic[j] = clustered_node
                    new_clusters.append(clustered_node)
                    correspondences_new_clusters[clustered_node.get_node_id()] = [i_node_id, j_node_id]

                    if verbose_prots_imotifs:
                        sys.stderr.write("two clusters (%s, %s) being joined in cluster %s with max value %s\n" %(i_node_id, j_node_id,
                                                                                                                  clustered_node.get_node_id(),max_value))
                        sys.stderr.write("proteins in one cluster %s are %s\n" %(i_node_id, protein_ids_i))
                        sys.stderr.write("proteins in one cluster %s are %s\n" %(j_node_id, protein_ids_j))


            # END OF for pair_i_j in positions_to_cluster:


            if verbose_prots_imotifs:
                sys.stderr.write("New clusters have been created!\n")
                for clustered_node in new_clusters:
                    sys.stderr.write("------\n")
                    for element in clustered_node.get_node_attribute_object().get_list_elements():
                        sys.stderr.write("clustered node %s contains protein %s\n" %(clustered_node.get_node_id(),
                                                                        element.get_node_id()))
                # END OF for clustered_node in new_clusters :


            # ---------------------------------------------
            # 5. Update the IPG graph with the new clusters
            # ---------------------------------------------

            # Now, add clusters in new_clusters to the ipg_graph. Add as well edges between this node and the intersection of nodes
            #  that were interacting with nodes that have been used to create this new cluster

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("a.")


            # for each new_cluster, find interactions that must be associated to it and add them to graph (add as well the new_cluster)
            for one_new_cluster in new_clusters:

                ipg_graph.add_node(one_new_cluster)

                # find the intersection of the sets of partners of all nodes that have been used to generate this one_new_cluster

                # - start with all partners for first node...
                first_node = ipg_graph.get_node(identifier=correspondences_new_clusters[one_new_cluster.get_node_id()][0] , get_mode="error")
                set_partners_first_node = Set(first_node.get_neighbour_ids())
                neighbours_intersection = copy.deepcopy(set_partners_first_node)

                for correspondent_node_id in correspondences_new_clusters[one_new_cluster.get_node_id()][1:]:
                    # for each previous node that has been used to generate this new cluster, find its edges and add them as well for one_new_cluster

                    another_node = ipg_graph.get_node(identifier=correspondent_node_id , get_mode="error")
                    set_partners_another_node = Set(another_node.get_neighbour_ids())
                    neighbours_intersection.intersection_update(set_partners_another_node)
                # END OF for correspondent_node_id in correspondences_new_clusters[one_new_cluster.get_node_id()][1:]:

                # only creating edges between the new cluster and the original nodes (non-original nodes cannot interact between
                #                                                                     themselves)
                original_neighbours_intersection = neighbours_intersection.intersection(set_original_node_ids)

                # before (by error?) it was like this: for one_neighbour_id in neighbours_intersection:
                for one_neighbour_id in original_neighbours_intersection:
                    # add each particular edge of the previous nodes that generated this new cluster

                    new_edge_attribute = GraphEdgeAttribute() # empty attribute: nothing to do with it at the moment
                    new_edge= GraphEdge(node1_id=one_neighbour_id, node2_id= one_new_cluster.get_node_id(),
                                        attribute_object= new_edge_attribute, graph=ipg_graph)
                    ipg_graph.add_edge(new_edge)
                # END OF for one_neighbour_id in neighbours_intersection:
            # END OF for one_new_cluster in new_clusters:


            if verbose_prots_imotifs:
                sys.stderr.write("REMOVING CLUSTERS THAT WERE USED TO CREATE A NEW CLUSTER (when they are not original)\n")
                sys.stderr.write("number of new clusters created: %s\n" %(len(new_clusters)))

            # --------------------------------------------------------------------
            # 6. remove non-original nodes that have been fused into a new cluster
            # --------------------------------------------------------------------
            #   if a non-original cluster A is has been used to create a new cluster B (by fusing it with another cluster), A
            #   must be removed from the graph

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("rm.")

            for one_used_node_id in used_nodes:
                # find to which previous cluster it was associated and remove the cluster if it is not original
                one_used_node = ipg_graph.get_node(identifier=one_used_node_id, get_mode="error")

                if one_used_node.is_original():
                    # delete the used node unless it is original
                    continue


                if verbose_prots_imotifs:
                    sys.stderr.write("removing used node %s\n" %(one_used_node_id))
                ipg_graph.rm_node(node_object= one_used_node )

            # END OF for one_used_node_id in used_nodes:

            if verbose_prots_imotifs_shallow:
                sys.stderr.write("x--")

        # END OF while(1)


        #after_file_alternative_fd = file("ipg_network.clustering_finished.dot", "w")
        #ipg_graph.output_dot_file(output_target=after_file_alternative_fd, use_alternative_id="yes")


    
    


    def create_go_clustered_network(self, output_target=None, term_type=None, score_threshold=None, sim_mode=None, level_threshold=None,
                                    distance_threshold=None, rep_term=None, print_id=None):
        """
        Creates a network of GO (Gene Ontology) terms from the piana graph
        Then, clusters the Go network using parameters provided
        Finally,  prints the clustered go network 
        
        "output_target" is the file object where the clustered network in DOT format will be printed
        
        
        - "term_type" sets the kind of GO terms that will be used for the clustering.
          -> term-type can be "molecular_function", "biological_process" or "cellular_component"
        
        - "score_threshold" is the lowest score obtained by the similarity function allowed for continuing the clustering
          -> can be any real number from 0 to 100 (0 will group all proteins, 100 will not group any proteins). To obtain a relevant clustered network
             use score thresholds between 0.1 and 1
    
        - "sim_mode" sets how to calculate distances between two clusters
           - "random" takes a random element from each cluster and evaluates similarity between them
           - "min" takes the minimal distance between elements of each cluster
           - "max" takes the maximal distance between elements of each cluster
           - "average" takes the average distance between all elements of each cluster
    
        - "level_threshold" is the lowest level of the go term in the cluster allowed for continuing the clustering
           -> GO is a hierarchy organized from a initial root level (ie. 0) that increasingly makes more specific the terms. 
              Therefore, the higher the level used the less clustering will be performed. To obtain a relevant clustered network 
              use level thresholds between 1 and 3. It all depends on how general you want to be in the interpretation of the network.
    
        - "distance_threshold" is the maximum distance allowed between two proteins in order to be clustered
           -> can be any integer between 1 and ...
    
        - "rep_term" sets which of the GO terms of the cluster will be used for printing output
           -> can be min (term of minimal depth in the hierarchy) or max (maximal depth)
    
        - "print_id" sets which id will be used for identifying the clusters in the printed output
           -> can be "no" (default id: go term name) or "yes" (a more complex id)
        """
        # TO DO!!! I should be able to create a go graph from a protein graph. And to work later with a network I clustered before... the clustered
        # graph must be kept somewhere in piana_api object

        # TO DO!!! improve this method to allow the user to do more things (print several levels, print composition, etc)
        
        # TO DO!!! Where do I print the cluster composition and the cluster interactions? I should pass a prefix instead of a target...
        
        
        self._check_arguments(method_name= "create_go_clustered_network",
                              file_object=output_target,
                              term_type=term_type,
                              score_threshold=score_threshold,
                              sim_mode=sim_mode,
                              level_threshold=level_threshold,
                              distance_threshold=distance_threshold,
                              rep_term=rep_term,
                              print_id=print_id)

        go_graph= GoGraph(piana_access_object = self.piana_access, term_type_value = term_type)

        if verbose:
            sys.stderr.write("Initializing GoGraph from a pianaGraph object \n")

        go_graph= go_graph.initialize_GoGraph_from_pianaGraph(pianaGraph = self.piana_graph)

        #result_network_file_name= file_name + ".initial_go_network"
        #sys.stderr.write("Printing GoGraph network to %s\n" %result_network_file_name)

        #go_graph.print_go_dot_file(output_target=file(result_network_file_name, "w"))

        if verbose:
            sys.stderr.write("Creating Clustering object\n")

        cluster_manager = Clustering()
        
        if verbose:
            sys.stderr.write("Creating GraphCluster generator\n")

        cluster_generator = ClusteredGraphGenerator()

        if verbose:
            sys.stderr.write("Creating sim object\n")
            
        sim_fun = GoClusteringSimilarityFunction(piana_access= self.piana_access,
                                                 term_type= term_type,
                                                 mode=sim_mode,
                                                 path_length_threshold=distance_threshold)

        if verbose:
            sys.stderr.write("Creating stop object\n")
            
        stop_cond= GoClusteringStopCondition(highest_depth=level_threshold, go_similarity_function= sim_fun, coeficient= score_threshold)

        if verbose:
            sys.stderr.write("Performing clustering\n")
            
        clustered_graph= cluster_manager.cluster_graph(graph_to_cluster=go_graph,
                                                       clustered_graph_generator= cluster_generator,
                                                       similarity_function=sim_fun,
                                                       stop_condition=stop_cond,
                                                       clustering_print_mode=None,
                                                       output_dir=None,
                                                       output_prefix=None,
                                                       root_protein=None, 
                                                       original_graph=None,
                                                       call_do_action= 0)

        if verbose:
            sys.stderr.write("Printing  clustering results\n")

        # the similarity function knows to which graph it belongs....
        sim_fun.print_go_graph_dot_file(output_target= output_target, representative_term=rep_term, use_alternative_id=print_id)

        #sys.stderr.write("Printing  clustering composition\n")
        #clustered_graph.print_cluster_composition()

        #sys.stderr.write("Printing  clustering interactions\n")
        #clustered_graph.print_cluster_interactions()

 
    # -------------------------------------
    #
    #  Other PIANA utilities
    # 
    # -------------------------------------


    def print_matched_pathways(self, pathways_dir, force_expression, output_file_object, results_dir, files_prefix, 
			       output_id_type, alternative_id_types, 
			       format_mode, pathway_id_type, tax_id, file_over_expressed= None,
			       file_infra_expressed= None, expression_id_type= None,
			       dic_special_proteins_file=None):
	"""
	Prints out for each pathway in pathways_dir, the number of proteins in the pathway (and percentage) that appear in the current piana_graph

	   --> this can be used to know which pathways are being 'affected' by the current network. For example, if you build a network for 
               proteins known to be involved in cancer, then you can check which pathways might be affected by the proteins in the cancer network

        "pathways_dir" is the directory that contains pathway files
                 -> a pathway file is a file that contains proteins that are known to be involved in a given pathway
		 -> the pathway filename must be indicative of the pathway: that name will be used for the output
		 -> all proteins in all pathway files must be in the same type of protein identifier.
		    If you have pathways that use different identifiers, you must call this method as many times as id types you have
	
	"force_expression" can be 'yes' or 'no' and is used to know which proteins of the network will be used to match pathways
             'no': all proteins in the current network are used to match pathways
	     'yes': only those proteins that are over/under expressed are used to match pathways

        "output_file_object" is the file where the results will be printed

        "output_id_type" is the type of code that will be used to print proteins
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "pathway_id_type" is the easy-to-remember type name that has been used in the pathways files
          -> Valid pathway_type are those listed in the PIANA reference card (see README.piana_tutorial)

        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "tax_id_value" sets which will be the species for the proteins translated
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

	"""

        self._check_arguments(method_name= "print_matched_pathways",
			      directory= pathways_dir,
			      force_expression= force_expression,
                              file_object= output_file_object,
			      id_type= output_id_type,
                              alternative_id_types=alternative_id_types,
                              another_id_type=pathway_id_type,
                              format_mode= format_mode,
                              file_over_expressed= file_over_expressed,
                              file_infra_expressed= file_infra_expressed,
                              expression_id_type= expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)

	self.piana_graph.match_pathways(pathways_dir=pathways_dir, force_expression=force_expression, 
					output_file_object=output_file_object, results_dir=results_dir, files_prefix=files_prefix, format_mode=format_mode,
					pathway_id_type= pathway_id_type, tax_id_value= tax_id, 
					output_id_type= output_id_type, alternative_id_types= alternative_id_types,
					user_protein_names = self.user_root_proteins,
					file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed, expression_id_type= expression_id_type, 
					dic_special_proteins_file=dic_special_proteins_file)
	
	    

    def print_matched_gos(self, force_expression, output_file_object, format_mode, output_id_type, tax_id, term_type, level_threshold, go_dbname, go_dbhost, go_dbuser, go_dbpass, 
			  file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, dic_special_proteins_file= None):
	"""
	Prints out the GO terms frequency for GO terms associated to proteins that appear in the current piana_graph

	   --> this can be used to know which GO terms are 'involved' in the current network. For example, if you build a network for 
               proteins known to be involved in cancer, then you can check which GO terms are related to the proteins in the cancer network

	"force_expression" can be 'yes' or 'no' and is used to know which proteins of the network will be used to match pathways
             'no': all proteins in the current network are used to match pathways
	     'yes': only those proteins that are over/under expressed are used to match pathways

        "output_file_object" is the file where the results will be printed

        "output_id_type" is the easy-to-remember type name that will be used to print out results
          -> Valid output_id_type are those listed in the PIANA reference card (see README.piana_tutorial)

        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "tax_id_value" sets which will be the species for the proteins translated
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        
        "term_type" sets the kind of GO terms that will be used for the clustering.
          -> term-type can be "molecular_function", "biological_process" or "cellular_component"
    
        "level_threshold" is the lowest level of the go terms that will be retrieved for calculating the frequency
	     -> all GO terms associated to each protein will be retrieved, then parents of those GO terms until reaching level_threshold will also be retrieved

             -> GO is a hierarchy organized from a initial root level (ie. 0) that increasingly makes more specific the terms. 
                Therefore, the higher the level used the more specific the GO terms will be. To obtain a relevant GO terms
                use level thresholds between 3 and 5. It all depends on how general you want to be in the interpretation of the network.

         "go_dbname", "go_dbhost", "go_dbuser" and "go_dbpass" are the mysql parameters for your external GO database
             -> attention! this database is different from your standard pianaDB. This is a different database that has to be created
                by following the instructions on README.populate_piana_db section "parse GO"
    
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }
	"""

        self._check_arguments(method_name= "print_matched_gos",
			      force_expression= force_expression,
                              file_object= output_file_object,
                              another_id_type=output_id_type,
                              format_mode= format_mode,
                              term_type=term_type,
                              level_threshold=level_threshold,
                              file_over_expressed= file_over_expressed,
                              file_infra_expressed= file_infra_expressed,
                              expression_id_type= expression_id_type,
			      dic_special_proteins_file=dic_special_proteins_file)


	# create connection to GO database
	if go_dbuser is None and go_dbpass is None:
	    go_db = MySQLdb.connect(db=go_dbname, host=go_dbhost)
	    
	elif go_dbpass is None and go_dbuser is not None:
	    go_db = MySQLdb.connect(user=go_dbuser, db=go_dbname, host=go_dbhost)
	    
	else:
	    go_db = MySQLdb.connect(user=go_dbuser, db=go_dbname, host=go_dbhost, passwd= go_dbpass )
	    
	go_cursor = go_db.cursor()


	# call piana_graph method
	self.piana_graph.match_gos(force_expression=force_expression, output_file_object=output_file_object, format_mode=format_mode, output_id_type=output_id_type, 
				   tax_id_value= tax_id, term_type=term_type, level_threshold=level_threshold, go_cursor= go_cursor,
				   file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed, expression_id_type= expression_id_type,
				   dic_special_proteins_file= dic_special_proteins_file)

	    

    def print_spot_protein_correspondence(self, spots_file_object, molecular_error_bounds, isoelectric_error_bounds,
                                          output_file_object, output_id_type, alternative_id_types, format_mode):
        """
        Finds correspondences between proteins in the piana_graph and spots in a 2D electrophoresis gel
          --> It does it by comparing molecular weights and isoeletric points of the spots with the mw and ip of protein sequences
          --> prints to "output_file_object" (a file object) the spots matches for each error allowed, using protein codes indicated by "id_type"
          --> (one spot can be assigned to several proteinPianas and viceversa (this is just a matching by mw and ip...)
               
        "spots_file_object": a file object with spots from a Gel, and their Molecular Weight and Isoelectric Point

               text file with spots must follows format (one spot per line):
        
                    spot_id<TAB>Molecular Weight<TAB>Isoelectric Point

               Attention!!! - Numbers must be in american style: 234234.45 and not 234234,45
                            - No headers or footers allowed


        "mw_error_bounds" and "ip_error_bounds" are lists of error bounds (they must have the same number of elements)

            the error bounds describe the percentage of error admitted when matching a spot mw or ip to the theoretical mw or ip of a protein
            for example:

                    mw_error_bounds   = [0.0, 0.0025, 0.005, 0.01]
                    ip_error_bounds   = [0.0, 0.0025, 0.005, 0.01]

        
        "output_id_type" is the easy-to-remember type name that will be used for printing the proteins that match the 2D gel
          -> Valid id_type are those listed in the PIANA reference card (see README.piana_tutorial)

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
          -> the user must provide a list of valid easy-to-remember type names
            alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             
            I suggest always placing md5 at the end of alternative types, so you never get a None in the output
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 
        """
        self._check_arguments(method_name= "print_spot_protein_correspondence",
                              another_file_object=spots_file_object,
                              molecular_error_bounds= molecular_error_bounds,
                              isoelectric_error_bounds= isoelectric_error_bounds,
                              file_object=output_file_object,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              format_mode=format_mode)
        
        spots_matches = self.piana_graph.match_spots_to_proteins(spots_file_object= spots_file_object,
                                                                 mw_error_bounds= molecular_error_bounds,
                                                                 ip_error_bounds= isoelectric_error_bounds,
                                                                 match_mode="all")

        self.piana_graph.print_spots_matches(spots_matches= spots_matches,
                                             output_target= output_file_object,
                                             format_mode= format_mode,
                                             id_type= output_id_type,
                                             alternative_id_types=alternative_id_types,
                                             mw_error_bound= molecular_error_bounds,
                                             ip_error_bound= isoelectric_error_bounds,
                                              user_protein_names = self.user_root_proteins)

   
    def protein_code_2_protein_code(self, input_file_object, input_id_type, output_file_object, output_id_type,
                                    alternative_id_types, format_mode , tax_id_value = 0):
        """
        Translates protein codes of a certain type into another type
        
        "input_file_object" is the file object with the protein codes to translate (one protein code per line)
        
        "input_id_type" is the type of code used to identify proteins in input_file_object
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)

        "output_file_object" is the file object where the translated codes will be written

        "output_id_type" is the target type of code
             -> valid protein-type values are those defined in the PIANA reference card (see README.piana_tutorial)
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html

        "tax_id_value" sets which will be the species for the proteins translated
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        
        """

        self._check_arguments(method_name= "protein_code_2_protein_code",
                              another_file_object=input_file_object ,
                              tax_id = tax_id_value,
                              id_type=input_id_type,
                              file_object=output_file_object,
                              another_id_type=output_id_type,
                              alternative_id_types=alternative_id_types,
                              format_mode=format_mode)
        
        # No name unification needed in this method, since the user is giving his list of input codes and the program is
        # returning all other codes associated to those names: no need to choose one in particular.
        
        list_input_protein_codes = utilities.return_proteins_from_file(file_object=input_file_object , id_type= input_id_type)


        if format_mode == "html":
            output_file_object.write("<table border=1>\n")
            
        for protein in list_input_protein_codes:

            list_other_codes = []
            
            list_proteinPiana = self.piana_access.get_list_proteinPiana(protein_id= protein.strip(),
                                                                        id_type=input_id_type,
                                                                        tax_id_value= tax_id_value,
                                                                        source_db_info= "no")

            for proteinPiana in list_proteinPiana:
                # each call to get_list_protein_external_codes returns the list of output_id_type codes associated to proteinPiana
                list_other_codes.extend( self.piana_access.get_list_protein_external_codes(proteinPiana=proteinPiana,
                                                                                           id_type= output_id_type,
                                                                                           alternative_id_types= alternative_id_types))
            # END OF for proteinPiana in list_proteinPiana:

            
            # print to output_file_object the result following format (with format_mode=='txt'):
            #   code_in_input_id_type<TAB>code1_in_output_id_type<TAB>code2_in_output_id_type<TAB>...

            if format_mode == "html":
                output_file_object.write("<tr><td>%s</td>" %(protein.strip()) )
            elif format_mode == "txt":
                output_file_object.write("%s" %(protein.strip()) )
                
            for other_code in list_other_codes:
                if format_mode == "txt":
                    output_file_object.write("%s%s" %( PianaGlobals.tab_separators[format_mode], other_code) )
                elif format_mode == "html":
                    output_file_object.write("<td>%s</td>" %( other_code) )

            if format_mode == "txt":
                output_file_object.write("%s" %PianaGlobals.line_separators[format_mode])
            elif format_mode == "html":
                output_file_object.write("</tr>")
          
        # END OF for protein  in list_input_protein_codes:
        if format_mode == "html":
            output_file_object.write("</table>\n")
    
