"""
 File       : PianaInsertSQL.py
 Author     : R. Aragues & D. Jaeggi
 Creation   : 2003
 Contents   : subclasses for generating insert sql statements
 Called from: PianaDBaccess

=======================================================================================================

This class is used in the same way as PianaSelectSQL

Each of the subclasses creates an sql query string that can be used to insert information into Piana

This is done this way to generalize as much as possible the interaction of users with the database pianaDB.

A user that needs to insert information into pianaDB will not create the SQL statement himself, but will
use one of these classes, together with method insert_db_content() of class PianaDB

class PianaDBaccess integrates into a single and clear interface all the access to piana databases
"""

# PianaInsertSQL.py: subclasses for generating insert sql statements
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


import PianaGlobals
import sys

verbose = 1


class PianaInsertSQL:
    """
    General class for generating Piana SQL INSERT queries. All SQL queries inserting data into piana should be extensions of this class
    """
    def get_sqlquery():
        """
        Override this with the sql query that inserts the desired values
        """
        return None
    
    get_sqlquery = staticmethod(get_sqlquery) 

#---------------------------------------------------------------------
#
# PROTEIN RELATED CLASSES
#
#---------------------------------------------------------------------

class UpdateTableColumn(PianaInsertSQL):
    """
    Generates sql statement to insert information on table protein
    """
    def get_sqlquery(proteinPiana= None, table= None, column= None, new_value= None):
                
        sqlquery = """UPDATE %s SET %s=%s where %s=%s """ \
        %(table, column, new_value, PianaGlobals.proteinPiana_col, proteinPiana )
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class InsertProtein(PianaInsertSQL):
    """
    Generates sql statement to insert information on table protein
    """
    def get_sqlquery(proteinPiana_value, tax_id_value, sequenceID_value, sourceDBID, buffer=None):

        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s) \
            VALUES (%s, %s, %s, %s) """ \
            %(PianaGlobals.protein_table,
              PianaGlobals.proteinPiana_col, PianaGlobals.sequenceID_col, PianaGlobals.taxID_col, PianaGlobals.sourceDBID_col,
              proteinPiana_value, sequenceID_value, tax_id_value, sourceDBID)

            return sqlquery
        else:
            buffer.insert2buffer(key= "InsertProtein", table= PianaGlobals.protein_table,
                                 columns= (PianaGlobals.proteinPiana_col, PianaGlobals.sequenceID_col, PianaGlobals.taxID_col, PianaGlobals.sourceDBID_col),
                                 values= (str(proteinPiana_value), str(sequenceID_value), str(tax_id_value), str(sourceDBID)))
            return None

    get_sqlquery = staticmethod(get_sqlquery)



class InsertSequence(PianaInsertSQL):
    """
    Generates sql statement to insert information on sequence

    Buffer cannot be used
    """
    def get_sqlquery(sequenceID_value, sequence_value, sequenceMD5_value,
                     sequenceLength_value, sequenceMW_value, sequenceIP_value, sourceDBID, buffer=None):

        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s, %s, %s) \
            VALUES (%s, "%s", "%s", %s, %s, %s, %s) """ \
            %(PianaGlobals.sequence_table,
              PianaGlobals.sequenceID_col, PianaGlobals.sequence_col, PianaGlobals.sequenceMD5_col, PianaGlobals.sequenceLength_col,
              PianaGlobals.sequenceMW_col, PianaGlobals.sequenceIP_col, PianaGlobals.sourceDBID_col,
              sequenceID_value, sequence_value, sequenceMD5_value, sequenceLength_value, sequenceMW_value, sequenceIP_value, sourceDBID)

            return sqlquery
        else:
            buffer.insert2buffer(key= "InsertSequence", table= PianaGlobals.sequence_table,
                                 columns= (PianaGlobals.sequenceID_col, PianaGlobals.sequence_col, PianaGlobals.sequenceMD5_col, PianaGlobals.sequenceLength_col,
                                           PianaGlobals.sequenceMW_col, PianaGlobals.sequenceIP_col, PianaGlobals.sourceDBID_col),
                                 values= (str(sequenceID_value), str(sequence_value), str(sequenceMD5_value), str(sequenceLength_value), str(sequenceMW_value), str(sequenceIP_value), str(sourceDBID)))
            return None

    get_sqlquery = staticmethod(get_sqlquery)

    
## class InsertProteinPianaCorrespondence(PianaInsertSQL):
##     """
##     Generates sql statement to generate the SQL statement that obtains the proteinPiana associated to a proteinMD5 in table proteinCorrespondence

##     This class is used to keep consistency of proteinPiana <--> proteinSequence between different versions of piana databases

##     Attention: not using it! Well... information is inserted, but coherence between different versions of piana databases is not maintained
##     """
##     def get_sqlquery(proteinPiana_value, proteinMD5_value, tax_id_value):
        
##         sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s) VALUES (%s, "%s", %s)""" %(PianaGlobals.proteinCorrespondence_table,
##                                                                                     PianaGlobals.proteinPiana_col,
##                                                                                     PianaGlobals.proteinMD5_col,
##                                                                                     PianaGlobals.taxID_col,
##                                                                                     proteinPiana_value,
##                                                                                     proteinMD5_value,
##                                                                                     tax_id_value)
##         return sqlquery
    
##    get_sqlquery = staticmethod(get_sqlquery)
    
class InsertProteinPianaCounter(PianaInsertSQL):
    """
    Generates sql statement to generate the SQL statement that obtains a new proteinPiana identifier from proteinPianaCounter table

    After getting the new proteinPiana identifier, the method calling this class must make sure that the counter in increased by 1
    """
    def get_sqlquery(updated_value, updated_col):
        
        sqlquery = """UPDATE %s set %s=%s""" %(PianaGlobals.proteinPianaCounter_table, updated_col,
                                               updated_value)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class InsertProteinSimilarity(PianaInsertSQL):
    """
    Generates sql statement to insert  protein similarity information into table proteinSimilarity
    """
    def get_sqlquery(proteinPianaA_value, proteinPianaB_value, source_db_value, buffer=None):

        # make sure proteinPianaA < proteinPianaB
        if proteinPianaA_value > proteinPianaB_value:
            temp = proteinPianaA_value
            proteinPianaA_value = proteinPianaB_value
            proteinPianaB_value = temp

        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s) \
            VALUES (%s, %s, %s) """ \
            %(PianaGlobals.proteinSimilarity_table,
              PianaGlobals.proteinPianaA_col, PianaGlobals.proteinPianaB_col,PianaGlobals.sourceDBID_col,
              proteinPianaA_value, proteinPianaB_value,source_db_value)
        
            return sqlquery
        else:
            buffer.insert2buffer( key=PianaGlobals.proteinSimilarity_table, table=PianaGlobals.proteinSimilarity_table,
                                  columns=(),
                                  values = (str(proteinPianaA_value), str(proteinPianaB_value), str(source_db_value)) )

            return None
    
    get_sqlquery = staticmethod(get_sqlquery)


#class InsertProteinSourceDB(PianaInsertSQL):
#    """
#    Generates sql statement to insert protein source DB information into table proteinSourceDB
#    """
#    def get_sqlquery(proteinPiana_value, codeSource_value, buffer=None):
#
#        if buffer is None:
#            sqlquery = """INSERT IGNORE INTO %s (%s, %s) \
#            VALUES (%s, "%s") """ \
#            %(PianaGlobals.proteinSourceDB_table,
#              PianaGlobals.proteinPiana_col, PianaGlobals.sourceDBID_col,
#              proteinPiana_value, codeSource_value)
#
#            return sqlquery
#        else:
#            buffer.insert2buffer( key = "InsertProteinSourceDB", table=PianaGlobals.proteinSourceDB_table,
#                                  columns = (PianaGlobals.proteinPiana_col, PianaGlobals.sourceDBID_col),
#                                  values = (str(proteinPiana_value), codeSource_value) )
#            return None
#    
#    get_sqlquery = staticmethod(get_sqlquery)
#


class InsertProteinExternalID(PianaInsertSQL):
    """
    Generates sql statement to insert protein ids on tables that follow the structure:

    codeID  (external code, for example uniprotEntry) (string)
    proteinPiana  (internal piana protein code)
    codeSource  (name of origin of the data being inserted)
    isPrimary [optional] (indicates whether it is a primary code or not
    
    Instead of writing one class for each table, this class is designed to handle all these similar tables. This is somehow
    dangerous, since we are hardcoding the order of the fields of the tables. If the order is changed, this class
    will not work... and we would need to de-generalize it to insert information separately in each table.

    All protein identifier tables are managed through this class
    """
    def get_sqlquery(table_name, code_value,  proteinPiana_value, codeSource_value, isPrimary_value=None, buffer=None ):

        if buffer is None:
            if isPrimary_value is None:
                sqlquery = """INSERT IGNORE INTO %s \
                VALUES ("%s", %s, "%s") """ \
                %(table_name, code_value,  proteinPiana_value, codeSource_value)
            else:
                sqlquery = """INSERT IGNORE INTO %s VALUES ("%s", %s, %s, "%s") """ \
                           %(table_name, code_value,  proteinPiana_value, isPrimary_value, codeSource_value)

            return sqlquery

        else:
            # Here, the dictionary key is a combination between table_name other string,
            # as the method can modify the different table in different form
            if isPrimary_value is None:
                buffer.insert2buffer( key = table_name+"noP", table= table_name,
                                      columns = (),
                                      values = (str(code_value),  str(proteinPiana_value), str(codeSource_value)) )
            else:
                buffer.insert2buffer( key= table_name+"P",table=table_name,
                                      columns = (),
                                      values = (str(code_value),  str(proteinPiana_value), str(isPrimary_value), str(codeSource_value)) )
            
            return None            

    get_sqlquery = staticmethod(get_sqlquery)



class InsertTemporalTable(PianaInsertSQL):
    """
    Generates sql statement to insert data in a temporal table:

    "db_temp_buffer" As all data inserted in temporal table must be processed before the connection to the database is closed, it is mandatory to pass a
    te buffer that contains the information about how to do the queries

    "conditions" is a list of the name of the columns that will be used to update parent table
    """
    
    def get_sqlquery(table_name, column_values, sourceDBID, db_temp_buffer, conditions, buffer=None ):

        num_columns = len(column_values)

        temp_columns = ["identifier"]
        temp_columns.extend(["attr%s" %(i+1) for i in xrange(num_columns)])
        temp_columns.append(PianaGlobals.sourceDBID_col)
        columns=tuple(temp_columns)
        
        table_columns = [i[0] for i in column_values]


        # Here, it is necessary to get a key for the temporal table, which will be used afterwards to update tables
        # To do this, we transform the table_columns to a integer in the following way: (it is better to be an integer than a string because of the indexing of the database)
        # It is a very ugly form to do this...BUG!!!! So, yes, it is very ugly...
        #identifier_str = str(table_columns)
        #identifier_int = len(identifier_str)+identifier_str.count(',')+identifier_str.count('a')+identifier_str.count('t')+identifier_str.count('2')+identifier_str.count('_')
        identifier_int = str(table_columns)

        temp_values = [str(identifier_int)]
        temp_values.extend([str(i[1]) for i in column_values])
        temp_values.append(str(sourceDBID))
        values = tuple(temp_values)
 
        db_temp_buffer.insert2tempbuffer( key = identifier_int,
                                          table = table_name,
                                          columns = table_columns,
                                          conditions = conditions)
        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s %s VALUES ("%s") """ \
            %(temporalData_table, columns, "\",\"".join(values))
            
            return sqlquery
        else:
            buffer.insert2buffer( key = "temptable%s" %(num_columns), table=PianaGlobals.temporalData_table,
                                  columns = columns,
                                  values = values )
            return None

    get_sqlquery = staticmethod(get_sqlquery)


class UpdateSourceDBIDfromTempTable(PianaInsertSQL):
    """
    """

    def get_sqlquery(identifier,table,conditions):

        # Construct conditions clause
        num_conditions = len(conditions)

        if num_conditions>5:
            raise ValueError("Too many conditions in the clause")

        conditions_clause = ["%s.attr%s=%s.%s" %(PianaGlobals.temporalData_table,i+1,table,conditions[i]) for i in xrange(num_conditions)]
        conditions_clause.append("identifier=\"%s\"" %(identifier))
        
        

        sqlquery = """UPDATE %s,%s SET %s.%s = %s.%s | %s.%s, temporalData.updated=1 WHERE %s""" %(table, PianaGlobals.temporalData_table, table, PianaGlobals.sourceDBID_col,
                                                                                                   table, PianaGlobals.sourceDBID_col,PianaGlobals.temporalData_table,
                                                                                                   PianaGlobals.sourceDBID_col, " and ".join(conditions_clause))

	sys.stderr.write("%s\n" %(sqlquery))
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class DeleteElementsFromTempTable(PianaInsertSQL):
    """
    """

    def get_sqlquery(identifier,table,conditions):

        # Construct conditions clause
        num_conditions = len(conditions)

        if num_conditions>5:
            raise ValueError("Too many conditions in the clause")

        conditions_clause = ["%s.attr%s=%s.%s" %(PianaGlobals.temporalData_table,i+1,table,conditions[i]) for i in xrange(num_conditions)]
        conditions_clause.append("identifier=\"%s\"" %(identifier))
        

        sqlquery = """DELETE FROM %s USING %s,%s WHERE %s""" %(PianaGlobals.temporalData_table,PianaGlobals.temporalData_table,table," and ".join(conditions_clause))

	sys.stderr.write("%s\n" %(sqlquery))
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)



class DeleteTempTableIdentifier(PianaInsertSQL):
    """
    """

    #def get_sqlquery(identifier):
    def get_sqlquery():
        #sqlquery = """DELETE FROM %s WHERE updated=1""" %(PianaGlobals.temporalData_table)
	sqlquery = """DELETE FROM %s""" %(PianaGlobals.temporalData_table)
	
	sys.stderr.write("%s\n" %(sqlquery))
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)

    

class InsertElementsFromTempTable(PianaInsertSQL):
    """
    """

    def get_sqlquery(table,columns,identifier):

        num_columns = len(columns)
        if num_columns>5:
            raise ValueError("Too many conditions in the clause")

        columns_clause = ["attr%s" %(i+1) for i in xrange(num_columns)]
        columns_clause.append(PianaGlobals.sourceDBID_col)

        columns.append(PianaGlobals.sourceDBID_col)

        sqlquery = """INSERT IGNORE INTO %s (%s) SELECT %s FROM %s WHERE identifier=\"%s\" and updated=0 """ %(table,",".join(columns),
                                                                                                           ",".join(columns_clause),PianaGlobals.temporalData_table,
                                                                                                           identifier)
	sys.stderr.write("%s\n" %(sqlquery))
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery) 


class InsertGeneralizedProteinExternalID(PianaInsertSQL):
    """
    Generates sql statement to insert protein ids on generalized table proteinExternalId
    """
    def get_sqlquery(table_name, externalId_type, externalId_column, code_value,  proteinPiana_value, codeSource_value, isPrimary_value=0, buffer=None ):

        if buffer is None:
            sqlquery = """ INSERT IGNORE INTO %s (%s,%s,%s,%s,%s) VALUES ("%s","%s","%s","%s","%s")""" \
                       %(table_name,
                         externalId_column,PianaGlobals.proteinPiana_col,PianaGlobals.sourceDBID_col,PianaGlobals.isPrimary_col,PianaGlobals.proteinTypeId_col,
                         code_value,proteinPiana_value,codeSource_value,isPrimary_value,externalId_type)

            return sqlquery

        else:
            # Here, the dictionary key is a combination between table_name other string
            buffer.insert2buffer( key="InsertGeneralizedProteinExternalID",
                                  columns = (externalId_column,PianaGlobals.proteinPiana_col,PianaGlobals.sourceDBID_col,
                                             PianaGlobals.isPrimary_col,PianaGlobals.proteinTypeId_col),
                                  values = (str(code_value),str(proteinPiana_value),str(codeSource_value),str(isPrimary_value),str(externalId_type)) )
            return None        

    get_sqlquery = staticmethod(get_sqlquery)


## class InsertProteinUniprotAcc(PianaInsertSQL):
##     """
##     Generates sql statement to insert protein information on tables that follow the structure:

##     codeID  (external code, for example uniprotAcc) (string)
##     proteinPiana  (internal piana protein code)
##     isPrimary (boolean field 1 (primary id) or 0 (secondary id)
##     codeSource  (name of origin of the data being inserted)

##     Instead of writing one class for each table, this class is designed to handle all these similar tables

##     current tables managed through this class are: swissAccession 
##     """
##     def get_sqlquery(table_name, code_value,  proteinPiana_value, codeSource_value):
                
##     return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)


## class InsertProteinEMBLCode(PianaInsertSQL):
##     """
##     Generates sql statement to insert protein information on tables that follow the structure:

##     code ID (external code, for example emblPID)
##     proteinPiana  (internal piana protein code)
##     version (number of version for the external code)
##     codeSource  (name of origin of the data being inserted)

##     Instead of writing one class for each table, this class is designed to handle all these similar tables

##     current tables managed through this class are: emblAcc, emblPID
##     """
##     def get_sqlquery(table_name, code_value, proteinPiana_value, version_value, codeSource_value):
        
##         sqlquery = """INSERT IGNORE INTO %s VALUES ("%s", %s, %s, "%s") """ \
##         %(table_name, code_value,  proteinPiana_value, version_value, codeSource_value)
        
##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)

class InsertProteinCode2(PianaInsertSQL):
    """
    Generates sql statement to insert protein information on tables that follow the structure:

    code ID (external code, for example interProID) (string)
    proteinPiana  (internal piana protein code)
    description (description of codeID) (string)
    codeSource  (name of origin of the data being inserted)

    Instead of writing one class for each table, this class is designed to handle all these similar tables

    current tables managed through this class are: interPro and pfam
    """
    def get_sqlquery(table_name, code_value, proteinPiana_value, description_value, codeSource_value, buffer=None):

        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s VALUES ("%s", %s, "%s", "%s") """ \
                       %(table_name, code_value,  proteinPiana_value, description_value, codeSource_value)
        
            return sqlquery
        else:
            buffer.insert2buffer( key= table_name+"interpro", table= table_name,
                                  columns = (),
                                  values = (str(code_value),  str(proteinPiana_value), description_value, str(codeSource_value)) )
            return None
    
    get_sqlquery = staticmethod(get_sqlquery)


class InsertProteinPdbCode(PianaInsertSQL):
    """
    Generates sql statement to insert protein information on tables that follow the structure:

    code ID (external code, for example PDB ID) (string)
    proteinPiana  (internal piana protein code)
    chain ID (specific to codeID) (string)
    codeSource  (name of origin of the data being inserted)

    Instead of writing one class for each table, this class is designed to handle all these similar tables

    current tables managed through this class are: pdb
    """
    def get_sqlquery(table_name, code_value, proteinPiana_value, chain_value, codeSource_value, combined_pdb_chain, range_value = None, buffer=buffer):

        if buffer is None:
            if range is None:
                sqlquery = """INSERT IGNORE INTO %s ("%s", "%s", "%s", "%s", "%s") VALUES ("%s", %s, "%s", "%s", "%s") """ \
                           %(table_name, PianaGlobals.pdbID_col, PianaGlobals.proteinPiana_col, PianaGlobals.chain_col, PianaGlobals.pdb_chain_col,
                             PianaGlobals.sourceDBID_col,
                             code_value,  proteinPiana_value, chain_value, combined_pdb_chain, codeSource_value)
            else:
                sqlquery = """INSERT IGNORE INTO %s VALUES ("%s", %s, "%s", "%s", "%s", "%s") """ \
                           %(table_name, code_value,  proteinPiana_value, chain_value, combined_pdb_chain, range_value, codeSource_value)
                    
                return sqlquery

        # with buffer:
        else:
            if range is None:
                buffer.insert2buffer( key = "InsertProteinPdbCode_norange", table= table_name,
                                      columns = (PianaGlobals.pdbID_col, PianaGlobals.proteinPiana_col, PianaGlobals.chain_col,
                                                             PianaGlobals.pdb_chain_col,PianaGlobals.sourceDBID_col,),
                                      values= (str(code_value),  str(proteinPiana_value), str(chain_value), combined_pdb_chain, str(codeSource_value)) )
            else:
                buffer.insert2buffer( key = "InsertProteinPdbCode_range", table= table_name,
                                      columns = (PianaGlobals.pdbID_col, PianaGlobals.proteinPiana_col, PianaGlobals.chain_col,
                                                             PianaGlobals.pdb_chain_col,PianaGlobals.range_col,PianaGlobals.sourceDBID_col),
                                      values = (str(code_value), str(proteinPiana_value), str(chain_value), combined_pdb_chain, str(range_value), str(codeSource_value)) )
            return None
    
    get_sqlquery = staticmethod(get_sqlquery)

#---------------------------------------------------------------------
#
# PROTEIN ATTRIBUTES RELATIONSHIP RELATED CLASSES
#
#---------------------------------------------------------------------

class InsertProteinAttributeRelationship(PianaInsertSQL):
    """
    Generates sql statement to insert information on protein attributes relationship tables
    
    the tables must follow the structure:
    
    (currently for tables:
    
    proteinCog_table
    proteinGo_table
    proteinEC_table 
    proteinSpecies_table 
    proteinDescription_table 
    proteinKeyword_table 
    proteinFunction_table 
    proteinSubcellularLocation_table
    )
    """
    def get_sqlquery( attribute_relationship_table, attribute_value, proteinPiana_value, attributeSource_value, buffer = None):

        if buffer is None:
            # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
            sqlquery = """INSERT IGNORE INTO %s   \
            VALUES ("%s",%s,"%s") """ \
            %(attribute_relationship_table,
              attribute_value, proteinPiana_value, attributeSource_value)

            return sqlquery
        
        else:
            # Here the key value is the table name, as this method can modify different tables
            buffer.insert2buffer( key=attribute_relationship_table, table= attribute_relationship_table,
                                  columns=(),
                                  values= (str(attribute_value), str(proteinPiana_value), str(attributeSource_value)) )

            return None
    
    get_sqlquery = staticmethod(get_sqlquery)

class InsertProteinDBAliRelationship(PianaInsertSQL):
    """
    Generates sql statement to insert information of DBAli clustering of pibase patches to table proteinDBAliCluster
    """

    def get_sqlquery( dbali_cluster_id_value, proteinPiana_value, clustering_method_value, patch_residues_value, protein_dbali_cluster_source):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s )  \
        VALUES ("%s", %s, "%s", "%s", "%s") """ \
        %(PianaGlobals.proteinDBAliCluster_table,
          PianaGlobals.DBAliclusterID_col, PianaGlobals.proteinPiana_col, PianaGlobals.DBAliclusteringMethod_col, PianaGlobals.patchResidues_col,
          PianaGlobals.sourceDBID_col,
          dbali_cluster_id_value , proteinPiana_value, clustering_method_value, patch_residues_value, protein_dbali_cluster_source )
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class InsertProteinScopRelationship(PianaInsertSQL):
    """
    Generates sql statement to insert information of SCOP to table proteinScop
    """
    def get_sqlquery( cf_value, sf_value, fa_value, proteinPiana_value, source_value, buffer = None):

        if buffer is None:
        
            # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s )  \
            VALUES (%s, %s, %s, %s, "%s") """ \
            %(PianaGlobals.proteinScop_table,
              PianaGlobals.cf_col, PianaGlobals.sf_col, PianaGlobals.fa_col, PianaGlobals.proteinPiana_col, PianaGlobals.proteinScopSource_col,
              cf_value, sf_value, fa_value, proteinPiana_value, source_value)
        
            return sqlquery

        else:
            buffer.insert2buffer( key="scop", table= PianaGlobals.proteinScop_table,
                                  columns=(PianaGlobals.cf_col, PianaGlobals.sf_col, PianaGlobals.fa_col, PianaGlobals.proteinPiana_col, PianaGlobals.proteinScopSource_col),
                                  values= (str(cf_value), str(sf_value), str(fa_value), str(proteinPiana_value), str(source_value)) )

            return None
    
    get_sqlquery = staticmethod(get_sqlquery)

class InsertProteinCathRelationship(PianaInsertSQL):
    """
    Generates sql statement to insert information of CATH to table proteinCath
    """
    def get_sqlquery( cathID_value, res_start_value, res_end_value, segmentID_value, proteinPiana_value, source_value):
        
        cath_fields = cathID_value.strip().split(".")

        if len(cath_fields) == 4:
            c_value = cath_fields[0]
            a_value = cath_fields[1]
            t_value = cath_fields[2]
            h_value = cath_fields[3]
            
        elif len(cath_fields) == 3:
            c_value = cath_fields[0]
            a_value = cath_fields[1]
            t_value = cath_fields[2]
            h_value = NULL
            
        elif len(cath_fields) == 2:
            c_value = cath_fields[0]
            a_value = cath_fields[1]
            t_value = NULL
            h_value = NULL
            
        elif len(cath_fields) == 1:
            c_value = cath_fields[0]
            a_value = NULL
            t_value = NULL
            h_value = NULL

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s  )  \
        VALUES (%s, %s, %s, %s, "%s","%s",%s, %s, %s, "%s") """ \
        %(PianaGlobals.proteinCath_table,
          PianaGlobals.c_col, PianaGlobals.a_col, PianaGlobals.t_col, PianaGlobals.h_col, PianaGlobals.cathID_col,
          PianaGlobals.segmentID_col, PianaGlobals.res_start_col,  PianaGlobals.res_end_col,
          PianaGlobals.proteinPiana_col, PianaGlobals.proteinCathSource_col,
          c_value, a_value, t_value, h_value, cathID_value,
          segmentID_value, res_start_value, res_end_value,
          proteinPiana_value, source_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

#---------------------------------------------------------------------
#
# PROTEIN ATTRIBUTES RELATED CLASSES
#
#---------------------------------------------------------------------

class InsertReactome(PianaInsertSQL):
    """
    Generates sql statement to insert information on table reactome
    """
    def get_sqlquery(reactome_id, reactomeDescription_value, reactomeSource_value):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s)  \
        VALUES ("%s", "%s", "%s") """ \
        %(PianaGlobals.reactome_table, PianaGlobals.reactomeID_col, PianaGlobals.reactomeDescription_col, PianaGlobals.reactomeSource_col, 
          reactome_id, reactomeDescription_value, reactomeSource_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



class InsertSpecies(PianaInsertSQL):
    """
    Generates sql statement to insert information on table species
    """
    def get_sqlquery(taxID_value, speciesName_value, speciesDescription_value,  speciesSource_value, speciesKingdom_value=None, buffer = None):

        if speciesKingdom_value is None:
            speciesKingdom_value = "unknown"
            
        if buffer is None:
            # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s)  \
            VALUES (%s, "%s", "%s", "%s", "%s") """ \
            %(PianaGlobals.species_table, PianaGlobals.taxID_col, PianaGlobals.speciesName_col, PianaGlobals.speciesDescription_col, 
              PianaGlobals.speciesSource_col, PianaGlobals.speciesKingdom_col,
              taxID_value, speciesName_value, speciesDescription_value, speciesSource_value, speciesKingdom_value)
        
            return sqlquery

        else:
            buffer.insert2buffer( key="InsertSpecies", table= PianaGlobals.species_table,
                                  columns=(PianaGlobals.taxID_col, PianaGlobals.speciesName_col, PianaGlobals.speciesDescription_col,
                                           PianaGlobals.speciesSource_col, PianaGlobals.speciesKingdom_col),
                                  values= (str(taxID_value), str(speciesName_value), str(speciesDescription_value), str(speciesSource_value), str(speciesKingdom_value) ))

            return None

            
    
    get_sqlquery = staticmethod(get_sqlquery)

class UpdateSpeciesKingdom(PianaInsertSQL):
    """
    Generates sql statement to update the kingdom field in species table
    """
    def get_sqlquery(taxID_value, speciesKingdom_value=None):

        if speciesKingdom_value is None:
            speciesKingdom_value = "unknown"
        
        sqlquery = """UPDATE IGNORE %s SET %s="%s" where %s=%s """ \
        %(PianaGlobals.species_table,
          PianaGlobals.speciesKingdom_col, speciesKingdom_value,
          PianaGlobals.taxID_col, taxID_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

    
class InsertGo(PianaInsertSQL):
    """
    Generates sql statement to insert information on table go
    """
    def get_sqlquery(goID_value, go_name_value, go_acc_value, go_term_type_value, go_distance2root_value, goSource_value,  buffer = None):

        if buffer is None:
            # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s, %s)  \
            VALUES ("%s", "%s", "%s", "%s", "%s", "%s") """ \
            %(PianaGlobals.go_table,
              PianaGlobals.goID_col, PianaGlobals.go_name_col, PianaGlobals.go_acc_col, PianaGlobals.go_term_type_col, PianaGlobals.go_distance2root,
              PianaGlobals.goSource_col,goID_value,    go_name_value,    go_acc_value,    go_term_type_value, go_distance2root_value,    goSource_value)
        
            return sqlquery

        else:
            buffer.insert2buffer( key="insertGo", table= PianaGlobals.go_table,
                                  columns=(PianaGlobals.goID_col, PianaGlobals.go_name_col, PianaGlobals.go_acc_col,
                                           PianaGlobals.go_term_type_col,PianaGlobals.go_distance2root, PianaGlobals.goSource_col ),
                                  values= (str(goID_value),    go_name_value,    go_acc_value,    go_term_type_value, str(go_distance2root_value),    str(goSource_value)) )

            return None
    
    get_sqlquery = staticmethod(get_sqlquery)


class InsertGo_term2term_distance(PianaInsertSQL):
    """
    Generates sql statement to insert distance between two terms
    """
    def get_sqlquery(term1_id_value, term2_id_value, distance_value, buffer= None):

        if term1_id_value > term2_id_value:
            # make sure order term1 < term2 is respected
            temp = term1_id_value
            term1_id_value = term2_id_value
            term2_id_value = temp

        if buffer is None:
            
            sqlquery= """ INSERT IGNORE INTO %s (%s,%s,%s)  VALUES ("%s","%s","%s")"""\
                      %(PianaGlobals.go_term2term_distance_table,
                        PianaGlobals.go_term2term_distance_term1_col,
                        PianaGlobals.go_term2term_distance_term2_col,
                        PianaGlobals.go_term2term_distance_col,
                        term1_id_value,
                        term2_id_value,
                        distance_value)
            return sqlquery
        else:
            buffer.insert2buffer( key="ins_t2t", table=PianaGlobals.go_term2term_distance_table,
                                  columns=(PianaGlobals.go_term2term_distance_term1_col, PianaGlobals.go_term2term_distance_term2_col,PianaGlobals.go_term2term_distance_col),
                                  values=(str(term1_id_value),str(term2_id_value),str(distance_value)) )
            return None
    
    get_sqlquery = staticmethod(get_sqlquery)


class InsertCog(PianaInsertSQL):
    """
    Generates sql statement to insert information on table cog
    """
    def get_sqlquery(cogID_value, cogDescription_value, cogFunction_value, cogSource_value, buffer = None):

        if buffer is None:
            # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s)  \
            VALUES ("%s", "%s", "%s", "%s") """ \
            %(PianaGlobals.cog_table,
              PianaGlobals.cogID_col, PianaGlobals.cogDescription_col, PianaGlobals.cogFunction_col, PianaGlobals.cogSource_col,
              cogID_value,             cogDescription_value,                cogFunction_value,            cogSource_value)
        
            return sqlquery
        
        else:
            buffer.insert2buffer( key="InsertCog", table= PianaGlobals.cog_table,
                                  columns=(PianaGlobals.cogID_col,PianaGlobals.cogDescription_col,PianaGlobals.cogFunction_col,PianaGlobals.cogSource_col),
                                  values=(cogID_value,cogDescription_value,cogFunction_value,str(cogSource_value)) )
            return None
    
    get_sqlquery = staticmethod(get_sqlquery)



#---------------------------------------------------------------------
#
# INTERACTION RELATED CLASSES
#
#---------------------------------------------------------------------

class DeleteInteraction(PianaInsertSQL):
    """
    Generates sql statement to delete an interaction entry from table interaction
    """
    def get_sqlquery(interactionPiana_value):

        sqlquery = "DELETE FROM %s where %s = %s" %(PianaGlobals.interaction_table, PianaGlobals.interactionPiana_col, interactionPiana_value)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    


# commented because this table does not exist anymore        
#class DeleteInteractionSourceDB(PianaInsertSQL):
#    """
#    Generates sql statement to delete an interaction entry from table interactionSourceDB
#
#    It can be used to delete a particular interactionPiana or just those entries that belong to a specific source db
#
#       - if interactionPiana_value 'not None' --> delete entries where interactionPiana = interactionPiana_value
#
##        - if sourceDBID_value 'not None' -->  delete entries where sourceDBID = sourceDBID_value

##     Only one argument can be 'not None'
##     """
##     def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None):

##         if interactionPiana_value is not None and sourceDBID_value is not None:
##             raise ValueError("in DeleteInteractionSourceDB, both arguments cannot be 'not None'")
        
##         if interactionPiana_value is not None:
##             sqlquery = """DELETE FROM %s where %s=%s """ %(PianaGlobals.interactionSourceDB_table, PianaGlobals.interactionPiana_col, interactionPiana_value)

##         elif sourceDBID_value is not None:
##             sqlquery = """DELETE FROM %s where %s="%s" """ %(PianaGlobals.interactionSourceDB_table, PianaGlobals.sourceDBID_col, sourceDBID_value)
            
##         else:
##             raise ValueError("in DeleteInteractionSourceDB, both arguments cannot be 'None'")

##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)
    
        
class DeleteInteractionMethod(PianaInsertSQL):
    """
    Generates sql statement to delete an interaction entry from table interactionMethod

    It can be used to delete a particular interactionPiana or just those entries that belong to a specific source db or those with a method ID

      - if interactionPiana_value 'not None' --> delete entries where interactionPiana = interactionPiana_value

      - if sourceDBID_value 'not None' -->  delete entries where sourceDBID = sourceDBID_value

      - if methodID_value 'not None' -->  delete entries where methodID = methodID_value

    Only one argument can be 'not None'
    """
    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None, methodID_value=None):

        if interactionPiana_value is not None and sourceDBID_value is not None and methodID_value is not None:
            raise ValueError("in DeleteInteractionMethod, all arguments cannot be 'not None'")
        
        elif interactionPiana_value is not None:
            sqlquery = """DELETE FROM %s where %s=%s """ %(PianaGlobals.interactionMethod_table, PianaGlobals.interactionPiana_col, interactionPiana_value)
            
        elif sourceDBID_value is not None:
            sqlquery = """DELETE FROM %s where %s="%s" """ %(PianaGlobals.interactionMethod_table, PianaGlobals.sourceDBID_col, sourceDBID_value)
            
        elif methodID_value is not None:
            sqlquery = """DELETE FROM %s where %s="%s" """ %(PianaGlobals.interactionMethod_table, PianaGlobals.methodID_col, methodID_value)
 
        else:
            raise ValueError("in DeleteInteractionMethod, all arguments cannot be 'None'")
            
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
class DeleteInteractionScores(PianaInsertSQL):
    """
    Generates sql statement to delete an interaction scores entry from table interactionScores

    It can be used to delete a particular interactionPiana or just those entries that belong to a specific source db

      - if interactionPiana_value 'not None' --> delete entries where interactionPiana = interactionPiana_value

      - if sourceDBID_value 'not None' -->  delete entries where sourceDBID = sourceDBID_value


    Only one argument can be 'not None'
    """
    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None):

        if interactionPiana_value is not None and sourceDBID_value is not None and methodID_value is not None:
            raise ValueError("in DeleteInteractionScores, all arguments cannot be 'not None'")
        
        elif interactionPiana_value is not None:
            sqlquery = """DELETE FROM %s where %s=%s """ %(PianaGlobals.interactionScores_table, PianaGlobals.interactionPiana_col, interactionPiana_value)
            
        elif sourceDBID_value is not None:
            sqlquery = """DELETE FROM %s where %s="%s" """ %(PianaGlobals.interactionScores_table, PianaGlobals.sourceDBID_col, sourceDBID_value)
            
        else:
            raise ValueError("in DeleteInteractionScores, all arguments cannot be 'None'")
            
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
        
class DeleteInteractionProteinSource(PianaInsertSQL):
    """
    Generates sql statement to delete an interaction protein source entry from table interactionProteinSource

    It can be used to delete a particular interactionPiana or just those entries that belong to a specific source db

      - if interactionPiana_value 'not None' --> delete entries where interactionPiana = interactionPiana_value

      - if sourceDBID_value 'not None' -->  delete entries where sourceDBID = sourceDBID_value

    Only one argument can be 'not None'
    """
    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None):

        if interactionPiana_value is not None and sourceDBID_value is not None and methodID_value is not None:
            raise ValueError("in DeleteInteractionProteinSource, all arguments cannot be 'not None'")
        
        elif interactionPiana_value is not None:
            sqlquery = """DELETE FROM %s where %s=%s """ %(PianaGlobals.interactionProteinSource_table, PianaGlobals.interactionPiana_col, interactionPiana_value)
            
        elif sourceDBID_value is not None:
            sqlquery = """DELETE FROM %s where %s="%s" """ %(PianaGlobals.interactionProteinSource_table, PianaGlobals.sourceDBID_col, sourceDBID_value)
            
        else:
            raise ValueError("in DeleteInteractionProteinsource, all arguments cannot be 'None'")
            
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class InsertInteraction(PianaInsertSQL):
    """
    Generates sql statement to insert information on table interaction
    """
    def get_sqlquery(interactionPiana_value, proteinPianaA_value, isSourceA_value, proteinPianaB_value, isSourceB_value, sourceDBID_value, interactionConfidence_value, buffer = None ):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error

        if buffer is None:
            sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s, %s, %s )\
            VALUES (%s, %s,%s,%s,%s,%s,"%s") """ %(PianaGlobals.interaction_table, PianaGlobals.interactionPiana_col,PianaGlobals.proteinPianaA_col, PianaGlobals.isSourceA_col,
                                                     PianaGlobals.proteinPianaB_col, PianaGlobals.isSourceB_col, PianaGlobals.sourceDBID_col,
                                                     PianaGlobals.interactionConfidence_col,
                                                     interactionPiana_value,
                                                     proteinPianaA_value, isSourceA_value, proteinPianaB_value, isSourceB_value,
                                                     sourceDBID_value, interactionConfidence_value )
            return sqlquery
        else:
            buffer.insert2buffer(key= "InsertInteraction", table= PianaGlobals.interaction_table,
                                 columns= (PianaGlobals.interactionPiana_col,PianaGlobals.proteinPianaA_col, PianaGlobals.isSourceA_col,
                                           PianaGlobals.proteinPianaB_col, PianaGlobals.isSourceB_col, PianaGlobals.sourceDBID_col,
					   PianaGlobals.interactionConfidence_col),
                                 values= (str(interactionPiana_value),str(proteinPianaA_value), str(isSourceA_value), str(proteinPianaB_value), 
					  str(isSourceB_value), str(sourceDBID_value), str(interactionConfidence_value)) )
            return None
            
    
    get_sqlquery = staticmethod(get_sqlquery)

#class InsertInteractionSourceDB(PianaInsertSQL):
#    """
#    Generates sql statement to insert information on table interactionSourceDB
#    """
#
#    def get_sqlquery(sourceDBID_value, confidenceAssignedSourceDB_value, interactionPiana_value ):

#        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
#        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s)\
#        VALUES ("%s",%s,%s) """ %(PianaGlobals.interactionSourceDB_table, PianaGlobals.sourceDBID_col,
#                                  PianaGlobals.confidenceAssigned_col,  PianaGlobals.interactionPiana_col,
#                                  sourceDBID_value, confidenceAssignedSourceDB_value, interactionPiana_value)
        
#        return sqlquery
#    
#    get_sqlquery = staticmethod(get_sqlquery)

class InsertInteractionMethod(PianaInsertSQL):
    """
    Generates sql statement to insert information on table interactionMethod
    """
    def get_sqlquery(methodID_value, interactionPiana_value, sourceDBID_value):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s)\
        VALUES ("%s",%s, "%s") """ %(PianaGlobals.interactionMethod_table,
                               PianaGlobals.methodID_col, PianaGlobals.interactionPiana_col, PianaGlobals.sourceDBID_col,
                               methodID_value, interactionPiana_value, sourceDBID_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class InsertInteractionFeatures(PianaInsertSQL):
    """
    Generates sql statement to insert information on table interactionMethod
    """
    def get_sqlquery(pubmedID_value, interactionPiana_value, sourceDBID_value):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s)\
        VALUES ("%s",%s, "%s") """ %(PianaGlobals.interactionFeatures_table,
                               PianaGlobals.pubmedID_col, PianaGlobals.interactionPiana_col, PianaGlobals.sourceDBID_col,
                               pubmedID_value, interactionPiana_value, sourceDBID_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
class InsertInteractionScores(PianaInsertSQL):
    """
    Generates sql statement to insert information on table interactionScores
    """
    def get_sqlquery(interactionPiana_value, sourceDBID_value, equiv_nscore_value, equiv_nscore_transferred_value, equiv_fscore_value,
                     equiv_pscore_value, equiv_hscore_value, array_score_value, array_score_transferred_value,
                     experimental_score_value, experimental_score_transferred_value, database_score_value, database_score_transferred_value,
                     textmining_score_value, textmining_score_transferred_value, combined_score_value):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
       
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)\
                                     VALUES (%s, "%s", %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s) """ %(
            PianaGlobals.interactionScores_table,
            PianaGlobals.interactionPiana_col,
            PianaGlobals.sourceDBID_col,
            PianaGlobals.equiv_nscore_col,
            PianaGlobals.equiv_nscore_transferred_col,
            PianaGlobals.equiv_fscore_col,
            PianaGlobals.equiv_pscore_col,
            PianaGlobals.equiv_hscore_col,
            PianaGlobals.array_score_col,
            PianaGlobals.array_score_transferred_col,
            PianaGlobals.experimental_score_col,
            PianaGlobals.experimental_score_transferred_col,
            PianaGlobals.database_score_col,
            PianaGlobals.database_score_transferred_col,
            PianaGlobals.textmining_score_col,
            PianaGlobals.textmining_score_transferred_col,
            PianaGlobals.combined_score_col,
            interactionPiana_value,
            sourceDBID_value,
            equiv_nscore_value,
            equiv_nscore_transferred_value,
            equiv_fscore_value,
            equiv_pscore_value,
            equiv_hscore_value,
            array_score_value,
            array_score_transferred_value,
            experimental_score_value,
            experimental_score_transferred_value,
            database_score_value,
            database_score_transferred_value,
            textmining_score_value,
            textmining_score_transferred_value,
            combined_score_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

    
class InsertInteractionProteinSource(PianaInsertSQL):
    """
    Generates sql statement to insert information on table interactionProteinSource
    """
    def get_sqlquery(interactionPiana_value = None, proteinPianaSource_value = None, sourceDBID_value=None):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s) VALUES (%s, %s, "%s") """ %(
            PianaGlobals.interactionProteinSource_table,
            PianaGlobals.interactionPiana_col,
            PianaGlobals.proteinPiana_col,
            PianaGlobals.sourceDBID_col,
            interactionPiana_value,
            proteinPianaSource_value,
            sourceDBID_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


#---------------------------------------
# BLAST / Protein Similarities related
#---------------------------------------

class InsertBlastResult(PianaInsertSQL):
    """
    Generates sql statement to insert blast result on table blastResults
    """
    def get_sqlquery(sequenceID_A, sequenceID_B, score, bit_score, start_A, end_A, start_B, end_B, identities, similarity, gaps, program, filter, buffer=None):

        values_str = (str(sequenceID_A), str(sequenceID_B), str(score), str(bit_score), str(start_A),
                      str(end_A), str(start_B), str(end_B), str(identities), str(similarity),str(gaps), str(program), str(filter))

        if buffer is None:
            sqlquery = """INSERT INTO %s (%s) VALUES ("%s")""" %(PianaGlobals.blastResults_table,
                                                               ",".join([PianaGlobals.sequenceID_A_col, PianaGlobals.sequenceID_B_col, PianaGlobals.score_col,
                                                                         PianaGlobals.bit_score_col, PianaGlobals.start_A_col,
                                                                         PianaGlobals.end_A_col, PianaGlobals.start_B_col,
                                                                         PianaGlobals.end_B_col, PianaGlobals.identities_col, PianaGlobals.similarity_col,
                                                                        PianaGlobals.gaps_col, PianaGlobals.program_col, PianaGlobals.filter_col]),
                                                               "\",\"".join(values_str))
            return sqlquery
        else:
            buffer.insert2buffer(key="Blast", table=PianaGlobals.blastResults_table,
                                 columns = (PianaGlobals.sequenceID_A_col, PianaGlobals.sequenceID_B_col, PianaGlobals.score_col,
                                            PianaGlobals.bit_score_col, PianaGlobals.start_A_col, PianaGlobals.end_A_col, PianaGlobals.start_B_col,
                                            PianaGlobals.end_B_col, PianaGlobals.identities_col, PianaGlobals.similarity_col, PianaGlobals.gaps_col,
                                            PianaGlobals.program_col, PianaGlobals.filter_col),
                                 values = values_str)
            return None



    get_sqlquery = staticmethod(get_sqlquery)

#---------------------------------------
# Database information classes
#---------------------------------------

class InsertNewProteinIdType(PianaInsertSQL):
    """
    Generates sql statement to insert new protein Type Id into database
    """

    def get_sqlquery(proteinTypeId=None,proteinTable=None,externalIdColumn=None,externalIdDescription=None):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s,%s) VALUES ("%s", "%s", "%s","%s" ) """ %( PianaGlobals.validProteinIds_table,
                                                                                                   PianaGlobals.proteinTypeId_col,
                                                                                                   PianaGlobals.proteinTable_col,
                                                                                                   PianaGlobals.externalIdColumn_col,
                                                                                                   PianaGlobals.externalIdDescription_col,
                                                                                                   proteinTypeId,
                                                                                                   proteinTable,
                                                                                                   externalIdColumn,
                                                                                                   externalIdDescription)
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class InsertNewExternalDatabase(PianaInsertSQL):
    """
    Generates sql statement to insert a new external database
    """

    def get_sqlquery(databaseName=None,databaseVersion=None,databaseFile=None,databaseDate=None,databaseDescription=None,databaseInteger=None):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s, %s,%s, %s, %s) VALUES ("%s", "%s", "%s","%s", "%s", %s ) """ %(PianaGlobals.externalDatabase_table,
                                                                                                                    PianaGlobals.databaseName_col,
                                                                                                                    PianaGlobals.databaseVersion_col,
                                                                                                                    PianaGlobals.databaseFile_col,
                                                                                                                    PianaGlobals.databaseDate_col,
                                                                                                                    PianaGlobals.databaseDescription_col,
                                                                                                                    PianaGlobals.databaseInteger_col,
                                                                                                                    databaseName, databaseVersion,
                                                                                                                    databaseFile,databaseDate,databaseDescription,
                                                                                                                    databaseInteger)
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class InsertExternalDatabaseInformationType(PianaInsertSQL):
    """
    Generates sql statement to insert a new type of information extracted from an external database
    """

    def get_sqlquery(databaseName,information_type,buffer=None):
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """INSERT IGNORE INTO %s (%s, %s) VALUES ("%s", "%s") """ %(PianaGlobals.externalDatabaseType_table,
                                                                               PianaGlobals.databaseName_col,
                                                                               PianaGlobals.databaseType_col,
                                                                               databaseName,information_type)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class DeleteRecords(PianaInsertSQL):
    """
    Generates sql statement that deletes those records in a database table with a column with a determined value
    """

    def get_sqlquery(table,field,value):
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """DELETE FROM %s WHERE %s=\"%s\"""" %(table,field,value)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class DeleteExternalDatabase(PianaInsertSQL):
    """
    Generates sql statement to delete an external database from databases table
    """

    def get_sqlquery(databaseName=None):
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """DELETE FROM %s WHERE %s=\"%s\"""" %(PianaGlobals.externalDatabase_table,PianaGlobals.databaseName_col,databaseName)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)



class UpdateSourceDBSet(PianaInsertSQL):
    """
    Generates sql statement to update SET datatype in a table in the database

    Used to insert or remove and element in a sourceDBID SET datatype
    
    """

    def get_sqlquery(table,field,values):
        """
        Used to insert or remove an element in sourceDBID SET datatype

        "field" is the SET field name

        "values" must be a list of all current values to be in the SET datatype

        """

        sqlquery = """ALTER TABLE %s MODIFY %s SET ('%s')""" %(table, field, "','".join(values) )

        return sqlquery


    get_sqlquery = staticmethod(get_sqlquery)


class UpdateENUMcolumn(PianaInsertSQL):
    """
    Generates sql statement to update an ENUM datatype in a table in the database
    """

    def get_sqlquery(table, column, list_enum):
        """
        "table" is the table that contains the ENUM column

        "column" is the name of the column to update

        "list" is the list with all the values of the ENUM Set
        
        """
        
        sqlquery = """ALTER TABLE %s MODIFY %s ENUM ('%s')""" %(table, column, "','".join(list_enum) )

        return sqlquery


    get_sqlquery = staticmethod(get_sqlquery)


class UpdateSourceDBID(PianaInsertSQL):
    """
    Generates sql statement to update SET datatype in a table in the database

    Used to insert or remove and element in a sourceDBID SET datatype
    
    """

    def get_sqlquery(table,field,value,conditions=[],type="insert"):
        """
        Used to insert or remove an element in sourceDBID SET datatype
    
        "type" can be  "insert"   to add a new element
                       "remove"   to remove an element

        "field" is the SourceDBID field name

        "value" is the value to add/remove (it must be in its decimal value!)

        "conditions" is a list of tuples of length 2, with the following format: (field_name,value)

        """

        #OLD CODE
        # Obtain a list of tuples (column_name,value)
        #comparisons = zip(condition_fields,condition_values)
        

        # Obtain the list of conditions:
        if len(conditions)>0:
            conditions_eq = ["%s=\"%s\"" %(tupla[0],tupla[1]) for tupla in conditions]
        else:
            conditions_eq = "";
        
        #if type=="insert":
        #    sqlquery = """UPDATE %s SET %s = CONCAT(%s,\",%s\") WHERE %s""" %(table,field,field,value," and ".join(conditions_eq))
        #elif type=="remove":
        #    sqlquery = """UPDATE %s SET %s = REPLACE(%s,'%s','') WHERE %s""" %(table,field,field,value," and ".join(conditions_eq))
        #else:
        #    raise ValueError("Type must be \"insert\" or \"remove\"")

        if type=="insert":
            sqlquery = """UPDATE %s SET %s = %s | %s""" %(table,field,field,value)
        elif type=="remove":
            sqlquery = """UPDATE %s SET %s = %s & ~%s""" %(table,field,field,value)
        else:
            raise ValueError("Type must be \"insert\" or \"remove\"")

        if len(conditions)>0:
            sqlquery = "%s WHERE %s" %(sqlquery," and ".join(conditions_eq))

        return sqlquery


    get_sqlquery = staticmethod(get_sqlquery)
    

 # OLD CODE:
## class InsertBuffer(PianaInsertSQL):
##     """
##     Generates sql statement to insert information on a key in the dictionary buffer
##     """
##     def get_sqlquery( query_data= None ):

##         if query_data is None:
##             raise ValueError("query_data cannot be None. It must be a dictionary of a buffer element")
##         elif len(query_data)<4:
##             raise ValueError("query_data has to have at least 4 elements: buffer_size, table_name,(columns),(values)")
##         else:
##             column_separator=","
##             values_separator="\",\""

##             values_string = []
##             for i in query_data[3:]:
##                 values_string.append("(\"%s\")" %(values_separator.join(i)))
            
##             sqlquery = "INSERT IGNORE INTO %s (%s) VALUES %s" %(query_data[1],
##                                                               column_separator.join(query_data[2]),
##                                                               column_separator.join(values_string) )

##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)
 


##  # NEW CODE:
## class InsertBuffer(PianaInsertSQL):
##     """
##     Generates sql statement to insert information on a key in the dictionary buffer
##     """
##     def get_sqlquery( query_data= None ):

##         if query_data is None:
##             raise ValueError("query_data cannot be None. It must be a dictionary of a buffer element")
##         else:
##             column_separator=","
##             values_separator="\",\""

#             values_string = []
##             values = query_data["values"].keys()
##             for i in values:
##                 values_string.append("(\"%s\")" %(values_separator.join(i)))
            
##             sqlquery = "INSERT IGNORE INTO %s (%s) VALUES %s" %(query_data["table"],
##                                                               column_separator.join(query_data["columns"]),
##                                                               column_separator.join(values_string) )

##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)


# NEW CODE:
class MultipleInsert(PianaInsertSQL):
    """
    Generates sql statement to insert information on a Buffer Element Object
    """
    def get_sqlquery( table,columns,values ):
        """
        "columns" must be a tuple with the name of the columns

        "values" must be a dictionary which keys are the tuples of values to be inserted
        """

        column_separator=","
        values_separator="\",\""

        values_string = []
        values = values.keys()
        for i in values:
            values_string.append("(\"%s\")" %(values_separator.join(i)))
            
        sqlquery = "INSERT IGNORE INTO %s (%s) VALUES %s" %(table,
                                                            column_separator.join(columns),
                                                            column_separator.join(values_string) )

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class CreateIndice(PianaInsertSQL):
    """
    Generates sql statement to insert information on a Buffer Element Object
    """
    def get_sqlquery( table,indice_name,columns ):
        """
        "columns" must be a tuple with the name of the columns

        """

        column_separator=","

            
        sqlquery = "CREATE INDEX %s ON %s (%s) " %(indice_name,
                                                   table,
                                                   column_separator.join(columns))

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class DropIndice(PianaInsertSQL):
    """
    Generates the sql statement to drop an index in the database
    """
    def get_sqlquery( table,indice_name ):


        column_separator=","

            
        sqlquery = "DROP INDEX %s ON %s " %(indice_name,
                                            table)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

