"""
 File       : PianaSelectSQL.py
 Author     : R. Aragues & D. Jaeggi
 Creation   : 2003
 Contents   : subclasses for generating select sql statements
 Called from: PianaDBaccess

=======================================================================================================

This class is used in the same way as PianaInsertSQL

Each of the subclasses creates an sql query string that can be used to select information from Piana

This is done this way to generalize as much as possible the interaction of users with the database pianaDB.

A user that needs to select information into pianaDB will not create the SQL statement himself, but will
use one of these classes, together with method select_db_content() of class PianaDB

class PianaDBaccess integrates into a single and clear interface all the access to piana databases

"""

# PianaSelectSQL.py: subclasses for generating select sql statements
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys

import PianaGlobals
import utilities

verbose = 0


class PianaSelectSQL:
    """
    General class for Piana SQL SELECT queries. All SQL queries reading piana should be extensions of this class
    """
    def get_sqlquery():
        """
        Override this with the sql query that returns the value being researched
        """
        return None
    
    get_sqlquery = staticmethod(get_sqlquery) 

#---------------------------------------------------------------------
# MISCELANEOUS SQL ACTIONS RELATED WITH PARSERS
#---------------------------------------------------------------------

class LockTables(PianaSelectSQL):
    """
    Generates the SQL statement that locks tables indicated in argument

    Method used to return the SQL query that locks access to mysql tables. Insertions are faster if tables are previously locked.

    table_list is a list of table names
    """
    def get_sqlquery(table_list):

        tables_to_lock = ""

        number_of_tables = len(table_list)
        
        for i in range(number_of_tables - 1):
            
            tables_to_lock +=table_list[i]
            tables_to_lock +=" WRITE, "

        # adding last table to lock. This looks weird because apparently we
        # are adding the same element as in the loop, but in fact range() returns
        # numbers from 0 to number_of_elements -1
        # therefore, number_of_tables-1 has not been added in the loop
        tables_to_lock += table_list[number_of_tables-1]
        tables_to_lock +=" WRITE "

        sqlquery = """lock tables %s """  %tables_to_lock

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class UnlockTables(PianaSelectSQL):
    """
    Generates the SQL statement that unlocks tables previously locked

    Method used to return the SQL query that locks access to mysql tables. Insertions are faster if tables are previously locked.

    table_list is a list of table names
    """
    def get_sqlquery():

        sqlquery = """unlock tables""" 

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class GetTableNames(PianaSelectSQL):
    """
    Generates the SQL statement that gets all database table names

    Method used to select tables to lock in parsers. However, it should not be used: it is better to lock only those tables that will be used

    Also used to select all tables to update sourceDBID field

    """

    def get_sqlquery():

        sqlquery = """show tables"""

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class GetTableDescription(PianaSelectSQL):
    """
    Generates the SQL statement to obtain the description of a table

    Method used to search those tables with sourceDBID field, to update them
    
    """

    def get_sqlquery(tableName):

        sqlquery = """DESC %s""" %(tableName)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class GetMaxPacket(PianaSelectSQL):
    """
    Generates the SQL statment that returns the maximum size of the allowed packet in database inserts
    """
    def get_sqlquery():
        
        #sqlquery = "show variables where variable_name=\"max_allowed_packet\""
        sqlquery = "show variables like \"max_allowed_packet\""

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



#---------------------------------------------------------------------
# DATABASE INFORMATION CLASSES
#---------------------------------------------------------------------

class SelectValidProteinIds(PianaSelectSQL):
    """
    Generates the SQL statment that returns the current protein identifiers stored in database
    """

    def get_sqlquery(validProteinIds_table):

        sqlquery = "select * from %s" %(validProteinIds_table)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)

class SelectValidSourceDBs(PianaSelectSQL):
    """
    Generates the SQL statment that returns the current external databases integrated into PIANA

    "database_type" is used to select only a type of database in function of the information it gives
    """

    def get_sqlquery(database_type=None):

        if database_type is None:
            sqlquery = "select * from %s" %(PianaGlobals.externalDatabase_table)
        else:
            sqlquery = "select * from %s,%s where %s=\"%s\" and %s.%s=%s.%s" %(PianaGlobals.externalDatabase_table, PianaGlobals.externalDatabaseType_table,
                                                                               PianaGlobals.databaseType_col,database_type,
                                                                               PianaGlobals.externalDatabaseType_table,PianaGlobals.databaseName_col,
                                                                               PianaGlobals.externalDatabase_table,PianaGlobals.databaseName_col)
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class SelectDatabaseType(PianaSelectSQL):
    """
    Generates the SQL statment that returns the type of information a database inserts into PIANA
    """

    def get_sqlquery(database_name=None):

        sqlquery = "select %s from %s where %s=\"%s\"" %(PianaGlobals.databaseType_col, PianaGlobals.externalDatabaseType_table,PianaGlobals.databaseName_col,database_name)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)
    

#---------------------------------------------------------------------
# PROTEIN RELATED CLASSES
#---------------------------------------------------------------------


class SelectColumnValues(PianaSelectSQL):
    """
    General method to select the total number of entries in a table that have a determined combination of values
    
    """
    def get_sqlquery(tableName,tableColumns=[],values=[]):
        """
        "tableColumns" and "values" must have the same length
        """

        # Obtain a list of tuples: (column_name,value)
        comparisons = zip(tableColumns,values)

        # Obtain the list of conditions:
        conditions = ["%s=\"%s\"" %(tupla[0],tupla[1]) for tupla in comparisons]

        sqlquery = """select count(*) from %s where %s""" %(tableName," and ".join(conditions))
        
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class SelectAllProteinPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains all proteinPiana in database for a given tax_id

    if tax_id_value is 0, returns all proteinPianas in the database

    """
    def get_sqlquery(tax_id_value=0):

        if tax_id_value==0:
            sqlquery = """select %s from %s """ %(PianaGlobals.proteinPiana_col,
                                                  PianaGlobals.protein_table)
        else:
            ## sqlquery = """select p.%s from %s p, %s s where p.%s=s.%s and s.%s=%s""" %(PianaGlobals.proteinPiana_col,
##                                                                                        PianaGlobals.protein_table,
##                                                                                        PianaGlobals.proteinSpecies_table,
##                                                                                        PianaGlobals.proteinPiana_col,
##                                                                                        PianaGlobals.proteinPiana_col,
##                                                                                        PianaGlobals.taxID_col,
##                                                                                        tax_id_value                 )
            sqlquery = """select %s from %s where %s=%s""" %(PianaGlobals.proteinPiana_col,
                                                             PianaGlobals.protein_table,
                                                             PianaGlobals.taxID_col,tax_id_value)
            
            
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

def _normalize_pdb(code_value):
    """
    returns a normalized pdb.chain code for string code_value
    """

    # if table is pdb, there is a bit of preprocessing to do
    #   old pdb codes used NULL chain values
    #   pianaDB table pdb contains a column (pdb_chain) where NULL chain values have been transformed to "A"
    #   therefore, if the code coming in has chain value NULL, we'll look in piana for chain value "A"
    # I also protect against wrong usage of pdb code (not writing the "." for empty chain values)

    if len(code_value) == 6:
        # code was correctly passed (ie. pdb.chain)
        pdb_fields = code_value.split(".")
        pdb_code = pdb_fields[0].strip()
        pdb_chain = pdb_fields[1].strip()

    elif len(code_value) == 5:
        # code was uncorrectly passed: two different incorrect options: pdb. or pdbchain

        if code_value[-1] == ".":
            pdb_code = code_value[:-1]
            pdb_chain = "A"
        else:
            pdb_code = code_value[:-1]
            pdb_chain = code_value[-1]

    elif len(code_value) == 4:
        # code was uncorrectly passed: no chain given (we will consider it to be the only chain in the file)

        pdb_code = code_value
        pdb_chain = "A"

    else:
        raise ValueError("Weird pdb code used: %s\n" %(code_value))

    if pdb_chain == "" or pdb_chain == "_":
        pdb_chain = "A"

    pdb_code = pdb_code + "." + pdb_chain

    return pdb_code

    
    
class SelectProteinPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains proteinPiana from any other protein identifier (swissprot, ...)

    Method get_sqlquery used to return the SQL query that retrieves the proteinPianas equivalent to "protein_code" of type "code_type"

    Valid "code_type" value are those in PianaGlobals.xxxx_col that are in a table where there is as well a proteinPiana_col

    if protein_code being passed is a pdb code, the format of the code must be pdb_code.chain_id  If the chain_id is None, write pdb_code. (leaving the dot)

    "isPrimary" is used when only user wants to obtain proteinPiana that correspond to primary external code
                i.e. if an external code has assigned two proteinPianas, only those assigned to primary codes will be returned

        "source_db_info" determines if information about the sourceDB that inserted the proteinPiana is returned or not
           - "no" will simply return a list of proteinPianas
           - "yes" will return a list of tuples (proteinPiana, sourceDBID)

    "id_type" is only necessary when identifier type uses the generalized table

    """
           
    def get_sqlquery(protein_code= None, protein_code_column=None, protein_code_table=None, isPrimary=0, source_db_info= "no", id_type=None):

        if protein_code is None:
            raise ValueError("Trying to Select the proteinPiana of None arguments")

        if protein_code_table is None:
            raise ValueError("trying to Select the proteinPiana for a type that doesn't have a table associated: %s" %(code_type))

        if protein_code_table == PianaGlobals.pdb_table:
            # if dealing with pdb codes, normalization required...
            protein_code = _normalize_pdb(code_value=protein_code)
        # END OF if protein_code_table == PianaGlobals.pdb_table:

        if source_db_info == "no":
            # add empty string to select: do not get source db
            get_source_db = ""
        elif source_db_info == "yes":
            if protein_code_table == PianaGlobals.protein_table:
                # if proteinPiana to be returned from protein table (which has no source_db) we
                # have to modify the query so it returns the name of the table...
                get_source_db = ', "protein"'
            else:
                get_source_db = ", %s" %(PianaGlobals.sourceDBID_col)


        if isPrimary:
            primary_condition = " and %s=1" %(PianaGlobals.isPrimary_col)
        else:
            primary_condition = ""

        if id_type:
            id_type_condition = " and %s=\"%s\"" %(PianaGlobals.externalIdType_col,id_type)
        else:
            id_type_condition = ""
        

        
        sqlquery = """select %s%s from %s where %s = "%s" %s %s""" %(PianaGlobals.proteinPiana_col,
                                                                     get_source_db,
                                                                     protein_code_table,
                                                                     protein_code_column,
                                                                     protein_code,
                                                                     primary_condition,
                                                                     id_type_condition)


        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectProteinPianaGeneralized(PianaSelectSQL):
    """
    "id_type" is only necessary when identifier type uses the generalized table
    """
    
    def get_sqlquery(proteinCode_table, proteinCode_col, code_value, id_type=None):

        if id_type:
            id_type_condition = " and %s=\"%s\"" %(PianaGlobals.externalIdType_col,id_type)
        else:
            id_type_condition = ""
            
        
        sqlquery = """select %s from %s where %s = "%s" %s""" %(PianaGlobals.proteinPiana_col,
                                                                proteinCode_table,
                                                                proteinCode_col, code_value,
                                                                id_type_condition)
        
        return sqlquery
    
    
    get_sqlquery = staticmethod(get_sqlquery)
            



class SelectProteinPianaTax(PianaSelectSQL):
    """
    Generates the SQL statement that obtains proteinPiana (of a specific taxonomy) from any other protein identifier (swissprot, ...)

    Method get_sqlquery is used to return the SQL query that retrieves the proteinPianas (of a specific taxonomy)
    equivalent to "protein_code"
    
    Valid taxonomy_value values are those in NCBI taxonomy database

    "source_db_info" determines if information about the sourceDB that inserted the proteinPiana is returned or not
           - "no" will simply return a list of proteinPianas
           - "yes" will return a list of tuples (proteinPiana, sourceDBID)

    "isPrimary" is used when only user wants to obtain proteinPiana that correspond to primary external code
                i.e. if an external code has assigned two proteinPianas, only those assigned to primary codes will be returned

    "id_type" is only necessary when identifier type uses the generalized table
                
    """
    def get_sqlquery(protein_code= None, protein_code_column= None, protein_code_table=None, taxonomy_value= None, isPrimary=0, source_db_info= "no", id_type=None):

        if protein_code is None or taxonomy_value is None:
            raise ValueError("Trying to call SelectProteinPianaTax with None arguments")

        if taxonomy_value == 0:
            return SelectProteinPiana.get_sqlquery(protein_code= protein_code, code_type= code_type, source_db_info= source_db_info)

        if protein_code_table is None:
            raise ValueError("trying to Select the proteinPiana for a type that doesn't have a table associated")


        if protein_code_table == PianaGlobals.pdb_table:
            # if dealing with pdb codes, normalization required...
            protein_code = _normalize_pdb(code_value=protein_code)
        # END OF if protein_code_table == PianaGlobals.pdb_table:


        if source_db_info == "no":
            # add empty string to select: do not get source db
            get_source_db = ""
        elif source_db_info == "yes":
            if protein_code_table == PianaGlobals.protein_table:
                # if proteinPiana to be returned from protein table (which has no source_db) we
                # have to modify the query so it returns the name of the table...
                get_source_db = ', "protein"'
            else:
                #get_source_db = ", code_table.%s" %(PianaGlobals.sourceDBID_col)
                get_source_db = ", %s.%s" %(protein_code_table,PianaGlobals.sourceDBID_col)

        if isPrimary:
            primary_condition = " and %s=1" %(PianaGlobals.isPrimary_col)
        else:
            primary_condition = ""

        if id_type:
            id_type_condition = " and %s=\"%s\"" %(PianaGlobals.externalIdType_col,id_type)
        else:
            id_type_condition = ""

        # the query in this method is built so it only retrieves proteinPianas that have taxonomy taxonomy_value

        ## Modification by Javi (aliases didn't work with lock options in mysql...)
        ## I also eliminate the distinct clause (it is so slow and PianaDB eliminates duplicates)
        ## Otherwise, PianaDB also eliminates duplicates
        
        #    --> code_table and species_table are aliases used for simplifying the query...
        #sqlquery = """select distinct code_table.%s%s from %s code_table, %s species_table \
        #    where code_table.%s="%s" and code_table.%s=species_table.%s and species_table.%s= %s """ %(PianaGlobals.proteinPiana_col,
        #                                                                                               get_source_db,
        #                                                                                               protein_code_table,
        #                                                                                               PianaGlobals.proteinSpecies_table,
        #                                                                                               code_type,
        #                                                                                               protein_code,
        #                                                                                               PianaGlobals.proteinPiana_col,
        #                                                                                               PianaGlobals.proteinPiana_col,
        #                                                                                               PianaGlobals.taxID_col,
        #                                                                                               taxonomy_value)

        ## For the moment, this is the best one
        if protein_code_table != PianaGlobals.sequence_table:
            sqlquery = """SELECT %s.%s%s from %s, %s \
            where %s.%s="%s" and %s.%s=%s.%s and %s.%s= %s %s %s""" %(protein_code_table, PianaGlobals.proteinPiana_col,
                                                                      get_source_db,
                                                                      protein_code_table,
                                                                      PianaGlobals.protein_table,
                                                                      protein_code_table,protein_code_column,protein_code,
                                                                      protein_code_table,PianaGlobals.proteinPiana_col,
                                                                      PianaGlobals.protein_table,PianaGlobals.proteinPiana_col,
                                                                      PianaGlobals.protein_table,PianaGlobals.taxID_col,taxonomy_value,
                                                                      primary_condition, id_type_condition)
        else:
            sqlquery = """SELECT %s%s from %s, %s \
            where %s.%s="%s" and %s.%s=%s.%s and %s.%s= %s %s %s""" %(PianaGlobals.proteinPiana_col,
                                                                      get_source_db,
                                                                      protein_code_table,
                                                                      PianaGlobals.protein_table,
                                                                      protein_code_table,protein_code_column,protein_code,
                                                                      protein_code_table,PianaGlobals.sequenceID_col,
                                                                      PianaGlobals.protein_table,PianaGlobals.sequenceID_col,
                                                                      PianaGlobals.protein_table,PianaGlobals.taxID_col,taxonomy_value,
                                                                      primary_condition, id_type_condition)
        #sqlquery = """select %s.%s from %s \
        #                       where %s.%s="%s" and %s.%s IN (SELECT %s FROM %s WHERE %s="%s") """ %(PianaGlobals.proteinSpecies_table,PianaGlobals.proteinPiana_col,
        #                                                                                          PianaGlobals.proteinSpecies_table,
        #                                                                                          PianaGlobals.proteinSpecies_table,PianaGlobals.taxID_col,taxonomy_value,
        #                                                                                          PianaGlobals.proteinSpecies_table,PianaGlobals.proteinPiana_col,
        #                                                                                          PianaGlobals.proteinPiana_col,
        #                                                                                          protein_code_table,
        #                                                                                          code_type,protein_code)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



class SelectProteinSimilarity(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a pair of similar proteins, if they exist

    Selects two elements, must be taken into account when making the call to PianaDB
    """

    def get_sqlquery(proteinPianaA_value, proteinPianaB_value):

        # make sure proteinPianaA < proteinPianaB
        if proteinPianaA_value > proteinPianaB_value:
            temp = proteinPianaA_value
            proteinPianaA_value = proteinPianaB_value
            proteinPianaB_value = temp
        
        sqlquery = """select %s, %s from %s where %s = %s and %s=%s""" %(PianaGlobals.proteinPianaA_col, PianaGlobals.proteinPianaB_col,
                                                                         PianaGlobals.proteinSimilarity_table,
                                                                         PianaGlobals.proteinPianaA_col, proteinPianaA_value,
                                                                         PianaGlobals.proteinPianaB_col, proteinPianaB_value)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectSimilarProteins(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list of proteins that are similar to a proteinPiana

    "mode" can be:

       - "higher" : searches in the column proteinPianaB
       - "lower"  : searches in the column proteinPianaA

    To get all similar proteins you need to do two queries, one with mode higher and another with mode lower. PianaDBaccess.get_similar_proteins_dic does it...
    """
    def get_sqlquery(proteinPiana_value, mode= None):

        if mode == "higher":
            sqlquery = """select %s from %s where %s = %s""" %(PianaGlobals.proteinPianaB_col, 
                                                                         PianaGlobals.proteinSimilarity_table,
                                                                         PianaGlobals.proteinPianaA_col, proteinPiana_value)
        elif mode == "lower":
            sqlquery = """select %s from %s where %s = %s""" %(PianaGlobals.proteinPianaA_col, 
                                                                         PianaGlobals.proteinSimilarity_table,
                                                                         PianaGlobals.proteinPianaB_col, proteinPiana_value)
        else:
            raise ValueError("Incorrent mode (%s) used: valid values are higher and lower" %(mode))
            

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectProteinID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains protein ids from protein tables

    This is a generalization that can be used for any protein table that has at least these two fields:

       proteinPiana
       proteinID_X
    
    Currently, this class can be used for all ID tables

    "id_type" is only necessary when the table being used is the generalized table

    """
    def get_sqlquery(proteinCode_table, proteinCode_col, proteinPiana_value, source_db_info="no", return_primary=None, id_type=None):

        if id_type:     id_type_condition = " and %s=\"%s\"" %(PianaGlobals.externalIdType_col,id_type)
        else:           id_type_condition = ""

        if not return_primary:

            if source_db_info == "no":
                sqlquery = """select %s from %s where %s = %s %s""" %(proteinCode_col, proteinCode_table, 
                                                                      PianaGlobals.proteinPiana_col , proteinPiana_value,
                                                                      id_type_condition)
            else:
                sqlquery = """select %s, %s from %s where %s = %s %s""" %(proteinCode_col, PianaGlobals.sourceDBID_col,
                                                                          proteinCode_table,
                                                                          PianaGlobals.proteinPiana_col , proteinPiana_value,
                                                                          id_type_condition)
        else:
            if source_db_info == "no":
                sqlquery = """select %s from %s where %s = %s and %s = %s %s""" %(proteinCode_col, proteinCode_table,
                                                                                  PianaGlobals.proteinPiana_col , proteinPiana_value,
                                                                                  PianaGlobals.isPrimary_col, return_primary,
                                                                                  id_type_condition)
            else:
                # source_db_info required
                sqlquery = """select %s, %s from %s where %s = %s and %s = %s %s""" %(proteinCode_col, PianaGlobals.sourceDBID_col,
                                                                                      proteinCode_table,
                                                                                      PianaGlobals.proteinPiana_col , proteinPiana_value,
                                                                                      PianaGlobals.isPrimary_col, return_primary,
                                                                                      id_type_condition )
        # END OF else: (if not return_primary:)


            
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


## class SelectSwissAccession(PianaSelectSQL):
##     """
##     Generates the SQL statement that obtains swiss accession numbers corresponding to proteinPiana

##     get_sqlquery method used to return the SQL query that retrieves the swiss accession number corresponding to proteinPiana

##         if return_primary == 1, then returns only primary swiss accessions
##         if return_primary == 0, then returns all swiss accession numbers
        
##     """
    
##     def get_sqlquery(proteinPiana_value, return_primary=1, source_db_info="no"):

##         if source_db_info == "no":
##             if return_primary == 1:
##                 sqlquery = """select %s from %s where %s = %s and %s = %s""" %(PianaGlobals.swissAccessionID_col, PianaGlobals.swissAccession_table,
##                                                                                PianaGlobals.proteinPiana_col , proteinPiana_value,
##                                                                                PianaGlobals.isPrimary_col, return_primary )
##             else:
##                 sqlquery = """select %s from %s where %s = %s""" %(PianaGlobals.swissAccessionID_col, PianaGlobals.swissAccession_table,
##                                                                    PianaGlobals.proteinPiana_col , proteinPiana_value)
##         else:
##             # source_db_info required
##             if return_primary == 1:
##                 sqlquery = """select %s, %s from %s where %s = %s and %s = %s""" %(PianaGlobals.swissAccessionID_col, PianaGlobals.sourceDBID_col,
##                                                                                    PianaGlobals.swissAccession_table,
##                                                                                    PianaGlobals.proteinPiana_col , proteinPiana_value,
##                                                                                    PianaGlobals.isPrimary_col, return_primary )
##             else:
##                 sqlquery = """select %s, %s from %s where %s = %s""" %(PianaGlobals.swissAccessionID_col, PianaGlobals.sourceDBID_col,
##                                                                        PianaGlobals.swissAccession_table,
##                                                                        PianaGlobals.proteinPiana_col , proteinPiana_value)
            

##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)


class SelectProteinTableColumn(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a column value of table protein, given a proteinPiana identifier

    This is a generalization that can be used for any column of table protein
    """
    def get_sqlquery(proteinTable_column_value, proteinPiana_value):
        
        sqlquery = """select %s from %s where %s = %s""" %(proteinTable_column_value, PianaGlobals.protein_table
                                                           , PianaGlobals.proteinPiana_col , proteinPiana_value)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



class SelectSequencePropertyFromProteinPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a column value of sequence table, given a proteinPiana identifier

    This is a generalization that can be used for any column of sequence protein
    """
    def get_sqlquery(sequenceTable_column_value, proteinPiana_value):
        
        sqlquery = """select %s from %s,%s where %s = %s and %s.%s=%s.%s""" %(sequenceTable_column_value, PianaGlobals.protein_table, PianaGlobals.sequence_table,
                                                                              PianaGlobals.proteinPiana_col, proteinPiana_value,
                                                                              PianaGlobals.protein_table, PianaGlobals.sequenceID_col,
                                                                              PianaGlobals.sequence_table, PianaGlobals.sequenceID_col)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



class SelectSequenceSequenceMD5FromSequenceID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a column value of sequence table, given a proteinPiana identifier

    This is a generalization that can be used for any column of sequence protein
    """
    def get_sqlquery(sequenceID_value):
        
        sqlquery = """select %s,%s from %s where %s = %s""" %( PianaGlobals.sequenceMD5_col, PianaGlobals.sequence_col,
                                                               PianaGlobals.sequence_table,
                                                               PianaGlobals.sequenceID_col,sequenceID_value )
                                                                               
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectSequenceFromSequenceID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a column value of sequence table, given a proteinPiana identifier

    This is a generalization that can be used for any column of sequence protein
    """
    def get_sqlquery(sequenceID_value):
        
        sqlquery = """select %s from %s where %s = %s""" %( PianaGlobals.sequence_col,
                                                            PianaGlobals.sequence_table,
                                                            PianaGlobals.sequenceID_col,sequenceID_value )
                                                                               
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


## class SelectProteinPianaCorrespondence(PianaSelectSQL):
##     """
##     Generates the SQL statement that obtains the proteinPiana associated to a proteinMD5 in table proteinCorrespondence

##     This class is use to keep consistency of proteinPiana <--> proteinSequence between different versions of piana databases
##     """
##     def get_sqlquery(proteinMD5_value, tax_id_value):
        
##         sqlquery = """select %s from %s where %s = "%s" and %s=%s """ %(PianaGlobals.proteinPiana_col, PianaGlobals.proteinCorrespondence_table,
##                                                                         PianaGlobals.proteinMD5_col , proteinMD5_value,
##                                                                         PianaGlobals.taxID_col, tax_id_value)
##         return sqlquery
    
##     get_sqlquery = staticmethod(get_sqlquery)


class SelectProteinPianaFromTaxID_and_sequenceID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the proteinPiana associated to a proteinMD5 and taxID in table protein

    This class is use to keep consistency of proteinPiana <--> proteinSequence and taxID between different versions of piana databases
    """
    def get_sqlquery(sequenceID_value, tax_id_value):
        
        sqlquery = """select %s from %s where %s=%s and %s=%s """ %(PianaGlobals.proteinPiana_col, PianaGlobals.protein_table,
                                                                    PianaGlobals.sequenceID_col , sequenceID_value,
                                                                    PianaGlobals.taxID_col, tax_id_value)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)



class SelectProteinPianaFromTaxID_and_MD5(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the proteinPiana associated to a proteinMD5 and taxID in tables protein and sequence

    This class is use to keep consistency of proteinPiana <--> proteinSequence and taxID between different versions of piana databases
    """
    def get_sqlquery(sequenceMD5_value, tax_id_value):
        
        sqlquery = """select %s from %s,%s where %s="%s" and %s.%s =%s.%s and %s=%s """ %(PianaGlobals.proteinPiana_col,
                                                                                          PianaGlobals.protein_table,PianaGlobals.sequence_table,
                                                                                          PianaGlobals.sequenceMD5_col,sequenceMD5_value,
                                                                                          PianaGlobals.protein_table,PianaGlobals.sequenceID_col,
                                                                                          PianaGlobals.sequence_table,PianaGlobals.sequenceID_col,
                                                                                          PianaGlobals.taxID_col, tax_id_value)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)

class SelectSequenceIDFromMD5(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the sequenceID associated to a sequenceMD5
    """
    def get_sqlquery(sequenceMD5_value):
        
        sqlquery = """select %s from %s where %s = "%s" """ %(PianaGlobals.sequenceID_col, PianaGlobals.sequence_table,
                                                              PianaGlobals.sequenceMD5_col , sequenceMD5_value)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)



class SelectAllDistinctMD5(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list of all distinct sequenceMD5 in table sequence

    This class is used to generate FASTA file of PIANA sequences
    """
    def get_sqlquery():
        
        sqlquery = """select %s from %s""" %(PianaGlobals.sequenceMD5_col,
                                             PianaGlobals.sequence_table)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)


class SelectAllDistinctSequenceID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list of all distinct sequenceID in table sequence

    This class is used to generate FASTA file of PIANA sequences
    """
    def get_sqlquery(taxID=0):

        if( taxID==0 ):
            
            sqlquery = """select %s from %s""" %(PianaGlobals.sequenceID_col,
                                                 PianaGlobals.sequence_table)
        else:
            sqlquery = """select %s from %s WHERE %s=%s""" %(PianaGlobals.sequenceID_col,
                                                             PianaGlobals.protein_table,
                                                             PianaGlobals.taxID_col,
                                                             taxID)
            
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)



class SelectProteinPianaFromMD5(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list of all proteinPianas that correspond to a sequence (represented by its MD5 value)

    This class is used to generate FASTA file of PIANA sequences
    """
    def get_sqlquery(proteinMD5_value):

        sqlquery = """select %s from %s,%s where %s="%s" and %s.%s=%s.%s""" %(PianaGlobals.proteinPiana_col,
                                                                              PianaGlobals.protein_table,PianaGlobals.sequence_table,
                                                                              PianaGlobals.proteinMD5_col,proteinMD5_value,
                                                                              PianaGlobals.protein_table, PianaGlobals.sequenceID_col,
                                                                              PianaGlobals.sequence_table, PianaGlobals.sequenceID_col)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class SelectAllProteinPiana_TaxID_and_MD5(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the relation (proteinMD5,taxID,proteinPiana) in table protein

    This class is use to keep consistency of proteinPiana <--> proteinSequence and taxID between different versions of piana databases
    """
    def get_sqlquery():
        
        sqlquery = """select %s,%s,%s from %s""" %(PianaGlobals.proteinMD5_col,
                                                     PianaGlobals.taxID_col,
                                                     PianaGlobals.proteinPiana_col,
                                                     PianaGlobals.protein_table)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)
    

class SelectAllProteinPiana_sequenceID_and_taxID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the relation (sequenceID, taxID,proteinPiana) in table protein

    This class is used to get all the relation to opmitze parsers (avoid querying database each time)
    """
    
    def get_sqlquery():
        
        sqlquery = """select %s,%s,%s from %s""" %(PianaGlobals.sequenceID_col,
                                                   PianaGlobals.taxID_col,
                                                   PianaGlobals.proteinPiana_col,
                                                   PianaGlobals.protein_table)
        return sqlquery
 
    get_sqlquery = staticmethod(get_sqlquery)


class SelectNewProteinPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a new proteinPiana identifier from proteinPianaCounter table

    After getting the new proteinPiana identifier, the method calling this class must make sure that the counter in increased by 1
    """
    def get_sqlquery():
        
        sqlquery = """select %s from %s""" %(PianaGlobals.proteinPianaCounter_col, PianaGlobals.proteinPianaCounter_table)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectNewInteractionPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a new interactionPiana identifier from proteinPianaCounter table

    After getting the new interactionPiana identifier, the method calling this class must make sure that the counter in increased by 1
    """
    def get_sqlquery():
        
        sqlquery = """select %s from %s""" %(PianaGlobals.interactionPianaCounter_col, PianaGlobals.proteinPianaCounter_table)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    

class SelectNewSequenceID(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a new sequenceID identifier from proteinPianaCounter table

    After getting the new sequenceID identifier, the method calling this class must make sure that the counter in increased by 1
    """
    def get_sqlquery():
        
        sqlquery = """select %s from %s""" %(PianaGlobals.sequenceIDCounter_col, PianaGlobals.proteinPianaCounter_table)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectNewInteractionPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a new sequenceID identifier from proteinPianaCounter table

    After getting the new sequenceID identifier, the method calling this class must make sure that the counter in increased by 1
    """
    def get_sqlquery():
        
        sqlquery = """select %s from %s""" %(PianaGlobals.interactionCounter_col, PianaGlobals.proteinPianaCounter_table)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectMaxProteinPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the maximum proteinPiana identifier
    
    Is used when an external database is dropped and we want to reuse these proteinPianas
    """
    def get_sqlquery():

        sqlquery = """select MAX(%s) from %s""" %(PianaGlobals.proteinPiana_col, PianaGlobals.protein_table)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)
    
    
class SelectCommonCodeProteins(PianaSelectSQL):
    """
    Generates the SQL statement that obtains those proteins that share a code with a given protein

    The list will not contain the protein passed as reference
    """
    def get_sqlquery( proteinCode_table, proteinCodeID_col, proteinPiana_value):

        sqlquery = """select t2.%s from %s t1, %s t2 where t1.%s = %s and t1.%s=t2.%s and t2.%s <> %s""" %(PianaGlobals.proteinPiana_col,
                                                                                                           proteinCode_table,
                                                                                                           proteinCode_table,
                                                                                                           PianaGlobals.proteinPiana_col, proteinPiana_value,
                                                                                                           proteinCodeID_col, proteinCodeID_col,
                                                                                                           PianaGlobals.proteinPiana_col, proteinPiana_value)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectCommonCharacteristicProteins(PianaSelectSQL):
    """
    Generates the SQL statement that obtains those proteins with characteristics described to get_sqlquery

    This class is used for example to get all proteins that belong to a certain species, by passing the approapiate arguments
    as shown:  get_sqlquery(proteinCode_table= PianaGlobals.proteinSpecies_table,
                            proteinCodeID_col= PianaGlobals.taxID_col,
                            proteinCodeID_value = taxonomy_value)

              will retrieve all those proteinPiana of table proteinSpecies_table where column taxID_col is taxonomy_value
    """
    # TO CHECK!!! I am not sure this will work for string and int values!!! Do I need to put the "" around %s???
    #             (I think integers do not mind to be between quotes)


    def get_sqlquery(proteinCode_table=None, proteinCodeID_col= None, proteinCodeID_value=None ):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
        if proteinCodeID_value is None:
            raise ValueError("code value needed to retrieve proteins from table")
        
        sqlquery = """select %s from %s where %s="%s" """ %(PianaGlobals.proteinPiana_col,
                                                         proteinCode_table,
                                                         proteinCodeID_col,
                                                         proteinCodeID_value)
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


#---------------------------------------------------------------------
# PROTEIN ATTRIBUTES RELATIONSHIP RELATED CLASSES
#---------------------------------------------------------------------

class SelectProteinAttributeRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about protein attributes relationship tables
    
    the tables must follow the structure:
    
    (currently for tables:

    protein_table
    proteinCog_table 
    proteinEC_table 
    proteinSpecies_table 
    proteinDescription_table 
    proteinKeyword_table 
    proteinFunction_table 
    proteinSubcellularLocation_table
    )
    """
    def get_sqlquery( attribute_relationship_table, attribute_relationship_col, proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(attribute_relationship_col, attribute_relationship_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectAllDBAliClusterRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about dbali cluster information
    """
    def get_sqlquery( clustering_method_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        
        sqlquery = """SELECT %s, %s from %s where %s="%s" """ \
        %(PianaGlobals.proteinPiana_col, PianaGlobals.DBAliclusterID_col,
          PianaGlobals.proteinDBAliCluster_table,
          PianaGlobals.DBAliclusteringMethod_col, clustering_method_value)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectDBAliClusterRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about dbali cluster information
    """
    def get_sqlquery( proteinPiana_value, clustering_method_value, source_db_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error

        if source_db_value == "all":
            sqlquery = """SELECT %s from %s where %s = %s and %s="%s"  """ \
                       %(PianaGlobals.DBAliclusterID_col,
                         PianaGlobals.proteinDBAliCluster_table,
                         PianaGlobals.proteinPiana_col, proteinPiana_value,
                         PianaGlobals.DBAliclusteringMethod_col, clustering_method_value)
        else:
            # if a specific source_db has been chosen, fix it in the sql query
            sqlquery = """SELECT %s from %s where %s = %s and %s="%s" and %s="%s"  """ \
                       %(PianaGlobals.DBAliclusterID_col,
                         PianaGlobals.proteinDBAliCluster_table,
                         PianaGlobals.proteinPiana_col, proteinPiana_value,
                         PianaGlobals.DBAliclusteringMethod_col, clustering_method_value,
                         PianaGlobals.sourceDBID_col, source_db_value)
            
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectPairMethodDBAliClusterRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about dbali cluster information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s, %s from %s where %s = %s  """ \
        %(PianaGlobals.DBAliclusterID_col, PianaGlobals.DBAliclusteringMethod_col,
          PianaGlobals.proteinDBAliCluster_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
     
    
class SelectProteinsWithScop(PianaSelectSQL):
    """
    Generates the sql statement to select proteins that have a particular scop information
    """
    def get_sqlquery( cf, sf, fa):

        # create the condition that the proteins to return have to respect by looking to the cf, sf, fa values
        condition_to_respect = ""
        
        if cf is not None:
            condition_to_respect += "%s=%s" %(PianaGlobals.cf_col, cf)

        if sf is not None:
            if condition_to_respect != "":
                # cf was not None and something was written to condition_to_respect: write the 'and'
                condition_to_respect += " and "
            condition_to_respect += "%s=%s" %(PianaGlobals.sf_col, sf)

        if fa is not None:
            if condition_to_respect != "":
                # cf was not None and something was written to condition_to_respect: write the 'and'
                condition_to_respect += " and "
            condition_to_respect += "%s=%s" %(PianaGlobals.fa_col, fa)

        # use the condition_to_respect to build the sql query that will return all proteinPianas that respect the arguments set by user
        sqlquery = """SELECT %s from %s where %s; """ \
        %(PianaGlobals.proteinPiana_col,
          PianaGlobals.proteinScop_table,
          condition_to_respect)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
   
    
class SelectProteinScopRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about scop information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s,%s,%s from %s where %s = %s """ \
        %(PianaGlobals.cf_col, PianaGlobals.sf_col, PianaGlobals.fa_col,
          PianaGlobals.proteinScop_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
class SelectProteinScopCfRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about scop fold information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(PianaGlobals.cf_col,
          PianaGlobals.proteinScop_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
class SelectProteinScopSfRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about scop superfamily information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(PianaGlobals.sf_col,
          PianaGlobals.proteinScop_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)
    
class SelectProteinScopFaRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about scop family information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(PianaGlobals.fa_col,
          PianaGlobals.proteinScop_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectProteinResiduesCathRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about cath information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s, %s, %s from %s where %s = %s """ \
        %(PianaGlobals.cathID_col, PianaGlobals.res_start_col, PianaGlobals.res_end_col,
          PianaGlobals.proteinCath_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectProteinCathRelationship(PianaSelectSQL):
    """
    Generates the sql statement to select information about cath information
    """
    def get_sqlquery( proteinPiana_value):
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(PianaGlobals.cathID_col,
          PianaGlobals.proteinCath_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectTaxIDfromSpeciesName(PianaSelectSQL):
    """
    Generates the sql statement to select the taxID value for a given speciesName speciesName_value
    """
    def get_sqlquery(speciesName_value= None ):

        if speciesName_value is None:
            raise ValueError("speciesName cannot be None")
        
        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = "%s" """ \
        %(PianaGlobals.taxID_col, PianaGlobals.species_table,
          PianaGlobals.speciesName_col, speciesName_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


class SelectPathWayDescriptionFromReactomeID(PianaSelectSQL):
    """
    Generates the sql statement to select the pathWay description for a given ReactomeID
    """
    def get_sqlquery(reactomeID_value= None) :

        if reactomeID_value is None:
            raise ValueError("taxID cannot be None")

        sqlquery = """SELECT %s from %s where %s = "%s" """ \
                   %(PianaGlobals.reactomeDescription_col, PianaGlobals.reactome_table,
                     PianaGlobals.reactomeID_col, reactomeID_value)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class SelectSpeciesNamefromTaxID(PianaSelectSQL):
    """
    Generates the sql statement to select the speciesName value for a given taxID_value
    """
    def get_sqlquery(taxID_value= None, only_scientific="no"  ):


        if taxID_value is None:
            raise ValueError("taxID_value cannot be None")

        if only_scientific == "yes":
            restriction = """ and speciesDescription="scientific name"  """
        else:
            restriction = ""
            
        sqlquery = """SELECT %s from %s where %s = %s %s""" %(PianaGlobals.speciesName_col,PianaGlobals.species_table,
                                                              PianaGlobals.taxID_col, taxID_value,
                                                              restriction)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectReactomeNamefromReactomeID(PianaSelectSQL):
    """
    Generates the sql statement to select the speciesName value for a given taxID_value
    """
    def get_sqlquery(reactome_id_value= None):


        if reactome_id_value is None:
            raise ValueError("reactome_id_value cannot be None")

        sqlquery = """SELECT %s from %s where %s = "%s" """ %(PianaGlobals.reactomeDescription_col, PianaGlobals.reactome_table,
                                                              PianaGlobals.reactomeID_col, reactome_id_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectProteinKingdom(PianaSelectSQL):
    """
    Generates the sql statement to select the kingdom value for a given proteinPiana
    """
    def get_sqlquery(proteinPiana_value= None ):
        
##         sqlquery = """SELECT s.%s from %s s, %s ps where ps.%s=%s and ps.%s=s.%s """ \
##         %(PianaGlobals.speciesKingdom_col,
##           PianaGlobals.species_table, PianaGlobals.proteinSpecies_table,
##           PianaGlobals.proteinPiana_col, proteinPiana_value,
##           PianaGlobals.taxID_col,PianaGlobals.taxID_col )

        sqlquery = """SELECT s.%s from %s s, %s p where p.%s=%s and s.%s=s.%s """ \
        %(PianaGlobals.speciesKingdom_col,
          PianaGlobals.species_table, PianaGlobals.protein_table,
          PianaGlobals.proteinPiana_col, proteinPiana_value,
          PianaGlobals.taxID_col,PianaGlobals.taxID_col )
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)


#---------------------------------------------------------------------
# INTERACTION RELATED CLASSES
#---------------------------------------------------------------------

def build_sql_db_restriction(list_source_dbs= None, inverse_dbs="no", valid_source_dbs={}, intersection="no", alias=""):
    """
    function used to build a restriction from a list of databases

    this restriction will limit the sql query to retrieve only those interactionPiana that belong to specific databases

    "intersection" is used to make the restriction a combination of all possible databases (only can be used if inverse_dbs="no"
	if it is "no", an OR command will be executed
	if it is "yes", and "AND" command will be executed

    "alias" can be used to set to which table will be restriction be applied
             -> write here the name of the table onto which the restriction will be applied (or its alias)
    """

    # Check if there is any external available database
    if len(valid_source_dbs.keys())==0:
	raise ValueError("Cannot restrict by database as there are not valid source databases")  

    if alias:
        alias = "%s." %alias
        
    search_value=0

    for actual_source in list_source_dbs:
        try:
            search_value += valid_source_dbs[actual_source]
        except KeyError:
            raise ValueError("You cannot restrict by %s database: It does not exist in the database" %(actual_source) )

    if inverse_dbs=="no":
        
        if intersection=="no":
            dbs_restriction = "%s%s & %s" %(alias, PianaGlobals.sourceDBID_col, search_value)
            
        elif intersection=="yes":
            # before dbs_restriction += "%s & %s = %s" %(PianaGlobals.sourceDBID_col, search_value, search_value)
            dbs_restriction = "%s%s & %s = %s" %(alias, PianaGlobals.sourceDBID_col, search_value, search_value)
            
        else:
            raise ValueError("intersection argument must be \"yes\" or \"no\"")
    else:
        dbs_restriction = "%s%s & ~%s" %(alias, PianaGlobals.sourceDBID_col, search_value)

    return dbs_restriction


    # depending on whether the user wants interactions from a set of dbs, or all interactions except for those in a set of dbs, the
    # sign and logic to use when building the restriction is different
    #if inverse_dbs == "yes":
    #    sign_to_use = "!="
    #    logic_to_use = "and"
    #else:
    #    sign_to_use = "="
    #    logic_to_use = "or"
    #
    #dbs_restriction = "("
    
    #if not list_source_dbs or not isinstance(list_source_dbs, list):
    #    raise ValueError("Doesn't make sense to restrict interactions to those from list %s source database" %(list_source_dbs) )
    #else:

        # build sql restriction to get only interactions from the list of databases in list_source_dbs
        #   -> to build restriction, start with first database and add the others afterwards
    #    dbs_restriction += """ s.sourceDBID%s"%s" """ %(sign_to_use, list_source_dbs[0])

    #    if len(list_source_dbs) > 1:
    #        for database in list_source_dbs[1:]:
    #            dbs_restriction += """ %s s.sourceDBID%s"%s" """ %(logic_to_use, sign_to_use, database)
            # END OF for database in list_source_dbs[1:]:
        # END OF if len(list_source_dbs) >1:
    # END OF else: (if not list_source_dbs:)

    #dbs_restriction += ")"

    return dbs_restriction

def build_sql_method_restriction(list_source_methods= None, inverse_methods="no"):
    """
    function used to build a restriction from a list of methods

    this restriction will limit the sql query to retrieve only those interactionPiana that where found by specific methods
    """

    # depending on whether the user wants interactions from a set of methods, or all interactions except for those in a set of methods, the
    # sign and logic to use when building the restriction is different
    if inverse_methods == "yes":
        sign_to_use = "!="
        logic_to_use = "and"
    else:
        sign_to_use = "="
        logic_to_use = "or"
        
    methods_restriction = "("
    if not list_source_methods or not isinstance(list_source_methods, list):
        raise ValueError("Doesn't make sense to restrict interactions to those from list %s source methods" %list_source_methods)
    else:

        # build sql restriction to get only interactions from the list of methods in list_source_methods
        #   -> to build restriction, start with first database and add the others afterwards
#        methods_restriction += """ %s.methodID%s"%s" """ %(PianaGlobals.methodID_col, sign_to_use, list_source_methods[0])
        methods_restriction += """ m.%s%s"%s" """ %(PianaGlobals.methodID_col, sign_to_use, list_source_methods[0])

        if len(list_source_methods) > 1:
            for method in list_source_methods[1:]:
                methods_restriction += """ %s m.%s%s"%s" """ %(logic_to_use, PianaGlobals.methodID_col, sign_to_use, method)                
#                methods_restriction += """ %s %s.methodID%s"%s" """ %(logic_to_use, PianaGlobals.methodID_col, sign_to_use, method)
            # END OF for method in list_source_methods[1:]:
        # END OF if len(list_source_methods) >1:
    # END OF else: (if not list_source_methods:)

    methods_restriction += ")"

    return methods_restriction

class SelectListInteractionPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains interaction internal  ids where a given protein is involved
    """
    def get_sqlquery(proteinPiana_value, list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no", valid_source_dbs={}):

        # instead of generalizing with one single query, speed up things by separating queries for all databases and those
        # that need to take into account the source database for the interaction

        if verbose:
            sys.stderr.write("in PianaSelectSQL: list source dbs is %s and list source methods is %s\n" %(list_source_dbs,list_source_methods))

        if list_source_dbs == "all" and list_source_methods == "all":
            # ----------------------------------------------
            # no restrictions to apply: get all interactions
            # ----------------------------------------------
            sqlquery = """select %s from %s where %s=%s or %s=%s""" %(PianaGlobals.interactionPiana_col, PianaGlobals.interaction_table
                                                                      , PianaGlobals.proteinPianaA_col, proteinPiana_value
                                                                      , PianaGlobals.proteinPianaB_col, proteinPiana_value)

        else:
            # ------------------------------------
            # there are restrictions to be applied
            # ------------------------------------

            # set them to "no": will be changed depending on which list is not empty
            apply_db_restriction= "no"
            apply_method_restriction= "no"

            # build restrictions for lists that have elements in them
            if list_source_dbs != "all" and list_source_dbs:
                db_restriction = build_sql_db_restriction(list_source_dbs= list_source_dbs,
                                                          inverse_dbs=inverse_dbs,
							  valid_source_dbs=valid_source_dbs,
                                                          alias = "i")
                apply_db_restriction = "yes"
                
            if list_source_methods != "all" and list_source_methods:
                method_restriction = build_sql_method_restriction(list_source_methods,
                                                                  inverse_methods=inverse_methods)
                apply_method_restriction = "yes"


            if apply_db_restriction == "yes" and apply_method_restriction == "yes":

                #
                # SQL query when there are source db and method restrictions
                #
                sqlquery = """select i.%s from %s i, %s m where (i.%s=%s or i.%s=%s) and ( (m.%s=i.%s) """ %(
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interaction_table,
                    PianaGlobals.interactionMethod_table,
                    PianaGlobals.proteinPianaA_col,
                    proteinPiana_value,
                    PianaGlobals.proteinPianaB_col,
                    proteinPiana_value,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col )


                sqlquery += "and " + db_restriction
                sqlquery += " and " + method_restriction
                
            elif  apply_db_restriction == "yes":
                
                #
                # SQL query when there is source db restriction
                #
                sqlquery = """select i.%s from %s i where (i.%s=%s or i.%s=%s) and (  """ %(
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interaction_table,
                    PianaGlobals.proteinPianaA_col,
                    proteinPiana_value,
                    PianaGlobals.proteinPianaB_col,
                    proteinPiana_value)


                sqlquery += db_restriction
                
            elif  apply_method_restriction == "yes":
                
                #
                # SQL query when there is source method restriction
                #
                sqlquery = """select i.%s from %s i, %s m where (i.%s=%s or i.%s=%s) and ( (m.%s=i.%s) """ %(
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interaction_table,
                    PianaGlobals.interactionMethod_table,
                    PianaGlobals.proteinPianaA_col,
                    proteinPiana_value,
                    PianaGlobals.proteinPianaB_col,
                    proteinPiana_value,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col)
                
                sqlquery += "and " + method_restriction
                
            # END OF elif  apply_method_restriction == "yes": (if apply_db_restriction == "yes" and apply_method_restriction == "yes":)
            else:
                raise ValueError("Trying to apply restrictions with empty lists of source dbs (%s) and methods (%s)" %(list_source_dbs, list_source_methods)) 

            sqlquery += ")"
            
        # END OF else (if list_source_dbs == "all" and list_source_methods == "all":)

        if verbose:
            sys.stderr.write("SelectListInteractionPiana returning query \n%s\n" %sqlquery)
 
        return sqlquery
     
    get_sqlquery = staticmethod(get_sqlquery)

class SelectInteractionPiana(PianaSelectSQL):
    """
    Generates the SQL statement that obtains interaction internal id interactionPiana for the interaction between two specific proteins

    Before looking for the interaction, it makes sure that the order proteinPianaA < proteinPianaB is respected. If it is not the case, it
    changes the query (not the objects passed, they remain unchanged) to execute it in the right order
    """
    def get_sqlquery(proteinPianaA_value, proteinPianaB_value, list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",valid_source_dbs={}):
 
        if proteinPianaA_value > proteinPianaB_value:
            temp = proteinPianaA_value
            proteinPianaA_value = proteinPianaB_value
            proteinPianaB_value = temp

        if list_source_dbs == "all" and list_source_methods == "all":
            # ---------------------------------------------------------------------------------
            # no restrictions to apply: get interactionPiana regardless of db and method source
            # ---------------------------------------------------------------------------------
            sqlquery = """select %s from %s where (%s=%s) and (%s=%s)""" %(PianaGlobals.interactionPiana_col, PianaGlobals.interaction_table
                                                                           , PianaGlobals.proteinPianaA_col, proteinPianaA_value
                                                                           , PianaGlobals.proteinPianaB_col, proteinPianaB_value)
        else:
            # ------------------------------------
            # there are restrictions to be applied
            # ------------------------------------

            #set them to "no": will be changed depending on which list is not empty
            apply_db_restriction= "no"
            apply_method_restriction= "no"

            # build restrictions for lists that have elements in them
            if list_source_dbs != "all" and list_source_dbs:
                db_restriction = build_sql_db_restriction(list_source_dbs,
                                                          inverse_dbs=inverse_dbs,
							  valid_source_dbs=valid_source_dbs,
                                                          alias="%s" %(PianaGlobals.interaction_table))
                apply_db_restriction = "yes"
                
            if list_source_methods != "all" and list_source_methods:
                method_restriction = build_sql_method_restriction(list_source_methods,
                                                                  inverse_methods=inverse_methods)
                apply_method_restriction = "yes"

            if apply_db_restriction == "yes" and apply_method_restriction == "yes":

                #
                # SQL query when there are source db and method restrictions
                #
                #sqlquery = """select i.%s from %s i, %s s, %s m where (i.%s=%s and i.%s=%s) and ( (s.%s=i.%s and m.%s=i.%s) """ %(
                #   PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interaction_table,
                #    PianaGlobals.interactionSourceDB_table,
                #    PianaGlobals.interactionMethod_table,
                #    PianaGlobals.proteinPianaA_col,
                #    proteinPianaA_value,
                #    PianaGlobals.proteinPianaB_col,
                #    proteinPianaB_value,
                #    PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interactionPiana_col  )


                #sqlquery += "and " + db_restriction
                #sqlquery += "and " + method_restriction

		sqlquery = """select %s from %s,%s where (%s.%s=%s and %s.%s=%s) and (%s.%s=%s.%s) """ %(
		   PianaGlobals.interactionPiana_col,
		   PianaGlobals.interaction_table,
		   PianaGlobals.interactionMethod_table,
		   PianaGlobasl.interaction_table,
		   PianaGlobals.proteinPianaA_col,
		   proteinPianaA_value,
		   PianaGlobals.interaction_table,
		   PianaGlobals.proteinPianaB_col,
		   proteinPianaB_value,
		   PianaGlobals.interaction_table,
		   PianaGlobals.interactionPiana_col,
		   PianaGlobals.interactionMethod_table,
		   PianaGlobals.interactionPiana_col  )
		
		sqlquery += "and " + db_restriction    
		sqlquery += " and " + method_restriction    
		
                
            elif  apply_db_restriction == "yes":
                
                #
                # SQL query when there is source db restriction
                #
                #sqlquery = """select i.%s from %s i, %s s where (i.%s=%s and i.%s=%s) and ( (s.%s=i.%s) """ %(
                #    PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interaction_table,
                #    PianaGlobals.interactionSourceDB_table,
                #    PianaGlobals.proteinPianaA_col,
                #    proteinPianaA_value,
                #    PianaGlobals.proteinPianaB_col,
                #    proteinPianaB_value,
                #    PianaGlobals.interactionPiana_col,
                #    PianaGlobals.interactionPiana_col)


                #sqlquery += "and " + db_restriction

                sqlquery = """select %s from %s,%s where (%s.%s=%s and %s.%s=%s) """ %(
                   PianaGlobals.interactionPiana_col,
                   PianaGlobals.interaction_table,
                   PianaGlobals.interactionMethod_table,
                   PianaGlobasl.interaction_table,
                   PianaGlobals.proteinPianaA_col,
                   proteinPianaA_value,
                   PianaGlobals.interaction_table,
                   PianaGlobals.proteinPianaB_col,
                   proteinPianaB_value)

		sqlquery += "and " + db_restriction
   
                
            elif  apply_method_restriction == "yes":
                
                #
                # SQL query when there is source method restriction
                #
                sqlquery = """select i.%s from %s i, %s m where (i.%s=%s and i.%s=%s) and ( (m.%s=i.%s) """ %(
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interaction_table,
                    PianaGlobals.interactionMethod_table,
                    PianaGlobals.proteinPianaA_col,
                    proteinPianaA_value,
                    PianaGlobals.proteinPianaB_col,
                    proteinPianaB_value,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col)
                
                sqlquery += "and " + method_restriction
                
            # END OF elif  apply_method_restriction == "yes": (if apply_db_restriction == "yes" and apply_method_restriction == "yes":)
            else:
                raise ValueError("Trying to apply restrictions with empty lists of source dbs (%s) and methods (%s)" %(list_source_dbs, list_source_methods)) 

            sqlquery += ")"
            
        # END OF else: (if list_source_dbs == "all" and list_source_methods == "all":)
        
        return sqlquery
     
    get_sqlquery = staticmethod(get_sqlquery)

class SelectInteractors(PianaSelectSQL):
    """
    Generates the SQL statement that obteins the interactors for a given interactionPiana)
    """
    def get_sqlquery(interactionPiana_value):

        sqlquery = """SELECT %s,%s FROM %s WHERE %s=%s""" %(PianaGlobals.proteinPianaA_col,PianaGlobals.proteinPianaB_col,
                                                            PianaGlobals.interaction_table, PianaGlobals.interactionPiana_col, interactionPiana_value)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)
                                                            

class SelectAllProteinProteinInteractions(PianaSelectSQL):
    """
    Generates the SQL statement that obtains all triplets (proteinPianaA, proteinPianaB, interactionPiana) for list_source_dbs and list_source_methods

    if only_id is "yes", then it only returns the sql query for retrieving interactionPianas (and not the proteinPianas of the interaction)

    "tax_id" can only contain a taxID value. It can be "all" or a taxID value
    """
    def get_sqlquery(use_self_ints="yes", list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                     only_id = "yes", valid_source_dbs={}, tax_id_value="all"):


        if list_source_dbs == "all" and list_source_methods == "all" and tax_id_value == "all":
            # ----------------------------------------------------------------------------------------------
            # no db or method restrictions to apply: get interactionPiana regardless of db and method source
            # ----------------------------------------------------------------------------------------------

            if use_self_ints == "no":
                # if user asked not to use self interactions, build restriction that does it..
                self_restriction = " where %s!=%s" %(PianaGlobals.proteinPianaA_col, PianaGlobals.proteinPianaB_col)
            else:
                # no restriction to apply
                self_restriction = ""

            
            if only_id == "yes":
                sqlquery = """select %s from %s %s""" %(PianaGlobals.interactionPiana_col,
                                                        PianaGlobals.interaction_table, self_restriction)
                
            else:
                sqlquery = """select %s, %s, %s from %s %s""" %(PianaGlobals.proteinPianaA_col, PianaGlobals.proteinPianaB_col,
                                                                PianaGlobals.interactionPiana_col,
                                                                PianaGlobals.interaction_table, self_restriction)
 
        else:
            # ------------------------------------
            # there are restrictions to be applied
            # ------------------------------------

            if use_self_ints == "no":
                # if user asked not to use self interactions, build restriction that does it..
                self_restriction = " and %s!=%s" %(PianaGlobals.proteinPianaA_col, PianaGlobals.proteinPianaB_col)
            else:
                # no restriction to apply
                self_restriction = ""

            #set them to "no": will be changed depending on which list is not empty
            apply_db_restriction= "no"
            apply_method_restriction= "no"
            apply_taxid_restriction= "no"

            # build restrictions for lists that have elements in them
            if list_source_dbs != "all" and list_source_dbs:
                db_restriction = build_sql_db_restriction(list_source_dbs,
                                                          inverse_dbs=inverse_dbs,
							  valid_source_dbs=valid_source_dbs,
                                                          alias="i")
                apply_db_restriction = "yes"
                
            if list_source_methods != "all" and list_source_methods:
                method_restriction = build_sql_method_restriction(list_source_methods,
                                                                  inverse_methods=inverse_methods)
                apply_method_restriction = "yes"

            if tax_id_value != "all":
                tax_id_restriction_tables = ",%s p1,%s p2 " %(PianaGlobals.protein_table,PianaGlobals.protein_table)
                #javi modified
                tax_id_restriction_where =  " p1.%s=%s AND p2.%s=%s AND i.%s=p1.%s AND i.%s=p2.%s " %(PianaGlobals.taxID_col,tax_id_value,
                                                                                                      PianaGlobals.taxID_col,tax_id_value,
                                                                                                      PianaGlobals.proteinPianaA_col,PianaGlobals.proteinPiana_col,
                                                                                                      PianaGlobals.proteinPianaB_col,PianaGlobals.proteinPiana_col)
                apply_taxid_restriction= "yes"
            else:
                tax_id_restriction_tables = ""
                tax_id_restriction_where = ""

            if apply_db_restriction == "yes" and apply_method_restriction == "yes":

                #
                # SQL query when there are source db and method restrictions
                #
                if only_id == "yes":
                    sqlquery = """select i.%s from %s i, %s m %s where  ( (m.%s=i.%s) """ %(
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        PianaGlobals.interactionMethod_table,
                        tax_id_restriction_tables,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interactionPiana_col  )

                else:
                    sqlquery = """select i.%s, i.%s, i.%s from %s i, %s m %s where  ( (m.%s=i.%s) """ %(
                        PianaGlobals.proteinPianaA_col,
                        PianaGlobals.proteinPianaB_col,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        PianaGlobals.interactionMethod_table,
                        tax_id_restriction_tables,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interactionPiana_col  )


                sqlquery += "and " + db_restriction
                sqlquery += " and " + method_restriction
                #javi modified:
                if apply_taxid_restriction == "yes":
                    sqlquery += " and " + tax_id_restriction_where
                #sqlquery += tax_id_restriction_where
                sqlquery += self_restriction
                
            elif  apply_db_restriction == "yes":

                #
                # SQL query when there is source db restriction
                #
                if only_id == "yes":
                    sqlquery = """select i.%s  from %s i %s where ( """ %(
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        tax_id_restriction_tables)
                    
                else:
                    sqlquery = """select i.%s, i.%s, i.%s  from %s i %s where ( """ %(
                        PianaGlobals.proteinPianaA_col,
                        PianaGlobals.proteinPianaB_col,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        tax_id_restriction_tables)


                sqlquery += db_restriction
                sqlquery += self_restriction
                #javi modified
                #sqlquery += tax_id_restriction_where
                if apply_taxid_restriction == "yes":
                    sqlquery += " and " + tax_id_restriction_where
                
            elif  apply_method_restriction == "yes":
                
                #
                # SQL query when there is source method restriction
                #
                if only_id == "yes":
                    sqlquery = """select i.%s  from %s i, %s m %s where ( (m.%s=i.%s) """ %(
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        PianaGlobals.interactionMethod_table,
                        tax_id_restriction_tables,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interactionPiana_col)
                else:
                    sqlquery = """select i.%s, i.%s, i.%s  from %s i, %s m %s where ( (m.%s=i.%s) """ %(
                        PianaGlobals.proteinPianaA_col,
                        PianaGlobals.proteinPianaB_col,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        PianaGlobals.interactionMethod_table,
                        tax_id_restriction_tables,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interactionPiana_col)
                    
                sqlquery += "and " + method_restriction
                #javi modified
                if apply_taxid_restriction == "yes":
                    sqlquery += " and " + tax_id_restriction_where
                #sqlquery += tax_id_restriction_where
                sqlquery += self_restriction
                
            # END OF elif  apply_method_restriction == "yes": (if apply_db_restriction == "yes" and apply_method_restriction == "yes":)
            elif tax_id_value != "all":
                #
                # SQL query when there is not source db nor method restriction, but yes tax-id
                #
                if only_id == "yes":
                    sqlquery = """select i.%s  from %s i %s where ( """ %(
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        tax_id_restriction_tables)
                    
                else:
                    sqlquery = """select i.%s, i.%s, i.%s  from %s i %s where ( """ %(
                        PianaGlobals.proteinPianaA_col,
                        PianaGlobals.proteinPianaB_col,
                        PianaGlobals.interactionPiana_col,
                        PianaGlobals.interaction_table,
                        tax_id_restriction_tables)

                sqlquery += tax_id_restriction_where
                sqlquery += self_restriction
            else:
                raise ValueError("Trying to apply restrictions with empty lists of source dbs (%s) and methods (%s)" %(list_source_dbs, list_source_methods)) 

            sqlquery += ")"
            
        # END OF else: (if list_source_dbs == "all" and list_source_methods == "all":)

        print sqlquery
        return sqlquery
     
    get_sqlquery = staticmethod(get_sqlquery)



class SelectPartners(PianaSelectSQL):
    """
    Generates the SQL statement that obtains interaction partners of a given protein


        Used to return the SQL query that retrieves the interaction partners of a protein (only those on side A or B)
        
        This class is not to be used externally: the side (A or B)
        where the protein is placed should be trasparent to the
        external user. Piana makes this trasparent to the user by
        creating a method (get_all_partners) under PianaDBaccess that
        performs two queries to obtain all partners.  PianaDBaccess.get_all_partners
        will call twice select_db_content with SelectPartners
        

        partner_side is either "A" or "B", indicating if the partner will be searched on side A or B.
        this is done this way because there is no single sqlquery that will return a complete list
        of partners of a given protein, given the fact that those partners can be on side A or B

        this is an artifact of the database design: an interaction is formed by a sideA and a sideB.
 
    """
    def get_sqlquery(proteinPiana_value, partner_side, list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",valid_source_dbs={},get_interactionPiana="no"):

        if partner_side == "A":

            partner_column = PianaGlobals.proteinPianaA_col
            this_protein_column = PianaGlobals.proteinPianaB_col
            
        elif partner_side == "B":

            partner_column = PianaGlobals.proteinPianaB_col
            this_protein_column = PianaGlobals.proteinPianaA_col
        else:
            raise ValueError("partner_side incorrect: it must be either A or B")

        if get_interactionPiana=="yes":
            interactionPiana_cond = ",i.%s" %(PianaGlobals.interactionPiana_col)
            
        elif get_interactionPiana == "no":
            interactionPiana_cond = ""
        else:
            raise ValueError("get_interactionPiana must be \"yes\" or \"no\"")
    
        
        if list_source_dbs == "all" and list_source_methods == "all":

            # -------------------------------------------------------------------------
            # no restrictions to apply: get partners regardless of db and method source
            # -------------------------------------------------------------------------
            
            sqlquery = """select i.%s%s from %s i where i.%s = %s""" %(partner_column, interactionPiana_cond, PianaGlobals.interaction_table,
                                                                       this_protein_column , proteinPiana_value)
            
        else:

            # ------------------------------------
            # there are restrictions to be applied
            # ------------------------------------

            #set them to "no": will be changed depending on which list is not empty
            apply_db_restriction= "no"
            apply_method_restriction= "no"

            # build restrictions for lists that have elements in them
            if list_source_dbs != "all" and list_source_dbs:
                db_restriction = build_sql_db_restriction(list_source_dbs,
                                                          inverse_dbs=inverse_dbs,
							  valid_source_dbs=valid_source_dbs,
                                                          alias="i")
                apply_db_restriction = "yes"
                
            if list_source_methods != "all" and list_source_methods:
                method_restriction = build_sql_method_restriction(list_source_methods,
                                                                  inverse_methods=inverse_methods)
                apply_method_restriction = "yes"

            if apply_db_restriction == "yes" and apply_method_restriction == "yes":

                #
                # SQL query when there are source db and method restrictions
                #
   
                sqlquery = """ select i.%s%s from %s i, %s m where (i.%s = %s) and ( (m.%s=i.%s) """ %(
                    partner_column,
                    interactionPiana_cond,
                    PianaGlobals.interaction_table,
                    PianaGlobals.interactionMethod_table,
                    this_protein_column,
                    proteinPiana_value,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col)


                sqlquery += "and " + db_restriction
                sqlquery += " and " + method_restriction

                      
            elif  apply_db_restriction == "yes":
                
                #
                # SQL query when there is source db restriction
                #
   
                sqlquery = """ select i.%s%s from %s i where (i.%s = %s) and ( """ %(
                    partner_column,
                    interactionPiana_cond,
                    PianaGlobals.interaction_table,
                    this_protein_column,
                    proteinPiana_value)


                sqlquery += db_restriction
          
                  
            elif  apply_method_restriction == "yes":
                
                #
                # SQL query when there is source method restriction
                #
   
                sqlquery = """ select i.%s%s from %s i, %s m where (i.%s = %s) and ( (m.%s=i.%s) """ %(
                    partner_column,
                    interactionPiana_cond,
                    PianaGlobals.interaction_table,
                    PianaGlobals.interactionMethod_table,
                    this_protein_column,
                    proteinPiana_value,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col )


                sqlquery += "and " + method_restriction
                 
            # END OF elif  apply_method_restriction == "yes": (if apply_db_restriction == "yes" and apply_method_restriction == "yes":)
            else:
                raise ValueError("Trying to apply restrictions with empty lists of source dbs (%s) and methods (%s)" %(list_source_dbs, list_source_methods)) 

            sqlquery += ")"
            
        # END OF else: (if list_source_dbs == "all" and list_source_methods == "all":)

        if verbose:
            sys.stderr.write("SelectPartners returning query \n%s\n" %sqlquery)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectPartnersProteinSharingCog(PianaSelectSQL):
    """
    Generates the SQL statement that obtains interaction partners for proteins that share cog with a certain proteinPiana

        This class is not to be used externally: the side (A or B)
        where the protein is placed should be trasparent to the
        external user. Piana makes this trasparent to the user by
        creating a method (get_partners_of_proteins_sharing_cog) under PianaDBaccess that
        performs two queries to obtain all partners.  PianaDBaccess.get_partners_of_proteins_sharing_cog
        will call twice select_db_content with SelectPartnersProteinSharingCog
        

        partner_side is either "A" or "B", indicating if the partner will be searched on side A or B.
        this is done this way because there is no single sqlquery that will return a complete list
        of partners of a given protein, given the fact that those partners can be on side A or B

        this is an artifact of the database design: an interaction is formed by a sideA and a sideB.

    """
    def get_sqlquery(proteinPiana_value, partner_side, list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",valid_source_dbs={}):

        if partner_side == "A":

            partner_column = PianaGlobals.proteinPianaA_col
            this_protein_column = PianaGlobals.proteinPianaB_col
            
        elif partner_side == "B":

            partner_column = PianaGlobals.proteinPianaB_col
            this_protein_column = PianaGlobals.proteinPianaA_col
        else:
            raise ValueError("partner_side incorrect: it must be either A or B")

        if list_source_dbs == "all" and list_source_methods == "all":

            # -------------------------------------------------------------------------
            # no restrictions to apply: get partners regardless of db and method source
            # -------------------------------------------------------------------------
            
            sqlquery = """select i.%s from %s i, %s pc1, %s pc2 where pc1.%s = %s and pc1.%s = pc2.%s and pc2.%s = i.%s""" %(
                partner_column,
                PianaGlobals.interaction_table,
                PianaGlobals.proteinCog_table,
                PianaGlobals.proteinCog_table,
                PianaGlobals.proteinPiana_col,
                proteinPiana_value,
                PianaGlobals.cogID_col,
                PianaGlobals.cogID_col,
                PianaGlobals.proteinPiana_col,
                this_protein_column )
            
        else:
            # ------------------------------------
            # there are restrictions to be applied
            # ------------------------------------

            #set them to "no": will be changed depending on which list is not empty
            apply_db_restriction= "no"
            apply_method_restriction= "no"

            # build restrictions for lists that have elements in them
            if list_source_dbs != "all" and list_source_dbs:
                db_restriction = build_sql_db_restriction(list_source_dbs,
                                                          inverse_dbs=inverse_dbs,
							  valid_source_dbs={},
                                                          alias="i")
                apply_db_restriction = "yes"
                
            if list_source_methods != "all" and list_source_methods:
                method_restriction = build_sql_method_restriction(list_source_methods,
                                                                  inverse_methods=inverse_methods)
                apply_method_restriction = "yes"

            if apply_db_restriction == "yes" and apply_method_restriction == "yes":

                #
                # SQL query when there are source db and method restrictions
                #

                sqlquery = """select i.%s from %s i, %s pc1, %s pc2, %s m where (pc1.%s = %s and pc1.%s = pc2.%s and pc2.%s = i.%s) \
                and ( (m.%s=i.%s)""" %(
                    partner_column,
                    PianaGlobals.interaction_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.interactionMethod_table,
                    PianaGlobals.proteinPiana_col,
                    proteinPiana_value,
                    PianaGlobals.cogID_col,
                    PianaGlobals.cogID_col,
                    PianaGlobals.proteinPiana_col,
                    this_protein_column,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col )

                sqlquery += "and " + db_restriction
                sqlquery += " and " + method_restriction
                      
            elif  apply_db_restriction == "yes":
                
                #
                # SQL query when there is source db restriction
                #

                sqlquery = """select i.%s from %s i, %s pc1, %s pc2 where (pc1.%s = %s and pc1.%s = pc2.%s and pc2.%s = i.%s) \
                and (""" %(
                    partner_column,
                    PianaGlobals.interaction_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.proteinPiana_col,
                    proteinPiana_value,
                    PianaGlobals.cogID_col,
                    PianaGlobals.cogID_col,
                    PianaGlobals.proteinPiana_col,
                    this_protein_column ) 


                sqlquery += "and " + db_restriction
                  
            elif  apply_method_restriction == "yes":
                
                #
                # SQL query when there is source method restriction
                #

                sqlquery = """select i.%s from %s i, %s pc1, %s pc2, %s m where (pc1.%s = %s and pc1.%s = pc2.%s and pc2.%s = i.%s) \
                and ( (m.%s=i.%s)""" %(
                    partner_column,
                    PianaGlobals.interaction_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.proteinCog_table,
                    PianaGlobals.interactionMethod_table,
                    PianaGlobals.proteinPiana_col,
                    proteinPiana_value,
                    PianaGlobals.cogID_col,
                    PianaGlobals.cogID_col,
                    PianaGlobals.proteinPiana_col,
                    this_protein_column,
                    PianaGlobals.interactionPiana_col,
                    PianaGlobals.interactionPiana_col  )


                sqlquery += "and " + method_restriction

            # END OF elif  apply_method_restriction == "yes": (if apply_db_restriction == "yes" and apply_method_restriction == "yes":)
            else:
                raise ValueError("Trying to apply restrictions with empty lists of source dbs (%s) and methods (%s)" %(list_source_dbs, list_source_methods)) 

            sqlquery += ")"
            
        # END OF else: (if list_source_dbs == "all" and list_source_methods == "all":)

        if verbose:
            sys.stderr.write("SelectPartnersProteinSharingCog returning query \n%s\n" %sqlquery)
           
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectInteractionPartner(PianaSelectSQL):
    """
    Generates the SQL statement that obtains the interaction partner of a protein for a given interactionID
    """

    def get_sqlquery(proteinPiana_value, interactionPiana_value, partner_side):

        if partner_side == "A":

            partner_column = PianaGlobals.proteinPianaA_col
            this_protein_column = PianaGlobals.proteinPianaB_col
            
        elif partner_side == "B":

            partner_column = PianaGlobals.proteinPianaB_col
            this_protein_column = PianaGlobals.proteinPianaA_col
        else:
            raise ValueError("partner_side incorrect: it must be either A or B")

        
        sqlquery = """select %s from %s where (%s=%s) and (%s=%s)""" \
                   %(partner_column, PianaGlobals.interaction_table
                     ,this_protein_column , proteinPiana_value, PianaGlobals.interactionPiana_col, interactionPiana_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectInteractionSourceDB(PianaSelectSQL):
    """
    Generates the SQL statement that obtains entries from table interaction in relation with its sourceDB

    if interactionPiana_value not None --> get the list of sourceDBID for this interactionPiana

    if sourceDBID_value not None --> get the list of interactionPiana for this sourceDBID

    Only one of the arguments can be "not None"
    """

    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None):

        if interactionPiana_value is not None and sourceDBID_value is not None:
            raise ValueError("in SelectInteractionSourceDB, both arguments cannot be 'not None'")
        
        elif interactionPiana_value is not None:
            sqlquery = """select %s from %s where (%s=%s)""" %(PianaGlobals.sourceDBID_col, PianaGlobals.interaction_table,
                                                               PianaGlobals.interactionPiana_col, interactionPiana_value)

        elif sourceDBID_value is not None:
            sqlquery = """select %s from %s where (%s="%s")""" %(PianaGlobals.interactionPiana_col, PianaGlobals.interaction_table,
                                                                 PianaGlobals.sourceDBID_col, sourceDBID_value)
        else:
            raise ValueError("in SelectInteractionSourceDB, both arguments cannot be 'None' Why here????")
            

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectInteractionMethod(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list  of methodID of an interactionPiana

    if sourceDB_value == None, then it returns the SQL statement that obtains all methodIDs for interactionPiana
    if sourceDB_value is a database, then it returns the SQL statement that obtains all methodIDs for interactionPiana for a particular sourceDB

    Attention! sourceDB_value is the integer associated to the source DB

    Attention: list_source_dbs and inverse_dbs not being used!!! 
    """

    # TO DO!!! Remove sourceDBID and get methods for all dbs in list_source_dbs (taking into account inverse_dbs)
    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None, list_source_dbs= "all", inverse_dbs="no"):

        if interactionPiana_value is not None and sourceDBID_value is not None:
            # return methods for a given interaction in a particular database

 
            # TO CHECK!!! Is this correct?
            #              I think that sourceDB & sourceDBID returns methods for a given interactionPiana that appears in a certain sourceDB
            
            sqlquery = """select %s from %s where (%s=%s) and (%s&%s) """ %(PianaGlobals.methodID_col, PianaGlobals.interactionMethod_table,
                                                                              PianaGlobals.interactionPiana_col, interactionPiana_value ,
                                                                              PianaGlobals.sourceDBID_col, sourceDBID_value )

            # code before adapating to new sourceDBID way of managing
            #sqlquery = """select %s from %s where (%s=%s) and (%s="%s") """ %(PianaGlobals.methodID_col, PianaGlobals.interactionMethod_table,
            #                                                                  PianaGlobals.interactionPiana_col, interactionPiana_value ,
            #                                                                  PianaGlobals.sourceDBID_col, sourceDBID_value )

        elif interactionPiana_value is not None:
            # return methods for a given interaction, regardless of the database
            sqlquery = """select %s from %s where (%s=%s)""" %(PianaGlobals.methodID_col, PianaGlobals.interactionMethod_table
                                                               , PianaGlobals.interactionPiana_col, interactionPiana_value)

        else:
            raise ValueError("in SelectInteractionMethod, interactionPiana_value cannot be 'None' ")
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)



class SelectInteractionPubmed(PianaSelectSQL):
    """
    Generates the SQL statement that obtains a list  of pubmedID of an interactionPiana

    if sourceDB_value == None, then it returns the SQL statement that obtains all pubmedIDs for interactionPiana
    if sourceDB_value is a database, then it returns the SQL statement that obtains all pubmedIDs for interactionPiana for a particular sourceDB

    Attention! sourceDB_value is the integer associated to the source DB

    Attention: list_source_dbs and inverse_dbs not being used!!! 
    """

    def get_sqlquery(interactionPiana_value=None, sourceDBID_value=None, list_source_dbs= "all", inverse_dbs="no"):

        if interactionPiana_value is not None and sourceDBID_value is not None:
            # return pubmeds for a given interaction in a particular database

 
            # TO CHECK!!! Is this correct?
            #              I think that sourceDB & sourceDBID returns methods for a given interactionPiana that appears in a certain sourceDB
            
            sqlquery = """select %s from %s where (%s=%s) and (%s&%s) """ %(PianaGlobals.pubmedID_col, PianaGlobals.interactionFeatures_table,
                                                                            PianaGlobals.interactionPiana_col, interactionPiana_value ,
                                                                            PianaGlobals.sourceDBID_col, sourceDBID_value )
        elif interactionPiana_value is not None:
            # return methods for a given interaction, regardless of the database
            sqlquery = """select %s from %s where (%s=%s)""" %(PianaGlobals.pubmedID_col, PianaGlobals.interactionFeatures_table
                                                               , PianaGlobals.interactionPiana_col, interactionPiana_value)

        else:
            raise ValueError("in SelectInteractionPubmed, interactionPiana_value cannot be 'None' ")
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

  
class SelectXFromGoTermId(PianaSelectSQL):
    """
    Generates the sql statement to select something (proteinPiana_col or go_acc_col)
    from one of the two go tables (proteinGo_table or go_table) from a
    given GO term_id
    """
    def get_sqlquery( x_col, term_id_value):

        # Those %s referring to character types should be surrounded by quotes. Otherwise, mysql gives error
        sqlquery = """SELECT %s from %s where %s = %s """ \
        %(x_col, PianaGlobals.go_table,
          PianaGlobals.goID_col, term_id_value)
        
        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class SelectGoTermIdUsingTermType(PianaSelectSQL):
    """
    Generates the sql statement to select go term id associated to a proteinPiana specifing the term type value
    because one proteinPiana could be associated to more than one Go term id depending the term type:

          - "molecular_function"
          - "biological_process"
          - "cellular_component"
    """
    def get_sqlquery( proteinPiana,term_type):

        sqlquery= """SELECT protGo.%s from  %s protGo, %s go \
        where protGo.%s = %s  and protGo.%s = go.%s and go.%s = "%s" """ \
        %(PianaGlobals.goID_col, PianaGlobals.proteinGo_table, PianaGlobals.go_table,
          PianaGlobals.proteinPiana_col, proteinPiana,
          PianaGlobals.goID_col,PianaGlobals.goID_col,
          PianaGlobals.go_term_type_col, term_type)

        return sqlquery
    
    get_sqlquery = staticmethod(get_sqlquery)

class Select_go_term2term_distance(PianaSelectSQL):
   """
   Generates the sql statement to select the distance between two go term ids
   """
   def get_sqlquery(term1_id, term2_id):
       
       if term1_id is None or term2_id is None:
           raise ValueError("Either term1(%s) or term2(%s) are None" %(term1_id, term2_id))
       
       if term1_id > term2_id:
           temp = term1_id
           term1_id = term2_id
           term2_id = temp
           
       sqlquery="""SELECT %s from %s where %s = %s and %s = %s """\
                 %(PianaGlobals.go_term2term_distance_col, PianaGlobals.go_term2term_distance_table,
                   PianaGlobals.go_term2term_distance_term1_col, term1_id, PianaGlobals.go_term2term_distance_term2_col, term2_id)

       return sqlquery

   get_sqlquery = staticmethod(get_sqlquery)
        



#javi:
class SelectAllSimilarProteins(PianaSelectSQL):
    """
    Generates the sql statement to select all protein-protein similarities for a given database source
    """
    def get_sqlquery(sourceDB):

        sqlquery = """SELECT %s,%s FROM %s WHERE %s & %s """ %(PianaGlobals.proteinPianaA_col,
                                                               PianaGlobals.proteinPianaB_col,
                                                               PianaGlobals.proteinSimilarity_table,
                                                               PianaGlobals.sourceDBID_col,
                                                               sourceDB)

        print sqlquery
        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)


class SelectCompleteTable(PianaSelectSQL):
    """
    """
    def get_sqlquery(table,attributes):

        if attributes:
            attributes_string = ",".join(attributes)
        else:
            attributes_string = "*"

        sqlquery= """SELECT %s FROM %s""" %(attributes_string, table)

        return sqlquery

    get_sqlquery = staticmethod(get_sqlquery)
        

