# File     : create_piana_tables.sql
# Author   : Ramon Aragues & Joan Planas
# Creation : 12.12.2003
# Contents : script that creates the tables used for PIANA
# Usage    : mysql --host=sefarad --database=pianaDB < create_piana_tables.sql
# Comments :
#
# To create tables on your database, do:
#
# mysql --host=sefarad --database=pianaDB < create_piana_tables.sql
#
# For a complete description on how to create a PIANA database, please
# refer to README.populate_piana_db
#
# Before running this script you must drop the previous database
# (if there was already one with the name you want to use)
#
# command line options host and database are needed
# otherwise, the script doesn t know where to create the tables
#


####################################################################
#                                                                  #
#                        PROTEIN DESCRIPTORS                       #
#                                                                  #
####################################################################



# ------------------------------------------------------------------
# create table sequence
# ------------------------------------------------------------------
# This table will hold all sequences entered into this DB as well as 
# the information intrinsically related to the sequence.
# Internal identifier, the sequence itself, Md5 code of that 
# sequence (obtained from an encriptation digest) and the source of 
# the entry
#
# Theoretical mollecular weight (MW) and Isoellectric Point (IP) are
# also stored in this table for each sequence
# ------------------------------------------------------------------

create table sequence(
	
	sequenceID integer unsigned not null,
	sequenceMD5 char(40) not null,
	sequence mediumtext not null,
	sequenceLength mediumint unsigned not null,
	sequenceMW float not null default 0,
	sequenceIP float not null default 0,

	sourceDBID bigint unsigned not null,

	index(sequenceID),

	primary key (sequenceMD5)

     	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table protein
# ------------------------------------------------------------------
# This table will hold all sequences entered into this DB as well as 
# the information intrinsically related to the sequence.
# Internal identifier, the sequence itself, Md5 code of that 
# sequence (obtained from an encriptation digest), the source of 
# the entry and its taxonomy id are stored in this table. 
#
#
# proteinPiana is the unique identifier
#   --> in fact, proteinPiana is just an integer that is unique
#       for each pair (sequence, taxID)
#       Therefore, a protein entity is defined by its 
#       sequence and species.
#
# when taxID is not known, use 0 as 'any species'
#
# Theoretical mollecular weight (MW) and Isoellectric Point (IP) are
# also stored in this table for each sequence
# ------------------------------------------------------------------

create table protein(

	proteinPiana integer unsigned not null,
	sequenceID integer unsigned not null,
	taxID mediumint unsigned not null,

	sourceDBID bigint unsigned not null,
	
	index(proteinPiana,taxID),

	primary key (sequenceID,taxID)
	

	#proteinPiana integer unsigned not null, # up to 2K millions of entries
	#proteinSequence mediumtext not null,
        #taxID mediumint unsigned not null,
	#proteinMD5 char(40) not null,
	#proteinSequenceLength mediumint unsigned not null,
	#proteinMW float not null default 0,
	#proteinIP float not null default 0,
	#sourceDBID bigint unsigned not null,
	#index(proteinMD5,taxID),
	#primary key (proteinPiana)

     	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinSimilarity
# ------------------------------------------------------------------
# This table holds pairs of proteins (ie. sequences) that appear to  
# be the same protein (very similar sequences, same codes, ...)
#
# Not very complete at this point: in the future the idea is to
# do a blast all for proteinPianas and then fill this table
# Right now, only completion (complete_piana.py) algorithms
#  insert information into this table.
# 
# proteinPianaA is always lower than proteinPianaB
#
# ------------------------------------------------------------------

create table proteinSimilarity(

	proteinPianaA integer unsigned not null,
	proteinPianaB integer unsigned not null,

	sourceDBID bigint unsigned not null,

	primary key (proteinPianaA, proteinPianaB)	
	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinPianaCounter
# ------------------------------------------------------------------
# This table holds the value that a new proteinPiana will take
# It is controlled from PianaDBaccess to make sure that it is 
# increased by 1 when a new proteinPiana enters the database
#
# Initially, it will contain the number of entries in 
# proteinCorrespondence + 1 
# ------------------------------------------------------------------

create table proteinPianaCounter(

	proteinPianaCounter integer not null,
	sequenceIDCounter integer not null,
	interactionCounter integer not null,

	primary key(proteinPianaCounter)
	
	) ENGINE = MYISAM;


# Filling the counter table with the maximum proteinPiana found in the proteinCorrespondence table
# uncomment this if using proteinCorrespondence
#insert into proteinPianaCounter select max(proteinPiana)+1 from proteinCorrespondence;

insert into proteinPianaCounter values(0,0,0);







####################################################################
#                                                                  #
#              DATABASES RELATED TABLES	                           #
#                                                                  #
#                                                                  #
####################################################################


# ------------------------------------------------------------------
# create table externalDatabase
# ------------------------------------------------------------------
# This table will hold information about all databases parsed and 
# integred into piana
# ------------------------------------------------------------------
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#  (see column order to follow above)

create table externalDatabase(

	databaseName varchar(255) not null,
	databaseVersion varchar(255) not null,
	parsed_file varchar(255),
	#date date not null,
	date char(10) not null,
	databaseDescription varchar(255) not null,
	databaseInteger bigint unsigned not null,

	primary key (databaseName)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table externalDatabaseType
# ------------------------------------------------------------------
# This table will hold information about which type of information 
# gives to PIANA each external database
# ------------------------------------------------------------------
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#  (see column order to follow above)


create table externalDatabaseType(

	databaseName varchar(255) not null,
	databaseType ENUM ("protein sequences","protein attributes","protein-protein interactions","identifiers cross-references"),

	primary key(databaseName,databaseType)

	) ENGINE = MYISAM;

####################################################################
#                                                                  #
#              EXTERNAL PROTEIN CODES  TABLES                      #
#                                                                  #
# All protein identifiers obtained from external databases.        # 
# The following section will relate these codes with proteinPianas #
#                                                                  #
#                                                                  #
#  If you change the structure of these tables, you ll have to     #
#  change accordingly classes that manage their access:            #
#   - PianaInsertSQL - InsertProteinExternalCode                   #
#   - PianaInsertSQL - InsertProteinSwissAccessionCode             #
#  and modify PianaGlobals with new names                          #
#                                                                  #
#   so ... DO NOT CHANGE THE ORDER OF THE COLUMNS!!!               #
#     --> methods in PianaInsertSQL  won't work if you do it       #
#                                                                  #
#   Column order is db_id, proteinPiana and sourceDBID             #
#                                                                  #
####################################################################


# ------------------------------------------------------------------
# create table valid_protein_ids
# ------------------------------------------------------------------
# This table will hold all valid protein identifiers stored in piana
# 
# used field indicates if the protein_id has been introduced into
# piana database or not
# ------------------------------------------------------------------
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#  (see column order to follow above)

create table valid_protein_ids(

	proteinTypeId varchar(15) not null, 
	proteinTable varchar(20) not null,
	externalIdColumn varchar(20) not null,
	externalIdDescription varchar(255) not null,

	index(proteinTypeId),

	primary key (proteinTypeId)

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table proteinExternalId
# ------------------------------------------------------------------
# This table will hold external protein Identifiers that do not have
# a specific table in the database
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)

create table proteinExternalId(

	proteinExternalId varchar(255) not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,
	isPrimary integer not null default 0,
	externalIdType ENUM ("unknown") not null,

	index(proteinPiana),

	primary key (proteinExternalId, proteinPiana, externalIdType)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table uniprotEntry 
# ------------------------------------------------------------------
# This table will hold uniprot entry names vs proteinPiana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: AOFA_HUMAN
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#  (see column order to follow above)

create table uniprotEntry(

	uniprotEntry varchar(15) not null, 
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,
	

	index(proteinPiana),

	primary key (uniprotEntry, proteinPiana)

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table uniprotAcc
# ------------------------------------------------------------------
# This table will hold uniprot accessions vs proteinPiana
#
# if isPrimary == 1: primary accession number used by uniprot
# if isPrimary == 0: secondary accesion numbers
# ------------------------------------------------------------------
# EXAMPLE OF CODE: P21397
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)
	
create table uniprotAcc(

	uniprotAcc char(6) not null, 
	proteinPiana integer unsigned not null,  
	isPrimary integer not null default 0,
	sourceDBID bigint unsigned not null,

	index(proteinPiana,isPrimary),
	index(uniprotAcc,isPrimary),

	primary key (uniprotAcc, proteinPiana)

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table geneName
# ------------------------------------------------------------------
# This table will hold all gen names used in external DBs having a 
# protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: MAO-A
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)
	
create table geneName(

	geneName varchar(100) not null,
	proteinPiana integer unsigned not null, 
	isPrimary integer not null default 0,
	sourceDBID bigint unsigned not null,

	index(proteinPiana,isPrimary),
	index(geneName,isPrimary),

	primary key (geneName, proteinPiana)


	) ENGINE = MYISAM;



# ------------------------------------------------------------------
# create table pdb
# ------------------------------------------------------------------
# This table will hold pdb information
# referring to proteins in table protein
# ------------------------------------------------------------------
# EXAMPLE OF CODE: 101m (pdb code) T(chain)
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)

create table pdb(
	pdbID varchar(5) not null, 
	proteinPiana integer unsigned not null,  
	chain varchar(4), 
	pdb_chain varchar(10) not null,
	range varchar(13),
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (pdb_chain, proteinPiana)

	) ENGINE = MYISAM;	



# ------------------------------------------------------------------
# create table gi 
# ------------------------------------------------------------------
# This table will hold gi codes used in external DBs and referring 
# any protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: 63879, 63880
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table gi(

	giID integer not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (giID, proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table accessionNumber
# ------------------------------------------------------------------
# This table will hold Accession Number used in external DBs and referring 
# any protein of those stored in piana
# From the collaboration between GenBank, EMBL and DDBJ. Original source
# not taken into account
# ------------------------------------------------------------------
# EXAMPLE OF CODE: BAC39240, CAH69632
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table accessionNumber(

	accessionNumber varchar(15) not null, 
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (accessionNumber, proteinPiana)

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table geneID
# ------------------------------------------------------------------
# This table will hold geneID codes used in external DBs and referring 
# any protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: XXXXXXXXXXXXXXX
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table geneID(

	geneID integer unsigned not null, 
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (geneID, proteinPiana)

	) ENGINE = MYISAM;	

# ------------------------------------------------------------------
# create table unigene
# ------------------------------------------------------------------
# This table will hold unigene codes used in external DBs and referring 
# any protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: Hs.505267
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table unigene(

	unigene  varchar(10) not null, 
	proteinPiana integer unsigned not null,  
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (unigene, proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table ensembl
# ------------------------------------------------------------------
# This table will hold unigene codes used in external DBs and referring 
# any protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: ENSG00000145649
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table ensembl(

	ensembl  varchar(40) not null, 
	proteinPiana integer unsigned not null,  
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (ensembl, proteinPiana)

	) ENGINE = MYISAM;
	
# ------------------------------------------------------------------
# create table refseq
# ------------------------------------------------------------------
# This table will hold refseq codes used in external DBs and referring 
# any protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE OF CODE: XXXXXXXXXXXXXXXXXXXX
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! (see column order to follow above)


create table refseq(

	refseq  varchar(15) not null, 
	proteinPiana integer unsigned not null,  
	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (refseq, proteinPiana)

	) ENGINE = MYISAM;	




####################################################################
#                                                                  #
#                 PROTEIN ATRIBUTES TABLES                         #
#                                                                  #
# These tables contain more in detail descriptions of protein      #
# attributes. For example, according to proteinSpecies one finds   #
# that a given protein belongs to one Species. If more information #
# is needed about this species, the user will look into the table  #
# species created below                                            #
#                                                                  #
####################################################################



# ------------------------------------------------------------------
# create table reactome
# ------------------------------------------------------------------
# This table will hold the correspondence between reactome code
# and the name of the pathway it represents
#
# ------------------------------------------------------------------

create table reactome(

	reactomeID varchar(15) not null,
	pathwayDescription varchar(100) not null,
	sourceDBID bigint unsigned not null,

	primary key (reactomeID) 

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table species
# ------------------------------------------------------------------
# This table will hold all scientific names and their related NCBI
# taxon code for further identification
#
# up to 1.67 milion of possible species 
# THIS FIELD REPRESENTS NCBI TAXON CODE FOR EACH SPECIES
# ------------------------------------------------------------------

create table species(

	taxID mediumint unsigned not null, 
	speciesName varchar(100) not null,
	speciesDescription varchar(255),
	speciesKingdom varchar(40) not null,
	sourceDBID bigint unsigned not null,

	index(speciesName),

	primary key (taxID, speciesName) 

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table ec
# ------------------------------------------------------------------
# This table will hold all ec codes and their description
#
# ------------------------------------------------------------------

create table ec(

	ecID varchar(10) not null, 
	ecDescription varchar(255),
	sourceDBID bigint unsigned not null,
	
	primary key (ecID) 

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table go
# ------------------------------------------------------------------
# This table will hold all go terms and their description
#
#
# ------------------------------------------------------------------

create table go(

	goID integer unsigned not null,
	
	name varchar(255),
	acc  varchar(12),
	#term_type varchar(50),
	term_type ENUM ("universal","molecular_function","cellular_component","biological_process"),
	distance2root integer unsigned not null,
	sourceDBID bigint unsigned not null,

	primary key (goID) 

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table go_term2term_distance
# ------------------------------------------------------------------
# This table will contains the distance between two go terms
# distance is the number of steps that we have to follow going from
# go_term1 to go_term2
# ------------------------------------------------------------------

create table go_term2term_distance(
	go_term1 integer unsigned not null,
	go_term2 integer unsigned not null,
	distance integer unsigned not null,
	index(go_term1),
	index(go_term2),
	primary key (go_term1, go_term2)
	)ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table cog
# ------------------------------------------------------------------
# This table will hold all cog codes and their description
#
# cogFunction is a letter: if you want to know to which function it
# corresponds to, you have to check it by hand in file fun.txt
#
# ------------------------------------------------------------------

create table cog(

	cogID varchar(10) not null, 
	cogDescription varchar(255),
	cogFunction varchar(10),
	sourceDBID bigint unsigned not null,

	primary key (cogID) 

	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table mim
# ------------------------------------------------------------------
# This table will hold all MIM codes and their description
#
# ------------------------------------------------------------------
	
create table mim(

	mimID integer unsigned not null,
	mimTitle varchar(255),
	mimKeywords varchar(255),

	sourceDBID bigint unsigned not null,

	primary key (mimID)


	) ENGINE = MYISAM;


####################################################################
#                                                                  #
#              PROTEIN-ATTRIBUTES RELATIONSHIP TABLES              #
#                                                                  #
#   The following tables are devoted to stablishing relationships  #
# between different protein attributes from external databases     #
# and a given unitary entity in this database, a protein identified# 
# by the internal piana identifier                                 #
#    The ID suffix is reserved for these codes that are unique     #
#                   identifiers of their own DB                    #
#                                                                  #
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!                      #
#     --> method PianaInsertSQL.InsertProteinAttributeRelationship #
#                                                                  #
#         does not work if these tables do not follow the order:   #
#                                                                  #
#                 - attribute code                                 #
#                 - proteinPiana                                   #
#                 - sourceDBID                                     #
#                                                                  #
####################################################################



# ------------------------------------------------------------------
# create table proteinReactome
# ------------------------------------------------------------------
# This table will hold Reactome Identifiers  used in pianaDB 
# and its relationship whith pianaDB proteins
# ------------------------------------------------------------------
# EXAMPLE OF CODE: REACT_1675.1
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#    (see column order to follow above)

create table proteinReactome(

	reactomeID varchar(15) not null, 
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana),
	
	primary key(reactomeID, proteinPiana, sourceDBID)
	
	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinGo
# ------------------------------------------------------------------
# This table will hold GO terms used in pianaDB and its relationship
# whith pianaDB proteins
# ------------------------------------------------------------------
# EXAMPLE OF CODE: 
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#    (see column order to follow above)

create table proteinGo(

	goID integer unsigned not null, 
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana,goID),
	
	primary key(goID, proteinPiana, sourceDBID)
	
	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinCog
# ------------------------------------------------------------------
# This table will hold COG codes used in pianaDB and its 
# relationship whith pianaDB proteins
# ------------------------------------------------------------------
# EXAMPLE OF CODE: COG0001
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#     (see column order to follow above)

create table proteinCog(

	cogID varchar(10) not null, 
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana),
	
	primary key(cogID, proteinPiana, sourceDBID)
	
	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinDescription
# ------------------------------------------------------------------
# This table will hold all protein descriptions used in 
# external DBs having  a protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE: Monoamine oxidase
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!   
#     (see column order to follow above)
	
create table proteinDescription(

	proteinDescription text(400) not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana)


	) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table proteinMIM
# ------------------------------------------------------------------
# This table will hold MIM codes associated with proteins
# ------------------------------------------------------------------
# EXAMPLE: 601647
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!   
#     (see column order to follow above)
	
create table proteinMIM(

	mimID integer unsigned not null,
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key(mimID,proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinKeyword
# ------------------------------------------------------------------
# This table will hold all protein keywords used in 
# external DBs having  a protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE: 
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!   
#      (see column order to follow above)
	
create table proteinKeyword(

	proteinKeyword varchar(50) not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinDisease
# ------------------------------------------------------------------
# This table will hold all protein related diseases used in 
# external DBs having  a protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE: Defects in PPP2R1B might be a cause of some lung and
#           colorectal cancers.
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!   
#     (see column order to follow above)
	
create table proteinDisease(

	proteinDisease text (2000) not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,

	index(proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinFunction
# ------------------------------------------------------------------
# This table will hold all protein functions described in 
# external DBs having a protein of those stored in piana
#
# proteinFunction removed from primary key and index because it 
# has to be of type text, and text types cannot go into primary key
# ------------------------------------------------------------------
# EXAMPLE: 
# ------------------------------------------------------------------
# removed the primary key because of type... 
# ...text cannot be primary key
#
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!   
#     (see column order to follow above)                    

create table proteinFunction(

	proteinFunction text not null,
	proteinPiana integer unsigned not null,
	sourceDBID bigint unsigned not null,


	index(proteinPiana)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinSubcellularLocation
# ------------------------------------------------------------------
# This table will hold all protein functions described in 
# external DBs having a protein of those stored in piana
# ------------------------------------------------------------------
# EXAMPLE: 
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!     
#    (see column order to follow above) 
	
create table proteinSubcellularLocation(

	proteinSubcellularLocation text(400) not null,
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana)


	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinEC	
# ------------------------------------------------------------------
# This table will hold known EC codes for proteins proteinPiana
#
# One protein can have multiple ecID
# One EC code can have multiple proteinPiana
# 
# 
# ------------------------------------------------------------------
# EXAMPLE OF CODE: EC 1.4.3.4
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!    
#     (see column order to follow above)  
	
create table proteinEC(

	ecID varchar(15) not null,
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,
	
	index(proteinPiana,ecID),
	index(ecID,proteinPiana)


	) ENGINE = MYISAM;




# ------------------------------------------------------------------
# create table proteinProsite
# ------------------------------------------------------------------
# This table will hold Prosite ids for proteins
# ------------------------------------------------------------------
# This table is the correspondence between proteinPiana and prosite ids
#
# ------------------------------------------------------------------
#   DO NOT CHANGE THE ORDER OF THE COLUMNS!!!    
#     (see column order to follow above)  

create table proteinProsite(

	prositeID  varchar(20) not null, 
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(prositeID),

	primary key (proteinPiana, prositeID, sourceDBID)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table interPro
# ------------------------------------------------------------------
# This table will hold interPro information
# referring to proteins in table protein
# ------------------------------------------------------------------
# EXAMPLE OF CODE: IPR001254
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#     (see column order to follow above)

create table interPro(
	interProID varchar(12) not null, 
	proteinPiana integer unsigned not null,  
	interProDescription varchar(40) not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (interProID, proteinPiana, sourceDBID)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table pfam
# ------------------------------------------------------------------
# This table will hold pfam information
# referring to proteins in table protein
# ------------------------------------------------------------------
# EXAMPLE OF CODE: IPR001254
# ------------------------------------------------------------------
#  DO NOT CHANGE THE ORDER OF THE COLUMNS!!! 
#     (see column order to follow above)

create table pfam(
	pfamID varchar(12) not null, 
	proteinPiana integer unsigned not null,  
	pfamDescription varchar(40) not null,

	sourceDBID bigint unsigned not null,

	index(proteinPiana),

	primary key (pfamID, proteinPiana, sourceDBID)

	) ENGINE = MYISAM;	


# ------------------------------------------------------------------
# create table proteinDBAliCluster
# ------------------------------------------------------------------
# This table will hold DBAli clusters for binding sites 
# of pibase
# ------------------------------------------------------------------
# This table is a bidirectional link table between protein 
# and dbali cluster 
#
# Attention! cluster ids are not relevant by themselves... 
# the only important thing is to know if two proteins share 
# the cluster or not 
#
# clusteringMethod can be used for determining the method used,
# but clusterIDs from different clusteringMethod values cannot 
# be compared
# 
#
# ------------------------------------------------------------------ 

create table proteinDBAliCluster(

	DBAliclusterID  varchar(20) not null, 
	proteinPiana integer unsigned not null,
	patchResidues text,
	DBAliclusteringMethod varchar(20) not null,
	sourceDBID varchar(20) not null, 

	index(DBAliclusterID),
	index(DBAliclusteringMethod),

	primary key (proteinPiana, DBAliclusterID, DBAliclusteringMethod)

	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinScop
# ------------------------------------------------------------------
# This table will hold SCOP codes used in pianaDB and its 
# relationship whith pianaDB proteins
# ------------------------------------------------------------------
# EXAMPLE OF CODE:
# ------------------------------------------------------------------
# This table is different from other protein attributes, 
# since there are three relevant codes: cf, sf, fa
# Instead of creating three tables, we will create methods 
# that handle this particularity

create table proteinScop(

	cf integer not null,
	sf integer not null, 
	fa integer not null,  
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(sf),
	index(fa),
	index(proteinPiana),
	
	primary key(cf, sf, fa, proteinPiana, sourceDBID)

	
	) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table proteinCath
# ------------------------------------------------------------------
# This table will hold CATH codes used in pianaDB and its 
# relationship whith pianaDB proteins
#
# res_start and res_end define the residues that delimit this 
# segment a cath domain can have several segments
#
# Attention! Even if the source database does not give information 
# about segmentID, insert something (eg. 1) since the primary key 
#  needs it
# ------------------------------------------------------------------
# EXAMPLE OF CODE:
# ------------------------------------------------------------------
# This table is different from other protein attributes, since  
# there are four relevant codes:  C, A, T and H
# Instead of creating three tables, we will create methods 
# that handle this particularity

create table proteinCath(

	c integer ,
	a integer , 
	t integer , 
	h integer , 
	cathID varchar(20) not null,
	segmentID integer unsigned not null,
	res_start integer unsigned not null,
        res_end   integer unsigned not null,   
	proteinPiana integer unsigned not null,

	sourceDBID bigint unsigned not null,

	index(c),
	index(a),
	index(t),
	index(h),
	#index(cathID),
	index(proteinPiana),
	
	primary key(cathID, segmentID, proteinPiana, sourceDBID, res_start, res_end)

	
	) ENGINE = MYISAM;


####################################################################
#                                                                  #
#              EXTERNAL PROTEIN DATABASES TABLES                   #
#                                                                  #
#   The following tables are devoted to storing information found  #
#   on external protein databases.                                 #
#                                                                  #
#   Iḿ not sure why am I creating a new table for it instead of    #
#   inserting information in  external codes tables... maybe a     #
#   matter of separating information... maybe something else...    #
#                                                                  #
####################################################################


####################################################################
#                                                                  #
#                        INTERACTION TABLES                        #
#                                                                  #
#  Tables containing information about the interactions            #
####################################################################




# ------------------------------------------------------------------
# create table interaction
# ------------------------------------------------------------------
# This table will hold all the interactions for which we have 
# information
#
# Contains the two proteins involved in the 
# interaction, as well as interactionConfidence.
#
# The order proteinPianaA < proteinPianaB must be respected
#
# interactionConfidence is a value that gives a measure of the 
# trustworthiness of this interaction. It is dynamically calculated 
# from field interactionSourceDB.confidenceAssigned
# and parameters of PianaGlobals: confidenceDB and confidenceMethod
#  (currently, interactionConfidence is always 1: it will be 
#   calculated as decribed above in the future)
#
# isSourceA and isSourceB are used to indicate the direction of the 
# interaction
# if isSourceA ==1 and isSourceB ==1 then it is a bidirectional 
# interaction
# if isSourceA ==0 or isSourceB ==0 then it is just monodirectional
#  (currently, both proteins are always considered as being Source)
# ------------------------------------------------------------------

create table interaction (
     

     interactionPiana integer not null,
     proteinPianaA integer unsigned not null,
     isSourceA integer not null default 1,
     proteinPianaB integer unsigned not null,
     isSourceB integer not null default 1,
     interactionConfidence float not null,
     sourceDBID bigint unsigned not null,

     index(proteinPianaA, proteinPianaB ),
     index(proteinPianaB, proteinPianaA ),

     primary key (interactionPiana)

     ) ENGINE = MYISAM;




# ------------------------------------------------------------------
# create table interactionMethod
# ------------------------------------------------------------------
# This table will hold (interactionPiana, methodName)
# For each interactionPiana, it will state the methodName that 
# was used to find it. It can have several methods for one 
# interactionPiana (i.e. an interaction that was found through 
# several different methods)
# 
# methodID values are controlled by method PianaDBaccess.get_methodID()
# and variable PianaGlobals.method_names 
# methodID should never be inserted without a previous call to 
# get_methodID() Otherwise you risk of introducing a methodID that 
# will not be recognized by the system
#
# The confidence given to methods is a parameter of PianaGlobals
# ------------------------------------------------------------------

create table interactionMethod(

     methodID varchar(20) not null,
     interactionPiana integer not null,
     sourceDBID bigint unsigned not null,

     index(methodID, interactionPiana),

     primary key (interactionPiana, methodID)

     ) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table interactionFeatures
# ------------------------------------------------------------------
# This table will hold (interactionPiana, feature1, feature2, ..)
# For each interactionPiana, it will have an entry with different 
# features we might be interested in
# 
# ------------------------------------------------------------------
# Note: in case you add more features to this table, you ll have
#       to pay attention to the primary key, since right now
#       PianaDBaccess.insert_interaction relies on -insert ignore-
#       in order not to insert duplicated rows... if there are
#       more than one feature, this cannot be handled this way

create table interactionFeatures(

     pubmedID varchar(20) not null,
     interactionPiana integer not null,
	
     sourceDBID bigint unsigned not null,

     index(pubmedID, interactionPiana),


     primary key (interactionPiana, pubmedID)

     ) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table interactionScores
# ------------------------------------------------------------------
# This table will hold confidence scores given by source databases
# (e.g. STRING gives scores for several potential interaction predictors)
# ------------------------------------------------------------------

create table interactionScores(

     sourceDBID bigint unsigned not null,
     interactionPiana integer not null,
     equiv_nscore  integer,
     equiv_nscore_transferred  integer,
     equiv_fscore integer,
     equiv_pscore   integer,
     equiv_hscore  integer,
     array_score  integer,
     array_score_transferred  integer,
     experimental_score  integer,
     experimental_score_transferred  integer,
     database_score  integer,
     database_score_transferred  integer,
     textmining_score  integer,
     textmining_score_transferred  integer,
     combined_score integer,

     index(sourceDBID),
     index(interactionPiana),

     primary key (interactionPiana, sourceDBID)

     ) ENGINE = MYISAM;

# ------------------------------------------------------------------
# create table interactionProteinSource
# ------------------------------------------------------------------
# This table will hold the protein source of expansion 
# interaction predictions
# 
# ------------------------------------------------------------------

create table interactionProteinSource(

     sourceDBID bigint unsigned not null,

     interactionPiana integer not null,
     proteinPiana integer unsigned not null,


    index(interactionPiana),
    index(proteinPiana),
    index(sourceDBID),

    primary key (interactionPiana, proteinPiana, sourceDBID)

    ) ENGINE = MYISAM;
	



# ------------------------------------------------------------------
# create table temporalData
# ------------------------------------------------------------------
# This table will hold temporal data
# This table should be always empty, only must contain data when a 
# parser is running. When database connection is closed, this data
# should be deleted
# 
# ------------------------------------------------------------------

create table temporalData(

	#id integer unsigned not null auto_increment,
	#identifier mediumint unsigned not null,
	identifier varchar(100) not null,
	attr1 varchar(100) not null,
	attr2 varchar(100),
	attr3 varchar(100),
	attr4 varchar(100),
	attr5 varchar(100),
	sourceDBID integer unsigned not null,
	updated tinyint default 0

	#index (updated,identifier),

	#primary key (identifier,attr1,attr2,attr3)
	#primary key (id)

) ENGINE = MYISAM;




# ------------------------------------------------------------------
# create table blastResults
# ------------------------------------------------------------------
# This table will hold results obtained with blast
# ------------------------------------------------------------------

create table blastResults(

	#blast_id integer unsigned not null auto_increment,

	sequenceID_A integer unsigned not null,
	sequenceID_B integer unsigned not null,
	score smallint(2) unsigned not null,
	bit_score float(4) unsigned not null,
	start_A smallint(2) unsigned not null,
	end_A smallint(2) unsigned not null,
	start_B smallint(2) unsigned not null,
	end_B smallint(2) unsigned not null,
	identities tinyint(1) unsigned not null,
	similarity tinyint(1) unsigned not null,
	gaps tinyint(1) unsigned not null,
	program ENUM("bl2seq","blastall"),
	filter ENUM("T","F")

	#primary key (blast_id)

) ENGINE = MYISAM;


# ------------------------------------------------------------------
# create table blastResults
# ------------------------------------------------------------------
# This table will hold the clusterID for which a proteinPiana belongs to
# ------------------------------------------------------------------

create table proteinSimilarityClusters(

	proteinPiana integer unsigned not null,
	clusterID integer unsigned not null,

	primary key(proteinPiana)


) ENGINE = MYISAM;


# Insert default valid_protein_ids

INSERT INTO valid_protein_ids (proteinTypeId, proteinTable, externalIdColumn, externalIdDescription) VALUES
	("proteinPiana", "protein", "proteinPiana", "internal PIANA identifier (eg. 12812)"),
	("sequence", "sequence", "sequence",  "the sequence of the protein"),
	("fasta","sequence","sequence","the sequence of the protein preceded by a title line"),
	("md5","sequence","sequenceMD5","the md5 checksum of the protein sequence followed by 8 aminoacids (first 4 and last 4 of the sequence)"),
	("unientry","uniprotEntry","uniprotEntry","the uniprot entry of the protein\n(eg. BCLX_HUMAN)"),
	("uniacc","uniprotAcc","uniprotAcc","the uniprot accession number of the protein (eg. Q07817)"),
	("geneName","geneName","geneName","the gene name for the protein (eg. bclx)"),
	("pdb.chain","pdb","pdb_chain","the pdb code for the protein (eg. 1r2d.A)"),
	("gi","gi","giID","the NCBI gi code for the protein (eg. 1622940)"),
	("geneID","geneID","geneID","the NCBI geneID code for the protein (eg. xxxxxxx)"),
	("unigene","unigene","unigene","the NCBI unigene code for the protein (eg. xxxxxxx)"),
	("refseq","refseq","refseq","the NCBI  RefSeq accession for the protein (eg. xxxxxxx)"),
	("ensembl","ensembl","ensembl","the ensembl code for the protein (eg. xxxxxxx)"),
	("ncbiacc","accessionNumber","accessionNumber","the NCBI accession Number code for the protein"),
	("sequenceID", "sequence", "sequenceID", "sequenceID value");




#
# END
#
