"""
File        : biogrid2piana.py
Author      : Ramon Aragues
Creation    : 11.03.07
Contents    : inserts information from a Biogrid tab file database into PIANA
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from a biogrid database

--> Biogrid database must be in txt tabulated format

--> download it from http://www.thebiogrid.org/downloads.php (in tabulated format)

"""

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline
import MySQLdb


from PianaDBaccess import *
from parserUtilities import *

import time


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("input-file=",None,"file name of input file containing biogrid txt data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="biogrid2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from a biogrid database",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


arguments_dic = parser.parseArguments()

input_file= arguments_dic["input-file"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
log_file = arguments_dic["log-file"]
time_control = arguments_dic["time-control"]
database_information = arguments_dic["database-information"]





# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.proteinPianaCounter_table]


# --------
# --------
#  Main()               
# --------                               
# --------

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )


# Lock_number is used to calculate the frequency to do unlock...
line_number=0


if verbose:
    sys.stderr.write("Reading data from biogrid file...\n")

start_parsing = 0

for line in input_file_fd:
    


    if start_parsing == 0:
        if line.startswith("INTERACTOR_A"):
            start_parsing = 1
        continue
        
    
    parser.add_to_log("Number of lines")
    line_number += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if line_number%20000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))


    line_fields = line.split('\t')  # [0] INTERACTOR_A                                             
                                    # [1] INTERACTOR_B
                                    # [2] OFFICIAL_SYMBOL_A
                                    # [3] OFFICIAL_SYMBOL_B
                                    # [4] ALIASES_FOR_A
                                    # [5] ALIASES_FOR_B
                                    # [6] EXPERIMENTAL_SYSTEM
                                    # [7] SOURCE (last name of author)
                                    # [8] PUBMED_ID
                                    # [9] ORGANISM_A_ID
                                    # [10] ORGANISM_B_ID


    # First step: obtain list of proteinPianas for interactors
    geneName_intA = line_fields[2]
    geneName_intB = line_fields[3]

    list_detection_methods = [line_fields[6]]
    list_pubmed_ids = [line_fields[8]]


    taxID_intA = int(line_fields[9])
    taxID_intB = int(line_fields[10])

    
    list_proteinPianas_intA = piana_access.get_list_proteinPiana(protein_id=geneName_intA,
                                                                 id_type="geneName",
                                                                 tax_id_value=taxID_intA, source_db_info="no")
    
    list_proteinPianas_intB = piana_access.get_list_proteinPiana(protein_id=geneName_intB,
                                                                 id_type="geneName",
                                                                 tax_id_value=taxID_intB, source_db_info="no")

    if len(list_proteinPianas_intA)==0 or len(list_proteinPianas_intB)==0:
        parser.add_to_log("Number of interactions not inserted because geneName not found")


    if not list_detection_methods:
        parser.add_to_log("Number of interactions without detection method")
        #sys.stderr.write("%s\n" %line)
        #sys.stderr.write("How is it possible not to have detection method?\n")

    if not list_pubmed_ids:
        parser.add_to_log("Number of interactions without pubmed id")
        
        #sys.stderr.write("%s\n" %line)
        #sys.stderr.write("How is it possible not to have pubmed reference?\n")
                
    
    # Insert into database the interactions:
    for proteinPianaA in list_proteinPianas_intA:
        for proteinPianaB in list_proteinPianas_intB:
            # We don't have to worry is proteinPianaA is smaller than proteinPianaB: piana_access will do it for us
            piana_access.insert_interaction(proteinPianaA_value = proteinPianaA,
                                            isSourceA_value = 1,
                                            proteinPianaB_value = proteinPianaB,
                                            isSourceB_value = 1,
                                            interactionConfidence_value = 1,
                                            methodDescription_value = list_detection_methods,
                                            sourceDBDescription_value = sourceDB,
                                            confidenceAssignedSourceDB_value=1,
                                            pubmed_id_value= list_pubmed_ids)



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to make sure that all inserts are correctly done, as well as unlock tables
   

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

if verbose:
    sys.stderr.write("%s\t%s\n" %(line_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
    sys.stderr.write(parser.get_log_string())
