"""
File        : xog2piana.py
Author      : Ramon Aragues
Creation    : 18.10.2004
Contents    : fills up tables in database piana with information from files whog or kog
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from files whog or kog

files whog and kog can be downloaded from ftp://ftp.ncbi.nih.gov/pub/COG

These files contain COG and KOG identifiers for proteins and protein domains. 

The format they follow is:

[H] COG0001 Glutamate-1-semialdehyde aminotransferase
  Afu:  AF1241
  Hbs:  VNG2326G
  Mac:  MA0581
  Mth:  MTH228
  Mja:  MJ0603 MJ0604 MJ0605
  Mka:  MK0280
  Tac:  Ta0571
  Tvo:  TVN0635 TVN0634
 
_______
 
[E] COG0002 Acetylglutamate semialdehyde dehydrogenase
  Afu:  AF2071
  Mac:  MA3566
  Mth:  MTH846
  Mja:  MJ1096
  Mka:  MK1077
  Pho:  PH1720
 

Attention!!! In some cases, the protein_name is a gi_code... make sure to check for gi_codes as well if nothing is found in table geneName

Before running, myva=gb and kyva=gb have to be parsed, to fill pianaDB with correspondences between gi codes and gene names


"""

# xog2piana.py: fills up tables in database piana with information from files whog or kog
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import MySQLdb

from PianaDBaccess import *
from parserUtilities import *

import time



# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file (whog or kog)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]
                           
# Define the parser information to print in the help:

parser = ParserObject( script_name="xog2piana.py",
                       script_description="This program fills up tables in database piana with information from files  myva=gb and kyva=gb",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)


arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]
database_information= arguments_dict["database-information"]



# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')




# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess


# this dictionary is used to transform from cog abbreviations to taxonomy id
dic_species_eq = { 
    "Afu":   2234,
    "Hbs":  64091,
    "Mac": 188937,
    "Mth": 145262,
    "Mja":   2190,
    "Mka": 190192,
    "Tac":   2303,
    "Tvo":  50339,
    "Pho":  53953,
    "Pab":  29292,
    "Pya":  13773,
    "Sso":   2287,
    "Ape":  56636,
    "Sce":   4932,
    "Spo":   4896,
    "Ecu":   6035,
    "Aae":  63363,
    "Tma":   2336,
    "Syn":   1148,
    "Nos": 103690,
    "Fnu": 190304,
    "Dra":   1299,
    "Cgl":   1718,
    "Mtu":  83332,
    "MtC":  83331,
    "Mle":   1769,
    "Cac":   1488,
    "Lla":   1360,
    "Spy": 160490,
    "Spn": 170187,
    "Sau": 158879,
    "Lin":   1642,
    "Bsu":   1423,
    "Bha":  86665,
    "Eco":  83333,
    "EcZ": 155864,
    "Ecs":  83334,
    "Ype":    632,
    "Sty":  99287,
    "Buc": 107806,
    "Vch":    666,
    "Pae":    287,
    "Hin":  71421,
    "Pmu":    747,
    "Xfa": 160492,
    "Nme": 122586,
    "NmA": 122587,
    "Rso":    305,
    "Hpy":  85962,
    "jHp":  85963,
    "Cje":    197,
    "Atu": 181661,
    "Sme":    382,
    "Bme":  29459,
    "Mlo":    381,
    "Ccr": 155892,
    "Rpr":    782,
    "Rco":    781,
    "Ctr":    813,
    "Cpn": 115713,
    "Tpa":    160,
    "Bbu":    139,
    "Uur":   2130,
    "Mpu":   2107,
    "Mpn":   2104,
    "Mge":   2097,
    "ath":  3702,
    "cel":  6239,
    "dme":  7227,
    "hsa":  9606,
    "sce":  4932,
    "spo":  4896,
    "ecu":  6035 }

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.gi_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.proteinCog_table,
                  PianaGlobals.cog_table,
		  PianaGlobals.temporalData_table,
		  PianaGlobals.protein_table]	
                  #PianaGlobals.proteinSpecies_table,]


if verbose:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)

# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

correspondences_not_inserted = 0

input_file_fd =  file(input_file, 'r')


line_number = 0


if verbose:
    sys.stderr.write( "Processing file\n")

cogs_not_inserted = 0
cog_code = None

# a xog file looks like this:
#
## [H] COG0001 Glutamate-1-semialdehyde aminotransferase
##   Afu:  AF1241
##   Hbs:  VNG2326G
##   Mac:  MA0581
##   Mth:  MTH228
##   Mja:  MJ0603 MJ0604 MJ0605
##   Mka:  MK0280
##   Tac:  Ta0571
##   Tvo:  TVN0635 TVN0634
  
# while there are lines in the file, get correspondence protein_name gi_code
for line in input_file_fd:

    line_number += 1


    if time_control:
        if not line_number%20000:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if line[0] == "[":
        # it is a new COG or KOG entry

        
        line_atoms = line[1:].split("]")  # line_atoms[0] will contains the function letters (there can be more than one)
                                          # line_atoms[1:] contains the cog code and its description

        cog_letter_function = line_atoms[0]

        rest_of_line = ""
        for atom in line_atoms[1:]:
            rest_of_line += atom

        line_fields = rest_of_line.split()  # line_fields[0] contains the cog code
                                            # line_fields[1:] contains the cog description

        cog_code = line_fields[0].strip()

        cog_description = ""
        for atom_description in line_fields[1:]:
            cog_description += atom_description

        if verbose:
            sys.stderr.write("Inserting new cog with code %s and function %s and description %s\n" %(cog_code, cog_letter_function, cog_description))

        # insert information into table cog
        piana_access.insert_cog(cog_id=cog_code , cog_description=cog_description , cog_function=cog_letter_function , source_db= sourceDB)

    # END OF if line[0] == "[":
    
    else:

        if line[0:7] == "_______":
            # new entry sign
            cog_code = None
        else:


            line_fields = line.split(":")  # line_fields[0] contains the species abbreviation
                                           # line_fields[1] contains space-separated protein names

            # check that it is a line with info (there are also empty lines)
            if len(line_fields) > 1:

                if cog_code is None:
                    raise ValueError("Cog code cannot be None... error parsing file")

                list_protein_names = line_fields[1].strip().split()

                species_abb = line_fields[0].strip()

                if dic_species_eq.has_key(species_abb):
                    tax_id_value= dic_species_eq[species_abb]
                else:
                    raise ValueError("Species abbreviation %s unknown: add the abbreviation and tax id to dictionary dic_species_eq\n" %(species_abb) )


                for protein_name in list_protein_names:

                    list_proteinPiana= piana_access.get_list_proteinPiana(protein_id= protein_name.strip(),
                                                                           id_type= "geneName",
                                                                           tax_id_value= tax_id_value,
                                                                           source_db_info="no")

                    if not list_proteinPiana:
                        pass
                        # if nothing was found under geneName, try with gi_code (cog uses both codes indistinctively)
                        list_proteinPiana= piana_access.get_list_proteinPiana(protein_id= protein_name.strip(),
                                                                               id_type= "gi",
                                                                               tax_id_value= tax_id_value,
                                                                               source_db_info="no")
                    # END OF if not list_proteinPiana:
                    
                    if list_proteinPiana:
                        for proteinPiana in list_proteinPiana:

                            if verbose:
                                sys.stderr.write("Inserting cog %s for proteinPiana %s\n" %(cog_code, proteinPiana))

                            piana_access.insert_protein_cog(cog_id= cog_code, proteinPiana_value= proteinPiana, proteinCogSource_value= sourceDB)
                            parser.add_to_log("Protein-COGs inserted")

                        # END OF for proteinPiana in list_proteinPiana:
                    # END OF if list_proteinPiana:
                        
                    else:
                        if verbose:
                            sys.stderr.write("cog not inserted because no proteinPiana was found for protein_name %s\n" %protein_name)
                        parser.add_to_log("COGs not inserted because no proteinPiana was found")
                    # END OF (if list_proteinPiana:)
                        
                # END OF for protein_name in list_protein_names:

            # END OF if len(line_fields) > 1:
            
        # END OF else(if line[0:7] == "_______":)
        
    # END OF else: (if line[0] == "[":)
    
# END OF for line in input_file_fd:

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if verbose:
    sys.stderr.write(parser.get_log_string())

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
