"""
File        : xyva_gb2piana.py
Author      : Ramon Aragues
Creation    : 18.10.2004
Contents    : fills up tables in database piana with information from files myva=gb or kyva=gb
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from files myva=gb or kyva=gb

files myva=gb and kyva=gb can be downloaded from ftp://ftp.ncbi.nih.gov/pub/COG

These files contain correspondences between protein names and gi codes. We need to parse these files because
the COG and KOG listings (whog and kog) use protein names instead of gi codes

The format they follow is pretty simple:

protein_name<spaces>gi_code

Attention!!! In some cases, the protein_name is also a gi_code... avoid inserting that proteinPiana (in parser xog2piana.py we make sure to
check for gi_codes as well if nothing is found in table geneName

Before running, genpept and nr have to be parsed, to fill pianaDB with correspondences between gi codes and proteinPianas

"""

# xyva_gb2piana.py: fills up tables in database piana with information from files myva=gb or kyva=gb
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import MySQLdb

from PianaDBaccess import *
from parserUtilities import *

import time


# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing xyva_gb data (kyva=gb or myva=gb)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]
                           
# Define the parser information to print in the help:




parser = ParserObject( script_name="xyva_gb2piana.py",
                       script_description="This program fills up tables in database piana with information from xyva_gb",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)


arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]
database_information= arguments_dict["database-information"]


# Start time counter
if time_control:
    initial_time = time.time()


# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')



# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.gi_table,
                  PianaGlobals.geneName_table,
		  PianaGlobals.temporalData_table]


if verbose:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)

# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

input_file_fd =  file(input_file, 'r')

line_number = 0


if verbose:
    sys.stderr.write( "Processing file\n")

# while there are lines in the file, get correspondence protein_name gi_code
for line in input_file_fd:

    line_number += 1
    parser.add_to_log("Number of lines processed")


    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if not line_number%20000:
            sys.stderr.write("%s processed lines in %s seconds\n" %(line_number,time.time()-initial_time))



    line_fields = line.split()  # line_fields[0] is the protein_name
                                # line_fields[1] is the gi_code
    

    if len(line_fields) > 1:
        
        # if the protein_name (line_fields[0]) being used is a gi_code (it happens!!!) skip the entry: we do not want to insert gi_codes 
        # in table geneNames 
        proteinPiana_exists = piana_access.get_list_proteinPiana(protein_id=line_fields[0],
                                                                  id_type= "gi",
                                                                  tax_id_value= 0,
                                                                  source_db_info="no")

        if not proteinPiana_exists:

            # get the proteinPianas associated with this gi code (line_fields[1])
            list_proteinPiana = piana_access.get_list_proteinPiana(protein_id=line_fields[1],
                                                                    id_type= "gi",
                                                                    tax_id_value= 0,
                                                                    source_db_info="no")

            if not list_proteinPiana:
                parser.add_to_log("Correspondences not inserted because no gi code found")
                if verbose:
                    sys.stderr.write("No protein found for this gi_code %s\n" %(line_fields[1]))
            # END OF if list_inserted_proteinPiana:

            else:
                for proteinPiana in list_proteinPiana:

                    if verbose:
                        sys.stderr.write("correpondence between proteinPiana %s and geneName %s\n" %(proteinPiana, line_fields[0].strip()))

                    piana_access.insert_geneName_code(geneName_code_value= line_fields[0].strip() ,
                                                      proteinPiana_value= proteinPiana,
                                                      geneName_source_value= sourceDB,
                                                      isPrimary_value=0)
                # END OF for proteinPiana in list_proteinPiana:
            # END OF else: (if not list_inserted_proteinPiana:)
        # END OF if not proteinPiana_exists:
        else:
            if verbose:
                sys.stderr.write("Entry skipt because protein name was a gi code\n")
            parser.add_to_log("Entries skipt because protein name was a gi code")
    # END OF if line_fields > 1:
# END OF for line in input_file_fd:


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if verbose:
    sys.stderr.write(parser.get_log_string())

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

if time_control:
    sys.stderr.write("%s\t%s\n" %(line_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
