"""
File        : scop_ints2piana.py
Author      : Ramon Aragues
Creation    : 15.02.2006
Contents    : script that fills up tables in database piana related to external DB "scop_ints"
Called from : 

=======================================================================================================

This program parsers a text file (following scop_ints format) with interactions and inserts them into piana
(it can also be used to create a file with scop ints)

"""

# scop_ints2piana.py: script that fills up tables in database piana related to external DB "scop_ints"
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys
import getopt

import re
import readline

from sets import *

import MySQLdb

import utilities

from PianaDBaccess import *

verbose = 1

# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------


# ----------------------
# Function usage()
# ----------------------
def usage():
    print "--------------------------------------------------------------------------------------------------------------"
    print "This program fills up tables in database piana related to external DB 'scop_ints'\n"
    print "  -> by looking to an input file with SCOP-SCOP interactions from the PDB, this program can be used to:\n"
    print "           1. predict interactions (and insert them into a piana DB) by assumming that if SCOPa and SCOPb "
    print "              interact in the PDB, then all proteins with SCOPa will interact with proteins with SCOPb"
    print "           2. infer protein-protein interactions from the input file with PDB info and insert them into"
    print "              the piana DB"
    print "           3. write to an output file (default name scop_scop_interactions.fa.txt) the SCOP fa-SCOP fa interactions"
    print "              when the interaction involves different proteins\n"
    print "Usage: python scop_ints2piana.py --piana-dbname=piana_dbname --piana-dbhost=piana_dbhost"
    print "                                 --piana-dbuser=piana_dbuser --piana-dbpass=piana_dbpass "
    print "                                 --scop_ints-file=scop_ints_file --method-name=method_name"
    print "                                 --only-prot-ints --no-species --min-num-contacts=min_num_contacts [--help] [--verbose]"
    print "\nwhere:"
    print "     piana_dbname   : name of database piana to be used (required)"
    print "     piana_dbhost   : name of host where database piana to be used is placed (required)"
    print "     piana_dbuser   : username accessing the database (not required in most systems)"
    print "     piana_dbpass   : password of username accessing the database (not required in most systems)"
    print "     scop_ints_file : the name of the scop_ints input file"
    print "     min_num_contacts: how many pairs of residues must be within 5 amstrong if order to consider that two domains interact? "
    print "     method_name    : the name of the method that will label this interaction in the database (scop_fa_pred or pdb_int)"
    print "                      sourceDBID will be expansion by default (change in code if you want to modify it)"
    print "     --only-prots-ints: set this flag if you only want interactions seen in the PDB between proteins (no predictions made)"
    print "                  Attention: if you do not set this flag (is. ask for predictions) PDB ints will not be inserted"
    print "                  Attention! When you set this flag, you can chose between the following (both cannot be done at the same time): "
    print "                              "
    print "                             1 - inserting interactions into piana database "
    print "                                    -> this is the default behaviour of flag  --only-prots-ints"
    print "                             2 - printing an output file with SCOP-SCOP interactions when SCOPs belong to different proteins "
    print "                                    -> in order to do this, you must also set flag --create-scop-scop-file "
    print "                                    -> output will be written to file scop_scop_interactions.fa.txt "
    print "     scop_ints_file : the name of the scop_ints input file"
    print "     --no-species   : inserts all interactions, regardless of the species of proteins (this only applies when --only-prots-ints is not set"
    print "     --help         : prints this message and exits"
    print "     --verbose      : prints process info to stdout"
    print "--------------------------------------------------------------------------------------------------------------"
    print "                 REMEMBER TO REMOVE THE HEADER LINE FROM THE INPUT FILE!!!!"
    print "--------------------------------------------------------------------------------------------------------------"
        

# ---------------------------
# Function parseArguments()                                               
# --------------------------- 

def parseArguments():
    
    global piana_dbname
    global piana_dbhost
    global piana_dbuser
    global piana_dbpass
    
    global min_num_contacts
    global method_name
    
    global scop_ints_file
    global create_scop_scop_file
    
    global only_prots_ints
    global no_species
    
    global verbose
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["verbose","help","scop_ints-file=","method-name=","only-prots-ints", "create-scop-scop-file",
                                                      "piana-dbname=", "piana-dbhost=", "piana-dbuser=", "piana-dbpass=", "min-num-contacts="])
    except getopt.GetoptError:
        # print help information and exit:
        usage()
        sys.exit(2)
     
    for option,value in opts:
        
        if option == "--scop_ints-file":
            scop_ints_file = value
            
        elif option == "--piana-dbname":
            piana_dbname = value
            
        elif option == "--piana-dbhost":
            piana_dbhost = value
            
        elif option == "--piana-dbuser":
            piana_dbuser = value
            
        elif option == "--piana-dbpass":
            piana_dbpass = value
            
        elif option == "--min-num-contacts":
            min_num_contacts = int(value)
            
        elif option == "--method-name":
            method_name = value
            
        elif option == "--only-prots-ints":
            only_prots_ints = 1
            
        elif option == "--create-scop-scop-file":
            create_scop_scop_file = 1
            
        elif option == "--no-species":
            no_species = 1
            
        elif option == "--verbose":
            verbose = 1
            
        elif option == "--help":
            # print help information and exit
            usage()
            sys.exit(2)
            
    # END OF for option,value in opts:
    
    if scop_ints_file is None or method_name is None:
        sys.stderr.write("You didn't set a scop_ints file name (%s) or method_name (%s)\n" %(scop_ints_file, method_name))
        usage()
        sys.exit(2)
        
    if not min_num_contacts:
        raise ValueError("You didn't set a minimun number of contacts")
        
# --------
# --------
#  Main()               
# --------                               
# --------

piana_dbname = None
piana_dbuser = None
piana_dbhost = None
piana_dbpass = None

# file containing external DB "scop_ints" must be set on command line
scop_ints_file = None

min_num_contacts = None

method_name = None
create_scop_scop_file = 0
only_prots_ints = 0
no_species = 0 # by default 0, will not insert interactions between proteins of different species


source_db = "expansion"

# parsing arguments from the command line
parseArguments()

# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)


scop_ints_fd = open(scop_ints_file,"r")

#
# Reading external DB "scop_ints" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from scop_ints file...\n")

number_of_interactions_added = 0
number_of_interactions_dif_species = 0
number_of_interactions_no_id = 0

if create_scop_scop_file:
    scop_scop_file_fd = file("scop_scop_interactions.fa.txt", "w")
    scop_pairs_already_inserted = {}

already_processed = {} # dictionary used to know which pairs scop_1-scop_2 have already been processed

for line in scop_ints_fd:

    # Retrieving different fields in line PDB chain1 class1 fold1 sf1 fa1 dm1 sp1 px1 chain2 class2 fold2 sf2 fa2 dm2 sp2 px2
    line_fields = line.split()   # line_fields[0] is PDB
                                 # line_fields[1] is chain1:res-res
                                 # line_fields[2] is class1
                                 # line_fields[3] is fold1
                                 # line_fields[4] is sf1
                                 # line_fields[5] is fa1
                                 # line_fields[6] is dm1
                                 # line_fields[7] is sp1
                                 # line_fields[8] is px1
                                 # line_fields[9] is PDB
                                 # line_fields[10] is chain2:res-res
                                 # line_fields[11] is class2
                                 # line_fields[12] is fold2
                                 # line_fields[13] is sf2
                                 # line_fields[14] is fa2
                                 # line_fields[15] is dm2
                                 # line_fields[16] is sp2
                                 # line_fields[17] is px2
                                 # line_fields[18] is ContactsNumber

    if len(line_fields) == 19:
        # if the number of fields matches the format we are following...

        if int(line_fields[18]) < min_num_contacts:
            # ignore interactions that do not respect the minimun number of contacts
            continue


        if only_prots_ints:
            # if the user only wants interactions that are in the PDB, then
            # insert those interactions and don't do predictions
            fields_chain_prot_1 = line_fields[1].split(":")
            if len(fields_chain_prot_1)<2:
                # if there is no ':' in the line, it means only residues appear: assign A to this protein
                chain_prot_1= "A"
            else:
                chain_prot_1= fields_chain_prot_1[0]
                

            pdb_chain_protein_1 = line_fields[0] + "." + chain_prot_1
            
            fields_chain_prot_2 = line_fields[10].split(":")
            
            if len(fields_chain_prot_2)<2:
                # if there is no ':' in the line, it means only residues appear: assign A to this protein
                chain_prot_2= "A"
            else:
                chain_prot_2= fields_chain_prot_2[0]
                

            pdb_chain_protein_2 = line_fields[0] + "." + chain_prot_2


            set_proteinPiana_1 = Set(piana_access.get_list_proteinPiana(protein_id= pdb_chain_protein_1,
                                                                        id_type= "pdb.chain",
                                                                        tax_id_value= 0, source_db_info= "no"))
            
            set_proteinPiana_2 = Set(piana_access.get_list_proteinPiana(protein_id= pdb_chain_protein_2,
                                                                        id_type= "pdb.chain",
                                                                        tax_id_value= 0, source_db_info= "no"))


            if verbose:
                if not set_proteinPiana_1 or not set_proteinPiana_2:

                
                    sys.stderr.write("proteinPiana not found for pdb %s (%s) or pdb %s (%s) \n" %(pdb_chain_protein_1, set_proteinPiana_1,
                                                                                                  pdb_chain_protein_2, set_proteinPiana_2))
                    number_of_interactions_no_id += 1
                
            # END OF if verbose:


            if set_proteinPiana_1.intersection(set_proteinPiana_2):
                # if both PDBs refer to the same protein, skip interaction
                continue

            
            for one_protein_1 in set_proteinPiana_1:
                for one_protein_2 in set_proteinPiana_2:

                    if create_scop_scop_file:
                        # if user only wanted to get a file with scop-scop interactions, print it

                        if line_fields[5] <= line_fields[14]:
                            key = "%s.%s" %(line_fields[5],line_fields[14] )
                        else:
                            key = "%s.%s" %(line_fields[14],line_fields[5] )

                        if scop_pairs_already_inserted.has_key(key):
                            continue
                        
                        if verbose:
                            sys.stderr.write( "writing to file interaction between SCOP fas %s and %s\n" %(line_fields[5], line_fields[14]))
                            
                        scop_pairs_already_inserted[key] = None
                        
                        scop_scop_file_fd.write("%s\t%s\n" %(line_fields[5], line_fields[14]))

                    else:
                        if verbose:
                            sys.stderr.write( "inserting interaction: %s (%s) --> %s (%s) with method %s\n" %(one_protein_1,pdb_chain_protein_1,
                                                                                                              one_protein_2,pdb_chain_protein_2,
                                                                                                              method_name))
                        number_of_interactions_added += 1
                        # otherwise, the user wants to insert the interactions into the database
                        interactionPiana = piana_access.insert_interaction(proteinPianaA_value = one_protein_1,
                                                                           isSourceA_value = 1,
                                                                           proteinPianaB_value = one_protein_2,
                                                                           isSourceB_value = 1,
                                                                           interactionConfidence_value = 1,
                                                                           methodDescription_value = method_name,
                                                                           sourceDBDescription_value = "expansion",
                                                                           confidenceAssignedSourceDB_value=1      )
                # END OF for one_protein_2 in list_proteinPiana_1:
            # END OF for one_protein_1 in list_proteinPiana_1:


            # the user chose to insert only interactions from the PDB: skip the rest of the code...
            continue


        # ---
        # If user chose to do predictions:
        # ---
        # do:
        #    1. get all proteins that have scop fa1
        #    2. get all proteins that have scop fa2
        #    3. insert interactions between all proteins with fa1 and all proteins with fa2, checking for species
        scop_fa_1 = line_fields[5]
        scop_fa_2 = line_fields[14]

        # check that this pair has not already been processed
        if scop_fa_1 <= scop_fa_2:       key = "%s.%s" %(scop_fa_1, scop_fa_2)
        else:                            key = "%s.%s" %(scop_fa_2, scop_fa_1)
        
        if already_processed.has_key(key):
            continue

        already_processed[key] = None

        # 1. and 2.
        list_proteins_fa_1 = piana_access.get_proteins_with_scop(cf=None, sf=None, fa=scop_fa_1)
        list_proteins_fa_2 = piana_access.get_proteins_with_scop(cf=None, sf=None, fa=scop_fa_2)

        for protein_fa_1 in list_proteins_fa_1:
            set_species_1 = Set(piana_access.get_protein_taxonomy_ids(proteinPiana_value = protein_fa_1))

            for protein_fa_2 in list_proteins_fa_2:
                
                set_species_2 = Set(piana_access.get_protein_taxonomy_ids(proteinPiana_value = protein_fa_2))

                # check if proteins are of the same species. If not, interaction not inserted (except command line option says other)
                same_species = 0
                if set_species_1 and set_species_2:
                    if set_species_1.intersection(set_species_2):
                        same_species=1
                else:
                    if verbose:
                        sys.stderr.write( "Warning! No taxonomy id found for proteinPiana %s (%s) or proteinPiana %s (%s)\n" %(protein_fa_1,
                                                                                                                               set_species_1,
                                                                                                                               protein_fa_2,
                                                                                                                               set_species_2))

                if same_species or no_species:
                    # inserting the interaction if proteins belong to same species, or if command line specified to insert all interactions

                    number_of_interactions_added += 1

                    if verbose:
                        sys.stderr.write( "inserting interaction: %s --> %s with method %s\n" %(protein_fa_1, protein_fa_2 , method_name))

                    interactionPiana = piana_access.insert_interaction(proteinPianaA_value = protein_fa_1,
                                                                       isSourceA_value = 1,
                                                                       proteinPianaB_value = protein_fa_2,
                                                                       isSourceB_value = 1,
                                                                       interactionConfidence_value = 1,
                                                                       methodDescription_value = method_name,
                                                                       sourceDBDescription_value = "expansion",
                                                                       confidenceAssignedSourceDB_value=1      )


                else:
                    # different species for proteins
                    number_of_interactions_dif_species +=1
                    if verbose:
                        sys.stderr.write( "proteinA %s (species = %s) and proteinB %s (species =%s) are of different species\n" %(protein_fa_1,
                                                                                                                                  set_species_1,
                                                                                                                                  protein_fa_2,
                                                                                                                                  set_species_2) )
                
            # END OF for protein_fa_2 in list_proteins_fa_2:
        # END OF for protein_fa_1 in list_proteins_fa_1:
                
        
        if verbose:
            if not list_proteins_fa_1 or not list_proteins_fa_2:

                
                sys.stderr.write("proteinPiana not found for scop fa %s (%s) or scop fa %s (%s) \n" %(scop_fa_1, list_proteins_fa_1,
                                                                                                      scop_fa_2, list_proteins_fa_2))
                number_of_interactions_no_id += 1
                
        # END OF if verbose:
        
    # END OF if len(line_fields) == 19:
    
    else:
        sys.stderr.write("Line was not well formatted! Check format used for file! README.expansion_data_format\n")
        
# END OF for line in expansion_fd:

scop_scop_file_fd.close()

if verbose:
   sys.stderr.write( "\n------------------------------------------------------------------------\n")
   sys.stderr.write( "All done! Number of ints added: %s.\n" %(number_of_interactions_added))
   sys.stderr.write( "Number of interactions not inserted: between different species: %s\n" %(number_of_interactions_dif_species))
   sys.stderr.write( "Number of interactions with proteins unknown to PIANA: %s\n" %(number_of_interactions_no_id))
   sys.stderr.write( "------------------------------------------------------------------------\n")
