"""
File        : expansion2piana.py
Author      : Ramon Aragues
Creation    : 18.02.2004
Modification: Javier Garcia Garcia
Contents    : script that fills up tables in database piana related to external DB "expansion"
Called from : 

=======================================================================================================

This program parsers a text file (following expansion format) with interactions and inserts them into piana

The format that this parser accepts is described in README.expansion_data_format

"""

# expansion2piana.py: script that fills up tables in database piana related to external DB "expansion"
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline

import MySQLdb

import utilities

from PianaDBaccess import *
from parserUtilities import *

verbose = 0

# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------

no_species = 0 # by default 0, will not insert interactions between proteins of different species

## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("expansion-file=",None,"the name of the expansion input file"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database"),
                              ("input-id-type=",None,"the code type being used in expansion input file (normally, it should be md5)"),
                              ("num-expansions=",None,"number of expansions performed to get the expansions file (1 or 2)")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("no-species",0,"inserts all interactions, regardless of the species of proteins"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("database-version=","not defined","version of the database to be inserted"),
                           ("database-description=","Expansions using PIANA","Description of the database to be inserted")]

# Define the parser information to print in the help:

parser = ParserObject( script_name="expansion2piana.py",
                       script_description="This program fills up tables in database piana related to external DB 'expansion'",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters


arguments_dict = parser.parseArguments()


expansion_file = arguments_dict["expansion-file"]
input_id_type=arguments_dict["input-id-type"]
num_expansions = arguments_dict["num-expansions"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
no_species = arguments_dict["no-species"]



# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)


try:
    # Introduce database info into PIANA database
    piana_access.insert_new_external_database(databaseName = sourceDB,
                                              databaseVersion = database_version,
                                              parsedFile = expansion_file.split("/")[-1],
                                              databaseDescription = database_description,
                                              databaseInformation = "protein-protein interactions")
except:
    print "External database %s existed previously. Interactions are going to be inserted under the same \"external database\"" %(sourceDB)



expansion_fd = open(expansion_file,"r")

#
# Reading external DB "expansion" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from expansion file...\n")


number_of_interactions_added = 0
number_of_interactions_dif_species = 0
number_of_interactions_no_id = 0

for line in expansion_fd:

    # Retrieving different fields in line
    line_fields = line.split()   # line_fields[0] is protein_code_a
                                 # line_fields[1] is protein_code_b
                                 # line_fields[2] is expansion_name
                                 # line_fields[3] is source_edge_db_id (not using it: read file comments above)
                                 # line_fields[4] is source_protein_code
    
    if len(line_fields) == 5:
        
        protein_code_a = line_fields[0]
        protein_code_b = line_fields[1]
        protein_code_source = line_fields[4]

        expansion_type = piana_access.get_from_dict(dict_name= PianaGlobals.expansion_types, description_value=line_fields[2] )
        method_name = "exp" + expansion_type + num_expansions
            

        list_proteinPiana_a = piana_access.get_list_proteinPiana(protein_id= protein_code_a,
                                                                  id_type= input_id_type,
                                                                  tax_id_value= 0,
                                                                  source_db_info= "no")
        
        list_proteinPiana_b = piana_access.get_list_proteinPiana(protein_id= protein_code_b,
                                                                  id_type= input_id_type,
                                                                  tax_id_value= 0,
                                                                  source_db_info= "no")
        
        list_proteinPiana_source = piana_access.get_list_proteinPiana(protein_id= protein_code_source,
                                                                       id_type= input_id_type,
                                                                       tax_id_value= 0,
                                                                       source_db_info= "no")



        for proteinPiana_a in list_proteinPiana_a:
            for proteinPiana_b in list_proteinPiana_b:
                for proteinPiana_source in list_proteinPiana_source:
                    
                    list_species_a = piana_access.get_protein_taxonomy_ids(proteinPiana_value = proteinPiana_a)
                    list_species_b = piana_access.get_protein_taxonomy_ids(proteinPiana_value = proteinPiana_b)

                    # check if proteins are of the same species. If not, interaction not inserted (except command line option says other)
                    same_species = 0
                    if list_species_a and list_species_b:
                        for species_a in list_species_a:
                            if species_a in list_species_b:
                                same_species = 1
                                break
                    else:
                        if verbose:
                            sys.stderr.write( "Warning! No taxonomy id found for proteinPiana %s (%s) or proteinPiana %s (%s)\n" %(proteinPiana_a,
                                                                                                                                   list_species_a,
                                                                                                                                   proteinPiana_b,
                                                                                                                                   list_species_b))

                    if same_species or no_species:
                        # inserting the interaction if proteins belong to same species, or if command line specified to insert all interactions
                        
                        number_of_interactions_added += 1

                        if verbose:
                            sys.stderr.write( "inserting interaction: %s --> %s with method %s\n" %(proteinPiana_a, proteinPiana_b , method_name))

                        last_interaction_id = piana_access.insert_interaction(proteinPianaA_value = proteinPiana_a,
                                                                              isSourceA_value = 1,
                                                                              proteinPianaB_value = proteinPiana_b,
                                                                              isSourceB_value = 1,
                                                                              interactionConfidence_value = 1,
                                                                              methodDescription_value = method_name,
                                                                              sourceDBDescription_value = sourceDB,
                                                                              confidenceAssignedSourceDB_value=1)

                        # as we have kept the interactionPiana, we now insert specific source protein for it
                        if last_interaction_id is not None:
                            piana_access.insert_interaction_protein_source(interactionPiana_value = last_interaction_id,
                                                                           proteinPianaSource_value = proteinPiana_source,
                                                                           sourceDBDescription_value = sourceDB)
                        else:
                            sys.stderr.write("Weird: last_intearction_id is None. How can this be?\n")

                    # END OF if same_species or no_species:

                    else:
                        number_of_interactions_dif_species +=1
                        if verbose:
                            sys.stderr.write( "proteinA %s (species = %s) and proteinB %s (species =%s) are of different species\n" %(proteinPiana_a,
                                                                                                                                      list_species_a,
                                                                                                                                      proteinPiana_b,
                                                                                                                                      list_species_b) )
                    # END OF else: (if same_species or no_species:)

                # END OF for proteinPiana_source in list_proteinPiana_source:
            # END OF for proteinPiana_b in list_proteinPiana_b:
        # END OF for proteinPiana_a in list_proteinPiana_a:
        
        if verbose:
            if not list_proteinPiana_a or not list_proteinPiana_b:

                
                sys.stderr.write("proteinPiana not found for %s (%s) or %s (%s) \n" %(protein_code_a, list_proteinPiana_a,
                                                                                      protein_code_b, list_proteinPiana_b))
                number_of_interactions_no_id += 1
                
        # END OF if verbose:
        
    # END OF if len(line_fields) == 5:
    
    else:
        sys.stderr.write("Line was not well formatted! Check format used for file! README.expansion_data_format\n")
        
# END OF for line in expansion_fd:

        
if verbose:
   sys.stderr.write( "\n------------------------------------------------------------------------\n")
   sys.stderr.write( "All done! Number of ints added: %s.\n" %(number_of_interactions_added))
   sys.stderr.write( "Number of interactions not inserted: between different species: %s\n" %(number_of_interactions_dif_species))
   sys.stderr.write( "Number of interactions with proteins unknown to PIANA: %s\n" %(number_of_interactions_no_id))
   sys.stderr.write( "------------------------------------------------------------------------\n")

piana_access.close()
