"""
File        : genpept2piana.py
Author      : Ramon Aragues
Creation    : 9.2004
Modifications: 2.2007 (Javier Garcia)
Contents    : fills up tables in database piana with information from genpept
Called from : 

=======================================================================================================

This file implements a program that fills up tables in database piana with information from genpept using biopython
libraries and methods

genpept can be downloaded from ftp://ftp.ncbi.nih.gov/genbank/relXXX.fsa_aa.gz

Before running, taxonomy table of piana must be populated (use taxonomy2piana for that)

It is advisable to parse swissprot and trembl before genpept

"""

# genpept2piana.py: fills up tables in database piana with information from genpept
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import MySQLdb

import utilities

import time

from PianaDBaccess import *
from parserUtilities import *

from Bio import Fasta  # needed to read the genpept file (which is in fasta format)

# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing genpept data (relxxx_fsa_aa)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("tax-id-file=",None,"a text file with two tab-separated columns: 1st is gi code and 2nd is taxonomy id"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]
                           
# Define the parser information to print in the help:

parser = ParserObject( script_name="genpept2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from genpept using biopython libraries and methods)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
tax_id_file= arguments_dict["tax-id-file"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]



# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')


# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

# All these tags will be considered to be pointing to id type Accession
accepted_coll_accessions = { "gb": None,
                             "dbj": None,
                             "emb": None}

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.protein_table,
                  PianaGlobals.sequence_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.species_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.proteinDescription_table,
                  PianaGlobals.temporalData_table]                             

            
# --------
# --------
#  Main()               
# --------                               
# --------

num_seqs_tax_not_known = 0

if verbose:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )



genpept_parser = Fasta.RecordParser()
genpept_input_file =  utilities.open_file(input_file, 'r')
genpept_iterator = Fasta.Iterator(genpept_input_file, genpept_parser)

genpept_record = genpept_iterator.next()


protein_number=0


if verbose:
    sys.stderr.write("Loading taxonomies\n")
# get dictionary for gis and tax ids
dic_gi_tax = utilities.return_dic_gi_vs_tax(file_name=tax_id_file)

if verbose:
    sys.stderr.write( "Processing file\n")
# while record read is not None, parse the record and insert data into piana
while genpept_record is not None:

    protein_number += 1

    parser.add_to_log("Number of proteins processed")

    if verbose:
        sys.stderr.write( "===================================\n")
        sys.stderr.write( "            NEW ENTRY\n")
        sys.stderr.write( "===================================\n")


    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if protein_number%20000==0:
            sys.stderr.write("%s proteins done in %s seconds\n" %(protein_number,time.time()-initial_time))


    protein_title_line = genpept_record.title
    protein_sequence = genpept_record.sequence.strip()

    if verbose:
        #sys.stderr.write("record is : %s\n" %(genpept_record))
        sys.stderr.write("title line is : %s\n" %(protein_title_line)) 
        #sys.stderr.write("sequence is : %s\n" %(protein_sequence))
        
    # title looks like this: gi|gi_id|sourceDB|acc| protein name [species]
    #                        - gi|4105707|gb|AAD02507.1| carbamate kinase [Trichomonas vaginalis]
    #                        - gi|4105707|gb|AAD02507.1| carbamate kinase [Trichomonas vaginalis]
    #                        - gi|4105707|gb|AAD02507.1| carbamate kinase
    #                        - gi|4105707|dbj|BBAS02507.1| carbamate kinase 4243
    #                        
    #    (character '>' from input file has been removed)
        
    title_atoms = protein_title_line.split('|')

    gi_id = int(title_atoms[1])

    if accepted_coll_accessions.has_key(title_atoms[2]):
        # we consider accessions from sources in accepted_coll_accessions
        coll_acc = title_atoms[3].strip()

        # Split Accession Number and Version
        try:
            splited = re.split("\.",coll_acc)
            accessionNum = splited[0]
        except:
            accessionNum = coll_acc

    else:
        accessionNum = ""

    temp_protein_name = title_atoms[4] # this atom is a bit more complex to process...


    # code to parse genpept titles provided by Wan Kyu Kim and Christoph Winter
    p = re.compile("(\S+)?\s+(.*?)(\[(.*?)\])?\s*$")
    r = p.match(temp_protein_name)

    # embl_accession = r.group(1)      # ignoring it: I think the new genpept version does not include it
    rest_of_protein_name = r.group(2)
    species_name= r.group(4)

    protein_name = rest_of_protein_name.strip().replace("\\", " ").replace('"'," ").replace("'"," ")    # remove " and ' characters
        
    if dic_gi_tax.has_key(gi_id):
        # we give priority to the dictionary with tax ids for gis
        list_tax_id_value = [dic_gi_tax[gi_id]]
    else:
        if species_name:
            # if no tax_id was known for this gi, try to get it from the species name
            list_tax_id_value = piana_access.get_taxonomies_from_species_name(species_name_value = species_name)
        else:
            list_tax_id_value = []

    num_taxonomies = len(list_tax_id_value)
    if num_taxonomies == 1:
        tax_id_value= list_tax_id_value[0]
        if tax_id_value==0:
            parser.add_to_log("Number of sequences with tax id 0 taken from gi-taxID file")
    else:
        tax_id_value = 0
        num_seqs_tax_not_known += 1
        parser.add_to_log("Number of sequences with tax not known (no tax id associated or more than one)")
	#sys.stderr.write("Species per entry gi:%s accession: %s species: %s\n" %(gi_id,accessionNum,list_tax_id_value))
        #genpept_record = genpept_iterator.next()
	#continue

        
    if verbose:
        sys.stderr.write("Title processed and fields obtained are:\n")
        sys.stderr.write("   - gi_id: %s\n" %(gi_id))
        sys.stderr.write("   - Accession Number: %s\n" %(accessionNum))
        sys.stderr.write("   - protein_name: %s\n" %(protein_name))
        sys.stderr.write("   - species_name: %s\n" %(species_name))
        sys.stderr.write("   - tax_id: %s (extracted from file:%s)\n" %(tax_id_value, dic_gi_tax.has_key(gi_id)))

    if protein_sequence:
        # Insert protein: insert_protein takes care of checking whether the (sequence, tax) was already there or not
        inserted_proteinPiana = piana_access.insert_protein(proteinSequence_value= protein_sequence,
                                                            tax_id_value= tax_id_value,
							    sourceDBID=sourceDB	)
        
        parser.add_to_log("Number of proteins inserted")


    """
    Now, we have a proteinPiana for the (sequence, tax_id), either the one already existing before or a new one.

    Just add protein information to database tables.

    TO DO!! this would be a good place to check for consistency between uniprot and genbank, using embl accessions
    """
    
    # --- table gi ---

    # inserting gi ID  for current protein
    if gi_id:
        piana_access.insert_gi_code(gi_code_value= gi_id, proteinPiana_value= inserted_proteinPiana,
                                    gi_source_value = sourceDB)
        parser.add_to_log("Number of gi codes inserted")
        

    # --- table accessionNumber ---

    # inserting accession Number for current protein
    if accessionNum != "":
        piana_access.insert_accessionNumber(accessionNumber_value= accessionNum, proteinPiana_value= inserted_proteinPiana,
                                            accessionNumber_source_value = sourceDB)
        parser.add_to_log("Number of accessionNumber inserted")

        
    # --- table embl accession ---

    # inserting embl  Accession  for current protein
    ## if coll_ac:

##         # I am not 100% sure how genpept works now... from what I understand, they have
##         # unified embl accession, genbank accession and DNA dbj
##         # For the time being, I consider this accession as an embl accession...
##         # --> insert_emblAccession_code takes care of processing the version after the point (if it is there)
##         piana_access.insert_emblAccession_code(emblAccession_code_value= coll_ac,
##                                                proteinPiana_value= inserted_proteinPiana,
##                                                emblAccession_source_value= sourceDB)
        
        
     # --- table description ---

    # inserting description (protein name) for current protein
    if protein_name != "":

        piana_access.insert_protein_description(description= protein_name ,
                                                proteinPiana_value= inserted_proteinPiana,
                                                proteinDescriptionSource_value= sourceDB)
        parser.add_to_log("Number of protein descriptions inserted")
        
    # --- table species ---

    # inserting species for current protein
    #if tax_id_value:
        # if a tax_id was already found for this protein, use it
        #piana_access.insert_protein_species(tax_id =  tax_id_value,
        #                                    proteinPiana_value = inserted_proteinPiana,
        #                                    proteinSpeciesSource_value = sourceDB)
        #parser.add_to_log("Number of protein species inserted")

    #elif species_name:
        # in case tax_id is 0, it can mean several species were found for the protein...
        # insert them...

        #for species_tax_id in list_tax_id_value:
        #    
        #    if verbose:
        #        sys.stderr.write( "adding protein_tax_id %s (species_name=%s) for proteinPiana %s (gi=%s)\n" %(
        #            species_tax_id,
        #            species_name,
        #            inserted_proteinPiana,
        #            gi_id))

         #   piana_access.insert_protein_species(tax_id =  species_tax_id,
         #                                       proteinPiana_value = inserted_proteinPiana,
         #                                       proteinSpeciesSource_value = sourceDB)
         #   parser.add_to_log("Number of protein species inserted")
        # END OF for species_tax_id in list_species_tax_id:

    
    # reading next record
    if verbose:
        sys.stderr.write( "reading next record\n")
        
    genpept_record = genpept_iterator.next()
    
# END OF while genpept_record is not None



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(protein_number, (time.time()-initial_time) ))

if verbose:
    sys.stderr.write("All done! Number of sequences for which the tax_id was unknown (no tax id associated or more than one): %s\n" %(num_seqs_tax_not_known))


if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
