"""
File        : go2piana.py
Author      : Pablo Boixeda & Ramon Aragues
Modifications: Javier Garcia, March 2007
Creation    : 06.2005
Contents    : program that fills up tables in database piana with information from GO
Called from : command line

=======================================================================================================

"""

# go2piana.py: program that fills up tables in database piana with information from GO
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys
import getopt

import re
import readline
import MySQLdb

from PianaGraph import *
from PianaGraphBuilder import *
from PianaDBaccess import *

from PianaDBaccess import *
import PianaGlobals

from parserUtilities import *

import time


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("go-dbname=",None,"name of the go mysql database"),
                              ("go-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("go-dbuser=",None,"name of the mysql go username"),
                           ("go-dbpass=",None,"name of the mysql go username"),
                           ("insert-protein-go",None,"if this flag is set, then insert relationship between proteins and go terms"),
                           ("insert-go-info",None,"if this flag is set, then insert go information"),
                           ("insert-go-distance",None,"if this flag is set, then calculate distance between go terms"),
                           ("threshold=",None,"This is the maximum distance that will be calculated (once reached, distance(term1, term2) will be set to INFINITE"),
                           ("input-file=",None,"these parameter (and input-proteins-type) is only required if you don't want to insert info for all proteins. Calculates distance only for gos associated to proteins in the network (depth 1) built from the root proteins in the file"),
                           ("input-proteins-type=",None,"type of the proteins inside the input_file. Attention! Do not use geneName to identify these proteins: they are ambiguous and no species information is available"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]



# Define the parser information to print in the help:

parser = ParserObject( script_name="go2piana.py",
                       script_description="This program fills up tables in database piana with information from GO",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )



# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)


arguments_dict = parser.parseArguments()

insert_protein_go = arguments_dict["insert-protein-go"]
insert_go_info = arguments_dict["insert-go-info"]
insert_go_distance = arguments_dict["insert-go-distance"]

input_proteins_type = arguments_dict["input-proteins-type"]
threshold = arguments_dict["threshold"]
input_file = arguments_dict["input-file"]

go_dbname = arguments_dict["go-dbname"]
go_dbhost = arguments_dict["go-dbhost"]
go_dbuser = arguments_dict["go-dbuser"] 
go_dbpass = arguments_dict["go-dbpass"]


piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]

sourceDB= arguments_dict["database-name"]

database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
database_information= arguments_dict["database-information"]

help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]




verbose_detailed = 0

initial_time = time.time()


INFINITE_DISTANCE = 100


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.go_table,
                  PianaGlobals.proteinGo_table,
                  PianaGlobals.go_term2term_distance_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.pdb_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.temporalData_table]




# TO DO!!! Let the user set a tax_id so he can use any type of protein code
#          Then, this tax_id will be given to PianaGraphBuilder as an argument



   

def get_fathers(gocursor=None,term_id=None):
   
   sqlquery="select term1_id from term2term where term2_id=%s"%term_id
   gocursor.execute(sqlquery)
   fathers=gocursor.fetchall()
   fathers_list=[]
   for father in fathers:
      fathers_list.append(father[0])


   return fathers_list

def calculate_distance(distance_up_to_here=None, term_id1=None, term_id2=None, gocursor=None):

   global threshold

   if gocursor is None:
      raise ValueError("No gocursor")

   if distance_up_to_here>= threshold:
      # if distance found is higher than the threshold, stop the search and return INFINITE_DISTANCE
      return INFINITE_DISTANCE

   if term_id1==term_id2:
      return distance_up_to_here

   else :
      fathers_term_id1=get_fathers(gocursor, term_id1)
      fathers_term_id2=get_fathers(gocursor, term_id2)

      results=[]

      for father_term1 in fathers_term_id1:
         results.append(calculate_distance(distance_up_to_here=distance_up_to_here+1,
                                           term_id1=father_term1,
                                           term_id2=term_id2,
                                           gocursor=gocursor))
      for father_term2 in fathers_term_id2:
         results.append(calculate_distance(distance_up_to_here=distance_up_to_here+1,
                                           term_id1=term_id1,
                                           term_id2=father_term2,
                                           gocursor=gocursor))

      if not results:
         return distance_up_to_here
      else :
         return min(results)
         

def insert_go_table(gocursor, piana_access, go_terms, depth, source_db):


   if not go_terms:
      pass      # no more sons... end recursion
   
   elif go_terms[0]==1:
      
      sqlquery = "select id, name, term_type, acc from term where is_obsolete=0 and id=%s" %go_terms[0]
      gocursor.execute(sqlquery)
      go_rows= gocursor.fetchall()

      # go_rows is a list with:
      #     [0][0] -> term id
      #     [0][1] -> name
      #     [0][2] -> term type
      #     [0][3] -> go accession
      
      piana_access.insert_go(go_id=go_rows[0][0],go_name=go_rows[0][1], term_type=go_rows[0][2], acc=go_rows[0][3], distance2root=depth, source_db= source_db)

      sqlquery="select term2_id from term2term where term1_id=%s" %go_rows[0][0]
      gocursor.execute(sqlquery)
      goSons=gocursor.fetchall()

      # TO CHECK!!!! goSons is a list or a list of lists?

      insert_go_table(gocursor=gocursor, piana_access=piana_access, go_terms=goSons, depth=depth+1, source_db=source_db)

   else :
      
      for go_term in go_terms:
         sqlquery = "select id, name, term_type, acc from term where is_obsolete=0 and id= %s" %go_term
         gocursor.execute(sqlquery)
         go_rows= gocursor.fetchall()

         for goRow in go_rows:

            if verbose_detailed:
               sys.stderr.write("Inserting go_id=%s whith depth=%s\n"%(goRow[0],depth))

            piana_access.insert_go(go_id=goRow[0], go_name=goRow[1], term_type=goRow[2], acc=goRow[3], distance2root=depth, source_db=source_db)
            
            sqlquery="select term2_id from term2term where term1_id=%s"%goRow[0]
            gocursor.execute(sqlquery)
            goSons=gocursor.fetchall()
            
            insert_go_table(gocursor=gocursor,piana_access=piana_access,go_terms=goSons,depth=depth+1, source_db= source_db)
         # END OF for goRow in go_rows
      # END OF for go_term in go_terms
   # END OF else: (if go_terms[0]==1:)
   

def go2piana(dbname, dbhost, dbuser, dbpassword,go_dbname,go_dbhost,go_dbuser,go_dbpass, input_file, input_proteins_type, tables_to_lock):
   # Initialisating connection to piana

   if verbose:
      sys.stderr.write("Opening connection to piana\n")

   piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

   # Introduce database info into PIANA database
   parsed_file = "mysql_database_%s" %(go_dbname)
   
   try:
      piana_access.insert_new_external_database(databaseName = sourceDB,
                                                databaseVersion = database_version,
                                                parsedFile = parsed_file,
                                                databaseDescription = database_description,
                                                databaseInformation = database_information)
   except:
      print "Database %s existed previously. Using it." %(sourceDB)


   # Locking tables to be updated
   piana_access.lock_tables(table_list = tables_to_lock )
   


   

    
   # opening connection to MySQL GO database and create a cursor to work with the database
   if verbose:
      sys.stderr.write("Opening connection to go\n")

   if go_dbuser is None and go_dbpass is None:
      godb = MySQLdb.connect(db=go_dbname, host=go_dbhost)

   elif go_dbpass is None and go_dbuser is not None:
      godb = MySQLdb.connect(user=go_dbuser, db=go_dbname, host=go_dbhost)

   else:
      godb = MySQLdb.connect(user=go_dbuser, db=go_dbname, host=go_dbhost, passwd= go_dbpass )

   gocursor = godb.cursor()


   number_genes_without_proteinPiana = 0


   counter = 0

   
   # --------
   # 1. inserting associations between proteins and go terms
   # --------
   if insert_protein_go:
      
      if verbose:
         sys.stderr.write("Retrieving protein go association\n")

      # First of all, obtain the relation between gene_product_id and term_id in table association
      sqlquery= "select term_id, gene_product_id from association"
      gocursor.execute(sqlquery)
      go_gene_product_id=gocursor.fetchall()

      if verbose:
         num_associations = len(go_gene_product_id)
         sys.stderr.write("%s associations found\n" %(num_associations))

      i = 0

      
      if verbose:
         sys.stderr.write("Retrieving Internal Go references for sequences\n")
      # Second, obtain from table term, those values that term_type is sequence, in order to obtain the id that identifies tye type_id in gene_product table
      # This will allow to known what kind of code they are giving to us
      sqlquery= "select id, name from term where term_type=\"sequence\""
      gocursor.execute(sqlquery)
      term_type_id_relation = gocursor.fetchall()

      # Tranform this data into a dictionary that contains the table where the code must be searched
      code_type = {}
      for term_type in term_type_id_relation:
         if term_type[1] == "gene":
            geneName_identifier = term_type[0]   #used to see if we have to check for the specie or not
            # If term type is gene, it is supposed to be a geneName
            code_type[term_type[0]] = "geneName"
         elif term_type[1] == "protein":
            uniprot_identifier = term_type[0]
            # If term type is protein, it is supposed to be uniprotEntry code (or uniprotAcc)
            code_type[term_type[0]] = ["unientry", "uniacc"]
         elif term_type[1] == "protein_structure":
            pdb_identifier = term_type[0]        #used to transform to Piana PDB nomenclature
            # If term type is protein_structre, it is supposed to be a PDB_CHAIN code
            code_type[term_type[0]] = "pdb.chain"
         elif term_type[1] == "transcript":
            # TODO (Really? In last version only there were 6 cases...)
            code_type[term_type[0]] = None
         elif term_type[1] == "complex":
            # TODO (Really? In the last version only were 12 cases...)
            code_type[term_type[0]] = None


      # Then, foreach relation gene_product - go_term, insert it into the database
      for pair in go_gene_product_id:
        # pair[0] -> term_id
        # pair[1] -> gene_product_id


        counter += 1

        # Check the frequency of lock/unlocking tables
        piana_access.check_lock_frequency(tables_to_lock)


        if not counter%20000:
           current_time = time.time()
           sys.stderr.write("%s\t%s\n" %(counter,current_time-initial_time) )

        #sqlquery="select symbol, species_id, type_id, dbxref_id from gene_product where id=%s" %pair[1]
        sqlquery="select symbol, species_id, type_id, dbxref_id from gene_product where id=%s" %pair[1]
        gocursor.execute(sqlquery)
        gene_to_symbol =  gocursor.fetchall() 

        list_proteinPiana = [] # initialize list of proteinPianas to which this GO term will be associated
        for symbol in gene_to_symbol:

           # Check if it is a geneName, to check for the specie
           if symbol[2]==geneName_identifier:
              # tax id has to be retrieved from table species
              sqlquery="select ncbi_taxa_id from species where id=%s" %symbol[1]
              gocursor.execute(sqlquery)
              tax_id =  gocursor.fetchall()

              # TO DO!!!
              # Attention! I am ignoring the association gene - term id when
              #            GO is not correctly assigning tax_id to the geneName
              #            For example, GO says that AAH1 is Candida Albicans (tax id=5476)
              #            when in fact, AAH1 does not exist in that species but in another
              #            one called Candida albicans SC5314  (tax id=237561)

              # Get the list of proteinPianas
              list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id=symbol[0].replace('"'," ").strip(),
                                                                          id_type= code_type[symbol[2]],
                                                                          tax_id_value=tax_id[0][0], source_db_info= "no", only_primary="yes"))


           elif symbol[2]==pdb_identifier:
              # GO PDB format is like 9PAP_@,9GPB_D, and Piana PDB format is like 9PAP.A,9GPB.D... So, it is necessary to adapt the format
              # Get the list of proteinPianas
              list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id=symbol[0].replace('"'," ").replace('_','.').replace('@','A').strip(),
                                                                          id_type= code_type[symbol[2]],
                                                                          tax_id_value=0, source_db_info= "no"))
              
           elif symbol[2]==uniprot_identifier:
              # Get the list of proteinPianas
              uniprot_proteinPianas = piana_access.get_list_proteinPiana(protein_id=symbol[0].replace('"'," ").strip(),
                                                                         id_type= code_type[symbol[2]][0],
                                                                         tax_id_value=0, source_db_info= "no")

              # Check if it has been found the uniprotEntry. If not, search for the uniprotAccession:
              if not uniprot_proteinPianas:
                 # Search the uniprotAcc correspondences in dbxref table
                 sqlquery="select xref_key from dbxref where id=%s" %symbol[3]
                 gocursor.execute(sqlquery)
                 uniprotAcc = gocursor.fetchall()
                 uniprot_proteinPianas = piana_access.get_list_proteinPiana(protein_id=uniprotAcc[0][0].replace('"'," ").strip(),
                                                                            id_type = code_type[symbol[2]][1],
                                                                            tax_id_value=0, source_db_info="no")

                 

              list_proteinPiana.extend(uniprot_proteinPianas)
              
           else:
              # It is a code not taken into account (transcript or complex)
              pass



        # END OF for symbol in gene_to_symbol:

        if verbose_detailed:
           sys.stderr.write("Inserting go for proteinPianas %s (symbol %s, tax %s) and go_term_id %s\n" %(list_proteinPiana,symbol[0], tax_id_value,
                                                                                                          pair[0]))
           
        
        # Insert GO information for proteinPianas associated to symbol
        for proteinPiana in list_proteinPiana:

           piana_access.insert_protein_go(go_id= pair[0], proteinPiana_value= proteinPiana, proteinGoSource_value=sourceDB)
           parser.add_to_log("Number of inserted proteinGos")

        # END OF for proteinPiana in list_proteinPiana:

        if not list_proteinPiana:
           number_genes_without_proteinPiana += 1
           parser.add_to_log("NUmber of genes without proteinPiana")

        if verbose_detailed:
           if not list_proteinPiana:
              sys.stderr.write("No proteinPiana found for symbol %s tax id %s\n" %(symbol[0], tax_id[0][0]))


        # END OF for symbol in gene_to_symbol:
      # END OF for pair in go_gene_product_id:

      
   # END OF if insert_protein_go:


   # Unlocking and locking tables before starting the next step
   # and empty buffer
   piana_access.execute_insert_buffer()
   piana_access.unlock_tables()

   sys.stderr.write("proteinGo inserted...")
   
   piana_access.lock_tables(table_list = tables_to_lock )
   
   
   # --------
   # 2. inserting information into table go
   # --------

   if insert_go_info:

   
      if verbose:
         sys.stderr.write("calling insert_go_table\n")
      
      # filling go table into piana go_table
      insert_go_table(gocursor=gocursor, piana_access=piana_access, go_terms=[1], depth=0, source_db=sourceDB)
      
   # END OF if insert_go_info:


   # Unlocking and locking tables before starting the next step
   # and empty buffer
   piana_access.execute_insert_buffer()
   piana_access.unlock_tables()

   sys.stderr.write("go table inserted...")
   
   piana_access.lock_tables(table_list = tables_to_lock )

   # --------
   # 3. inserting distance between go terms
   # --------

   if insert_go_distance:


      ## HERE THE LOCK TABLES IS NOT USED, AS IT USES ALIAS IN SOME QUERIES... IT SHOULD BE CHANGED...

      piana_access.unlock_tables()


      #   term_types = ["biological_process",
      #                 "molecular_function",
      #                 "celullar_component"]

      term_types = ["molecular_function"]  # this is here only to speed up the process... normally, do parsing for all term types

      for term_type in term_types:

         if input_file is None:
            # if no list provided, calculate distances for all gos
            sqlquery="""select distinct id from term where is_obsolete=0 and term_type="%s" """ %term_type
            
            gocursor.execute(sqlquery)
            go_ids=gocursor.fetchall()
         else:
            query_proteins = []

            for line in file(input_file,  "r"):
               query_proteins.append( line.strip() )

            # building a network to then get the proteins in the network (and then get the go terms for these proteins)
            temp_piana_graph = PianaGraph("Temp Graph", piana_access_object= piana_access)  
            temp_piana_builder = PianaGraphBuilder(piana_access_object=piana_access,
                                                   use_self_ints='yes',
                                                   depth=1,
                                                   hub_threshold= 0,
                                                   list_protein_codes= query_proteins,
                                                   id_type = input_proteins_type,
                                                   tax_id_value=0 ,
                                                   list_source_dbs= "all",
                                                   list_source_methods= "all")
            
            temp_piana_graph.build_graph(temp_piana_builder)


            
            proteinPiana_list=temp_piana_graph.get_node_ids_list()
            go_list=[]

            if verbose:
               sys.stderr.write("getting go terms for proteinPianas %s\n" %(proteinPiana_list))
            
            for proteinPiana_id in proteinPiana_list:
               go_list.extend(piana_access.get_protein_go_term_id(proteinPiana_value=proteinPiana_id,
                                                                  term_type_value=term_type))


            if verbose:
               sys.stderr.write("Go list obtained is %s\n" %(go_list))
               
            go_ids = []
            gos_already_appended = {}  # used to avoid duplication of go terms
            for go_term in go_list:
               if not gos_already_appended.has_key(go_term):
                  # we have to simulate a fetchall list (ie a list of lists, where [0] in each sublist is the go term
                  go_ids.append( (int(go_term), 1) )
                  gos_already_appended[go_term] = None
         # END OF else: (if input_file is None:)

         num_gos = len(go_ids)


         for i in range(0, num_gos):
            for j in range(i+1, num_gos):

               if verbose:
                  sys.stderr.write("%s->i=%s j=%s=" %(num_gos, i, j))

               term2term_distance=calculate_distance(distance_up_to_here=0, term_id1=go_ids[i][0], term_id2=go_ids[j][0], gocursor=gocursor)

               if verbose:
                  sys.stderr.write("%s--" %(term2term_distance))

               if term2term_distance != INFINITE_DISTANCE:
                  # if distance is infinite, do not insert info...
                  if verbose_detailed:
                     sys.stderr.write("Inserting distance %s for go terms %s and %s\n" %(term2term_distance,go_ids[i][0],
                                                                                         go_ids[j][0] ))
                     
                  piana_access.insert_go_term2term_distance(term1_id= go_ids[i][0], term2_id=go_ids[j][0], distance=term2term_distance)
            # END OF for j in range(i+1, num_gos):
         # END OF for i in range(num_gos):

      # END OF for term_type in term_types:
   # END OF if insert_go_distance:

   parser.add_to_log("Number of genes for which a proteinPiana could not be found")
   if verbose:
      sys.stderr.write("Number of genes for which a proteinPiana could not be found: %s\n" %(number_genes_without_proteinPiana))


   piana_access.close()

# --------
# --------
#  Main()
# --------
# --------

threshold= INFINITE_DISTANCE   # if nothing is said on command line, set no threshold (1000 is the same as setting no threshold...)

go2piana(piana_dbname, piana_dbhost, piana_dbuser, piana_dbpass, go_dbname, go_dbhost, go_dbuser, go_dbpass, input_file, input_proteins_type, tables_to_lock)

sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

#piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s seconds\n" %(time.time()-initial_time))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

