"""
File        : hprd2piana.py
Author      : Javier Garcia
Creation    : April 2007
Contents    : inserts information from an hprd  txt file database into PIANA
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from a HPRD database

--> HPRD database must be in txt tabulated format

"""

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline
import MySQLdb


from PianaDBaccess import *
from parserUtilities import *


import time


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("input-file-name=",None,"file name of input file containing HPRD tabulated txt data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="hprd2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from a HPRD database (in tabulated format)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

arguments_dic = parser.parseArguments()

input_file = arguments_dic["input-file-name"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
time_control = arguments_dic["time-control"]
log_file = arguments_dic["log-file"]
database_information = arguments_dic["database-information"]


# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()



if log_file:
    log_file_fd = file(log_file, 'w')


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.geneID_table,
                  PianaGlobals.validProteinIds_table,
                  PianaGlobals.protein_table]

# --------
# --------
#  Main()               
# --------                               
# --------

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Insert a new protein identifier type: Intact
piana_access.insert_new_id_type(proteinTypeId="hprd", externalIdDescription="HPRD identifier for the protein")


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )


# Lock_number is used to calculate the frequency to do unlock...
line_number=0
header=0


# Format in April 2007:
#Protein A HPRD identifier    Protein A Entrez Gene   Protein B HPRD identifier       Protein B Entrez Gene   PubMed  Experiment Type

# Example:
#HPRD ID:00010   Entrez Gene:43  HPRD ID:00010   Entrez Gene:43  7085778 in vivo
#HPRD ID:00010   Entrez Gene:43  HPRD ID:00100   Entrez Gene:351 12427014        in vitro
#HPRD ID:00010   Entrez Gene:43  HPRD ID:00359   Entrez Gene:1282        12524166        in vitro
#HPRD ID:00010   Entrez Gene:43  HPRD ID:01029   Entrez Gene:3912        15474030        in vivo; in vitro; yeast-two hybrid
#HPRD ID:00010   Entrez Gene:43  HPRD ID:01033   Entrez Gene:284217      12524166        in vitro
#HPRD ID:00010   Entrez Gene:43  HPRD ID:01057   Entrez Gene:1939        7085778 in vivo


if verbose:
    sys.stderr.write("Reading data from hprd file...\n")


for line in input_file_fd:

    # Skip header line
    if header==0:
        header=1
        continue
    else:
        line_number += 1

    parser.add_to_log("Number of processed lines")

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if line_number%1000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))


    line_fields = line.split('\t')  # line_fields[0] is Protein A HPRD identifier
                                    # line_fields[1] is Protein A Entrez Gene
                                    # line_fields[2] is Protein B HPRD identifier
                                    # line_fields[3] is Protein B Entrez Gene
                                    # line_fields[4] is PubMed
                                    # line_fields[5] is Experiment Type

    # First step: obtain list of proteinPianas for interactors
    name_hprd_intA = None
    name_hprd_intB = None
    name_geneID_intA = None
    name_geneID_intB = None
    
    hprd_re = re.search("HPRD ID:(\d+)",line_fields[0])
    if hprd_re:
        name_hprd_intA = hprd_re.group(1)
    hprd_re = re.search("HPRD ID:(\d+)",line_fields[2])
    if hprd_re:
        name_hprd_intB = hprd_re.group(1)

    geneID_re = re.search("Entrez Gene:(\d+)",line_fields[1])
    if geneID_re:
        name_geneID_intA = geneID_re.group(1)
        
    geneID_re = re.search("Entrez Gene:(\d+)",line_fields[3])
    if geneID_re:
        name_geneID_intB = geneID_re.group(1)

    if name_geneID_intA is None or name_geneID_intB is None:
        print line
        parser.add_to_log("Interaction not inserted because GeneID not found [%s]" %(line) )
        continue
        
    list_proteinPianas_intA = []
    list_proteinPianas_intB = []

    list_pubmed_ids = line_fields[4].split(',')
    list_detection_methods = [x.strip() for x in line_fields[5].split(';')]


    # Search proteinPianas for interactor A (we fix taxID as it must be human proteins)
    list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana( protein_id= name_geneID_intA,
                                                                        id_type="geneID",
                                                                        tax_id_value=9606, source_db_info="no") )

    # Search proteinPianas for interactor B (we fix taxID=9606 as it must be human proteins)
    list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana( protein_id= name_geneID_intB,
                                                                        id_type="geneID",
                                                                        tax_id_value=9606, source_db_info="no") )

    if( len(list_proteinPianas_intA)==0 ):
        if verbose:
            sys.stderr.write("No proteinPiana found for geneID %s and species human" %(name_geneID_intA) )
        parser.add_to_log("Number of geneID not found in database")
    if( len(list_proteinPianas_intB)==0 ):
        if verbose:
            sys.stderr.write("No proteinPiana found for geneID %s and species human" %(name_geneID_intA) )
        parser.add_to_log("Number of geneID not found in database")
    
                                 
    ## The next step is to introduce interaction into database

    # Insert into database correspondences between intact code and proteinPianas:
    # TODO INSERT THE INTACT CODE IN externalProteinId table...	  
    for proteinPianaA in list_proteinPianas_intA:
        for proteinPianaB in list_proteinPianas_intB:
            piana_access.insert_interaction(proteinPianaA_value = proteinPianaA,
                                            isSourceA_value = 1,
                                            proteinPianaB_value = proteinPianaB,
                                            isSourceB_value = 1,
                                            interactionConfidence_value = 1,
                                            methodDescription_value = list_detection_methods,
                                            sourceDBDescription_value = sourceDB,
                                            confidenceAssignedSourceDB_value=1,
                                            pubmed_id_value= list_pubmed_ids)
            parser.add_to_log("Number of interactions inserted")

    # Insert hprd identifiers for these proteins
    for proteinPianaA in list_proteinPianas_intA:
        piana_access.insert_external_id_code(external_id_type="hprd",
                                             external_id_value=name_hprd_intA,
                                             proteinPiana_value = proteinPianaA,
                                             sourceDBID = sourceDB)
        parser.add_to_log("Number of hprd identifiers added")
        
    for proteinPianaB in list_proteinPianas_intB:
        piana_access.insert_external_id_code(external_id_type="hprd",
                                             external_id_value=name_hprd_intB,
                                             proteinPiana_value = proteinPianaB,
                                             sourceDBID = sourceDB)
        parser.add_to_log("Number of hprd identifiers added")
        

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
   

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

if verbose:
    sys.stderr.write("%s\t%s\n" %(line_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
    sys.stderr.write(parser.get_log_string())
