"""
File        : intact2piana.py
Author      : Javier Garcia
Creation    : March 2007
Contents    : inserts information from a Intact txt file database into PIANA
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from a intact database

--> Intact database must be in txt tabulated format

"""

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline
import MySQLdb


from PianaDBaccess import *
from parserUtilities import *


import time


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("input-file-name=",None,"file name of input file containing intact txt data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="intact2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from a intact database",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

arguments_dic = parser.parseArguments()

input_file = arguments_dic["input-file-name"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
time_control = arguments_dic["time-control"]
log_file = arguments_dic["log-file"]
database_information = arguments_dic["database-information"]


# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()



if log_file:
    log_file_fd = file(log_file, 'w')


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.ensembl_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.proteinPianaCounter_table]

# --------
# --------
#  Main()               
# --------                               
# --------

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Insert a new protein identifier type: Intact
piana_access.insert_new_id_type(proteinTypeId="intact", externalIdDescription="intact identifier for the protein")


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )


# Lock_number is used to calculate the frequency to do unlock...
line_number=0


# Format in March 2007:
# ID interactor A ID interactor B Alt. ID interactor A    Alt. ID interactor B    Alias(es) interactor A  Alias(es) interactor B  Interaction detection me
#thod(s) Publication 1st author(s)       Publication Identifier(s)       Taxid interactor A      Taxid interactor B      Interaction type(s)     Source d
#atabase(s)      Interaction identifier(s)       Confidence value(s)

# Example:
# ID interactor A: uniprotkb:Q08641
# ID interactor B: uniprotkb:P26781
# Alt. ID interactor A: uniprotkb:ABP140(gene name)
# Alt. ID Interactor B: uniprotkb:RPS11A(gene name)
# Alias(es) interactor A: uniprotkb:YOR239W/YOR240W
# Alias(es) interactor B: uniprotkb:RPS18A|uniprotkb:YDR025W|uniprotkb:YD9813.03|uniprotkb:RPS11B|uniprotkb:RPS18B|uniprotkb:YBR048W|uniprotkb:YBR0501
# Interaction detection method(s): MI:0676(tap)
# Publication 1st author(s): -
# Publication Identifier(s): pubmed:16429126
# Taxid interactor A: taxid:4932(yeast)
# Taxid interactor B: taxid:4932(yeast)
# Interaction type(s): MI:0218(physical interaction)
# Source database(s): MI:0469(intact)
# Interaction identifier(s): intact:EBI-785909
# Confidence value(s): -

if verbose:
    sys.stderr.write("Reading data from intact file...\n")


for line in input_file_fd:

    line_number += 1

    parser.add_to_log("Number of processed lines")

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if line_number%20000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))


    line_fields = line.split('\t')  # line_fields[0] is ID Interactor A|cross_references
                                    # line_fields[1] is ID interactor B|cross_references
                                    # line_fields[2] is Alt. ID interactor A
                                    # line_fields[3] is Alt. ID interactor B
                                    # line_fields[4] is Alias(es) interactor A
                                    # line_fields[5] is Alias(es) interactor B
                                    # line_fields[6] is Interaction detection method(s)
                                    # line_fields[7] is Publication 1st author(s)
                                    # line_fields[8] is Publication Identifier(s)
                                    # line_fields[9] is Taxid interactor A
                                    # line_fields[10] is Taxid interactor B
                                    # line_fields[11] is Interaction type(s)
                                    # line_fields[12] is Source database(s) (in version March 2007 all is MI:0469(intact))
                                    # line_fields[13] is Interaction identifier(s)
                                    # line_fields[14] is confidence value(s) (in version March 2007 there is not any value)

    # First step: obtain list of proteinPianas for interactors
    names_intA = line_fields[0].split('|')
    names_intB = line_fields[1].split('|')

    list_proteinPianas_intA = []
    list_proteinPianas_intB = []
    list_intact_codes_intA = []
    list_intact_codes_intB = []

    list_detection_methods = []
    list_pubmed_ids = []


    # Get the taxID for the interactors # Is it necessary? To check... I don't know if a refseq or a uniprot can have more than one species...
    taxID_re = re.search("taxid:(\d+)",line_fields[9])
    if taxID_re:
        taxID_intA = taxID_re.group(1)
    else:
        taxID_intA = 0

    taxID_re = re.search("taxid:(\d+)",line_fields[10])
    if taxID_re:
        taxID_intB = taxID_re.group(1)
    else:
        taxID_intB = 0

    # For interactor A:
    for interactorA in names_intA:
        interactor = re.search("(\w+):([_\d\w]+)",interactorA)
        if interactor is not None:
            if interactor.group(1) == "refseq":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="refseq",
                                                                                    tax_id_value=taxID_intA, source_db_info="no") )
            elif interactor.group(1) == "uniprotkb":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="uniacc",
                                                                                    tax_id_value=taxID_intA, source_db_info="no") )
            elif interactor.group(1) == "intact":
                list_intact_codes_intA.append(interactor.group(2))

            elif interactor.group(1) == "genbank_protein_gi":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                   id_type="gi",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )

            elif interactor.group(1) == "genbank":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2).strip().split(".")[0],
                                                                                   id_type="ncbiacc",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )

            elif interactor.group(1) == "ensembl":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                   id_type="ensembl",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )
            else:
                parser.add_to_log("Cross-reference %s not identified" %(interactor.group(1)))
                #sys.stderr.write("%s\n%s: Cross-reference %s not known\n" %(line,interactorA,interactor.group(1)) )

    # For interactor B:
    for interactorB in names_intB:
        interactor = re.search("(\w+):([\_\d\w\-\.]+)",interactorB)
        if interactor is not None:
            if interactor.group(1) == "refseq":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="refseq",
                                                                                    tax_id_value=taxID_intB, source_db_info="no") )
            elif interactor.group(1) == "uniprotkb":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type= "uniacc",
                                                                                    tax_id_value=taxID_intB, source_db_info="no") )

            elif interactor.group(1) == "genbank_protein_gi":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                   id_type="gi",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )
            elif interactor.group(1) == "intact":
                list_intact_codes_intB.append(interactor.group(2))

            
            elif interactor.group(1) == "genbank":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2).strip().split(".")[0],
                                                                                   id_type="ncbiacc",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )

            elif interactor.group(1) == "ensembl":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                   id_type="ensembl",
                                                                                   tax_id_value=taxID_intA, source_db_info="no") )
            else:
                parser.add_to_log("Cross-reference %s not identified" %(interactor.group(1)))
                #sys.stderr.write("%s\n%s: Cross-reference %s not known\n" %(line,interactorB,interactor.group(1)) )


    # Once obtained the proteinPianas, is it the same with all cross-references? If all is correct, it should be yes... In case not... introduce more cross-references???
    # For the moment, we are NOT Inserting new cross_references found here

    # So, the next step is to eliminate the repetitions in proteinPianas:
    if list_proteinPianas_intA:
        uniq_list_proteinPianas_intA = dict(map(None,list_proteinPianas_intA,[None])).keys()
    else:
        uniq_list_proteinPianas_intA = []

    if list_proteinPianas_intB:
        uniq_list_proteinPianas_intB = dict(map(None,list_proteinPianas_intB,[None])).keys()
    else:
        uniq_list_proteinPianas_intB = []


    ## Now, we have the uniq list of proteinPianas for interactorA, for interactorB, and we have the INTACT ID for interactors
    ## The next step is to introduce it into database

    # Insert into database correspondences between intact code and proteinPianas:
    # TODO INSERT THE INTACT CODE IN externalProteinId table...	  
    for proteinPiana in uniq_list_proteinPianas_intA:
        for intact_id in list_intact_codes_intA:
            piana_access.insert_external_id_code(external_id_type="intact",
                                                 external_id_value=intact_id,
                                                 proteinPiana_value = proteinPiana,
                                                 sourceDBID = sourceDB)
            parser.add_to_log("Number of external identifiers added")
            
    for proteinPiana in uniq_list_proteinPianas_intB:
        for intact_id in list_intact_codes_intB:
            piana_access.insert_external_id_code(external_id_type="intact",
                                                 external_id_value=intact_id,
                                                 proteinPiana_value=proteinPiana,
                                                 sourceDBID=sourceDB)
            parser.add_to_log("Number of external identifiers added")


    # Second step: obtain detection method

    # There are some cases where there is more than one detection method
    raw_list_detection_method = line_fields[6].split('|')

    for actual_method in raw_list_detection_method:
        #detection_method = re.search("MI:(\d+)\(([\w\s\_\-\.]+)\)",actual_method)
        detection_method = re.search("MI:(\d+)\((.+)\)\s*$",actual_method)
        if detection_method is not None:
            list_detection_methods.append(detection_method.group(2))
            if verbose:
                sys.stderr.write("Detection method: %s\n" %(detection_method.group(2)) )
        else:
            if verbose:
                sys.stderr.write("Detection method not found?\n")

    # Third step: obtain pubmed id for this interaction
    raw_list_pubmed_ids = line_fields[8].split('|')

    for actual_pubmed_id in raw_list_pubmed_ids:
        pubmed_id = re.search("pubmed:(\d+)",actual_pubmed_id)
        if pubmed_id is not None:
            list_pubmed_ids.append(pubmed_id.group(1))
            if verbose:
                sys.stderr.write("Pubmed ID: %s\n" %pubmed_id.group(1) )
        else:
            if verbose:
                sys.stderr.write("Pubmed ID not found?\n")

    if not list_detection_methods:
        #sys.stderr.write("%s\n" %line)
        #sys.stderr.write("How is it possible not to have detection method?\n")
        parser.add_to_log("Number of lines without detection method")

    if not list_pubmed_ids:
        #sys.stderr.write("%s\n" %line)
        #sys.stderr.write("How is it possible not to have pubmed reference?\n")
        parser.add_to_log("Number of lines without pubmed reference")
                
    
    # Insert into database the interactions:
    for proteinPianaA in uniq_list_proteinPianas_intA:
        for proteinPianaB in uniq_list_proteinPianas_intB:
            # We don't have to worry is proteinPianaA is smaller than proteinPianaB: piana_access will do it for us
            piana_access.insert_interaction(proteinPianaA_value = proteinPianaA,
                                            isSourceA_value = 1,
                                            proteinPianaB_value = proteinPianaB,
                                            isSourceB_value = 1,
                                            interactionConfidence_value = 1,
                                            methodDescription_value = list_detection_methods,
                                            sourceDBDescription_value = sourceDB,
                                            confidenceAssignedSourceDB_value=1,
                                            pubmed_id_value= list_pubmed_ids)
            parser.add_to_log("Number of interactions inserted")



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
   

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

if verbose:
    sys.stderr.write("%s\t%s\n" %(line_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
    sys.stderr.write(parser.get_log_string())
