"""
File        : mint2piana.py
Author      : Ramon Aragues
Creation    : 11.03.07
Contents    : inserts information from a Mint tab file database into PIANA
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from a mint database

--> Mint database must be in txt tabulated format

--> download it from http://mint.bio.uniroma2.it/mint/download.do (in tabulated format)

"""

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline
import MySQLdb


from PianaDBaccess import *
from parserUtilities import *

import time

verbose = 0
verbose_detailed = 0
verbose_methods = 0   # set this to 1 to get a list of MINT methods (ints won't be inserted into database)

## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("input-file-name=",None,"file name of input file containing mint txt data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
		           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="mint2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from a mint database",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

arguments_dic = parser.parseArguments()

input_file = arguments_dic["input-file-name"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
time_control = arguments_dic["time-control"]
log_file = arguments_dic["log-file"]
database_information = arguments_dic["database-information"]



# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.sequence_table,
                  PianaGlobals.proteinPianaCounter_table,
		  PianaGlobals.temporalData_table]

# --------
# --------
#  Main()               
# --------                               
# --------

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
					   databaseInformation = database_information)

# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )


# Lock_number is used to calculate the frequency to do unlock...
line_number=0

if verbose_methods:
    methods_dic = {}
# Format in March 2007:

# ID interactor A	ID interactor B	Alt. ID interactor A	Alt. ID interactor B	Alias(es) interactor A	Alias(es) interactor B	Interaction detection method(s)	Publication 1st author(s)	Publication Identifier(s)	Taxid interactor A	Taxid interactor B	Interaction type(s)	Source database(s)	Interaction identifier(s)	Confidence value(s)	bioSource_taxid:celltype:tissue	figures	url	negative	experiment_role_1	biological_role_1	expression_level_1	protein_identification_1	sample_process_1	experimental_preparation_1	binding_sites_1	modifications_1	mutations_1	experiment_role_2	biological_role_2	expression_level_2	protein_identification_2	sample_process_2	experimental_preparation_2	binding_sites_2	modifications_2	mutations_2	mint protein group 1 (taxon)	mint protein group 2 (taxon)	MINT confidence score


# [0] ID interactor A
# [1] ID interactor B
# [2] Alt. ID interactor A
# [3] Alt. ID interactor B
# [4] Alias(es) interactor A
# [5] Alias(es) interactor B
# [6] Interaction detection method(s)
# [7] Publication 1st author(s)
# [8] Publication Identifier(s)
# [9] Taxid interactor A
# [10] Taxid interactor B
# [11] Interaction type(s)
# [12] Source database(s)
# [13] Interaction identifier(s)
# [14] Confidence value(s) not used by PIANA
# [15] bioSource_taxid:celltype:tissue not used by PIANA
# [16] figures not used by PIANA
# [17] url not used by PIANA
# [18] negative not used by PIANA
# [19] experiment_role_1 not used by PIANA
# [20] biological_role_1 not used by PIANA
# [21] expression_level_1 not used by PIANA
# [22] protein_identification_1 not used by PIANA
# ...
# [] not used by PIANA
# [] not used by PIANA


# Example:
# [0] ID interactor A: flybase:Q08641
# [1] ID interactor B: uniprotkb:P26781
# [2] Alt. ID interactor A: uniprotkb:ABP140(gene name)
# [3] Alt. ID Interactor B: refseq:RPS11A(gene name)
# [4] Alias(es) interactor A: uniprotkb:YOR239W/YOR240W
# [5] Alias(es) interactor B: uniprotkb:RPS18A|uniprotkb:YDR025W|uniprotkb:YD9813.03|uniprotkb:RPS11B|uniprotkb:RPS18B|uniprotkb:YBR048W|uniprotkb:YBR0501
# [6] Interaction detection method(s): MI:0676(tap)
# [7] Publication 1st author(s): -
# [8] Publication Identifier(s): pubmed:16429126
# [9] Taxid interactor A: taxid:4932(yeast)
# [10] Taxid interactor B: taxid:4932(yeast)
# [11] Interaction type(s): MI:0218(physical interaction)
# [12] Source database(s): MI:0469(mint)
# [13] Interaction identifier(s): mint:EBI-785909
# [14] Confidence value(s): -
# [15] ....
# [16] ...

if verbose:
    sys.stderr.write("Reading data from mint file...\n")

start_parsing = 0

for line in input_file_fd:

    if start_parsing == 0:
        if line.startswith("ID interactor A"):
            start_parsing = 1
        continue
        
    

    line_number += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if line_number%20000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))


    line_fields = line.split('\t')  # [0] ID interactor A
                                    # [1] ID interactor B
                                    # [2] Alt. ID interactor A
                                    # [3] Alt. ID interactor B
                                    # [4] Alias(es) interactor A
                                    # [5] Alias(es) interactor B
                                    # [6] Interaction detection method(s)
                                    # [7] Publication 1st author(s)
                                    # [8] Publication Identifier(s)
                                    # [9] Taxid interactor A
                                    # [10] Taxid interactor B
                                    # [11] Interaction type(s)
                                    # [12] Source database(s)
                                    # [13] Interaction identifier(s)
                                    # [14] Confidence value(s)
                                    # [15] bioSource_taxid:celltype:tissue
                                    # [16] figures
                                    # [17] url
                                    # [18] negative
                                    # [19] experiment_role_1
                                    # [20] biological_role_1
                                    # [21] expression_level_1
                                    # [22] protein_identification_1
                                    # ...


    # First step: obtain list of proteinPianas for interactors
    names_intA = line_fields[0].split('|')
    names_intB = line_fields[1].split('|')

    list_proteinPianas_intA = []
    list_proteinPianas_intB = []

    list_detection_methods = []
    list_pubmed_ids = []


    # Get the taxID for the interactors        # Is it necessary? To check... I don't know if a refseq or a uniprot can have more than one species...
    taxID_re = re.search("taxid:(\d+)",line_fields[9])
    if taxID_re:
        taxID_intA = taxID_re.group(1)
    else:
        taxID_intA = 0

    taxID_re = re.search("taxid:(\d+)",line_fields[10])
    if taxID_re:
        taxID_intB = taxID_re.group(1)
    else:
        taxID_intB = 0

    # For interactor A:
    for interactorA in names_intA:
        interactor = re.search("(\w+):([_\d\w]+)",interactorA)
        if interactor is not None:
            if interactor.group(1) == "refseq":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="refseq",
                                                                                    tax_id_value=taxID_intA, source_db_info="no") )
            elif interactor.group(1) == "uniprotkb":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="uniacc",
                                                                                    tax_id_value=taxID_intA, source_db_info="no") )
            elif interactor.group(1) == "flybase":
                list_proteinPianas_intA.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="flybase",
                                                                                    tax_id_value=taxID_intA, source_db_info="no") )
                
            else:
                sys.stderr.write("Cross-reference %s not known\n" %(interactor.group(1)) )

    # For interactor B:
    for interactorB in names_intB:
        interactor = re.search("(\w+):([\_\d\w\-\.]+)",interactorB)
        if interactor is not None:
            if interactor.group(1) == "refseq":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type="refseq",
                                                                                    tax_id_value=taxID_intB, source_db_info="no") )
            elif interactor.group(1) == "uniprotkb":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                    id_type= "uniacc",
                                                                                    tax_id_value=taxID_intB, source_db_info="no") )
            elif interactor.group(1) == "flybase":
                list_proteinPianas_intB.extend( piana_access.get_list_proteinPiana(protein_id=interactor.group(2),
                                                                                   id_type="flybase",
                                                                                   tax_id_value=taxID_intB, source_db_info="no") )

            else:
                sys.stderr.write("Cross-reference %s not known\n" %(interactor.group(1)) )


    # Once obtained the proteinPianas, is it the same with all cross-references? If all is correct, it should be yes... In case not... introduce more cross-references???
    # For the moment, we are NOT Inserting new cross_references found here

    # So, the next step is to eliminate the repetitions in proteinPianas:
    if list_proteinPianas_intA:
        uniq_list_proteinPianas_intA = dict(map(None,list_proteinPianas_intA,[None])).keys()
    else:
        uniq_list_proteinPianas_intA = []

    if list_proteinPianas_intB:
        uniq_list_proteinPianas_intB = dict(map(None,list_proteinPianas_intB,[None])).keys()
    else:
        uniq_list_proteinPianas_intB = []


    ## Now, we have the uniq list of proteinPianas for interactorA, for interactorB
        
    # Second step: obtain detection method

    # There are some cases where there is more than one detection method
    raw_list_detection_method = line_fields[6].split('|')

    for actual_method in raw_list_detection_method:
        #detection_method = re.search("MI:(\d+)\(([\w\s\_\-\.]+)\)",actual_method)
        detection_method = re.search("[MI|MINT]:(\d+)\((.+)\)\s*$",actual_method)
        if detection_method is not None:
            list_detection_methods.append(detection_method.group(2))
            if verbose:
                sys.stderr.write("Detection method: %s\n" %(detection_method.group(2)) )
        else:
            if verbose:
                sys.stderr.write("Detection method not found?\n")

    # Third step: obtain pubmed id for this interaction
    raw_list_pubmed_ids = line_fields[8].split('|')

    for actual_pubmed_id in raw_list_pubmed_ids:
        pubmed_id = re.search("pubmed:(\d+)",actual_pubmed_id)
        if pubmed_id is not None:
            list_pubmed_ids.append(pubmed_id.group(1))
            if verbose:
                sys.stderr.write("Pubmed ID: %s\n" %pubmed_id.group(1) )
        else:
            if verbose:
                sys.stderr.write("Pubmed ID not found?\n")

    if not list_detection_methods:
        if verbose_detailed:
            sys.stderr.write("%s\n" %line)
            sys.stderr.write("No detection method given to this interaction (the one in the previous line)\n")

    if not list_pubmed_ids:
        if verbose_detailed:
            sys.stderr.write("%s\n" %line)
            sys.stderr.write("No pubmed fiven to this interaction (the one in the previous line)\n")


    if verbose_methods:
        methods_dic.update(dict(map(None, list_detection_methods, [])))
        continue
    
    # Insert into database the interactions:
    for proteinPianaA in uniq_list_proteinPianas_intA:
        for proteinPianaB in uniq_list_proteinPianas_intB:
            # We don't have to worry is proteinPianaA is smaller than proteinPianaB: piana_access will do it for us
            if verbose_detailed:
                sys.stderr.write("interaction between %s (%s) and %s (%s), with methods %s, sourceDB %s and pubmeds %s\n" %(proteinPianaA, names_intA,
                                                                                                                          proteinPianaB, names_intB,
                                                                                                                          list_detection_methods,
                                                                                                                          sourceDB,
                                                                                                                          list_pubmed_ids ))

            piana_access.insert_interaction(proteinPianaA_value = proteinPianaA,
                                            isSourceA_value = 1,
                                            proteinPianaB_value = proteinPianaB,
                                            isSourceB_value = 1,
                                            interactionConfidence_value = 1,
                                            methodDescription_value = list_detection_methods,
                                            sourceDBDescription_value = sourceDB,
                                            confidenceAssignedSourceDB_value=1,
                                            pubmed_id_value= list_pubmed_ids)



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
   

piana_access.close()

if verbose_methods:
    for one_method in methods_dic:
        print one_method

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

if verbose:
    sys.stderr.write("%s\t%s\n" %(line_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
