"""
File        : gene2accession2piana.py
Author      : Javier Garcia and Ramon Aragues
Creation    : February 2007
Contents    : script that fills up tables in database piana related to  external DB "Gene"
Called from : 

=======================================================================================================

This file parsers gene2accession NCBI file and inserts information into PianaDb

Command line option '--help' describes usage of this program

For more details on how to use it, read piana/README.populate_piana_db

"""


#=======================================================================================================
# This file parsers a Gene NCBI text file and inserts information into PianaDb
# Command line option '--help' describes usage of this program
# For more details on how to use it, read piana/README.populate_piana_db
#
# gene2accession2piana.py: script that fills up tables in database piana related to external DB "Gene" from NCBI
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline

import MySQLdb

import time

from PianaDBaccess import *
from parserUtilities import *


# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing genpept data (relxxx_fsa_aa)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("tax-id-file=",None,"a text file with two tab-separated columns: 1st is gi code and 2nd is taxonomy id"),
                             ("database-name=","ncbi_gene","internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]



# Define the parser information to print in the help:

parser = ParserObject( script_name="gene2accession_parser.py",
                       script_description="This program fills up tables in database piana related to external DB 'Gene'",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dict = parser.parseArguments()

gene2accession_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
tax_id_file = arguments_dict["tax-id-file"]
source_db = arguments_dict["database-name"]
database_version = arguments_dict["database-version"]
database_description = arguments_dict["database-description"]
piana_dbuser = arguments_dict["piana-dbuser"]
piana_dbpass = arguments_dict["piana-dbpass"]
help = arguments_dict["help"]
verbose = arguments_dict["verbose"]
time_control = arguments_dict["time-control"]
log_file= arguments_dict["log-file"]
database_information = arguments_dict["database-information"]

parser.parseArguments()




# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.gi_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.geneID_table,
                  PianaGlobals.temporalData_table]

# --------
# --------
#  Main()               
# --------                               
# --------


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = source_db,
                                          databaseVersion = database_version,
                                          parsedFile = gene2accession_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

gene_fd = open(gene2accession_file,"r")

#
# Reading external DB "Gene" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from gene2accession file...\n")

line_counter=0


# This dictionary will contain all associations between geneID and proteinPiana inserted (because there are some entries repeated in gene2accession file)
# For example: different geneID entries with status provisional... (Maybe it would be better not taking into account this entries...)
# It is used in order to not duplicate inserts
inserted_geneID_dict = {}

for line in gene_fd:

    line_counter += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if not line_counter%20000:
            sys.stderr.write("%s lines in %s seconds\n" %(line_counter,time.time()-initial_time))

    if line[0] != '#':

        # Parsing the line using the regular expression followed by taxonomy data
        line_fields = line.split("\t")

        # line_fields is a vector with
        #                              (tax_id, GeneID, status,
        #                               RNA_nucleotide_accession.version, RNA_nucleotide_gi,
        #                               protein_accession.version, protein_gi,
        #                               genomic_nucleotide_accession.version, genomic_nucleotide_gi,
        #                               start_position_on_the_genomic_accession, end_position_on_the_genomic_accession,
        #                               orientation assembly)
        #      line_fields[0] = tax_id
        #      line_fields[1] = GeneID
        #      line_fields[2] = status
        #      line_fields[3] = RNA_nucleotide_accession.version
        #      line_fields[4] = RNA_nucleotide_gi
        #      line_fields[5] = protein_accession.version
        #      line_fields[6] = protein_gi
        #      line_fields[7] = genomic_nucleotide_accession.version
        #      line_fields[8] = genomic_nucleotide_gi
        #      line_fields[9] = start_position_on_the_genomic_accession
        #      line_fields[10]= end_position_on_the_genomic_accession
        #      line_fields[11]= orientation assembly

        
        if line_fields:

            # I don't know if use gi or proteinAccession to obtain proteinPiana... Is it the same? In principle yes... But different gi codes can have the same
            # proteinAccession (as a proteinAccession can have different versions, and each version has a different gi code)
            # To check these correspondences... I think different gi's can have the same proteinPiana, and an accession number can have different proteinPianas
            # If this is true... what is better to use?

            # Should I have to save all accessionNumbers? Or only protein_accession? Uniprot give us genomic_nucleotide_accession...
            # I can use this file to check and save all correspondences between these codes
            
            if( line_fields[5] == "-" and line_fields[6] == "-" ):   # Normally, both are codes or both are "-" (but not always... in March 2007, in 2 cases not)
                parser.add_to_log("Not inserted geneID (because no gi nor protein accessionNumber found)")
            else:
                tax_id = int( line_fields[0].strip() )
                geneID = int( line_fields[1].strip() )
                if line_fields[5] != "-":
                    proteinAccession = line_fields[5].strip().split('.')[0]
                else:
                    proteinAccession = None
                if line_fields[6] != "-":
                    proteinGI = int( line_fields[6].strip() )
                else:
                    proteinGI = None

                if geneID != "":

                    proteinPiana_list = []

                    proteinPiana_gi_list=[]
                    proteinPiana_accession_list=[]

                    if proteinGI:
                                           
                        # Get the proteinPiana list that correspond to this gi code
                        proteinPiana_gi_list = piana_access.get_list_proteinPiana(protein_id=proteinGI,
                                                                                   id_type="gi",
                                                                                   tax_id_value=0, source_db_info="no")

                    if proteinAccession:

                        #Get the proteinPiana list that correspond to this proteinAccession Code
                        proteinPiana_accession_list = piana_access.get_list_proteinPiana(protein_id=proteinAccession,
                                                                                         id_type="ncbiacc",
                                                                                         tax_id_value=0, source_db_info="no" )

                    ## We compare the two lists (theoretically, there would be the same proteinPiana codes... if not, add cross-references between them in database
                    if proteinPiana_gi_list:
                        proteinPiana_gi_dict = dict(map(None,proteinPiana_gi_list,[]))
                    else:
                        proteinPiana_gi_dict = {}

                    for proteinPiana_acc in proteinPiana_accession_list:
                        if not proteinPiana_gi_dict.has_key(proteinPiana_acc):
                            # This accessionNumber for this gi code was not found previously, so, we add it into database
                            piana_access.insert_gi_code(gi_code_value=proteinGI, proteinPiana_value=proteinPiana_acc, gi_source_value=source_db)
                            parser.add_to_log("New gi references added")

                    if proteinPiana_accession_list:
                        proteinPiana_acc_dict = dict(map(None,proteinPiana_accession_list,[]))
                    else:
                        proteinPiana_acc_dict = {}

                    for proteinPiana_gi in proteinPiana_gi_list:
                        if not proteinPiana_acc_dict.has_key(proteinPiana_gi):
                            # This gi for this accession Number code was not found previously, so, we add it into database
                            piana_access.insert_accessionNumber(accessionNumber_value=proteinAccession, proteinPiana_value=proteinPiana_gi,
                                                                accessionNumber_source_value=source_db)
                            parser.add_to_log("New protein accessionNumber references added")

                    # Now, add the GeneID code for all proteinPiana found

                    # First, unify list of proteinPianas:
                    proteinPiana_accession_list.extend(proteinPiana_gi_list)
                    if proteinPiana_accession_list:
                        proteinPiana_list = dict(map(None,proteinPiana_accession_list,[])).keys()
                    else:
                        proteinPiana_list = []

                    # If no protein Piana is found... should we search by mRNA gi or genomic gi? For the moment, we don't do this...
                    if len(proteinPiana_list)==0:
                        if verbose:
                            sys.stderr.write( " No proteinPiana was found for gi: %s nor protein accession: %s!  \n" %(proteinGI,proteinAccession) )
                        parser.add_to_log("Not inserted geneID (because no gi nor protein accessionNumber found)")

                    else:
                        parser.add_to_log("Inserted geneID")
                        for actual_proteinPiana in proteinPiana_list:
                            # Check if this relations has been previously inserted
                            if not inserted_geneID_dict.has_key((geneID,actual_proteinPiana)):
                                piana_access.insert_geneID_code( geneID_code_value=geneID, proteinPiana_value=actual_proteinPiana, geneID_source_value=source_db )
                                inserted_geneID_dict[(geneID,actual_proteinPiana)] = None
                                if verbose:
                                    sys.stderr.write(" geneID-proteinPiana previously inserted\n ")
                            if verbose:
                                sys.stderr.write(" Inserting code geneID: %s!!!\n" %(geneID) )
                else:
                    parser.add_to_log("Not inserted geneID (because no gi nor protein accessionNumber found)")


                # END OF if geneID != ""...
        
                if verbose:
                    sys.stderr.write( "geneID  is <%s> -- tax_id is <%s> -- proteinAccession is <%s> -- proteinGI is <%s>  \n" %(geneID, tax_id, proteinAccession, proteinGI))
  
            # END OF if line_fiels[5]...
        # END OF if line_fields:
    # END OF line[0]!='#'
# END OF for line in gene_fd:
                    
                              

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )

if verbose:
    sys.stdour.write("%s\n" %(parser.get_log_string()) )

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
