"""
File        : swissprot2piana.py
Author      : Ramon Aragues
Creation    : 29.11.2004
Contents    : fills up tables in database piana with information from ncbi cross-references
              between swissprot and gi
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from ncbi swissprot (gi with sprots)

ncbi swissprot can be downloaded from ftp://ftp.ncbi.nih.gov/blast/db/FASTA/swissprot.gz

Before running, taxonomy table of piana must be populated (use taxonomy2piana for that)

It is advisable to parse swissprot and trembl and genpept and nr before parsing swissprot

"""

# swissprot2piana.py: fills up tables in database piana with information from ncbi swissprot
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES


import sys
import getopt

import re
import readline
import MySQLdb

import utilities

from PianaDBaccess import *
from parserUtilities import *

import time

from Bio import Fasta  # needed to read the swissprot file (which is in fasta format)




# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values
# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#

compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing genpept data (relxxx_fsa_aa)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("tax-id-file=",None,"a text file with two tab-separated columns: 1st is gi code and 2nd is taxonomy id"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]
                           
# Define the parser information to print in the help:

parser = ParserObject( script_name="swissprot2piana.py",
                       script_description="This program fills up tables in database piana with information from swissprot",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dic = parser.parseArguments()

input_file = arguments_dic["input-file-name"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
tax_id_file = arguments_dic["tax-id-file"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
time_control = arguments_dic["time-control"]
log_file = arguments_dic["log-file"]
database_information= arguments_dic["database-information"]

verbose_detailed = 0


# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')

protein_number=0


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.protein_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.sequence_table,
                  PianaGlobals.species_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.temporalData_table]
# --------
# --------
#  Main()               
# --------                               
# --------

if verbose_detailed:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)

# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )


swissprot_parser = Fasta.RecordParser()
swissprot_input_file =  utilities.open_file(input_file, 'r')
swissprot_iterator = Fasta.Iterator(swissprot_input_file, swissprot_parser)

swissprot_record = swissprot_iterator.next()


if verbose:
    print "Processing file"

# get dictionary for gis and tax ids
dic_gi_tax = utilities.return_dic_gi_vs_tax(file_name=tax_id_file)

# while record read is not None, parse the record and insert data into piana
while swissprot_record is not None:

    protein_number += 1
    parser.add_to_log("Number of proteins in the input file")

    swissaccession = None
    swissprot = None
    gi_id = None

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    
    if time_control:
        if protein_number%20000==0:
            sys.stderr.write("%s proteins done in %s seconds\n" %(protein_number,time.time()-initial_time))

    if verbose_detailed:
        sys.stderr.write( "===================================\n")
        sys.stderr.write( "            NEW ENTRY\n")
        sys.stderr.write( "===================================\n")

    protein_title_line = swissprot_record.title
    protein_sequence = swissprot_record.sequence

    if verbose_detailed:
        sys.stderr.write("title line is : %s\n" %(protein_title_line)) 
        sys.stderr.write("sequence is : %s\n" %(protein_sequence))


    # title can look like this: 
    #                    gi|51315741|sp||P62809_1 [Segment 1 of 3] Cyclic phosphodiesterase (CPDase)
    #                    gi|50400705|sp|Q9ZES3|APAG_RHIET Protein apaG
    #                    gi|50401406|sp||P84036_1 [Segment 1 of 2] Plasminogen-activating proteinase (LV-PA)
    #                    gi|50401529|sp|Q7WQ22|TRMB_BORBR tRNA (guanine-N(7)-)-methyltransferase (tRNA(m7G46)-methyltransferase)$gi|50401527|sp
#|Q7WC23|TRMB_BORPA tRNA (guanine-N(7)-)-methyltransferase (tRNA(m7G46)-methyltransferase)

    #                           
    #    (character '>' from input file has been removed by biopython)
    #    (where character $ is actually octal character 001)
    #
    # in misc ncbi swissprot (as opposed to genpept) a title can have several gi identifiers for the same sequence


    # So, given the messy format of this file, first of all split by octal character 001 to see if there is more than one gi
    title_entries = protein_title_line.split("\001")

    # for each of the title entries (normally just one) split it, retrieve information and insert codes, using same sequence for all)
    for title_entrie in title_entries:

        
        title_atoms = title_entrie.split('|')

        # title_atom[1] (should) is always the gi code
        gi_id = int(title_atoms[1])

        # title_atom[3] is the swiss accession
        # if swiss accession doesn't appear on position 3, then position 4 is a swiss accession with segment identifier
        # if swiss accession appears on position 3, then position 4 is a swissprot id
        #

        if title_atoms[3] == "" or title_atoms[3] is None:
            # no swiss accession given, get it from position 4 (removing the segment part)
            swissaccession = title_atoms[4].split()[0][0:6]
            swissprot      = None
        else:
            swissaccession = title_atoms[3]
            swissprot      = title_atoms[4].split()[0]


        if dic_gi_tax.has_key(gi_id):
            # we give priority to the dictionary with tax ids for gis
            tax_id_value = dic_gi_tax[gi_id]
        else:
            # get tax_id for this gi (tax id was previously inserted by genpept or nr)
            # 1. get proteinPianas associated to the gi (using 0 for tax id will get all proteinPianas regardless of the species)
            #                                            a gi is normally associated to a single species... it should work ok)
            pps_for_this_gi = piana_access.get_list_proteinPiana(protein_id=gi_id,
                                                                 id_type= "gi",
                                                                 tax_id_value = 0,
                                                                 source_db_info="no")
            # 2. get tax_ids associated to those proteinPianas
            dic_tax_id = {}
            for this_proteinPiana in pps_for_this_gi:
                list_tax_id = piana_access.get_protein_taxonomy_ids(proteinPiana_value=this_proteinPiana )

                for this_tax_id in list_tax_id:
                    if this_tax_id != 0:
                        dic_tax_id[this_tax_id] = None

            # END OF for this_proteinPiana in pps_for_this_gi:

            # 3. check if we can use this tax id (ie. there is only one, and it is a valid one)
            if len(dic_tax_id)== 1:
                tax_id_value = dic_tax_id.keys()[0]
            else:
                # if nothing found, use dummy value for this sequence
                tax_id_value = 0
        # END OF else: (if dic_gi_tax.has_key(gi_id):)

        if tax_id_value == 0:
            # for the time being, do not allow ncbi swissprot to insert new sequences is the tax id is unknown...
            #  (moreover, when the tax id is known, the (sequence, tax_id) will probably be already there...)
            # I prefer to avoid introducing noise...
            continue
        
        #----
        # first of all, get a proteinPiana for the (sequence, tax_id) that we are processing
        #----


        ## New (sequence,tax_id) found will not be inserted!!! We only use this file to take database cross-references
        ## So, if no proteinPiana is found, cross-references will not be introduced! (Javi)
        
        if protein_sequence != "":
            #inserted_proteinPiana = piana_access.insert_protein(proteinSequence_value= protein_sequence,
            #                                                    tax_id_value= tax_id_value)
            inserted_proteinPiana = piana_access.check_protein(proteinSequence_value=protein_sequence,
                                                               tax_id_value=tax_id_value)
        else:
            raise ValueError("Trying to insert an empty sequence from ncbi_swissprot into Piana")


        if verbose:
            sys.stderr.write("Inserting gi <%s>, uniacc <%s> and unientry <%s> for proteinPiana <%s>\n" %(gi_id,
                                                                                                          swissaccession,
                                                                                                          swissprot,
                                                                                                          inserted_proteinPiana))



        
        """
        Now, we have a proteinPiana for the sequence, only the one already existing before (if it did not exist, it won't be introduced!
        We also have the protein swissprot info

        insert this info to database
        """
        if inserted_proteinPiana is None:
            parser.add_to_log("Not inserted proteins")
        else:
            
            # --- table gi ---

            # inserting gi ID  for current protein
            if gi_id != "" and gi_id is not None:
                piana_access.insert_gi_code(gi_code_value= gi_id, proteinPiana_value= inserted_proteinPiana,
                                            gi_source_value = sourceDB)
            


            # --- table uniprotAcc ---

            # inserting uniprotAccession  for current protein
            if swissaccession != "" and swissaccession is not None:
                # Length control error
                if len(swissaccession)>6:
                    sys.stderr.write("ERROR: uniprotAccession not correct? %s (length superior to 6)\n" %(swissaccession) )
                else:
                    piana_access.insert_uniprotAcc(swissaccession,inserted_proteinPiana,sourceDB, isPrimary_value= 0)

            # --- table uniprotEntry ---

            # inserting  uniprot Entry for current protein
            if swissprot != "" and swissprot is not None:
                piana_access.insert_uniprotEntry(swissprot,inserted_proteinPiana,sourceDB)

          
            # --- table species ---

            # tax_id_value comes from the section where we checked if there was a species name in the title line
            if verbose_detailed:
                sys.stderr.write( "adding protein_tax_id %s (species_name=%s) for proteinPiana %s (gi=%s)\n" %(tax_id_value,
                                                                                                               species_name,
                                                                                                               inserted_proteinPiana,
                                                                                                               gi_id))
            # --- table proteinSpecies ---
            #piana_access.insert_protein_species(tax_id =  tax_id_value,
            #                                    proteinPiana_value = inserted_proteinPiana,
            #                                    proteinSpeciesSource_value = sourceDB)

            # --- table proteinSourceDB ---
            #piana_access.insert_protein_source(proteinPiana_value = inserted_proteinPiana,
            #                                   protein_source_value = sourceDB)
            

    # END OF for title_entrie in title_entries

    
    # reading next record
    if verbose_detailed:
        print "reading next record"
        
    swissprot_record = swissprot_iterator.next()
    
# END OF while swissprot_record is not None



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if verbose:
    sys.stderr.write(parser.get_log_string())


if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()


if time_control:
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )
