"""
File        : nr2piana.py
Author      : Ramon Aragues
Modified    : Javier Garcia February 2007
Creation    : 10.2004
Contents    : fills up tables in database piana with information from nr
Called from : 

=======================================================================================================

This file implements a program that fills up tables in database piana with information from nr

nr can be downloaded from ftp://ftp.ncbi.nih.gov/blast/db/FASTA/nr.gz

Before running, taxonomy table of piana must be populated (use taxonomy2piana for that)

It is advisable to parse swissprot and trembl and genpept before parsing nr

"""

# nr2piana.py:  fills up tables in database piana with information from nr
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues



## STEP 1: IMPORT NECESSARY MODULES


import sys
import getopt

import re
import readline
import MySQLdb

import utilities

import time

from PianaDBaccess import *
from parserUtilities import *

from Bio import Fasta  # needed to read the nr file (which is in fasta format)



# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing nr data (nr)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("tax-id-file=",None,"a text file with two tab-separated columns: 1st is gi code and 2nd is taxonomy id"),
                             ("database-name=","ncbi_nr","internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\""),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)")]

# Define the parser information to print in the help:

parser = ParserObject( script_name="nr2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from nr)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
tax_id_file= arguments_dict["tax-id-file"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]


verbose_all = 0

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.protein_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.sequence_table,
                  PianaGlobals.species_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.pdb_table,
                  PianaGlobals.proteinDescription_table,
                  PianaGlobals.temporalData_table]


# Start time counter
if time_control:
    initial_time = time.time()


# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')
   
            
# --------
# --------
#  Main()               
# --------                               
# --------

if verbose:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)

# Initialize protein number (to the output printing...)
protein_number=0         # Counter of total of protein lines in file
total_entries=0          # Counter of total of sequence-taxid entries in file
total_inconsistencies=0  # counter of inconsistencies found (inconsistencies with identifiers cross-references)

piana_access.lock_tables(table_list = tables_to_lock )

nr_parser = Fasta.RecordParser()
nr_input_file =  utilities.open_file(input_file, 'r')
nr_iterator = Fasta.Iterator(nr_input_file, nr_parser)

nr_record = nr_iterator.next()


if verbose:
    print "Processing file"

# get dictionary for gis and tax ids
if verbose:
    print "Reading gi/taxonomy information..."
dic_gi_tax = utilities.return_dic_gi_vs_tax(file_name=tax_id_file)



# Dictionary to save inconsistencies (when a proteinPiana does not exists but exists any external Identifier...)
inconsistencies = {}

# while record read is not None, parse the record and insert data into piana
while nr_record is not None:

    protein_number += 1
    parser.add_to_log("Number of proteins in input file")

    if verbose:
        sys.stderr.write( "===================================\n")
        sys.stderr.write( "            NEW ENTRY\n")
        sys.stderr.write( "===================================\n")


    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if protein_number%20000==0:
            sys.stderr.write("%s proteins\t%s seq-taxid entries\tin %s seconds\n" %(protein_number,total_entries,time.time()-initial_time))


    protein_title_line = nr_record.title
    protein_sequence = nr_record.sequence.strip()

    if verbose_all:
        sys.stderr.write("title line is : %s\n" %(protein_title_line)) 
        sys.stderr.write("sequence is : %s\n" %(protein_sequence))

    """
    Now, we have a proteinPiana that will be asigned to all other codes that we find in the title line
    """
    # title can look like this: gi|1346670|sp|P04175|NCPR_PIG NADPH--cytochrome P450 reductase (CPR) (P450R)
    #                           gi|5832586|dbj|BAA84019.1| maturase [Eucharis grandiflora]
    #                           gi|223154|prf||0601198A polymerase beta,RNA
    #                           gi|17538019|ref|NP_496216.1| surfeit 5 (18.1 kD) (2K591) [Caenorhabditis elegans]$gi|2497033|sp|Q23679|YWM3_CAEEL Hypothetical protein ZK970.3 in chromosome II$gi|3881897|emb|CAA88887.1| Hypothetical protein ZK970.3 [Caenorhabditis elegans]$gi|7511441|pir||T28127 hypothetical protein ZK970.3 - Caenorhabditis elegans
    #                           
    #    (character '>' from input file has been removed)
    #    (where character $ is actually octal character 001)
    #
    # in nr (as opposed to genpept) a title can have several gi identifiers for the same sequence
    # however, nr is a mess, so it is easier to get species and so on from genpept (that is why we parse it first)


    # So, given the messy format of nr, first of all split by octal character 001 to see if there is more than one gi
    title_entries = protein_title_line.split("\001")

    # for each of the title entries (normally just one) split it, retrieve information and insert codes, using same sequence for all)
    for title_entrie in title_entries:

        total_entries += 1

        gi = None

        if verbose:
            sys.stderr.write("One entry of the title is: %s\n" %(title_entrie))
        
        title_atoms = title_entrie.split('|')

        # title_atom[1] (should) is always the gi code
        #if( title_atoms[0] == "gi" ):
        #    gi_id = int(title_atoms[1])

        #
        #  title_atom[2] can be (exhaustive list as of 10.2004):
        #
        #  - gb: for EMBL accessions
        #  - emb: for EMBL protein id
        #  - pir: for PIR accessions
        #  - sp: for swissprot
        #  - dbj: dna japan
        #  - prf: ???
        #  - ref: ref_seq
        #  - pdb: pdb code
        #  - tpg: ???
        #  - tpe: ???

        ### UPDATED: February 2007:

        #  - gb: for GenBank acession Number
        #  - emb: for EMBL accession Number
        #  - dbj: for DDBJ, DNA Database of Japan
        #  - pir: for NBRF PIR
        #  - prf: for Protein Research Foundation
        #  - sp:  for Swiss-Prot Entries
        #  - pdb: for Brookhaven Protein Data Bank
        #  - pat: for Patents
        #  - bbs: for GenInfo Backbone Id
        #  - gnl: for General database Identifier
        #  - ref: for NCBI Reference Sequence
        #  - lcl: for Local Sequence Identifier

        # title_atom[3] is the value (a protein id) indicated in title_atom[2]

        # In theory, only one has to exist... (because of the EMBL/GenBank/DDBJ nucleotide sequence database...
        #gb_accession = None
        #embl_accession = None
        #dbj_accession = None

        accessionNumber = None
        
        uniprot_accession = None  
        uniprot_entry = None
        ref_seq = None
        pdb_code = None
        pdb_chain = None

        description = None

        species_name = None

        # Read all the interesting cross-references:
        for i in xrange(len(title_atoms)):
            if title_atoms[i] == "gi":
                if gi is None:
                    gi = int(title_atoms[i+1])
                else:
                    sys.stderr.write(" Error parsing... how an entry can have several gi values? \n")
            elif title_atoms[i] == "gb" or title_atoms[i] == "emb" or title_atoms[i] == "dbj":
                if accessionNumber is None:
                    accession = re.search("(\w+)\.\d+",title_atoms[i+1])
                    if accession:
                        accessionNumber = accession.group(1)
                else:
                    sys.stderr.write(" Error parsing... how an entry can have several accesionNumber.version values? \n")
            elif title_atoms[i] == "sp":
                if uniprot_accession is None:
                    uniprot_accession = title_atoms[i+1][0:6]
                else:
                    sys.stderr.write(" There are more than one uniprot accession for the same gi... \n")
                if uniprot_entry is None:
                    # "Chapuza" because there is one entry that has no uniprotEntry name...and the parser does not work correctly
                    # This entry is: gi|20148613|gb|AAM10197.1| similar to high affinity sulfate transporter 2|sp|P53392 [Arabidopsis thaliana]
                    if len(title_atoms) > (i+2):
                        uniprot_entry = (title_atoms[i+2].split())[0].strip()
                else:
                    sys.stderr.write(" There are more than one uniprot entry for the same gi... \n")
            elif title_atoms[i] == "ref":
                if ref_seq is None:
                    ref_seq = title_atoms[i+1].strip().split(".")[0]
                else:
                    sys.stderr.write(" There are more than one uniprot RefSeq for the same gi... \n")
            elif title_atoms[i] == "pdb":
                if pdb_code is None:
                    pdb_code = title_atoms[i+1].strip()
                    pdb_chain = title_atoms[i+2][0].strip()
                else:
                    sys.stderr.write(" There are more than one PDB for the same gi... \n")

        # Take the description
        has_description = re.search("\|([^\|]*)\[",title_entrie)
        if has_description:
            description = has_description.group(1)
            #description = has_description.group(1).replace('"','').replace("(","").strip()

        # Take the specie:
        has_specie = re.search("\[(.*)\]$",title_entrie)
        if has_specie:
            species_name = has_specie.group(1).replace('"','').replace("(","").strip()


        # Now, get the tax_ID associated with the specie
        if dic_gi_tax.has_key(gi):
            # we give priority to the dictionary with tax ids for gis
            list_tax_ids = [dic_gi_tax[gi]]
        elif species_name:
            # if no value found in dic, get the tax_id associated to that species name
            list_tax_ids= piana_access.get_taxonomies_from_species_name(species_name_value = species_name)
        else:
            list_tax_ids=[]


        if verbose:
            sys.stderr.write( "ENTRY:\n\tgi: %s \n\tAccessionNumber: %s\n\tuniprot_acc: %s\n\tuniprot_entry: %s\n\trefseq: %s\n\tpdb code: %s\n\tpdb chain: %s\n\t\n\tDescription: %s\n\tspecie: %s\n\n"
                              %(gi, accessionNumber, uniprot_accession, uniprot_entry, ref_seq, pdb_code, pdb_chain, description, species_name) )

        # If any specie has been found, assign it to taxID=0
        if len(list_tax_ids)==0:
            list_tax_ids = [0]
        if len(list_tax_ids)>1:
            sys.stderr.write("More than one tax_id for this entry: %s!\n" %(title_entrie) )
        for tax_id_iter in list_tax_ids:
            proteinPiana = piana_access.insert_protein(proteinSequence_value=protein_sequence,sourceDBID=sourceDB,tax_id_value=tax_id_iter)
                    
            # --- table gi ---
            if gi:
                piana_access.insert_gi_code(gi,proteinPiana,sourceDB)
                
            # --- table accessionNumber ---
            if accessionNumber:
                piana_access.insert_accessionNumber(accessionNumber,proteinPiana,sourceDB)

            # --- table uniprotAcc ---
            if uniprot_accession:
                # I'm not sure if the uniprot_Accession that is here is primary or not... (to check) (Javi)
                piana_access.insert_uniprotAcc(uniprot_accession,proteinPiana,sourceDB, isPrimary_value=0)

            # --- table uniprotEntry ---
            if uniprot_entry:
                piana_access.insert_uniprotEntry(uniprot_entry,proteinPiana,sourceDB)

            # --- table refseq ---
            if ref_seq:
                piana_access.insert_refseq_code(ref_seq,proteinPiana,sourceDB)

            # --- table pdb ---
            if pdb_code:
                piana_access.insert_pdb_code(pdb_code,proteinPiana,pdb_chain,None,sourceDB)

            # --- table description ---
            if description:
                if description  != "":
                    piana_access.insert_protein_description(
                        description = description.replace('"'," ").replace("\n", " ").replace("'"," ").replace('\\', " ").strip(),
                        proteinPiana_value = proteinPiana,
                        proteinDescriptionSource_value = sourceDB )
                    

            # Check if the protein exists in the database:
            #proteinPiana = piana_access.check_protein(proteinSequence_value=protein_sequence, tax_id_value=tax_id_iter)
                
            #if not proteinPiana:
            #    # The proteinPiana does not exist in the database... we have to introduce it
            #    # But before it, we are going to check previously for inconsistencies with external references... (there are some inconsistencies)

            #    # Search if this gi code exists in the database
            #    gi_protein_pianas = piana_access.get_list_proteinPiana(protein_id=gi, id_type="gi")

            #    if gi_protein_pianas:
            #        inconsistencies.setdefault('gi',[]).append(gi)

            #    # Search if this accessionNumber code exists in the database (if exists accessionNumber)
            #    accessionNumber_protein_pianas=[]
            #    if accessionNumber:
            #        accessionNumber_protein_pianas = piana_access.get_list_proteinPiana(protein_id=accessionNumber,
            #                                                                            id_type="ncbiacc")
            #        if accessionNumber_protein_pianas:
            #            inconsistencies.setdefault('accNum',[]).append(accessionNumber)
                    
                            
            #    # Search if the uniprotAccession exists in the database
            #    uniprotAcc_protein_pianas=[]
            #    if uniprot_accession:
            #        uniprotAcc_protein_pianas = piana_access.get_list_proteinPiana(protein_id=uniprot_accession, id_type="uniacc")
            #        if uniprotAcc_protein_pianas:
            #            inconsistencies.setdefault('uniAcc',[]).append(uniprot_accession)


            #    # search if the uniprotEntry exists in the database
            #    uniprotEntry_protein_pianas=[]
            #    if uniprot_entry:
            #        uniprotEntry_protein_pianas = piana_access.get_list_proteinPiana(protein_id=uniprot_entry, id_type="unientry")
            #        if uniprotEntry_protein_pianas:
            #            inconsistencies.setdefault('uniEntry',[]).append(uniprot_entry)

            #    # If there is any reference that previously existed in PIANA database, do not insert this protein, otherwise yes
            #    # In the case the new proteinPiana is inserted, insert the description too
            #    if gi_protein_pianas or accessionNumber_protein_pianas or uniprotAcc_protein_pianas or uniprotEntry_protein_pianas:
            #        total_inconsistencies += 1
            #        parser.add_to_log("Total of inconsistencies")
            #    else:
            #        proteinPiana = piana_access.insert_protein(proteinSequence_value=protein_sequence,sourceDBID=sourceDB,tax_id_value=tax_id_iter)

            #    # --- table proteinDescription ---
            #    if description:
            #        if description  != "":
            #            piana_access.insert_protein_description(
            #                description = description.replace('"'," ").replace("\n", " ").replace("'"," ").replace('\\', " ").strip(),
            #                proteinPiana_value = proteinPiana,
            #                proteinDescriptionSource_value = sourceDB )

            
##             # Now, if proteinPiana previously existed or has been introduced, we have to introduce external references
##         #if proteinPiana:
##         # Update the proteinSource for this proteinPiana:
##                     proteinPiana = piana_access.insert_protein(proteinSequence_value=protein_sequence,sourceDBID=sourceDB,tax_id_value=tax_id_iter)
                    
##                     # --- table gi ---
##                     if gi:
##                         piana_access.insert_gi_code(gi,proteinPiana,sourceDB)

##                     # --- table accessionNumber ---
##                     if accessionNumber:
##                         piana_access.insert_accessionNumber(accessionNumber,proteinPiana,sourceDB)

##                     # --- table uniprotAcc ---
##                     if uniprot_accession:
##                         # I'm not sure if the uniprot_Accession that is here is primary or not... (to check) (Javi)
##                         piana_access.insert_uniprotAcc(uniprot_accession,proteinPiana,sourceDB, isPrimary_value=0)

##                     # --- table uniprotEntry ---
##                     if uniprot_entry:
##                         piana_access.insert_uniprotEntry(uniprot_entry,proteinPiana,sourceDB)

##                     # --- table refseq ---
##                     if ref_seq:
##                         piana_access.insert_refseq_code(ref_seq,proteinPiana,sourceDB)

##                     # --- table pdb ---
##                     if pdb_code:
##                         piana_access.insert_pdb_code(pdb_code,proteinPiana,pdb_chain,None,sourceDB)

##                     # --- table species ---
##                     #piana_access.insert_protein_species(tax_id = tax_id_iter,
##                     #                                    proteinPiana_value = proteinPiana,
##                     #                                    proteinSpeciesSource_value = sourceDB )

    # reading next record
    if verbose:
        sys.stderr.write( "\n---------------reading next record---------------\n")
    nr_record = nr_iterator.next()
    
# END OF while nr_record is not None

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
     sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(protein_number, (time.time()-initial_time) ))

if verbose:
    sys.stderr.write(" Inconsistencies: %s\n" %(inconsistencies) )
    sys.stderr.write(" Number of entries with inconsistencies (not introduced to PIANA) %s\n" %(total_inconsistencies) )
    sys.stderr.write("%s\t%s\n" %(protein_number, (time.time()-initial_time) ))
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.write("\n\n List of code inconsistencies: \n%s\n" %(inconsistencies))
    log_file_fd.close()

