"""
File        : ori2piana.py
Author      : Ramon Aragues
Creation    : 25.02.2004
Contents    : script that fills up tables in database piana related to external DB "ori"
Called from : build_database.py

=======================================================================================================

This program parsers a text file with interactions (ori format) and inserts them into piana

The input file to this parser can be found at piana/data/externalDBs/oriDB/interact.dat

format of the input file is:

Entry 1:
|Heteromer|1cpc_AB|3.7|DIP=0|STRING=0|Homol=0|
5H2C_HUMAN (P28335) 5-hydroxytryptamine 2C receptor (     6.7    1.2e+03 ! ?
FBN1_HUMAN (P35555) Fibrillin 1 precursor.                3.5      1e+04 ! ?
Entry 2:
|Heteromer|1cpc_AB|3.7|DIP=0|STRING=0|Homol=1|
PHA1_FREDI (P07122) C-phycocyanin-1 alpha chain.          4.4    5.7e+03   X
PHB1_FREDI (P07119) C-phycocyanin-1 beta chain.          12.6         20   X
Entry 3:
|Heteromer|1cpc_AB|3.7|DIP=0|STRING=0|Homol=1|
PHA1_FREDI (P07122) C-phycocyanin-1 alpha chain.          4.4    5.7e+03   X
PHB3_FREDI (P14877) C-phycocyanin-3 beta chain.          12.6         20   X
E
........................

Command line option '--help' describes usage of this program

"""

# ori2piana.py: script that fills up tables in database piana related to external DB "ori"
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys
import getopt

import re
import readline

import MySQLdb

from parserUtilities import *

from PianaDBaccess import *

verbose = 0

# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
# These will be the values taken by the program when called directly from build_database.py

# Location of file containing external DB "ori"
ori_file = "../../../data/externalDBs/oriDB/interact.dat"

compulsory_arguments_list = [ ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]

parser = ParserObject( script_name="ori2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with interactions predicted from distant seq/str patterns\n",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

arguments_dic = parser.parseArguments()

piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
log_file = arguments_dic["log-file"]
time_control  = arguments_dic["time-control"]
database_information = arguments_dic["database-information"]

# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()
    
# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.proteinPianaCounter_table]

# --------
# --------
#  Main()               
# --------                               
# --------

if log_file:
    log_file_fd = file(log_file, 'w')


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

try:
    # Introduce database info into PIANA database
    piana_access.insert_new_external_database(databaseName = sourceDB,
                                              databaseVersion = database_version,
                                              parsedFile = ori_file.split("/")[-1],
                                              databaseDescription = database_description,
					      databaseInformation = database_information)
except:
    pass # ignore error because we accept to have several files with the same database label


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

# opening file descriptor to input file
ori_fd = open(ori_file,"r")


# 
# Reading external DB "ori" and inserting its data into piana DB
# 

if verbose:
    print "Reading data from ori file..."

# initialize variables
new_entry_count = 0
swissacc_a = None
swissacc_b = None
proteinPiana_a = None
proteinPiana_b = None

number_of_interactions_null_swissacc = 0
number_of_interactions_no_id = 0

# Lock_number is used to calculate the frequency to do unlock...
line_number=0

for line in ori_fd:

    line_number += 1

    # since there is no separator character, use single space to detect the swissaccession
    # this means that we need to change double spaces into single spaces, so split() always
    # partitions the line in the same way
    line = re.sub( "[ ]+", " ", line)
    
    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)
    
    if time_control:
        if line_number%20000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))

    # Retrieving fields from the line (separated by space)
    line_fields = line.split(" ")

    if line_fields[0] == "Entry":
        # new entry reached:
        #     1 - insert information about previous entry
        #     2 - initialize variables

        # 1 - insert information about previous entry
        if swissacc_a is not None and swissacc_b is not None:

            # TO CHECK!!! Why can swissacc be None? I don't remember the reason for this test...
            if verbose:
                print "Interaction found between %s and %s" %(swissacc_a, swissacc_b)

            list_proteinPiana_a = piana_access.get_list_proteinPiana(protein_id=swissacc_a,
                                                                     id_type="uniacc",
                                                                     tax_id_value= 0,
                                                                     only_primary="no",
                                                                     source_db_info= "no")
            
            list_proteinPiana_b = piana_access.get_list_proteinPiana(protein_id=swissacc_b,
                                                                     id_type="uniacc",
                                                                     tax_id_value= 0,
                                                                     only_primary="no",
                                                                     source_db_info= "no")
            


            
            for proteinPiana_a in list_proteinPiana_a:
                for proteinPiana_b in list_proteinPiana_b:
                    
                    if verbose:
                        sys.stderr.write( "   - inserting interaction: %s --> %s\n" %(proteinPiana_a, proteinPiana_b))


                    
                        
                    piana_access.insert_interaction(proteinPianaA_value = proteinPiana_a,
                                                    isSourceA_value = 1,
                                                    proteinPianaB_value = proteinPiana_b,
                                                    isSourceB_value = 1,
                                                    interactionConfidence_value = 1,
                                                    methodDescription_value = "dist_patt",
                                                    sourceDBDescription_value = sourceDB,
                                                    confidenceAssignedSourceDB_value=1)

                # END OF for proteinPiana_b in dic_proteinPiana_b:
            # END OF for proteinPiana_a in dic_proteinPiana_a:
            if verbose:
 
                if not list_proteinPiana_a or not list_proteinPiana_b:
                    sys.stderr.write("proteinPiana not found for %s or %s \n" %(swissacc_a, swissacc_b))
                    number_of_interactions_no_id += 1

            
        # END OF if swissacc_a is not None and swissacc_b is not None:
        else:
            number_of_interactions_null_swissacc += 1
        # END OF else: (if swissacc_a is not None and swissacc_b is not None:)

        # 2 - initialize variables for next entry
        new_entry_count = 1
        swissacc_a = None
        swissacc_b = None
        proteinPiana_a = None
        proteinPiana_b = None
     
    elif new_entry_count == 1:
        # skipping second line of the entry (not relevant to parse)
        new_entry_count +=1
        
    elif new_entry_count == 2:
        # first protein of the interaction being parsed
        swissacc_a = line_fields[1]
        swissacc_a = swissacc_a.strip("(").strip(")")
        new_entry_count +=1
        
    elif new_entry_count == 3:
        # second protein of the interaction being parsed
        swissacc_b = line_fields[1]
        swissacc_b = swissacc_b.strip("(").strip(")")
        new_entry_count +=1

    # END OF if line_fields[0] == "Entry": elif new_entry_count == 1: .....
   

if verbose:
    sys.stderr.write( "All done! There were %s interactions without swissaccession and %s without a proteinPiana\n\n" %(
        number_of_interactions_null_swissacc,
        number_of_interactions_no_id))


piana_access.close()
