"""
File        : pdbsprotec2piana.py
Author      : Ramon Aragues
Creation    : 10.2004
Contents    : fills up tables in database piana with information from pdbsprotec
Called from : 

=======================================================================================================

This file implements a program that fills up tables in database piana with information from pdbsprotec, which contains
a mapping between pdb codes, swissprot and ec codes

pdbsprotec can be downloaded from http://www.bioinf.org.uk/pdbsprotec/

Before running, uniprot (swissprot and trembl) has to be in pianaDB

Files that can be parsed using pdbsprotec2piana.py look like this:


 pdbcode | chainid | res1 | res2 | sprot  |     ec
---------+---------+------+------+--------+------------
 101m    |         | 1    | 153  | P02185 | 0.0.0.0
 101m    |         | 1    | 153  | P02185 | 0.0.0.0
 102l    |         | 1    | 162  | P00720 | 3.2.1.17
 102m    |         | 1    | 153  | P02185 | 0.0.0.0 
 104l    | A       | 1    | 162  | P00720 | 3.2.1.17
 104l    | B       | 1    | 162  | P00720 | 3.2.1.17
 104m    |         | 1    | 153  | P02185 | 0.0.0.0

 

"""

# pdbsprotec2piana.py: fills up tables in database piana with information from pdbsprotec
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import MySQLdb

import time

from PianaDBaccess import *
from parserUtilities import *



# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing pdbsprotec data"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="pdbsprotec2piana.py",
                       script_description="This program fills up tables in database piana with information from pdbsprotec",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]




# Start time counter
if time_control:
    initial_time = time.time()

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.uniprotAcc_table,
                  PianaGlobals.pdb_table,
                  PianaGlobals.proteinEC_table,
                  PianaGlobals.temporalData_table]

if verbose:
    print "Arguments read are: input-file= %s || piana_dbname= %s || piana_dbhost= %s" %(input_file, piana_dbname, piana_dbhost)

# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


#Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)



pdbsprotec_input_file_fd =  file(input_file, 'r')

number_of_lines = 0

# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

if verbose:
    sys.stderr.write( "Processing file\n")

# while there are lines, insert info into piana
for line in pdbsprotec_input_file_fd:

    number_of_lines += 1
    parser.add_to_log("Number of lines")


    if time_control:
        if (number_of_lines%20000)==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(number_of_lines,time.time()-initial_time))
    
    if number_of_lines == 1 or number_of_lines == 2:
        # skip two first header lines
        continue


    line_fields = line.split("|")  # line_fields[0] is pdbcode
                                   # line_fields[1] is chainid 
                                   # line_fields[2] is res1   (starting residue number)
                                   # line_fields[3] is res2   (ending residue number)
                                   # line_fields[4] is sprot
                                   # line_fields[5] is ec 

    if len(line_fields) != 6:
        continue
    
    pdbcode = line_fields[0].strip()
    chainid = line_fields[1].strip()
    res1    = line_fields[2].strip()
    res2    = line_fields[3].strip()
    sprot   = line_fields[4].strip()
    ec      = line_fields[5].strip()

    if verbose:
        sys.stderr.write("Entry processed and fields obtained are:\n")
        sys.stderr.write("   -pdbcode : %s\n" %(pdbcode))
        sys.stderr.write("   -chainid : %s\n" %(chainid))
        sys.stderr.write("   -sprot : %s\n" %(sprot))
        sys.stderr.write("   -ec : %s\n" %(ec))
        

    # check if the pdbsprotec protein sequence already exists in piana or not
    list_sprot_proteinPiana = piana_access.get_list_proteinPiana(protein_id=sprot,
                                                                 id_type= "uniacc",
                                                                 tax_id_value = 0)


    for sprot_proteinPiana in list_sprot_proteinPiana:
        # for each proteinPiana, insert a new pdb entry and a new ec code

        if pdbcode != "":

            if verbose:
                sys.stderr.write("Inserting correspondence between proteinPiana <%s> and pdcode <%s>, chain <%s>\n" %(sprot_proteinPiana,
                                                                                                                      pdbcode,
                                                                                                                      chainid))
            parser.add_to_log("number_of_pds_inserted")

            if res1!= "" and res2!="":
                range = "%s-%s" %(res1,res2)
            else:
                range = None
                
            piana_access.insert_pdb_code(pdb_code_value= pdbcode , proteinPiana_value= sprot_proteinPiana, chain_value= chainid,
                                         range_value = range,pdb_source_value= sourceDB )

        if ec != "0.0.0.0":
            # don't insert if ec doen't carry content

            if verbose:
                sys.stderr.write("Inserting correspondence between proteinPiana <%s> and ec <%s>\n" %(sprot_proteinPiana,
                                                                                                      ec))
                
            piana_access.insert_protein_ec(ec_id= ec, proteinPiana_value= sprot_proteinPiana, proteinECSource_value= sourceDB)

    # END OF for sprot_proteinPiana in list_sprot_proteinPiana:

    if not list_sprot_proteinPiana:
        parser.add_to_log("Number of sprots without a proteinPiana")
        if verbose:
            sys.stderr.write ("No proteinPiana found for sprot %s\n" %(sprot) )
    
    
# END OF for line in pdbsprotec_input_file_fd:


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if verbose:
    sys.stderr.write(parser.get_log_string())

if time_control:
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )


if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
