"""
File        : piana_text_int2piana.py
Author      : Ramon Aragues
Creation    : 12.11.2004
Contents    : script that fills up tables in database piana from a piana text file describing interactions
Called from :  build_database.py

=======================================================================================================

This program parsers a piana text file with interactions and inserts them into pianaDB

The format for files accepted as input files for this parser is described in README.piana_interaction_data_format


Command line option '--help' describes usage of this program

"""

# piana_text_int2piana.py: script that fills up tables in database piana from a piana text file describing interactions
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline

import MySQLdb

from PianaDBaccess import *
from parserUtilities import *

verbose = 0

## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

compulsory_arguments_list = [ ("input-file=",None,"file name of input file containing interactions in PIANA txt format"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted"),
                              ("input-id-type=",None,"Identifier type of proteins in your input file"),
                              ("tax-id=",None,"""taxonomy id for the proteins in your file\n
                                               --> valid values are 0 (unknown or not important) and NCBI taxon ids\n
                                               --> if your file uses gene names to identify proteins, setting the tax id is required (gene names are highly ambiguous)""")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("no-species",0,"inserts all interactions, regardless of the species of proteins"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]


# Define the parser information to print in the help:


parser = ParserObject( script_name="piana_text_int2piana.py",
                       script_description="""This file implements a program that fills up tables in database piana with interactions from a file in PIANA text format\n
                       --> Read README.piana_interaction_data_format to learn about the PIANA text format for interactions\n/
                       --> Attention! The method name you are giving to the predictions must exist in PianaGlobals.method_names (e.g. y2h)""",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


arguments_dic = parser.parseArguments()

input_file = arguments_dic["input-file"]
piana_dbname = arguments_dic["piana-dbname"]
piana_dbhost = arguments_dic["piana-dbhost"]
sourceDB = arguments_dic["database-name"]
database_version = arguments_dic["database-version"]
database_description = arguments_dic["database-description"]
input_id_type = arguments_dic["input-id-type"]
tax_id = int(arguments_dic["tax-id"])
piana_dbuser = arguments_dic["piana-dbuser"]
piana_dbpass = arguments_dic["piana-dbpass"]
no_species = arguments_dic["no-species"]
help = arguments_dic["help"]
verbose = arguments_dic["verbose"]
log_file = arguments_dic["log-file"]
time_control  = arguments_dic["time-control"]
database_information = arguments_dic["database-information"]




# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

if time_control:
    initial_time = time.time()

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.interaction_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.proteinPianaCounter_table]

# --------
# --------
#  Main()               
# --------                               
# --------

if log_file:
    log_file_fd = file(log_file, 'w')



# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

try:
    # Introduce database info into PIANA database
    piana_access.insert_new_external_database(databaseName = sourceDB,
                                              databaseVersion = database_version,
                                              parsedFile = input_file.split("/")[-1],
                                              databaseDescription = database_description,
					      databaseInformation = database_information)
except:
    pass # ignore error because we accept to have several files with the same database label


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

number_of_interactions_added = 0
number_of_interactions_no_id = 0

input_file_fd = file(input_file,"r")

#
# Reading "piana_text_int" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write("Reading data from input file %s...\n" %input_file)


number_of_interactions_added = 0
number_of_interactions_dif_species = 0
number_of_interactions_no_id = 0

# Lock_number is used to calculate the frequency to do unlock...
line_number=0

for line in input_file_fd:

    line_number += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if line_number%20000==0:
            sys.stderr.write("%s lines done in %s seconds\n" %(line_number,time.time()-initial_time))


    # Retrieving different fields in line
    line_fields = line.split()
    
    if len(line_fields) == 4:
        
        # line_fields.groups() is a tuple with (tax_id, tax_name)
        protein_a = line_fields[0].strip()
        protein_b = line_fields[1].strip()
        source_db = line_fields[2].strip()
        method = line_fields[3].strip()
        #confidence = int(line_fields[4].strip())
        confidence = 1
        

        if verbose:
            sys.stderr.write( "interaction read is %s --> %s with method %s\n" %(protein_a, protein_b, method ))

        
        list_proteinPiana_a = piana_access.get_list_proteinPiana(protein_id= protein_a,
                                                                   id_type=input_id_type ,
                                                                   tax_id_value= tax_id, source_db_info= "no")
        
        list_proteinPiana_b = piana_access.get_list_proteinPiana(protein_id= protein_b,
                                                                  id_type=input_id_type ,
                                                                  tax_id_value= tax_id, source_db_info= "no")
       


        for proteinPiana_a in list_proteinPiana_a:
            for proteinPiana_b in list_proteinPiana_b:
            
                list_species_a = piana_access.get_protein_taxonomy_ids(proteinPiana_value = proteinPiana_a)
                list_species_b = piana_access.get_protein_taxonomy_ids(proteinPiana_value = proteinPiana_b)
                
                # check if proteins are of the same species. If not, interaction not inserted (except command line option says other)
                same_species = 0
                
                if list_species_a and list_species_b:
                    for species_a in list_species_a:
                        if species_a in list_species_b:
                            same_species = 1
                            break
                else:
                    sys.stderr.write( "Warning! No taxonomy id found for proteinPiana %s (%s) or proteinPiana %s (%s)\n" %(proteinPiana_a,
                                                                                                                           list_species_a,
                                                                                                                           proteinPiana_b,
                                                                                                                           list_species_b))
                    # when no taxonomy found, we assume that is a real interaction by seting same_species to 1
                    same_species = 1
                # END OF else: (if list_species_a and list_species_b:)
                    
                if same_species or no_species:
                    if verbose:
                        sys.stderr.write( "inserting interaction: %s --> %s with method %s\n" %(proteinPiana_a, proteinPiana_b ,
                                                                                                piana_access.get_methodID(method) ))

                        
                    piana_access.insert_interaction(proteinPianaA_value = proteinPiana_a,
                                                    isSourceA_value = 1,
                                                    proteinPianaB_value = proteinPiana_b,
                                                    isSourceB_value = 1,
                                                    interactionConfidence_value = confidence,
                                                    methodDescription_value = method,
                                                    sourceDBDescription_value = source_db,
                                                    confidenceAssignedSourceDB_value=1)
                    number_of_interactions_added += 1
                # END OF if same_species or no_species:

                else:
                    number_of_interactions_dif_species +=1
                    if verbose:
                        sys.stderr.write( "proteinA %s (species = %s) and proteinB %s (species =%s) are of different species\n" %(proteinPiana_a,
                                                                                                                                  list_species_a,
                                                                                                                                  proteinPiana_b,
                                                                                                                                  list_species_b) )
                # END OF else: (if same_species or no_species:)

            # END OF for proteinPiana_b in list_proteinPiana_b:
        # END OF for proteinPiana_a in list_proteinPiana_a:
 
        if verbose:
            if not list_proteinPiana_a or not list_proteinPiana_b:
                
                sys.stderr.write("proteinPiana not found for %s (%s) or %s (%s) \n" %(protein_a, list_proteinPiana_a,
                                                                                      protein_b, list_proteinPiana_b))
                number_of_interactions_no_id += 1
        # END OF if verbose:
    
    # END OF if len(line_fields) == 5:
    
    else:
        if verbose:
            sys.stderr.write("Line was not well formatted! Check format used for file! README.piana_interaction_data_format\n")
            sys.stderr.write("(this error might be due to empty lines in your input file: don't worry if it is the case, the data is being inserted anyway)\n")

# END OF for line in input_file_fd


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
   

piana_access.close()
   
if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(line_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
  
   
if verbose:
   sys.stderr.write( "\n------------------------------------------------------------------------\n")
   sys.stderr.write( "All done! Number of ints added: %s.\n" %(number_of_interactions_added))
   sys.stderr.write( "Number of interactions not inserted: between different species: %s\n" %(number_of_interactions_dif_species))
   sys.stderr.write( "Number of interactions with proteins unknown to PIANA: %s\n" %(number_of_interactions_no_id))
   sys.stderr.write( "------------------------------------------------------------------------\n")
