"""
File        : psi2piana.py
Author      : Ramon Aragues
Creation    : 24.10.2005
Modification: Javier Garcia
Contents    : inserts information from a PSI mysql database into a piana database
Called from : command line

=======================================================================================================

This file implements a program that fills up tables in database piana with information from a psi database

--> a PSI database must exist already with format shown in create_psi_tables.sql (directory code/dbCreation)

    this existing PSI database was previously populated using parsePSI.py
"""

# psi2piana.py: inserts information from a PSI mysql database into a piana database
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import MySQLdb

from PianaDBaccess import *
from parserUtilities import *
import PianaGlobals


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("psi-dbname=",None,"name of the psi mysql database"),
                              ("psi-dbhost=",None,"name of the machine with psi mysql server"),
                              ("psi-db=",None,"""the psi database you are parsing (unfortunately, the standards are not that respected...)
                                                   - mips: when parsing the MIPS Mammalian Protein-Protein Database
                                                   - bind: when parsing the Biomolecular Interaction Network Database"""),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("psi-dbuser=",None,"name of the mysql psi username"),
                           ("psi-dbpass=",None,"password of the mysql psi user"),
                           ("insert-psidb-ids",0,""),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]

# Define the parser information to print in the help:

parser = ParserObject( script_name="uniprot2piana.py",
                       script_description="This program fills up tables in database piana with information from swissprot or TrEMBL (no differenttreatment required)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)


arguments_dict = parser.parseArguments()

psi_dbname = arguments_dict["psi-dbname"]
psi_dbhost = arguments_dict["psi-dbhost"]
psi_dbuser = arguments_dict["psi-dbuser"]
psi_dbpass = arguments_dict["psi-dbpass"]
psi_db = arguments_dict["psi-db"]
insert_psidb_ids = arguments_dict["insert-psidb-ids"]

piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]


# Start time counter
if time_control:
    initial_time = time.time()


if log_file:
    log_file_fd = file(log_file, 'w')







# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
# Table names are defined in piana/code/PianaDB/PianaGlobals.py
# If you don't known which tables are being used, the list can be empty: in this case, all tables will be locked
# While tables are locked, anyone else can access to them

tables_to_lock = []

verbose_detailed = 0

counting_discarded = 0


# Control to check if the database has repeated information...
inserted_interactions = {}
inserted_interactionMethods = {}
inserted_interactionPubmed = {}

   
def get_associated_list_proteinPianas(psi_cursor, piana_access, node_psi_id, parser):
    """
    from a psi node of database connected via psi_cursor, return a list of (proteinPiana, sourceDB) associated to it

    """
    # I have removed an argument ext_codes_wanted that was not being used in this file...

    # -----------------------------------------------------------------------
    #  1. find the external identifiers for node_psi_id
    # -----------------------------------------------------------------------
    #     after the queries, we'll have:
    #
    #         list_ext_codes  --> list of tuples (uniacc, emblacc, gi id)
    #
    #
    #     the database has string "None" when there is no value for an external code

    sqlquery = """select spAcc_id, emblAcc_id, gi_id, description, sequence from proteinCodes where counter_id = %s""" %(node_psi_id)
    psicursor.execute(sqlquery)
    list_ext_codes = psicursor.fetchall()


    sqlquery = """select tax_id from proteinFeatures where counter_id = %s""" %(node_psi_id)
    psicursor.execute(sqlquery)
    value_retrieved = psicursor.fetchall()
    if value_retrieved:
        tax_id = int(value_retrieved[0][0])
    else:
        tax_id = 0
        parser.add_to_log("Number of taxID not found in psi database")


    # instead of dealing directly with the sql result, create a dictionary like this:
    #                             { 'spAcc_id': [P23423, q23423, ...],
    #                               'emblAcc_id': [...]
    #                               'gi_id': [2342,234234, ...]
    #                               'description': [sdfsdfsdfd,sdfsdfdsf, ...]
    #                               'sequence': [sdfsdfsdfsdfsdf,sdfsdfsdfsdfsdf]
    #                             }
    # this dictionary will be then used to search for the proteinPianas (by giving priority to certain codes over the others)

    dic_all_codes = { 'spAcc_id': [],
                      'emblAcc_id': [],
                      'gi_id': [],
                      'description': [],
                      'sequence': []
                      }
    
    for group_codes in list_ext_codes:
        protein_swissacc=   group_codes[0]
        protein_emblacc =   group_codes[1]
        protein_gi =   group_codes[2]
        protein_description =   group_codes[3]
        protein_sequence =   group_codes[4]

        if protein_swissacc and protein_swissacc != "NULL" and protein_swissacc != "None":
            dic_all_codes['spAcc_id'].append(protein_swissacc)
            
        if protein_emblacc and protein_emblacc != "NULL":
            dic_all_codes['emblAcc_id'].append(protein_emblacc)
            
        if protein_gi and protein_gi != "NULL":
            dic_all_codes['gi_id'].append(protein_gi)
            
        if protein_description and protein_description != "NULL":
            dic_all_codes['description'].append(protein_description)
            
        if protein_sequence and protein_sequence != "NULL":
            dic_all_codes['sequence'].append(protein_sequence)
            
    # END OF for group_codes in list_ext_codes:
        
    # Now, use the dictionary to search for proteinPianas associated to those codes...
    #          1. all proteinPianas associated to spAcc_id emblAcc_id gi_id sequence
    #          2. if nothing found, search for geneName using description
    #                 --> if only one proteinPiana associated: use it relying on it!
    #                 --> if more than one proteinPiana associated, use it setting it to _c
    

    temp_list_proteinPiana = []


    for spAcc_id_value in dic_all_codes['spAcc_id']:
        temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= spAcc_id_value,
                                                                           id_type= "uniacc",
                                                                           tax_id_value= tax_id, source_db_info= "no")         )

    for emblAcc_id_value in dic_all_codes['emblAcc_id']:
        temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= emblAcc_id_value,
                                                                           id_type= "ncbiacc",
                                                                           tax_id_value= tax_id, source_db_info= "no")         )

    for gi_id_value in dic_all_codes['gi_id']:
        temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= gi_id_value,
                                                                           id_type= "gi",
                                                                           tax_id_value= tax_id, source_db_info= "no")         ) 
        
    for sequence_value in dic_all_codes['sequence']:
        temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= sequence_value,
                                                                           id_type= "sequence",
                                                                           tax_id_value= tax_id, source_db_info= "no")         )
        
        
    if not temp_list_proteinPiana:
        # nothing was found for 'good' codes.... now check how many proteinPianas associated to this description there are...
        # We are searching if the description is a PRIMARY geneName
        # If there is only one, we will use it (and trust it)
        temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= protein_description,
                                                                          id_type= "geneName",
                                                                          tax_id_value= tax_id, source_db_info= "no", only_primary="yes") )
        if temp_list_proteinPiana:
            parser.add_to_log("Number of proteinPianas found using geneName")
    else:
        parser.add_to_log("Number of proteinPianas found from uniprotAcc, ncbiacc, giID or sequence")
        
                                     
    # If Not proteinPiana has been found, try to find it without fixing tax_id_value for type ids uniprotAccession, ncbi accessionNumber and gi:
    if not temp_list_proteinPiana:
        for spAcc_id_value in dic_all_codes['spAcc_id']:
            temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= spAcc_id_value,
                                                                              id_type= "uniacc",
                                                                              tax_id_value= 0, source_db_info= "no")         )

        for emblAcc_id_value in dic_all_codes['emblAcc_id']:
            temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= emblAcc_id_value,
                                                                              id_type= "ncbiacc",
                                                                              tax_id_value= 0, source_db_info= "no")         )

        for gi_id_value in dic_all_codes['gi_id']:
            temp_list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id= gi_id_value,
                                                                              id_type= "gi",
                                                                              tax_id_value= 0, source_db_info= "no")         )
        if temp_list_proteinPiana:
            parser.add_to_log("Number of proteinPianas found using uniacc,ncbiacc or giID without using taxID")
        

    # remove duplicates
    if temp_list_proteinPiana:
        list_proteinPiana = dict(map(None,temp_list_proteinPiana,[None]))
    else:
        list_proteinPiana = []

    return list_proteinPiana



    
testing_uniq  = {} # just testing how many different interactions are inserted...


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

# opening connection to MySQL PSI database and create a cursor to work with the database
if psi_dbuser is None and psi_dbpass is None:
    psidb = MySQLdb.connect(db=psi_dbname, host=psi_dbhost)
                
elif psi_dbpass is None and psi_dbuser is not None:
    psidb = MySQLdb.connect(user=psi_dbuser, db=psi_dbname, host=psi_dbhost)
    
else:
    psidb = MySQLdb.connect(user=psi_dbuser, db=psi_dbname, host=psi_dbhost, passwd= psi_dbpass )

 
psicursor = psidb.cursor()


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = "temporal_mysql_database",
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)


piana_access.lock_tables(table_list = tables_to_lock )


if insert_psidb_ids:
    # if user asked to insert protein internal ids for the psi database, do it
    if psi_db == "hprd":
        # but only if it is HPRD (the other databases do not have 'interesting' internal ids)


        piana_access.insert_new_id_type(proteinTypeId="hprd", externalIdDescription="HPRD identifier for the protein")
        
        sqlquery = """ select counter_id, db_id from proteins"""
        psicursor.execute(sqlquery)
        all_pairs_protein_ids = psicursor.fetchall()

        for one_pair_protein_id in all_pairs_protein_ids:
            # one_pair_protein_id is [counter_id, db_id]
            all_proteinPiana= get_associated_list_proteinPianas(psi_cursor= psicursor, piana_access= piana_access,
                                                                node_psi_id=one_pair_protein_id[0],
                                                                parser=parser)


            for one_proteinPiana in all_proteinPiana:
                # insert the dip_uid code into the piana database
                piana_access.insert_external_id_code(external_id_type = "hprd",
                                                     external_id_value = one_pair_protein_id[1].strip(),
                                                     proteinPiana_value = one_proteinPiana,
                                                     sourceDBID = sourceDB )

# END OF if insert_psidb_ids:

# Obtaining all links in database PSI

if verbose:
    sys.stderr.write("Inserting interactions from PSI\n")

sqlquery = """ select interaction_key, proteinA_counter_id, proteinB_counter_id from interactions"""
psicursor.execute(sqlquery)
psi_interactions = psicursor.fetchall()


# For each interaction in PSI we do the following:
#  1. find the method of the interaction
#  2. find the psi external identifier for proteins involved in the interaction
#  3. find the piana internal identifier for those proteins
#  4. insert a piana interaction with the internal identifiers
if verbose:
    num_total_ints = len(psi_interactions)
    sys.stderr.write("Number of psi interactions is: %s \n" %(num_total_ints))

log_fd = log_file_fd

num_psi_interaction=0

for psi_interaction in psi_interactions:
    #   psi_interaction[0] --> interaction_key
    #   psi_interaction[1] --> proteinA_counter_id
    #   psi_interaction[2] --> proteinB_counter_id

    num_psi_interaction += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if num_psi_interaction%1000==0:
            sys.stderr.write("%s interactions done in %s seconds\n" %(num_psi_interaction,time.time()-initial_time))

    
    # -----------------------------------------------------------------------
    #  1. find the method and pubmed id of the interaction
    # -----------------------------------------------------------------------
    #     after the query, we'll have a record with the following info:
    #          list_methods -->  list of tuples (method,)
    #          list_pmids -->  list of tuples (pmid,)
    #         

    if verbose_detailed:
        sys.stderr.write("getting method and pubmed from psi database\n" )
        
    sqlquery = """select method from interactionMethod where interaction_key = '%s' """ %(psi_interaction[0])
    psicursor.execute(sqlquery)
    list_methods_temp = psicursor.fetchall()

    # Transform the tuple of tuples of list_methods_temp to a list
    list_methods = [x[0] for x in list_methods_temp]

    
    sqlquery = """select pubmed_id from interactionFeatures where interaction_key = '%s' """ %(psi_interaction[0])
    psicursor.execute(sqlquery)
    list_pmids_temp = psicursor.fetchall()

    # Transform the tuple of tuples of list_methods_temp to a list
    list_pmids = [x[0] for x in list_pmids_temp]


    # -----------------------------------------------------------------------
    #  2. get a list of protein pianas for psi nodes "from:" and "to:"
    # -----------------------------------------------------------------------
    if verbose_detailed:
        sys.stderr.write("getting proteinPianas for psi A counter_id %s\n" %(psi_interaction[1]))
        
    list_proteinPiana_a= get_associated_list_proteinPianas(psi_cursor= psicursor, piana_access= piana_access, node_psi_id=psi_interaction[1],parser=parser)
    if verbose_detailed:
        sys.stderr.write("getting proteinPianas for psi B counter_id %s\n" %(psi_interaction[2]))
        
    list_proteinPiana_b= get_associated_list_proteinPianas(psi_cursor= psicursor, piana_access= piana_access, node_psi_id=psi_interaction[2],parser=parser)

    # list_proteinPiana_x is a list of proteinPiana

    if verbose_detailed:
        sys.stderr.write("finding source for proteinPianas\n" )
        

    # -----------------------------------------------------------------------
    #  3. insert a piana interaction with the internal identifiers
    # -----------------------------------------------------------------------

    if verbose:
        sys.stderr.write("int %sof%s: %s (pps %s) and %s (pps %s) and methods %s\n" %(num_psi_interaction,
                                                                                      num_total_ints,
                                                                                      psi_interaction[1],
                                                                                      list_proteinPiana_a,
                                                                                      psi_interaction[2],
                                                                                      list_proteinPiana_b,
                                                                                      list_methods))
        
    # TO DO!!! Confidence value is now set to 1!!! Change to proper value. In fact, interactionConfidence_value will not be an argument
    #          since it will be an internal value calculated from several factors (sourceDB, method used, intersection of DBs, ...)

    for proteinPiana_a in list_proteinPiana_a:
        for proteinPiana_b in list_proteinPiana_b:

            if verbose:
                if proteinPiana_a <= proteinPiana_b:
                    a= proteinPiana_a
                    b= proteinPiana_b
                else:
                    a= proteinPiana_b
                    b= proteinPiana_a
                
                key = "%s.%s" %(a, b)
                testing_uniq[key] = None
            # END OF if verbose:
                
            if len(list_methods) == 0:
                # there was no method associated... set a unclassified method
                list_methods = ["unclassified"]

            if len(list_pmids) == 0:
                list_pmids = [0]

            if verbose:
                if verbose_detailed:
                    sys.stderr.write("Inserting piana interaction %s -- %s with methods %s, pubmed ids %s and source db %s\n" %(proteinPiana_a,
                                                                                                                                proteinPiana_b,
                                                                                                                                list_methods,
                                                                                                                                list_pmids, source_db ))


            piana_access.insert_interaction(proteinPianaA_value = proteinPiana_a,
                                            isSourceA_value = 1,
                                            proteinPianaB_value = proteinPiana_b,
                                            isSourceB_value = 1,
                                            interactionConfidence_value = 1,
                                            methodDescription_value = list_methods,
                                            sourceDBDescription_value = sourceDB,
                                            confidenceAssignedSourceDB_value=1,
                                            pubmed_id_value= list_pmids)

            # Control
            if proteinPiana_a>proteinPiana_b:
                temp = proteinPiana_a
                proteinPiana_a=proteinPiana_b
                proteinPiana_b=temp
            if inserted_interactions.has_key((proteinPiana_a,proteinPiana_b)):
                #print "Interaction repeated: psi interaction: %s  Protein Pianas: %s - %s " %(psi_interaction,proteinPiana_a,proteinPiana_b)
                parser.add_to_log("Number of repeated interacions")
            else:
                inserted_interactions[(proteinPiana_a,proteinPiana_b)] = None
            for actual_method in list_methods:
                if inserted_interactionMethods.has_key((proteinPiana_a,proteinPiana_b,actual_method)):
                    parser.add_to_log("Number of repeated interactions with repeated method")
                    #print "Interaction Method repeated: psi interaction: %s  Protein Pianas: %s - %s Method %s" %(psi_interaction,proteinPiana_a,proteinPiana_b,actual_method)
                else:
                    inserted_interactionMethods[(proteinPiana_a,proteinPiana_b,actual_method)] = None

            parser.add_to_log("Number of inserted interactions")

        # END OF for proteinPiana_b in list_proteinPiana_b:
    # END OF for proteinPiana_a in list_proteinPiana_a:

    if not list_proteinPiana_a or not list_proteinPiana_b:
        parser.add_to_log("Number of interactions not inserted because not found proteinPiana")
        
        if verbose:
            sys.stderr.write("Int not inserted (no proteinPiana found) -->  PSI int %s between <<%s>> and <<%s>> \n" %(psi_interaction[0],
                                                                                                                       psi_interaction[1],
                                                                                                                       psi_interaction[2]))

# END OF for psi_interaction in psi_interactions:


if verbose:
    sys.stderr.write("%s" %(parser.get_log_string()) )


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s interactions done in %s seconds\n" %(num_psi_interaction, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
                             

