"""
File        : refseq2piana.py
Author      : Javier Garcia and Ramon Aragues
Creation    : 2.2007
Contents    : script that fills up tables in database piana related to  external DB "RefSeq"
Called from : Command line

=======================================================================================================

This file parsers RefSeq-releaseXX.catalog  NCBI file and inserts information into PianaDb

Command line option '--help' describes usage of this program

For more details on how to use it, read piana/README.populate_piana_db

"""


#=======================================================================================================
# This file parsers a RefSeq Complete Release Catalog NCBI text file and inserts information into PianaDb
# Command line option '--help' describes usage of this program
# For more details on how to use it, read piana/README.populate_piana_db
#
# gene2accession2piana.py: script that fills up tables in database piana related to external DB "Gene" from NCBI
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


## STEP 1: IMPORT NECESSARY MODULES


import sys
import getopt

import re
import readline

import MySQLdb

import time

from PianaDBaccess import *
from parserUtilities import *


# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"the name of the RefSeq input file (RefSeq Catalog file)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("tax-id-file=",None,"a text file with two tab-separated columns: 1st is gi code and 2nd is taxonomy id"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]


# Define the parser information to print in the help:

parser = ParserObject( script_name="refseq2piana.py",
                       script_description="This file implements a program that fills up tables in database piana with information from refseq",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dic = parser.parseArguments()

refseqcatalog_file = arguments_dic["input-file-name"]
piana_dbname =arguments_dic["piana-dbname"]
piana_dbhost =arguments_dic["piana-dbhost"]
tax_id_file=arguments_dic["tax-id-file"]
sourceDB=arguments_dic["database-name"]
database_version=arguments_dic["database-version"]
database_description=arguments_dic["database-description"]
piana_dbuser=arguments_dic["piana-dbuser"]
piana_dbpass=arguments_dic["piana-dbpass"]
help=arguments_dic["help"]
verbose=arguments_dic["verbose"]
time_control=arguments_dic["time-control"]
log_file=arguments_dic["log-file"]
database_information=arguments_dic["database-information"]


# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')


# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.gi_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.protein_table]


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = refseqcatalog_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

refseq_fd = open(refseqcatalog_file,"r")

#
# Reading external DB "Gene" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from %s  file...\n" %(refseqcatalog_file) ) 

line_number=0

for line in refseq_fd:

    line_number += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if not line_number%20000:
            sys.stderr.write("%s\t%s\n" %(line_number,time.time()-initial_time))

    # Parsing the line, obtain fields:
    # 1. tax_id
    # 2. species name
    # 3. RefSeq accession.version
    # 4. gi
    # 5. FTP directories data is provided in
    # 6. RefSeq status code
    # 7. sequence length

    line_fields = line.split("\t")

    # line_fields is a vector with
    #                              (tax_id, species_name, RefSeq_accession.version, gi, FTP_directory,
    #                               RefSeq_status, sequence_length)
    
    if line_fields:

        if( line_fields[2] != "-" and line_fields[3] != "-" ):
            tax_id = int( line_fields[0].strip() )
            refseq_accession_version = line_fields[2].strip()
            gi_id = int( line_fields[3].strip() )

            p = re.compile("(\w+)\.\d+")
            r = p.match(refseq_accession_version)

            refseq_accession = r.group(1)
            #refseq_version = r.group(2)    Not taken into account

            if verbose:
                sys.stderr.write( "taxID  is <%s> -- refseq_accession is <%s> -- proteinGI is <%s>  \n" %(tax_id, refseq_accession, gi_id))
                
            parser.add_to_log("Number of RefSeq references inserted")

            if( tax_id == "" or tax_id == "-" ):
                tax_id = 0

            # Get the proteinPiana list that correspond to this gi code
            proteinPiana_list = piana_access.get_list_proteinPiana(protein_id=gi_id, id_type="gi",
                                                                    tax_id_value=tax_id, source_db_info="no")

            # If not proteinPiana are found, search without specie:
            if len(proteinPiana_list)==0:
                proteinPiana_list = piana_access.get_list_proteinPiana(protein_id=gi_id, id_type="gi",
                                                                       tax_id_value=0, source_db_info="no")
            
            if len(proteinPiana_list)==0:
                if verbose:
                    sys.stderr.write( "gi code %s not found!  \n" %(gi_id) )
                parser.add_to_log("Number of RefSeq references NOT inserted")
            else:
                parser.add_to_log("Number of RefSeq references inserted")
                for actual_proteinPiana in proteinPiana_list:
                    piana_access.insert_refseq_code( refseq_code_value = refseq_accession, proteinPiana_value=actual_proteinPiana, refseq_source_value=sourceDB)
                    if verbose:
                        sys.stderr.write(" Inserting code %s!!!\n" %(refseq_accession) )


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables


piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(line_number, (time.time()-initial_time) ))


if verbose:
    sys.stderr.write(parser.get_log_string())

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()


### Using release21.accession2geneid file  I could check for the consistencies between proteinGI and proteinAccession... (previously inserted in the database)
### If the correspondence is found, continue with the parsing
### If the correspondence is not found... What should I do???
### Should I do this...? For the moment, we do not do this...
