"""
File        : scop2piana.py
Author      : Ramon Aragues
Creation    : 8.11.2004
Contents    : script that fills up tables in database piana from  "scop" 
Called from : 

=======================================================================================================

This program parsers a text file with scop information and inserts them into piana

The specific files from scop that are parsed by this parser are: dir.cla.scop.txt_xxx


Command line option '--help' describes usage of this program


"""

# scop2piana.py: script that fills up tables in database piana from  "scop" 
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline

import time

from PianaDBaccess import *
from parserUtilities import *




# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"the name of the scop input file (dir.cla.scop.txt_xxx)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]
                           
# Define the parser information to print in the help:

parser = ParserObject( script_name="scop2piana.py",
                       script_description="This program fills up tables in database piana related to external DB 'scop'",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dict = parser.parseArguments()

input_file_name = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]





# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')




# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.proteinScop_table,
                  PianaGlobals.pdb_table,
		  PianaGlobals.temporalData_table]


# --------
# --------
#  Main()               
# --------                               
# --------

number_of_lines = 0

# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file_name.split("/")[-1],
                                          databaseDescription = database_description)

scop_fd = file(input_file_name,"r")



#
# Reading external DB "scop" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write("Reading data from scop file...\n")

for line in scop_fd:

    number_of_lines += 1
    parser.add_to_log("Number of lines processed")

    if line[0] == "#":
        # skip comments
        continue

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if not number_of_lines%20000:
            sys.stderr.write("%s lines done in %s seconds\n" %(number_of_lines,time.time()-initial_time) )

    
    # Retrieving different fields in line
    # line example:
    # d1dlya_ 1dly    A:            a.1.1.1 14983   cl=46456,cf=46457,sf=46458,fa=46459,dm=46460,sp=46462,px=14983
    # d1dlya_ 1dly    A:12-123      a.1.1.1 14983   cl=46456,cf=46457,sf=46458,fa=46459,dm=46460,sp=46462,px=14983

    line_fields = line.split()  # line_fields[0] is complete pdb entry
                                # line_fields[1] is pdb code
                                # line_fields[2] is pdb chain follow by : XX-YY (optional)
                                # line_fields[3] is ???
                                # line_fields[4] is ???
                                # line_fields[5] is comma-separated codes


    if len(line_fields) != 6:
        # skip incomplete lines
        parser.add_to_log("Number of incomplete lines")
        continue
    
    pdb_code = line_fields[1]

    pdb_chain_fields = line_fields[2].split(":")
    if len(pdb_chain_fields) == 2:
        # standard line, chain:something (where something can be nothing or xx-yy)
        pdb_chain = pdb_chain_fields[0]
    else:
        # if pdb chain format is incorrect, set to A
        pdb_chain = "A"

    if pdb_chain == "-":
        # in piana, there are no empty chains: set to A
        pdb_chain = "A"
    
    pdb_code_chain = pdb_code + "." + pdb_chain
    list_code_values = line_fields[5].split(",")
    
    for code_value in list_code_values:
        
        code_value_pair = code_value.split("=")
        
        if code_value_pair[0] == "cf":
            cf_value = code_value_pair[1]
            
        elif code_value_pair[0] == "sf":
            sf_value = code_value_pair[1]
            
        elif code_value_pair[0] == "fa":
            fa_value = code_value_pair[1]
    # END OF for code_value in list_codes:

    # pdb.chain points directly to a species: no need to check tax id
    list_proteinPiana = piana_access.get_list_proteinPiana(protein_id=pdb_code_chain,
                                                            id_type= "pdb.chain",
                                                            tax_id_value =0,
                                                            source_db_info= "no")
    
    
    for proteinPiana in list_proteinPiana:

        if verbose:
            sys.stderr.write("Inserting cf=%s sf=%s fa=%s for proteinPiana %s of pdb_chain %s\n" %(cf_value, sf_value, fa_value,
                                                                                                   proteinPiana, pdb_code_chain) )
              

        piana_access.insert_protein_scop(cf= cf_value, sf= sf_value, fa= fa_value, proteinPiana_value= proteinPiana, proteinScopSource_value=sourceDB)

        
    # END OF for proteinPiana_a in list_proteinPiana_a:
    if not list_proteinPiana:
        parser.add_to_log("Number of proteins without proteinPiana")
        if verbose:
            sys.stderr.write("proteinPiana not found for %s \n" %(pdb_code_chain))
            

# END OF for line in scop_fd:


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s lines done in %s seconds\n" %(number_of_lines, (time.time()-initial_time) ))

if verbose:
    sys.stderr.write(parser.get_log_string())

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

