"""
File        : string2piana.py
Author      : Ramon Aragues
Creation    : 21.01.2005
Modifications: Javier Garcia May 2007
Contents    : script that fills up tables in database piana related to external DB "string"
Called from : 

=======================================================================================================

This program reads database STRING. Then, inserts relevant information into database PIANA 


option --start-at=interaction_number permits the partitioning of the piana insertions. This has been done to
allow the user to update piana in several steps, since it takes so long to do all of them. 


Command line option '--help' describes usage of this program


"""

# string2piana.py: script that fills up tables in database piana related to external DB "string"
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline

import MySQLdb

from PianaDBaccess import *
from parserUtilities import *
import utilities



## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("string-dbname=",None,"name of the string mysql database"),
                              ("string-dbhost=",None,"name of host where string mysql database is placed"),
                              ("score-thres=",None,"minimum STRING combined score required for inserting interaction into piana_dbname"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("string-dbuser=",None,"name of the mysql string username (not required in most systems)"),
                           ("string-dbpass=",None,"name of the mysql string username (not required in most systems)"),
                           ("insert-cogs",None,"inserts cogs (and kogs) for proteins in string"),
                           ("insert-ints",None,"inserts string interactions"),
                           ("interaction-number=",0,"number of interaction from which the program will start piana insertions"),
                           ("number-seqs-threshold=",100,"avoid introducing noise to PIANA by setting a threshold on how many sequences can be associated to a STRING code"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]

# Define the parser information to print in the help:

parser = ParserObject( script_name="string2piana.py",
                       script_description="This program fills up tables in database piana related to external DB 'string'",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters


arguments_dict = parser.parseArguments()

interaction_number = int(arguments_dict["interaction-number"])
string_dbname = arguments_dict["string-dbname"]
string_dbhost = arguments_dict["string-dbhost"]
string_dbuser = arguments_dict["string-dbuser"]
string_dbpass = arguments_dict["string-dbpass"]
    
score_thres = int(arguments_dict["score-thres"])
number_seqs_threshold = int(arguments_dict["number-seqs-threshold"])
    
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
piana_dbuser = arguments_dict["piana-dbuser"]
piana_dbpass = arguments_dict["piana-dbpass"]

insert_cogs = arguments_dict["insert-cogs"]
insert_ints = arguments_dict["insert-ints"]
    
verbose = arguments_dict["verbose"]

log_file = arguments_dict["log-file"]
time_control= arguments_dict["time-control"]

sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
database_information= arguments_dict["database-information"]

interaction_number =  int(arguments_dict["interaction-number"])
insert_cogs =  arguments_dict["insert-cogs"]
insert_ints =  arguments_dict["insert-ints"]

# Start time counter
if time_control:
    initial_time = time.time()


string_id_proteinPianas = {}        # This dictionary will contain those proteinPianas associated to a stringDB identifier. It is used
                                    # to avoid multiple queries with the same identifiers, and so, speed up the parser
                                    # Key: stringDB identifier
                                    # Value: List of proteinPianas



verbose_detailed = 0
verbose_cog = 0



# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
# Table names are defined in piana/code/PianaDB/PianaGlobals.py
# If you don't known which tables are being used, the list can be empty: in this case, all tables will be locked
# While tables are locked, anyone else can access to them

tables_to_lock = [PianaGlobals.protein_table,
                  PianaGlobals.proteinCog_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.ensembl_table,
                  PianaGlobals.refseq_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.interactionScores_table,
                  PianaGlobals.interaction_table,
                  PianaGlobals.interactionMethod_table,
                  PianaGlobals.interactionFeatures_table,
                  PianaGlobals.proteinPianaCounter_table]




# setting string that describes method used for finding interaction
methodDescription_value = "predicted"
   
def get_associated_proteinPianas(piana_access, string_cursor, return_source_db="no", string_protein_id=None, number_seqs_threshold=1000):
    """
    returns a list of proteinPianas associated to protein "protein_id" of species "tax_id"

    The type of code being used is 'guessed' using regexps: if no type code can be inferred from the protein value, returns empty list
    """

    global string_id_proteinPianas

    try:
        return string_id_proteinPianas[string_protein_id]
    except KeyError:
        pass
    
    # get the code and the tax id
    fields_protein_id = string_protein_id.split(".")

    if len(fields_protein_id) == 2:
        protein_id= fields_protein_id[1]
        tax_id = int(fields_protein_id[0])
    else:
        return []

    # OLD CODE: Not used now, because it is very relative. STRING does not tell what id type is. Trying to guess it is difficult. So, we will search for these codes
    # in all possible type ids, using the taxID, and append all the results
    # guess the type of identifier used in this id (a list is returned with potential types being used)
    #list_id_types = utilities.get_id_type(protein_id)

    #list_proteinPiana = []

    #if list_id_types:
    #    
    #    for id_type in list_id_types:

    #        if id_type == "geneName":
    #            # string has _ instead of . for gene names
    #            protein_id = protein_id.replace("_", ".")

    #        list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
    #                                                                    id_type = id_type,
    #                                                                    tax_id_value= tax_id, source_db_info= return_source_db))

        # END OF for id_type in list_id_types:
    # END OF if list_id_types:


    list_proteinPiana = []

    # Try for code type uniEntry:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type = "unientry",
                                                                tax_id_value = tax_id))

    # Try for code type refseq:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type="refseq",
                                                                tax_id_value = tax_id))

    # Try for code type ensembl:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type="ensembl",
                                                                tax_id_value = tax_id))

    # Try for code type uniAcc:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type="uniacc",
                                                                tax_id_value = tax_id,
                                                                only_primary="yes"))

    # Try for code type accessionNumber:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type="ncbiacc",
                                                                tax_id_value = tax_id))

    # Try for code type geneName:
    list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                id_type ="geneName",
                                                                tax_id_value = tax_id,
                                                                only_primary="yes"))
                                                                

                             
    if not list_proteinPiana:
        # if nothing found for the code given, search in STRING table identifiers_proteins

        list_proteinPiana= []
        
        string_sqlquery = """SELECT identifier FROM identifiers_proteins where protein_id="%s" """ %(string_protein_id)
        string_cursor.execute(string_sqlquery)
        protein_identifiers = string_cursor.fetchall()

        
        for one_identifier in protein_identifiers:
            protein_id = one_identifier[0]
            #OLD CODE:
            #list_id_types = utilities.get_id_type(one_identifier[0])

            #if not list_id_types:
            #    continue
            #for id_type in list_id_types:
                
                #if id_type == "geneName":
                #
                #    # if it is a geneName... ignore: they introduce to much noise...
                #    continue
                    
                #    # string has _ instead of . for gene names
                #    #protein_id  = one_identifier[0].replace("_", ".")
                #else:
                #    protein_id  = one_identifier[0]
                    
                #list_proteinPiana.extend( piana_access.get_list_proteinPiana(protein_id = protein_id,
                #                                                              id_type = id_type,
                #                                                              tax_id_value= tax_id,
                #                                                              source_db_info= return_source_db))
                
                # Try for code type uniEntry:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type = "unientry",
                                                                        tax_id_value = tax_id))

            # Try for code type refseq:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type="refseq",
                                                                        tax_id_value = tax_id))
            
            # Try for code type ensembl:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type="ensembl",
                                                                        tax_id_value = tax_id))
            
            # Try for code type uniAcc:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type="uniacc",
                                                                        tax_id_value = tax_id,
                                                                        only_primary="yes"))

            # Try for code type accessionNumber:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type="ncbiacc",
                                                                        tax_id_value = tax_id))
            
            # Try for code type geneName:
            list_proteinPiana.extend(piana_access.get_list_proteinPiana(protein_id = protein_id,
                                                                        id_type ="geneName",
                                                                        tax_id_value = tax_id,
                                                                        only_primary="yes"))
                
    # END OF if not list_proteinPiana:

    # avoid returning entries which are too ambigous... for example, 601.EFTS_SALTY corresponds to more than
    # 200 proteinPianas (there is an emblAccession associated to it (AL627266) which identifies a whole chromosome)
    # We avoid these cases in a not-so-clean way, but I cannot think of another way of doing it... we are just
    # skipping these cases by applying a threshold to the number of proteinPianas that a code can be associated to

    if len(list_proteinPiana) > number_seqs_threshold:
        string_id_proteinPianas[string_protein_id] = []
        return []

    # removes duplicates and add the list to the dictionary
    return_list = dict(map(None,list_proteinPiana,[])).keys()
    string_id_proteinPianas[string_protein_id] = return_list
    return return_list



# ---------------------------
# Function get_species()                                               
# --------------------------- 

# TO DO!!! Not using STRING species information!
# TO DO!!! There are lots of information in STRING that could be used in PIANA...

def get_species(string_protein_id):
    sqlquery = """select species_id from proteins_species where gene_id = "%s";""" %(string_protein_id)
    cursorString.execute(sqlquery)
    auxspecies = cursorString.fetchall()
    if auxspecies:
        return auxspecies[0][0]
    else:
        return None
 
# ---------------------------
# method insert_protein_cogs()                                               
# --------------------------- 
   
def insert_protein_cogs(string_cursor):
    """
    insert COG codes from STRING into pianaDB
    """

    string_sqlquery = """SELECT protein_id, orthgroup_id FROM proteins_orthgroups"""
    string_cursor.execute(string_sqlquery)
    protein_ids_ary = string_cursor.fetchall()

    for protein_id_info_fields in protein_ids_ary:

        # get protein name from the string protein id (which looks like this: 198094.Q81LD0 or  2214.GSA_METAC or ...)
        string_protein_id = protein_id_info_fields[0]
        protein_ids_fields = re.split("\.", string_protein_id)
        protein_id = protein_ids_fields[1].strip()

        # get the species for this protein
        tax_id = int(protein_ids_fields[0].strip())

        
        for cog_code in protein_id_info_fields[1:]:
            cog_code = cog_code.strip()

            list_proteinPiana = get_associated_proteinPianas(piana_access= piana_access,
                                                             string_cursor= string_cursor,
                                                             return_source_db="yes",
                                                             string_protein_id=string_protein_id,
                                                             number_seqs_threshold=number_seqs_threshold)

            for protein_piana in list_proteinPiana:
                # protein_piana is a pair (proteinPiana, source_db)

                if verbose_detailed:
                    sys.stderr.write( "inserting cog id %s for proteinPiana %s (code is %s of type %s)" %(cog_code, protein_piana[0],
                                                                                                              protein_id, id_type))

                piana_access.insert_protein_cog( cog_id = cog_code,
                                                 proteinPiana_value = protein_piana,
                                                 proteinCogSource_value= sourceDB)
                #if protein_piana[1] == "completion":
                #    piana_access.insert_protein_cog( cog_id= cog_code,
                #                                     proteinPiana_value= protein_piana[0],
                #                                     proteinCogSource_value= "string_c")
                #else:
                #    piana_access.insert_protein_cog( cog_id= cog_code,
                #                                     proteinPiana_value= protein_piana[0],
                #                                     proteinCogSource_value= "string")
                        
            # END OF for protein_piana in list_proteinPiana:
        # END OF for cog_code in protein_id_info_fields[1:]:
    # END OF for protein_id_info_fields in protein_ids_ary:

# --------
# --------
#  Main()               
# --------                               
# --------




number_of_interactions_no_id = 0 # will count interactions not found because proteinPiana not found
number_of_interactions_low_score = 0 # will count interactions not found because proteinPiana not found
number_of_interactions_added = 0 # will count number of interactions added

# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

# opening connection to MySQL STRING database and create a cursor to work with the database
if string_dbuser is None and string_dbpass is None:
    stringdb = MySQLdb.connect(db=string_dbname, host=string_dbhost)
elif string_dbpass is None:
    stringdb = MySQLdb.connect(user=string_dbuser, db=string_dbname, host=string_dbhost)
else:
    stringdb = MySQLdb.connect(user=string_dbuser, db=string_dbname, host=string_dbhost, passwd= string_dbpass )

cursorString = stringdb.cursor()


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = "MySQL database %s" %(string_dbname),
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

if insert_cogs:
    if verbose:
        sys.stderr.write( "inserting cogs for all proteins in string\n")
    insert_protein_cogs(string_cursor= cursorString)
# END OF if insert_cogs:


if not insert_ints:
    sys.exit()

if verbose:
    sys.stderr.write( "Obtaining the list of all STRING interactions\n")


# finding out how many queries will be needed to get all String interactions
#  --> if the query is done in one single step, the computer runs out of memory and gets blocked...
sqlquery= """select count(*) from precomputed_protein_links;"""
cursorString.execute(sqlquery)
number_of_interactions = cursorString.fetchall()[0][0]   # fetachall answer of a count comes in [0][0]

#last_interaction_retrieved = -1   # last interaction retrieved (init to -1 so the first one we get is 0)
last_interaction_retrieved = -1
query_size = 1000000                 # how many interactions will be retrieved in each query (needed because of memory limitations)
interactions_to_retrieve = number_of_interactions  # how many interactions haven't been processed

number_interactions_processed=0

while interactions_to_retrieve > 0:
    # Obtaining the list of all interactions in database STRING


    if verbose:
        sys.stderr.write("%sto%s-" %(last_interaction_retrieved+1, last_interaction_retrieved+1+query_size))
        sys.stderr.flush()

    sqlquery = """select protein_id_a, protein_id_b, equiv_nscore, equiv_nscore_transferred, equiv_fscore, equiv_pscore, equiv_hscore, array_score, array_score_transferred, experimental_score, experimental_score_transferred, database_score, database_score_transferred, textmining_score, textmining_score_transferred, combined_score  from precomputed_protein_links limit %s,%s;""" %(last_interaction_retrieved+1, query_size)
    cursorString.execute(sqlquery)
    list_interactions = cursorString.fetchall()



    number_of_interactions = len(list_interactions)
    i=0

    for gene_pair in list_interactions:

        number_interactions_processed += 1

        piana_access.check_lock_frequency(tables_to_lock)

        if time_control:
            if number_interactions_processed%20000==0:
                sys.stderr.write("%s interactions processed in %s seconds\n" %(number_interactions_processed, time.time()-initial_time))

        # Note: variables are called gene_* because old versions of STRING used gene_id_a and gene_id_b instead of protein_id_a and protein_id_b
        #    --> if you are trying to parse an old version of STRING that still uses gene*, you have to change the following in this parser:
        #                 protein_id_a --> gene_id_a
        #                 protein_id_b --> gene_id_b
        #                 precomputed_protein_links --> precomputed_gene_links
        #                 proteins_orthgroups --> genes_orthgroups
        #                 protein_id --> gene_id
        #                 proteins_species --> genes_species
        #
        # gene_pair[0] = gene_id_a
        # gene_pair[1] = gene_id_b
        # gene_pair[2] = equiv_nscore
        # gene_pair[3] = equiv_nscore_transferred_col
        # gene_pair[4] = equiv_fscore_col
        # gene_pair[5] = equiv_pscore
        # gene_pair[6] = equiv_hscore
        # gene_pair[7] = array_score
        # gene_pair[8] = array_score_transferred
        # gene_pair[9] = experimental_score
        # gene_pair[10] = experimental_score_transferred
        # gene_pair[11] = database_score
        # gene_pair[12] = database_score_transferred
        # gene_pair[13] = textmining_score
        # gene_pair[14] = textmining_score_transferred
        # gene_pair[15] = combined_score

        # TO DO!!! Either set a command line option to choose threshold or work on piana to set confidence thresholds when retrieving interactions
        # so we can insert all interactions in STRING but only use those that are reliable

        if int(gene_pair[15]) < score_thres:
            # this is temporary: trying to parse string faster... takes too long to insert all interactions regardless of their
            # combined score. Therefore, I only insert those interactions above a certain threshold
            
            if verbose_detailed:
                sys.stderr.write( "%sto%s-Skipping  interaction number %s (score %s) of a total of %s\n" %(last_interaction_retrieved+1,
                                                                                                           last_interaction_retrieved+1+query_size,
                                                                                                           i, gene_pair[15], number_of_interactions))
            i = i+1
            number_of_interactions_low_score += 1
            parser.add_to_log("Number of interactions with low score")
            continue
        # END OF if int(gene_pair[15]) < score_thres:

        if verbose_detailed:
            sys.stderr.write( "%sto%s-%s(p1=%s, p2=%sscore %s)of%s\n" %(last_interaction_retrieved+1,
                                                                        last_interaction_retrieved+1+query_size,
                                                                        i, gene_pair[0], gene_pair[1], gene_pair[15], number_of_interactions))
        i = i+1


        # before, there was an  "if" here... permiting to re-start the piana update at a given interaction_number
        # it was used for those cases that the parser was halt due to an error, so it didn't have to restart all over again
        # why has it been removed? Don't know :-) But if I needed again, I just have to place it here, don't I?


        # retrieving the swissprotID/swissAccession/trembl from String protein ID
        # string_protein_id is an internal code of STRING that looks something like: 117.3MGH_RHOBA
        # the first term is the protein taxonomy id
        # the second term is the protein identifier (swissprot ID, swissprot Accession, ...)

        list_proteinPiana_a = get_associated_proteinPianas(piana_access=piana_access,
                                                           string_cursor=cursorString,
                                                           return_source_db="yes",
                                                           string_protein_id=gene_pair[0],
                                                           number_seqs_threshold=number_seqs_threshold)
        
        list_proteinPiana_b = get_associated_proteinPianas(piana_access=piana_access,
                                                           string_cursor=cursorString,
                                                           return_source_db="yes",
                                                           string_protein_id=gene_pair[1],
                                                           number_seqs_threshold=number_seqs_threshold)
        


        # remove redundant codes (redundancies happen between different sourceDBs and find out if it is
        # a code that comes from a completion (ie. it only comes from a completion)

        #      create a dic of unique proteinPianas associated to protein A
        #dic_proteinPiana_a = {}
        #for proteinPiana_a in list_proteinPiana_a:
            # proteinPiana_a is a tuple 
            #    proteinPiana_a[0] --> proteinPiana
            #    proteinPiana_a[1] --> source_dd

            #if proteinPiana_a[1] != "completion":
            #    dic_proteinPiana_a[proteinPiana_a[0]] = 'standard'
            #elif proteinPiana_a[1] == "completion":
            #    if not dic_proteinPiana_a.has_key(proteinPiana_a[0]):
            #        dic_proteinPiana_a[proteinPiana_a[0]] = 'completion'
        # END OF for proteinPiana_a in list_proteinPiana_a:

        #      create a dic of unique proteinPianas associated to protein B
        #dic_proteinPiana_b = {}
        #for proteinPiana_b in list_proteinPiana_b:
            # proteinPiana_b is a tuple 
            #    proteinPiana_b[0] --> proteinPiana
            #    proteinPiana_b[1] --> source_dd

            #if proteinPiana_b[1] != "completion":
            #    dic_proteinPiana_b[proteinPiana_b[0]] = 'standard'
            #elif proteinPiana_b[1] == "completion":
            #    if not dic_proteinPiana_b.has_key(proteinPiana_b[0]):
            #        dic_proteinPiana_b[proteinPiana_b[0]] = 'completion'
        # END OF for proteinPiana_b in list_proteinPiana_b:


        list_methods=[]
        if gene_pair[2]>=score_thres:
            list_methods.append("neighborhood")
        if gene_pair[3]>=score_thres:
            list_methods.append("interologn")
        if gene_pair[4]>=score_thres:
            list_methods.append("fusion")
        if gene_pair[5]>=score_thres:
            list_methods.append("ocurrence")
        if gene_pair[7]>=score_thres:
            list_methods.append("coexpress")
        if gene_pair[8]>=score_thres:
            list_methods.append("interologcoexp")
        if gene_pair[9]>=score_thres:
            list_methods.append("experimental")
        if gene_pair[10]>=score_thres:
            list_methods.append("interologexp")
        if gene_pair[11]>=score_thres:
            list_methods.append("databases")
        if gene_pair[12]>=score_thres:
            list_methods.append("interologdb")
        if gene_pair[13]>=score_thres:
            list_methods.append("textmining")
        if gene_pair[14]>=score_thres:
            list_methods.append("interologtm")
        if gene_pair[15]>=score_thres:
            list_methods.append("combinedsc")
            
        for proteinPiana_a in list_proteinPiana_a:
            for proteinPiana_b in list_proteinPiana_b:
                    
                if verbose_detailed:
                    sys.stderr.write( "%sto%s-   - inserting interaction number %s: %s(%s) --> %s(%s)\n" %(last_interaction_retrieved+1,
                                                                                                           last_interaction_retrieved+1+query_size,
                                                                                                           i,
                                                                                                           gene_pair[0], proteinPiana_a,
                                                                                                           gene_pair[1], proteinPiana_b))

                #if dic_proteinPiana_a[proteinPiana_a] == "completion" or dic_proteinPiana_b[proteinPiana_b] == "completion":
                #    source_db = "string_c"
                #else:
                #    source_db = "string"

                # TO DO! piana_combined_score must somehow be calculated internally for each interaction

                last_interaction_id = piana_access.insert_interaction(proteinPianaA_value =proteinPiana_a,
                                                                      isSourceA_value =1,
                                                                      proteinPianaB_value =proteinPiana_b,
                                                                      isSourceB_value =1,
                                                                      interactionConfidence_value = 1,
                                                                      methodDescription_value = list_methods,
                                                                      sourceDBDescription_value= sourceDB,
                                                                      confidenceAssignedSourceDB_value = gene_pair[15])
                    
                # as we have kept the interactionPiana, we now insert specific scores for it
                piana_access.insert_interaction_scores(interactionPiana_value = last_interaction_id,
                                                       sourceDBDescription_value= sourceDB ,
                                                       equiv_nscore_value = gene_pair[2],
                                                       equiv_nscore_transferred_value = gene_pair[3],
                                                       equiv_fscore_value = gene_pair[4],
                                                       equiv_pscore_value = gene_pair[5],
                                                       equiv_hscore_value = gene_pair[6],
                                                       array_score_value = gene_pair[7],
                                                       array_score_transferred_value = gene_pair[8],
                                                       experimental_score_value = gene_pair[9],
                                                       experimental_score_transferred_value = gene_pair[10],
                                                       database_score_value = gene_pair[11],
                                                       database_score_transferred_value = gene_pair[12],
                                                       textmining_score_value = gene_pair[13],
                                                       textmining_score_transferred_value = gene_pair[14],
                                                       combined_score_value = gene_pair[15])
                number_of_interactions_added += 1
                parser.add_to_log("Number of interactions added")
            # END OF for proteinPiana_b in dic_proteinPiana_b:
        # END OF for proteinPiana_a in dic_proteinPiana_a:

        if not list_proteinPiana_a or not list_proteinPiana_b:
                if verbose_detailed:
                    sys.stderr.write("%sto%s-proteinPiana not found for %s or %s \n" %(last_interaction_retrieved+1,
                                                                                       last_interaction_retrieved+1+query_size,
                                                                                       gene_pair[0], gene_pair[1]))
                number_of_interactions_no_id += 1
                parser.add_to_log("Number of interactions with no proteinPiana")

    # END OF for gene_pair in list_interactions:

    last_interaction_retrieved = last_interaction_retrieved + 1 + query_size   # last interaction retrieved (init to -1 so the first one we get is 0)
    interactions_to_retrieve = interactions_to_retrieve - query_size # how many interactions haven't been processed
    
# END OF while interactions_to_retrieve > 0

if verbose:
    sys.stderr.write( "All done! Number of ints added: %s. Number of ints without proteinPiana: %s. Num of ints low score: %s\n\n" %(
        number_of_interactions_added,
        number_of_interactions_no_id,
        number_of_interactions_low_score
        ))



## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(number_interactions_processed, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()
