"""
File        : taxonomy2piana.py
Author      : Daniel Jaeggi and Ramon Aragues
Creation    : 8.2003
Modified by : Javier Garcia (April 2007)
Contents    : script that fills up tables in database piana related to  external DB "taxonomy"
Called from : 

=======================================================================================================

This file parsers a taxonomy text file and inserts information into PianaDb

Command line option '--help' describes usage of this program

For more details on how to use it, read piana/README.populate_piana_db

"""

# taxonomy2piana.py: script that fills up tables in database piana related to  external DB "taxonomy"
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES


import sys

import re
import readline


import MySQLdb


from PianaDBaccess import *
from parserUtilities import *


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:


# Location of file containing external DB "taxonomy"
taxonomy_file = "../../../data/externalDBs/taxonomyDB/names.dmp"

compulsory_arguments_list = [ ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-version=",None,"version of the database to be inserted")]

optional_arguments_list = [("taxonomy-file=",taxonomy_file,"the name of the taxonomy input file"),
                           ("database-name=","ncbi_taxonomy","internal identifier name to this database (it must be unique in the database)"),
                           ("database-description=","NCBI Species information","Description of the database to be inserted"),
                           ("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)")]

# Create a parser object with the information of the parser

parser = ParserObject( script_name="taxonomy2piana.py",
                       script_description="This program fills up tables in database piana related to external DB 'taxonomy'",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


# Parse arguments:

arguments_dict = parser.parseArguments()



piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]


database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]

taxonomy_file = arguments_dict["taxonomy-file"]

source_db = arguments_dict["database-name"]

help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
log_file= arguments_dict["log-file"]


# --------
# --------
#  Main()               
# --------                               
# --------

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
# Table names are defined in piana/code/PianaDB/PianaGlobals.py
# If you don't known which tables are being used, the list can be empty: in this case, all tables will be locked
# While tables are locked, anyone else can access to them

tables_to_lock = [PianaGlobals.species_table]


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)


# Open file descriptors

taxonomy_fd = file(taxonomy_file,"r")

if log_file:
    log_file_fd = file(log_file, 'w')


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = source_db,
                                          databaseVersion = database_version,
                                          parsedFile = taxonomy_file.split("/")[-1],
                                          databaseDescription = database_description)
#
# Reading external DB "taxonomy" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from database dump file...\n")

for line in taxonomy_fd:

    # Parsing the line using the regular expression followed by taxonomy data
    line_fields = line.split("|")

    # line_fields is a tuple with (tax_id, tax_name, ?, tax_comment)
    
    if line_fields:
        
        tax_id = int( line_fields[0].strip() )
        tax_name = line_fields[1].strip().replace('"','').replace("/","").replace("\\","")
        tax_comment = line_fields[3].strip().replace('"','').replace("/","").replace("\\","")
        
        if verbose:
            sys.stderr.write( "tax_id is <%s> -- tax_name is <%s> -- tax comment is <%s>\n" %(tax_id, tax_name, tax_comment))

        if tax_id != "" and tax_name != "":
            piana_access.insert_species(tax_id= tax_id, tax_name= tax_name, tax_comment=tax_comment, source_db= source_db)
            parser.add_to_log("Number of species inserted")
            
        else:
            parser.add_to_log("Number of species not inserted")
        
    # END OF if line_fields:
# END OF for line in taxonomy_fd:


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
piana_access.close()
        
if verbose:
    sys.stderr.write("All done!\n%s\n" %(parser.get_log_string()))


if log_file:
    log_file_fd.write("%s" %(parser.get_log_string()))
    log_file_fd.close()


