"""
File        : parser_model.py
Author      : Javier Garcia
Creation    : April 2007
Contents    : This is a model of a parser file
Called from : 

=======================================================================================================

This file implements a model to develop a parser to be used in PIANA

The methods to integrate external databases into PIANA are stored in PianaDBaccess class (see documentation)

For more details on how to use it, read piana/README.populate_piana_db
"""

#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


## STEP 1: IMPORT NECESSARY MODULES
## Here you must import all external modules necessaries to the parser

import sys
import re
import readline
import MySQLdb
import time

from PianaDBaccess import *
from parserUtilities import *


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# In all parsers there must be some compulsory parameters, at least they have default values:
# Arguments related with database information:
#	- database-name
#	- database-version
#	- database-description
# Arguments related with database connection:
#	- piana-dbname
#	- piana-dbhost

# Example of arguments:

compulsory_arguments_list = [ ("input-file=",None,"file name of input file containing swissprot or TrEMBL data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log_file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser")]

# Create a parserObject with the parser information. This object is used to store parser parameters and some methods (parseArguments, print_help... (see documentation))

parser = parserObject( script_name="uniprot2piana.py",
                          script_description="This program fills up tables in database piana with information from swissprot or TrEMBL (no differenttreatment required)",
                          compulsory_arguments=compulsory_arguments_list,
                          optional_arguments=optional_arguments_list )

# Read parser parameters

arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]


# Start time counter
if time_control:
    initial_time = time.time()


# Define which tables will be used in the parser
# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
# Table names are defined in piana/code/PianaDB/PianaGlobals.py
# If you don't known which tables are being used, the list can be empty: in this case, all tables will be locked
# While tables are locked, anyone else can access to them

tables_to_lock = [PianaGlobals.protein_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.proteinSpecies_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.species_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.pdb_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.interPro_table,
                  PianaGlobals.unigene_table,
                  PianaGlobals.ensembl_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.proteinReactome_table,
                  PianaGlobals.proteinEC_table,
                  PianaGlobals.proteinCog_table,
                  PianaGlobals.proteinGo_table,
                  PianaGlobals.proteinDescription_table,
                  PianaGlobals.proteinKeyword_table,
                  PianaGlobals.proteinFunction_table,
                  PianaGlobals.proteinSubcellularLocation_table,
                  PianaGlobals.proteinDisease_table,
                  PianaGlobals.proteinMIM_table,
                  PianaGlobals.reactome_table]

		  
# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess
# This method has huge variations between parsers, but some methods must be in all of them
# In this part, those methods are explained
                  
################################
##         MAIN           ######
################################


# Open file descriptors:

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )


# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = sourceDB,
                                          databaseVersion = database_version,
                                          parsedFile = input_file.split("/")[-1],
                                          databaseDescription = database_description)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )



# PARSER ITSELF
# BUCLE
#
# If you want to use a log file, you only have to add, elsewhere you want in the parser bucle, the following method:

parser.add_to_log("Number of proteins inserted"). #This method is a counter, so you can counter the number an event is performed

# When you print the log file, all the counters will be printed with the label inserted






## LAST STEP: CLOSE DATABASE CONNECTION    ¡¡¡¡ IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables
piana_access.close()


if time_control:
    sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(protein_number, (time.time()-initial_time) ))

    
# Print the log file
if log_file:
    sys.log_file_fd.write(parser.get_log_string())
    log_file_fd.close()



