"""
File        : uniprot2piana.py
Author      : Ramon Aragues
Creation    : 16.3.2004
Modified    : Javier Garcia Garcia April 2007
Contents    : fills up tables in database piana with information from uniprot
Called from : 

=======================================================================================================

This file implements a program that fills up tables in database piana with information of uniprot databases

This parser uses biopython libraries and methods

Command line option '--help' describes usage of this program

For more details on how to use it, read piana/README.populate_piana_db
"""

# uniprot2piana.py: fills up tables in database piana with information from uniprot
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt
import re
import readline
import MySQLdb

import time

from Bio.SwissProt import SProt
from Bio import File

from PianaDBaccess import *
from parserUtilities import *


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values

# Define the arguments:

compulsory_arguments_list = [ ("input-file=",None,"file name of input file containing swissprot or TrEMBL data"),
                              ("piana-dbname=",None,"name of database piana to be used"),
                              ("piana-dbhost=",None,"name of host where database piana to be used is placed"),
                              ("mode=",None,"sets mode to be used by parser. Valid modes are: \"scratch\" (piana database is empty, create it from scratch) or \"tables\" (fill only tables indicated in tables_to_fill (see code)\""),
                              ("database-name=",None,"internal identifier name to this database (it must be unique in the database)"),
                              ("database-version=",None,"version of the database to be inserted"),
                              ("database-description=",None,"Description of the database to be inserted")]

optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interactions\"")]

# Define the parser information to print in the help:

parser = ParserObject( script_name="uniprot2piana.py",
                       script_description="This program fills up tables in database piana with information from swissprot or TrEMBL (no differenttreatment required)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )

# Read parser parameters


arguments_dict = parser.parseArguments()

input_file = arguments_dict["input-file"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
sourceDB= arguments_dict["database-name"]
database_version= arguments_dict["database-version"]
database_description= arguments_dict["database-description"]
piana_dbuser= arguments_dict["piana-dbuser"]
piana_dbpass= arguments_dict["piana-dbpass"]
database_information= arguments_dict["database-information"]
help= arguments_dict["help"]
verbose= arguments_dict["verbose"]
time_control= arguments_dict["time-control"]
log_file= arguments_dict["log-file"]
mode= arguments_dict["mode"]



# Start time counter
if time_control:
    initial_time = time.time()



# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess


# if using mode 'tables' these will be the tables that will be populated. In mode 'scrath' this list is ignored
#tables_to_fill = [PianaGlobals.pfam_table, PianaGlobals.proteinExternalId_table]   # valid tables are those *_table in PianaGlobals
#tables_to_fill = [PianaGlobals.geneName_table]
#tables_to_fill = [PianaGlobals.proteinDescription_table,
#                  PianaGlobals.proteinKeyword_table,
#                  PianaGlobals.proteinMIM_table,
#                  PianaGlobals.proteinDisease_table,
#                  PianaGlobals.proteinFunction_table]
tables_to_fill = [PianaGlobals.ensembl_table,
                  PianaGlobals.unigene_table]

# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
# Table names are defined in piana/code/PianaDB/PianaGlobals.py
# If you don't known which tables are being used, the list can be empty: in this case, all tables will be locked
# While tables are locked, anyone else can access to them

tables_to_lock = [PianaGlobals.sequence_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.proteinPianaCounter_table,
                  PianaGlobals.uniprotEntry_table,
                  PianaGlobals.uniprotAcc_table,
                  PianaGlobals.species_table,
                  PianaGlobals.geneName_table,
                  PianaGlobals.pdb_table,
                  PianaGlobals.gi_table,
                  PianaGlobals.interPro_table,
                  PianaGlobals.pfam_table,
                  PianaGlobals.unigene_table,
                  PianaGlobals.ensembl_table,
                  PianaGlobals.accessionNumber_table,
                  PianaGlobals.proteinReactome_table,
                  PianaGlobals.proteinEC_table,
                  PianaGlobals.proteinCog_table,
                  PianaGlobals.proteinGo_table,
                  PianaGlobals.proteinDescription_table,
                  PianaGlobals.proteinKeyword_table,
                  PianaGlobals.proteinFunction_table,
                  PianaGlobals.proteinSubcellularLocation_table,
                  PianaGlobals.proteinDisease_table,
                  PianaGlobals.proteinMIM_table,
                  PianaGlobals.reactome_table,
                  PianaGlobals.proteinExternalId_table,
                  PianaGlobals.externalDatabaseType_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.validProteinIds_table]

                  


################################
##         MAIN           ######
################################

# Open file descriptors:

input_file_fd = file(input_file, 'r')

if log_file:
    log_file_fd = file(log_file, 'w')

    


# Initializing connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost= piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass )

if mode == "scratch":
    # starting from scratch
    
    # Introduce database info into PIANA database
    piana_access.insert_new_external_database(databaseName = sourceDB,
                                              databaseVersion = database_version,
                                              parsedFile = input_file.split("/")[-1],
                                              databaseDescription = database_description,
                                              databaseInformation = database_information)


    # Locking tables to be updated
    piana_access.lock_tables(table_list = tables_to_lock )
    
    # insert the new protein id types that are going to be referenced in this parser (default id types are already in the database)
    piana_access.insert_new_id_type(proteinTypeId="flybase", externalIdDescription="FlyBase identifier for the protein")

else:
    # only a few tables are going to be filled
    piana_access.lock_tables(table_list =  tables_to_lock)
    if PianaGlobals.proteinExternalId_table in tables_to_fill:
        piana_access.insert_new_id_type(proteinTypeId="flybase", externalIdDescription="FlyBase identifier for the protein")
    

# Initializing bipoython swissprot parser
uniprot_parser = SProt.RecordParser()
uniprot_iterator = SProt.Iterator(handle= input_file_fd, parser= uniprot_parser)

# reading first record
uniprot_record = uniprot_iterator.next()


protein_number=0

# while record read is not None, parse the record and insert data into piana
while uniprot_record is not None:

    if verbose:
        sys.stderr.write( "===================================\n")
        sys.stderr.write( "            NEW ENTRY\n")
        sys.stderr.write( "===================================\n")
        
    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if protein_number%20000==0:
            sys.stderr.write("%s proteins done in %s seconds\n" %(protein_number,time.time()-initial_time))

    # since a protein is defined by its sequence and tax_id, first thing
    # to clarity is whether we have a tax_id for this protein or not
    #  Special cases:  - can we have a sequence assigned to two species?
    #                  -
    # If there is no taxonomy_id or there is more than one, we use a dummy
    # tax_id for this protein... which means that it will correspond to
    # whichever tax_id
    
    list_tax_id = [int(one_tax_id) for one_tax_id in uniprot_record.taxonomy_id]

    if not len(list_tax_id) == 1:
        # checking that the number of tax id is just one
        
        if len(list_tax_id) == 0:
            #raise ValueError("No tax id value assigned to this protein: %s" %(uniprot_record.entry_name))
	    sys.stderr.write("No tax id value assigned to this protein: %s" %(uniprot_record.entry_name))
            uniprot_record = uniprot_iterator.next()
	    continue
        elif len(list_tax_id) > 1:
            #raise ValueError("Many tax id values assigned to this protein: %s" %(uniprot_record.entry_name))
	    sys.stderr.write("Many tax id values assigned to this protein: %s" %(uniprot_record.entry_name))
            uniprot_record = uniprot_iterator.next()
	    continue
    
    tax_id_value= list_tax_id[0]
    
    if verbose:
        sys.stderr.write("parsing mode is: %s \n" %mode )
        sys.stderr.write("sourceDB is: %s\n " %sourceDB )
        sys.stderr.write("protein sequence is: %s\n" %uniprot_record.sequence )   
        sys.stderr.write("tax id is: %s\n" %tax_id_value )         

    # --- table protein ---
    if mode == "tables":
        # since we are only updating certain tables, get the protein piana that should already exist in the database for this sequence

        
        list_inserted_proteinPiana = piana_access.get_list_proteinPiana(protein_id=uniprot_record.sequence,
                                                                        id_type= "sequence",
                                                                        tax_id_value = tax_id_value,
                                                                        source_db_info="no")


        num_of_pps = len(list_inserted_proteinPiana)
        
        # check that only one proteinPiana was found for the sequence (as it should be)
        if num_of_pps == 0:
            raise ValueError("trying to update tables for a (sequence, tax_id) that is not in the database")
        
        elif num_of_pps > 1:
            raise ValueError("trying to update tables for a (sequence, tax_id) that has multiple proteinPianas associated")
        else:
            # we will work with that proteinPiana from now on
            inserted_proteinPiana = list_inserted_proteinPiana[0]

    elif mode == "scratch":
        # inserting sequence  for current protein into piana. insert_protein fills all fields of table protein
        # if inserting a sequence that was already in the database, it will update fields without creating a new proteinPiana

        # TO CHECK!!! What's the use of updating fields that are automatically generated from the sequence... if the sequence
        # is the same, then the fields will be the same... then, why updating instead of "doing nothing"? This has to be changed
        # in PianaDBaccess.insert_protein if found to be a reasonable argument... maybe updating is good when we didn't calculate
        # correctly the MW, IP or MD5... 
        if uniprot_record.sequence != "":
            inserted_proteinPiana = piana_access.insert_protein(proteinSequence_value= uniprot_record.sequence,
                                                                tax_id_value = tax_id_value,
                                                                sourceDBID=sourceDB)
            parser.add_to_log("Number of proteins in the database")


    # --- table uniprotEntry ---
    if mode == "scratch" or (PianaGlobals.uniprotEntry_table in tables_to_fill and mode =="tables"):
        if verbose:
            sys.stderr.write( "uniprotEntry is: %s \n" %uniprot_record.entry_name ) 

        # inserting uniprotEntry  for current protein
        if uniprot_record.entry_name != "":
            piana_access.insert_uniprotEntry(uniprotEntry= uniprot_record.entry_name, proteinPiana_value= inserted_proteinPiana, 
                                               sourceDBID = sourceDB )
            parser.add_to_log("Number of uniprotEntry codes inserted")

    # --- table uniprotAcc ---
    if mode == "scratch" or (PianaGlobals.uniprotAcc_table in tables_to_fill and mode =="tables"):
   
        if verbose:
            sys.stderr.write( "uniprotAcc numbers are: %s \n"  %uniprot_record.accessions)
            
            
        isPrimary= 1    # isPrimary will be true only for the first accession number found
    
        for uniprot_accession in uniprot_record.accessions:

            if uniprot_accession != "":
                piana_access.insert_uniprotAcc(uniprotAcc_value= uniprot_accession,
                                               proteinPiana_value= inserted_proteinPiana,
                                               sourceDBID = sourceDB,
                                               isPrimary_value= isPrimary )
                isPrimary= 0
                parser.add_to_log("Number of uniprotEntry codes inserted")

    # --- table geneName ---
    if mode == "scratch" or (PianaGlobals.geneName_table in tables_to_fill and mode =="tables"):
   
        if verbose:
            sys.stderr.write( "uniprot_record.gene_name is: %s \n"  %uniprot_record.gene_name )

        # parse_string_field_value gets pairs [field_name, field_value] from a string where
        # format looks something like "field_name=field_value;field_name=field_value...."
        gene_name_fields = utilities.parse_string_field_value(input_string= uniprot_record.gene_name,
                                                              separator_field_value="=",
                                                              global_separators=[';','\n','"']) 
        
        # inserting geneName code for current protein
        if gene_name_fields:

            for gene_name_field in gene_name_fields:

                if gene_name_field[0] == "Name":
                    if verbose:
                        sys.stderr.write( "individual gene name (primary) is: <%s>" %(gene_name_field[1].strip().strip("'").strip('"')) )

                    geneName_value= gene_name_field[1].strip().strip("'").strip('"')
                    
                    if geneName_value and not re.match('^[0-9]+$', geneName_value):
                        # avoid inserting empty values or numbers
                        piana_access.insert_geneName_code(geneName_code_value= geneName_value ,
                                                          proteinPiana_value= inserted_proteinPiana,
                                                          geneName_source_value= sourceDB,
                                                          isPrimary_value=1 )
                        parser.add_to_log("Number of geneName codes inserted")
                    
                elif gene_name_field[0] == "Synonyms" or gene_name_field[0] == "OrderedLocusNames" or gene_name_field[0] == "ORFNames":
                    gene_synonyms = gene_name_field[1].split(",")
                    
                    for gene_synonim in gene_synonyms:

                        gene_synomim_value = gene_synonim.strip().strip("'").strip('"')
                        if gene_synomim_value and not re.match('^[0-9]+$', gene_synomim_value ):
                            piana_access.insert_geneName_code(geneName_code_value= gene_synomim_value,
                                                              proteinPiana_value= inserted_proteinPiana,
                                                              geneName_source_value= sourceDB,
                                                              isPrimary_value=0 )
                            parser.add_to_log("Number of uniprotEntry codes inserted")
                    # END OF for gene_synonim in gene_name_field[1]:
                # END OF elif gene_name_field[0] == "Synonyms":

            # END OF for gene_name_field in gene_name_fields:
        # END OF  if gene_name_fields:

    # --- table species ---
    if mode == "scratch" or (PianaGlobals.species_table in tables_to_fill and mode =="tables"):

        if verbose:
            sys.stderr.write( "uniprot record taxonomy: %s \n" %uniprot_record.taxonomy_id)

        # inserting taxonomy code for current protein
       ##  if uniprot_record.taxonomy_id:

##             for one_taxonomy_id in uniprot_record.taxonomy_id:

##                 if one_taxonomy_id != "":
##                     piana_access.insert_protein_species(tax_id =  one_taxonomy_id,
##                                                         proteinPiana_value = inserted_proteinPiana,
##                                                         proteinSpeciesSource_value = sourceDB )
##                     parser.add_to_log("Number of protein species inserted")


    # --- table proteinDescription ---
    if mode == "scratch" or (PianaGlobals.proteinDescription_table in tables_to_fill and mode =="tables"):

        if verbose:      
            sys.stderr.write( "description is: %s \n"  %uniprot_record.description)

        # example of field: Corticotropin-lipotropin precursor (Pro-opiomelanocortin) (POMC) Contains: NPP; Melanotropin gamma (Gamma-MSH); 
 
        # inserting description of the protein
        if uniprot_record.description  != "":
            list_descriptions = uniprot_record.description.split(";")
            # do some preprocessing of the description
            for one_description in list_descriptions:

                piana_access.insert_protein_description(
                    description= one_description.replace('"'," ").replace("\n", " ").replace("'"," ").replace('\\', " ").strip(),
                    proteinPiana_value= inserted_proteinPiana,
                    proteinDescriptionSource_value= sourceDB )
                parser.add_to_log("Number of protein Descriptions inserted")

    # --- table proteinEC ---
    if mode == "scratch" or (PianaGlobals.proteinEC_table in tables_to_fill and mode =="tables"):

        # inserting EC code of the protein (EC code is found between parenthesis inside the description field, preceded by "EC")
        # example: synthase (EC 6.3.5.5); Aspartate carbamoyltransferase (EC 2.1.3.2)
        if uniprot_record.description  != "":
            enzymes = re.findall("\(EC \d+\.\d+\.\d+\.\d+\)", uniprot_record.description)

            for enzyme in enzymes:

                enzyme = re.sub("[(^\(EC )(\)$)]", "", enzyme).strip()

                if enzyme != "":
                    piana_access.insert_protein_ec(ec_id= enzyme,
                                                   proteinPiana_value= inserted_proteinPiana,
                                                   proteinECSource_value= sourceDB )
                    parser.add_to_log("Number of EC codes inserted")
            
    for comment in uniprot_record.comments:

        comment_fields = comment.split(":")

        # Bug: if the comment contained the character ":" all the content after the character was removed. To solve this bug:
        comment_fields[1] = ":".join(comment_fields[1:])

        # comment_fields[0] can be of type:
        #
        # ALLERGEN  Information relevant to allergenic proteins
        # ALTERNATIVE PRODUCTS Description of the existence of related protein sequence(s) produced by alternative splicing of the same gene
        #                      or by the use of alternative initiation codons
        # BIOTECHNOLOGY Description of the use of a specific protein in a biotechnological process
        # CATALYTIC ACTIVITY Description of the reaction(s) catalyzed by an enzyme [1]
        # CAUTION Warning about possible errors and/or grounds for confusion
        # COFACTOR Description of an enzyme cofactor
        # DATABASE Description of a cross-reference to a network database/resource for a specific protein; see 3.11.2
        # DEVELOPMENTAL STAGE Description of the developmentally-specific expression of a protein
        # DISEASE Description of the disease(s) associated with a deficiency of a protein
        # DOMAIN Description of the domain structure of a protein
        # ENZYME REGULATION Description of an enzyme regulatory mechanism
        # FUNCTION General description of the function(s) of a protein
        # INDUCTION Description of the compound(s) or condition(s) that stimulate the synthesis of a protein
        # MASS SPECTROMETRY Reports the exact molecular weight of a protein or part of a protein as determined by mass spectrometric methods
        # MISCELLANEOUS Any comment which does not belong to any of the other defined topics
        # PATHWAY Description of the metabolic pathway(s) with which a protein is associated
        # PHARMACEUTICAL Description of the use of a protein as a pharmaceutical drug
        # POLYMORPHISM Description of polymorphism(s)
        # PTM Description of a posttranslational modification
        # RNA EDITING Description of any type of RNA editing that leads to one or more amino acid changes
        # SIMILARITY Description of the similaritie(s) (sequence or structural) of a protein with other proteins
        # SUBCELLULAR LOCATION Description of the subcellular location of the mature protein
        # SUBUNIT Description of the quaternary structure of a protein
        # TISSUE SPECIFICITY Description of the tissue specificity of a protein


        if comment_fields[0] == "SUBCELLULAR LOCATION":
            
            # --- table proteinsubcellularLocation ---
            if mode == "scratch" or (PianaGlobals.proteinSubcellularLocation_table in tables_to_fill and mode =="tables"):

                if verbose:
                    sys.stderr.write( "comment of type %s is: %s \n" %(comment_fields[0], comment_fields[1]))
               
                piana_access.insert_protein_subcellularLocation(subcellularLocation = comment_fields[1].replace('"'," ").replace("\n", " ").strip(),
                                                                proteinPiana_value = inserted_proteinPiana,
                                                                proteinSubcellularLocationSource_value= sourceDB )
                parser.add_to_log("Number of Subcellular Locations inserted")
                
        elif comment_fields[0] == "FUNCTION":
            
            # --- table proteinFunction ---
            if mode == "scratch" or (PianaGlobals.proteinFunction_table in tables_to_fill and mode =="tables"):

                if verbose:
                    sys.stderr.write( "comment of type %s is: %s \n" %(comment_fields[0], comment_fields[1]))
               
                piana_access.insert_protein_function(function= comment_fields[1].replace('"'," ").replace("\n", " ").strip(),
                                                     proteinPiana_value= inserted_proteinPiana,
                                                     proteinFunctionSource_value= sourceDB )
                parser.add_to_log("Number of protein functions inserted")

        elif comment_fields[0] == "DISEASE":

            # --- table proteinDisease ---
            if mode == "scratch" or (PianaGlobals.proteinDisease_table in tables_to_fill and mode =="tables"):

                if verbose:
                    sys.stderr.write( "comment of type %s is: %s \n" %(comment_fields[0], comment_fields[1]))

                piana_access.insert_protein_disease(disease= comment_fields[1].replace('"'," ").replace("\n", " ").strip(),
                                                    proteinPiana_value= inserted_proteinPiana,
                                                    proteinDiseaseSource_value= sourceDB )
                parser.add_to_log("Number of protein diseases inserted")
                

            
    # END OF for comment in uniprot_record.comments:
    
    for cross_reference in uniprot_record.cross_references:

        # cross_reference[0] can be of type:
        #
        # EMBL  Nucleotide sequence database of EMBL/EBI (see 3.12.6)  Nucleic Acids Res. 32:D27-D30(2004); PMID: 14681351
        # Aarhus/Ghent-2DPAGE Human keratinocyte 2D gel protein database from Aarhus and Ghent universities FEBS Lett. 430:64-72(1998); PMID: 9678596
        # ANU-2DPAGE Australian National University 2-DE database Proteomics 1:1149-1161(2001); PMID: 11990509
        # COMPLUYEAST-2DPAGE 2-DE database at Universidad Complutense de Madrid J. Chromatogr. B. Biomed. Sci. Appl. 787:129-148(1993); PMID: 12659738
        # DictyBase Dictyostelium discoideum online informatics resource Nucleic Acids Res. 32:D332-D333(2004); PMID: 14681427
        # ECO2DBASE Escherichia coli gene-protein database (2D gel spots) (ECO2DBASE) Electrophoresis 20:2149-2159(1999); PMID: 9298644
        # EcoGene Escherichia coli K12 genome database (EcoGene) Nucleic Acids Res. 28:60-64(2000); PMID: 10592181
        # FlyBase Drosophila genome database (FlyBase) Nucleic Acids Res. 32:D418-D420(2004); PMID: 14681446
        # GeneDB_SPombe Schizosaccharomyces pombe GeneDB Nucleic Acids Res. 32:D339-D343(2004); PMID: 14681429
        # Genew Human gene nomenclature database (Genew) Nucleic Acids Res. 32:D255-D257(2004); PMID: 14681406
        # GermOnline GermOnline database Nucleic Acids Res. 32:D560-D567(2004); PMID: 14681481
        # GK Genome Knowledgebase (GK) http://www.genomeknowledge.org/
        # GlycoSuiteDB Database of glycan structures (GlycoSuiteDB) Nucleic Acids Res. 31:511-513(2003); PMID: 12520065
        # GO Gene Ontology (GO) database Nucl. Acids. Res. 32:D258-D261(2004); PMID: 14681407
        # Gramene Comparative mapping resource for grains (Gramene) Plant Physiol. 130:1606-1613(2002); PMID: 12481044
        # HAMAP Database of microbial protein families (HAMAP) Comput. Biol. Chem. 27:49-58(2003)
        # HIV HIV sequence database Kuiken C.L. et al., In: Theoretical Biology and Biophysics Group, Los Alamos National Laboratory, Los Alamos, NM.
        # HSC-2DPAGE Harefield hospital 2D gel protein databases (HSC-2DPAGE) Electrophoresis 18:471-479(1997); PMID: 9150926
        # HSSP Homology-derived secondary structure of proteins database (HSSP) Nucleic Acids Res. 27:244-247(1999); PMID: 9847191
        # InterPro Integrated resource of protein families, domains and functional sites (InterPro) Nucleic Acids Res. 31:315-318(2003); PMID: 12520011
        # Leproma Mycobacterium leprae genome database (Leproma) Lepr. Rev. 72:470-477(2001); PMID: 11826483
        # ListiList Listeria innocua and Listeria monocytogenes genomes database Microbiology 141:261-268(1995); PMID: 7704253
        # MaizeDB Maize Genetics/Genomics Database (MaizeGDB) Nucleic Acids Res. 32:D393-D397(2004); PMID: 14681441
        # Maize-2DPAGE Maize genome 2D Electrophoresis database (Maize-2DPAGE) Theor. Appl. Genet. 93:997-1005(1996)
        # MEROPS Peptidase database (MEROPS) Nucleic Acids Res. 32:D160-D164(2004); PMID: 14681384
        # MGD Mouse genome database (MGD) Nucleic Acids Res. 32:D476-D481(2004); PMID: 14681461
        # MIM Mendelian Inheritance in Man Database (MIM) Nucleic Acids Res. 30:52-55(2002); PMID: 11752252
        # MypuList Mycoplasma pulmonis genome database (MypuList) http://genolist.pasteur.fr/MypuList/
        # PDB 3D-macromolecular structure Protein Data Bank (PDB) Nucleic Acids Res. 31:489-491(2003); PMID: 12520059
        # Pfam Pfam protein domain database Nucleic Acids Res. 32:D138-141(2004); PMID: 14681378
        # PHCI-2DPAGE Parasite host cell interaction 2D-PAGE database http://www.gram.au.dk/
        # PhosSite Phosphorylation Site Database for prokaryotic proteins (In) Leslie M. (ed.); NetWatch. Science 294:1623-1623(2001) 
        # PhotoList Photorhabdus luminescens genome database Microbiology 141:261-268(1995); PMID: 7704253
        # PIR Protein sequence database of the Protein Information Resource (PIR) Nucleic Acids Res. 30:35-37(2002); PMID: 11752247
        # PIRSF PIR SuperFamilies of iProClass Nucleic Acids Res. 32:D112-D124(2004); PMID: 14681371
        # PMMA-2DPAGE Purkyne Military Medical Academy 2D-PAGE database http://www.pmma.pmfhk.cz/
        # PRINTS Protein Fingerprint database (PRINTS) Nucleic Acids Res. 30:239-241(2002); PMID: 11752304
        # ProDom ProDom protein domain database Brief. Bioinform. 3:246-251(2002); PMID: 12230033
        # PROSITE PROSITE protein domain and family database (see 3.12.7) Nucleic Acids Res. 32:D134-137(2004); PMID: 14681377
        # REBASE Restriction enzymes and methylases database (REBASE) Nucleic Acids Res. 31:418-420(2003); PMID: 12520038
        # RGD Rat Genome Database (RGD) Nucleic Acids Res. 30:125-128(2002); PMID: 11752273
        # SagaList Streptococcus agalactiae NEM316 / Serotype III genome database Microbiology 141:261-268(1995); PMID: 7704253
        # SGD Saccharomyces Genome Database (SGD) Nucleic Acids Res. 32:D311-D314(2004); PMID: 14681421
        # Siena-2DPAGE 2D-PAGE database from the Department of Molecular Biology, University of Siena, Italy http://www.bio-mol.unisi.it/
        # SMART Simple Modular Architecture Research Tool (SMART) Nucleic Acids Res. 32:D142-D144(2004); PMID: 14681379
        # StyGene Salmonella typhimurium LT2 genome database (StyGene)  
        # SubtiList Bacillus subtilis 168 genome database (SubtiList) Nucleic Acids Res. 30:62-65(2002); PMID: 11752255
        # SWISS-2DPAGE 2D-PAGE database from the Geneva University Hospital (SWISS-2DPAGE) Nucleic Acids Res. 28:286-288(2000); PMID: 10592248
        # TIGR The bacterial databases of 'The Institute of Genome Research' (TIGR) Nucleic Acid Res. 29:159-164(2001); PMID: 11125077
        # TIGRFAMs TIGR protein family database (TIGRFAMs) Nucleic Acids Res. 29:41-43(2001); PMID: 11125044
        # TRANSFAC Transcription factor database (TRANSFAC) Nucleic Acids Res. 31:374-378(2004); PMID: 12520026
        # TubercuList Mycobacterium tuberculosis H37Rv genome database (TubercuList) FEBS Lett. 452:7-10(1999); PMID: 10376668
        # WormPep Caenorhabditis elegans genome sequencing project protein database (WormPep) Genomics 46:200-216(1997); PMID: 9417907
        # ZFIN Zebrafish Information Network genome database (ZFIN) Nucleic Acids Res. 31:241-243(2003); PMID: 12519991


        #if cross_reference[0] == "PIR":
        #    # cross_reference[1] is PIR accession number
        #    # cross_reference[2] is PIR Entry name
        #    # both can be unknown (coded as "-")

        #    if cross_reference[1]  != "" and cross_reference[1] <> "-":
                
        #        # --- table pirAccession ---
        #        if mode == "scratch" or (PianaGlobals.pirAccession_table in tables_to_fill and mode =="tables"):

        #            if verbose:
        #                sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )
               
        #            piana_access.insert_pirAccession_code(pirAccession_code_value= cross_reference[1],
        #                                                  proteinPiana_value= inserted_proteinPiana,
        #                                                  pirAccession_source_value= sourceDB)
                
        #    if cross_reference[2]  != "" and cross_reference[2] <> "-":
                
        #        # --- table pirEntry ---
        #        if mode == "scratch" or (PianaGlobals.pirEntry_table in tables_to_fill and mode =="tables"):

        #            if verbose:
        #                sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )
                
        #            piana_access.insert_pirEntry_code(pirEntry_code_value= cross_reference[2],
        #                                              proteinPiana_value= inserted_proteinPiana,
        #                                              pirEntry_source_value= sourceDB)
                
        if cross_reference[0] == "EMBL":
            # cross_reference[1] is EMBL/GenBank/DDBJ accession number
            # cross_reference[2] is EMBL/GenBank/DDBJ Protein_ID             Not taken into account
            # cross_reference[3] is Status_Identifier                        Not taken into account
            # cross_reference[4] is Moleculte Type (Genomic_DNA, mRNA...)    Not taken into account
            # both can be unknown (coded as "-")

            if cross_reference[1] != "" and cross_reference[1] <> "-":


                # --- table accessionNumber ---

                if mode == "scratch" or (PianaGlobals.accessionNumber_table in tables_to_fill and mode == "tables" ):

                    # inserting accession Number for current protein
                    piana_access.insert_accessionNumber(accessionNumber_value= cross_reference[1], proteinPiana_value= inserted_proteinPiana,
                                                        accessionNumber_source_value = sourceDB )
                    parser.add_to_log("Number of accessionNumber inserted")
                # END OF if mode...
            # END OF if cross_reference...

        # TO DO!!! Add tables and code to populate other external references...
        elif cross_reference[0] == "GO":
            pass
                    
        elif cross_reference[0] == "PROSITE":
            pass
        elif cross_reference[0] == "TRANSFAC":
            pass
        elif cross_reference[0] == "ProDom":
            pass
        elif cross_reference[0] == "FlyBase":
            # cross_reference[1] is the flybase Identifier
            # cross_reference[2] is the gene name used in flybase

            if mode == "scratch" or (PianaGlobals.proteinExternalId_table in tables_to_fill and mode == "tables"):

                if verbose:
                    sys.stderr.write( "cross reference to database %s is: %s   \n" %(cross_reference[0], cross_reference[1:]) )

                piana_access.insert_external_id_code(external_id_type = "flybase",
                                                     external_id_value = cross_reference[1],
                                                     proteinPiana_value = inserted_proteinPiana,
                                                     sourceDBID = sourceDB )
                parser.add_to_log("FlyBase reference inserted")
            
        elif cross_reference[0] == "Reactome":
            # cross_reference[1] is the Reactome Identifier
            # cross_reference[2] is the name of the pathway

            if mode == "scratch" or (PianaGlobals.reactome_table in tables_to_fill and mode == "tables"):

                if verbose:
                    sys.stderr.write( "cross reference to database %s is: %s   \n" %(cross_reference[0], cross_reference[1:]) )

                piana_access.insert_protein_reactome(reactome_id_value = cross_reference[1].strip(),
                                                     proteinPiana_value = inserted_proteinPiana,
                                                     reactome_pathwayname_value = cross_reference[2].strip(),
                                                     proteinReactomeSource_value = sourceDB )
                parser.add_to_log("Reactome reference inserted")
                
        elif cross_reference[0] == "PDB":
            # cross_reference[1] is the PDB code, entry name
            # cross_reference[2] is the structure detemination method (X-ray, NRM, EM...)
            # cross_reference[3] is the chain and range of which the structure has been determined (it can be "-" if unknown)

            if cross_reference[1] != "" and cross_reference[1] != "-":

                # --- table pdb ---
                if mode == "scratch" or (PianaGlobals.pdb_table in tables_to_fill and mode == "tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )

                    # Parse the chain and residues range (residues range is not very important here, but I think I can need this information
                    # in the future...
                    if cross_reference[3] == "-":
                        actual_chain = ""
                        fragment_range = None
                        piana_access.insert_pdb_code(pdb_code_value = cross_reference[1],
                                                     proteinPiana_value = inserted_proteinPiana,
                                                     chain_value = actual_chain,
                                                     range_value = fragment_range,
                                                     pdb_source_value = sourceDB )
                        parser.add_to_log("Number of pdb references inserted")
                    else:
                        splited = re.split(",",cross_reference[3])

                        for fragment in splited:
                            splited_fragment = re.split("=",fragment.strip())
                            splited_fragment[0].strip()
                            fragment_range = splited_fragment[1].strip()
                            #sys.stderr.write( "Fragment %s  \n" %fragment_range )

                            if fragment_range == "-":
                                fragment_range = None

                            if splited_fragment[0] == '@':
                                chain_value = ""
                            else:
                                splited_chain = re.split("\/",splited_fragment[0])
                                
                                for actual_chain in splited_chain:
                                    piana_access.insert_pdb_code(pdb_code_value = cross_reference[1],
                                                                 proteinPiana_value = inserted_proteinPiana,
                                                                 chain_value = actual_chain,
                                                                 range_value = fragment_range,
                                                                 pdb_source_value = sourceDB )
                                    parser.add_to_log("Number of pdb references inserted")
                                    
                                # END OF for actual_chain
                        # END OF for fragment in splited
                    # END OF if cross_reference[3]
            # END OF if cross_reference[1]
            

        elif cross_reference[0] == "UniGene":
            # cross_reference[1] is the Unique Identifier for Unigene Database
            # cross_reference[2] not used. Always contains "-"

            if cross_reference[1] != "" and cross_reference[1] <> "-":

                # --- table unigene ---
                if mode == "scratch" or (PianaGlobals.unigene_table in tables_to_fill and mode == "tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )

                    piana_access.insert_unigene_code(unigene_code_value = cross_reference[1],
                                                     proteinPiana_value = inserted_proteinPiana,
                                                     unigene_source_value = sourceDB )
                    parser.add_to_log("Number of unigene references inserted")
                
        elif cross_reference[0] == "Ensembl":
            # cross_reference[1] is Ensembl Unique Identifier for the gene
            # cross_reference[2] is the species of origin (Not taken into account in this parser)

            if cross_reference[1] != "" and cross_reference[1] <> "-":

                # --- table ensembl ---
                if mode == "scratch" or (PianaGlobals.ensembl_table in tables_to_fill and mode == "tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )

                    piana_access.insert_ensembl_code(ensembl_code_value = cross_reference[1],
                                                     proteinPiana_value = inserted_proteinPiana,
                                                     ensembl_source_value = sourceDB )
                    parser.add_to_log("Number of ensembl references inserted")
            
        elif cross_reference[0] == "MIM":
            # cross_reference[1] is MIM ID number
            # cross_reference[2] is "Gene" or "Phenotype" (Not taken into account now)

            if cross_reference[1] != "" and cross_reference[1] <> "-":

                # --- table proteinMIM ---
                if mode == "scratch" or (PianaGlobals.proteinMIM_table in tables_to_fill and mode == "tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )

                    piana_access.insert_protein_mim(mimID_value= cross_reference[1],
                                                    proteinPiana_value = inserted_proteinPiana,
                                                    proteinMIMSource_value= sourceDB )
                    parser.add_to_log("Number of mim references inserted")
                    

        elif cross_reference[0] == "InterPro":

            if cross_reference[1]  != "" and cross_reference[1] <> "-":
                
                # --- table interPro ---
                if mode == "scratch" or (PianaGlobals.interPro_table in tables_to_fill and mode =="tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )
                        
                    # interPro description is not necessarily there... make arrangements to avoid errors
                    if len(cross_reference) > 2 and cross_reference[2] != "" and cross_reference[2] <> "-":
                        interPro_description = cross_reference[2]
                    else:
                        interPro_description= ""
                        
                    piana_access.insert_interPro_code(interProID_code_value= cross_reference[1],
                                                      proteinPiana_value= inserted_proteinPiana,
                                                      interProDescription_value= interPro_description,
                                                      interPro_source_value= sourceDB )
                    parser.add_to_log("Number of interpro references inserted")
        
        elif cross_reference[0] == "Pfam":

            if cross_reference[1]  != "" and cross_reference[1] <> "-":
                
                # --- table pfam ---
                if mode == "scratch" or (PianaGlobals.pfam_table in tables_to_fill and mode =="tables"):

                    if verbose:
                        sys.stderr.write( "cross reference to database %s is: %s  \n" %(cross_reference[0], cross_reference[1:]) )
                        
                    # pfam description is not necessarily there... make arrangements to avoid errors
                    if len(cross_reference) > 2 and cross_reference[2] != "" and cross_reference[2] <> "-":
                        pfam_description = cross_reference[2]
                    else:
                        pfam_description= ""
                                                
                    piana_access.insert_pfam_code(pfamID_code_value= cross_reference[1],
                                                  proteinPiana_value= inserted_proteinPiana,
                                                  pfamDescription_value= pfam_description,
                                                  pfam_source_value= sourceDB )
                    parser.add_to_log("Number of pfam references inserted")

    
    # END OF for cross_reference in uniprot_record.cross_references:
    
    # --- table proteinKeyword ---
    if mode == "scratch" or (PianaGlobals.proteinKeyword_table in tables_to_fill and mode =="tables"):
        
        # inserting keywords into piana
        for keyword in uniprot_record.keywords:

            # The KW (KeyWord) lines provide information that can be
            # used to generate indexes of the sequence entries based on
            # functional, structural, or other categories.

            if keyword != "":
                piana_access.insert_protein_keyword(keyword= keyword,
                                                    proteinPiana_value= inserted_proteinPiana,
                                                    proteinKeywordSource_value= sourceDB )
                parser.add_to_log("Number of proteinKeywords inserted")

        # END OF for keyword in uniprot_record.keywords:
            
    for feature in uniprot_record.features:

        # The FT (Feature Table) lines provide a precise but simple
        # means for the annotation of the sequence data. The table
        # describes regions or sites of interest in the sequence. In
        # general the feature table lists posttranslational
        # modifications, binding sites, enzyme active sites, local
        # secondary structure or other characteristics reported in
        # the cited references. Sequence conflicts between
        # references are also included in the feature table.
        if mode == "scratch":

            # TO DO!!! Do something about this field!
            
            if verbose:
                sys.stderr.write( "feature of type %s is: %s  \n" %(feature[0] , feature[1:]))
            
    # END OF for feature in uniprot_record.features:
        
    for reference in uniprot_record.references:

        if mode == "scratch":
            # These lines comprise the literature citations. The citations indicate the sources from which the data has been abstracted.
            if verbose:
                sys.stderr.write("      reference positions: %s \n" %(reference.positions ))
                sys.stderr.write("      reference comments: %s \n" %(reference.comments   ) ) 
                sys.stderr.write("      reference references: %s \n" %(reference.references ))
                sys.stderr.write("      reference authors: %s \n" %(reference.authors   ))
                sys.stderr.write("      reference title: %s \n" %(reference.title    ) )
                sys.stderr.write("      reference location: %s \n" %(reference.location ))
                
    # END OF  for reference in uniprot_record.references:

    # reading next record

    protein_number += 1
    if verbose:
        sys.stderr.write("-- reading next record\n")
    uniprot_record = uniprot_iterator.next()
    
# END OF while uniprot_record is not None


## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("Total time: %s proteins done in %s seconds\n" %(protein_number, (time.time()-initial_time) ))

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()



