"""
 File       : piana.py
 Author     : R. Aragues
 Creation   : 12.1.2004
 Contents   : command line interface to PIANA (batch or interactive execution available)
 Called from: command line

=======================================================================================================
This program is the command line interface to PIANA (batch or interactive execution available)


To learn more about what you can do with piana.py type piana.py --help, read conf_files/general_template.piana_conf

You've got some examples on how to use piana.py on piana/README.piana_examples

You should also read piana/README.piana_tutorial

If you encounter problems running PIANA, please read piana/README.piana_requirements and README.piana_installation

"""

# piana.py: command line interface to PIANA tool
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import sys
import getopt
import time
import re
import readline
import cPickle
from sets import *

from ndict import seqdict # implements ordered dictionaries
                          # package distributed within PIANA but
                          # developed by someone else
                          # read piana/code/utilities/seqdict/seqdict_webpage.txt
                          # for more info on this package

from PianaApi import *


# VERBOSE VARIABLES: used to print out information about the execution process
time_control = 0
verbose = 0
verbose_graph_builder = 0
verbose_protein = 0
verbose_command_sequence = 1
verbose_specBuild = 1
verbose_parse_conf = 0
verbose_build = 0

# -------------------------------------
# GLOBAL VARIABLES USED BY THIS PROGRAM
# -------------------------------------

# This ordered dictionary contains execution commands that can be given to piana, their description and which arguments does it need
# in this ordered dictionary, first you have to give the keys (in the order you want) and then the dictionary
#   ATTENTION: respect the order of the commands: it must be the same in both list and dictionary!!!

valid_execution_commands= seqdict(

    [
     "create-report",
     "reset-network",
     "save-network",
     "load-network",
     "add-protein",
     "add-proteins-file",
     "add-interactions-file",
     "species-network",
     "database-method-network",
     "print-table",
     "print-table-db-intersection",
     "print-network",
     "print-network-db-intersection",
     "print-all-prots-info",
     "print-root-prots-info",
     "print-connect-prots-info",
     "classify-network-proteins",
     "protein-code-2-protein-code",
     "protein-code-2-all-info",
     "create-random-files",
     "print-parameters",
     "expand-interactions",
     "find-shortest-route",
     "find-distance-group",
     "match-proteins-to-pathways",
     "match-proteins-to-gos",
     "match-proteins-to-spots",
     "cluster-by-go-terms",
     "calculate-imotifs",
     "exit"],
    {
    "create-report":                  [ "tells PIANA to produce an HTML report with links to results files", None],
    "reset-network":                  [ "start a new network  (doesn't change parameters, use option 'modif' for that)", None],
    "save-network":                   [ "save current network in a disc file. You'll be able to load it in the future", None],
    "load-network":                   [ "loads a network in memory from a disc file (overwrites current network)\n", None],
    "add-protein":                    [ "add a protein (and its interactions) to current network", None],
    "add-proteins-file":              [ "add a list of proteins (and their interactions) from an text file to current network", None],
    "add-interactions-file":          [ "add a list of interactions from an text file to current network", None],
    "species-network":                [ "build a network for all proteins in a given species", None],
    "database-method-network":        [ "build a network for all protein interactions in a given database and/or a given method\n", None],
    "print-table":                    [ "print table with all interactions in the network", None],
    "print-table-db-intersection":    [ "print table using only interactions that appear at the same time in several databases\n", None],
    "print-network":                  [ "print network in DOT format", None],
    "print-network-db-intersection":  [ "print network  in DOT format using only interactions that appear at the same time in several databases\n", None],
    "print-all-prots-info":           [ "prints information about all proteins in network", None],
    "print-root-prots-info":          [ "prints information about root proteins in network", None],
    "print-connect-prots-info":       [ "prints information about proteins that connect root_nodes", None],
    "classify-network-proteins":      [ "prints all proteins with a label indicating whether they are root, linker or partner", None],
    "protein-code-2-protein-code":    [ "transforms codes from input-file to output_proteins_code (doesn't create a network)", None],
    "protein-code-2-all-info":        [ "gets information of codes from input-file (doesn't create a network)\n", None],
    "create-random-files":            [ "creates files that contain random proteins (doesn't create a network)\n", None],
    "print-parameters":               [ "print parameters used to create current network", None],
    "expand-interactions":            [ "predicts interactions of proteins in the network using expansion", None],
    "find-shortest-route":            [ "prints the route between two given proteins (ie. intermediate proteins)\n", None],
    "find-distance-group":            [ "prints proteins that are at distance N of a query protein\n", None],
    "match-proteins-to-pathways":     [ "finds which pathways appear in the current network\n", None],
    "match-proteins-to-gos":          [ "finds the frequency of the GO terms appearing in the current network\n", None],
    "match-proteins-to-spots":        [ "identifies spots in a 2D gel by matching MW and/or IP to proteins in the network\n", None],
    "cluster-by-go-terms":            [ "clusters the protein interaction network using go terms\n", None],
    "calculate-imotifs":              [ "calculates iMotifs for network proteins", None],
    "exit":                           [ "exit", None]
})

"""
OPTIONS CURRENTLY DE-ACTIVATED:

     "find-cure-for",
     "modify-parameters",
     "filter",

    "find-cure-for":                  [ "finds a cure for whatever illness you are interested in", None],
    "modify-parameters":              [ "changes current parameters (depth, code types, ...) (!!!NOT WORKING!!!)\n", None],
    "filter":                         [ "filters proteins from the network (!!!NOT WORKING!!!)", None],
"""
    
# .................................
# Index of argument strings used to set values in commands of piana configuration file
# .................................

# arguments used in piana configuration file
arg_protein_name      = "protein-name"
arg_protein_a_name    = "protein-a-name"
arg_protein_b_name    = "protein-b-name"
arg_id_type      = "id-type"
arg_file_name         = "file-name"
arg_species_name      = "species-name"
arg_piana_db          = "piana-db"
arg_tax_id            = "tax-id"
arg_output_target     = "output-target"
arg_print_mode        = "print-mode"
arg_format_mode       = "format-mode"
arg_database_name     = "database-name"
arg_method_name       = "method-name"
arg_list_dbs          = "list-dbs"
arg_output_mode       = "output-mode"
arg_clustering_steps  = "clustering-steps"
arg_score_threshold   = "score-threshold"
arg_function_name     = "function-name"
arg_clustering_level  = "clustering-level"
arg_ranked_parameters_file  = "ranked-parameters-file"
arg_expansion_type    = "expansion-type"
arg_expansion_threshold = "expansion-threshold"
arg_expansion_nodes   = "expansion-nodes"
arg_exp_output_mode   = "exp-output-mode"
arg_spots_file_name   = "spots-file-name"
arg_list_mw_error     = "list-mw-error"
arg_list_ip_error     = "list-ip-error"
arg_save_mode         = "save-mode"
arg_disc_name         = "disc-name"
arg_remove_redundant  = "remove-redundant"
arg_distance          = "distance"
arg_info              = "info"
arg_term_type         = "term-type"
arg_sim_mode          = "sim-mode"
arg_level_threshold   = "level-threshold"
arg_distance_threshold= "distance-threshold"
arg_rep_term          = "rep-term"
arg_print_id          = "print-id"
arg_similarity_mode   = "similarity-mode"
arg_minimum_score     = "min-score"
arg_imotif_mode       = "imotif-mode"
arg_num_ints_thres    = "num-ints-thres"
arg_common_ints_thres = "common-ints-thres"
arg_global_map        = "global-map"
arg_pathways_dir      = "pathways-dir"
arg_force_expression  = "force-expression"
arg_go_dbname         = "go-dbname" 
arg_go_dbhost         = "go-dbhost" 
arg_go_dbuser         = "go-dbuser" 
arg_go_dbpass         = "go-dbpass" 
arg_num_files         = "num-files" 
arg_num_prots         = "num-prots" 
arg_force_ints        = "force-ints" 
arg_files_prefix      = "files-prefix" 
arg_check_special     = "check-special" 
arg_check_expression  = "check-expression" 

# .................................
# Asking for input from the user: define which sentences will be used
# .................................

string_get_id_type      = "Write protein code type (unientry, uniacc, geneName, gi, pdb.chain, sequence or md5): "
string_get_file_name         = "Write input file name with one protein per line: "
string_get_int_file_name     = "Write input file name with one interaction per line: "
string_get_depth             = "Write depth of network: "
string_get_hub_threshold     = "Write maximum number of interactions per protein (0 for ignoring thresholds): "
string_get_protein           = "Write protein name: "
string_get_net_species_name  = "Write species name: (eg. blank (if using tax id) or human, yeast, fruit fly, ...) "
string_get_net_tax_id        = "Write taxonomy id: (eg. blank (if using species name) or 9606, 4932, 7227, ...) "
string_get_species_name      = "Write species name: (eg. all, human, yeast, saccharomyces cerevisiae, ...) "
string_get_piana_dbname      = "Write the name of piana database (eg. pianaDB_limited): "
string_get_piana_dbhost      = "Write the name of piana host (eg. localhost): "
string_get_piana_dbuser      = "Write the name of piana user: "
string_get_piana_dbpass      = "Write the name of piana password: "
string_get_results_prefix    = "Write the prefix all result files will have: "
string_get_results_dir       = "Write the directory where results will be saved (must end with '/'): "
string_get_results_directory = "Write directory where result files will be placed: "
string_get_output_name       = "Write file name where output will be saved (screen prints to stdout): "
string_get_print_mode        = "Write which proteins are allowed to appear in output (all, all_root or only_root): "
string_get_format_mode       = "Write format desired for output files (txt or html for tables: dot or sif for network): "
string_get_output_mode       = "Write output mode: compact (one line info) or extended (all information): "
string_get_int_dbname        = "Write an interaction database name  (dip, string, ori, posas, expansion, pibase or mips) to be included: (end to exit) "
string_get_dbname            = "Choose an interaction database name  (dip, string, ori, posas, expansion, pibase or mips): "
string_get_methodname        = "Choose an interaction method name  (y2h, pibase_dbalig75c0, ): "
string_get_expansion_type    = "Choose expansion type you want to apply (ec, cog, scop or interpro): "
string_get_expansion_threshold="Set the expansion threshold (max number of nodes sharing expansion type (0 means no threshold): "
string_get_nodes_to_expand   = "Write the nodes you want to predict interactions for: (all or root): "
string_get_exp_output_mode   = "Write if expansion predictions should be added to network or printed to stdout (add or print): "
string_get_clustering_steps  = "Set how many clustering steps you want to perform: "
string_get_function_name     = "Set scoring function name that you want to use (quotient, lineal, ...): "
string_get_clustering_level  = "Set which clustering level to be printed:  "
string_get_spots_file        = "Write file name with spots ids and their MW and IP: "
string_get_save_file         = "Write file name where network will be saved: "
string_get_load_file         = "Write file name containing the network you want to load:  "
string_get_remove_redundant  = "Set if redundant codes should be removed from output (yes or no): "
string_get_distance          = "Set distance from protein query that will be used to retrieve distance-group proteins: "
string_get_info              = "Set which info will be printed next to the distance group proteins (blank, all, scop or cath): "
string_get_term_type         = "Set the GO term type you want to use for clustering (molecular_function, biological_process or cellular_component): "
string_get_sim_mode          = "Set the mode used for calculating differences between clusters (min, max, random, average): "
string_get_go_score_threshold= "Set the score threshold for considering two clusters similar (eg. 0.2): "
string_get_level_threshold   = "Set the lowest level of the go term in the cluster allowed for continuing clustering (eg. 2): "
string_get_distance_threshold= "Set the maximum distance allowed between two proteins in order to be clustered (eg. 3): "
string_get_rep_term          = "Set which of the GO terms in the cluster will be used for printing, max level or min level (min or max): "
string_get_print_id          = "Sets if clusters will be labeled in the output with and id or just the name (yes or no): "
string_get_protein_name      = "Set the protein name: "
string_get_pathways_dir      = "Write the path for the directory that holds the pathway files (one protein per line): "
string_get_go_dbname         = "Write the name of your GO external database" 
string_get_go_dbhost         = "Write the host of your GO external database" 
string_get_go_dbuser         = "Write the username of your GO external database" 
string_get_go_dbpass         = "Write the password of your GO external database" 


# ----
# Setting default values for piana parameters (for those parameters that were not set through command line and for those
#                                              parameters that cannot be set throught command line in interactive mode)
# ----
default_output_mode = "compact"

# set here the molecular error bounds for which you want to perform the matching
default_molecular_error_bounds   = [0.0, 0.0025, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3]

# set here the isoelectric error bounds for which you want to perform the matching
default_isoelectric_error_bounds = [0.0, 0.0025, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3]


def usage():
    """
    Describes the use of piana: 
    """
    
    sys.stdout.write( "---------------------------------------------------           \n" )
    sys.stdout.write( "  PIANA: Protein Interactions and Network Analysis            \n" )
    sys.stdout.write( "---------------------------------------------------         \n\n")
    
    sys.stdout.write( " There is an interactive mode, where commands to be           \n" )
    sys.stdout.write( " executed are requested to the user,  and a batch             \n" )
    sys.stdout.write( " mode where the sequence of commands is determined            \n" )
    sys.stdout.write( " in a piana configuration file.                             \n\n")
    sys.stdout.write( " Please, read piana/README.piana_tutorial for more            \n" )
    sys.stdout.write( " information on how to use this program                     \n\n" )
    sys.stdout.write( " Atention! The interactive mode does not implement            \n" )
    sys.stdout.write( "           many of the batch mode capabilities                \n" )
    
    sys.stdout.write( "Usage: python piana.py --exec-mode=exec-mode-value            \n" )
    sys.stdout.write( "       --memory-usage=memory-usage-value                      \n" )
    sys.stdout.write( "       --configuration-file=configuration-file-value          \n" )
    sys.stdout.write( "       --input-file=input-file-value                          \n" )
    sys.stdout.write( "       --input-id-type=input-id-type-value                    \n" )
    sys.stdout.write( "       --input-proteins-species=input-proteins-species-value  \n" )
    sys.stdout.write( "       --depth=depth-value                                    \n" )
    sys.stdout.write( "       --hub-threshold=hub-threshold-value                    \n" )
    sys.stdout.write( "       --output-id-type=output-id-type-value                  \n" )
    sys.stdout.write( "       --output-proteins-species=output-proteins-species-value\n" )
    sys.stdout.write( "       --piana-dbname=piana-dbname-value                      \n" )
    sys.stdout.write( "       --piana-dbhost=piana-dbhost-value                      \n" )
    sys.stdout.write( "       --piana-dbuser=piana-dbuser-value                      \n" )
    sys.stdout.write( "       --piana-dbpass=piana-dbpass-value                      \n" )
    sys.stdout.write( "       --results-prefix=results-prefix-value                  \n" )
    sys.stdout.write( "       --results-dir=results-dir-value                        \n" )
    sys.stdout.write( "       --file-over-expressed=file-over-expressed-value        \n" )
    sys.stdout.write( "       --file-infra-expressed=file-infra-expressed-value      \n" )
    sys.stdout.write( "       --expression-id-type=expression-id-type-value          \n" )
    sys.stdout.write( "       --special-proteins=special_proteins_string             \n" )
    sys.stdout.write( "          (see general_template.piana_conf for string format  \n" )
    sys.stdout.write( "       --help --verbose --print-reference-card                \n" )
    
    sys.stdout.write( "\n" )

    sys.stdout.write( "   --help               : prints this message and exits                 \n" )
    sys.stdout.write( "   --verbose            : prints process info to stdout                 \n" )
    sys.stdout.write( "   --print-reference-card : prints PIANA's reference card               \n" )
    sys.stdout.write( "   exec-mode-value      : interactive or batch (default: interactive)   \n" )
    sys.stdout.write( "                          batch mode requires a piana_configuration_file\n" )
    sys.stdout.write( "   configuration-file-value : file that sets piana parameters (both exec\n" )
    sys.stdout.write( "                              modes) and execution commands (only batch)\n" )
    sys.stdout.write( "                              -> conf_files/general_template.piana_conf \n" )
    sys.stdout.write( "                                 describes how to write this conf file  \n" )
    sys.stdout.write( "   memory-usage         : it can be \"high\" or \"low\". \"high\" should\n" )
    sys.stdout.write( "                          be used if the network is not big enough. If  \n" )
    sys.stdout.write( "                          the network is very large, use \"low\" insted.\n" )
    sys.stdout.write( "   input-file-value     : file name containing one protein per line     \n" )
    sys.stdout.write( "   depth-value          : depth of the network to be built              \n" ) 
    sys.stdout.write( "   hub-threshold-value  : max number of interactions allowed per protein\n" )
    sys.stdout.write( "                           - if num_ints_of_protein >= hub_threshold    \n" )
    sys.stdout.write( "                              --> no interactions added for the protein \n" )
    sys.stdout.write( "                           - if hub_threshold== 0, no thresholds applied\n" )
    sys.stdout.write( "   piana-dbname-value   : name of the PIANA database to be used         \n" )
    sys.stdout.write( "   piana-dbhost-value   : host where the PIANA database is placed       \n" )
    sys.stdout.write( "   piana-dbuser-value   : username accessing the database               \n" )
    sys.stdout.write( "                          (not required in most systems)                \n" )
    sys.stdout.write( "   piana-dbpass-value   : password of username accessing the database   \n" )
    sys.stdout.write( "                          (not required in most systems)                \n" )
    sys.stdout.write( "   results-prefix-value : prefix of results files, indicating a         \n" )
    sys.stdout.write( "                          representative name of your experiment        \n" )
    sys.stdout.write( "   results-dir-value    : directory where results will be saved         \n" )
    sys.stdout.write( "                          (must end with '/')                           \n" ) 
    sys.stdout.write( "\n file-over-expressed-value: file containing one protein per line      \n" )
    sys.stdout.write( "                          that has been found over expressed            \n" )  
    sys.stdout.write( "   file-infra-expressed-value: file containing one protein per line     \n" )
    sys.stdout.write( "                          that has been found under expressed           \n" ) 
    sys.stdout.write( "\n special-proteins: (see general_template.piana_conf to learn which    \n" )
    sys.stdout.write( "                      format has to be used for this parameter)         \n" )  
    sys.stdout.write( "\n spots-file-name-value: input file containing one spot per line       \n" )
    sys.stdout.write( "                          with its isoelect point and molecular weight  \n" )
    sys.stdout.write( "   input-proteins-species-value:species of proteins in input_file_name  \n" )
    sys.stdout.write( "                             (fixes species for ids with many species)  \n" )  
    sys.stdout.write( "   input-id-type-value : type of  protein database identifier in        \n" )
    sys.stdout.write( "                               input_file_name                          \n" )  
    sys.stdout.write( "   output-id-type-value: type of protein database identifier that \n" )
    sys.stdout.write( "                               will be used for output                  \n" ) 
    sys.stdout.write( "   output-proteins-species-value:species of proteins in the output      \n" )
    sys.stdout.write( "                             (fixes species for ids with many species)  \n" )  
    sys.stdout.write( "   expression-id-type-value: type of protein database identifier  \n" )
    sys.stdout.write( "                               that is used in the expression files     \n" )    
    sys.stdout.write( "\n  -> attention! geneNames are not reliable at all: we strongly suggest\n" )
    sys.stdout.write( "                that you find their equivalent codes in Uniprot entries,\n" )    
    sys.stdout.write( "\n              uniprot accession numbers or NCBI geneID                \n" )   
    sys.stdout.write( "-------------------------------------------------------------------------\n" )
         

def set_parameter_value(parameter_name= None, value=None):
    """
    checks if parameter name "parameter_name" has a related parameter, and in case it does, assigns "value" to it

    It is used by both parseArguments() and parse_configuration_file()

    Parameter assignment has been centralized here so it can be used by the two
    ways we have of setting parameters: command line and configuration file

    command line always has preference over configuration file: command line arguments overwrite those
    values that are set in the configuration file

    """
    global spots_file_name
    global input_file_name
    
    global input_id_type
    global expression_id_type
    global input_proteins_species
    global hub_threshold
    
    global output_id_type
    global output_proteins_species
    global alternative_id_types
    global list_keywords
    
    global file_over_expressed
    global file_infra_expressed
    global expression_id_type

    global special_proteins
    
    global depth
    global hub_threshold
    global use_self_ints
    global list_source_dbs
    global inverse_dbs
    global list_source_methods
    global inverse_methods

    global results_prefix
    global results_dir
    global exec_mode
    global memory_usage
    global configuration_file
    
    global piana_dbname
    global piana_dbhost
    global piana_dbuser
    global piana_dbpass

    global use_secondary_db
    global argument_print_reference_card
    global verbose
    global verbose_command_sequence


    # finding the variable associated to parameter_name and assigning value to it
    # (to make sure command line has preference over configuration file, before
    #  assigning the new value, make sure a value was not already assigned before (ie is  None) )
    #
    # When parsing the configuration file, many values will be "blank": a configuration file parameter value
    # will only be set if it is different from blank and it was not set through the command line.
    
    if parameter_name == "input-file":
        if input_file_name is None:
            if value != "blank":
                input_file_name =  value

    if parameter_name == "memory-usage":
        if value == "blank":
            memory_usage="high"
        if value != "blank":
            if value=="high":
                memory_usage="memory"
            elif value=="low":
                memory_usage="speed"
            else:
                raise ValueError("memory usage can only be \"high\" or \"low\"")
    
             
    elif parameter_name == "piana-dbname":
        if piana_dbname is None:
            if value != "blank":
                piana_dbname = value
        
    elif parameter_name == "piana-dbhost":
        if piana_dbhost is None:
            if value != "blank":
                piana_dbhost = value
        
    elif parameter_name == "piana-dbuser":
        if piana_dbuser is None:
            if value != "blank":
                piana_dbuser = value
        
    elif parameter_name == "piana-dbpass":
        if piana_dbpass is None:
            if value != "blank":
                piana_dbpass = value
            
    elif parameter_name == "use-secondary-db":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "no"
        if value != "blank":
            use_secondary_db = value
        else:
            use_secondary_db = "no"
        
    elif parameter_name == "input-id-type":
        if input_id_type is None:
            if value != "blank":
                input_id_type = value

    elif parameter_name == "expression-id-type":
        if expression_id_type is None:
            if value != "blank":
                expression_id_type = value
        
    elif parameter_name == "input-proteins-species":
        if input_proteins_species is None:
            if value != "blank":
                input_proteins_species = value
            else:
                input_proteins_species = "all" # ---> default value for input_proteins_species
        
    elif parameter_name == "output-id-type":
        if output_id_type is None:
            if value != "blank":
                output_id_type = value

    elif parameter_name == "output-proteins-species":
        if output_proteins_species is None:
            if value != "blank":
                output_proteins_species = value
            else:
                output_proteins_species = "all" # ---> default value for output_proteins_species
                
        
    elif parameter_name == "depth":
        if depth is None:
            if value != "blank":
                depth = int(value)
        
    elif parameter_name == "hub-threshold":
        if hub_threshold is None:
            # if the threshold was not set on the command line, then retrieve it from configuration file
            if value != "blank":
                hub_threshold = int(value)
            else:
                hub_threshold = 0 # default value for hub threshold is 0 (ie. no hub threshold used)
        
    elif parameter_name == "use-self-ints":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "yes"
        if value != "blank":
            use_self_ints = value
        else:
            use_self_ints = "yes" # ---> default value for use_self_ints
        
    elif parameter_name == "list-source-dbs":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "all"
        if value != "blank" and value != "all":
            list_source_dbs = value.split(":")
        else:
            list_source_dbs = "all" # ---> default value for list_source_dbs
        
    elif parameter_name == "inverse-dbs":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "no"
        if value != "blank":
            inverse_dbs = value
        else:
            inverse_dbs = "no" # ---> default value for inverse_dbs
            
    elif parameter_name == "list-source-methods":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "all"
        if value != "blank" and value != "all":
            list_source_methods = value.split(":")
        else:
            list_source_methods = "all"
        
    elif parameter_name == "inverse-methods":
        # this parameter can only be set through a configuration file: no need to check if it is None
        #   if configuration file doesn't have a value for it, set to "no"
        if value != "blank":
            inverse_methods = value
        else:
            inverse_methods = "no" # ---> default value for inverse_methods
        
    elif parameter_name == "spots-file-name":
        if spots_file_name is None:
            if value != "blank":
                spots_file_name = value
        
    elif parameter_name == "file-over-expressed":
        if file_over_expressed is None:
            if value != "blank":
                file_over_expressed = value
        
    elif parameter_name == "file-infra-expressed":
        if file_infra_expressed is None:
            if value != "blank":
                file_infra_expressed = value
        
    elif parameter_name == "expression-id-type":
        if expression_id_type is None:
	    if value != "blank":
		expression_id_type = value

    elif parameter_name == "special-proteins":
        if special_proteins is None:
            if value != "blank":
                special_proteins = value
        
    elif parameter_name == "results-prefix":
        if results_prefix is None:
            if value != "blank":
                results_prefix = value
        
    elif parameter_name == "results-dir":
        if results_dir is None:
            if value != "blank":
                results_dir = value
            else:
                results_dir = "./" # ---> default value for results_dir
        
    elif parameter_name == "exec-mode":
        if exec_mode is None:
            if value != "blank":
                exec_mode = value
        
        if exec_mode != "interactive" and exec_mode !="batch":
            sys.stderr.write("\nincorrect execution mode set: %s\n" %exec_mode)
            usage()
            sys.exit(2)
        
    elif parameter_name == "configuration-file":
        # this parameter will only be set through the command line... no need to check if it is None
        configuration_file = value
        
    elif parameter_name == "alternative-id-types":
        # this parameter will only be set through the configuration file... no need to check if it is None
        if value != "blank":
            alternative_id_types = value.split(":")
        else:
            alternative_id_types = [] # ---> default value for 
        
    elif parameter_name == "list-keywords":
        # this parameter will only be set through the configuration file... no need to check if it is None
        if value != "blank":
            list_keywords = value.split(":")
        else:
            list_keywords = [] # ---> default value for list_keywords
        
    elif parameter_name == "verbose":
        verbose = 1
        verbose_command_sequence = 1
        
    elif parameter_name == "print-reference-card":
        argument_print_reference_card = 1
        
    elif parameter_name == "help":
        # print help information and exit
        usage()
        sys.exit(2)
        
def parseArguments():
    """
    parses command line.

    piana.py --help for command line options
    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], "",                     
                                   ["verbose","help","print-reference-card","piana-dbname=", "piana-dbhost=","piana-dbuser=","piana-dbpass=", "depth=", 
                                    "hub-threshold=","input-file=","input-id-type=","input-proteins-species=", "output-id-type=", "output-proteins-species=",
                                    "spots-file-name=", "expression-id-type=", "file-over-expressed=", "file-infra-expressed=", "special-proteins=",
                                    "exec-mode=","configuration-file=","results-prefix=", "results-dir=","memory-usage="])
    except getopt.GetoptError, msg:
        # print help information and exit:
        sys.stderr.write( "\n\n--\ncommand line arguments are not correct: %s\n--\n\n" %(msg))
        sys.exit(2)
     
    for option,value in opts:

        # this function checks what the option is, and sets the corresponding parameter to "value"
        set_parameter_value(parameter_name= option.strip("--"), value= value)
    # ENDOF for option,value in opts:


def parse_configuration_file(input_configuration_file= None):
    """
    parses a piana configuration file

    configuration_file is required in batch_mode. In interactive_mode, it can be used to set parameters values,
    but execution commands will be ignored.

    parameters set by configuration file do not overwrite those that were read from the command line
    (ie. command line has preference over configuration file)

    To create your own configuration file, follow instructions on file general_template.piana_conf
    """

    global execution_sequence

    counter = 0

    parsing_section = "parameters" # variable used to indicate if we are parsing the parameters section of the configuration file
                                   # or the execution commands section. The transition is marked in the file with a line starting
                                   # with ">>>"

    if input_configuration_file.lower() == "stdin":
        configuration_file_fd = sys.stdin
    else:
        configuration_file_fd = file(input_configuration_file, "r")


    # Modified: as (for line in fd) uses an internal buffer, it is not wanted when using stdin. So, it is necessary to change it with readline()
    #for conf_line in configuration_file_fd:
    conf_line = configuration_file_fd.readline()
    while( conf_line ):

        if conf_line[0] !="#" and conf_line.strip() != "":
            # if the configuration line is not a comment or an empty line, read it
            
            if verbose_parse_conf:
                sys.stderr.write("parsing configuration line %s\n" %(conf_line) )

            if parsing_section == "parameters":

                if conf_line[0:3] == ">>>":
                    # check for transition between parameters and execution commands
                    parsing_section = "execution"
                else:
                    # if we are parsing the parameters section, just get the value after the "=" and assing it to the parameter
                    conf_line_fields = conf_line.split("=")
                    parameter_name =conf_line_fields[0].strip()
                    parameter_value = conf_line_fields[1].strip()

                    #if parameter_value != "blank":  THIs used to be here, but I have removed to make sure the parameters get their default values
                    #                                Previously, they were just being left to None

                    
                    # this function checks what the option is, and sets the corresponding parameter to "value"
                    set_parameter_value(parameter_name= parameter_name, value= parameter_value)
                # END OF else: (if conf_line[0:3] == ">>>":)

            # END OF if parsing_sectin == "parameters":
            
            elif parsing_section == "execution":
                # if it is an execution line, obtain the command (first element before the ;) and set the arguments accordingly
                #  (execution_sequence follows the structure described in Main section of program)
                
                execution_line_fields = conf_line.strip().split(";")  # execution_line_fields[0] is the command name
                                                                      # execution_line_fields[1:] are the arguments "name_argument=value_argument"

                if not execution_line_fields[0] in valid_execution_commands.keys():
                    raise ValueError("Unknown execution command <%s> set in configuration file <%s>\n" %(execution_line_fields[0],
                                                                                                         input_configuration_file))
                
                if verbose_parse_conf:
                    sys.stderr.write("parsing execution command fields %s\n" %(execution_line_fields) )

                execution_sequence[counter] = {}
                execution_sequence[counter][ execution_line_fields[0] ] = {}

                if execution_line_fields[1].strip() != "":
                    # if there are arguments (second element of split(";") is not null), retrieve them...
                    for execution_argument in execution_line_fields[1:]:
                        # for each argument, get name and value (which are separated by "=")
                        execution_argument_fields = execution_argument.split("=") # execution_argument_fields[0] is the argument name
                                                                                  # execution_argument_fields[1] is the argument value
                        if len(execution_argument_fields) > 1:
                            # if there was an "=" in the argument, then set execution_sequence
                            execution_sequence[counter][execution_line_fields[0]][execution_argument_fields[0]] = execution_argument_fields[1]
                    # END OF for execution_argument in execution_line_fields[1:]:
                # END OF if len(execution_line_fields) > 1:

                counter += 1
            # END OF elif parsing_section == "execution":
            
        # END OF if conf_line[0] !="#" and conf_line.strip() != "":
        conf_line = configuration_file_fd.readline()
    # END OF for conf_line in configuration_file_fd:

    # reversing the ordered dictionary of execution commands, so we can use pop() to get one by one the dictionary elements
    execution_sequence.reverse()






#  ------------------------------------------------------------------------------------------------
#  ------------------------------------------------------------------------------------------------
#                                                    Main               
#  ------------------------------------------------------------------------------------------------
#  ------------------------------------------------------------------------------------------------

# --
# General variables that will be instantiated with values from the command line or from the configuration file
# --
# To avoid problems with old configuration files, new parameters have default values assigned
# However, to avoid using non-desired values, it is recommended that the configuration file explicitly sets values for all parameters

piana_dbname= None            # defines the piana database parameters
piana_dbhost= None
piana_dbuser= None
piana_dbpass= None

use_secondary_db = None


input_file_name = None
input_id_type = None
input_proteins_species = None

output_id_type = None
output_proteins_species = None

alternative_id_types = []
list_keywords = []

file_infra_expressed = None
file_over_expressed = None
expression_id_type= None

special_proteins = None

spots_file_name = None

depth = None

hub_threshold = None
use_self_ints = None

list_source_dbs = None
inverse_dbs = None
list_source_methods = None
inverse_methods = None

results_prefix= None          # prefix that all results produced by piana will contain in their name
results_dir= None             # directory where results will be saved
exec_mode = None              # type of execution that will be performed: interactive or batch
memory_usage = "memory"       # Memory management used to construct and store networks. It can be "memory" or "speed"
configuration_file = None     # file that sets values for piana parameters and execution commands

argument_print_reference_card = 0

initial_time=time.time()

# --
# Global variables of the program
# --

execution_sequence = seqdict()         # An ordered dictionary describing how the batch execution of piana will be done
                                       # (it is filled by the configuration file parser)
                                       # 
                                       # I've been obliged to add a counter as a key, so the user can execute the same
                                       # command many times. If the execution_command is the key, then each command
                                       # can only appear once in the execution_sequence
                                       #
                                       # structure of execution_sequence is: (if real dictionaries were ordered...)
                                       #
                                       #                     { 0:{execution_command1: {argument11:value11, argument12:value12, ...}},
                                       #                       1:{execution_command2: {argument21:value21, argument22:value22, ...}},
                                       #                       2:{...........................................................
                                       #                     }
                                       #

# parsing arguments from the command line
parseArguments()

report_file_object = None

if configuration_file is not None:
    # parsing configuration file with instructions on how to run piana
    #   (remember that command line options have preference over parameter values in configuration file)

    try:
        parse_configuration_file(configuration_file)
        
    except Exception, inst:
        sys.stderr.write("\n******************\nError reported: %s\n" %(inst))
        sys.stderr.write("******************\n")
        sys.stderr.write("Your configuration file has errors: check that you respect the format shown in general_template.piana_conf\n")
        sys.stderr.write(" --> read the error reported above\n\n")
        sys.stderr.write("Some typical errors:  \n")
        sys.stderr.write("    - you forgot the ';' after the command name or you wrote a ';' after the last argument of the command\n")
        sys.stderr.write("    - you are using an old configuration file that does not respect the new format\n")
        sys.stderr.write("    - you misspelled the name of the parameters\n")
        sys.stderr.write("    - you are using an invalid command\n")
        sys.stderr.write("    - your configuration file does not exist\n")
        sys.stderr.write("    - your input files do not exist\n\n")
        sys.exit()
    
else:
    # no configuration file given...
    
    if exec_mode is None:
        sys.stderr.write("\nWarning! exec-mode is None --> setting exec-mode to default (interactive)\n\n")
        exec_mode = "interactive"
        
    if exec_mode == "batch":
        # if exec mode is batch, configuration_file is mandatory
        sys.stderr.write( "\nbatch mode requires a configuration_file \n")
        sys.exit(2)
        
    elif exec_mode == "interactive":
        # set here defaults for variables that cannot be set through command line or interactive mode
        
        use_secondary_db="no"

        list_source_dbs="all"
        use_self_ints = "yes"
        inverse_dbs="no"
        list_source_methods="all"
        inverse_methods="no"
        
        alternative_id_types=["uniacc", "unientry", "gi", "geneName", "md5", "proteinPiana"]
        list_keywords=[]
        
        file_infra_expressed = None
        file_over_expressed = None
        expression_id_type= None

        special_proteins = None
        
# END OF if configuration_file is not None:

# -------------------------
# Make sure we have all parameters we need
# -------------------------

if use_secondary_db is None: use_secondary_db = "no"  # if the user didn't say anything about
                                                      # the secondary db, assume he doesn't want to use it

# in interactive mode, a db name and host is required
if exec_mode == "interactive" and piana_dbname is None:
    piana_dbname = raw_input(string_get_piana_dbname)
    
if exec_mode  == "interactive" and piana_dbhost is None:
    piana_dbhost = raw_input(string_get_piana_dbhost)

if results_dir is None:
    results_dir = "./"


# create a name for this network (we do it from the input_file_name, in case it exists...
# TO DO!!! Give the possibility to the user to give a name to the network (in the input parameters of the configuration file)
if input_file_name is not None:
    network_name = os.path.split(input_file_name)[1].split(".")[0] # keep only the file name without path of extension
else:
    network_name = "no_name"

# ---------------------------------
# open an interface to the piana api
if verbose_command_sequence:
    sys.stderr.write("Opening connection to Piana database %s on server %s (use_secondary_db=%s)\n" %(piana_dbname, piana_dbhost, use_secondary_db))
    
piana_api = PianaApi(piana_dbname=piana_dbname, piana_dbhost= piana_dbhost, piana_dbuser=piana_dbuser, piana_dbpass= piana_dbpass,
                     piana_dbport= piana_configuration_parameters.piana_dbport, 
		     use_secondary_db= use_secondary_db, network_name=network_name, memory_mode=memory_usage )
# ---------------------------------

if argument_print_reference_card:
    # if the user asked for the reference card, print it and exit
    piana_api.write_reference_card(file_object=sys.stdout)
    sys.exit()


# convert input_proteins_species and output_proteins_species to input_tax_id and output_tax_id

if input_proteins_species:
    input_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=input_proteins_species)

if output_proteins_species:
    output_tax_id= piana_api.get_one_tax_id_from_species_name(species_name=output_proteins_species)

# parse the special-proteins line to create the dictionary of special proteins 
dic_special_proteins_file = {}   # it is dic that contains special proteins and their labels
                                  #  follows structure:
                                  #      { label1: (id_type1, path_to_file1),
                                  #        label2: (id_type2, path_to_file2),
                                  #        ...............
                                  #      }

if special_proteins:
    special_proteins_tokens = special_proteins.split(":")
    for one_token in special_proteins_tokens:
	# for each file of special proteins, get path to file, label and type of id
	token_fields = one_token.split(",")
	# token_fields[0] -> path_to_file1
	# token_fields[1] -> id_type1
	# token_fields[2] -> label1
	dic_special_proteins_file[token_fields[2]] = (token_fields[1], token_fields[0])



# TO DO!!!!!!!!!!
# in some cases (eg. protein_code2protein_code) we do not want to build the network for input_file
# how do I make the difference between those cases and the others??????
#  -> I could set a flag on the configuration file to indicate that the PPI network is not needed

if input_file_name is not None and depth != -1:
    # ----------------------------------------------------------------------------------------------------------
    # this is a very specific case for piana.py... normally, building a network is just a matter of adding files,
    # proteins and interaction files. However, when the user gives an input_file_name through the command line
    # we must build a network directly for that input_file_name. Why? Because in many cases, the user just wants
    # to work with one proteins file, and instead of adding the file through the interactive menu (or through the
    # configuration file) we let him speed up things by setting it directly through this command line parameter.
    # Briefly: when the parameter input_file_name is not None, build the network for it making appropiate calls...
    #
    # The user can avoid this step by setting depth to -1. This is useful for commands that do not need a network
    # to be built (e.g. translating between protein identifiers)
    # ----------------------------------------------------------------------------------------------------------
    
    if exec_mode == "batch":
        if input_id_type is None or depth is None or hub_threshold is None:
            sys.stderr.write("input-id-type(%s), hub_threshold (%s) and depth (%s) cannot be None\n" %(
                input_id_type,
                hub_threshold,
                depth))
            raise ValueError("In batch mode, when giving an input-file, all parameters stated above must be passed as well\n")

        this_file_id_type= input_id_type
        this_file_tax_id= input_tax_id
        this_file_depth= depth
        this_file_hub_threshold= hub_threshold
    else:
        # exec_mode is interactive
        if input_id_type is None:  this_file_id_type = raw_input("Input file identifier type: " + string_get_id_type)
        else:                            this_file_id_type= input_id_type
                
        if input_proteins_species is None: this_file_protein_species = raw_input("Input file protein species: "+string_get_species_name)
        else:                              this_file_protein_species = input_proteins_species

        this_file_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=input_proteins_species)
                
        if depth is None:   this_file_depth = int(raw_input(string_get_depth))
        else:               this_file_depth = depth
                
        if hub_threshold is None:    this_file_hub_threshold = int(raw_input(string_get_hub_threshold))
        else:                        this_file_hub_threshold = hub_threshold


    if verbose_command_sequence:
        sys.stderr.write("Adding proteins in input file %s of type %s with threshold %s and depth %s\n" %(input_file_name, this_file_id_type,
                                                                                                          this_file_hub_threshold, depth ))

    file_object = file(input_file_name, "r")

    if time_control:
        initial_time = time.time()
        
    piana_api.add_file_proteins_to_piana_graph(file_object= file_object,
                                               id_type= this_file_id_type,
                                               tax_id_value= this_file_tax_id,
                                               depth = this_file_depth,
                                               hub_threshold= this_file_hub_threshold,
                                               use_self_ints=use_self_ints,
                                               list_source_dbs= list_source_dbs,
                                               inverse_dbs= inverse_dbs,
                                               list_source_methods= list_source_methods,
                                               inverse_methods= inverse_methods,
					       force_secondary_db="no")         # input_file will never use the secondary DB... otherwise, it wouldn't be secondary


    if time_control:
         sys.stderr.write("Adding input file proteins to graph took %s seconds\n" %(time.time() - initial_time))
        

        
    file_object.close()
# END OF if input_file_name is not None:


"""
the piana_object has been initialized: get PIANA command (interactively or from configuration file) to be executed and proceed


1. Get command from menu (in interactive mode) or from configuration file (in batch mode)
2. Execute command

"""
option = None
while option!= "exit":


    # 1. Get command to be executed

    if exec_mode == "interactive":
        if not option or option == "?":
            # If interactive mode, print list of available options
            sys.stdout.write( "=============================================================\n")
            sys.stdout.write( "PIANA is under GNU GPL\n")
            sys.stdout.write( "piana v. 1.0, Copyright (C) 2005 Ramon Aragues\n")
            sys.stdout.write( "piana comes with ABSOLUTELY NO WARRANTY; for details\n")
            sys.stdout.write( "read the license.  This is free software, and you are welcome\n")
            sys.stdout.write( "to redistribute it under certain conditions; read the license \n")
            sys.stdout.write( "for details.\n")

            sys.stdout.write( "==============================================================\n")
            sys.stdout.write( "PIANA commands are:\n\n")

            for execution_command in valid_execution_commands.keys():
                sys.stdout.write("%s --> %s\n" %(execution_command.ljust(30), valid_execution_commands[execution_command][0] ))

            sys.stdout.write( "==============================================================\n")
            
        option = raw_input("\nType your command (? for help):> ")

            

        sys.stdout.write("option chosen was: %s\n" %option)
    # END OF if exec_mode == "interactive":

    
    elif exec_mode == "batch":
        # if batch mode, get the execution command and arguments from the execution_sequence ordered dictionary
        current_command = execution_sequence.pop()
        
        option = (current_command.values()[0]).keys()[0]  # option is a string containing the command (must be in valid_execution_commands)
        arguments = current_command.values()[0][option]   # arguments is a dictionary with {argument_name=argument_value, ...} for this command
    # END OF elif exec_mode == "batch":



    # 2. Execute the command
    
    """
    In either mode, batch or interactive, option is the command that piana must execute

    In interactive mode, the parameters for the can be already set
    (through command line or configuration file) or obtained from the
    user interactively. They are obtained interactively in case their
    value is None.

    In batch mode, input parameters must be set in advance (through command line or configuration file). 
      --> read arguments from configuration file (which has been previously parsed)
          ( arguments is a dictionary with {argument_name=argument_value, ...} )


    The code is identical for all commands:

    1. get arguments values from user (either interactively or from the configuration file)
    2. check and prepare (if necessary) arguments values
    3. call a PianaApi method 
    """

    # -------------------------------------------------------------------------------------------------
    # option == create-report
    # tells PIANA to create a report with results files
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: create-report;file-name=file_name
    #
    if option == "create-report":

        if exec_mode == "interactive":
            report_file_name = raw_input("Global report file:" + string_get_save_file)  

        elif exec_mode == "batch":
            try:
                if arguments[arg_file_name] != "blank":  report_file_name  = arguments[arg_file_name] + ".html"
                else:
                    if results_prefix:
                        report_file_name  = results_prefix + "."  + option + ".html"
                    else:
                        raise ValueError("Trying to give a default name to the report without giving a results prefix")
                    
            except ValueError, msg:
                
                sys.stderr.write("configuration file error: make sure your command create-report respects the format described in general_template.piana_conf\n")
                sys.stderr.write("This error might also be raised when not providing a results-prefix and asking PIANA to create default output file names\n")
                sys.exit()
                
        if verbose_command_sequence:
            sys.stderr.write("PIANA will create a report (%s) for the following commands\n" %(report_file_name))
            
        
	try:
	    # In case it exists, close  previous report file (it might be open if the user asked for more than one report)
	    report_file_object.close()
	except:
	    pass

        # save the current network
        report_file_object = file(results_dir + report_file_name , "wb")

	report_file_object.write(""" <title>PIANA report: %s</title>\n""" %(report_file_name))
	report_file_object.write(""" <table border=1><tr><td><big><b> PIANA report: %s </b></big></td></tr></table><br>\n""" %(report_file_name))
	report_file_object.write(""" This report links to results obtained by PIANA.<br>\n""")
	report_file_object.write("""<hr>\n""")
	report_file_object.write("""<b>Attention:</b><ul>\n""")
	report_file_object.write("""<li>All tables and images in this report can be downloaded by right-clicking on them and chosing "save link target as..."</li> \n""")
	report_file_object.write("""<li>In this report, we call "root proteins" to those proteins that were given to PIANA as seeds to create the protein interaction network</li> \n""")
	report_file_object.write("""<li>In this report, we call "linker proteins" to those proteins that are connected to more than one root protein</li> \n""")
	report_file_object.write("""</ul><hr> \n""")
	report_file_object.write("""<b>INPUT PARAMETERS</b><br><br>\n""")
	report_file_object.write("""<ul>\n""")
	report_file_object.write("""<li>PIANA database being used is %s (on server %s)</li><br>\n""" %(piana_dbname, piana_dbhost))

	if use_secondary_db == "yes":
	    report_file_object.write("""<li>Using a secondary database %s (on server %s)<br>\n""" %(piana_configuration_parameters.piana_secondary_dbname, 
												  piana_configuration_parameters.piana_secondary_dbhost ))
	    report_file_object.write("""For those proteins with no interactions on primary database, interactions were retrieved from this secondary DB: %s (server %s)</li><br>\n""" %(piana_configuration_parameters.piana_secondary_dbname, 
												  piana_configuration_parameters.piana_secondary_dbhost ))


	report_file_object.write("""<li>PIANA used interactions from the following public repositories of interactions: %s\n""" %(list_source_dbs))
	if list_source_dbs == "all":
	    report_file_object.write("""<br> (%s)</li><br>\n""" %(piana_api.get_external_databases_in_pianaDB(database_type="protein-protein interactions")))
	else:
	    report_file_object.write("""</li><br>\n""" %())

	report_file_object.write("""<li>PIANA used interactions detected using the following methods: %s</li><br>\n""" %(list_source_methods))


	if input_file_name:
	    report_file_object.write("""<li>A initial network was created using proteins in <a href="%s">this file</a> as root<br><ul>\n""" %(input_file_name.split("/")[-1]))
	    report_file_object.write("""<li> This initial network was created by retrieving interaction partners for proteins in the above mentioned databases.</li><br>\n""")
	    report_file_object.write("""<li> The type of identifier is this file is %s, considering that proteins are of species %s</li></ul></li><br>\n""" %(input_id_type, input_proteins_species))

	report_file_object.write("""<li>Partners were retrieved up to depth %s<br>\n""" %(depth))
	report_file_object.write(""" <i>The depth represents how far away from the root proteins we went to retrieve partners (eg. depth 2 means retrieving partners of partners of root proteins)</i></li><br>\n""")

	if hub_threshold != 0:
	    report_file_object.write("""<li>A hub threshold of %s was used\n""" %(hub_threshold))
	else:
	    report_file_object.write("""<li>No hub threshold was used<br>\n""")
	report_file_object.write(""" <i>For some experiments, it is useful to avoid adding proteins to the network when they have too many interactions.<br>\n""")
	report_file_object.write(""" The hub threshold fixes which is the maximum number of partners a protein can have in order to be added to the network</i> </li><br>\n""")
	
	if use_self_ints == "no":
	    report_file_object.write("""<li>Self interactions were not added<br>(however, due to the ideosynchrasy of PIANA, you might observe some cases where are protein interacts with itself)</li><br>\n""")
	    

	if file_infra_expressed or file_over_expressed:
	    report_file_object.write("""<li>Expression data being used in this report was retrieved from the following files: (type of protein identifier is %s)<br><ul>\n""" %(expression_id_type))
	    report_file_object.write("""<li> Proteins in <a href="%s">this file</a> were considered as over expressed</li>\n""" %(file_over_expressed.split("/")[-1]))
	    report_file_object.write("""<li> Proteins in <a href="%s">this file</a> were considered as under expressed</li>\n""" %(file_infra_expressed.split("/")[-1]))
	    
	report_file_object.write("""</li><br></ul></ul><hr>\n""")

	report_file_object.write("""<b>LINKS to RESULTS</b><br><ul>\n""")
	
           
    # END OF elif option == "create-report":

    # -------------------------------------------------------------------------------------------------
    # option == reset-network
    # start a new network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: reset-network;
    #
    elif option == "reset-network":

        if verbose_command_sequence:
            sys.stderr.write("Reseting network\n")
            
	if report_file_object:
	    report_file_object.write("""<li>Network was resetted at this point: proteins previously added are not considered from here on.</li><br>\n""")
	
        piana_api.reset_piana_graph()
        
        
           
    # END OF elif option == "reset-network":

    # -------------------------------------------------------------------------------------------------
    # option == save-network
    # saves the current network into a disc file
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: save-network;file-name=file_name
    #
    elif option == "save-network":

        if exec_mode == "interactive":
            save_file_name = raw_input(string_get_save_file)  

        elif exec_mode == "batch":
            try:
                if arguments[arg_file_name] != "blank":    save_file_name = arguments[arg_file_name]
                else:
                    sys.stderr.write("Trying to save the network without providing a file name!\n")
                    sys.stderr.write("You must set the destination file name in the arguments of your command save-network\n")
                    sys.exit()
            except:
                sys.stderr.write("configuration file error: make sure your command save-network respects the format described in general_template.piana_conf\n")
                sys.exit()
                
        # END OF elif exec_mode == "batch": (if exec_mode == "interactive":)

        
        if verbose_command_sequence:
            sys.stderr.write("Saving network\n")
            
        # save the current network
        file_object = file(results_dir + save_file_name, "wb")

        cPickle.dump(piana_api, file_object , 2) 
        file_object.close()

    # END OF elif option == "save-network":

    # -------------------------------------------------------------------------------------------------
    # option == load-network
    # loads network from a file, overwriting current network in memory
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: load-network;file-name=file_name
    #
    elif option == "load-network":

        if exec_mode == "interactive":
            load_file_name = raw_input(string_get_load_file)  

        elif exec_mode == "batch":
            try:
                if arguments[arg_file_name] != "blank":  load_file_name = arguments[arg_file_name]
                else:  
                    sys.stderr.write("Trying to load a network without providing a file name!\n")
                    sys.stderr.write("You must set the destination file name in the arguments of your command save-network\n")
                    sys.exit()  
            except:
                sys.stderr.write("configuration file error: make sure your command load-network respects the format described in general_template.piana_conf\n")
                sys.exit()            
            
        # END OF elif exec_mode == "batch": (if exec_mode == "interactive":)


        if verbose_command_sequence:
            sys.stderr.write("Loading network\n")
          
  
        file_object = file(results_dir + load_file_name, "rb")
        piana_api = cPickle.load(file_object)
        file_object.close()

	if report_file_object:
	    report_file_object.write("""<li>A new network was loaded from file %s. Previous actions are not considered from here on.</li><br>\n""" %(results_dir + load_file_name))
   
    # END OF elif option == "load-network":

    # -------------------------------------------------------------------------------------------------
    # option == add-protein
    # add a protein to the network (and its interaction partners at depth X)
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: add-protein;protein-name=protein_name;id-type=this_protein_code_type
    #
    elif option == "add-protein":

        if exec_mode == "interactive":

            # protein name cannot be set through the command line:always ask for it
            protein_name = raw_input("Protein to add to network: " + string_get_protein)  

            if input_id_type is None:  this_id_type = raw_input("Input identifier type: " + string_get_id_type)
            else:                            this_id_type = input_id_type
                
            if input_proteins_species is None: this_protein_species = raw_input("Input protein species: " + string_get_species_name)
            else:                              this_protein_species = input_proteins_species
            
            
            if depth is None:          depth = int(raw_input(string_get_depth))
                
            if hub_threshold is None:  hub_threshold = int(raw_input(string_get_hub_threshold))

	    this_force_secondary_db = "no"  # setting by default in interactive mode not to use the secondary PIANA db

        elif exec_mode == "batch":
            
            protein_name = arguments[arg_protein_name]
            
            if arguments[arg_id_type] != "blank":   this_id_type = arguments[arg_id_type]
            else:                                        this_id_type = input_id_type
                
            if arguments[arg_species_name] != "blank":   this_protein_species = arguments[arg_species_name]
            else:                                        this_protein_species = input_proteins_species

	    if arguments.has_key(arg_piana_db):
		if arguments[arg_piana_db] == "secondary":   this_force_secondary_db = "yes"
		else:                                        this_force_secondary_db = "no"
	    else:
		this_force_secondary_db = "no"
		

        # END OF elif exec_mode == "batch": (if exec_mode == "interactive":)


        # check required parameters
        if protein_name is None or this_id_type is None or this_protein_species is None or depth is None or hub_threshold is None:
            raise ValueError("To add a protein, protein name (%s), identifier type (%s), protein species (%s) depth(%s) and hub_threshold (%s) are required\n" %(
                protein_name,
                this_id_type,
                this_protein_species,
                depth,
                hub_threshold))
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)

        if verbose_command_sequence:
            sys.stderr.write("Adding protein %s of type name %s with tax %s\n" %(protein_name, this_id_type, this_tax_id))
          
	if report_file_object:
	    report_file_object.write("""<li>A new protein was added to the network: %s (identifier type is %s, species is %s)</li><br>\n""" %(protein_name, this_id_type, this_tax_id))
   
        piana_api.add_protein_to_piana_graph(protein_code=protein_name,
                                             id_type = this_id_type,
                                             tax_id_value= this_tax_id,
                                             depth=depth,
                                             hub_threshold=hub_threshold,
                                             use_self_ints = use_self_ints,
                                             list_source_dbs=list_source_dbs,
                                             inverse_dbs= inverse_dbs,
                                             list_source_methods=list_source_methods,
                                             inverse_methods= inverse_methods,
					     force_secondary_db= this_force_secondary_db)

    # END OF elif option == "add-protein":

    # -------------------------------------------------------------------------------------------------
    # option == add-proteins-file
    # add a list of proteins from an text file to current network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: add-proteins-file;file-name=complete_path_to_file;id-type=this_file_proteins_code_type
    #
    elif option == "add-proteins-file":

        if exec_mode == "interactive":
            file_name = raw_input( string_get_file_name)
            
            if input_id_type is None:  this_file_id_type = raw_input("File identifier type: " + string_get_id_type)
            else:                            this_file_id_type= input_id_type
                
            if input_proteins_species is None: this_file_protein_species = raw_input("Input protein species: "+string_get_species_name)
            else:                              this_file_protein_species = input_proteins_species
                
            if depth is None:   depth = int(raw_input(string_get_depth))
                
            if hub_threshold is None:    hub_threshold = int(raw_input(string_get_hub_threshold))

	    this_force_secondary_db = "no"  # setting by default in interactive mode not to use the secondary PIANA db

        elif exec_mode == "batch":
                
            if arguments[arg_file_name] != "blank":         file_name = arguments[arg_file_name]
            else:                raise ValueError("Trying to add a proteins file without setting its name in configuration file\n")
            
            if arguments[arg_id_type] != "blank":      this_file_id_type = arguments[arg_id_type]
            else:                                           this_file_id_type = input_id_type

            if arguments[arg_species_name] != "blank":      this_file_protein_species = arguments[arg_species_name]
            else:                                           this_file_protein_species = input_proteins_species

	    if arguments.has_key(arg_piana_db):
		if arguments[arg_piana_db] == "secondary":   this_force_secondary_db = "yes"
		else:                                        this_force_secondary_db = "no"
	    else:
		this_force_secondary_db = "no"
		
        # END OF elif exec_mode == "batch":

        # check required parameters
        if file_name is None or this_file_id_type is None or this_file_protein_species is None:
            raise ValueError("To add proteins from a file, file name (%s), identifier type (%s) and protein species (%s) are required\n" %(
                file_name,
                this_file_id_type,
                this_file_proteins_species))

        this_file_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_file_protein_species)
        
        if verbose_command_sequence:
            sys.stderr.write("Adding proteins in file %s of type %s with threshold %s\n" %(file_name, this_file_id_type, hub_threshold))

	if report_file_object:
	    report_file_object.write("""<li>A <a href="%s">new file</a> with proteins was added (identifier type is %s, species is %s)</li><br>\n""" %(file_name, this_file_id_type, this_file_tax_id))

        file_object = file(file_name, "r")
        piana_api.add_file_proteins_to_piana_graph(file_object= file_object,
                                                   id_type= this_file_id_type,
                                                   tax_id_value= this_file_tax_id,
                                                   depth=depth,
                                                   hub_threshold=hub_threshold,
                                                   use_self_ints = use_self_ints,
                                                   list_source_dbs=list_source_dbs,
                                                   inverse_dbs= inverse_dbs,
                                                   list_source_methods=list_source_methods,
                                                   inverse_methods= inverse_methods,
						   force_secondary_db= this_force_secondary_db)
        file_object.close()
    # END OF elif option == "add-proteins-file":

    
    # -------------------------------------------------------------------------------------------------
    # option == add-interactions-file
    # add a list of interactions from an text file to current network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: add-interactions-file;file-name=complete_path_to_file;id-type=this_file_interactions_code_type
    #
    elif option == "add-interactions-file":

        if exec_mode == "interactive":
            file_name = raw_input(string_get_int_file_name)
            
            if input_id_type is None:
                this_file_id_type = raw_input("identifiers type used in interactions: " + string_get_id_type)
            else:
                this_file_id_type= input_id_type

        elif exec_mode == "batch":

            if arguments[arg_file_name] != "blank":
                file_name = arguments[arg_file_name]
            else:
                raise ValueError("Trying to add a interactions file without setting its name in configuration file\n")

            if arguments[arg_id_type] != "blank":
                this_file_id_type = arguments[arg_id_type]
            else:
                this_file_id_type = input_id_type

        # END OF elif exec_mode == "batch":


        if file_name is None or this_file_id_type is None:
            raise ValueError("To add interactions, file name (%s) and a identifier type (%s) are required\n" %(file_name,
                                                                                                            this_file_id_type ))
            
        if verbose_command_sequence:
            sys.stderr.write("Adding interactions in file %s with proteins of type %s\n" %(file_name, this_file_id_type))

	if report_file_object:
	    report_file_object.write("""<li>A new file with interactions was added to the network: %s</li>\n""" %(file_name))

        file_object = file(file_name, "r")
        
        piana_api.add_file_interactions_to_piana_graph(file_object= file_object, id_type= this_file_id_type)
       
    # END OF elif option == "add-interactions-file":

    # -------------------------------------------------------------------------------------------------
    # option == species-network
    # buids a protein-protein network for all proteins in a given species
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: species-network;species-name=species_name;tax-id=tax_id
    #
    elif option == "species-network":

        # TO DO!!! let the user choose the taxonomy directly, instead of going through species name
        taxonomy_value = None
        
        if exec_mode == "interactive":
            temp_species_name = raw_input( "Build network of species: " + string_get_net_species_name)
            temp_tax_id = raw_input( "Build network of taxonomy: " + string_get_net_tax_id)

            if temp_species_name == "blank":      species_name = None
            else:                                 species_name =temp_species_name

            if temp_tax_id == "blank":            tax_id= None
            else:                                 tax_id= temp_tax_id
                
            if hub_threshold is None:  hub_threshold = int(raw_input(string_get_hub_threshold))
                
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_species_name] != "blank":        species_name = arguments[arg_species_name]
            else:                                             species_name = None
                
            if arguments[arg_tax_id] != "blank":              tax_id = int(arguments[arg_tax_id])
            else:                                             tax_id = None
        # END OF elif exec_mode == "batch":
        
        if species_name is None and tax_id is None or hub_threshold is None:
            raise ValueError("Trying to build species network without setting a species (species name(%s) or taxonomy id(%s)) or hub threshold (%s)\n" %(
                species_name, tax_id, hub_threshold))
            
        if verbose_command_sequence:
            sys.stderr.write("Building the network for (species <%s> -- tax_id <%s>) \n" %(species_name, tax_id))

	if report_file_object:
	    report_file_object.write("""<li>Network was built for a specific species (species_name=%s, tax_id=%s)</li>\n""" %(species_name, tax_id))


        piana_api.create_species_piana_graph(species_name= species_name,
                                             tax_id= tax_id,
                                             hub_threshold= hub_threshold,
                                             use_self_ints = use_self_ints,
                                             list_source_dbs= list_source_dbs,
                                             inverse_dbs= inverse_dbs,
                                             list_source_methods= list_source_methods,
                                             inverse_methods= inverse_methods)
        
        
       
    # END OF elif option == "species-network":

    # -------------------------------------------------------------------------------------------------
    # option == database-method-network
    # buids a protein-protein network for all interactions in a given database and/or a given method
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: database-method-network;database-name=database_name;method-name=method_name
    #
    elif option == "database-method-network":
        
        if exec_mode == "interactive":
            database_name = raw_input( "Database to be used: " + string_get_dbname)
            method_name = raw_input( "Method to be used: " + string_get_methodname)
            species_name = raw_input( "Species to be used: " + string_get_species_name)
                
            if hub_threshold is None:  hub_threshold = int(raw_input(string_get_hub_threshold))
            
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_database_name] != "blank":
                database_name = arguments[arg_database_name]
            else:
                raise ValueError("Trying to build network without setting the database name in configuration file\n")
            
            if arguments[arg_method_name] != "blank":
                method_name = arguments[arg_method_name]
            else:
                raise ValueError("Trying to build  network without setting the method name in configuration file\n")

            if arguments[arg_species_name] != "blank":
                species_name = arguments[arg_species_name]
            else:
                raise ValueError("Trying to build species network without setting the species name in configuration file\n")

        # END OF elif exec_mode == "batch":

        if species_name is None:
            raise ValueError("Trying to build database-method network without setting a species name (%s)\n" %(species_name))
        
        if database_name != "all":
            list_source_dbs = [database_name]
            inverse_dbs= "no"

            
            
        if method_name != "all":
            list_source_methods = [method_name]
            inverse_methods= "no"

        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=species_name)
        
        
        if verbose_command_sequence:
            sys.stderr.write("Building the network for species %s source db %s and method %s\n" %(species_name,
                                                                                                  database_name,
                                                                                                  method_name))

	if report_file_object:
	    report_file_object.write("""<li>Network was built for all interactions of a specific external database and detection methods (see input parameters above)</li>\n""")

        # TO DO!!! Why am I not using hub_threshold in this method?
        piana_api.create_database_method_piana_graph(tax_id_value= this_tax_id,
                                                     use_self_ints = use_self_ints,
                                                     list_source_dbs= list_source_dbs,
                                                     inverse_dbs= inverse_dbs,
                                                     list_source_methods = list_source_methods,
                                                     inverse_methods= inverse_methods)
        
    # END OF elif option == "database-method-network":

    
    # -------------------------------------------------------------------------------------------------
    # option == print-table or print-network
    #
    # print network, in several output flavors:
    #
    #   print-table: standard table with protein interactions
    #   print-network: standard DOT file
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: print-table;output-target=output_target;id-type=protein_code_type;print-mode=print_mode;format-mode=format_mode
    # configuration file syntax: print-network;output-target=output_target;id-type=protein_code_type;print-mode=print_mode;format-mode=format_mode

    elif option == "print-table" or option == "print-network":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            print_mode = raw_input(string_get_print_mode)
 
	    format_mode = raw_input(string_get_format_mode)

            
            if output_id_type is None:            this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:                                       this_print_id_type = output_id_type
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species


        # END OF if exec_mode == "interactive":
        
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")

            if arguments[arg_print_mode] != "blank":             print_mode = arguments[arg_print_mode]
            else:                                                print_mode = "all"
            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:         
		if format_mode is None:
		    raise ValueError("You must set a format mode for your output")
                

            if arguments[arg_output_target] != "blank":          
		output_name = arguments[arg_output_target]
		network_attributes_prefix = output_name + "."
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                # keep the network_attributes_prefix for cases where the user asked to have SIF format
		#   -> it will be used to create files that cytoscape uses for loading information for nodes and edges
		network_attributes_prefix = results_dir + results_prefix + "." + print_mode + "." + option + "."
		output_name = network_attributes_prefix + format_mode

            if output_proteins_species:         this_protein_species = output_proteins_species
            else:                               this_protein_species = "all"
        # END OF elif exec_mode == "batch":


        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name , "w")


        if option == "print-table":

            if verbose_command_sequence:
                sys.stderr.write("Printing out table to file %s with output type %s\n" %(output_name, this_print_id_type))

  
	    if report_file_object:

		if print_mode == "all":
		    report_file_object.write("""<li>This table contains <a href="%s">all interactions</a> in the network<br><ul>\n""" %(
                        output_name.split("/")[-1]))

		elif print_mode == "all_root" :
		    report_file_object.write("""<li>This table contains <a href="%s">all interactions in the network where at least one root protein is involved</a><br><ul>\n""" %(
			    output_name.split("/")[-1]))

		elif print_mode == "only_root":
		    report_file_object.write("""<li>This table contains <a href="%s">all interactions in the network between root proteins</a><br><ul>\n""" %(output_name.split("/")[-1]))
		    report_file_object.write("""<li>Only interactions where both members are root proteins are shown</li>\n""")

		elif print_mode == "connecting":
		    report_file_object.write("""<li>This table contains <a href="%s">all interactions in the network between root and linker proteins</a><br><ul>\n""" %(output_name.split("/")[-1]))
		    report_file_object.write("""<li>Only interactions where both members are a root or a linker protein are shown</li>\n""")

		report_file_object.write("""<li>Proteins whose description/function contain a keyword appear in red (keywords are listed on top of the table)</li>\n""")
		report_file_object.write("""<li>Over/under expressed proteins have a pink/green label under their 'information' column</li>\n""")
		report_file_object.write("""<li>Cellular location information is not very reliable: do not trust it too much.</li>\n""")
		report_file_object.write("""</ul><br></li>""")
                # END OF elif print_mode == "connecting":
            # END OF if report_file_object:

		
            
            if time_control:
                initial_time = time.time()
            # Output to stdout the interactions of the network 
            piana_api.print_interactions( id_type= this_print_id_type,
                                          output_file_object= output_target,
                                          output_format= "table",
                                          print_mode = print_mode,
                                          format_mode = format_mode,
                                          alternative_id_types= alternative_id_types,
                                          tax_id_value= this_tax_id,
                                          use_self_ints=use_self_ints,
                                          list_keywords= list_keywords,
                                          file_over_expressed= file_over_expressed,
                                          file_infra_expressed= file_infra_expressed,
                                          expression_id_type= expression_id_type,
					  dic_special_proteins_file= dic_special_proteins_file,
					  network_attributes_prefix = network_attributes_prefix)
            if time_control:
                sys.stderr.write("Printing interactions took %s seconds\n" %(time.time() - initial_time))
        

        elif option == "print-network":

            if verbose_command_sequence:
                sys.stderr.write("Printing out 'imageable' network to file %s\n" %(output_name))
            
            # Output to stdout the network file in DOT format
            piana_api.print_interactions( id_type= this_print_id_type,
                                          output_file_object= output_target,
                                          output_format= "network",
                                          print_mode = print_mode,
                                          format_mode = format_mode,
                                          alternative_id_types= alternative_id_types,
                                          tax_id_value= this_tax_id,
                                          use_self_ints=use_self_ints,
                                          list_keywords= list_keywords,
                                          file_over_expressed= file_over_expressed,
                                          file_infra_expressed= file_infra_expressed,
                                          expression_id_type= expression_id_type,
					  dic_special_proteins_file= dic_special_proteins_file,
					  network_attributes_prefix = network_attributes_prefix)
            
        # END OF elif option == "print-network":
 
        if output_name != "screen":
            output_target.close()

    # END OF elif option == "print-table" or option == "print-network":


    # -------------------------------------------------------------------------------------------------
    # option == print-table-db-intersection or print-network-db-intersection
    #
    # prints interactions that are in several databases in different flavors:
    #
    #    - print-table-db-intersection: prints table with interactions that appear in several databases
    #    - print-network-db-intersection: prints network with interactions that appear in several databases 
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: print-table-db-intersection;output-target=output_target;id-type=protein_code_type;print-mode=print_mode;list-dbs=dbname1:dbname2:...;format-mode=format_mode;format-mode=format_mode
    # configuration file syntax: print-network-db-intersection;output-target=output_target;id-type=protein_code_type;print-mode=print_mode;list-dbs=dbname1:dbname2:...;format-mode=format_mode;format-mode=format_mode
    #
    elif option == "print-table-db-intersection" or option == "print-network-db-intersection":


        db_list =[]

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            print_mode = raw_input(string_get_print_mode)
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species
            
            format_mode = raw_input(string_get_format_mode)

            
            if output_id_type is None:
                # if no output proteins code type was given, ask user to give one
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type= output_id_type

            # Ask the user to provide the databases where an interaction must appear in order to be printed
            ext_db_name = None

            while (ext_db_name !="end"):
                ext_db_name = raw_input(string_get_int_dbname)
                if ext_db_name != "end":
                    db_list.append(ext_db_name)
            # END OF while (db_name !="end"):

        # END OF if exec_mode == "interactive":
        
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_list_dbs] != "blank":
                colon_separated_dbs = arguments[arg_list_dbs]
                db_list = colon_separated_dbs.strip().split(":")
            else:
                raise ValueError("Trying to %s without providing any database in the configuration file" %option)


            if arguments[arg_print_mode] != "blank":            print_mode = arguments[arg_print_mode]
            else:                                               print_mode = "all"
            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:         
		if format_mode is None:
		    raise ValueError("You must set a format mode for your output")
            
            
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + print_mode + "." + option +  "." + format_mode 
                for dbname in db_list:
                    output_name = output_name + "_" + dbname
            
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")

            this_protein_species = output_proteins_species
        # END OF elif exec_mode == "batch":
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")


        if option == "print-table-db-intersection":
            
            if verbose_command_sequence:
                sys.stderr.write("Printing out intersection table to file %s with output type %s\n" %(output_name, this_print_id_type))
   
	    if report_file_object:
		report_file_object.write("""<li>This table contains <a href="%s">interactions in the network common to specific databases</a> (%s)</li>\n""" %(
			output_name.split("/")[-1], db_list))
              
                
            piana_api.print_interactions( id_type= this_print_id_type,
                                          output_file_object= output_target,
                                          output_format= "table",
                                          intersection_dbs = db_list,
                                          print_mode = print_mode,
                                          format_mode = format_mode,
                                          alternative_id_types= alternative_id_types,
                                          tax_id_value= this_tax_id,
                                          use_self_ints=use_self_ints,
                                          list_keywords= list_keywords,
                                          file_over_expressed= file_over_expressed,
                                          file_infra_expressed= file_infra_expressed,
                                          expression_id_type= expression_id_type,
					  dic_special_proteins_file= dic_special_proteins_file)
            
            
        elif option == "print-network-db-intersection":
            
            if verbose_command_sequence:
                sys.stderr.write("Printing out intersection network to file %s with output type %s\n" %(output_name, this_print_id_type))
                
            piana_api.print_interactions( id_type= this_print_id_type,
                                          output_file_object= output_target,
                                          output_format= "network",
                                          intersection_dbs = db_list,
                                          print_mode = print_mode,
                                          format_mode = format_mode,
                                          alternative_id_types= alternative_id_types,
                                          tax_id_value= this_tax_id,
                                          use_self_ints=use_self_ints,
                                          list_keywords= list_keywords,
                                          file_over_expressed= file_over_expressed,
                                          file_infra_expressed= file_infra_expressed,
                                          expression_id_type= expression_id_type,
					  dic_special_proteins_file= dic_special_proteins_file)
        # END OF elif option == "print-network-db-intersection":
            
        if output_name != "screen":
            output_target.close()

    # END OF elif option == "print-table-db-intersection" or option == "print-network-db-intersection":


    
    # -------------------------------------------------------------------------------------------------
    # option == print-all-prots-info or print-root-prots-info or print-connect-prots-info 
    #
    #  prints information about proteins in the network in different flavors:
    #
    #    - print-all-prots-info: about all proteins in the network
    #    - print-root-prots-info: about root proteins
    #    - print-connect-prots-info: about proteins that connect root nodes of the network
    # 
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: print-all-prots-info;output-target=output_target;id-type=protein_code_type;output-mode=output_mode;format-mode=format_mode
    # configuration file syntax: print-root-prots-info;output-target=output_target;id-type=protein_code_type;output-mode=output_mode;format-mode=format_mode
    # configuration file syntax: print-connect-prots-info;output-target=output_target;id-type=protein_code_type;output-mode=output_mode;format-mode=format_mode

    elif option == "print-all-prots-info" or option == "print-root-prots-info" or option == "print-connect-prots-info":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            
            if output_id_type is None:
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type = output_id_type

                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species
            
            output_mode = raw_input(string_get_output_mode)
            format_mode = raw_input(string_get_format_mode)

        # END OF if exec_mode == "interactive":
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

                
            if arguments[arg_output_mode] != "blank":            output_mode = arguments[arg_output_mode]
            else:                                                output_mode = default_output_mode
            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"
            
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + output_mode + "." + option + "." + format_mode 
            
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")

            this_protein_species = output_proteins_species
        # END OF elif exec_mode == "batch":
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")


        if verbose_command_sequence:
            sys.stderr.write("Printing out info for proteins to file %s\n" %(output_name) )


        if option == "print-all-prots-info" :

  
	    if report_file_object:
		report_file_object.write("""<li>This table contains <a href="%s">information for all network proteins</a>\n""" %(output_name.split("/")[-1]))
		
              
 

            piana_api.print_all_proteins_information( id_type=this_print_id_type,
                                                      output_file_object= output_target,
                                                      output_mode= output_mode,
                                                      format_mode = format_mode,
                                                      alternative_id_types=alternative_id_types,
                                                      list_keywords= list_keywords,
                                                      tax_id_value= this_tax_id,
                                                      file_over_expressed= file_over_expressed,
                                                      file_infra_expressed= file_infra_expressed,
                                                      expression_id_type= expression_id_type,
						      dic_special_proteins_file= dic_special_proteins_file)
        elif option == "print-root-prots-info":
  
	    if report_file_object:
		report_file_object.write("""<li>This table contains <a href="%s">information for  network root proteins</a>\n""" %(output_name.split("/")[-1]))
		

            piana_api.print_root_proteins_information( id_type=this_print_id_type,
                                                       output_file_object= output_target,
                                                       output_mode= output_mode,
                                                       format_mode = format_mode,
                                                       alternative_id_types=alternative_id_types,
                                                       list_keywords= list_keywords,
                                                       tax_id_value= this_tax_id,
                                                       file_over_expressed= file_over_expressed,
                                                       file_infra_expressed= file_infra_expressed,
                                                       expression_id_type= expression_id_type,
						       dic_special_proteins_file= dic_special_proteins_file)
            
        elif option == "print-connect-prots-info":
  
	    if report_file_object:
		report_file_object.write("""<li>This table contains <a href="%s">information for network linker proteins</a><br>\n""" %(output_name.split("/")[-1]))
		report_file_object.write("""<ul><li> A linker protein is one that is connected to at least two root proteins, and therefore,\n""")
		report_file_object.write(""" it is highly probable to be involved in the biological problem you are studying</li></ul>\n""")
		

            piana_api.print_connecting_proteins_information( id_type=this_print_id_type,
                                                             output_file_object= output_target,
                                                             output_mode= output_mode,
                                                             format_mode = format_mode,
                                                             alternative_id_types=alternative_id_types,
                                                             list_keywords= list_keywords,
                                                             tax_id_value= this_tax_id,
                                                             file_over_expressed= file_over_expressed,
                                                             file_infra_expressed= file_infra_expressed,
                                                             expression_id_type= expression_id_type,
							     dic_special_proteins_file= dic_special_proteins_file)
            
        # END OF elif option ==  "print-connect-prots-info":
  
	if report_file_object:
	    report_file_object.write("""<ul>""" )
	    report_file_object.write("""<li>Proteins whose description/function contain a keyword are in red (keywords are listed on top of the table)</li>\n""" )
	    report_file_object.write("""<li>Over/under expressed proteins have a pink/gree label under column 'other information' </li>\n""" )
	    report_file_object.write("""</ul></li><br>""" )

        if output_name != "screen":
            output_target.close()
        
                   
    # END OF elif option == "print-all-prots-info" or option == "print-root-prots-info" or
    #             option == "print-connect-prots-info":

 
    
    # -------------------------------------------------------------------------------------------------
    # option == classify-network-proteins
    #
    #  prints information about proteins in the network in different flavors:
    #
    #    - print-all-prots-info: about all proteins in the network
    #    - print-root-prots-info: about root proteins
    #    - print-connect-prots-info: about proteins that connect root nodes of the network
    # 
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: print-all-prots-info;output-target=output_target;id-type=protein_code_type

    elif option == "classify-network-proteins":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            
            if output_id_type is None:
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type = output_id_type

                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species
            
        # END OF if exec_mode == "interactive":
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + option 
            
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")

            this_protein_species = output_proteins_species
        # END OF elif exec_mode == "batch":
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")


        if verbose_command_sequence:
            sys.stderr.write("Printing out proteins and labels to file %s\n" %(output_name) )


        if report_file_object:
            report_file_object.write("""<li>This file contains <a href="%s">all network proteins with labels</a>\n""" %(output_name.split("/")[-1]))

        piana_api.print_all_proteins_with_network_class( id_type=this_print_id_type,
                                                         output_file_object= output_target,
                                                         alternative_id_types=alternative_id_types,
                                                         tax_id_value= this_tax_id,
                                                         file_over_expressed= file_over_expressed,
                                                         file_infra_expressed= file_infra_expressed,
                                                         expression_id_type= expression_id_type,
                                                         dic_special_proteins_file= dic_special_proteins_file)
  
	if report_file_object:
	    report_file_object.write("""<ul>""" )
	    report_file_object.write("""<li>class values can be:\n""" )
	    report_file_object.write("""<ul>""" )
	    report_file_object.write("""<li>root: it is a protein given by the user</li>\n""" )
	    report_file_object.write("""<li>linker: connects two or more root proteins</li>\n""" )
	    report_file_object.write("""<li>partner: only connected to one root protein</li>\n""" )
	    report_file_object.write("""</ul></li>""" )
	    report_file_object.write("""<li>special labels values can be those you set in your configuration file:\n""" )
	    report_file_object.write("""</li>""" )
	    report_file_object.write("""<li>expression values can be yes is over/under expressed) or no (no info available)\n""" )
	    report_file_object.write("""</li>""" )
	    report_file_object.write("""</ul></li><br>""" )

        if output_name != "screen":
            output_target.close()
        
                   
    # END OF elif option == "classify-network-proteins"

   

    # -------------------------------------------------------------------------------------------------
    # option == protein-code-2-protein-code
    # transforms codes from input-file (which are of type input-id-type) to output-id-type
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: protein-code-2-protein-code;output-target=blank
    #
    elif option == "protein-code-2-protein-code":

        if exec_mode == "interactive":

            output_name = raw_input(string_get_output_name)

            if input_file_name is None:
                input_file_name = raw_input(string_get_file_name)
           
            if input_id_type is None:
                input_id_type = raw_input("Input code: " + string_get_id_type)

            if output_id_type is None:
                output_id_type = raw_input("Output code: " + string_get_id_type)
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species
                
            format_mode = raw_input(string_get_format_mode)
               

        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)
            #  ( arguments is a dictionary with {argument_name=argument_value, ...} )

            if input_file_name is None or input_id_type is None or output_id_type is None or results_prefix is None:
                raise ValueError("input_file_name (%s), input_id_type (%s) and output_id_type (%s) and results_prefix (%s) cannot be None\n" %(
                    input_file_name,
                    input_id_type,
                    output_id_type,
                    results_prefix))
            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"
   
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                output_name = results_dir + results_prefix + "." + option + "."  + input_id_type + "2" + output_id_type + "." + format_mode

            this_protein_species = output_proteins_species
        # END OF elif exec_mode == "batch":
           
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
             
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")

        input_file_object = file(input_file_name, "r")
            
        """
        At this point, all required parameters have a value assigned to them. Proceed...
        """
            
        if verbose_command_sequence:
            sys.stderr.write("Printing out code equivalences of proteins in %s to file %s\n" %(input_file_name, output_name))
	
	if report_file_object:
	    report_file_object.write("""<li>This table contains <a href="%s">the translation from %s to %s of proteins in the initial input file of root proteins</a></li>\n""" %(
		    output_name.split("/")[-1],
		    input_id_type, output_id_type))

        piana_api.protein_code_2_protein_code(input_file_object= input_file_object,
                                              input_id_type= input_id_type,
                                              output_file_object= output_target,
                                              format_mode = format_mode,
                                              output_id_type= output_id_type,
                                              alternative_id_types= alternative_id_types,
                                              tax_id_value = this_tax_id)

        if output_name != "screen":
            output_target.close()
        
    # END OF elif option == "protein-code-2-protein-code":

    # -------------------------------------------------------------------------------------------------
    # option == protein-code-2-all-info
    # transforms codes from input-file (which are of type input-id-type) to output-id-type
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #
    elif option == "protein-code-2-all-info":

        if exec_mode == "interactive":

            output_name = raw_input(string_get_output_name)

            if input_file_name is None:
                input_file_name = raw_input(string_get_file_name)
           
            if input_id_type is None:
                input_id_type = raw_input("Input code: " + string_get_id_type)

            if output_id_type is None:
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type = output_id_type

                
            format_mode = raw_input(string_get_format_mode)
         
            output_mode = raw_input(string_get_output_mode)

            
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)
            #  ( arguments is a dictionary with {argument_name=argument_value, ...} )

            if input_file_name is None or input_id_type is None  or results_prefix is None:
                raise ValueError("input_file_name (%s), input_id_type (%s) and results_prefix (%s) cannot be None\n" %(
                    input_file_name,
                    input_id_type,
                    results_prefix))
             
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")
                
            if arguments[arg_output_mode] != "blank":            output_mode = arguments[arg_output_mode]
            else:                                                output_mode = default_output_mode
            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"
            
            if arguments[arg_output_target] != "blank":  output_name = arguments[arg_output_target]
            else:                                        output_name = results_dir + results_prefix  + "." + option + "." + format_mode

        # END OF elif exec_mode == "batch":

        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")

        input_file_object = file(input_file_name, "r")

        """
        At this point, all required parameters have a value assigned to them. Proceed...
        """
        if verbose_command_sequence:
            sys.stderr.write("Printing out info for proteins to file %s\n" %(output_name) )
            
	if report_file_object:
	    report_file_object.write("""<li>This table contains <a href="%s">information for proteins in the initial file of root proteins</a></li>\n""" %(output_name.split("/")[-1]))

        piana_api.print_file_proteins_information(input_file_object= input_file_object,
                                                  input_id_type=input_id_type,
                                                  output_file_object= output_target,
                                                  output_id_type=this_print_id_type,
                                                  output_mode = output_mode,
                                                  format_mode = format_mode,
                                                  list_keywords= list_keywords,
                                                  alternative_id_types= alternative_id_types,
                                                  file_over_expressed= file_over_expressed,
                                                  file_infra_expressed= file_infra_expressed,
                                                  expression_id_type= expression_id_type,
						  dic_special_proteins_file= dic_special_proteins_file)

        if output_name != "screen":
            output_target.close()

    # END OF elif option == "protein-code-2-all-info":

    # -------------------------------------------------------------------------------------------------
    # option == create-random-files
    # creates files that contain random proteins
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: create-random-files;num-files=num_files;num-prots=num_prots
    #
    elif option == "create-random-files":

        if exec_mode == "interactive":

            raise ValueError("This command cannot be used in interactive mode")
            
        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)
            #  ( arguments is a dictionary with {argument_name=argument_value, ...} )

            if arguments[arg_num_files] != "blank":
                num_files = int(arguments[arg_num_files])
            else:
                num_files = 1
                
            if arguments[arg_num_prots] != "blank":
                num_prots = int(arguments[arg_num_prots])
            else:
                num_prots = 1
                
            if arguments[arg_force_ints] != "blank":
                force_ints = arguments[arg_force_ints]
            else:
                force_ints = 1
                
            if arguments[arg_files_prefix] != "blank":
                files_prefix = arguments[arg_files_prefix]
            else:
                files_prefix = results_dir + results_prefix
                
            if arguments[arg_check_special] != "blank":
                check_special = arguments[arg_check_special]
            else:
                check_special = "no"
                
            if arguments[arg_check_expression] != "blank":
                check_expression = arguments[arg_check_expression]
            else:
                check_expression = "no"

        # END OF elif exec_mode == "batch":

        current_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=output_proteins_species)
        """
        At this point, all required parameters have a value assigned to them. Proceed...
        """
        if verbose_command_sequence:
            sys.stderr.write("Creating %s files with random proteins (output file names are %s.x.%s)\n" %(num_files,
                                                                                                          results_dir + results_prefix , option) )


        if check_special == "yes":
            # if user asked to have info about special proteins, set the dic to be used
            this_dic_special_proteins_file = dic_special_proteins_file
        else:
            # setting special proteins to None will make this call to produce a file without special info
            this_dic_special_proteins_file = None

        if check_expression == "yes":
            # if user asked to have info about expression, set the files to be used
            this_file_over_expressed = file_over_expressed
            this_file_infra_expressed = file_infra_expressed
            this_expression_id_type= expression_id_type
        else:
            # 
            this_file_over_expressed = None
            this_file_infra_expressed = None
            this_expression_id_type= None
             

        # user asked to create num_files: do it
        for i in range(num_files):

            if verbose_command_sequence:
                sys.stderr.write("%sof%s." %(i, num_files))
                
            this_output_name = files_prefix  + "." + str(i) + "." + option
            
            file_object = file(this_output_name, "w")
            
            new_random_dic = piana_api.get_random_protein_dic(tax_id_value=current_tax_id, dic_size=num_prots,
                                                              id_type= output_id_type,
                                                              alternative_id_types= alternative_id_types,
                                                              force_ints=force_ints, use_self_ints=use_self_ints,
                                                              list_source_dbs= list_source_dbs, inverse_dbs=inverse_dbs,
                                                              list_source_methods= list_source_methods, inverse_methods=inverse_methods,
                                                              file_over_expressed= this_file_over_expressed,
                                                              file_infra_expressed= this_file_infra_expressed,
                                                              expression_id_type= this_expression_id_type,
                                                              dic_special_proteins_file=this_dic_special_proteins_file )

            # new_random_dic is:
            #
            #       { protein1: {'expression':expression_value, 'labels':[label1, label2, ...],
            #         protein2: {'expression':expression_value, 'labels':[label1, label2, ...],
            #         ..................
            #       }

            # print results (one protein or the protein with its special labels) to output file
            for one_protein in new_random_dic:
                file_object.write("%s" %one_protein )

                if check_expression == "yes":
                    file_object.write("\texpression=%s" %(new_random_dic[one_protein]['expression']))
                    
                if check_special == "yes":
                    for one_item in new_random_dic[one_protein]['labels']:
                        file_object.write("\tlabel=%s" %one_item)

                file_object.write("\n")
                
            # END OF for one_protein in new_random_list:
            
        # END OF for i in range(number_of_files):
      
        if verbose_command_sequence:
            sys.stderr.write("\n")
    # END OF elif option == "create-random-files":


    # -------------------------------------------------------------------------------------------------
    # option == expand-interactions 
    # apply expansion to network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: expand-interactions;expansion-type=expansion_type;expansion-nodes=expansion_nodes;expansion-threshold=expansion_threshold
    #
    elif option == "expand-interactions":

        if exec_mode=="interactive":

            expansion_type = raw_input(string_get_expansion_type )
            expansion_nodes = raw_input(string_get_nodes_to_expand)
            exp_output_mode = raw_input(string_get_exp_output_mode)
            expansion_threshold = int(raw_input(string_get_expansion_threshold))
            
            if exp_output_mode == "print":
                # name and type only needed when printing information
                output_name = raw_input(string_get_output_name)

                if output_id_type is None:
                    this_print_id_type = raw_input("Expansion output:" + string_get_id_type)
                else:
                    this_print_id_type = output_id_type
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species

        # END OF if exec_mode=="interactive":
        
        elif exec_mode=="batch":
            
            if arguments[arg_expansion_type] != "blank":
                expansion_type = arguments[arg_expansion_type]
            else:
                raise ValueError("Trying to expand interactions without giving a expansion-type in configuration file")

            if arguments[arg_expansion_nodes] != "blank":
                expansion_nodes = arguments[arg_expansion_nodes]
            else:
                raise ValueError("Trying to expand interactions without setting expansion nodes in configuration file")
            
            if arguments[arg_expansion_threshold] != "blank":
                expansion_threshold = int(arguments[arg_expansion_threshold])
            else:
                raise ValueError("Trying to expand interactions without setting expansion expansion_threshold in configuration file")
  
            if arguments[arg_exp_output_mode] != "blank":
                exp_output_mode = arguments[arg_exp_output_mode]
            else:
                raise ValueError("Trying to expand interactions without setting output mode in configuration file")
                       
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:
                    this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")

            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + option + "." + expansion_type + "_thres" + str(expansion_threshold) + "." + expansion_nodes

            this_protein_species = output_proteins_species
        # END OF elif exec_mode=="batch":

        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)
        
        # TO DO!!! Introduce interaction confidence based on: - how many proteins share that certain characteristic
        #                                                     - how many interactions coming from a single protein


        if exp_output_mode == "print":
            # if output mode of expansion is print, open the file where results will be printed
            if output_name == "screen":
                output_target = sys.stdout
            else:
                output_target = file(output_name,"w")
        else:
            output_target = None

        """
        At this point, we have all the parameters required. Proceed...
        """
        if verbose_command_sequence:
            sys.stderr.write("Expanding interactions for current network\n")
            
	if report_file_object:

	    if exp_output_mode == "add":
		report_file_object.write("""<li>Predictions by 'propagating interactions from %s' were added to the network</li>\n""" %(expansion_type))

	    elif exp_output_mode == "print":
		report_file_object.write("""<li>This table contains <a href="%s">predictions by 'propagating interactions from %s'</a></li>\n""" %(output_name.split("/")[-1], expansion_type))
				     
			     

        piana_api.expand_piana_graph_interactions(expansion_type=expansion_type, expansion_mode=expansion_nodes,
                                                  expansion_threshold=expansion_threshold, hub_threshold=hub_threshold,
                                                  exp_output_mode=exp_output_mode,  output_file_object=output_target,
                                                  id_type=this_print_id_type,
                                                  output_tax_id= this_tax_id,
                                                  alternative_id_types=alternative_id_types,
                                                  use_self_ints = use_self_ints,
                                                  list_source_dbs=list_source_dbs,
                                                  inverse_dbs= inverse_dbs,
                                                  list_source_methods=list_source_methods,
                                                  inverse_methods= inverse_methods)

  
        if exp_output_mode == "print" and output_name != "screen":
            output_target.close()


    # END OF elif option == "expand-interactions":

    # -------------------------------------------------------------------------------------------------
    # option == match-proteins-to-pathways
    # finds which pathways appear in the current network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #  match-proteins-to-pathways;pathways-dir=pathways_dir;id-type=blank;force-expression=force_expression;format-mode=format_mode
    #
    elif option == "match-proteins-to-pathways":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            
            if input_id_type is None:
                # if no output proteins code type was given, ask user to give one
                pathway_id_type = raw_input("Pathway proteins: " + string_get_id_type)
            else:
                pathway_id_type = output_id_type

            if output_id_type is None:
                # if no output proteins code type was given, ask user to give one
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type = output_id_type
                
            format_mode = raw_input(string_get_format_mode)
            pathways_dir = raw_input(string_get_pathways_dir)
   
	    
        # END OF if exec_mode == "interactive":

        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_pathways_dir] != "blank":            pathways_dir = arguments[arg_pathways_dir]
            else:                                                raise ValueError("pathways-dir  must be set when using command match-proteins-to-pathways")

            if arguments[arg_pathway_type] != "blank":   pathway_id_type = arguments[arg_pathway_type]
            else:        
		if input_id_type is not None:
		    pathway_id_type = input_id_type
                else:
                    raise ValueError(" identifiers type for pathway proteins cannot be None (use either input identifiers type or command arguments to set it")

            if arguments[arg_id_type] != "blank":   this_print_id_type = arguments[arg_id_type]
            else:        
		if output_id_type is not None:
		    this_print_id_type = output_id_type
                else:
                    raise ValueError(" identifiers type for outputting matched pathway proteins cannot be None (use either output identifiers type or command arguments to set it")
 

            if arguments[arg_force_expression] != "blank":       force_expression = arguments[arg_force_expression]
            else:                                                raise ValueError("force-expression must be set when using command match-proteins-to-pathways") 

            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"

            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix  + "." + option + ".expr_" + force_expression + "." + format_mode
            

        # END OF elif exec_mode == "batch":
        
        if output_name == "screen":            output_target = sys.stdout
        else:                                  output_target = file(output_name,"w")
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=input_proteins_species)

        """
        At this point, all parameters are set... Proceed..
        """

        if verbose_command_sequence:
            sys.stderr.write("\nPrinting pathway matches to output file %s\n" %(output_name))

	if report_file_object:
	    report_file_object.write("""<li>This table contains <a href="%s">information on which known pathways have overlapping proteins with the current network</a><ul>\n""" %(output_name.split("/")[-1]))
	    if force_expression == "yes":
		report_file_object.write("""<li> Only proteins that were under/over expressed were used to create this table</li>\n""")
	    elif force_expression == "no":
		report_file_object.write("""<li> All network proteins were used to create this table, regardless of their expression</li>\n""")

	    report_file_object.write("""<li> All lists of proteins in this table can be clicked to obtain their full description.</li>\n""")
	    report_file_object.write("""</ul></li><br>""")
				     
	piana_api.print_matched_pathways(pathways_dir=pathways_dir,
					 force_expression=force_expression,
					 output_file_object=output_target,
					 results_dir= results_dir,
					 files_prefix=results_prefix,
					 format_mode = format_mode,
					 pathway_id_type=pathway_id_type,
					 output_id_type=this_print_id_type,
					 alternative_id_types=alternative_id_types,
					 tax_id= this_tax_id,
					 file_over_expressed= file_over_expressed,
					 file_infra_expressed= file_infra_expressed,
					 expression_id_type= expression_id_type,
					 dic_special_proteins_file= dic_special_proteins_file)


        if output_name != "screen":
            output_target.close()

	# TO DO!!!!!!! It would make a lot more sense to read the expression files just once, and update the node attributes of the graph with that
	# information so then each nodes carries with it the expression info and we don't have to do it multiple times. 
	# The way it is done right now is completely inefficient!!!


    # END OF elif option == "match-proteins-to-pathways":

    # -------------------------------------------------------------------------------------------------
    # option == match-proteins-to-gos
    # finds the frequency of the GO terms that appear in the current network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #  match-proteins-to-gos;id-type=blank;force-expression=force_expression;format-mode=format_mode;term-type=term_type;level-threshold=level_threshold;go-dbname=go_dbname;go-dbhost=go_dbhost;go-dbuser=go_dbuser;go-dbpass=go_dbpass
    #
    elif option == "match-proteins-to-gos":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            
            if input_id_type is None:
                # if no output proteins code type was given, ask user to give one
                this_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_id_type = output_id_type
                
            format_mode = raw_input(string_get_format_mode)

            term_type = raw_input(string_get_term_type)
            level_threshold = int(raw_input(string_get_level_threshold))

            go_dbname = raw_input(string_get_go_dbname)
            go_dbhost = raw_input(string_get_go_dbhost)
            go_dbuser = raw_input(string_get_go_dbuser)
            go_dbpass = raw_input(string_get_go_dbpass)
   
	    
        # END OF if exec_mode == "interactive":

        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_term_type] != "blank":        term_type = arguments[arg_term_type]
            else:                                          raise ValueError("Trying to cluster by go terms without giving a term type")
            
            if arguments[arg_level_threshold] != "blank":  level_threshold = int(arguments[arg_level_threshold])
            else:                                          raise ValueError("Trying to cluster by go terms without giving a level threshold")
            
            if arguments[arg_id_type] != "blank":   this_id_type = arguments[arg_id_type]
            else:        
		if input_id_type is not None:
		    this_id_type = output_id_type
                else:
                    raise ValueError(" output identifiers type for printing matched GO terms cannot be None (use either output identifiers type or command arguments to set it")
 

            if arguments[arg_force_expression] != "blank":       force_expression = arguments[arg_force_expression]
            else:                                                raise ValueError("force-expression must be set when using command match-proteins-to-gos") 

            if arguments[arg_go_dbname] != "blank":              go_dbname = arguments[arg_go_dbname]
            else:                                                raise ValueError("go_dbname must be set when using command match-proteins-to-gos") 

            if arguments[arg_go_dbhost] != "blank":              go_dbhost = arguments[arg_go_dbhost]
            else:                                                raise ValueError("go_dbhost must be set when using command match-proteins-to-gos") 

            if arguments[arg_go_dbuser] != "blank":              go_dbuser = arguments[arg_go_dbuser]
            else:                                                go_dbuser = None

            if arguments[arg_go_dbpass] != "blank":              go_dbpass = arguments[arg_go_dbpass]
            else:                                                go_dbpass = None

            
            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"

            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix  + "." + option + ".type_" + term_type + ".expr_" + force_expression + "." + format_mode
            

        # END OF elif exec_mode == "batch":
        
        if output_name == "screen":            output_target = sys.stdout
        else:                                  output_target = file(output_name,"w")
        
        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=input_proteins_species)

        """
        At this point, all parameters are set... Proceed..
        """

        if verbose_command_sequence:
            sys.stderr.write("\nPrinting GO terms frequencies to output file %s\n" %(output_name))

	if report_file_object:
	    report_file_object.write("""<li>This table contains <a href="%s">information on which GO terms (type %s) have overlapping proteins with the current network</a></li>\n""" %(
		    output_name.split("/")[-1], term_type))
				     
	piana_api.print_matched_gos(force_expression=force_expression,
				    output_file_object=output_target,
				    format_mode = format_mode,
				    output_id_type=this_id_type,
				    tax_id= this_tax_id,
				    term_type=term_type,
				    level_threshold=level_threshold, 
				    go_dbname=go_dbname, go_dbhost=go_dbhost, go_dbuser=go_dbuser, go_dbpass=go_dbpass,
				    file_over_expressed= file_over_expressed,
				    file_infra_expressed= file_infra_expressed,
				    expression_id_type= expression_id_type,
				    dic_special_proteins_file= dic_special_proteins_file)


        if output_name != "screen":
            output_target.close()

	# TO DO!!!!!!! It would make a lot more sense to read the expression files just once, and update the node attributes of the graph with that
	# information so then each nodes carries with it the expression info and we don't have to do it multiple times. 
	# The way it is done right now is completely inefficient!!!


    # END OF elif option == "match-proteins-to-gos":
      
    # -------------------------------------------------------------------------------------------------
    # option == match-proteins-to-spots
    # identifies spots in a 2D gel by matching MW and/or IP to proteins in the network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #  match-proteins-to-spots;output-target=output_target;id-type=protein_code_type;list-mw-error=mw_%er1:..:mw_%erN;list-ip-error=ip_%er1:..:ip_%erN
    #
    elif option == "match-proteins-to-spots":

        if exec_mode == "interactive":
            output_name = raw_input(string_get_output_name)
            
            if output_id_type is None:
                # if no output proteins code type was given, ask user to give one
                this_print_id_type = raw_input("Output: " + string_get_id_type)
            else:
                this_print_id_type = output_id_type
                
            if spots_file_name is None:
                spots_file_name = raw_input(string_get_spots_file )

            format_mode = raw_input(string_get_format_mode)
   
            # In interactive mode we do not let the user choose the error bounds
            molecular_error_bounds   = default_molecular_error_bounds  
            #       - set here the isoelectric error bounds for which you want to perform the matching
            isoelectric_error_bounds = default_isoelectric_error_bounds

        # END OF if exec_mode == "interactive":

        elif exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)

            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"
            
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix  + "." + option + "." + format_mode
            
            if arguments[arg_id_type] != "blank":
                this_print_id_type = arguments[arg_id_type]
            else:
                if output_id_type is not None:   this_print_id_type = output_id_type
                else:
                    raise ValueError("output identifiers type cannot be None")
                
            if spots_file_name is None:
                if arguments[arg_spots_file_name] != "blank":   spots_file_name = arguments[arg_spots_file_name]
                else:
                    raise ValueError("Trying to match proteins to spots without giving a spots file in command line or configuration file")

            # get error bounds, either from default or from configuration file
            if  arguments[arg_list_mw_error] == "blank":   molecular_error_bounds   = default_molecular_error_bounds
            else:
                tmp_molecular_error_bounds = arguments[arg_list_mw_error].strip().split(":")
                molecular_error_bounds = []
                for mw_error in tmp_molecular_error_bounds:
                    molecular_error_bounds.append(float(mw_error))

            if arguments[arg_list_ip_error] == "blank":   isoelectric_error_bounds   = default_isoelectric_error_bounds
            else:
                tmp_isoelectric_error_bounds = arguments[arg_list_ip_error].strip().split(":")
                isoelectric_error_bounds = []
                for ip_error in tmp_isoelectric_error_bounds:
                    isoelectric_error_bounds.append(float(ip_error))

        # END OF elif exec_mode == "batch":
        
        if output_name == "screen":            output_target = sys.stdout
        else:                                  output_target = file(output_name,"w")

        spots_file_object = file(spots_file_name, "r")
        
        """
        At this point, all parameters are set... Proceed..
        """

        if verbose_command_sequence:
            sys.stderr.write("\nPrinting spot matches to output file %s\n" %(output_name))

	if report_file_object:
	    report_file_object.write("""<li>This table contains <a href="%s">information on which 2D spots have corresponding proteins in the current network</a></li>\n""" %(
		    output_name.split("/")[-1]))

        piana_api.print_spot_protein_correspondence(spots_file_object= spots_file_object,
                                                    molecular_error_bounds=molecular_error_bounds, isoelectric_error_bounds=isoelectric_error_bounds,
                                                    output_file_object=output_target,
                                                    format_mode = format_mode,
                                                    output_id_type=this_print_id_type, alternative_id_types=alternative_id_types)


        spots_file_object.close()
        
        if output_name != "screen":
            output_target.close()
    # END OF elif option == "match-proteins-to-spots"

    # -------------------------------------------------------------------------------------------------
    # option == find-shortest-route 
    # find the route between two proteins (ie. intermediate proteins)
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: 
    # find-shortest-route;protein-a-name=protein_a_name;protein-b-name=protein_a_name;id-type=blank;output-target=blank;format-mode=format_mode

    elif option == "find-shortest-route":

        if exec_mode=="interactive":

            protein_a_name = raw_input("protein A: " + string_get_protein)
            protein_b_name = raw_input("protein B: " + string_get_protein)

            if input_id_type is None:
                this_id_type = raw_input("Input identifier type: " + string_get_id_type)
            else:
                this_id_type = input_id_type
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species

        # END OF if exec_mode=="interactive":
        
        elif exec_mode=="batch":
            
            protein_a_name = arguments[arg_protein_a_name]
            protein_b_name = arguments[arg_protein_b_name]
            
            if arguments[arg_id_type] != "blank":           this_id_type = arguments[arg_id_type]
            else:                                           this_id_type = input_id_type
                
            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + option + "." + protein_a_name + "_2_" + protein_b_name

            this_protein_species = output_proteins_species
        # END OF elif exec_mode=="batch":

        
        if protein_a_name is None or protein_b_name is None or this_id_type is None or output_id_type is None or this_protein_species is None:
            raise ValueError("In find-shortest-route, protein a name (%s) , protein b name (%s) , identifier type (%s) protein species (%s) and output type (%s) are required\n" %(
                protein_a_name,
                protein_a_name,
                this_id_type,
                this_protein_species,
                output_id_type))

        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)

        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")

	if report_file_object:
	    report_file_object.write("""<li>This file describes the <a href="%s">shortest route between protein %s and protein %s</a></li>\n""" %(
		    output_name.split("/")[-1], protein_a_name, protein_b_name))

        piana_api.print_shortest_route(protein_a_name=protein_a_name, protein_b_name=protein_b_name,
                                       input_id_type=this_id_type, output_id_type=output_id_type,
                                       alternative_id_types=alternative_id_types,
                                       output_file_object=output_target, format_mode="txt", tax_id_value=this_tax_id)

    # -------------------------------------------------------------------------------------------------
    # option == find-distance-group 
    # find proteins that are at a certain distance from a query protein
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #      find-distance-group;protein-name=protein_name;id-type=blank;distance=distance;output-target=blank;info=blank;format-mode=format_mode
    #
    elif option == "find-distance-group":

        if exec_mode=="interactive":

            query_protein = raw_input(string_get_protein)
            distance = int( raw_input(string_get_distance) )
            info = raw_input(string_get_info)

            if input_id_type is None:
                this_id_type = raw_input("Input identifier type: " + string_get_id_type)
            else:
                this_id_type = input_id_type
                
            if output_proteins_species is None: this_protein_species = raw_input("Output protein species: "+ string_get_species_name)
            else:                               this_protein_species = output_proteins_species

        # END OF if exec_mode=="interactive":
        
        elif exec_mode=="batch":
            
            query_protein = arguments[arg_protein_name]

            if arguments[arg_format_mode] != "blank":            format_mode = arguments[arg_format_mode]
            else:                                                format_mode = "txt"
            
            if arguments[arg_id_type] != "blank":           this_id_type = arguments[arg_id_type]
            else:                                                this_id_type = input_id_type
                
            if arguments[arg_distance] != "blank":
                if arguments[arg_distance] != "all":     distance = int(arguments[arg_distance])
                else:                                    distance = "all"
            else:
                raise ValueError("distance required for find-distance-group\n")
                
            if arguments[arg_info] != "blank":
                info = arguments[arg_info]
            else:
                raise ValueError("You must set argument info for 'command distance' in your configuration file")

            if arguments[arg_output_target] != "blank":
                output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + option + "." + str(distance) + "." + format_mode

            this_protein_species = output_proteins_species
        # END OF elif exec_mode=="batch":

        
        if query_protein is None or this_id_type is None or output_id_type is None or this_protein_species is None:
            raise ValueError("In find-distance-group, protein name (%s) , identifier type (%s) protein species (%s) and output type (%s) are required\n" %(
                protein_name,
                this_id_type,
                this_protein_species,
                output_id_type))

        this_tax_id = piana_api.get_one_tax_id_from_species_name(species_name=this_protein_species)

        # chapucilla para poder debuggear usando proteinPianas...
        if this_id_type == "proteinPiana":
            query_protein = int(query_protein)
        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")

        """
        At this point we have all the required parameters... Proceed...
        """
            
        if verbose_command_sequence:
            sys.stderr.write("\nPrinting proteins at distance %s from %s to output file %s\n" %(distance, query_protein ,output_name))

	if report_file_object:
	    report_file_object.write("""<li>This file describes the <a href="%s">proteins that are at distance %s from protein %s </a></li>\n""" %(
		    output_name.split("/")[-1], distance, query_protein))

        try:
            piana_api.print_proteins_at_distance_x(query_protein=query_protein, distance=distance, input_id_type=this_id_type,
                                                   output_id_type=output_id_type, alternative_id_types=alternative_id_types,
                                                   output_file_object= output_target, format_mode=format_mode, info= info,
                                                   tax_id_value=this_tax_id)
        except:
            sys.stderr.write("We cannot find proteins at distance %s from protein %s because this protein is not in the network\n" %(distance,
                                                                                                                                     query_protein ))
            

        
        if output_name != "screen":
            output_target.close()

    # END OF elif option == "find-distance-group":
    

   
    # -------------------------------------------------------------------------------------------------
    # option ==  find-cure-for;illness=illness
    # This is obviously an inside joke that you can use to amuse your colleagues
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #
    elif option == "find-cure-for":

        if exec_mode == "batch":
            if arguments["illness"] != "blank":
                illness = arguments["illness"]
            else:
                raise ValueError("Trying to cure an illness with PIANA without telling me the illness to be cured")

        
        sys.stdout.write("\nUsing PIANA to find a cure for %s\n" %illness)
        time.sleep(4)
        sys.stdout.write("\nWaw! I was fast! I've found the cure for %s\n" %illness)
        time.sleep(4)
        sys.stdout.write("\nThe answer is.... ")
        sys.stdout.flush()
        time.sleep(3)
        sys.stdout.write(".... ")
        sys.stdout.flush()
        time.sleep(3)
        sys.stdout.write(".... \n")
        time.sleep(3)
        sys.stdout.write("\nArjjjj!!!! Somebody threw me his knife! Arrjjjjj!\n")
        time.sleep(2)
        sys.stdout.write("\nI'm going to tell you the cure before they arrive\n")
        time.sleep(2)
        sys.stdout.write("\nArjjjj!!!! Again! Another knife\n")
        time.sleep(2)
        sys.stdout.write("\nThe cure for %s is...\n" %illness)
        time.sleep(2)
        sys.stdout.write("\nSorry, I'm dead: run PIANA again for another illness\n" )
        time.sleep(2)

    # END OF elif option == "find-cure-for":
   
    # -------------------------------------------------------------------------------------------------
    # option ==  calculate-imotifs
    #
    # prints out all iMotifs found for all levels of the clustering, with results at each level
    # this command is used to print out the results that will be later used to find out which
    # are the best parameters for finding CIRs and perform the evaluation
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:
    #
    elif option == "calculate-imotifs":

        if exec_mode == "batch":
            # if batch mode, read arguments from configuration file (which has been previously parsed)
            
            if arguments[arg_similarity_mode] != "blank":
                similarity_mode = arguments[arg_similarity_mode]
            else:
                raise ValueError("Trying to %s without setting the similarity_mode in configuration file" %option)
            
            if arguments[arg_minimum_score] != "blank":
                minimum_score = float(arguments[arg_minimum_score])
            else:
                raise ValueError("Trying to %s without setting the similarity_mode in configuration file" %option)
            
            if arguments[arg_imotif_mode] != "blank":
                imotif_mode = arguments[arg_imotif_mode]
            else:
                raise ValueError("Trying to %s without setting the %s in configuration file" %(option, arg_imotif_mode))
            
            if arguments[arg_num_ints_thres] != "blank":
                num_ints_thres = int(arguments[arg_num_ints_thres])
            else:
                raise ValueError("Trying to %s without setting the %s in configuration file" %(option, arg_num_ints_thres))
            
            if arguments[arg_common_ints_thres] != "blank":
                common_ints_thres = int(arguments[arg_common_ints_thres])
            else:
                raise ValueError("Trying to %s without setting the %s in configuration file" %(option, arg_common_ints_thres))
            
            if arguments[arg_global_map] != "blank":
                global_map = arguments[arg_global_map]
            else:
                global_map = None

        else:
            raise ValueError("This command cannot be used in interactive mode\n")

        if not results_prefix:
            results_prefix= "imotif_calculation"



        if verbose_command_sequence:
            sys.stderr.write("\nCreating iMotif results files --")
        
        piana_api.do_imotif_clustering(similarity_mode=similarity_mode,
                                       minimum_score=minimum_score,
                                       imotif_mode=imotif_mode,
                                       num_ints_thres=num_ints_thres,
                                       common_ints_thres=common_ints_thres,
                                       global_map=global_map,
                                       results_prefix=results_prefix,
                                       results_dir= results_dir,
                                       use_self_ints = use_self_ints,
                                       list_source_dbs=list_source_dbs , inverse_dbs=inverse_dbs,
                                       list_source_methods=list_source_methods, inverse_methods=inverse_methods,
                                       hub_threshold = hub_threshold,
                                       input_id_type=input_id_type,
                                       input_tax_id=input_tax_id )

    # -------------------------------------------------------------------------------------------------
    # option ==  cluster-by-go-terms
    # clusters the ppi network using go terms
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:cluster-by-go-terms;output-target=blank;term-type=blank;score-threshold=blank;level-threshold=blank;distance-threshold=blank;rep-term=blank;print-id=blank
    #
    elif option == "cluster-by-go-terms":

        if exec_mode=="interactive":

            output_name = raw_input(string_get_output_name )
            term_type = raw_input(string_get_term_type)
            go_score_threshold = float(raw_input(string_get_go_score_threshold))
            sim_mode = raw_input(string_get_sim_mode)
            level_threshold = int(raw_input(string_get_level_threshold))
            distance_threshold = int(raw_input(string_get_distance_threshold))
            rep_term = raw_input(string_get_rep_term)
            print_id = raw_input(string_get_print_id)

        # END OF if exec_mode=="interactive":
        
        elif exec_mode=="batch":

            if arguments[arg_term_type] != "blank":        term_type = arguments[arg_term_type]
            else:                                          raise ValueError("Trying to cluster by go terms without giving a term type")
            
            if arguments[arg_score_threshold] != "blank":  go_score_threshold = float(arguments[arg_score_threshold])
            else:                                          raise ValueError("Trying to cluster by go terms without giving a score threshold")
            
            if arguments[arg_sim_mode] != "blank":         sim_mode = arguments[arg_sim_mode]
            else:                                          raise ValueError("Trying to cluster by go terms without giving a mode for similarity function")
            
            if arguments[arg_level_threshold] != "blank":  level_threshold = int(arguments[arg_level_threshold])
            else:                                          raise ValueError("Trying to cluster by go terms without giving a level threshold")
            
            if arguments[arg_distance_threshold] != "blank":  distance_threshold = int(arguments[arg_distance_threshold])
            else:                                             raise ValueError("Trying to cluster by go terms without giving a distance threshold")
  
  
            if arguments[arg_rep_term] != "blank":         rep_term = arguments[arg_rep_term]
            else:                                          raise ValueError("Trying to cluster by go terms without giving a rep term ")
            
            if arguments[arg_print_id] != "blank":         print_id = arguments[arg_print_id]
            else:                                          raise ValueError("Trying to cluster by go terms without giving a print id")

            if arguments[arg_output_target] != "blank":    output_name = arguments[arg_output_target]
            else:
                if results_prefix is None:
                    raise ValueError("Using default output name without giving a results prefix")

                output_name = results_dir + results_prefix + "." + str(go_score_threshold) + "." + term_type + "." + str(level_threshold) + "." + sim_mode +"." + str(distance_threshold) + "." + option
        # END OF elif exec_mode=="batch":

        
        if output_name == "screen":
            output_target = sys.stdout
        else:
            output_target = file(output_name,"w")

        
        """
        At this point, all parameters are set... Proceed..
        """

        if verbose_command_sequence:
            sys.stderr.write("\nClustering PPI network to a GO clustered network in output file %s\n" %(output_name))

        piana_api.create_go_clustered_network(output_target=output_target,
                                              term_type=term_type, score_threshold=go_score_threshold,
                                              sim_mode=sim_mode, level_threshold=level_threshold,
                                              distance_threshold=distance_threshold, rep_term=rep_term, print_id=print_id)

        if output_name != "screen":
            output_target.close()
    
    # -------------------------------------------------------------------------------------------------
    # option == print-parameters
    # print parameters used to create current network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax:  no configuration command provided for printing parameters
    #
    elif option == "print-parameters":

        sys.stdout.write("-----------------------------------------------------------\n")
        sys.stdout.write("Network depth set to: %s\n" %depth)
        sys.stdout.write("Hub threshold set to: %s\n" %hub_threshold)
        sys.stdout.write("Input identifiers type set to: %s\n" %input_id_type)
        sys.stdout.write("Output identifiers type set to: %s\n" %output_id_type)
        sys.stdout.write("-----------------------------------------------------------\n")
        
    # END OF elif option == "print-parameters":

    # -------------------------------------------------------------------------------------------------
    # option == modify-parameters  !!NOT WORKING!!!
    #  modifies current parameters 
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: under work... doesn't currently work in batch mode
    #
    elif option == "modify-parameters":
        
        if exec_mode == "interactive":
            sys.stdout.write("Write here the new parameter values (leave empty if you want to leave it unchanged)\n")
            
            temp_input_id_type = raw_input("Type for input: " + string_get_id_type)
            if temp_input_id_type != "":
                input_id_type   = temp_input_id_type
                
            temp_output_id_type = raw_input("Type for output: " + string_get_id_type)
            if temp_output_id_type != "":
                output_id_type =temp_output_id_type
               
            temp_depth = int(raw_input(string_get_depth))
            if temp_depth != "":
               depth =temp_depth
               
            temp_hub_threshold = int(raw_input(string_get_hub_threshold))
            if temp_hub_threshold != "":
               hub_threshold =temp_hub_threshold
               
            # TO DO!!! Right now, I don't let the user change pianaDB during execution (it has no effect on piana_access)
            temp_piana_dbname = raw_input(string_get_piana_dbname)
            if temp_piana_dbname != "":
              piana_dbname  =temp_piana_dbname
              
            temp_piana_dbhost = raw_input(string_get_piana_dbhost)
            if temp_piana_dbhost != "":
               piana_dbhost =temp_piana_dbhost
               
            temp_piana_dbuser = raw_input(string_get_piana_dbuser)
            if temp_piana_dbuser != "":
               piana_dbuser =temp_piana_dbuser
               
            temp_piana_dbpass = raw_input(string_get_piana_dbpass)
            if temp_piana_dbpass != "":
               piana_dbpass =temp_piana_dbpass

            temp_results_prefix = raw_input(string_get_results_prefix)
            if temp_results_prefix != "":
               results_prefix =temp_results_prefix

            temp_results_directory = raw_input(string_get_results_directory)
            if temp_results_directory != "":
               results_directory =temp_results_directory
            
        elif exec_mode == "batch":
            # TO DO!!!! modify parameters by reading arguments
            pass
        
                    
    # END OF elif option == "modify-parameters":

    # -------------------------------------------------------------------------------------------------
    # option == filter  !!NOT WORKING!!!
    # apply filters to network
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: under work... doesn't currently work in batch mode
    #
    elif option == "filter":

        sys.stdout.write("Function still not implemented\n")
        continue
        list_dbs = []

        if include_dip:
            list_dbs.append("dip")

        if include_string:
            list_dbs.append("string")

        if include_posas:
            list_dbs.append("posas")

        if include_ori:
            list_dbs.append("ori")

        if include_bind:
            list_dbs.append("bind")

        db_filter = FilterInteractionSourceDB(list_dbs=list_dbs, belong_mode="filter_belonging")

        filters = []
        filters.append(db_filter)

        if verbose:
            sys.stderr.write( "Applying filters for db list: %s \n" %list_dbs)

        piana_graph.filter(filters)
        
    # END OF elif option == "filter":

    # -------------------------------------------------------------------------------------------------
    # option == ?
    # prints PIANA commands
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: this is not to be called from configuration file: only used in interactive mode
    #

    elif option == "?":
        pass
 
    # -------------------------------------------------------------------------------------------------
    # option == exit
    # exits the program
    # -------------------------------------------------------------------------------------------------
    # configuration file syntax: exit;
    #
    elif option == "exit":

        print "total time: %s seconds" %(time.time()-initial_time)
        
        sys.stderr.write("Exiting....\n")
    
	if report_file_object:

	    report_file_object.write("""</ul><br><br><hr><br> \n""" )
	    report_file_object.write("""  If you use these results, please do not forget to make reference to our paper in bioinformatics:<br><br>  \n""")
	    report_file_object.write(""" R. Aragues, D. Jaeggi and B. Oliva</b><br> "PIANA: Protein Interactions and Network Analysis" <br>
                                     <i>Bioinformatics. 2006 Apr 15;22(8):1015-7</i> (2006) <a href="http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=Retrieve&db=pubmed&list_uids=16510498">[PubMed]</a>\n""")
	    report_file_object.write("""<br><br> <small>Results generated at %s</small> \n""" %(time.strftime(" %X - %B %d, %Y")))
	    report_file_object.close()

    # END OF elif option == "exit"

    # -------------------------------------------------------------------------------------------------
    # option unknown
    # -------------------------------------------------------------------------------------------------
    else:

        if exec_mode == "interactive":
            sys.stdout.write("\tCommand <%s> is unknown\n" %option)

        elif exec_mode == "batch":
            raise ValueError("Trying to execute unknown command %s\n" %option)
        
    # END OF else: (if option == xxx)
        
        
    
# END OF while(option != exit)

sys.exit(os.EX_OK)


