"""
File        : predict_interactions_using_common_feature.py
Author      : Ramon Aragues 
Creation    : 14.2.2007
Contents    : predicts protein-protein interactions based
              on proteins having a common feature (e.g. SCOP,
              COG, ...)
Called from : command line 

=======================================================================================================
Refer to README.populate_piana_db for a detailed description of this predictor

"""


# predict_interactions_using_common_feature.py: predicts protein-protein
#             interactions based
#             on proteins having a common feature (e.g. SCOP,
#             COG, ...)
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline
import cPickle
import copy

# import numarray
from plplot import *

import GOApi

import PianaGlobals

from PianaDBaccess import *
from GraphPatchGroup import *

import utilities
from parserUtilities import *


verbose = 1


def get_key_from_property_pair(property_a, property_b):
    """
    given two strings property_a and property_b, returns the key that identifies them
    """
    if property_a <= property_b:
        return "%s@%s" %(property_a, property_b)
    else:
        return "%s@%s" %(property_b, property_a)

        
def get_property_pair_from_key(key):
    """
    returns a tuple of strings from the key
    """
    key_fields = key.split("@")
    
    return (key_fields[0], key_fields[1])


def get_protein_properties(proteinPiana, property_type, piana_access):
    """
    returns a list of properties for a given protein proteinPiana
    """
    global dic_protein_properties
    global dictionary_properties
    
    if dic_protein_properties.has_key(proteinPiana):
        return dic_protein_properties[proteinPiana]
    else:
        these_properties = source_piana_access.get_code_value(proteinCode_table= dictionary_properties[property_type][0],
                                                       proteinCode_col= dictionary_properties[property_type][1],
                                                       proteinPiana_value= proteinPiana,
                                                       answer_mode = "list",
                                                       source_db_info="no")
        
        dic_protein_properties[proteinPiana] = these_properties
        
        return these_properties



def get_proteins_with_property(property_value, property_type, piana_access):
    """
    returns the list of proteins that have a given property property_value of type property_type
    """
    global dic_proteins_with_property
    
    if dic_proteins_with_property.has_key(property_value):
        return dic_proteins_with_property[property_value]
    else:
        these_proteinPianas = piana_access.get_proteinPianas_with_value(value= property_value,
                                                                        proteinCode_table= dictionary_properties[property_type][0],
                                                                        proteinCode_col= dictionary_properties[property_type][1])
        
        dic_proteins_with_property[property_value] = these_proteinPianas
        return these_proteinPianas

        
    


    
# --------
# --------
#  Main()               
# --------                               
# --------

dictionary_properties = {"scop_fa":[PianaGlobals.proteinScop_table, PianaGlobals.fa_col],
                         "scop_sf":[PianaGlobals.proteinScop_table, PianaGlobals.sf_col],
                         "scop_cf":[PianaGlobals.proteinScop_table, PianaGlobals.cf_col],
                         "pfam":[PianaGlobals.pfam_table, PianaGlobals.pfamID_col],
                         "cog":[PianaGlobals.proteinCog_table, PianaGlobals.cogID_col]}


## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

compulsory_arguments_list = [ ("source-piana-dbname=", None, "name of PIANA database with interactions to be 'transferred'(required)"),
                              ("source-piana-dbhost=", None, "name of host where source piana database is located (required)"),
                              ("property-type=", None, "type of property that will be used for predicting interactions (scop_fa, scop_sf, scop_cf, cog, pfam, ...) Attention! this property type must be hardcoded in the code within dictionary dictionary_properties"),
                              ("method-tag=",None,"label that will be used to identify the method of prediction for these interactions"),
                              ("db-tag=",None,"label that will be used to identify the source database for these interactions")]

optional_arguments_list = [("source-piana-dbuser=",None,"username accessing the source database (not required in most systems)"),
                           ("source-piana-dbpass=",None,"password of username accessing the source database (not required in most systems"),
                           ("target-piana-dbname=",None,"name of PIANA database where predictions will be inserted (if not set, predictions are written to stdout)"),
                           ("target-piana-dbhost=",None,"name of host where target piana database is located"),
                           ("target-piana-dbuser=",None,"username accessing the target database (not required in most systems)"),
                           ("target-piana-dbpass=",None,"password of username accessing the target database (not required in most systems"),
                           ("output-id-type=",None," type of identifier for the output (if writing to text file)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("log_file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser")]

# Define the parser information to print in the help:

this_parser = ParserObject( script_name="predict_interactions_using_common_feature.py",
                       script_description="This script predicts protein-protein interactions based on proteins having a common feature (e.g. SCOP, COG, ...)\n Attention! The method name you are giving to the predictions must exist in PianaGlobals.method_names (e.g. interologscop)",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


arguments_dict = this_parser.parseArguments()

source_piana_dbname = arguments_dict["source-piana-dbname"]
source_piana_dbhost = arguments_dict["source-piana-dbhost"]
property_type = arguments_dict["property-type"]
method_tag = arguments_dict["method-tag"]
db_tag = arguments_dict["db-tag"]
source_piana_dbuser = arguments_dict["source-piana-dbuser"]
source_piana_dbpass = arguments_dict["source-piana-dbpass"]
target_piana_dbname = arguments_dict["target-piana-dbname"]
target_piana_dbhost = arguments_dict["target-piana-dbhost"]
target_piana_dbuser = arguments_dict["target-piana-dbuser"]
target_piana_dbpass = arguments_dict["target-piana-dbpass"]
output_id_type = arguments_dict["output-id-type"]
help = arguments_dict["help"]
verbose = arguments_dict["verbose"]
log_file = arguments_dict["log_file"]
time_control = arguments_dict["time-control"]


# Read parser parameters


# Start time counter
if time_control:
    initial_time = time.time()



# STEP 3: THE PARSER ITSELF. IT MUST USE THE METHODS OF PianaDBaccess

# Initialisating connections to piana
source_piana_access = PianaDBaccess(dbname=source_piana_dbname, dbhost=source_piana_dbhost, dbuser=source_piana_dbuser, dbpassword= source_piana_dbpass)

if target_piana_dbname:
    target_piana_access = PianaDBaccess(dbname=target_piana_dbname, dbhost=target_piana_dbhost, dbuser=target_piana_dbuser, dbpassword= target_piana_dbpass)

    # if we are going to insert interactions into a target DB, we need to make sure that the database tag given by the user exists in the target database

    try:
        # avoid the error given by piana_access when a databaseName under this name already exists. This has been done to prevent parsing
        # the same database twice into a PIANA db. But in this case, we are just tagging the predictions as being of an imaginary database
        # given by the user, and therefore, even if he asks to call the database with a name that was already used before to
        # insert other predictions, we must allow him to do so
        target_piana_access.insert_new_external_database(databaseName = db_tag,
                                                         databaseVersion = "",
                                                         parsedFile = input_file.split("/")[-1],
                                                         databaseDescription = "Database with predictions using common features")
    except:
        # if the database existed already, just ignore this error
        pass



dic_property_ints = {}  # dic with interactions described between properties
                        # keys are "property_a@property_b", with a <= b

dic_protein_properties = {} # keeps properties for those proteins for which we already retrieved their properties
                            # keys are the proteinPianas and contents a list of properties

dic_proteins_with_property = {} # keeps which proteins have each property
                                # keys are the properties and contents lists of proteins with that property 


# 1 for each interaction, infer property-property interactions
all_triplets = source_piana_access.get_all_protein_protein_interactions( use_self_ints="yes",
                                                                             list_source_dbs= "all", inverse_dbs="no",
                                                                             list_source_methods="all",
                                                                             inverse_methods="no",
                                                                             only_id="no")

for one_triplet in all_triplets:

    proteinPiana_a = one_triplet[0]
    proteinPiana_b = one_triplet[1]

    # get properties of proteinPiana_a
    properties_a = get_protein_properties(proteinPiana=proteinPiana_a, property_type=property_type, piana_access=source_piana_access)
        
    # get properties of proteinPiana_b
    properties_b = get_protein_properties(proteinPiana=proteinPiana_b, property_type=property_type, piana_access=source_piana_access)

    for one_property_a in properties_a:
        for one_property_b in properties_b:
            dic_property_ints[ get_key_from_property_pair(one_property_a, one_property_b) ] = None
        # END OF
    # END OF
# END OF

del all_triplets


# 2. for each property-property interaction, predict protein-protein interactions

already_printed_ints = {} # used to avoid repeating the same interaction multiple times

for one_property_interaction in dic_property_ints:
    
    (property_a, property_b) = get_property_pair_from_key(one_property_interaction)

    # get proteins with property_a
    proteins_property_a = get_proteins_with_property(property_value=property_a, property_type=property_type, piana_access=source_piana_access)
    
    # get proteins with property_b
    proteins_property_b = get_proteins_with_property(property_value=property_b, property_type=property_type, piana_access=source_piana_access)

    # predict interactions for all combinations between a and b
    for one_protein_property_a in proteins_property_a:
        for one_protein_property_b in proteins_property_b:

            if one_protein_property_a <= one_protein_property_b:
                this_key = "%s.%s" %(one_protein_property_a, one_protein_property_b)
            else:
                this_key = "%s.%s" %(one_protein_property_b, one_protein_property_a)

            if already_printed_ints.has_key(this_key):
                continue
            try:
                already_printed_ints[this_key] = None
            except:
                sys.stderr.write("Memory error with key %s and size of dic %s" %(this_key, len(already_printed_ints)))

            if verbose:
                sys.stderr.write("%s\t%s\t%s_predicted\n" %(one_protein_property_a, one_protein_property_b, property_type))


            if target_piana_dbname:
                target_piana_access.insert_interaction(proteinPianaA_value=one_protein_property_a, isSourceA_value=1,
                                                       proteinPianaB_value=one_protein_property_b, isSourceB_value=1,
                                                       interactionConfidence_value=1,
                                                       methodDescription_value=method_tag,
                                                       sourceDBDescription_value=db_tag,
                                                       confidenceAssignedSourceDB_value=1,
                                                       pubmed_id_value="unknown"  )
            else:
                # print predicted interaction to stdout (using output_id_type)
                protein_id_a = source_piana_access.get_list_protein_external_codes(proteinPiana=one_protein_property_a,
                                                                                   id_type= output_id_type,
                                                                                   alternative_id_types= ["md5"],
                                                                                   answer_mode= "single")
                protein_id_b = source_piana_access.get_list_protein_external_codes(proteinPiana=one_protein_property_b,
                                                                                   id_type= output_id_type,
                                                                                   alternative_id_types= ["md5"],
                                                                                   answer_mode= "single")
                
                sys.stdout.write("%s\t%s\t%s\t%s\n" %(protein_id_a, protein_id_b, db_tag, method_tag))
                                                                                   
            
        # END OF
    # END OF
# END OF
        
## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

source_piana_access.close()


if time_control:
    sys.stderr.write("Total time: %s seconds\n" %(time.time()-initial_time))
