"""
 File       : parserUtilities.py
 Author     : Javier Garcia
 Creation   : April 2007
 Contents   : miscelaneous utilities used by piana parsers
=======================================================================================================

"""

import sys,getopt,re


### METHODS USED IN PARSERS ###




class ParserObject(object):
    """
    Class that stores the information of the script (script utility, parameters,...)
    
    """

    def __init__ (self, script_name=None, script_description=None, optional_arguments = [], compulsory_arguments = [] ):
        """
        "arguments" are list of arguments with the following format: a tuple with (argument, default_value, argument_description)
        """

        self.script_name = script_name
        self.script_description = script_description
        self.optional_arguments = optional_arguments
        self.compulsory_arguments = compulsory_arguments

        self.log = {}


    def _indent(self,max_num_tabulators, initial_length):

        num_tabulators = max_num_tabulators - (initial_length+1)/8

        values_to_return = []

        #print "num tabulators: %s" %(num_tabulators)

        for i in xrange(num_tabulators):
            values_to_return.append("\t")

        return "".join(values_to_return)

    def _splitsize(self, string, size):
        """
        Split a string in substrings with a determined size
        """

        list_return = []

        final_position=0

        if len(string)<=size:
            list_return = [string]
        else:
            for i in xrange(len(string)/size):
                initial_position = i*size + final_position - i*size
                final_position = (i+1)*size
                while final_position<len(string) and string[final_position] != " " and  string[final_position] != "\t":
                    final_position += 1
                list_return.append(string[initial_position:final_position])

        return list_return
                
                
                

    def print_help(self):

        print "--------------------------------------------------------------------------------------------------------------"
        print "DESCRIPTION:"
        print "\t"+self.script_description

        usage = "\tpython %s " %(self.script_name)

        for argument in self.compulsory_arguments:
            if re.search("=",argument[0]):
                usage = usage + "--%s%s " %(argument[0],argument[0].rstrip("="))
            else:
                usage = usage + "--%s " %(argument[0])

        for argument in self.optional_arguments:
            if re.search("=",argument[0]):
                usage = usage + "[--%s=%s] " %(argument[0],argument[0].rstrip("="))
            else:
                usage = usage + "[--%s] " %(argument[0])

        print "\n"
        print "USAGE:"
        print usage

        print "\nWHERE:\n"

        if len(self.compulsory_arguments)>0:
            print "COMPULSORY ARGUMENTS:"

            for argument in self.compulsory_arguments:
                sys.stdout.write("\t%s:" %(argument[0].rstrip("=")))
                sys.stdout.write("%s" %(self._indent(3,len(argument[0])-1)))
                argument_description = self._splitsize(string=argument[2],size=80)
                if len(argument_description)==1:
                    sys.stdout.write("%s\n" %(argument[2]))
                else:
                    sys.stdout.write("%s\n" %(argument_description[0]))
                    for i in range(1,len(argument_description)):
                        sys.stdout.write("\t\t\t\t%s\n" %(argument_description[i]))

        if len(self.optional_arguments)>0:
            print
            print "OPTIONAL ARGUMENTS:"

            for argument in self.optional_arguments:
                sys.stdout.write("\t%s:" %(argument[0].rstrip("=")))
                sys.stdout.write("%s" %(self._indent(3,len(argument[0])-1)))
                argument_description = self._splitsize(string=argument[2]+" [default: %s]" %(argument[1]),size=80)
                if len(argument_description)==1:
                    sys.stdout.write("%s [default: %s]\n" %(argument[2],argument[1]))
                else:
                    sys.stdout.write("%s\n" %(argument_description[0]))
                    for i in range(1,len(argument_description)):
                        sys.stdout.write("\t\t\t\t%s\n" %(argument_description[i]))
                
        print "--------------------------------------------------------------------------------------------------------------"


    def parseArguments(self):
        """
        Method that returns a dictionary with the values of the arguments
        
        """
        
        arguments = self.compulsory_arguments+self.optional_arguments
        
        # Set all default values
        #return_values = [i[1] for i in arguments]
        return_dict = {}
        for i in arguments:
            return_dict[i[0].replace("=","")] = i[1]
        
        # Obtain a list with the names of all arguments
        list_arguments = [argument[0] for argument in arguments]
        # It can be of the following way because it contains "=" digit
        #list_arguments = return_dict.keys()
        
        # Parse arguments
        try:
            opts, args = getopt.getopt(sys.argv[1:], "", list_arguments)
            
        except getopt.GetoptError, bad_opt:
            # return error in parsing parameters, and return void list
            sys.stderr.write("%s\n" %(bad_opt.__str__()) )
            return {}

        # If there is no error, continue with the parsing
        for option,value in opts:
            if option=="--help":
                self.print_help()
                sys.exit(2)
            for actual_argument  in list_arguments:
                # Delete the "=" value if it has
                temp_arg = actual_argument.replace("=","")
                if option=="--"+temp_arg:
                    if value=="":
                        return_dict[temp_arg]=1
                    else:
                        return_dict[temp_arg]=value

        # Check for all compulsory arguments:
        for comp_arg in self.compulsory_arguments:
            if return_dict[comp_arg[0].replace("=","")] is None:
                sys.stderr.write("%s argument is not defined!\n" %(comp_arg[0].replace("=","")))
                self.print_help()
                sys.exit(2)

        return return_dict


    def add_to_log(self,key):
        """
        Increment the counter of a log dictionary for a given key
        
        Used in parsers
        
        """

        try:
            self.log[key] += 1
        except KeyError:
            self.log[key] = 1


    def get_log_string(self):
        """
        Returns a string with the content of the log dictionary

        Format: key: value
        """

        string_list = []

        for log_element in self.log.keys():
            string_list.append("%s: %s" %(log_element,self.log[log_element]))

        return "\n".join(string_list)

