"""
File        : PianaGraph.py
Author      : Ramon Aragues
Creation    : 2003
Contents    : implementation of graph handling protein interaction networks (subclass of Graph.py)
Called from : programs/classes that need to use piana graphs

=======================================================================================================

This file implements class PianaGraph
"""

# PianaGraph.py: implements a class for managing protein-protien interaction networks
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

import types
import copy
import time

from Graph import *
from PianaDBaccess import *
from PianaGraphEdgeAttribute import *
from PianaGraphNodeAttribute import *

from sets import *

import GOApi
import utilities


verbose = 0
verbose_output_graph = 0
verbose_patches = 0
verbose_spots = 0
verbose_expression= 0
verbose_pathway= 0


#----------------------------------------------------------------------------------------------
class PianaGraph(Graph):
#----------------------------------------------------------------------------------------------
    """
    Implementation of graph handling protein interaction networks (subclass of Graph.py)
    """
    def __init__(self, graphID = None, piana_access_object = None, piana_secondary_access_object = None, memory_mode="memory"):


        self.graph_id = graphID
	self.piana_access = piana_access_object
	self.piana_secondary_access = piana_secondary_access_object   # TO CHECK! I think I don't need the
                                                                      # secondary access on this class...
                                                                      # since I am only using it when
                                                                      # building the graph (ie. PianaGraphBuilder)
                                                                      # and when outputting the edge info (which
                                                                      # is handled by the edge attribute, which
                                                                      # always contains the piana_access that was
                                                                      # used to find it)


        self.ext_code_connecting = None  # dictionary used in connecting mode to keep linkers and the roots
                                         # that it connects (using ext_code instead of proteinPiana) 
                                         # 
                                         # structure is: { linker_protein:[root_protein, root_protein, ...],
                                         #                 linker_protein:[root_protein, root_protein, ...],
                                         #                 ......................
                                         #               }
                                         # Attention: it is not only about using ext code instead of proteinPiana: the
                                         #            linkers are different as well! Because a proteinPiana can be connecting 
                                         #            two root proteinPianas that in fact the same ext code, and therefore that 
                                         #            it not considered a linker when using ext_codes. In fact this dictionary 
                                         #            is the only one that should be used to decide if a protein is a linker or 
                                         #            not... 
	
	

                                         
	self.interaction_table_strings = {}  # used to keep the HTML lines of the interaction table that are associated to each ext code
                                             #  structure is: { 'header': "html code for interactions table header"
                                             #                  'footer': "html code for interactions table footer"
	                                     #                  protein_ext_code1: [interaction line 1, interaction line 2, ...]
                                             #                  protein_ext_code2: [interaction line 1, interaction line 2, ...]
                                             #                  ..................
                                             #                }
                                             # It is filled in method print_table_interaction, so unless it is called (via output_interactions) 
                                             # it will be empty 
                                             # TO DO!!!
                                             # Attention! there is one problem: if the user in match-proteins-to-pathways is asking for 
                                             #            an output in a type of protein identifier different from the one used for 
                                             #            outputting the interaction table, this will not work... In the current implementation, 
                                             #            both ID types must be the same or no info will be found for the proteins.



        # These variables used to unify names used for the proteins:
        
        self.last_key_list = -1               # initializing to a key that does not match to any list of proteins
        self.number_of_nodes_names_filled = 0 # used to know if nodes where added since dic_protein_naming was filled
        self.last_output_type= None
        self.dic_protein_naming = None  # dictionary used for controlling which name is assigned to each protein
                                        # has structure: { protein_name: unique protein name to be used,
                                        #                  protein_name: unique protein name to be used,
                                        #                  .......................
                                        #                }
                                        
        self.proteinPiana_naming = {}   # used for controlling which name is assigned to each proteinPiana
                                        # has structure: { proteinPiana: unique protein name to be used,
                                        #                  proteinPiana: unique protein name to be used,
                                        #                  .......................
                                        #                }


        self.unified_id_type= None      # used to check whether the unification was already done or not
        self.first_unified_node= None     # used to check whether the unification was already done or not
        self.last_unified_node= None      # used to check whether the unification was already done or not
        self.number_of_unified_nodes = 0  # used to check whether the unification was already done or not

        self.num_unified_neighbours = {}
        
        self.ext_code_unified_node_correspondence = None  # a dictionary used for creating unified nodes
                                                          # follows structure:
                                                          #                     { ext_code: unified_node,
                                                          #                       ext_code: unified_node,
                                                          #                       ...................... }
                                                        
        self.number_of_unified_edges = 0                                
        self.edge_key_unified_edge_correspondence = {}   # a dictionary used for creating unified edges
                                                        # follows structure:
                                                        #                     { edge_id: unified_edge,
                                                        #                       edge_id: unified_edge,
                                                        #                       ...................... }
                                                        # If "speed" mode is used, instead of having the unified edge as value, it contains a list with associated
                                                        # interactionPianas


        # These variables used to keep track of over/infra expressed proteins:
        

        self.over_expressed_proteinPianas = {}  # used for keeping a list of which proteinPianas are over expressed according to a expression file
        self.infra_expressed_proteinPianas = {}  # used for keeping a list of which proteinPianas are over expressed according to a expression file
	
	self.dic_special_proteinPianas = {} # keeps track of special proteinPianas
	                                    # follows structure:
                                            #      { label1: {proteinPiana:None, proteinPiana:None, ...},
                                            #        label2: {proteinPiana:None, proteinPiana:None, ...},
                                            #        ...............
                                            #      }
	                                    # It is filled by calling to _fill_special_proteins_dic
                                            # It is used to highlight or label proteins that belong to specific lists
                                            
	self.dic_special_ext_codes = {} # keeps track of special proteinPianas
	                                # follows structure:
                                        #      { ext_code1: [label1, label2, ...],
                                        #      { ext_code2: [label3, label2, ...],
                                        #        ...............
                                        #      }
                                        # It is filled by calling to _fill_special_proteins_dic
                                        # It is used to highlight or label proteins that belong to specific lists
	

        # initialize graph
        Graph.__init__(self, graphID= self.graph_id,memory_mode=memory_mode)

    # ---------
    # Methods needed for pickling and unpickling 
    # ---------

    def __getstate__(self):

        odict = self.__dict__.copy() # copy the dict since we change it
        #del odict['db']              # remove filehandle entry
        return odict

    def __setstate__(self, dict):

        self.__dict__ = dict
        dict.__class__.__init__(dict)
    
    def __getnewargs__(self):

        return (self.graph_id, self.piana_access, self.piana_secondary_access)

    # ------------------------------------------
    # Overwritte some methods
    # ------------------------------------------
    def build_graph(self, user_graph_builder, complete_taxID=None):
        
        user_graph_builder.set_graph(graph_object= self)
        user_graph_builder.build_graph(complete_taxID=complete_taxID)
    

    # ------------------------------------------
    #  PianaGraph Methods: get information about the graph and its componets
    #------------------------------------------

    def get_nodes_molecular_weights(self):
        """
        returns a dictionary (hashed by nodeID (ie proteinPiana)) of molecular weights:
                                                      { nodeID1:molecularWeight1, nodeID2:molecularWeight2 , ...}
        """

        molecular_weights = {}

        for node in self.get_node_object_list():

            node_attribute = node.get_node_attribute_object()

            molecular_weights[node.get_node_id()]= node_attribute.get_molecular_weight()
        # END OF for node in self.get_node_object_list():

        return molecular_weights

    def get_nodes_isoelectric_points(self):
        """

        returns a dictionary (hashed by nodeID (ie proteinPiana)) of isoelectric points:
                                                       {nodeID1:isoelectricPoint1, nodeID2:isoelectricPoint2 , ...}
        """

        isoelectric_points = {}

        for node in self.get_node_object_list():

            node_attribute = node.get_node_attribute_object()

            isoelectric_points[node.get_node_id()]= node_attribute.get_isoelectric_point()
        # END OF for node in self.get_node_object_list():

        return isoelectric_points


    def add_interaction(self, protein_a= None, protein_b=None, id_type= None, tax_id_value=0 , source_db=None, method=None, confidence=None):
        """
        Adds an interaction to the current PianaGraph object

        "protein_a" is the protein code of one side of the interaction

        "protein_b" is the protein code of the other side of the interaction

        "id_type" is the type of protein code being used for protein_a and protein_b
           --> valid types of protein code are listed in the PIANA reference card (read README.piana_tutorial)

        "tax_id_value" sets the species of the interactions that are being added (can be used for eliminating ambiguities between codes across species)
           --> valid species names are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        
        "source_db", "method" and "confidence" not being used: that data is obtained directly from the database instead
        """
        
        # Search for the db identifiers of protein_a and protein_b
        # We don't allow to add interactions between proteins that are not in the database
        list_proteinPiana_a = self.piana_access.get_list_proteinPiana(protein_id= protein_a,
                                                                      id_type= id_type,
                                                                      tax_id_value= tax_id_value,
                                                                      source_db_info="no")
        list_proteinPiana_b = self.piana_access.get_list_proteinPiana(protein_id= protein_b,
                                                                      id_type=id_type ,
                                                                      tax_id_value= tax_id_value,
                                                                      source_db_info="no")
        

        for proteinPiana_a in list_proteinPiana_a:
            for proteinPiana_b in list_proteinPiana_b:
                
                # 1.0 first of all, we create an edge attribute describing the propagation of the edge
                new_edge_attribute = PianaGraphEdgeAttribute(interactionPiana_value= None, mem_mode = "inMemory", piana_access = self.piana_access)

                # create_edge returns a tuple [edge_object, new] where new==1 means edge didn't exist in the graph before
                #             -> it doesn't add the edge to the graph
                new_edge = self.create_edge(node_id1= proteinPiana_a,
                                            node_id2= proteinPiana_b,
                                            attribute_object= new_edge_attribute)

                if new_edge[1] == 1:
                    if verbose:
                        sys.stderr.write(" adding edge between %s (%s) and %s (%s) \n" %(proteinPiana_a, protein_a, proteinPiana_b, protein_b))

                    # edge has to be added: create nodes, add them and then add the edge
                    new_node_a_attribute = PianaGraphNodeAttribute(proteinPiana_value=proteinPiana_a, mem_mode = "inMemory",
                                                                   piana_access = self.piana_access)
                    new_node_b_attribute = PianaGraphNodeAttribute(proteinPiana_value=proteinPiana_b, mem_mode = "inMemory",
                                                                   piana_access = self.piana_access)
                    
                    new_node_a = self.get_node(identifier=proteinPiana_a, attribute=new_node_a_attribute)
                    new_node_b = self.get_node(identifier=proteinPiana_b, attribute=new_node_b_attribute)
                    
                    # we do not know if the nodes are already in the graph: add them (add_node takes care of merging attributes if different)
                    self.add_node(new_node_a)
                    self.add_node(new_node_b)
                    self.add_edge(new_edge[0])

                # END OF if new_edge[1] == 1:
                else:
                    if verbose:
                        sys.stderr.write(" edge between %s (%s) and %s (%s) already existed in network\n" %(proteinPiana_a,
                                                                                                            protein_a,
                                                                                                            proteinPiana_b,
                                                                                                            protein_b))
            # END OF for proteinPiana_b in list_proteinPiana_b:
        # END OF for proteinPiana_a in list_proteinPiana_a:

        if not list_proteinPiana_a or not list_proteinPiana_b:
            sys.stderr.write("No proteinPiana found for protein (%s->%s) or protein (%s->%s)\n" %(protein_a,
                                                                                                  list_proteinPiana_a,
                                                                                                  protein_b,
                                                                                                  list_proteinPiana_b))
 
    # ----------------------------------------------------------------------------------------------
    #  PianaGraph Methods: Internal methods for unifying edges and nodes, deciding names to use, etc
    #-----------------------------------------------------------------------------------------------
   
    def _get_edge_key(self, this_edge, id_type, alternative_id_types ):
        """
        returns a tuple (unique key for an edge, node_1_ext_code, node_2_ext_code)

        The edge key is generated considering its external codes instead of the proteinPianas

        the ext_codes are taken using the dic_protein_naming
        """
        proteinPiana_start, proteinPiana_end = this_edge.get_node_ids()

        node_1_ext_code = self.get_unified_name( proteinPiana= proteinPiana_start,
                                                 id_type=id_type, alternative_id_types=alternative_id_types)
        node_2_ext_code = self.get_unified_name( proteinPiana= proteinPiana_end,
                                                 id_type=id_type, alternative_id_types=alternative_id_types)
        
        if node_1_ext_code <= node_2_ext_code:    edge_key = "%s.%s" %(node_1_ext_code, node_2_ext_code)
        else:                                     edge_key = "%s.%s" %(node_2_ext_code, node_1_ext_code)

        return (edge_key, node_1_ext_code, node_2_ext_code)

    def get_unified_name(self, proteinPiana, id_type, alternative_id_types):
        """
        returns the name (of code type "id_type") that is assigned to "proteinPiana" as the unified name

        , another_dic_prot_naming=None
	if another_dic_prot_naming is not None, uses this dictionary instead of self.dic_protein_naming


        """
        # --> Attention: PIANA is currently not using under any circustance the dictionary another_dic_prot_naming


        if self.proteinPiana_naming.has_key(proteinPiana):
            return self.proteinPiana_naming[proteinPiana]

        # if the dictionary with proteinPiana names to be used does not have the entry, find it (and update dictionary self.proteinPiana_naming)
        
        # get an ext_code  for this proteinPiana
        one_code = self.piana_access.get_list_protein_external_codes(proteinPiana= proteinPiana,
                                                                     id_type= id_type,
                                                                     alternative_id_types= alternative_id_types,
                                                                     answer_mode="single")

        # convert to uppercase in case it is not an int (we always work with uppercase names...)
        if not hasattr(one_code, "__int__"):     code_to_use = one_code.upper()
        else:                                    code_to_use = one_code

        # get the unified name
        if not code_to_use:     ext_code = "no_code_found"
        else:                
	    #if another_dic_prot_naming is None:
            try:
                ext_code= self.dic_protein_naming[code_to_use]
            except:
                raise ValueError("Strange1: trying to get a unified name for a non-existing protein (proteinPiana=%s) (%s does not appear in %s)" %(
                    proteinPiana,
                    code_to_use,
                    self.dic_protein_naming))
	    #else:
            #    try:
            #       ext_code= another_dic_prot_naming[code_to_use]
            #    except:
            #        raise ValueError("Strange2: trying to get a unified name for a non-existing protein (proteinPiana=%s) (%s does not appear in %s)" %(
            #            proteinPiana,
            #            code_to_use,
            #            another_dic_prot_naming))
		
        self.proteinPiana_naming[proteinPiana] = ext_code
        
        return ext_code
                
        
    def _create_unified_network(self, id_type, alternative_id_types, user_keywords=[], list_proteinPianas= None):
        """

        Attention!!! It is mandatory to call self._fill_dic_protein_naming before calling this method!!!!
                       (and self._fill_expression_dics  in case you are calling it from a method that requires expression information)

        This method is here to solve the following problems:

           - N nodes with a different node_id (ie proteinPiana) can have the same ext_code
             Each of these nodes can have different characteristics (be root or not, have a keyword or not, ...)
             If we print the nodes directly to output, the characteristics that will be used for the output for all nodes with the same ext_code
             will be those of the first node printed, since we avoid repeating the same ext_code twice
             However, the correct thing to do is to print just one ext_code, but one that has all the characteristics of the proteinPianas that correspond
             to it
             Therefore, we must unify the nodes by their ext_codes, and the new nodes must have all the characteristics of the corresponding proteinPianas

           - something similar happens with the edges
             Edges are defined by the proteinPianas that they join. However, proteinPianas can correspond to the same ext_codes, and therefore, if we
             only print one edge for two ext_codes, we are loosing the characteristics of the other edges

             eg:  edge 1 is between proteinPianas 11 and 12, and is given by DIP
                  edge 2 is between proteinPianas 21 and 22, and is given by STRING
                  edge 3 is between proteinPianas 31 and 32, and is given by HPRD

                  then, if proteinPiana 11 has the same ext_code as proteinPiana 31, and proteinPiana 12 has the same ext_code as proteinPiana 32
                        we should be printing the edge for DIP and HPRD. If this method didn't exist, we would only consider DIP, since the second
                        edge would be ignore because it is between the same ext_codes

        What this method does is create a list of unified nodes and another one of unified edges, and creates correspondences between the
        piana_graph nodes and the unified nodes and the same for the edges. Then, when printing, we will use the characteristics of the
        unified nodes and edges instead of using the piana_graph nodes and edges. There are a number of node characteristics that are not filled
        up when creating the network (eg. taxonomies) but are used for printing. These characteristics are stored in PianaGraphNodeAttribute but
        are only filled up in this method. If we have a network in memory, its nodes will not contain those characteristics, they will only be
        used when creating 'unified' nodes to storing info associated to several proteinPianas that correspond to the same ext_code

        "user_keywords" is a list of keywords that the user wants to highlight when printing information


        "list_proteinPianas" is usually None, but... if you are printing information for proteins that are not in the network (ie. using
        method output_list*) then, since there are no Nodes associated to the proteins that we want to print, we need to create the nodes
        from scratch and not based on a previous one. In those cases, set here the list of proteinPianas for which you want to do the unification.
        Setting this parameter also makes the method to skip the edge unification, since there are no edges associated to these proteins.
        
        """

        initial_time = time.time()

        number_of_nodes_in_current_graph = len(self.get_node_ids_list())
        number_of_edges_in_current_graph = len(self.get_edge_ids_list())

        
        if list_proteinPianas:
            # we are unifying nodes that do not appear in a network...(they are being provided via a list of proteins)
            #   -> create GraphNode objects for these proteinPianas with the corresponding attributes

            if len(list_proteinPianas) == self.number_of_unified_nodes and list_proteinPianas[0]== self.first_unified_node and (
                list_proteinPianas[-1]== self.last_unified_node and id_type== self.unified_id_type):
                # do not redo the unification if we already did it for this list of proteinPianas
                # the way of checking it is not very nice, but I think it works almost always and doesn't take too much time to do the check
                # In order to do a perfect check on whether the unification must be done again or not, we would have to create a unique
                # identifier for the list for which the unification is happening, and redo it only if the unique identifier changes
                # In the way I am doing the reunification it will fail (ie. it won't do the unification when in fact, it was needed) in the following
                # case: the number of elements in the new list is the same as in the old list, and the elements 0, and last of the list
                # are identical in both lists
                return 0
            
            
            self.first_unified_node= list_proteinPianas[0]
            self.last_unified_node= list_proteinPianas[-1]

            list_of_nodes = []
            for proteinPiana in list_proteinPianas:
                new_attribute = PianaGraphNodeAttribute(proteinPiana_value= proteinPiana,  piana_access= self.piana_access, mem_mode = "onDemand")
                new_node = GraphNode(nodeID=proteinPiana,
                                     attribute = new_attribute,
                                     isRoot =1, graph = None, expanded_from=None, expansion_type=None, ishidden=0, alternative_id= None)

                list_of_nodes.append(new_node)

        else:
            # unifying nodes of the network
            #number_of_nodes_in_current_graph = len(self.get_node_ids_list())
            #number_of_edges_in_current_graph = len(self.get_edge_ids_list())
            
            if self.ext_code_unified_node_correspondence is not None and number_of_nodes_in_current_graph == self.number_of_unified_nodes and (
                number_of_edges_in_current_graph == self.number_of_unified_edges and id_type== self.unified_id_type):
                # do not redo the unification if it was already done (unless there has been a change in the number of nodes/edges in the graph)
                
                # TO DO!!! This is not exactly correct: if there has been a change in the graph, but the number of
                #          nodes and edges is still the same (eg. an edge was added and another edge was removed), this 'if' will
                #          consider that dics self.*_unified_*_correspondence do not need recalculating, which is false...
                #          This is not very likely to happen, but it would be nice to think of a way that checks in a perfect manner
                #          whether the dics have to be recalculated or not: how do I create a unique key of the graph that was used
                #          to fill the dics? A string with all node ids and edge ids? Ummm.... too complicated and time costly...
                return 0
            
            
            list_of_nodes = self.get_node_object_list()
        # END OF else: (if list_proteinPianas:)

        # network has to be unified: initialize variables

        self.number_of_unified_nodes = 0
        self.number_of_unified_edges = 0
        self.unified_id_type = id_type
        self.ext_code_unified_node_correspondence = {}                              
        self.edge_key_unified_edge_correspondence = {}
        self.num_unified_neighbours = {}

        total_nodes_unified = 0
        total_edges_unified = 0
        
        # ----
        # node unification
        # ---
        node_initial_time = time.time()
        #print "Node unification...%s nodes" %(number_of_nodes_in_current_graph)
        node_unif_counter = 0
        for current_node in list_of_nodes:

            node_unif_counter += 1

            # for each node in the graph, create (or merge attributes with a previous node) a unique node for the ext_code that corresponds to it

            node_proteinPiana = current_node.get_node_id()
            
            # set the characteristics of the attribute that are not filled up when creating the network
            #   -> in case we afterwards find out that there was no node already created for this ext_code, these characteristics will be the
            #      characteristics of the initial unified node. Otherwise, they will be merge with the characteristics that existed previously
            #   -> these _set_* methods change the node in the PianaGraph, but it is not a problem because the modifications are made to
            #      variables that are only used by the unification process
            #javi:
            #current_node.get_node_attribute_object()._set_associated_proteinPianas(proteinPiana= node_proteinPiana)
            #current_node.get_node_attribute_object()._set_associated_nodes(node=current_node)
            #current_node.get_node_attribute_object()._set_dic_functions()
            #current_node.get_node_attribute_object()._set_dic_descriptions()
            #current_node.get_node_attribute_object()._set_dic_taxonomies()
            current_node.get_node_attribute_object()._set_keywords_appearing(user_keywords=user_keywords)
            current_node.get_node_attribute_object()._set_union_neighbours(neighbours_set=current_node.get_neighbour_ids_set())

            if self.over_expressed_proteinPianas.has_key(node_proteinPiana) and self.infra_expressed_proteinPianas.has_key(node_proteinPiana):
                node_expression = PianaGlobals.over_and_infra_expressed_protein
            elif self.over_expressed_proteinPianas.has_key(node_proteinPiana):         node_expression = PianaGlobals.over_expressed_protein
            elif self.infra_expressed_proteinPianas.has_key(node_proteinPiana):        node_expression = PianaGlobals.infra_expressed_protein
            else:                                                                      node_expression = None
            
            current_node.get_node_attribute_object()._set_expression(expression=node_expression)

	    node_special_labels = self._get_special_labels(proteinPiana= node_proteinPiana, format="list")
	    current_node.get_node_attribute_object()._set_special_labels(list_special_labels=node_special_labels)

            
            # get the ext_code that is going to be used for this proteinPiana
            node_ext_code = self.get_unified_name( proteinPiana= node_proteinPiana,
                                                   id_type=id_type, alternative_id_types=alternative_id_types)

            if self.ext_code_unified_node_correspondence.has_key(node_ext_code):
                # a node for this ext_code was already created: merge unique node characteristics with the current node
                self.ext_code_unified_node_correspondence[node_ext_code].merge_nodes(node_object= current_node, ignore_ids=1)
                
            else:
                #print "creating a unified edge"
                # no node was created for this ext_code: create a unique node with the characteristics of the current node
                #   - we set node id to 0, because we are not going to use the ids and in this way, we avoid conflicts when merging nodes
                #   - other characteristics will be added to this node when another proteinPiana has the same ext_code

                current_attribute= copy.deepcopy(current_node.get_node_attribute_object())   # to avoid modifying the attribute of a node in the network
                                                                                             # PianaGraphNodeAttribute makes sure of not deepcopying the
                                                                                             # mysql connection piana_access

                current_attribute._set_associated_node_ids(proteinPiana=current_node.get_node_id())
                
                # create new node and associate it to the ext_code
                new_node = GraphNode(nodeID= 0,
                                     attribute = current_attribute,
                                     isRoot = current_node.is_root(),
                                     graph = None,
                                     expanded_from= current_node.is_expanded()[0],
                                     expansion_type= current_node.is_expanded()[1],
                                     ishidden= current_node.is_hidden(),
                                     alternative_id= node_ext_code)
                
                self.ext_code_unified_node_correspondence[node_ext_code] = new_node

                #print self.get_unified_name(proteinPiana=current_node.get_node_id(),id_type="",alternative_id_types=[])
                total_nodes_unified +=1
            # END OF else: (if self.ext_code_unified_node_correspondence.has_key(node_ext_code):)
            
            self.number_of_unified_nodes += 1
            
        # END OF for node in self.get_node_object_list():

        if list_proteinPianas:
            # in case we have received a list of proteinPianas, skip the edge unification, since we are only interested in
            # unifying nodes because we are unifying nodes that do not appear in a network..
            return 0

        #print "Node unification: %s" %(time.time()-node_initial_time)


        # ----
        # edge unification
        # ---
        #print "Edge unification...%s edges!" %(number_of_edges_in_current_graph)
        edge_initial_time = time.time()
        #javi:
        #for current_edge in self.get_edge_object_list():
        #first, obtain the list of edge_ids, not the list of edge objects (it takes a lot of memory!):
        new_created=0
        unified = 0
        for current_edge_id in self.get_edge_ids_list():

            unified += 1

            current_edge = self.get_edge( identifier1 = current_edge_id, get_mode="error")

            (edge_key, node_1_ext_code, node_2_ext_code) = self._get_edge_key(this_edge= current_edge,
                                                                              id_type= id_type,
                                                                              alternative_id_types=alternative_id_types )
            
            if self.edge_key_unified_edge_correspondence.has_key(edge_key):
                # an edge was already created for this pair for ext_codes: merge current edge with the one already created
                #javi:
                #self.edge_key_unified_edge_correspondence[edge_key].merge_edges(new_edge= current_edge, ignore_ids= 1)
                if self.memory_mode=="memory":
                    self.edge_key_unified_edge_correspondence[edge_key].merge_edges(new_edge= current_edge, ignore_ids= 1)
                else:
                    self.edge_key_unified_edge_correspondence[edge_key].append(current_edge.get_edge_attribute_object().get_interactionPiana())
            else:
                # no edge has been created for this pair of ext_codes: create one and set it as the unified edge
                # in the unified edge, the node ids are never used: we set them to 0 (unified edges are used just to unify characteristics
                # of different edges that have the same ext_codes
                #javi

                total_edges_unified +=1

                if self.memory_mode=="memory":
                    current_attribute = current_edge.get_edge_attribute_object()
                
                    new_attribute = copy.deepcopy(current_attribute)  # to avoid modifying the attribute of an edge in the network
                                                                      # PianaGraphEdgeAttribute makes sure of not deepcopying the
                                                                      # mysql connection piana_access
                                                                      
                    new_edge = GraphEdge(node1_id= 0,
                                         node2_id= 0,
                                         attribute_object= new_attribute,
                                         graph= None,
                                         original= current_edge.is_original(),
                                         propagated= current_edge.is_propagated() ,
                                         extended= current_edge.is_extended() ,
                                         hidden = current_edge.is_hidden() )

                    self.edge_key_unified_edge_correspondence[edge_key] = new_edge

                else:                    
                    self.edge_key_unified_edge_correspondence[edge_key] = [current_edge_id,current_edge.get_edge_attribute_object().get_interactionPiana()]

                new_created+=1
            # END OF else:  (if self.edge_key_unified_edge_correspondence.has_key(edge_key):)
            
            self.number_of_unified_edges += 1
            
        # END OF for current_edge in self.get_edge_object_list():

        #print "Edge unification time: %s" %(time.time()-edge_initial_time)

        #print "TOTAL NODES UNIFIED: %s\n" %(total_nodes_unified)
        #print "TOTAL EDGES UNIFIED: %s\n" %(total_edges_unified)

    def _fill_dic_protein_naming(self, id_type, alternative_id_types, list_proteinPianas= None, key_list= None,
                                 user_protein_names = {}):
        """
        fills the naming dictionary, used to be coherent with the protein names used for output and avoid using different codes for the same protein

        it is called whenever protein names have to be used.

        "id_type" is the preferred type of code to be used and alternative_id_types are others to use in case the preferred is not found
          -> valid type names are found in the PIANA reference card (read README.piana_tutorial)

        "list_proteinPianas" is normally None, but can be set to a list of proteinPianas for those cases in which we are not printing proteins of the network
        but just proteins that the user asked information about. In case list_proteins is not None, then the names are unified for the proteins in the list
        instead of doing it for proteins of the network. This will not have an effect on successive calls asking information about proteins of the network,
        because the dictionary will be filled again when this method detects that the counter self.number_of_nodes_names_filled does not coincide with
        the number of nodes in the network (to make sure, we set self.number_of_nodes_names_filled to -1 when a protein list is passed as argument)
         (Attention! the list_proteins is a python list of proteinPianas. You can get a list of proteinPianas using PianaDBaccess.get_list_proteinPianas())

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}

        If the method was already called (and no more nodes have been added to the network) this call is ignored
        """

        # FIXING WHICH NAME WILL BE USED FOR EACH PROTEIN
        # -----------------------------------------------
        # In order to fix the naming problems with proteins, we create a dictionary that tries to set which will be the unique protein name to be
        # used for a protein that is represented by several nodes
        #
        # Attention: this is not enterely correct... the unique name that will be taken for a protein depends on the order in which the nodes are
        #            read, since many codes do not have "primary name", and the first returned from the database is the one that will be used.
        #            I do not know how to fix this arbitrary way of chosing the protein names... somebody would have to look to all protein names
        #            and decide which one to use... but still, not all biologists use the same name for the same protein...
        #            In any case, I guarantee that if one of those possible names was given by the user as a root protein, that will be the one used
        #
        # This looks very complicated (and it is) but choosing the name to be used for output is important...


        # ---------------
        # 1. First of all, decide whether we are going to unify names for all proteins in the network or just for those in a list given by the user
        #    --> detect as well if the unification was already established... if it was, exit the method
        # ---------------
        if not list_proteinPianas:
            # --> Unifying names for all proteins in the network
            
            number_of_nodes_in_current_graph = len(self.get_node_ids_list())

            if self.dic_protein_naming is not None and number_of_nodes_in_current_graph == self.number_of_nodes_names_filled and (
                self.last_output_type == id_type):
                # no nodes have been added and self.dic_protein_naming has already been populated: don't do anything
                # TO DO!!! This is not exactly correct: if there has been a change in the graph, but the number of
                #          nodes is still the same (eg. a node was added and another node was removed), this 'if' will
                #          consider that dic_protein_naming does not need recalculating, which is false...
                #          This is not very likely to happen, but it would be nice to think of a way that checks in a perfect manner
                #          whether self.dic_protein_naming has to be recalculated or not: how do I create a unique key of the graph that was used
                #          to fill the dic? A string with all node ids and edge ids? Ummm.... too complicated and time costly...
                return 0
            

            # initialize
            self.last_output_type = id_type
            self.number_of_nodes_names_filled = number_of_nodes_in_current_graph
            self.dic_protein_naming = {}    # has structure: { protein_name: unique protein name to be used,
                                            #                  protein_name: unique protein name to be used,
                                            #                  .......................
                                            #                }
            #list_proteinPianas = []
            #for node in self.get_node_object_list():
            #    list_proteinPianas.append(node.get_node_id())

            list_proteinPianas = [node.get_node_id() for node in self.get_node_object_list()] # changed to one line to optimize __mirar__

        # END OF if not list_proteinPianas:
        
        else:
            # --> Unifying names for proteins in list_proteinPianas
            
            if self.number_of_nodes_names_filled == -1 and key_list == self.last_key_list and self.last_output_type == id_type:
                # if the dic was previosly filled for this list, skip this method
                return 0
            
            # TO DO!!! Change the way list of proteins are handled
            #          This is completely inneficient, since the dictionary is going to be filled as many times as proteins there are in the list.
            #          Something has to be done to detect that the dictionary was already filled for this list of proteins. If it is a different list
            #          of proteins, it has to be filled again... Right now, I have solved it using the key_list argument, but this is not very clean
            #          and I am obliging the user to set unique keys for lists...
            self.last_output_type = id_type
            self.number_of_nodes_names_filled = -1
            self.last_key_list = key_list
            self.dic_protein_naming = {} 
        # END OF else:(if list_proteinPianas is None or not list_proteinPianas:)
            
        # ---------------
        # 2. Now, for each proteinPiana, fix which 'unified' external code will be used
        # ---------------
        for proteinPiana in list_proteinPianas:

            dic_unique_protein_name = {}   

            # 
            # 2.1 get all the protein names that can be given to this proteinPiana
            #
            temp_list_protein_names = self.piana_access.get_list_protein_external_codes(proteinPiana= proteinPiana,
                                                                                        id_type= id_type,
                                                                                        alternative_id_types= alternative_id_types,
                                                                                        answer_mode="list_primary",
                                                                                        source_db_info="no")
               
            list_protein_names = []
            # 2.1.1 convert all names to uppercase (to avoid considering two names as different when they just differ in case)
            for protein_name in temp_list_protein_names:
                # avoid errors when trying to do upper case over protein names that are ints
                if not hasattr(protein_name, "__int__"):   list_protein_names.append(protein_name.upper())
                else:                                      list_protein_names.append(protein_name)

            #
            # 2.2 Check if any of the names that can be given to this protein was already 'assigned to be another unique name'
            #     -> in case it was, keep it in dic_unique_protein_name so later on we know which is the 'unified name'
            #             -> we also initialize to 0 the frequency: later on it will be used to count the frequency for each unique name
            #
            for protein_name in list_protein_names:
                if self.dic_protein_naming.has_key(protein_name):
                    dic_unique_protein_name[ self.dic_protein_naming[protein_name] ] = 0 # we set it to Zero, it will be used later in cases with name conflict.
                                                                                         # the name with highest frequency is the one used... 
                                                                                         # if there are no conflicts, the number will not be used for anything  

            # END OF for protein_name in list_protein_names:


            #
            # 2.3 Check whether there is more than one unique name already fixed by another protein (2.3.1), there is only one (2.3.2) or None (2.3.3)
            # 
            number_of_unique_protein_names = len(dic_unique_protein_name)
            
            if number_of_unique_protein_names > 1:
                # 2.3.1
                # if there are more than one name, this means that two proteins with different lists of ext codes do share at least one ext code
                # My approach to this problem is a bit chapucero, but I cannot think of another way of doing it... I will take as unique code for the current 
                # protein and for all the names that previously had that unique name, the unique name that was associated to a larger number of ext codes
                # Of course, this doesn't apply if one of those unique names is in the list of names given originally by the user... user_protein_names
                # ... in this case, the name given by the user has preference over the other regardless of their frequencies of use
                # ... if more than one unique name is a user protein name, then frequency rules over the decision on which one to use
                unique_names_that_are_user_names = {}
                # check if one name in dic_unique_protein_name appear in the user_protein_names dictionary
                for protein_unique_name in dic_unique_protein_name:
                    if user_protein_names.has_key(protein_unique_name):
                        unique_names_that_are_user_names[protein_unique_name] = 0
                
                number_of_unique_names_that_are_user_names=  len(unique_names_that_are_user_names)
                
                if number_of_unique_names_that_are_user_names == 1:
                    # only one of the names is a user name, set that protein as the one to be used
                    protein_highest_freq = unique_names_that_are_user_names.popitem()[0] 
                else:
                    # none of the unique_names appears in the list of user names or more than one of the unique names is a user name (ie. freq rules)


                    # if there were no unique names that were user names, check frequency for all proteins in dic_unique_protein_name
                    # otherwise (ie. 2 or more unique names were user names) check frequency only for those
                    if number_of_unique_names_that_are_user_names > 1:
                        dic_unique_protein_name = copy.deepcopy(unique_names_that_are_user_names)  # change the dictionary to use for calculating freqs

                    # calculate frequencies
                    for ext_code in self.dic_protein_naming:
                        # self.dic_protein_naming[ext_code] is the unique code assigned to ext_code
			
                        if dic_unique_protein_name.has_key(self.dic_protein_naming[ext_code]):
                            dic_unique_protein_name[self.dic_protein_naming[ext_code]] += 1
                    # END OF for ext_code in self.dic_protein_naming:

                    # find the name with the highest frequency
                    max_freq= 0
                    protein_highest_freq= None
                    for name in dic_unique_protein_name:
                        if dic_unique_protein_name[name] > max_freq:
                            max_freq = dic_unique_protein_name[name]
                            protein_highest_freq = name
                    # END OF for name in dic_unique_protein_name:
                # END OF else: (if) DECIDING WHICH OF THE UNIQUE NAMES WILL BE USED

                # set all proteins associated to any of the unique codes in dic_unique_protein_name, to the highest frequency name found
                for name_from_complete_list in self.dic_protein_naming:
                    # self.dic_protein_naming[name_from_complete_list] is the unique name formerly assigned to name_from_complete_list

                    if dic_unique_protein_name.has_key( self.dic_protein_naming[name_from_complete_list] ):
                        self.dic_protein_naming[name_from_complete_list]= protein_highest_freq
                # END OF for name_from_complete_list in self.dic_protein_naming:

                
                # fix the unique protein name for all ext codes of this protein
                for protein_name in list_protein_names:
                    self.dic_protein_naming[protein_name] =  protein_highest_freq

            
            elif number_of_unique_protein_names == 1:
                # 2.3.2
                # there is just one name to be used: set all the ext codes to use this name

                # get the only pair (key, value) and fix the unique protein name for this ext code
                unique_name = dic_unique_protein_name.popitem()[0] 
                for protein_name in list_protein_names:
                    self.dic_protein_naming[protein_name] =  unique_name


            elif number_of_unique_protein_names == 0:
                # 2.3.3
                # in case no code was set already as unique code for any of the names of this protein, get a new single code and set it for the other names

                user_name_coincidences = {}
                # if any of the names given to this protein is a user name, use it. Otherwise, get one from the database using answer_mode 'single'
                for protein_name in list_protein_names:
                    if user_protein_names.has_key(protein_name):
                        user_name_coincidences[protein_name] = None

                number_of_coincidences = len(user_name_coincidences)

                if number_of_coincidences == 0 or number_of_coincidences > 1:
                    # if there are no coincidences (none of the protein names is a user name) or there are more than 1 (more than one of the names of the
                    # proteins is a user name), use answer_mode 'single' to retrieve which would be the preferred name to use according to the database
                    single_name = self.piana_access.get_list_protein_external_codes(proteinPiana= proteinPiana,
                                                                                    id_type= id_type,
                                                                                    alternative_id_types= alternative_id_types,
                                                                                    answer_mode="single")
                    if number_of_coincidences > 1:
                        # if there are 2 or more coincidences, choose which one to use:

                        if not hasattr(single_name, "__int__"):  name_to_use = single_name.upper()
                        else:                                    name_to_use = single_name
                        
                        if user_name_coincidences.has_key(name_to_use):
                            #   - if one of those coincidences corresponds to the db name, take that one
                            unique_name = name_to_use
                        else:
                            #   - if none of the coincidences corresponds to the db name, take a random one
                            #     (the user name has preference over the db preference over names)
                            unique_name = user_name_coincidences.popitem()[0]
                    else:
                        # no coincidences--> take the database single name
                        if not hasattr(single_name, "__int__"):    unique_name = single_name.upper()
                        else:                                    unique_name = single_name

                # END OF if coincidences == 0 or coincidences >1:
                
                else:
                    # only one of the protein names is also a user name: use it
                    unique_name = user_name_coincidences.popitem()[0]
                # END OF else: (if coincidences == 0 or coincidences >1:)

                    
                # fix the unique protein name for all ext codes of this protein
                for protein_name in list_protein_names:
                    self.dic_protein_naming[protein_name] =  unique_name

        # END OF for node in self.get_node_object_list(): (FIXING WHICH NAME WILL BE USED FOR EACH PROTEIN)

    # TO DO!!! Instead of fixing which name is going to be used for each external id, it would be faster to fix which
    #          unified name will be used for each proteinPiana. Doing it like this, we would not need to access the database
    #          each time we are looking for a unified identifier, since we would get it directly from a internal dictionary { proteinPiana:name_to_use, ....}



    def _fill_expression_dics(self, file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, tax_id_value=0):
        """
        fills self dictionaries that keep the proteinPianas that are over of infra expressed
        """

        if self.over_expressed_proteinPianas or self.infra_expressed_proteinPianas or (
                                                           file_over_expressed is None and file_infra_expressed is None):
            # do not do anything if the dictionaries are already filled or if no files were passed as argument
            return 0
        
          
        if expression_id_type is None:
            raise ValueError("You must give a protein code type for your expression files")  

        self.over_expressed_proteinPianas = {}  
        self.infra_expressed_proteinPianas = {} 
        self.over_expressed_ext_codes = {}  
        self.infra_expressed_ext_codes = {} 

        if file_over_expressed is not None:
            over_fd = file(file_over_expressed, "r")
            list_proteins = utilities.return_proteins_from_file(file_object= over_fd, id_type=expression_id_type)

            for protein_name in list_proteins:

                for proteinPiana in self.piana_access.get_list_proteinPiana(protein_id= protein_name,
                                                                            id_type= expression_id_type,
                                                                            tax_id_value= tax_id_value,
                                                                            source_db_info="no"):
                    self.over_expressed_proteinPianas[proteinPiana] = None
                    
            over_fd.close()
        # END OF if file_over_expressed is not None:

        if file_infra_expressed is not None:
            infra_fd = file(file_infra_expressed, "r")
            list_proteins = utilities.return_proteins_from_file(file_object= infra_fd, id_type=expression_id_type)

            for protein_name in list_proteins:

                for proteinPiana in self.piana_access.get_list_proteinPiana(protein_id= protein_name,
                                                                            id_type = expression_id_type,
                                                                            tax_id_value= tax_id_value,
                                                                             source_db_info="no"):
                    self.infra_expressed_proteinPianas[proteinPiana] = None
                    
            infra_fd.close()
        # END OF if file_infra_expressed is not None:

    def _fill_special_protein_dics(self, dic_special_proteins_file={}, tax_id_value = 0,
                                   id_type=None, alternative_id_types=None):
	"""
        Attention!!! It is mandatory to call self._fill_dic_protein_naming before calling this method!!!!
        
	fills the dictionary of special proteinPianas (ie. self.dic_special_proteinPianas) related to different labels
                    { label1:{proteinPiana:None, proteinPiana:None, ...}, label2: {proteinPiana1:None}....}
                    
        if id_type is not None, fills as well the dictionary of special ext codes (ie. self.dic_special_ext_codes)
                    { ext_code: [label1, label2, ...], ext_code2: [label1], ... }
           -> this dictionary uses external identifiers as keys, but ids are 'unified' into a single type of identifier
        
	These dictionaries are then used to highligth those special proteins and show their labels on the HTML tables

        "tax_id_value" is the tax id for the proteins in dic_special_proteins_file
        
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        "id_type" is the type of identifier that will be used for proteins in dictionary self.dic_special_ext_codes
            (or "alternative_id_types" used in case nothing found for id_type)
            
	"""
	if not dic_special_proteins_file or (self.dic_special_proteinPianas and not id_type) or self.dic_special_ext_codes:
	    # do not populate the dictionary if it has already been populated or the dictionary of files is empty
            return 0
        
	    
	self.dic_special_proteinPianas = {}

	for one_label in dic_special_proteins_file:
	    
	    id_type = dic_special_proteins_file[one_label][0]
	    special_file = dic_special_proteins_file[one_label][1]
	    
	    self.dic_special_proteinPianas[one_label] = {}

            # get special proteins from this file (along with the properties assigned to the protein in that file (if any))
	    dic_proteins = utilities.return_proteins_from_file(file_object= file(special_file, "r"), id_type=id_type, format="dic")
	    
	    for protein_name in dic_proteins:
		
		for proteinPiana in self.piana_access.get_list_proteinPiana(protein_id= protein_name,
                                                                            id_type= id_type,
                                                                            tax_id_value= tax_id_value,
                                                                            source_db_info="no"):
                    
		    self.dic_special_proteinPianas[one_label][proteinPiana] = dic_proteins[protein_name]

                    if id_type is None:
                        # if no protein id type was given, we do not need to fill the dictionary self.dic_special_ext_codes
                        continue

                    try:
                        # get unified name
                        unified_name = self.get_unified_name( proteinPiana= proteinPiana,
                                                              id_type=id_type,
                                                              alternative_id_types=alternative_id_types)

                        if dic_proteins[protein_name]:
                            label_to_print = "%s@%s" %(one_label, dic_proteins[protein_name])
                        else:
                            label_to_print = "%s" %(one_label)
                            
                        # create (or append) list of labels associated to unified name
                        if self.dic_special_ext_codes.has_key(unified_name):
                            if label_to_print not in self.dic_special_ext_codes[unified_name]:
                                self.dic_special_ext_codes[unified_name].append(label_to_print)
                        else:
                            self.dic_special_ext_codes[unified_name] = [label_to_print]
                    except:
                        # if we cannot get a unified name, it means that this special protein does not appear in the network
                        #  -> therefore, it can be ignored because its labels will never be printed out
                        continue
                    
		# END OF for proteinPiana in self.piana_access.get_list_proteinPiana(.....)
	    # END OF for protein_name in dic_proteins:
	# END OF for one_label in dic_special_proteins_file:
    
    def _special_labels_list_2_string(self, special_labels_list=[], tag="", separator=" "):
	"""
	transforms a list with special labels into a single string of labels separated by "separator"

        "tag" is written to the string before each label value (eg. label=)
	"""
	special_labels_string= ""
	if special_labels_list:
	    special_labels_string = tag + special_labels_list[0]
	    for one_label in special_labels_list[1:]:
		special_labels_string += separator + tag + one_label 

	return special_labels_string

    def _get_special_labels(self, proteinPiana=None, format="list"):
	"""
	returns the labels associated to "proteinPiana" in self.dic_special_proteinPianas

	if "format" is 'list' then it returns a list of strings
	if "format" is 'string' then it returns the labels in a single string, separated by '-'

        Attention! self.dic_special_proteinPianas must have been previously filled with command self._fill_special_proteins_dic

	"""
	list_labels = []
	for one_label in self.dic_special_proteinPianas:
	    if self.dic_special_proteinPianas[one_label].has_key(proteinPiana):
                if self.dic_special_proteinPianas[one_label][proteinPiana]:
                    this_label_to_print = "%s@%s" %(one_label, self.dic_special_proteinPianas[one_label][proteinPiana])
                else:
                    this_label_to_print = "%s" %(one_label)
                    
		list_labels.append(this_label_to_print)

	return list_labels

    def _count_expression_dic(self, expression_dic={}):
	"""
	returns a tuple (total number of proteins in expression_dic, number of overexpressed proteins, number of underexpressed proteins)

	
	in "expression_dic" keys are protein ext codes and values a tuple (is_over_expressed, is_under_expressed)


	"""
	number_of_prots= len(expression_dic)

	number_of_over_expressed_prots = 0
	number_of_under_expressed_prots =0
	number_of_expressed_prots=0

	for one_prot in expression_dic:
	    if expression_dic[one_prot][0]:
		number_of_over_expressed_prots += 1

	    if expression_dic[one_prot][1]:
		number_of_under_expressed_prots += 1

	    if expression_dic[one_prot][0] or expression_dic[one_prot][1]:
		number_of_expressed_prots += 1

	return (number_of_prots, number_of_expressed_prots, number_of_over_expressed_prots, number_of_under_expressed_prots)

    # ------------------------------------------
    #  PianaGraph Methods: Output methods
    #------------------------------------------

    def output_nodes_molecular_weights(self):

        string_molecular_weights = "Molecular weights for proteins in the network"

        node_molecular_weights = self.get_nodes_molecular_weights()

        for node_molecular_weight_key in node_molecular_weights.keys():
            
            string_molecular_weights += "Molecular weight for node %s is %s\n" %(node_molecular_weight_key,
                                                                                 node_molecular_weights[node_molecular_weight_key])
        # END OF for node_molecular_weight_key in node_molecular_weights.keys():
        
        return string_molecular_weights

    def output_nodes_isoelectric_points(self):

        string_isoelectric_points = "Isoelectric points for proteins in the network"

        node_isoelectric_points = self.get_nodes_isoelectric_points()

        for node_isoelectric_point_key in node_isoelectric_points.keys():
            
            string_isoelectric_points += "Isoelectric point for node %s is %s\n" %(node_isoelectric_point_key,
                                                                                   node_isoelectric_points[node_isoelectric_point_key])
        # END OF for node_isoelectric_point_key in node_isoelectric_points.keys():
        
        return string_isoelectric_points


    def _passes_species_test(self, required_protein_tax_id, dic_protein_taxonomies ):
        """
        internal function used to check if a given species "required_protein_tax_id" is in a list of taxonomies "list_protein_taxonomies"

        "required_protein_tax_id" is the taxonomy id we require the protein to be (eg. human, yeast, ...)
            --> valid tax id values are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy  

        "list_protein_taxonomies" is the list of current protein taxonomies
        
        returns 0 if species test is not passed. 1 otherwise
        """
        if required_protein_tax_id==0 or dic_protein_taxonomies.has_key(required_protein_tax_id):
            return 1
        else:
            return 0


    def _get_unified_neighbours(self, neighbours_proteinPianas, id_type= None, alternative_id_types=[]):
        """
        given a list of proteinPianas "neighbours_proteinPianas", returns a list with their unified names
        """
        return Set([ (self.get_unified_name( proteinPiana= neighbour_proteinPiana,
                                             id_type=id_type, alternative_id_types=alternative_id_types), None) for neighbour_proteinPiana in neighbours_proteinPianas])

        
    def _get_num_unified_networks(self, protein_extcode, protein_node, id_type, alternative_id_types):
        """
        returns the number of neighbours (unified under unique names) for a given protein external identifier
        """
        if not self.num_unified_neighbours.has_key(protein_extcode):

            neighbours_proteinPianas = protein_node.get_node_attribute_object().get_union_neighbours()

            neighbours_ext_codes= self._get_unified_neighbours(neighbours_proteinPianas= neighbours_proteinPianas,
                                                               id_type= id_type, alternative_id_types=alternative_id_types)

            self.num_unified_neighbours[protein_extcode] = len(neighbours_ext_codes)
            
        return self.num_unified_neighbours[protein_extcode]


        
    def _create_ext_code_connecting(self, id_type= None, alternative_id_types=[], connecting_proteinPianas= {}):
        """
        using the unified nodes and names, returns a dictionary {protein:[root_protein, root_protein, ...], protein: [], ...]}
        using id_type as type of protein code


        connecting_proteinPianas is the dictionary returned by Graph.get_connecting_nodes_dic()
        Now, this parameter it is not mandatory, as it is queried in this method if it is not suplied


        it is mandatory that _create_unified_network has been called before calling this method
        """

        number_of_nodes_in_current_graph = len(self.get_node_ids_list())
        number_of_edges_in_current_graph = len(self.get_edge_ids_list())
        
        if self.ext_code_connecting is not None and number_of_nodes_in_current_graph == self.number_of_unified_nodes and (
            number_of_edges_in_current_graph == self.number_of_unified_edges and id_type== self.unified_id_type):
            # do not create the dictionary if it was already done (unless there has been a change in the number of nodes/edges in the graph)
            
            # TO DO!!! This is not exactly correct: if there has been a change in the graph, but the number of
            #          nodes and edges is still the same (eg. an edge was added and another edge was removed), this 'if' will
            #          consider that dics self.*_unified_*_correspondence do not need recalculating, which is false...
            #          This is not very likely to happen, but it would be nice to think of a way that checks in a perfect manner
            #          whether the dics have to be recalculated or not: how do I create a unique key of the graph that was used
            #          to fill the dics? A string with all node ids and edge ids? Ummm.... too complicated and costly...
            return 0
        

        self.ext_code_connecting = {}

        
        for actual_node in self.get_node_object_list():

            new_set = Set([])

            node_ext_code = self.get_unified_name( proteinPiana= actual_node.get_node_id(),
                                                   id_type=id_type, alternative_id_types=alternative_id_types)

            root_connections = actual_node.get_all_root_connections()   #get all connections to root from the original graph

            for actual_root_connected in root_connections:
                root_ext_code = self.get_unified_name( proteinPiana= actual_root_connected,
                                                       id_type=id_type, alternative_id_types=alternative_id_types)

                # If this root ext code was not found previously in this node, add it
                if not root_ext_code in new_set:
                    new_set.add(root_ext_code)

            #If this unified node existed previously, merge both root connections (doing the union within the sets)
            if( self.ext_code_connecting.has_key(node_ext_code) ):
                self.ext_code_connecting[node_ext_code] = self.ext_code_connecting[node_ext_code].union(new_set)
            else:
                self.ext_code_connecting[node_ext_code] = new_set

        # Eliminate those ext codes that do not connect with more than a root
        # This is used to control the following case:
        # Now, remove those instances where the connecting node is only connected to one proteinPiana
        #   (this occurs when a node is connected to two nodes with different proteinPianas (and therefore, identified as connecting node)
        #    but those proteinPianas all correspond to the same ext_code... therefore, it is not a connecting node anymore)
        ext_code_connecting_list = self.ext_code_connecting.keys()
        
        for actual_ext_code_connecting in ext_code_connecting_list:
            if len(self.ext_code_connecting[actual_ext_code_connecting])<=1:
                del self.ext_code_connecting[actual_ext_code_connecting]


        # Uncommented by Javi, old method:
##         number_of_nodes_in_current_graph = len(self.get_node_ids_list())
##         number_of_edges_in_current_graph = len(self.get_edge_ids_list())
        
##         if self.ext_code_connecting is not None and number_of_nodes_in_current_graph == self.number_of_unified_nodes and (
##             number_of_edges_in_current_graph == self.number_of_unified_edges and id_type== self.unified_id_type):
##             # do not create the dictionary if it was already done (unless there has been a change in the number of nodes/edges in the graph)
            
##             # TO DO!!! This is not exactly correct: if there has been a change in the graph, but the number of
##             #          nodes and edges is still the same (eg. an edge was added and another edge was removed), this 'if' will
##             #          consider that dics self.*_unified_*_correspondence do not need recalculating, which is false...
##             #          This is not very likely to happen, but it would be nice to think of a way that checks in a perfect manner
##             #          whether the dics have to be recalculated or not: how do I create a unique key of the graph that was used
##             #          to fill the dics? A string with all node ids and edge ids? Ummm.... too complicated and costly...
##             return 0
        
##         #javi: # Before, the connecting_proteinPianas was introduced as a parameters (it is mantained still now), but most of the times it was not necessary
##                # as the unifyed connecting proteins was already generated before. So, now this dictionary is demanded here, to avoid generating it multiple
##                # innecessary times
##         if not connecting_proteinPianas:
##             connecting_proteinPianas = self.get_connecting_nodes_dic()
            
##         new_dictionary = {}
##         self.ext_code_connecting = {}
##         for connect_proteinPiana in connecting_proteinPianas:

##             # get the ext_code that is going to be used for this proteinPiana
##             node_ext_code = self.get_unified_name( proteinPiana= connect_proteinPiana,
##                                                    id_type=id_type, alternative_id_types=alternative_id_types)


##             if not new_dictionary.has_key(node_ext_code):
##                 new_dictionary[node_ext_code] = []

##             for root_proteinPiana in connecting_proteinPianas[connect_proteinPiana]:

##                 root_ext_code = self.get_unified_name( proteinPiana= root_proteinPiana,
##                                                        id_type=id_type, alternative_id_types=alternative_id_types)

##                 if root_ext_code not in new_dictionary[node_ext_code]:
##                     new_dictionary[node_ext_code].append(root_ext_code)
                    
##             # END OF for root_proteinPiana in connecting_proteinPianas[connect_proteinPiana]:
            
##         # END OF for connect_proteinPiana in connecting_proteinPianas:

##         # Now, remove those instances where the connecting node is only connected to one proteinPiana
##         #   (this occurs when a node is connected to two nodes with different proteinPianas (and therefore, identified as connecting node)
##         #    but those proteinPianas all correspond to the same ext_code... therefore, it is not a connecting node anymore)
##         for connecting_ext_code in new_dictionary:
##             if len(new_dictionary[connecting_ext_code]) > 1:
##                 self.ext_code_connecting[connecting_ext_code] = new_dictionary[connecting_ext_code]

                            
    def _passes_connecting_mode_test(self, id_a, is_root_a, id_b, is_root_b, connecting_prots_dic ):
        """
        internal function used to check is a protein passes the print mode='connecting' test

        this means interactions will be printed only in these cases:
           - interactions between root nodes
           - interaction between a root node and a node that connects more than 1 root nodes


        the only restriction is that id_a, id_b and connecting_prots_dic keys must use the same type of protein code

        returns 0 if connecting mode test is passed. 1 otherwise
        """
        if is_root_a and is_root_b:
            # both proteins are root: print the interaction
            result_test = 1
        
        elif not (is_root_a or is_root_b):
            # none of the proteins is root: do not print
            result_test = 0
        
        #elif self.ext_code_connecting.has_key(id_a) and is_root_b:
        elif id_a in self.ext_code_connecting and is_root_b:
            result_test = 1
            
        #elif self.ext_code_connecting.has_key(id_b) and is_root_a:
        elif id_b in self.ext_code_connecting and is_root_a:
            result_test = 1
            
        #elif self.ext_code_connecting.has_key(id_a) and self.ext_code_connecting.has_key(id_b):
        elif id_a in self.ext_code_connecting and id_b in self.ext_code_connecting:
            result_test = 1

        else:
            # this is an interaction between a root protein and some other protein (that is only connected to this root protein)
            # do not print it
            result_test = 0

         
        return result_test
        
    def _passes_root_mode_test(self, is_root_a, is_root_b, print_mode=None):
        """
        
        
        internal function used to check is a protein passes the print mode test

        if print_mode is all: all proteins pass it: return 1
        if print_mode is "all_root": one of the proteins must be root
        if print_mode is "only_root": both proteins must be root
        if print_mode is "connecting": error!!! This method should never be called in connecting mode: use instead _passes_connecting_mode_test
        
        """

        if print_mode == "all":
            return 1
        
        elif print_mode == "all_root":
            if is_root_a == 1 or is_root_b == 1:
                return 1
            else:
                return 0
            
        elif print_mode == "only_root":
            if is_root_a ==1 and is_root_b == 1:
                return 1
            else:
                return 0
            
        elif print_mode == "connecting":
            if (is_root_a or is_linker_a ) and (is_root_b == 1 or is_linker_b):
                return 1
            else:
                return 0

        else:
            raise ValueError("Incorrect print mode (%s) given. Valid values are: 'all', 'all_root', 'only_root' and 'connecting'" %(print_mode))


    def get_network_proteins(self, output_id_type, alternative_id_types= None, tax_id_value= 0, user_protein_names = {}):
	"""
	
	returns a list with the proteins in the network, using output_id_type to identify the proteins.
	

        "output_id_type" is the easy-to-remember type name that will be used for printing the protein
          -> valid id_type are those listed in the PIANA reference card (read README.piana_tutorial)

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide pairs a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output


        "tax_id_value" sets the species of the proteins being printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}
	"""

        # make sure self.dic_protein_naming is filled
	#   -> list_proteinPianas and key_list not being used... we are interested in the proteins that appear in the network
        self._fill_dic_protein_naming(id_type=output_id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= [], key_list=None, user_protein_names= user_protein_names)

        # make sure the unified nodes and edges are created
	#   -> list_proteinPianas and user_keywords not being used... we are interested in the proteins that appear in the network
        self._create_unified_network(id_type=output_id_type, alternative_id_types=alternative_id_types, user_keywords=[], list_proteinPianas= [] )

	
	#dic_protein_names = {}
        set_protein_names = Set()

	for proteinPiana_value in self.get_node_ids_list():
	    # find the unified name for each network protein
	    protein_name = self.get_unified_name( proteinPiana= proteinPiana_value,
						  id_type=output_id_type, alternative_id_types=alternative_id_types)

	    #dic_protein_names[protein_name] = None
            set_protein_names.add(protein_name)

	#return dic_protein_names.keys()
        return list(set_protein_names)




    def output_protein_information(self, proteinPiana_value, id_type, alternative_id_types= None, output_target= sys.stdout,
                                   output_mode= "compact", format_mode="txt", filter_mode="all", list_keywords= [],
                                   tax_id_value= 0, user_protein_names = {}, list_connects_nodes= None, 
				   list_proteinPianas= None, key_list= -1, 
                                   file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, dic_special_proteins_file=None):
        """
        writes information about protein "proteinPiana_value" to "output_target", following format "output_mode"


        "output_target" is a file object (sys.stdout to print to screen)


        "id_type" is the easy-to-remember type name that will be used for printing the protein
          -> valid id_type are those listed in the PIANA reference card (read README.piana_tutorial)
      
        "filter_mode" not working at the minute: it will be used for hidden, unhidden, all

        "output_mode" can be:
           - 'compact': all relevant information in one line
           - 'extended': all information in text paragraphs
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "list_connects_nodes" is a list of root nodes that are connected through protein 'proteinPiana_value'. It comes from
        one of the dictionary rows of self.get_connecting_nodes_dic(). If it is set to None, the method considers that the user does not
        want to know information about connected root nodes

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide pairs a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "tax_id_value" sets the species of the proteins being printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}
           
        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> when printing protein information, it prints in the protein information line user_keyword=word_found user_keyword=word_found ...
           -> currently, it is not used when printing the interactions in a table
           -> If you are interested in highlighting proteins related to cancer, list_keywords could be: ['cancer', 'onco', 'carcinoma', 'tumor']

        "list_proteinPianas" is normally None, but can be set to a list of proteinPianas for those cases in which we are not printing proteins of the network
        but just proteins that the user asked information about. In case list_proteins is not None, then the names are unified for the proteins in the list
        instead of doing it for proteins of the network. Look at PianaApi.print_list_proteins to better understand how this works

        "key_list" is a unique key for this list of proteins, used to know when the unification of names has to be performed again
           -> if you are making calls for different lists of proteins, just set the key to consecutive numbers so this method can
              differentiate between the lists
           -> a user key_list cannot be -1, because -1 is used as default to indicate that no key was used

        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed


        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        """
        
        # make sure self.dic_protein_naming is filled
        self._fill_dic_protein_naming(id_type=id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= list_proteinPianas, key_list=key_list, user_protein_names= user_protein_names)

        
        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value = tax_id_value)
        

	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                       id_type=None, alternative_id_types=None)

        # make sure the unified nodes and edges are created
        self._create_unified_network(list_proteinPianas= list_proteinPianas,
                                     id_type=id_type, alternative_id_types=alternative_id_types, user_keywords=list_keywords )
        
        protein_user_type = self.get_unified_name( proteinPiana= proteinPiana_value,
                                                   id_type=id_type, alternative_id_types=alternative_id_types)

        #if self.already_printed_proteins.has_key(protein_user_type):
        if protein_user_type in self.already_printed_proteins:
            # do not print information for a protein that has already been printed (the unified node contained all the information associated to this ext_code
            return 0
        
        #self.already_printed_proteins[protein_user_type] = None
        self.already_printed_proteins.add(protein_user_type)

        # Important!
        # get the unified node associated to this ext_code (the unique node for this ext_code, created by merging info from all proteinPianas with that ext_code 
        node = self.ext_code_unified_node_correspondence[protein_user_type]


        set_associated_proteinPianas = node.get_node_attribute_object().get_associated_proteinPianas()
        

        # getting all external codes of all types for proteinPianas associated to this unified node
        dic_ext_codes = {}
        for proteinPiana_value in set_associated_proteinPianas:
            list_protein_ext_code = self.piana_access.get_list_protein_external_codes(proteinPiana= proteinPiana_value, id_type= "all")
            for temp_ext_code in list_protein_ext_code:
                dic_ext_codes[temp_ext_code] = None
        list_protein_ext_code = dic_ext_codes.keys()

        # get taxonomies for this unified node
        dic_protein_taxonomies = node.get_node_attribute_object().get_dic_taxonomies()

        # check if the protein being processed respects the species restriction
        passes_species_test = self._passes_species_test(required_protein_tax_id= tax_id_value,
                                                        dic_protein_taxonomies= dic_protein_taxonomies)

        if passes_species_test == 1:

            # passes the species restriction: continue processing information for the protein
            list_protein_description = node.get_node_attribute_object().get_descriptions()
            list_protein_function = node.get_node_attribute_object().get_functions()
            list_protein_subcellularLocation = node.get_node_attribute_object().get_dic_cell_location().keys()
            list_protein_user_keywords = node.get_node_attribute_object().get_keywords_appearing()
            protein_expression = node.get_node_attribute_object().get_expression()
            protein_special_labels = node.get_node_attribute_object().get_special_labels()


	    if node.is_root():
		root_html_prefix= "<u><font color=FFFC17>"
		root_html_sufix= "</font></u>"
	    else:
		root_html_prefix= ""
		root_html_sufix= ""

            if list_protein_user_keywords:
                # if there were user keywords in this protein, write it underlined in the html table
                html_prefix = "<u><font color=CC0000>"
                html_sufix = "</font></u>"
            else:
                html_prefix = ""
                html_sufix = ""

            if protein_expression == "over_expressed":
                expression_html_prefix="<font color=FF33FF>"
                expression_html_sufix="</font>"
                
            elif protein_expression == "infra_expressed":
                expression_html_prefix="<font color=33FF00>"
                expression_html_sufix="</font>"
            else:
                expression_html_prefix=""
                expression_html_sufix=""

	    
	    special_labels_string = self._special_labels_list_2_string(special_labels_list= protein_special_labels, tag="", separator=" ")

	    if special_labels_string:
		special_labels_html_prefix = "<font color=CC0000>"
		special_labels_html_sufix = "</font>"
	    else:
		special_labels_html_prefix = ""
		special_labels_html_sufix = ""
		

            if output_mode == "extended":
                
                output_target.write("========================================================================%s" %PianaGlobals.line_separators[format_mode] )
                output_target.write("Information for protein %s (proteinPianas = %s)%s%s" %(protein_user_type,
                                                                                           set_associated_proteinPianas,
                                                                                           PianaGlobals.line_separators[format_mode],
                                                                                           PianaGlobals.line_separators[format_mode] ))
                if format_mode == "txt":
                    output_target.write("(other equivalent codes: %s)%s" %(list_protein_ext_code, PianaGlobals.line_separators[format_mode]) )
                elif format_mode == "html":
                    output_target.write("equivalent codes: %s%s" %(list_protein_ext_code, PianaGlobals.line_separators[format_mode] ))
            # END OF if output_mode == "extended":

            # in case a list of connecting nodes was passed , print information for the protein as well as which are the root nodes that it connects
            # list_connects_nodes can be None (no connecting node info desired) or a list (it could be an empty list. (ie. we wanted information but list is empty))
            if list_connects_nodes is not None:

                #number_of_connected_nodes = len(dic_connected_roots_extcode)
                number_of_connected_nodes = len(self.ext_code_connecting[protein_user_type])
                if number_of_connected_nodes > 1:
                    if output_mode == "compact":
                        if format_mode == "txt":
                            #output_target.write("%s%s connects %s root_proteins%s%s%sdatabases %s%sdescription=%s%sfunction=%s%sexpression=%s%sspecial_labels=%s%stax_ids=%s%s" %(
                            output_target.write("%s%s connects %s root_proteins%s%s%sdescription=%s%sfunction=%s%sexpression=%s%sspecial_labels=%s%stax_ids=%s%s" %(
                                PianaGlobals.line_separators[format_mode],
                                protein_user_type,                             number_of_connected_nodes,
                                PianaGlobals.tab_separators[format_mode],
                                #dic_connected_roots_extcode.keys(),                       PianaGlobals.tab_separators[format_mode],
                                list(self.ext_code_connecting[protein_user_type]),              PianaGlobals.tab_separators[format_mode],
                                #dic_all_dbs.keys(),                                       PianaGlobals.tab_separators[format_mode],
                                list_protein_description,                                 PianaGlobals.tab_separators[format_mode],
                                list_protein_function,                                    PianaGlobals.tab_separators[format_mode],
                                protein_expression,                                       PianaGlobals.tab_separators[format_mode],
                                special_labels_string,                                    PianaGlobals.tab_separators[format_mode],
				dic_protein_taxonomies.keys(),                            PianaGlobals.line_separators[format_mode]) )
                            
                        elif format_mode == "html":
                            output_target.write("""\n<tr><td>%s%s%s</td><td align=center><a href="%s">iHop</a><br></td><td>%s</td><td>%s</td><td>is_root=%s%s%s expression=%s%s%s<br>special_labels=%s%s%s</td><td>%s</td></tr>\n""" %(
                                html_prefix,
                                protein_user_type,
                                html_sufix,
				"http://www.ihop-net.org/UniPub/iHOP/?search=%s&field=all&ncbi_tax_id=%s" %(protein_user_type, tax_id_value),
                                list_protein_description,
                                list_protein_function,
				root_html_prefix, node.is_root(), root_html_sufix,
                                expression_html_prefix, protein_expression, expression_html_sufix,
				special_labels_html_prefix, special_labels_string, special_labels_html_sufix,
                                #dic_connected_roots_extcode.keys() ))
                                list(self.ext_code_connecting[protein_user_type])))
                            
                        # END OF elif format_mode == "html":
                    # END OF if output_mode == "compact":
                    elif output_mode == "extended":

                        if format_mode == "txt":
                            # this is printing the same as in 'compact' output_mode, but I prefer to keep it separate in case in the future
                            # we want to add some extra information in output_mode 'all'
                            #output_target.write("%s%s connects %s root_proteins%s%s%sdatabases %s%sdescription=%s%sfunction=%s%sexpression=%s%sspecial_labels=%s%s" %(
                            output_target.write("%s%s connects %s root_proteins%s%s%sdescription=%s%sfunction=%s%sexpression=%s%sspecial_labels=%s%s" %(
                                PianaGlobals.line_separators[format_mode],
                                protein_user_type,                             number_of_connected_nodes,
                                PianaGlobals.tab_separators[format_mode],
                                #dic_connected_roots_extcode.keys(),                       PianaGlobals.tab_separators[format_mode],
                                list(self.ext_code_connecting[protein_user_type]),        PianaGlobals.tab_separators[format_mode],
                                #dic_all_dbs.keys(),                                       PianaGlobals.tab_separators[format_mode],
                                list_protein_description,                                 PianaGlobals.tab_separators[format_mode],
                                list_protein_function,                                    PianaGlobals.tab_separators[format_mode],
                                protein_expression,                                       PianaGlobals.tab_separators[format_mode],
				special_labels_string,                                    PianaGlobals.line_separators[format_mode]) )

                        elif format_mode == "html":
                            output_target.write("%sLinker %s%s%s -- description %s -- function %s (is_root=%s%s%s expression=%s%s%s special_labels=%s) connects root proteins:%s%s\n" %(
                                PianaGlobals.line_separators[format_mode],
                                html_prefix,
                                protein_user_type,
                                html_sufix,
                                list_protein_description,
                                list_protein_function,
				root_html_prefix, node.is_root(), root_html_sufix,
                                expression_html_prefix, protein_expression, expression_html_sufix, 
				special_labels_string,
                                #dic_connected_roots_extcode.keys(),
                                list(self.ext_code_connecting[protein_user_type]),
                                PianaGlobals.line_separators[format_mode] ))

                    # END OF elif output_mode == "extended":
                # END OF if number_of_connected_nodes > 1:
            # END OF if list_connects_nodes is not None:

            else:
                # there is no list of connecting nodes (i.e info requested about all proteins, not just the linker proteins)

                if output_mode == "compact":
                    if format_mode == "txt":
                        # output in compact mode is (txt) external_code<TAB>protein description<TAB>protein function<TAB>alternative external codes<NEW LINE>
                        output_target.write("%s%s%s%s%s%sroot=%s%sexpression=%s%sspecial_labels=%s%stax_ids=%s" %(protein_user_type,
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  list_protein_description,
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  list_protein_function,
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  node.is_root(),
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  protein_expression,
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  special_labels_string,
                                                                                                                  PianaGlobals.tab_separators[format_mode],
                                                                                                                  dic_protein_taxonomies.keys()  ))
                        
                        for user_keyword in list_protein_user_keywords:
                            output_target.write("%suser_keyword=%s" %(PianaGlobals.tab_separators[format_mode], user_keyword))

                        for protein_ext_code in list_protein_ext_code:
                            type = protein_ext_code.split(":")[0]
                            if type != PianaGlobals.sequence_col and type != PianaGlobals.sequenceMD5_col:
                                output_target.write("%s%s" %(PianaGlobals.tab_separators[format_mode], protein_ext_code ))

                            #if type != PianaGlobals.proteinSequence_col and type != PianaGlobals.proteinMD5_col:
                            #    # do not print sequence and md5
                            #    output_target.write("%s%s" %(PianaGlobals.tab_separators[format_mode], protein_ext_code ))

                            # TO DO!!! Avoid printing identifiers that do not give useful information (md5? sequence?)


                        output_target.write("%s" %PianaGlobals.line_separators[format_mode])

                        
                    elif format_mode == "html":

                        output_target.write("""<tr><td>%s%s%s</td><td align=center><a href="%s">iHop</a></td><td>%s</td><td>%s</td><td>is_root=%s%s%s expression=%s%s%s<br>special_labels=%s</td></tr>\n""" %(
                            html_prefix,
                            protein_user_type,
                            html_sufix,
			    "http://www.ihop-net.org/UniPub/iHOP/?search=%s&field=all&ncbi_tax_id=%s" %(protein_user_type, tax_id_value),
                            list_protein_description,
                            list_protein_function,
                            root_html_prefix, node.is_root(), root_html_sufix,
                            expression_html_prefix, protein_expression, expression_html_sufix,
			    special_labels_string ))

                elif output_mode == "extended":

                    for protein_taxonomy in dic_protein_taxonomies:
                        output_target.write("%sprotein %s taxonomy: %s%s" %(PianaGlobals.line_separators[format_mode],
                                                                            protein_user_type, protein_taxonomy,
                                                                            PianaGlobals.line_separators[format_mode]) )

                    for description in list_protein_description:
                        output_target.write("%sprotein %s description: %s%s" %(PianaGlobals.line_separators[format_mode],
                                                                               protein_user_type, description,
                                                                               PianaGlobals.line_separators[format_mode]) )

                    for function in list_protein_function:
                        output_target.write("%sprotein %s function: %s%s" %(PianaGlobals.line_separators[format_mode],
                                                                            protein_user_type, function,
                                                                            PianaGlobals.line_separators[format_mode]) )

                    for subcellularLocation in list_protein_subcellularLocation:
                        output_target.write("%sprotein %s subCellularLocation: %s%s" %(PianaGlobals.line_separators[format_mode],
                                                                                       protein_user_type, subcellularLocation,
                                                                                       PianaGlobals.line_separators[format_mode]) )
                        
                    for user_keyword in list_protein_user_keywords:
                        output_target.write("%sprotein %s user keyword: %s%s" %(PianaGlobals.line_separators[format_mode],
                                                                                protein_user_type, user_keyword, PianaGlobals.line_separators[format_mode]) )

		    output_target.write("root=%s%sexpression=%s%sspecial_labels=%s%s" %(node.is_root(),
										 PianaGlobals.line_separators[format_mode],
										 protein_expression,
										 PianaGlobals.line_separators[format_mode],
										 special_labels_string,
										 PianaGlobals.line_separators[format_mode]))

                        
                    output_target.write("========================================================================%s" %PianaGlobals.line_separators[format_mode])
                # END OF elif output_mode == "extended": 
            # END OF else: (if list_connects_nodes is not None:)
        # END OF if passes_species_test == 1:

    def _print_table_headers_protein_information(self, output_target=None, connecting_nodes_dic= None, user_keywords=[]):
        """
        prints the table headers for protein information to output_target
        """
        output_target.write("<html><body><br><br>\n")
        output_target.write("Keywords used to highlight proteins: %s<br><br>\n" %user_keywords)
        output_target.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080">\n""")

        if connecting_nodes_dic is None:
            output_target.write("<tr><td align=center><b>Protein</b></td><td align=center><b>External <br>links</b></td><td align=center><b>Description</b></td><td align=center><b>Function</b></td><td align=center><b>Other <br>information</b></td></tr>\n")
        else:
            output_target.write("<tr><td align=center><b>Linker protein</b></td><td align=center><b>External <br>links</b></td><td align=center><b>Description</b></td><td align=center><b>Function</b></td><td align=center><b>Other <br>information</b></td><td><b>Roots <br>connected</b></td></tr>\n")

            
    def _print_table_closure_protein_information(self, output_target=None):
        """
        prints the table headers for protein information to output_target
        """
        output_target.write("\n</table></body></html>\n")


                
    def output_all_proteins_information(self,  id_type, output_target, output_mode= "compact", format_mode= "txt",
                                        filter_mode="all", alternative_id_types= [], tax_id_value= 0, list_keywords= [],
                                        user_protein_names = {}, file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
					dic_special_proteins_file=None):
        """
        writes information about all proteins in the network in file "output_target"
        
        The description of the arguments in this method is the same as in method output_protein_information()
        """

        initial_time = time.time()
        
        if output_mode == "compact" and format_mode == "html":
            self._print_table_headers_protein_information(output_target= output_target, connecting_nodes_dic=None, user_keywords=list_keywords)

        #self.already_printed_proteins = {} # used in output_protein_information to avoid printing information for the same protein many times
        self.already_printed_proteins = Set()
        
        for node in self.get_node_object_list():

            self.output_protein_information(proteinPiana_value= node.get_node_id(),
                                            id_type=id_type,
                                            output_target= output_target,
                                            output_mode= output_mode,
                                            format_mode= format_mode,
                                            filter_mode= filter_mode,
                                            alternative_id_types=alternative_id_types,
                                            list_keywords= list_keywords,
                                            tax_id_value= tax_id_value,
                                            list_connects_nodes= None,
                                            list_proteinPianas= None,
                                            key_list= -1,
                                            user_protein_names = user_protein_names,
                                            file_over_expressed= file_over_expressed,
                                            file_infra_expressed= file_infra_expressed,
                                            expression_id_type= expression_id_type,
					    dic_special_proteins_file= dic_special_proteins_file )

            
        if output_mode == "compact" and format_mode == "html":
             self._print_table_closure_protein_information(output_target=output_target)

        #print "Total time for printing proteins information: %s" %(time.time()-initial_time)

            
    def output_root_proteins_information(self, id_type,  output_target, output_mode= "compact", format_mode= "txt", filter_mode="all",
                                         alternative_id_types= [], tax_id_value= 0, list_keywords= [],
                                         user_protein_names = {}, file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
					 dic_special_proteins_file=None):
        """
        writes information about root proteins in the network in file "output_target"
        
        The description of the arguments in this method is the same as in method output_protein_information()
        """
        initial_time = time.time()
        if output_mode == "compact" and format_mode == "html":
            self._print_table_headers_protein_information(output_target= output_target, connecting_nodes_dic=None, user_keywords=list_keywords)

        #self.already_printed_proteins = {} # used in output_protein_information to avoid printing the same protein many times
        self.already_printed_proteins = Set()
            
        root_protein_ids = self.get_root_node_ids()

        for node_id in root_protein_ids:

            self.output_protein_information(proteinPiana_value= node_id,
                                            id_type=id_type,
                                            output_target= output_target,
                                            output_mode= output_mode,
                                            format_mode= format_mode,
                                            filter_mode= filter_mode,
                                            alternative_id_types=alternative_id_types,
                                            list_keywords= list_keywords,
                                            tax_id_value= tax_id_value,
                                            list_connects_nodes= None,
                                            list_proteinPianas= None,
                                            key_list= -1,
                                            user_protein_names = user_protein_names,
                                            file_over_expressed= file_over_expressed,
                                            file_infra_expressed= file_infra_expressed,
                                            expression_id_type= expression_id_type,
					    dic_special_proteins_file= dic_special_proteins_file)

        if output_mode == "compact" and format_mode == "html":
             self._print_table_closure_protein_information(output_target=output_target)

        #print "Total time for printing proteins information: %s" %(time.time()-initial_time)



    def output_connecting_proteins_information(self, id_type,  output_target, output_mode= "compact", format_mode= "txt", filter_mode="all",
                                               alternative_id_types= [], tax_id_value = 0, list_keywords= [],
                                               user_protein_names = {},
                                               file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, 
					       dic_special_proteins_file=None):
        """
        writes information  to file object "output_target" about linker proteins in the network
        
        The description of the arguments in this method is the same as in method output_protein_information()
        """
        initial_time = time.time()

        connecting_node_dic = self.get_connecting_nodes_dic() # returns a dictionary with proteinPiana:[root_nodes]

        if output_mode == "compact" and format_mode == "html":
            #self._print_table_headers_protein_information(output_target= output_target, connecting_nodes_dic=connecting_node_dic, user_keywords=list_keywords)
            self._print_table_headers_protein_information(output_target= output_target, connecting_nodes_dic=connecting_node_dic, user_keywords=list_keywords)

        #self.already_printed_proteins = {} # used in output_protein_information to avoid printing the same protein many times
        self.already_printed_proteins = Set()

        #javi added:
        # create the connecting dictionary using ext codes instead of proteinPianas
        self._create_ext_code_connecting(id_type= id_type,
                                         alternative_id_types= alternative_id_types,
                                         connecting_proteinPianas = connecting_node_dic)

        #javi:
        already_printed_nodes = Set([])

        for actual_node in self.get_node_object_list():
            if self.ext_code_connecting.has_key(self.get_unified_name( proteinPiana = actual_node.get_node_id(),
                                                                       id_type=id_type,
                                                                       alternative_id_types=alternative_id_types)):
                self.output_protein_information(proteinPiana_value= actual_node.get_node_id(),
                                                id_type=id_type,
                                                alternative_id_types=alternative_id_types,
                                                output_target= output_target,
                                                output_mode= output_mode,
                                                format_mode= format_mode,
                                                filter_mode= filter_mode,
                                                list_keywords= list_keywords,
                                                tax_id_value= tax_id_value,
                                                user_protein_names = user_protein_names,
                                                list_connects_nodes= connecting_node_dic[actual_node.get_node_id()],
                                                list_proteinPianas= None,
                                                key_list= -1,
                                                file_over_expressed= file_over_expressed,
                                                file_infra_expressed= file_infra_expressed,
                                                expression_id_type= expression_id_type,
                                                dic_special_proteins_file= dic_special_proteins_file)

        ## OLD RAMON METHOD: It was getting the connecting proteins from the original network, not from the unified one


##         for connecting_node_id in connecting_node_dic:

##             self.output_protein_information(proteinPiana_value= connecting_node_id,
##                                             id_type=id_type,
##                                             alternative_id_types=alternative_id_types,
##                                             output_target= output_target,
##                                             output_mode= output_mode,
##                                             format_mode= format_mode,
##                                             filter_mode= filter_mode,
##                                             list_keywords= list_keywords,
##                                             tax_id_value= tax_id_value,
##                                             user_protein_names = user_protein_names,
##                                             list_connects_nodes= connecting_node_dic[connecting_node_id],
##                                             list_proteinPianas= None,
##                                             key_list= -1,
##                                             file_over_expressed= file_over_expressed,
##                                             file_infra_expressed= file_infra_expressed,
##                                             expression_id_type= expression_id_type,
##                                             dic_special_proteins_file= dic_special_proteins_file)
            

        if output_mode == "compact" and format_mode == "html":
             self._print_table_closure_protein_information(output_target=output_target)

        #print "Total time for printing proteins information: %s" %(time.time()-initial_time)


    def output_list_proteins_information(self,  list_proteins, input_id_type, output_id_type, output_target, output_mode= "compact",
                                         format_mode= "txt", filter_mode="all", alternative_id_types= [], list_keywords= [], 
                                         tax_id_value= 0, key_list=-1, 
                                         file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, 
					 dic_special_proteins_file=None):
        """
        writes information  to file object output_target about all proteins in list "list_proteins" (of type "input_id_type")

        This method is used to print information about proteins that are not in a network!!!! (ie. it won't print information about interactions of the proteins)

        The description of the rest of the arguments of this method is the same as in method output_protein_information()
        """

        initial_time = time.time() 
        if output_mode == "compact" and format_mode == "html":
            self._print_table_headers_protein_information(output_target= output_target, connecting_nodes_dic=None, user_keywords=list_keywords)

        #self.already_printed_proteins = {} # used in output_protein_information() to avoid printing the same protein many times
        self.already_printed_proteins = Set()


        this_user_protein_names= {} # in output_list_* the user protein names are the same as those in the list, therefore it is
                                    # not required as an argument and it is populated inside this method
            
        # get all the proteinPianas associated to the proteins in the list
        dic_proteinPiana = {}
        
        for user_protein in list_proteins:
            this_user_protein_names[user_protein] = None # will be used to fix preference of user names over the other possible names

            temp_list_user_proteinPiana = self.piana_access.get_list_proteinPiana(protein_id= user_protein,
                                                                                  id_type= input_id_type,
                                                                                  tax_id_value= tax_id_value,
                                                                                  source_db_info="no" )

            for user_proteinPiana in temp_list_user_proteinPiana:
                dic_proteinPiana[user_proteinPiana] = None
        # END OF for user_protein in list_proteins:

        
        list_proteinPianas = dic_proteinPiana.keys()
        for proteinPiana in dic_proteinPiana:
            # list_proteinPianas is used to unify nodes and decide which names will be used for the proteins given by the user
            self.output_protein_information(proteinPiana_value= proteinPiana,
                                            id_type= output_id_type,
                                            output_target= output_target,
                                            output_mode= output_mode,
                                            format_mode= format_mode,
                                            filter_mode= filter_mode,
                                            list_connects_nodes= None,
                                            alternative_id_types= alternative_id_types,
                                            list_keywords= list_keywords,
                                            tax_id_value= tax_id_value,
                                            list_proteinPianas= list_proteinPianas,
                                            key_list= key_list,
                                            user_protein_names = this_user_protein_names,
                                            file_over_expressed= file_over_expressed,
                                            file_infra_expressed= file_infra_expressed,
                                            expression_id_type= expression_id_type,
					    dic_special_proteins_file= dic_special_proteins_file)
        # END OF for proteinPiana in list_proteinPiana:

        if output_mode == "compact" and format_mode == "html":
             self._print_table_closure_protein_information(output_target=output_target)

        #print "Total time for printing proteins information: %s" %(time.time()-initial_time)

    def output_proteins_with_network_class(self, id_type,  output_target, 
                                           alternative_id_types= [], tax_id_value = 0, 
                                           user_protein_names = {},file_over_expressed= None,
                                           file_infra_expressed= None, expression_id_type= None,
                                           dic_special_proteins_file={}):
        """
        writes to file object "output_target" all the proteins in the network with a label indicating whether they are root, linker or partner
        
        The description of the arguments in this method is the same as in method output_protein_information()
        """

        # make sure self.dic_protein_naming is filled
        self._fill_dic_protein_naming(id_type=id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= None, key_list=None, user_protein_names= user_protein_names)
        
        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value= tax_id_value)

        # make sure the unified nodes and edges are created
        self._create_unified_network( id_type=id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= None, user_keywords=[] )


	# fill the special proteins dics so that the information is available afterwards
	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                       id_type=id_type, alternative_id_types=alternative_id_types)


        #self.already_printed_proteins = {}
        self.already_printed_proteins = Set()

        #javi: Now, this dictionary is filled inner the method self._create_ext_code_connecting
        #connecting_prots = self.get_connecting_nodes_dic() # returns a dictionary with proteinPiana:[root_nodes] where root_nodes are proteinPianas
        

        # create the connecting dictionary using ext codes instead of proteinPianas
        self._create_ext_code_connecting(id_type= id_type,
                                         alternative_id_types= alternative_id_types)


        for one_node in self.get_node_object_list():

            
            protein_id = self.get_unified_name( proteinPiana= one_node.get_node_id(),
                                                id_type=id_type, alternative_id_types=alternative_id_types)
            
            #if self.already_printed_proteins.has_key(protein_id):
            if protein_id in self.already_printed_proteins:
                # do not print  a protein that has already been printed (the unified node contained all the information associated to this ext_code)
                continue
            
            #self.already_printed_proteins[protein_id] = None
            self.already_printed_proteins.add(protein_id)
            
            unified_node = self.ext_code_unified_node_correspondence[protein_id]

            # get taxonomies for this unified node
            dic_protein_taxonomies = unified_node.get_node_attribute_object().get_dic_taxonomies()

            # check if the protein being processed respects the species restriction
            passes_species_test = self._passes_species_test(required_protein_tax_id= tax_id_value,
                                                            dic_protein_taxonomies= dic_protein_taxonomies)

            if passes_species_test != 1:
                # skip protein if it doesn't pass the species test
                continue

            # get special labels for this protein
            if self.dic_special_ext_codes.has_key(protein_id):
                special_labels_string= self._special_labels_list_2_string(special_labels_list=self.dic_special_ext_codes[protein_id],
                                                                          tag="label=",
                                                                          separator="\t")
            else:
                special_labels_string= ""

            
            if unified_node.is_root():
                node_class = "root"
                
            elif self.ext_code_connecting.has_key(protein_id):
                number_of_connected_roots = len(self.ext_code_connecting[protein_id])
                node_class = "linker_" + str(number_of_connected_roots) 
            else:
                node_class = "partner"

            if special_labels_string:  chain_to_print = "\t" + special_labels_string
            else:                      chain_to_print = ""
                
                
                
            output_target.write("%s\tclass=%s\texpression=%s%s\n" %(protein_id,
                                                                    node_class,
                                                                    unified_node.get_node_attribute_object().get_expression(),
                                                                    chain_to_print   ))
                
        # END OF for one_node in self.get_node_object_list():
    

    def get_proteins_expression(self, list_proteins=[], id_type=None,
                                alternative_id_types= [], tax_id_value = 0, 
                                file_over_expressed= None, file_infra_expressed= None, expression_id_type= None):
        """
        given a list [protein1, protein2, ...] (of type "id_type") this method returns a dic
                     {protein1:'over_expressed', protein2:'over_and_under_expressed', protein3:...}

        Attention! If expression is None, the protein will not appear in the dictionary

        protein ext_codes in list_proteins do not need to be network proteins, neither unified
        """
        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value = tax_id_value)
        
        dic_expression_ext_codes = {}

        for one_protein in list_proteins:
            list_proteinPianas = self.piana_access.get_list_proteinPiana(protein_id= one_protein,
                                                                         id_type=id_type ,
                                                                         tax_id_value=tax_id_value,
                                                                         source_db_info="no")
            expression = None
            for one_proteinPiana in list_proteinPianas:

                if self.over_expressed_proteinPianas.has_key(one_proteinPiana) and self.infra_expressed_proteinPianas.has_key(one_proteinPiana):
                    expression = PianaGlobals.over_and_infra_expressed_protein
                    break
                
                elif self.over_expressed_proteinPianas.has_key(one_proteinPiana):
                    if expression == PianaGlobals.infra_expressed_protein:
                        expression = PianaGlobals.over_and_infra_expressed_protein
                        break
                    else:
                        expression = PianaGlobals.over_expressed_protein
                elif self.infra_expressed_proteinPianas.has_key(one_proteinPiana):
                    if expression == PianaGlobals.over_expressed_protein:
                        expression = PianaGlobals.over_and_infra_expressed_protein
                        break
                    else:
                        expression = PianaGlobals.infra_expressed_protein
            # END OF for one_proteinPiana in list_proteinPianas:

            if expression:
                dic_expression_ext_codes[one_protein] = expression
        # END OF for one_protein in list_proteins:

        return dic_expression_ext_codes
    

    def get_proteins_special_labels(self, list_proteins=[], id_type=None, 
                                    alternative_id_types= [], tax_id_value = 0, 
                                    file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, 
                                    dic_special_proteins_file={} ):
        """
        
        given a list [protein1, protein2, ...] (of type "id_type") this method returns a dictionary
        
        
             {  protein1:{'expression':expression_value, 'labels':[label1, label2, ...]},
                protein2: ....
             }
        
        
        "list_proteins" is a list of external codes that has been obtained by calling to another PianaGraph method
                -> you cannot set any list of proteins here. This list must be the result of some operation over a group
                   of unified names
        
	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
        follows structure:   { label1: (id_type1, path_to_file1),
        label2: (id_type2, path_to_file2),
        ...............
        }
        
        
        if the following parameters are set, then the expression info will be inserted in the results. Otherwise, expression info will be None
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed
        
        
        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed
        
        Attention! A name unification must have been made previously for the proteinPianas of ext_codes in list_proteins, using the
                   same id_type you are using now (either via a call to get_list_unified_names or other network
                   operations that call _fill_dic_protein_naming)
        """
        # fill the special proteins dics so that the information is available afterwards
	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                        id_type=id_type, alternative_id_types=alternative_id_types)
                            

        if file_over_expressed or file_infra_expressed:
        
            expression_dic_ext_codes = self.get_proteins_expression(list_proteins=list_proteins, id_type=id_type,
                                                                    alternative_id_types= alternative_id_types,
                                                                    tax_id_value =tax_id_value, 
                                                                    file_over_expressed= file_over_expressed,
                                                                    file_infra_expressed= file_infra_expressed,
                                                                    expression_id_type=expression_id_type )
        else:
            expression_dic_ext_codes = {}
               
        dic_to_return = {}             
        for unified_protein_name in list_proteins:

            dic_to_return[unified_protein_name] = {'expression':None, 'labels':[]}
            
            if expression_dic_ext_codes.has_key(unified_protein_name):
                dic_to_return[unified_protein_name]['expression'] = expression_dic_ext_codes[unified_protein_name]
                
            if self.dic_special_ext_codes.has_key(unified_protein_name):
                dic_to_return[unified_protein_name]['labels'].extend(self.dic_special_ext_codes[unified_protein_name])
        # END OF for one_protein in list_proteins:
        
        return dic_to_return
             

    def print_table_interaction(  self,
				  output_target= None,
                                  format_mode= "txt",
                                  root_connectivity= None,
                                  extcode_start= None,
                                  num_neigh_start= None,
                                  node_type_start= None,
                                  expression_start= None,
				  special_labels_string_start= None,
                                  list_protein_user_keywords_start= None,
                                  extcode_end= None,
                                  num_neigh_end= None,
                                  node_type_end= None,
                                  expression_end= None,
				  special_labels_string_end= None,
                                  list_protein_user_keywords_end= None,
                                  same_loc= None,
                                  same_species= None,
                                  list_interaction_dbs= None,
                                  list_interaction_methods= None,
                                  list_interaction_pubmeds= None,
                                  fill_interaction_table_strings=0):
        """

        Prints one interaction  between "extcode_start" and "extcode_end" in a table to "output_target"
          -> html table if "format_mode" is 'html' or tabulated text if "format_mode" is 'txt'

        "output_target" is a file object with writing rights

        "list_interaction_dbs" is a list of the database names where this interaction appears

        "list_interaction_pubmeds" is a list of the pubmed IDs where this interaction appears

        in format_mode 'html', the table is simplified to fit to the screen.
 
        """
        if format_mode == "txt":
            # printing root connectivity of the protein that is not the root protein
            if root_connectivity is not None:
                if root_connectivity != -1:     output_target.write("connectivity=%s" %(root_connectivity))
                else:                           output_target.write("connectivity=None")
            else:
                output_target.write("connectivity=None")

            # printing start protein and number of neighbours
            output_target.write("\tprotein_1=%s\tneighbours_1=%s" %(extcode_start, num_neigh_start) )
            
            # printing start protein root info
            output_target.write("\troot_1=%s" %(node_type_start) )
            
            # printing start protein expression information
            output_target.write("\texpression_1=%s" %(expression_start) )

            # printing start protein expression information
            output_target.write("\tspecial_labels_1=%s" %(special_labels_string_start) )

            # printing end protein and number of neighbours
            output_target.write("\tprotein_2=%s\tneighbours_2=%s" %(extcode_end, num_neigh_end) )
            
            # printing end protein root info
            output_target.write("\troot_2=%s" %(node_type_end) )
            
            # printing start protein and number of neighbours
            output_target.write("\texpression_2=%s" %(expression_end) )

            # printing start protein expression information
            output_target.write("\tspecial_labels_2=%s" %(special_labels_string_end) )

            # printing cellular localization and species information
            output_target.write("\tlocation=%s\tspecies=%s" %(same_loc, same_species) )

            # printing source databases
            #for interaction_database in list_interaction_dbs:
            #    output_target.write("\tdb=%s" %interaction_database)
            output_target.write("\tdb=%s" %("\tdb=".join(list_interaction_dbs)))
            
            # printing methods
            #for interaction_method in list_interaction_methods:
            #    output_target.write("\tmethod=%s" %interaction_method)
            output_target.write("\tmethod=%s" %("\tmethod=".join(list_interaction_methods)))

            # printing pubmedIDs (javi)
            if list_interaction_pubmeds:
                output_target.write("\tpubmed=%s" %("\tpubmed=".join(list_interaction_pubmeds)))
            else:
                output_target.write("\tpubmed=")

            # END OF for interaction_database in PianaGlobals.interaction_databases:
            output_target.write("\n")
        # END OF if format_mode == "txt":

        elif format_mode == "html":

            if list_protein_user_keywords_start:
                # if there were user keywords in this protein, write it underlined in the html table
                start_html_prefix = "<u><font color=CC0000>"
                start_html_sufix = "</font></u>"
            else:
                start_html_prefix = ""
                start_html_sufix = ""

            if expression_start == "over_expressed":
                expression_html_prefix_start="<font color=FF33FF>"
                expression_html_sufix_start="</font>"
                
            elif expression_start == "infra_expressed":
                expression_html_prefix_start="<font color=33FF00>"
                expression_html_sufix_start="</font>"
            else:
                expression_html_prefix_start=""
                expression_html_sufix_start=""
                
            if list_protein_user_keywords_end:
                # if there were user keywords in this protein, write it underlined in the html table
                end_html_prefix = "<u><font color=CC0000>"
                end_html_sufix = "</font></u>"
            else:
                end_html_prefix = ""
                end_html_sufix = ""

            if expression_end == "over_expressed":
                expression_html_prefix_end="<font color=FF33FF>"
                expression_html_sufix_end="</font>"
                
            elif expression_end == "infra_expressed":
                expression_html_prefix_end="<font color=33FF00>"
                expression_html_sufix_end="</font>"
            else:
                expression_html_prefix_end=""
                expression_html_sufix_end=""

	    if special_labels_string_start:
		to_print_special_labels_start = "<br>special_labels=%s" %(special_labels_string_start)
	    else:
		to_print_special_labels_start = ""

	    if special_labels_string_end:
		to_print_special_labels_end = "<br>special_labels=%s" %(special_labels_string_end)
	    else:
                to_print_special_labels_end = ""

	    string_to_print = "\n<tr>"
	    string_to_print += """<td>%s%s%s</td><td align=center><a href="%s">iHop</a><br></td>""" %(start_html_prefix, extcode_start, start_html_sufix,
									 "http://www.ihop-net.org/UniPub/iHOP/?search=%s&field=all&ncbi_tax_id=0" %(extcode_start))
	    string_to_print += "<td>is_root=%s expression=%s%s%s %s</td>" %(node_type_start,
                                                                                         expression_html_prefix_start, expression_start, expression_html_sufix_start,
											  to_print_special_labels_start  )
	    string_to_print += """<td>%s%s%s</td><td align=center><a href="%s">iHop</a><br></td>""" %(end_html_prefix, extcode_end, end_html_sufix,
									 "http://www.ihop-net.org/UniPub/iHOP/?search=%s&field=all&ncbi_tax_id=0" %(extcode_end))
	    string_to_print += "<td>is_root=%s expression=%s%s%s %s</td>" %(node_type_end,
                                                                                 expression_html_prefix_end, expression_end, expression_html_sufix_end,
										to_print_special_labels_end	 )
	    string_to_print += "<td><center>%s</center></td><td><center>%s<center></td>" %(same_loc, same_species)
	    string_to_print += "<td>"
            # printing source databases
            #for interaction_database in list_interaction_dbs:
            #     string_to_print += " %s " %interaction_database
            string_to_print += " ".join(list_interaction_dbs)
	    string_to_print += "</td>"
	    string_to_print += "<td>"

            # printing detection method
            #for interaction_method in list_interaction_methods:
	    #	string_to_print +=" %s " %interaction_method
            string_to_print +=" ".join(list_interaction_methods) 
	    string_to_print += "</td>"
	    string_to_print += "</tr>\n"


            # printing pubmedIDs (javi)
            if list_interaction_pubmeds:
                string_to_print +=" ".join(list_interaction_pubmeds)
            else:
                string_to_print += "</td>"
                string_to_print += "</tr>\n"

            output_target.write(string_to_print)
	    
            if fill_interaction_table_strings:
                # keep in the dictionary of strings the interaction lines associated to the two proteins
                if self.interaction_table_strings.has_key(extcode_start):
                    self.interaction_table_strings[extcode_start].append(string_to_print) 
                else:
                    self.interaction_table_strings[extcode_start] = [string_to_print]

                if self.interaction_table_strings.has_key(extcode_end):
                    self.interaction_table_strings[extcode_end].append(string_to_print) 
                else:
                    self.interaction_table_strings[extcode_end] = [string_to_print]

        # END OF elif format_mode == "html":

        else:
            raise ValueError("Incorrect format mode given for outputting table")
  



    def print_image_interaction(output_target= None,
				format_mode = None, 
				edge_object= None,
				piana_access = None,
				extcode_start= None,
				extcode_end= None,
				list_interaction_dbs= None,
				list_interaction_methods= None,
				network_attributes_prefix = None):
        """
        prints to "output_target" one interaction between "extcode_start" and "extcode_end" in format defined by "format_mode"

        "output_target" is a file object with write rights

        "format_mode" sets the format used for printing the network interaction (right now, the only valid format is 'dot')
        
        "edge_object" is the GraphEdge object of this interaction

        "list_interaction_dbs" are the database names where this interaction appears

        files written by this method can be used to create an image of the network using neato from package Graphviz:

           $> cat "output_of_this_method" | neato -Tgif -o output_file.gif
    
        "network_attributes_prefix" is the directory + file prefix that will be used to create the files that describe the information for this node
            -> these files can then be used by cytoscape to visualize specific properties of the node
        
        """
        if format_mode == "dot":        
	    # setting the edge color
	    number_of_dbs = len(list_interaction_dbs)

	    if number_of_dbs == 1:
		# if there is just one db with this interaction, print its colour
		interaction_style = piana_access.get_interaction_line_style("normal")
		interaction_color= piana_access.get_interaction_source_database_color(list_interaction_dbs[0])

	    elif number_of_dbs == 0:
		# If interaction does not belong to any database, then default color given... (it is a prediction made just now)
		interaction_color= piana_access.get_interaction_source_database_color("")
		# setting the line style of the edge to be printed
		if edge_object.is_extended() or edge_object.is_propagated():
		    if edge_object.is_extended():
			interaction_style = piana_access.get_interaction_line_style("extended")
		    else:
			interaction_style = piana_access.get_interaction_line_style("propagated")

		else:
		    interaction_style = piana_access.get_interaction_line_style("normal")

	    else:
		# if it is an interaction that appears in two or more different databases, print a special colour to indicate it
		interaction_style = piana_access.get_interaction_line_style("normal")
		interaction_color= piana_access.get_interaction_source_database_color("intersection")
            


            output_target.write('"%s" -- "%s" [len=%s, color=%s, style=%s];\n' %(extcode_start, extcode_end, PianaGlobals.dot_edge_lenght,
										 interaction_color, interaction_style ))
	# END OF if format_mode == "dot":

	elif format_mode == "sif":

	    # write interaction to output_target (ie the file that will hold the network in sif format)
	    output_target.write("%s\tpp\t%s\n" %(extcode_start, extcode_end))

	    # now, write to the files that hold the attributes for this edge
	    
	    # edges in cytoscape are represented by start_node(pp)end_node
	    edge_sif_id = "%s (pp) %s" %(extcode_start, extcode_end)

	    file_source_dbs_fd = file(network_attributes_prefix + "source_dbs.eda","a")
	    for one_db in list_interaction_dbs:
		file_source_dbs_fd.write("%s = %s\n" %(edge_sif_id, one_db))
	    file_source_dbs_fd.close()

	    file_source_methods_fd = file(network_attributes_prefix + "source_methods.eda","a")
	    for one_db in list_interaction_methods:
		file_source_methods_fd.write("%s = %s\n" %(edge_sif_id, one_db))
	    file_source_methods_fd.close()
	# END OF elif format_mode == "sif":

    print_image_interaction = staticmethod(print_image_interaction)

    def _check_special_node(self,  node_object, keywords):
        """
        returns 1 if node is a root protein, a expanded node or contains keywords in its description, function or name
        """

        if node_object.is_root() or node_object.is_expanded()[0] or \
           self.piana_access.check_keywords_in_protein(list_proteinPiana=[node_object.get_node_id()], keywords=keywords):
            return 1
        else:
            return 0

    def _print_node_image(self, output_target, node_object, node_ext_code, keywords= [], expression = None, format_mode = None,
			  network_attributes_prefix = None, special_labels_string="" ):
        """
        prints a node in dot format

        
        "output_target" is a file object (set to sys.stdout if you want the output printed to your screen)

        "format_mode" sets the format used for printing the network interaction (right now, 'dot' or 'sif')
        
        "node_object" is the node object (obviously)

        "node_ext_code" is the external code that you want to use for this node

        "keywords" is the list of keywords that will be used to highlight the node in the network (see argument list_keywords below in output_interactions)

        "expression" indicates if the protein is over of infra expressed. Valid values are:
           - 'over_expressed'
           - 'infra_expressed'
           - None --> not info available 
    
        "special_labels_string" is a string that combines into one chain all the special labels

        "network_attributes_prefix" is the directory + file prefix that will be used to create the files that describe the information for this node
            -> these files can then be used by cytoscape to visualize specific properties of the node

        """

        keywords_appearing = node_object.get_node_attribute_object().get_keywords_appearing()
        
	if format_mode == "sif":
	    # append to node attribute files info associated to this node

	    if expression:
		file_with_expression_fd = file(network_attributes_prefix + "expression.noa","a")
		file_with_expression_fd.write("%s = %s\n" %(node_ext_code, expression) )
		file_with_expression_fd.close()

	    if special_labels_string:
		file_with_special_labels_fd = file(network_attributes_prefix + "special_labels.noa","a")
		file_with_special_labels_fd.write("%s = %s\n" %(node_ext_code, special_labels_string) )
		file_with_special_labels_fd.close()

	    if node_object.is_root():
		file_with_roots_fd = file(network_attributes_prefix + "roots.noa","a")
		file_with_roots_fd.write("%s = root_protein\n" %(node_ext_code) )
		file_with_roots_fd.close()
	    
	    if keywords_appearing:
		file_with_keywords_fd = file(network_attributes_prefix + "have_keyword.noa","a")
		file_with_keywords_fd.write("%s = yes\n" %(node_ext_code) )
		file_with_keywords_fd.close()

		file_listing_keywords_fd = file(network_attributes_prefix + "list_keywords.noa","a")
		for one_keyword in keywords_appearing:
		    file_listing_keywords_fd.write("%s = %s\n" %(node_ext_code, one_keyword) )
		file_listing_keywords_fd.close()

            ## Added by Javi to print all root nodes in SIF format
            if node_object.is_root():
                output_target.write("%s\n" %( node_ext_code ))

                
	# END OF if format_mode == "sif":
	    
	elif format_mode == "dot":
	    # set the color of the node border depending on its expression information
	    if expression:
		bordercolor = self.piana_access.get_node_border_color(expression)
		if verbose_expression:
		    sys.stderr.write("protein %s has expression %s\n" %(node_ext_code, expression))
	    else:
		bordercolor =  self.piana_access.get_node_border_color("normal")

	    if node_object.is_root():
		if not keywords_appearing:
		    # node is root and keywords do not appear in it (expression info encoded in bordercolor)
		    output_target.write('"%s" [fillcolor = %s, color= %s]\n' %(node_ext_code, self.piana_access.get_node_fill_color("root"),
									       bordercolor))
		else:
		    # node is root and keywords appear in it (expression info encoded in bordercolor)
		    output_target.write('"%s" [fillcolor = %s, color= %s]\n' %(node_ext_code, self.piana_access.get_node_fill_color("root_keyword"),
										   bordercolor ))

	    elif keywords_appearing:
		# node is not root and keywords appear in it (expression info encoded in bordercolor)
		output_target.write('"%s" [fillcolor = %s, color= %s]\n' %(node_ext_code, self.piana_access.get_node_fill_color("keyword"),
										   bordercolor ))

	    elif expression:
		# node is not root, keywords do not appear in it, there is expression information associated to it
		output_target.write('"%s" [fillcolor = %s, color= %s]\n' %(node_ext_code, self.piana_access.get_node_fill_color("normal"),
									       bordercolor ))


	    elif node_object.is_expanded()[0] is not None:
		# node is not root, keywords do not appear in it, no expression information, node comes from an expansion
		output_target.write('"%s" [color = %s]\n' %(node_ext_code, self.piana_access.get_node_border_color("expanded")))
	# END OF if format_mode == "dot"
	else:
	    raise ValueError("Incorrect format_mode set for printing out node image information")
        
    def output_interactions(self,  id_type= None,
                            output_target= None, output_format= None, format_mode= "txt", intersection_dbs=None,
                            filter_mode="all", print_mode= "all", alternative_id_types=[],
                            tax_id_value= 0, use_self_ints="yes", list_keywords= [], file_over_expressed= None,
                            file_infra_expressed= None, expression_id_type= None, dic_special_proteins_file=None, user_protein_names = {}, 
			    network_attributes_prefix=None):
        """

        writes a (table or network in dot format) in file "output_target" describing graph information in the following formats:

        dot format: follows the standard dot language for describing networks
        table format:follows the format described in method print_table_interaction

        "id_type" is the easy-to-remember type name that will be used for printing the proteins in the network.
           --> Valid id_type are those listed in the PIANA reference card (read README.piana_tutorial)
        
        "output_target" is a file object (set to sys.stdout if you want the output printed to your screen)

        "output_format" is the format that will be followed for the output
           -"table": prints interactions in table format
           -"network": prints interactions in a network
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text (only valid for output_format 'table')
               - 'html' will print html  (only valid for output_format 'table')
               - 'dot' will print a DOT file (only valid for output_format 'network')

        "intersection_dbs" sets intersection mode, which only prints out interactions that appear in all dbs of the list being passed
            -> it can be None (no intersection mode applied) or a list of database names
            -> valid database names are those in PianaGlobals.interaction_databases

        "filter_mode" not being used at the minute... will be used for hidden, unhidden, all

        "print_mode" sets which proteins will be printed
           -> "all" will print all interactions in the network
           -> "all_root" will print all interactions in the network where at least one partner is a root protein
           -> "only_root" will print only interactions between root proteins
           -> "connecting" will print only interactions between root proteins and those proteins that connect more than one root protein

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide pairs a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "tax_id_value" sets the species of the interactions that are being added (can be used for eliminating ambiguities between codes across species)
           --> valid species names are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy  

        "use_self_ints" determines if self interactions of proteins will be printed out
            --> valid values are 'yes' (i.e. print all interactions) and 'no' (i.e. do not print self interactions)
            (a self interaction is an interaction of a protein with itsel, as seen in homodimers)

        "list_keywords" sets a list of keywords that will be used to highlight important proteins in the network
           -> currently, it highlights in red the proteins in the DOT file that contain at least one keyword in the function, description or name
           -> currently, it is not used when printing the interactions in a table
           -> If you are interested in highlighting proteins related to cancer, list_keywords could be: ['cancer', 'onco', 'carcinoma', 'tumor']
           
        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}


	"network_attributes_prefix" is used to create files with information about the nodes and edges
            -> these files can then be used from cytoscape to visualize certain characteristics of the network, nodes and edges
            """

        initial_method_time = time.time()
        # make sure self.dic_protein_naming is filled
        #print "Checking for _fill_dic... filled"

        self._fill_dic_protein_naming(id_type=id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= [], key_list=None, user_protein_names = user_protein_names)
        #print "Checking for _fill_dic... filled"
        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value= tax_id_value)

        #print "Checking for _fill_dic... filled"
 	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                        id_type=None, alternative_id_types=None)

        #print "Creating unified network..."
        self._create_unified_network(id_type=id_type, alternative_id_types=alternative_id_types,
                                     user_keywords=list_keywords )

        if print_mode=="connecting":
            # javi: Now, this dictionary is filled in the method _create_ext_code_connecting
            #connecting_prots = self.get_connecting_nodes_dic() # returns a dictionary with proteinPiana:[root_nodes] where root_nodes are proteinPianas

            # create the connecting dictionary using ext codes instead of proteinPianas
            self._create_ext_code_connecting(id_type= id_type,
                                             alternative_id_types= alternative_id_types)

        initial_time = time.time()

        if output_format == "network":

            if format_mode== "dot":
                # -----------------
                # PRINT DOT HEADERS
                # -----------------
                output_target.write("graph G { graph [orientation=%s, pack=%s, overlap=%s]" %(PianaGlobals.dot_orientation,
                                                                                              PianaGlobals.dot_pack,
                                                                                              PianaGlobals.dot_overlap) )
            
                output_target.write(" node [shape=%s, fontsize=%s, width=%s, height=%s,style=%s, fillcolor=%s];\n" %(PianaGlobals.dot_node_shape,
                                                                                                                     PianaGlobals.dot_node_font_size,
                                                                                                                     PianaGlobals.dot_node_width,
                                                                                                                     PianaGlobals.dot_node_height,
                                                                                                                     PianaGlobals.dot_node_style,
                                                                                                                     PianaGlobals.dot_standard_node_fill_color))
            # END OF if format_mode== "dot":
	    elif format_mode == "sif":
		# ----------------
		# PRINT HEADERS FOR NODE CYTOSCAPE ATTRIBUTES FILES
		# ----------------
		
		file_with_expression_fd = file(network_attributes_prefix + "expression.noa","w")
		file_with_expression_fd.write("expresion\n")
		file_with_expression_fd.close()

		file_with_special_labels_fd = file(network_attributes_prefix + "special_labels.noa","w")
		file_with_special_labels_fd.write("special_labels\n")
		file_with_special_labels_fd.close()

		file_with_roots_fd = file(network_attributes_prefix + "roots.noa","w")
		file_with_roots_fd.write("roots\n")
		file_with_roots_fd.close()

		file_with_keywords_fd = file(network_attributes_prefix + "have_keyword.noa","w")
		file_with_keywords_fd.write("have_keyword\n")
		file_with_keywords_fd.close()

		file_listing_keywords_fd = file(network_attributes_prefix + "list_keywords.noa","w")
		file_listing_keywords_fd.write("list_keywords\n")
		file_listing_keywords_fd.close()
            # END OF elif format_mode == "sif":
		

            # -------------------------------------
            # PRINT SPECIAL PROTEINS IN IMAGE FORMAT
            # -------------------------------------
            # for dot output format, set the root nodes to be of a different colour
            #                        set the nodes with keywords to be of a different colour
            #                        other nodes will have the standard color
            #
            

            #already_printed_nodes = {} # used in output_protein_information to avoid printing same unified node many times in the dot file
            already_printed_nodes = Set()
                                       

            # setting nodes output parameters depending on their attributes 
            for this_node in self.get_node_object_list():

                proteinPiana = this_node.get_node_id()

                # get the ext_code that is going to be used for this proteinPiana
                node_ext_code = self.get_unified_name( proteinPiana= proteinPiana,
                                                       id_type=id_type, alternative_id_types=alternative_id_types)
            
                #if already_printed_nodes.has_key(node_ext_code):
                if node_ext_code in already_printed_nodes:
                    # if node was printed, no need to continue: the unified node has printed all the info associated to proteinPianas associated to this ext_code
                    continue

                #already_printed_nodes[node_ext_code] = None # set as already printed
                already_printed_nodes.add(node_ext_code)

                # get the unified node associated to this ext_code
                node = self.ext_code_unified_node_correspondence[node_ext_code]

                # check if the proteins of the interaction respect the species restriction: if not, skip protein
                dic_protein_taxonomies = node.get_node_attribute_object().get_dic_taxonomies()
                passes_species_test = self._passes_species_test(required_protein_tax_id= tax_id_value,
                                                                dic_protein_taxonomies= dic_protein_taxonomies)

                expression = node.get_node_attribute_object().get_expression()

                special_labels_string = self._special_labels_list_2_string(special_labels_list= node.get_node_attribute_object().get_special_labels(),
                                                                           tag="", separator= " ")
                
                if not passes_species_test:
                    continue

                # check if it passes print_mode
                #          TO DO!!! in print_mode "all_root" I am printing all nodes, even if they don't have any interaction afterwards
                #                   this is wrong! I should only print the node if it has an interaction to a root node
                if print_mode == "only_root" and not node.is_root():
                    continue
                #elif print_mode == "all_root" and interaction doesn't have at least one root node:
                #    continue # The problem is that at this point I do not know whether this node will be connected to a root
                #               or not... If I want this to work, I have to either print the nodes when printing the edges
                #               or keep a list of nodes that are connected to root nodes and then check if this
                #               node is in that list
                elif print_mode == "connecting" and not ( node.is_root() or self.ext_code_connecting.has_key(node_ext_code) ):
                    continue

                self._print_node_image(output_target=output_target, format_mode=format_mode, node_object= node, node_ext_code=node_ext_code,
				       keywords= list_keywords, expression=expression, special_labels_string=special_labels_string, network_attributes_prefix=network_attributes_prefix )

            # END OF  for node in self.nodes:
            
        # END OF if output_format == "network":
           
        # print html table header
        if output_format == "table":


            if format_mode == "html":
                output_target.write("<html><body><br><br>\n")
                output_target.write("Keywords used to highlight proteins: %s<br><br>\n" %list_keywords)
		table_header = """<table border=1 style="border-collapse: collapse" bordercolor="#808080">\n"""
		table_header += "<tr><td><center><b>protein A</b></td><td><center><b>protein A <br>External Links</b></td><td><center><b>protein A <br>information</b></center></td><td><center><b>protein B</b></center></td><td><center><b>protein B <br>External Links</b></td><td><center><b>protein B <br>information</b></center></td><td><center><b>same cell <br>location</b></center></td><td><center><b>same <br>species</b></center></td><td><center><b>source <br>databases</b></center></td><td><center><b>detection <br>methods</b></center></td></tr>\n"
                output_target.write(table_header)
		self.interaction_table_strings['header'] = table_header

	elif output_format == "network":
		
	    if format_mode == "sif":
		# if using SIF format, create files with headers that will describe interaction attributes for cytoscape
		
		file_source_dbs_fd = file(network_attributes_prefix + "source_dbs.eda","w")
		file_source_dbs_fd.write("source_dbs\n")
		file_source_dbs_fd.close()

		file_source_methods_fd = file(network_attributes_prefix + "source_methods.eda","w")
		file_source_methods_fd.write("source_methods\n")
		file_source_methods_fd.close()
	    # END OF if format_mode == "sif":
	# END OF elif output_format == "network":


            
        root_connectivity = 0
        #already_printed_interactions = {} # keeps track of interactions already printed (in order to skip them)
        already_printed_interactions = Set()

        #javi
        #for edge in self.get_edge_object_list():
        printed=0
        for current_edge_id in self.get_edge_ids_list():
            edge = self.get_edge( identifier1 = current_edge_id, get_mode="error")

            printed += 1

            # ------------------
            # PRINT INTERACTIONS
            # ------------------
            # -----
            # Getting external codes for edge proteins and the edge_key
            # -----

            (edge_key, protein_extcode_start, protein_extcode_end) = self._get_edge_key(this_edge=edge,
                                                                                        id_type= id_type,
                                                                                        alternative_id_types= alternative_id_types )

            # skip interactions that have already been printed
            #if already_printed_interactions.has_key(edge_key):
            if edge_key in already_printed_interactions:
                #print "%s already printed!\n" %edge_key
                continue

            #print "%s not printed yet!\n" %edge_key
            
            # skip self interactions when asked by user
            if use_self_ints == "no":
                if protein_extcode_start == protein_extcode_end:
                    #print "self ints filtered!"
                    continue
                
            #already_printed_interactions[edge_key] = None
            already_printed_interactions.add(edge_key)

            # Important! setting the edge for this key to be the unified edge: this guarantees that all
            #            edges between these ext_codes (even if they come from different proteinPianas)
            #            have the same characteristics (source dbs, method, ...) and therefore
            #            they are not lost from the fact that we are using this type of code for the output
            #javi
            #edge = self.edge_key_unified_edge_correspondence[edge_key]
            edge = self.get_unified_edge(edge_key=edge_key)

            # Important! setting the nodes for the edge nodes to be their unified nodes: this guarantees that the nodes
            #            have all the characteristics of the proteinPianas that have the same ext_codes
            protein_node_start = self.ext_code_unified_node_correspondence[protein_extcode_start]
            protein_node_end = self.ext_code_unified_node_correspondence[protein_extcode_end]

            dic_protein_taxonomies_start = protein_node_start.get_node_attribute_object().get_dic_taxonomies()
            dic_protein_taxonomies_end = protein_node_end.get_node_attribute_object().get_dic_taxonomies()

            #if (protein_node_start=="RAF1" and protein_node_end=="MYC") or (protein_node_start=="MYC" and protein_node_end=="RAF1"):
            #print "protein_node_start: %s protein_node_end: %s\n" %(protein_node_start,protein_node_end)
            #if (self.proteinPiana_naming[protein_node_start.get_node_attribute_object().get_node_id()]=="RAF1" and self.proteinPiana_naming[protein_node_end.get_node_attribute_object().get_node_id()]=="MYC") or (self.proteinPiana_naming[protein_node_start.get_node_attribute_object().get_node_id()]=="MYC" and self.proteinPiana_naming[protein_node_end.get_node_attribute_object().get_node_id()]=="RAF1"):
            #    print "START: %s taxonomies: %s" %(protein_node_start, dic_protein_taxonomies_start)
            #    print "END: %s taxonomies: %s" %(protein_node_end, dic_protein_taxonomies_end)
            #    print "associated_nodes_start: %s" %(protein_node_start.get_node_attribute_object().get_associated_nodes())
            #    print "associated_nodes_end: %s" %(protein_node_end.get_node_attribute_object().get_associated_nodes())
            #    sys.exit()
            
            # -----
            # Checking if interaction respects species restrictions
            # -----
            # check if the proteins of the interaction respect the species restriction: if not, skip interaction
            if not self._passes_species_test(required_protein_tax_id= tax_id_value,
                                             dic_protein_taxonomies= dic_protein_taxonomies_start):
                #print "specie filtered!"
                continue

            if not self._passes_species_test(required_protein_tax_id= tax_id_value,
                                             dic_protein_taxonomies= dic_protein_taxonomies_end):
                #print "specie filtered!"
                continue

            # -----
            # Checking if interaction respects print_mode restrictions
            # -----
            if print_mode != "all":
                if print_mode == "connecting":
                    
                    # returns the connecting prots using ext_code id_type
                    
                    if not self._passes_connecting_mode_test(id_a= protein_extcode_start,
                                                             is_root_a = protein_node_start.is_root(),
                                                             id_b= protein_extcode_end,
                                                             is_root_b= protein_node_end.is_root(),
                                                             connecting_prots_dic= self.ext_code_connecting ):
                        #print "connecting_mode filtered!"
                        continue
                    
                else:
                    # print mode is not "connecting", check for root restrictions
                    if not self._passes_root_mode_test( is_root_a= protein_node_start.is_root(),
                                                        is_root_b= protein_node_end.is_root(),
                                                        print_mode= print_mode ):
                        #print "root_mode filtered!"
                        continue
                # END OF else: (if print_mode == "connecting":)
            # END OF if print_mode != "all":

            # -----
            # Checking database intersection: used when intersection_dbs is not None
            # -----
            # --> if interaction doesn't appear in all dbs, continue with next loop iteration
            
            edge_attribute = edge.get_edge_attribute_object()

            #print edge_attribute

            # getting detection methods for this interaction
            list_interaction_methods = edge_attribute.get_list_methodID()

            # getting pubmed IDs for this interaction (javi)
            list_interaction_pubmeds = edge_attribute.get_list_pubmedID()
            

            # get interaction dbs for this edge (removing duplicates)
            list_table_interaction_dbs = dict(map(None, edge_attribute.get_list_sourceDB_names(), [])).keys()
            
            if intersection_dbs is not None:
                # intersection mode activated... print only interactions that appear in all dbs in intersection_dbs

                if Set(intersection_dbs).difference(Set(list_table_interaction_dbs)):
                    # if those dbs required by user are not completely covered (i.e. difference is a Set with elements) by the dbs of this interaction, skip it
                    #print "intersection_dbs filtered!"
                    continue
                
                
## commented because code improved using Sets
##                 appears_in_all_dbs = 1
##                 for intersection_db in intersection_dbs:
##                     if intersection_db not in list_table_interaction_dbs:
##                         appears_in_all_dbs = 0
##                         break
##                 # END OF for intersection_db in intersection_dbs:

##                 if appears_in_all_dbs == 0:
##                     continue
                
            # END OF if intersection_dbs is not None:

            
            # -----
            # Checking if both proteins are in the same subcellular location
            # -----
            if Set(protein_node_start.get_node_attribute_object().get_cell_location()).intersection(Set(protein_node_end.get_node_attribute_object().get_cell_location())):
                same_subcellularLocation =  PianaGlobals.positive_location_value[format_mode]
            else:
                same_subcellularLocation = PianaGlobals.negative_location_value[format_mode]

            # -----
            # Checking if both proteins are in the same species
            # -----
            if Set(dic_protein_taxonomies_start).intersection(Set(dic_protein_taxonomies_end)):
                same_species = PianaGlobals.positive_species_value[format_mode]
            else:
                same_species = PianaGlobals.negative_species_value[format_mode]

##             # -----
##             # Calculating root connectivity
##             # -----
##             # for the protein of the pair that is not the root protein (if none of them is a root protein, connectivity=-1)
##             # find how many other root proteins it is connecting. It is an indication on how important is the role of
##             # this non-root protein in the network
 
##             if protein_node_start.is_root() and protein_node_end.is_root():
##                 root_connectivity = max( self.get_node_root_connectivity(node=protein_node_start), self.get_node_root_connectivity(node=protein_node_end))

##             elif not protein_node_start.is_root() and not protein_node_end.is_root():   root_connectivity = -1

##             elif protein_node_start.is_root():
##                 if self.ext_code_connecting.has_key(protein_extcode_end):  root_connectivity = len(self.ext_code_connecting[protein_extcode_end])
##                 else:                                                      root_connectivity = 0

##             elif protein_node_end.is_root():
##                 if self.ext_code_connecting.has_key(protein_extcode_start): root_connectivity = len(self.ext_code_connecting[protein_extcode_start])
##                 else:                                                       root_connectivity = 0

##             else:
##                 raise ValueError("Root connectivity cannot be calculated: case not taken into account")

            root_connectivity = 0 # Not calculating root_connectivity... it is not a useful value, and it takes time to calculate it

            if protein_node_start.is_root():                       node_type_start = PianaGlobals.positive_root_value[format_mode]
            else:                                                  node_type_start = PianaGlobals.negative_root_value[format_mode]

            if protein_node_end.is_root():                         node_type_end = PianaGlobals.positive_root_value[format_mode]
            else:                                                  node_type_end = PianaGlobals.negative_root_value[format_mode]
   
            list_protein_user_keywords_start = protein_node_start.get_node_attribute_object().get_keywords_appearing()
            list_protein_user_keywords_end = protein_node_end.get_node_attribute_object().get_keywords_appearing()

            # get number of neighbours for the unified start and end nodes
            num_neigh_start = self._get_num_unified_networks(protein_extcode= protein_extcode_start, protein_node=protein_node_start,
                                                             id_type=id_type, alternative_id_types=alternative_id_types)
            num_neigh_end = self._get_num_unified_networks(protein_extcode= protein_extcode_end, protein_node=protein_node_end,
                                                             id_type=id_type, alternative_id_types=alternative_id_types)

            #print "really printing to %s..." %output_target
            # -----
            # Printing interaction with info collected 
            # -----
            if output_format == "table":
                # TO DO!!! Number of neighbours information is not correct: it is not merging neighbours in the unified node...
		#       (I might have correct this... I don't remember right now whether I did it or not... CHECK!)
                self.print_table_interaction(output_target=output_target,
					     format_mode = format_mode,
					     root_connectivity= root_connectivity,
					     extcode_start= protein_extcode_start,
					     num_neigh_start= num_neigh_start,
					     node_type_start= node_type_start,
					     expression_start= protein_node_start.get_node_attribute_object().get_expression(),
					     special_labels_string_start=self._special_labels_list_2_string(special_labels_list= protein_node_start.get_node_attribute_object().get_special_labels(), tag="", separator=" "),
					     list_protein_user_keywords_start= list_protein_user_keywords_start ,
					     extcode_end= protein_extcode_end,
					     num_neigh_end= num_neigh_end,
					     node_type_end= node_type_end,
					     expression_end= protein_node_end.get_node_attribute_object().get_expression(),
					     special_labels_string_end=self._special_labels_list_2_string(special_labels_list= protein_node_end.get_node_attribute_object().get_special_labels(), tag="", separator=" "),
					     list_protein_user_keywords_end= list_protein_user_keywords_end,
					     same_loc= same_subcellularLocation,
					     same_species= same_species,
					     list_interaction_dbs= list_table_interaction_dbs,
					     list_interaction_methods= list_interaction_methods,
                                             list_interaction_pubmeds= list_interaction_pubmeds,
                                             fill_interaction_table_strings=0)
                # TO DO!!! fill_interaction_table_strings is fixed to 0. However, in some cases, we want to fill it (e.g. match-pathways). This has to be changed
                #          so that print_table_interactions keeps strings or not depending on the user request

            elif output_format == "network":
		
                PianaGraph.print_image_interaction(output_target= output_target,
						   format_mode = format_mode,
						   edge_object= edge,
						   piana_access = self.piana_access,
						   extcode_start= protein_extcode_start,
						   extcode_end= protein_extcode_end,
						   list_interaction_dbs= list_table_interaction_dbs,
                                                   list_interaction_methods= list_interaction_methods,
						   network_attributes_prefix = network_attributes_prefix)
                
            else:
                sys.stderr.write("unknown output format (%s) set" %(output_format))
        # END OF for edge in self.edges:
            

        # PRINT HTML TABLE CLOSURE
        if output_format == "table":
            if format_mode == "html":
		table_footer = "\n</table>"
                output_target.write(table_footer)
                output_target.write("</body></html>\n\n")
		self.interaction_table_strings['footer'] = table_footer
                
        if output_format == "network":
            if format_mode == "dot":
                # closing statement for  dot file
                output_target.write("}")

        #print "Total time for printing interactions: %s seconds" %(time.time()-initial_time)
        #print "Total for printing method: %s seconds" %(time.time()-initial_method_time)


    def match_pathways(self, pathways_dir, force_expression, output_file_object, results_dir, files_prefix, format_mode, 
		       pathway_id_type, tax_id_value, output_id_type, alternative_id_types, 
		       user_protein_names = {}, file_over_expressed= None, file_infra_expressed= None, expression_id_type= None,
		       dic_special_proteins_file=None):
	"""
	Prints out for each pathway in pathways_dir, the number of proteins in the pathway (and percentage) that appear in the current piana_graph

	   --> this can be used to know which pathways are being 'affected' by the current network. For example, if you build a network for 
               proteins known to be involved in cancer, then you can check which pathways might be affected by the proteins in the cancer network

        "pathways_dir" is the directory that contains pathway files
                 -> a pathway file is a file that contains proteins that are known to be involved in a given pathway
		 -> the pathway filename must be indicative of the pathway: that name will be used for the output
		 -> all proteins in all pathway files must be in the same type of protein identifier.
		    If you have pathways that use different identifiers, you must call this method as many times as id types you have
	
	"force_expression" can be 'yes' or 'no' and is used to know which proteins of the network will be used to match pathways
             'no': all proteins in the current network are used to match pathways
	     'yes': only those proteins that are over/under expressed are used to match pathways

        "output_file_object" is the file where the results will be printed

        "output_id_type" is the type of code that will be used to print proteins
             -> valid protein-type values are those defined in the PIANA reference card (read README.piana_tutorial)
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "pathway_id_type" is the easy-to-remember type name that has been used in the pathways files
            -> Valid pathway_type are those listed in the PIANA reference card (read README.piana_tutorial)

        "format_mode" sets the type of format that will be used for output
              - 'txt' will print flat text
              - 'html' will print html 

        "tax_id_value" is the tax_id of the proteins we are working on (proteins in pathway files must be of this tax_id)

               - set to 0 if you don't mind of the species, or to the tax id in case you do

        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}
	"""

	"""
	STEPS FOLLOWED IN THIS METHOD:


	1. find the proteinPianas that appear in each pathway file

	2. decide which unified names will be used along this method (for all proteins in the network and all proteins in the patway files)
            --> user names are the root proteins given by user and the names used in the pathway files

        3. retrieve all information for proteins (ext_code, whethere they are linkers or not, connected to which roots, special labels, ...)

	4. calculated expression statistics for different types of proteins (roots, linkers, partners, other)

	5. for each relevant protein, write its interactions to a file (so the user can click to see the interactions for a protein)


	6. create a file that contains all relevant proteins (so the user can click to see relevant proteins)

	7. create a file with results on overlaps between the network relevant proteins and proteins from the pathways (ie. pathway matches)
	   -> in this step we also create other files that are linked from the HTML results so that the user can access the information
	
	"""

	# -----------
	# 1.
	# -----------
	# get the pathway proteinPianas from the files provided by the user
	# -----------

	dic_pathways = {}   # dictionary that contains a set of proteinPianas for each pathway in pathways_dir
	                    #   -> the pathway name is created from the pathway filename
	                    #      it follows the structure:
	                    #               
	                    #               { pathway_name1: Set([prot1, prot2, ...]),
	                    #                 pathway_name2: Set([...]),
	                    #                 ..................
	                    #               }

	pathways_files = os.listdir(pathways_dir)

	for one_pathway_file in pathways_files:
	    # for each pathway, extract proteins and populate the dictionary
	    
	    if not os.path.isfile(pathways_dir + "/" + one_pathway_file):
		continue

	    pathway_key = os.path.split(one_pathway_file)[1].split(".")[0] # keep only the file name without path of extension

	    proteins_set = Set([])
	    
	    for one_line in file(pathways_dir + "/" + one_pathway_file , "r"):

		user_protein_names[one_line.strip()] = None # update names that user wants to fix (even if the pathways use a different type of identifier
		                                            # we place the names here... we don't really care, it is just to make sure that if 
		                                            # PIANA is deciding between two different names for one protein, the one used by the user will have priority

		proteins_set.update(self.piana_access.get_list_proteinPiana(protein_id= one_line.strip(),
                                                                            id_type= pathway_id_type,
                                                                            tax_id_value= tax_id_value,
                                                                            source_db_info= "no"))
	    # END OF for one_line in file(pathways_dir + "/" + one_pathway_file , "r"):

	    dic_pathways[pathway_key] = proteins_set
	    
	# END OF for one_pathway_file in os.listdir(pathways_dir):

	# ----
	# MISCELANEOUS INITIALIZATIONS
	# ----
	# fill the expression dics so that the information is available afterwards
        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value= tax_id_value)

	# fill the special proteins dics so that the information is available afterwards
	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                       id_type=None, alternative_id_types=None)



	# -----------
	# 2.
	# -----------
	# Decide which unified names will be used along this method
	# A name has to be decided for all proteins in the network and all proteins in the pathway files
	# On the other hand, we have to make sure of three things:
	#    -> we respect the names given by the user as root proteins (ensured by argument user_protein_names provided by PianaApi)
	#    -> we respect the names given by the user as pathway proteins (ensured by assing pathway protein names to user_protein_names (see 1.) )
	#    -> we respect the names decided by command print-table (otherwise, we might not find interactions associated to a protein)
	#       (ensured  by adding to user_protein_names all unique names in the previous dictionary self.dic_protein_naming)
	# ----------------

	# 2.1. create the list of all proteinPianas for which a name has to be decided (network proteinPianas + pathway proteinPianas)
	dic_proteinPianas = {}

	# -> first of all, place network proteinPianas
	for one_network_proteinPiana in self.get_node_ids_list():
	    dic_proteinPianas[one_network_proteinPiana] = None

	# -> now, place all pathways proteinPianas
	for pathway_key in dic_pathways:
	    for one_proteinPiana in dic_pathways[pathway_key]:
		dic_proteinPianas[one_proteinPiana] = None

	# -> now, add all unique names from the previous dictionary to user_protein_names
	for one_external_code in  self.dic_protein_naming:
	    user_protein_names[self.dic_protein_naming[one_external_code]] = None 

	# 2.2. fill self.dic_protein_naming with unique names for each proteinPiana that is in the network or in the pathways
	#      -> for the list of proteinPianas that appear in network or pathways
	#      -> respecting user root protein names, user pathway protein names and previously decided names
        self._fill_dic_protein_naming(id_type=output_id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= dic_proteinPianas.keys(), key_list="x", user_protein_names = user_protein_names )


	# -----------
	# 3.
	# -----------
	# Retrieve information for proteins
	# ----------------


	# ###
	# 3.1 get the linker proteins (using ext_codes just decided in the  call to _fill_dic_protein_naming)
	# ###

        # javi: Now, this dictionary is filled in the method _create_ext_code_connecting
        #connecting_prots = self.get_connecting_nodes_dic() # returns a dictionary with proteinPiana:[root_nodes] where root_nodes are proteinPianas

        # create the connecting dictionary using ext codes instead of proteinPianas
        self._create_ext_code_connecting(id_type= output_id_type,
                                         alternative_id_types= alternative_id_types)

	
		
	


	# ###
	# 3.2 find the relevant proteins
	# ###
	# 3.2.1 find network proteins (and in case user forced expression, select those over/under expressed)
	# 3.2.2 decide which external identifier will be used for each protein

	# 3.2.1
	if force_expression == "no":
	    # if expression not to be taken into account, get all proteins in the network
	    relevant_proteinPianas = Set(self.get_node_ids_list())

	elif force_expression == "yes":
	    # if expression required, use only those proteins that are over/under expressed
	    relevant_proteinPianas = Set([])

	    for one_proteinPiana in self.get_node_ids_list():
		if self.over_expressed_proteinPianas.has_key(one_proteinPiana) or self.infra_expressed_proteinPianas.has_key(one_proteinPiana):
		    relevant_proteinPianas.add(one_proteinPiana)
	    # END OF for one_proteinPiana in self.get_node_object_list():
	# END OF elif force_expression == "yes":
       

	# ###
	# 3.3 Several things are done simultaneously here:
	#        a -> all relevant proteinPianas are translated to external codes
	#        b -> dictionary of relevant ext codes that were seed proteins
	#        c -> dictionary for over/under expressed external codes is created
	#        d -> dictionary of special labels for each ext code is created
	#        e -> dictionary of roots connected to each external code is created
	# ###
	
	relevant_ext_codes = Set([])   # is a dic {proteinPiana: external identifier to be used, proteinPiana: exte....}
	dic_root_ext_codes = {}  # used to keep in a dictionary those external codes that are root proteins 
	dic_relevant_ext_codes = {}  # a dictionary for keeping which  proteinPianas are associated to each ext code
                                     # structure is the following:
	                             #     { ext_code: [pp1, pp2, ...],
	                             #       ext_code: [pp5, pp3, pp4, ...],
                                     #       .....
                                     #     }

	dic_over_exp_ext_codes = {}  # is a dic with key the ext code and content None. Only contains ext_codes that were found over expressed
	dic_infra_exp_ext_codes = {} # is a dic with key the ext code and content None. Only contains ext_codes that were found under expressed
	dic_root_linked = {} # is a dic with key the ext code and content a root protein. Only contains ext_codes that are directly connected to a root protein
	dic_special_labels_ext_codes = {} # is a dic with key the ext code and content a list of special labels associated to that protein

	for one_proteinPiana in relevant_proteinPianas:
	    
	    # 3.3.a get unified name for this relevant protein, add it to set of relevant codes and fill dictionary of correspodence relevant_code:[proteinPiana, proteinPiana, ...]

	    this_ext_code = self.get_unified_name( proteinPiana= one_proteinPiana,
						   id_type=output_id_type, alternative_id_types=alternative_id_types)
	    relevant_ext_codes.add(this_ext_code)

	    if dic_relevant_ext_codes.has_key(this_ext_code):
		dic_relevant_ext_codes[this_ext_code].append(one_proteinPiana)
	    else:
		dic_relevant_ext_codes[this_ext_code] = [one_proteinPiana]

	    # 3.3.b. find if this relevant protein was one of the seed proteins (ie. root)
	    if self.get_node(identifier=one_proteinPiana , get_mode="error").is_root():
		dic_root_ext_codes[this_ext_code] = None
		dic_root_linked[this_ext_code]= "is_root"  # if it is a root, it overwrites whatever value was assigned to it previously


	    # 3.3.c. find if this relevant protein is over/under expressed: fill dictionary with over/under expressed ext codes
	    if self.over_expressed_proteinPianas.has_key(one_proteinPiana):
		dic_over_exp_ext_codes[this_ext_code] = None

	    if self.infra_expressed_proteinPianas.has_key(one_proteinPiana):
		dic_infra_exp_ext_codes[this_ext_code] = None


	    # 3.3.d. find if this relevant protein has special labels: fill dictionary dic_special_labels_ext_codes with the excode:[label,...]
	    for one_special_label in self.dic_special_proteinPianas:
		# for each label, check if one_proteinPiana has it
		if self.dic_special_proteinPianas[one_special_label].has_key(one_proteinPiana):

		    if dic_special_labels_ext_codes.has_key(this_ext_code):
                        # if the external code is already associated to a label, this one will be appended to the list of labels

			if one_special_label not in dic_special_labels_ext_codes[this_ext_code]:
			    dic_special_labels_ext_codes[this_ext_code].append(one_special_label)
		    else:
                        # if this is the first label for this external code, create its list of labels
			dic_special_labels_ext_codes[this_ext_code] = [one_special_label]


	    # 3.3.e. find root proteins connected to this relevant protein
	    #      Attention! This information will only be used when the protein is not a linker protein
	    #                 For linker proteins, connected root have been collected in method self._fill_expression_dics()
	    if not dic_root_linked.has_key(this_ext_code):
		# -> only looking for linked roots if no value was previously set 
		#     (no point in finding a new value if it was decided it is connected to a root  (by a previous equivalent proteinPiana) or if it is "is_root")

		for one_neighbour_id in self.get_node(identifier=one_proteinPiana , get_mode="error").get_neighbour_ids():
		    
		    if self.get_node(identifier= one_neighbour_id, get_mode="error").is_root():
			dic_root_linked[this_ext_code] = self.get_unified_name( proteinPiana=one_neighbour_id ,
										id_type=output_id_type, alternative_id_types=alternative_id_types)
			break

	# END OF for one_proteinPiana in relevant_proteinPianas:
	

	num_relevant_proteins= len(relevant_ext_codes)	    

	# ###
	# 4. get statistics on how many proteins (divided in classes: root, linkers, partners, other) have an expression associated to them
	# ###

	#
	# 4.1 --> expression stats for network proteins (divided in classes: root, linkers, partners, other)
	#
	dic_root_expressions = {}     # keys are protein ext codes and values a tuple (is_over_expressed, is_under_expressed)
	dic_linkers_expressions = {}  # idem
 	dic_partners_expressions = {} # idem
 	dic_others_expressions = {} # idem

	for one_proteinPiana in self.get_node_ids_list():

	    is_over_expressed = False
	    is_under_expressed = False

	    if self.over_expressed_proteinPianas.has_key(one_proteinPiana):	is_over_expressed = True

	    if self.infra_expressed_proteinPianas.has_key(one_proteinPiana):	is_under_expressed = True
	
	    this_ext_code = self.get_unified_name( proteinPiana= one_proteinPiana,
						   id_type=output_id_type, alternative_id_types=alternative_id_types)


	    if self.get_node(identifier=one_proteinPiana , get_mode="error").is_root():
		# this is a root: calculate statistics for roots
		if dic_root_expressions.has_key(this_ext_code):
		    # if an entry for this ext code was already set, keep the expressions that are True
		    dic_root_expressions[this_ext_code] = (dic_root_expressions[this_ext_code][0] and is_over_expressed, 
							   dic_root_expressions[this_ext_code][1] and is_under_expressed )
		else:
		    
		    dic_root_expressions[this_ext_code] = (is_over_expressed, is_under_expressed)

	    elif self.ext_code_connecting.has_key(this_ext_code):
		# this is a linker: calculate statistics for linkers (elif guarantees that this is not a root)
		if dic_linkers_expressions.has_key(this_ext_code):
		    # if an entry for this ext code was already set, keep the expressions that are True
		    dic_linkers_expressions[this_ext_code] = (dic_linkers_expressions[this_ext_code][0] and is_over_expressed, 
							      dic_linkers_expressions[this_ext_code][1] and is_under_expressed )
		else:
		    dic_linkers_expressions[this_ext_code] = (is_over_expressed, is_under_expressed)

#	    elif dic_root_linked.has_key(this_ext_code): # commented because see below (1)
	    else:
		# this is a root partner: calculate statistics for partners (elif guarantees that this is not a root or a linker)
		if dic_partners_expressions.has_key(this_ext_code):
		    # if an entry for this ext code was already set, keep the expressions that are True
		    dic_partners_expressions[this_ext_code] = (dic_partners_expressions[this_ext_code][0] and is_over_expressed, 
							       dic_partners_expressions[this_ext_code][1] and is_under_expressed )
		else:
		    dic_partners_expressions[this_ext_code] = (is_over_expressed, is_under_expressed)

# (1) commented because everything else is not 'others'... dic_root_linked only has ext codes that are relevant proteins....
# if I want others to yield proper results, I need to find all proteins in the network that are connected to a root
#	    else:
#		# everything else is a network protein that is not a root and is not connected to a root protein (depth >1)
#		if dic_others_expressions.has_key(this_ext_code):
#		    # if an entry for this ext code was already set, keep the expressions that are True
#		    dic_others_expressions[this_ext_code] = (dic_others_expressions[this_ext_code][0] and is_over_expressed, 
#							     dic_others_expressions[this_ext_code][1] and is_under_expressed )
#		else:
#		    dic_others_expressions[this_ext_code] = (is_over_expressed, is_under_expressed)
	# END OF for one_proteinPiana in self.get_node_ids_list():

	
	stats_file_name= results_dir + files_prefix  + ".expression_statistics.txt"
	stats_file_fd = file(stats_file_name, "w")
	
	stats_file_fd.write("case=%s" %(files_prefix))

	(number_of_roots, number_of_expressed_roots, number_of_over_expressed_roots, number_of_under_expressed_roots) = self._count_expression_dic(expression_dic= dic_root_expressions)
	stats_file_fd.write("\troots=%s,%s,%s,%s" %(number_of_roots, number_of_expressed_roots, number_of_over_expressed_roots, number_of_under_expressed_roots ))

	(number_of_linkers, number_of_expressed_linkers, number_of_over_expressed_linkers, number_of_under_expressed_linkers) = self._count_expression_dic(expression_dic=dic_linkers_expressions )
	stats_file_fd.write("\tlinkers=%s,%s,%s,%s" %(number_of_linkers, number_of_expressed_linkers, number_of_over_expressed_linkers, number_of_under_expressed_linkers ))

	(number_of_partners, number_of_expressed_partners, number_of_over_expressed_partners, number_of_under_expressed_partners) = self._count_expression_dic(expression_dic=dic_partners_expressions )
	stats_file_fd.write("\tpartners=%s,%s,%s,%s" %(number_of_partners, number_of_expressed_partners, number_of_over_expressed_partners, number_of_under_expressed_partners ))

	(number_of_others, number_of_expressed_others, number_of_over_expressed_others, number_of_under_expressed_others) = self._count_expression_dic(expression_dic=dic_others_expressions )
	stats_file_fd.write("\tothers=%s,%s,%s,%s\n" %(number_of_others, number_of_expressed_others, number_of_over_expressed_others, number_of_under_expressed_others ))

        # TO DO!!! Add info on random groups of size len(roots), len(linkers), len(partners) for both force_ints yes and no
        #          Then, on parse_piana_pathways, add column with info for random groups by reading this new field


	# -----------
	# 5. 
	# -----------
	# print for each relevant protein a file with interactions where it is involved
	# ---------------------- 
	#  -> these files will be referenced from the HTML files so the user can click and see which are the interactions for that protein
	#     ( see utilities.print_proteins_to_file() )
	#     (if you change the name of these files, you have to change it as well on utilities.print_proteins_to_file() )

	for one_protein_extcode in relevant_ext_codes:
	    int_file_fd = file(results_dir + files_prefix + "." + one_protein_extcode.replace("/","_").replace(" ","_") + ".interaction_file.html" , "w")
	    int_file_fd.write(self.interaction_table_strings['header'])
	    if self.interaction_table_strings.has_key(one_protein_extcode):
		for one_int_line in self.interaction_table_strings[one_protein_extcode]:
		    int_file_fd.write(one_int_line)
            # END OF if self.interaction_table_strings.has_key(one_protein_extcode):
            
	    int_file_fd.write(self.interaction_table_strings['footer'])
            
        # END OF for one_protein_extcode in relevant_ext_codes:
	   
	# ----------- 
	# 6. 
	# -----------
	# print to a file all relevant proteins from the network (ie. either all of them or those that were over/under expressed)
	#   -> I am actually creating two files. Read below to learn why...
	# ---------------------- 

	# 6.1 print the HTML file with all relevant proteins
        #    -> we ask the method to create a pickle file so that we can afterwards load it when parsing multiple cases
	#       the object will be a dic with protein as key and content the string that must be printed associated to it.
	#       We only need to create the pickle file once, since these proteins cover all the proteins that the user wants to get information about
	this_file_name= results_dir + files_prefix  + ".all_relevant_proteins" + "." + force_expression + ".html"
	pickle_file_name= results_dir + files_prefix  + ".all_relevant_proteins"+ "." + force_expression  + ".pickle"
	utilities.print_proteins_to_file(file_object= file(this_file_name,"w"), proteins_list=relevant_ext_codes,
					 piana_access= self.piana_access, dic_ext_code_2_pp= dic_relevant_ext_codes, dic_root_ext_codes= dic_root_ext_codes,
					 dic_special_labels= dic_special_labels_ext_codes,
					 dic_linkers= self.ext_code_connecting, dic_over_expressed=dic_over_exp_ext_codes, dic_infra_expressed=dic_infra_exp_ext_codes,
					 dic_root_linked = dic_root_linked, format_mode="html", files_prefix = files_prefix, 
					 file_title= files_prefix + " (forcing expression was " + force_expression + ") all relevant proteins",
					 pickle_file=file(pickle_file_name, "wb")  )

	# 6.2 print the TXT file with all relevant proteins
	#         this is not the most beautiful code in the world, but I cannot think of another way:
	#            -> I need to pass to parse_piana_pathways the list of relevant proteins for each case
	#            -> I need the relevant proteins to be in HTML mode for outputting
	#            -> however, I need to parse the file with relevant proteins, and HTML is not the easiest to parse
	#          So... I am creating two files with the same information: once in HTML and another one in TXT
	txt_file_name= results_dir + files_prefix  + ".all_relevant_proteins"+ "." + force_expression  + ".txt"
	utilities.print_proteins_to_file(file_object= file(txt_file_name,"w"), proteins_list=relevant_ext_codes,
					 piana_access= self.piana_access, dic_ext_code_2_pp= dic_relevant_ext_codes, dic_root_ext_codes= dic_root_ext_codes,
					 dic_special_labels= dic_special_labels_ext_codes,
					 dic_linkers= self.ext_code_connecting, dic_over_expressed=dic_over_exp_ext_codes, dic_infra_expressed=dic_infra_exp_ext_codes,
					 dic_root_linked = dic_root_linked, format_mode="txt", files_prefix = files_prefix , file_title="",
					 pickle_file=None  )

	# ----------- 
	# 7. 
	# -----------
	# Create a file with overlaps between the network relevant proteins and proteins from the pathways (ie. pathway matches)
	# -----------------------

	# print headers for the file
	if format_mode == "html":
	    output_file_object.write("<b>PPI network was built for for %s</b><br><br>\n" %(self.get_graph_id()))
	    output_file_object.write("""Click here for the %s <a href="%s">relevant proteins in the network %s</a> (forcing expression=%s)<br><br>\n""" %(num_relevant_proteins, 
																			  this_file_name,
																			  self.get_graph_id(),  
																			  force_expression) )
	    output_file_object.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080">\n""")
	    output_file_object.write("<tr> <td>pathway</td> <td>pathway size</td> <td>matched</td> <td>percentage matched</td></tr>\n") 
	elif format_mode == "txt":
	    output_file_object.write("file_with_all_relevant_proteins=%s\n" %(this_file_name))

	# now, check how many of relevant proteins appear in each pathways and write result to the file
	for one_pathway in dic_pathways:

	    # convert pathway proteins to the type of identifier required by user
	    this_pathway_ext_codes= Set([])
	    for one_proteinPiana in dic_pathways[one_pathway]:
		
		path_ext_code = self.get_unified_name( proteinPiana= one_proteinPiana,
						       id_type=output_id_type, alternative_id_types=alternative_id_types)
		this_pathway_ext_codes.add(path_ext_code)

		if dic_relevant_ext_codes.has_key(path_ext_code):
		    dic_relevant_ext_codes[path_ext_code].append(one_proteinPiana)
		else:
		    dic_relevant_ext_codes[path_ext_code] = [one_proteinPiana]
	    # END OF for one_proteinPiana in dic_pathways[one_pathway]:

	    num_proteins_in_pathway = len(this_pathway_ext_codes)

	    # print to a file all proteins from the pathway so the user can click and see the pathway proteins
	    all_path_file_name= results_dir + files_prefix  + "." + one_pathway + ".all_pathway_proteins" + ".html"
	    utilities.print_proteins_to_file(file_object= file(all_path_file_name,"w"), proteins_list= this_pathway_ext_codes, piana_access= self.piana_access, 
					     dic_ext_code_2_pp= dic_relevant_ext_codes, dic_root_ext_codes= dic_root_ext_codes,
					     dic_special_labels= dic_special_labels_ext_codes,
					     dic_linkers= self.ext_code_connecting,  dic_over_expressed=dic_over_exp_ext_codes, dic_infra_expressed=dic_infra_exp_ext_codes,
					     dic_root_linked = dic_root_linked, format_mode="html", files_prefix = files_prefix, file_title= " all proteins for pathway " + one_pathway,
					     pickle_file=None  )

	    # calculate the intersection between the relevant proteins and the pathway proteins
	    common_proteins = this_pathway_ext_codes.intersection(relevant_ext_codes)
	    num_matched_proteins = len(common_proteins)

	    # write to a file all proteins that matched the pathway so the user can click and see the overlap between the pathway and the relevant proteins
	    matched_file_name= results_dir + files_prefix  + "." + one_pathway + ".matched_pathway_proteins." + force_expression + ".html"
	    utilities.print_proteins_to_file(file_object= file(matched_file_name,"w"), proteins_list=common_proteins, piana_access= self.piana_access, 
					     dic_ext_code_2_pp= dic_relevant_ext_codes, dic_root_ext_codes= dic_root_ext_codes,
					     dic_special_labels= dic_special_labels_ext_codes,
					     dic_linkers= self.ext_code_connecting,  dic_over_expressed=dic_over_exp_ext_codes, dic_infra_expressed=dic_infra_exp_ext_codes,
					     dic_root_linked = dic_root_linked, format_mode="html",  files_prefix = files_prefix, file_title= " overlap between case " + files_prefix + " and pathway " + one_pathway,
					     pickle_file=None   )

	    try:	per_matched_proteins = 100*num_matched_proteins/float(num_proteins_in_pathway)
	    except:	per_matched_proteins = 0

	    # write to the output results file the results for this pathway
	    if format_mode == "txt":
		output_file_object.write("network_name=%s\tnum_prots_in_network=%s\tpathway=%s\tnum_matched=%s\tpercentage_matched=%s\tnum_prots_in_pathway=%s\tforce_expression=%s\tpathway_file=%s\tmatched_file=%s\n" %(
			self.get_graph_id(),
			num_relevant_proteins,
			one_pathway,
			num_matched_proteins,
			per_matched_proteins,
			num_proteins_in_pathway,
			force_expression,
			all_path_file_name,
			matched_file_name))

	    elif format_mode == "html":
		output_file_object.write("<tr>")
		output_file_object.write("<td>%s</td>" %(one_pathway))
		output_file_object.write("""<td><a href="%s">%s</a></td>""" %(all_path_file_name, num_proteins_in_pathway))
		output_file_object.write("""<td><a href="%s">%s</a></td>""" %(matched_file_name , num_matched_proteins))
		output_file_object.write("<td>%.2f%%</td>" %(per_matched_proteins))
		output_file_object.write("</tr>\n")
		
	    # END OF elif format_mode == "html":
	# END OF for one_pathway in dic_pathways:

	if format_mode == "html":
	    output_file_object.write("</table>")
    
    
	# 8. do the overlap between the pathways and the expression information (regardless of the proteins being relevant or not)
	#    -> this is not taking the network into account for anything, but it is useful to do it here so we can use the
	#       name unification and produce outputs with this very same call
	#    -> the output of this file will be used by parse_piana_pathways to show the influence of various cases expression data on a pathway    
	
	all_expression_on_pathway_txt_file_name = results_dir + files_prefix  + ".expression_on_pathways.txt"
	all_expression_on_pathway_txt_fd = file(all_expression_on_pathway_txt_file_name, "w")
	all_expression_on_pathway_html_file_name = results_dir + files_prefix  + ".expression_on_pathways.html"
	all_expression_on_pathway_html_fd = file(all_expression_on_pathway_html_file_name, "w")
	

	# print headers for html file and txt file (txt file is only used by parse_piana_pathways)
	
	all_expression_on_pathway_html_fd.write("<b>PPI network was built for for %s</b><br><br>\n" %(self.get_graph_id()))
	all_expression_on_pathway_html_fd.write("""Click here for the %s <a href="%s">relevant proteins in the network %s</a> (forcing expression=%s)<br><br>\n""" %(num_relevant_proteins, 
																				     this_file_name,
																				     self.get_graph_id(),  
																				     force_expression) )
	all_expression_on_pathway_html_fd.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080">\n""")
	all_expression_on_pathway_html_fd.write("<tr> <td align=center><b>pathway name</b></td> <td align=center><b>pathway size</b></td> <td align=center><b>expressed</b><br> (%% of total)</td> <td align=center><b>over expressed</b><br> (%% of total)</td> <td align=center><b>under expressed</b><br> (%% of total)</td></tr>\n") 
	
	all_expression_on_pathway_txt_fd.write("case=%s\n" %(self.get_graph_id()))

	# now, check how many of relevant proteins appear in each pathways and write result to the file
	for one_pathway in dic_pathways:

	    dic_expression_on_pathway = {}

	    for one_proteinPiana in dic_pathways[one_pathway]:

		is_over_expressed= False
		is_under_expressed= False

		if self.over_expressed_proteinPianas.has_key(one_proteinPiana):	 is_over_expressed = True

		if self.infra_expressed_proteinPianas.has_key(one_proteinPiana): is_under_expressed = True
	
		
		path_ext_code = self.get_unified_name( proteinPiana= one_proteinPiana,
						       id_type=output_id_type, alternative_id_types=alternative_id_types)

		if dic_expression_on_pathway.has_key(path_ext_code):
		    dic_expression_on_pathway[path_ext_code] = (dic_expression_on_pathway[path_ext_code][0] and is_over_expressed, 
								dic_expression_on_pathway[path_ext_code][1] and is_under_expressed)

		else:
		    dic_expression_on_pathway[path_ext_code] = (is_over_expressed, is_under_expressed)
	    # END OF for one_proteinPiana in dic_pathways[one_pathway]:
	    
	    (number_of_proteins_in_pathway, number_of_expressed_prots, 
	     number_of_over_expressed_prots, number_of_under_expressed_prots) = self._count_expression_dic(expression_dic= dic_expression_on_pathway)

	    all_expression_on_pathway_txt_fd.write("pathway=%s\texpression=%s,%s,%s,%s\n" %(one_pathway, number_of_proteins_in_pathway, number_of_expressed_prots, 
											    number_of_over_expressed_prots, number_of_under_expressed_prots))

	    try:    percentage_expressed = 100*number_of_expressed_prots/number_of_proteins_in_pathway
	    except: percentage_expressed = 0

	    try:    percentage_over_expressed = 100*number_of_over_expressed_prots/number_of_proteins_in_pathway
	    except: percentage_over_expressed = 0

	    try:    percentage_under_expressed = 100*number_of_under_expressed_prots/number_of_proteins_in_pathway
	    except: percentage_under_expressed = 0

	    all_expression_on_pathway_html_fd.write("<tr> <td align=center>%s</td> <td align=center>%s</td> <td align=center>%s<br> (%.2f%%)</td> <td align=center>%s<br> (%.2f%%)</td> <td align=center>%s<br> (%.2f%%)</td></tr>\n" %(one_pathway, number_of_proteins_in_pathway, number_of_expressed_prots, percentage_expressed, number_of_over_expressed_prots, percentage_over_expressed, number_of_under_expressed_prots,percentage_under_expressed) )
	    

	# END OF for one_pathway in dic_pathways:
	all_expression_on_pathway_html_fd.write("</table>")


        
    def match_gos(self, force_expression, output_file_object, format_mode, output_id_type, tax_id_value, 
		  term_type, level_threshold, go_cursor,
		  file_over_expressed= None, file_infra_expressed= None, expression_id_type= None, dic_special_proteins_file=None):
	"""
	Prints out the GO terms frequency for GO terms associated to proteins that appear in the current piana_graph

	   --> this can be used to know which GO terms are 'involved' in the current network. For example, if you build a network for 
               proteins known to be involved in cancer, then you can check which GO terms are related to the proteins in the cancer network
	
	"force_expression" can be 'yes' or 'no' and is used to know which proteins of the network will be used to match pathways
             'no': all proteins in the current network are used to match pathways
	     'yes': only those proteins that are over/under expressed are used to match pathways

        "output_file_object" is the file where the results will be printed

        "pathway_id_type" is the easy-to-remember type name that has been used in the pathways files
          -> Valid pathway_type are those listed in the PIANA reference card (read README.piana_tutorial)

        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html 

        "tax_id_value" is the tax_id of the proteins we are working on (proteins in pathway files must be of this tax_id)

               - set to 0 if you don't mind of the species, or to the tax id in case you do
        
        "term_type" sets the kind of GO terms that will be used for the clustering.
          -> term-type can be "molecular_function", "biological_process" or "cellular_component"
    
        "level_threshold" is the lowest level of the go terms that will be retrieved for calculating the frequency
	     -> all GO terms associated to each protein will be retrieved, then parents of those GO terms until reaching level_threshold will also be retrieved

             -> GO is a hierarchy organized from a initial root level (ie. 0) that increasingly makes more specific the terms. 
                Therefore, the higher the level used the more specific the GO terms will be. To obtain a relevant GO terms
                use level thresholds between 3 and 5. It all depends on how general you want to be in the interpretation of the network.

        "go_cursor" is a mysql connection to your external GO database
	      -> look at PianaApi.print_matched_gos to learn how to create a mysql connection from python

        "file_over_expressed" and "file_infra_expressed" are names of files that contain proteins that are either over expressed or infra expressed

        "expression_id_type" is the type of protein code used in files file_over_expressed and file_infra_expressed

	"dic_special_proteins_file" is a dictionary used to highligth special proteins in your results files
                                    follows structure:   { label1: (id_type1, path_to_file1),
                                                           label2: (id_type2, path_to_file2),
                                                           ...............
                                                         }

	"""


        self._fill_expression_dics(file_over_expressed= file_over_expressed, file_infra_expressed= file_infra_expressed,
                                   expression_id_type= expression_id_type, tax_id_value= tax_id_value)

	self._fill_special_protein_dics(dic_special_proteins_file=dic_special_proteins_file, tax_id_value = tax_id_value,
                                       id_type=None, alternative_id_types=None)


	"""
	1. get all [relevant] proteins from the network
	     -> relevant means 'with over/under expression' in case the user has set force_expression to 'yes'
	2. get GO terms for these relevant proteins
	3. print frequency for each GO term
	"""
	
	if force_expression == "no":
	    # if expression not to be taken into account, get all proteins in the network
	    if verbose_pathway:
		sys.stderr.write("using all proteins\n" )

	    proteinPianas_in_network = Set(self.get_node_ids_list())

	elif force_expression == "yes":
	    # if expression required, use only those proteins that are over/under expressed

	    if verbose_pathway:
		sys.stderr.write("retrieving expressed proteins\n" )

	    proteinPianas_in_network = Set([])

	    for one_proteinPiana in self.get_node_ids_list():
	    
		if self.over_expressed_proteinPianas.has_key(one_proteinPiana) or self.infra_expressed_proteinPianas.has_key(one_proteinPiana):
		    proteinPianas_in_network.add(one_proteinPiana)

	    # END OF for one_proteinPiana in self.get_node_object_list():

	# END OF elif force_expression == "yes":


	# keep the number of proteins in the network being considered
	num_relevant_proteins= len(proteinPianas_in_network)

	if format_mode == "html":
	    output_file_object.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080">""")
	    output_file_object.write("<tr> <td>name</td> <td>go term</td> <td>fequency</td>  <td>expression forced</td> </tr>")

	# 2.1 get the GO terms directly associated to proteins in the network
	list_gos = []	    
	for one_proteinPiana in proteinPianas_in_network:
	    list_gos.extend(self.piana_access.get_protein_go_term_id(proteinPiana_value=one_proteinPiana, term_type_value=term_type))
	# END OF for one_proteinPiana in proteinPianas_in_network:

	# 2.2 get the GO parents for these GO terms
                                  
                                   
	list_parent_gos = GOApi.get_all_parents(list_go_terms=list_gos, dic_gos_at_level={}, 
					     gocursor=go_cursor, piana_access=self.piana_access, stop_level = level_threshold)

	# for the frequency study, we use all GOs regardless if they are parents or directly associated
	list_gos.extend(list_parent_gos)

	# 2.3 create dictionary with frequencies with these GO terms	

	dic_gos = {}   # dictionary that contains frequencies for each GO term in the network
	               #               { go_term: number_of_appearances,
	               #                 go_term: number_of_appearances,
	               #                 ..................
	               #               }

	number_of_gos = 0
	for one_go in list_gos:
	    number_of_gos += 1
	    if dic_gos.has_key(one_go):   dic_gos[one_go] += 1
	    else:                         dic_gos[one_go] = 1
	    
	# END OF for one_go in list_gos:


	# 3. Now, print the frequencies together with the GO name (instead of using the non-informative go term id)

	if format_mode == "html":
	    output_file_object.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080">""")
	    output_file_object.write("<tr> <td>GO name</td> <td>Frequency</td> <td>Percentage</td>  <td>expression forced</td> </tr>")

	num_relevant_proteins= len(proteinPianas_in_network)

	# 3.1 transform the dic into tuples so we can then order them by frequency
	ordered_go_tuples = []
	for one_go_term in dic_gos:
	    
	    ordered_go_tuples.append( (dic_gos[one_go_term], one_go_term) )
	# END OF for one_go_term in dic_gos:
	    
	ordered_go_tuples.sort()    # putting list in order
	ordered_go_tuples.reverse() # we want to print first the most frequent terms: reverse it

	for one_tuple in ordered_go_tuples:
	    # one_tuple[0] is the number of appearances of the go term
	    # one_tuple[1] is the GO term
	    #
	    
	    go_name = self.piana_access.get_protein_go_name(go_term_id_value=one_tuple[1] )
	    
	    if format_mode == "html":
		output_file_object.write("<tr> <td>%s</td> <td>%s</td> <td>%.2f%%</td> <td>%s</td> </tr>" %(go_name, one_tuple[0], 100*one_tuple[0]/float(number_of_gos), force_expression  ))
		
	    elif format_mode == "txt":
		output_file_object.write("network_name=%s\tnum_prots_in_network=%s\tgo=%s\tfreq=%s\tpercentage=%.2f%%\tnum_go_terms=%s\tforce_expression=%s\n" %(
			self.get_graph_id(), 
			num_relevant_proteins, 
			go_name, 
			one_tuple[0], 
			100*one_tuple[0]/float(number_of_gos),
			number_of_gos,
			force_expression))

	# END OF one_tuple in ordered_go_tuples:

	if format_mode == "html":
	    output_file_object.write("</table>")


    def match_spots_to_proteins(self, spots_file_object= None, mw_error_bounds= None, ip_error_bounds= None, match_mode="all"):
        """

        matches spots in a spot file with proteins in the network
        
        "spots_file_object": a file object with spots from a Gel, and their Molecular Weight and Isoelectric Point

               text file with spots must follows format (one spot per line):
        
                    spot_id<TAB>Molecular Weight<TAB>Isoelectric Point

               Attention!!! - Numbers must be in american style: 234234.45 and not 234234,45
                            - No headers or footers allowed



        returns a dictionary (with error levels as keys) with a list of tuples [spot_id, proteinPiana]

               { error_level1: [ [spot_id1, proteinPiana1], [spot_id2, proteinPiana2], ...],
                 error_level2: [ [spot_id1, proteinPiana1], [spot_id2, proteinPiana2], ...],
                 ......................
               }

               (one spot can be assigned to several proteinPianas and viceversa (this is just a matching by mw and ip...)


        "mw_error_bounds" and "ip_error_bounds" are lists of error bounds (they must have the same number of elements)

            the error bounds describe the percentage of error admitted when matching a spot mw or ip to the theoretical mw or ip of a protein
            for example:

                    mw_error_bounds   = [0.0, 0.0025, 0.005, 0.01]

                    ip_error_bounds   = [0.0, 0.0025, 0.005, 0.01]


        "match_mode" can be (default is "all")
          - "all" : matching performed by comparing molecular weight and isoelectric point
          - "mw"  : matching performed by comparing only molecular weight
          - "ip"  : matching performed by comparing only isoelectric point

        """


        if spots_file_object is None or mw_error_bounds is None or ip_error_bounds is None or match_mode is None:
            raise ValueError("all arguments needed to match spots to proteins: spots file=%s, mw errors=%s, ip errors=%s and match_mode=%s" %(spots_file_object,
                                                                                                                                        mw_error_bounds,
                                                                                                                                        ip_error_bounds,
                                                                                                                                        match_mode))

        # Important! molecular_error_bounds and isoelectric_error_bounds must have same number of elements
        # It is the case because the resulting matches will be given for "error levels", where each position in the list of
        # error bounds is a new "error level". Therefore, mw_error_bounds[n] will be in the same error level as ip_error_bounds[n], regardless
        # of the actual values
        if len(mw_error_bounds) != len(ip_error_bounds):
            raise ValueError("mw_error_bounds and ip_error_bounds must have the same number of elements")

        # Perform matching with spots file
        #    molecular weight and isoelectric point matching will determine the identity of the spots of a gel
        # -----------------------------------------------------------------------------------------------------
        # 1. Filtering by molecular weight
        # 2. Filtering by isoelectric point
        # 3. Printing out intersection of two filters

        spots_mw = {}  # dic used to store spots mol weights (keys are spot_id)
        molecular_n_match = 0    # counter of number of matches with molecular weight filter

        spots_ip = {}  # dic used to store spots isoelectric points (keys are spot_id)
        isoelectric_n_match = 0    # counter of number of matches with isoelectric point filter


        # then extract the spot data from the input file

        for line in spots_file_object.readlines():
            line = line.rstrip()
            values = line.split("\t")

            if verbose_spots:
                sys.stderr.write("for spot_id %s reading mw=%s (type %s) and ip=%s (type %s)" %(values[0],values[1], type(values[1]),values[2], type(values[2]) ))
            
            spots_mw[values[0]] = int( float(values[1]) * 1000) # conversion to units used by piana molecular weight
                                                               # piana has weights like: 10575.9, 132796, ...
                                                               # and the spots file 10.575, 132.796, ...
            spots_ip[values[0]] = float(values[2])
        # END OF for line in spots_file_object.readlines():

        # -----------------------------
        # 1. Filtering by molecular weight
        # -----------------------------

        # nodes_molecular_weights is a dictionary of the form:
        # {proteinPiana1:molecular_weight1, proteinPiana:molecular_weight2, ...}
        nodes_molecular_weights = self.get_nodes_molecular_weights()

        #print "NODES MOLECULAR WEIGHTS:\n"
        #print nodes_molecular_weights

        # nodes_isoelectric_points is a dictionary of the form:
        # {proteinPiana:[isoelectric_point1, isoelectric_point2, ...], proteinPiana:[], ...}
        nodes_isoelectric_points = self.get_nodes_isoelectric_points()

        #print "NODES ISOELECTRIC POINTS:\n"
        #print nodes_isoelectric_points


        # for each level of error do the matching
        # level error can have different values for mw_error_bounds and ip_error_bounds, but since len(xx_error_bounds) is equal
        # they both use the same index

        matchings_to_return = {}  # dictionary that will contain all spots matches to proteinPianas for each error level
                                  #  it follows structure:
                                  #                      { error_level1: { spot_id: proteinPiana,
                                  #                                        spot_id: proteinPiana,
                                  #                                          ......................},
                                  #                        error_level2: { spot_id: proteinPiana,
                                  #                                        spot_id: proteinPiana,
                                  #                                          ....................},
                                  #                        .....................................
                                  #                      }



        

        for i in range(len(mw_error_bounds)):

            matchings_to_return[i] = {}

            assignments_molecular_weight = []   # a list of tuples [spot_id, proteinPiana] matching spots and proteins
                                                # according to molecular weight

            assignments_isoelectric_point = []  # a list of tuples [spot_id, proteinPiana] matching spots and proteins 
                                                # according to isoelectric point

            mw_n_match = 0   # counts number of mw matches
            ip_n_match = 0 # counts number of mw matches

            # for each spot_id, check molecular weight
            for spot_id in spots_mw.keys():

                # ...compare with each  protein in the network
                for proteinPiana in nodes_molecular_weights.keys():
                    lower_bound = int( (1.0 - mw_error_bounds[i]) * spots_mw[spot_id] )
                    upper_bound = int( (1.0 + mw_error_bounds[i]) * spots_mw[spot_id] )

                    if (nodes_molecular_weights[proteinPiana] >= lower_bound) and (nodes_molecular_weights[proteinPiana] <= upper_bound):
                        mw_n_match += 1

                        assignments_molecular_weight.append( [spot_id, proteinPiana] )

                    # END OF if (nodes_molecular_weights[proteinPiana] > lower_bound) and (nodes_molecular_weights[proteinPiana] < upper_bound):
                # END OF for proteinPiana in nodes_molecular_weights.keys():
            # END OF for spot_id in spots_mw.keys():


            # for each spot_id, check isoelectric point
            for spot_id in spots_ip.keys():

                # ...compare with each linked protein
                for proteinPiana in nodes_isoelectric_points.keys():
                    lower_bound = int( (1.0 - ip_error_bounds[i]) * spots_ip[spot_id] )
                    upper_bound = int( (1.0 + ip_error_bounds[i]) * spots_ip[spot_id] )

                    if (nodes_isoelectric_points[proteinPiana] >= lower_bound) and (nodes_isoelectric_points[proteinPiana] <= upper_bound):
                        ip_n_match += 1

                        assignments_isoelectric_point.append( [spot_id, proteinPiana] )

                    # END OF if (nodes_isoelectric_points[proteinPiana] > lower_bound) and (nodes_isoelectric_points[proteinPiana] < upper_bound) 
                # END OF for proteinPiana in nodes_isoelectric_points.keys():
            # END OF for spot_id in spots_ip.keys():

            # loops finished: update dictionary with results obtained for this error level (update depends on match_mode)


            if match_mode == "mw":
                matchings_to_return[i] =  assignments_molecular_weight
                
            elif match_mode == "ip":
                matchings_to_return[i] =  assignments_isoelectric_point
                
            elif match_mode == "all":

                matchings_to_return[i] = []
                
                for assignment_molecular_weight in assignments_molecular_weight:

                    if assignment_molecular_weight in assignments_isoelectric_point:
                        # if tuple is in both filters, then append to matchings_to_return
                        matchings_to_return[i].append(assignment_molecular_weight)
                        
                # END OF for assignment_molecular_weight in assignments_molecular_weight:

            else:
                raise ValueError('match_mode must have a value "mw", "ip" or "all"')
            
        # for i in len(mw_error_bounds):

        return matchings_to_return


    def print_spots_matches(self, spots_matches= None, output_target= None, format_mode = "txt",
                            id_type= None, mw_error_bound= None, ip_error_bound= None, alternative_id_types=[],
                            user_protein_names = {} ):
        """
        prints to "output_target" (a file object) the spots maches for each level, using protein codes indicated by "id_type"
            
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html
               
        "id_type" is the easy-to-remember type name that will be used for printing the proteins in the network.
        Valid id_type are those listed in the PIANA reference card (read README.piana_tutorial)

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
        --> user must provide pairs a list of valid easy-to-remember type names
            alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             
            I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "user_protein_names" can be used to fix a set of protein names that have preference over other names when a protein has several names
           (eg. if proteinPiana has gene names CXCL1 and MGSA, and this dictionary has CXCL1, it is guaranteed that CXCL1 will be used for output)
           --> user_protein_names is a dictionary that follows the structure: { protein_name:None, protein_name:None, ...}

             
        spots_matches were obtained with method PianaGraph.match_spots_to_proteins

        spots_matches follow structure:
                                   { error_level1: { spot_id: proteinPiana,
                                                     spot_id: proteinPiana,
                                                       ......................},
                                     error_level2: { spot_id: proteinPiana,
                                                     spot_id: proteinPiana,
                                                       ....................},
                                     .....................................
                                   }

                                   
        """
        # make sure self.dic_protein_naming is filled
        self._fill_dic_protein_naming(id_type=id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= [], key_list=None, user_protein_names = user_protein_names)

        self._create_unified_network(id_type=id_type, alternative_id_types=alternative_id_types, user_keywords=[])
        
        dic_already_printed = {}   # used to avoid printing the same combination spot--> protein more than once

        if format_mode == "html":
            # print table headers
            output_target.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080"><tr><td>mw error</td><td>ip error</td><td>spot id</td><td>protein id</td></tr>""")

            
        for spot_match_error_level in spots_matches.keys():

            for spot_match in spots_matches[spot_match_error_level]:

                protein_ext_code = self.get_unified_name( proteinPiana= spot_match[1],
                                                          id_type=id_type, alternative_id_types=alternative_id_types)

                key = spot_match[0] + "." + protein_ext_code
                if dic_already_printed.has_key(key):
                    # avoid printing the same combination spot-protein
                    continue
                
                dic_already_printed[key] = None
                
                if verbose_spots:
                    sys.stderr.write("at error level %s (mw_error %s - ip_error %s) spot_id %s matched protein %s\n" %(spot_match_error_level,
                                                                                                                     mw_error_bound[spot_match_error_level],
                                                                                                                     ip_error_bound[spot_match_error_level],
                                                                                                                     spot_match[0],
                                                                                                                     protein_ext_code) )
                if format_mode == "txt":
                    output_target.write("error level %s (mw_error %s - ip_error %s) spot_id %s matches protein %s\n" %(spot_match_error_level,
                                                                                                                       mw_error_bound[spot_match_error_level],
                                                                                                                       ip_error_bound[spot_match_error_level],
                                                                                                                       spot_match[0],
                                                                                                                       protein_ext_code) )
                elif format_mode == "html":
                    output_target.write("<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n" %(mw_error_bound[spot_match_error_level],
                                                                                                    ip_error_bound[spot_match_error_level],
                                                                                                    spot_match[0],
                                                                                                    protein_ext_code))
            # END OF for spot_match in spots_matches[spot_match_error_level]:
        # END OF for spot_match_error_level in spots_matches.keys():
    
        if format_mode == "html":
            # print table closure
            output_target.write("\n</table></body></html>\n")

       
    def get_list_unified_names(self, list_proteinPianas,
                               output_id_type,
                               alternative_id_types,
                               key_list= 0):
        """
        returns a list with the unified names for proteinPianas in list_proteinPianas
        """
   
        self._fill_dic_protein_naming(id_type=output_id_type,
                                      alternative_id_types=alternative_id_types,
                                      list_proteinPianas= list_proteinPianas,
                                      key_list=key_list,
                                      user_protein_names= {} )

        return self.dic_protein_naming
        
    def output_shortest_route(self, protein_a_name, protein_b_name, input_id_type,
                              output_id_type, alternative_id_types, output_file_object,
                              format_mode, tax_id_value):
        """
        
        Prints to "output_file_object" the route from "protein_a_name" to "protein_b_name"
        

        "input_id_type" is the type of code used for protein query_protein
             -> valid protein-type values are those defined in the PIANA reference card (read README.piana_tutorial)

        "output_id_type" is the type of code that will be used to print proteins
             -> valid protein-type values are those defined in the PIANA reference card (read README.piana_tutorial)
        
        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
         --> user must provide a list of valid easy-to-remember type names
             alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             I suggest always placing md5 at the end of alternative types, so you never get a None in the output
             
        "output_file_object"is the file object (sys.stdout to print to screen) where the route will be printed
        
        
        "format_mode" sets the type of format that will be used for output  (Attention! currently not working!!!! mode is always TXT)
               - 'txt' will print flat text
               - 'html' will print html 

        "tax_id_value" sets the species of the protein that will be printed (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        """

        # make sure self.dic_protein_naming is filled
        self._fill_dic_protein_naming(id_type=output_id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= [], key_list=None, user_protein_names = {protein_a_name:None, protein_b_name:None})

        self._create_unified_network(id_type=output_id_type, alternative_id_types=alternative_id_types, user_keywords=[])
        
        list_proteinPiana_a = self.piana_access.get_list_proteinPiana(protein_id= protein_a_name,
                                                                      id_type= input_id_type,
                                                                      tax_id_value=tax_id_value,
                                                                      source_db_info="no")

        list_proteinPiana_b = self.piana_access.get_list_proteinPiana(protein_id= protein_b_name,
                                                                      id_type= input_id_type,
                                                                      tax_id_value=tax_id_value,
                                                                       source_db_info="no")


        min_shortest_route = (1000, []) # init to very high value so it is changed at the first route found
        
        # find shortest routes for all combinations of proteinPianas: we will keep the shortest one of them
        for proteinPiana_a in list_proteinPiana_a:
            for proteinPiana_b in list_proteinPiana_b:
                this_shortest_route= self.find_shortest_route(start_node_id=proteinPiana_a, end_node_id=proteinPiana_b)

                if this_shortest_route[0] < min_shortest_route[0]:
                    min_shortest_route = this_shortest_route
                
        # END OF for proteinPiana_a in list_proteinPiana_a:

        
        if min_shortest_route[1]:
            output_file_object.write("protein_a=%s\tprotein_b=%s\tdistance=%s\n" %(protein_a_name, protein_b_name, min_shortest_route[0]))

            ext_code = self.get_unified_name( proteinPiana= min_shortest_route[1][0].get_node_id(),
                                              id_type= output_id_type,
                                              alternative_id_types=alternative_id_types )

            output_file_object.write( "START=%s" %(ext_code))

            for node in min_shortest_route[1][1:-1]:
                # for each node in the route, print its external code in order


                ext_code = self.get_unified_name( proteinPiana= node.get_node_id(),
                                                  id_type= output_id_type,
                                                  alternative_id_types=alternative_id_types )


                output_file_object.write( "<-->%s" %(ext_code))
            # END OF for node in min_shortest_route[1][1:-1]:

            ext_code = self.get_unified_name( proteinPiana= min_shortest_route[1][-1].get_node_id(),
                                              id_type= output_id_type,
                                              alternative_id_types=alternative_id_types )

            output_file_object.write( "<-->END=%s\n" %(ext_code))
        # END OF if min_shortest_route[1]:
        
        else:
            output_file_object.write("protein_a=%s\tprotein_b=%s\tdistance=NO_ROUTE\n" %(protein_a_name, protein_b_name))

            
    

    
    def get_distance_groups(self, protein_ext_code=None, input_id_type= None, output_id_type= None, alternative_id_types=[],
                            user_protein_names= {}, tax_id_value= 0):
        """

        gets proteins that are at a certain distance from a given protein protein_ext_code
        
        returns a dictionary that follows the structure:

                            { 1:[list of protein ext codes at distance 1 from protein_ext_code],
                              2:[list of protein ext codes at distance 2 from protein_ext_code],
                              3:[list of protein ext codes at distance 3 from protein_ext_code],
                              4:[list of protein ext codes at distance 4 from protein_ext_code]
                            }
                            
        "protein_ext_code" is the protein for which we want to find proteins at a certain distance

        "input_id_type" is the easy-to-remember identifier type used for protein_ext_code
          --> valid values are those listed in the PIANA reference card (read README.piana_tutorial)

        "output_id_type" is the easy-to-remember identifier type that you want to use for results
          --> valid values are those listed in the PIANA reference card (read README.piana_tutorial)

        "alternative_id_types" can be used to set a list of alternative types in case no id_type code is found
        --> user must provide pairs a list of valid easy-to-remember type names
            alternative_id_types can for example look like this: ["gi", "uniacc", "md5"]
             
            I suggest always placing md5 at the end of alternative types, so you never get a None in the output

        "tax_id_value" sets the species of the proteins being returned (can be used for eliminating ambiguities between codes across species)
           --> valid tax ids are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
           
        """
        # make sure self.dic_protein_naming is filled
        self._fill_dic_protein_naming(id_type=output_id_type, alternative_id_types=alternative_id_types,
                                      list_proteinPianas= [], key_list=None, user_protein_names=user_protein_names)

        self._create_unified_network(id_type=output_id_type, alternative_id_types=alternative_id_types, user_keywords=[])

        distance_dic = {} # initialize the dictionary of distances that will be returned


        list_proteinPiana = self.piana_access.get_list_proteinPiana(protein_id= protein_ext_code,
                                                                    id_type=input_id_type,
                                                                    tax_id_value=tax_id_value,
                                                                    source_db_info="no")
        
        for proteinPiana in list_proteinPiana:
            # for each proteinPiana associated to the query protein, find which proteins in the network are connected to it
            
            # get_distances returns a dictionary { proteinPiana: distance, proteinPiana: distance, ....}
            protein_distances = self.get_distances(query_node_id= proteinPiana)


            proteins_at_distance_x_from_query= protein_distances.keys()
            
            for distant_proteinPiana in proteins_at_distance_x_from_query:
                # for each protein that is at a certain distance from the query protein, get the unified ext code and
                # add it to the dictionary of groups of proteins at a distance from the query protein
                
                ext_code = self.get_unified_name( proteinPiana= distant_proteinPiana,
                                                  id_type= output_id_type,
                                                  alternative_id_types= alternative_id_types)
               
                
                #       protein_distances[distant_proteinPiana] is an integer with the distance at which distant_proteinPiana can be found
                
                if distance_dic.has_key( protein_distances[distant_proteinPiana] ):
                    distance_dic[protein_distances[distant_proteinPiana]][ext_code] = None
                else:
                    distance_dic[protein_distances[distant_proteinPiana]] = {ext_code:None}
            # END OF for distant_proteinPiana in proteins_at_distance_x_from_query:
        # END OF for proteinPiana in list_proteinPiana:

        return distance_dic


    def print_distance_group(self, distance_group,  input_id_type,  output_target, info, format_mode="txt",
                             user_protein_names= {}, tax_id_value=0):
        """
        prints a distance group to output_target 

        "distance_group" is a list of protein codes of type input_id_type 
        
        "output_target" is a file object (sys.stdout to print to screen)

        "info" determines which information will be printed next to the proteins in the distance group
          (info will be preceded by a header stating the type of info being printed)
           - blank: nothing
           - all: all info available
           - scop: scop codes
           - cath: cath codes
           
        "format_mode" sets the type of format that will be used for output
               - 'txt' will print flat text
               - 'html' will print html

        "input_id_type" is the easy-to-remember identifier type used for protein_ext_code
          --> valid values are those listed in the PIANA reference card (read README.piana_tutorial)

        "tax_id_value" sets the species of the proteins that will be printed (can be used for eliminating ambiguities between codes across species)
           --> valid species names are 0 (do not take into account the species) and those taxonomy ids provided by ncbi taxonomy
        """

        # I think there is no need for this, since it has already been done when creating the distance group
        # this is just printing a list of protein ext codes that were found when creating the distance group, together with theis scops and caths
        
        # make sure self.dic_protein_naming is filled
        #self._fill_dic_protein_naming(id_type=output_id_type, alternative_id_types= alternative_id_types,
        #                              list_proteinPianas= [], key_list=None, user_protein_names=user_protein_names)

        #self._create_unified_network(id_type=id_type, alternative_id_types=alternative_id_types, user_keywords=[] )

        if format_mode == "html":
            output_target.write("""<table border=1 style="border-collapse: collapse" bordercolor="#808080"><tr><td>protein</td><td>scop codes</td>\<td>cath codes</td></tr>\n""")
        
        for protein in distance_group:
            list_proteinPiana = self.piana_access.get_list_proteinPiana(protein_id= protein,
                                                                        id_type= input_id_type,
                                                                         tax_id_value=tax_id_value,
                                                                         source_db_info="no")

            dic_scop = {}
            dic_cath = {}
            
            for proteinPiana in list_proteinPiana:

                list_scop= self.piana_access.get_protein_scop_cf_sf_fa(proteinPiana_value=proteinPiana)

                for one_scop in list_scop:
                    scop_key = str(one_scop[0]) + "." + str(one_scop[1]) + "." + str(one_scop[2])
                    dic_scop[scop_key] = (one_scop[0], one_scop[1], one_scop[2] )
                    
                list_cath= self.piana_access.get_protein_cath(proteinPiana_value=proteinPiana, residue_value=None)
                for one_cath in list_cath:
                    dic_cath[one_cath] = None

            # END OF for proteinPiana in list_proteinPiana:

            if format_mode == "txt":
                output_target.write("%s" %(protein) )
                
            elif format_mode == "html":
                output_target.write("<tr><td>%s</td>" %(protein) )
                
            if info== "all" or info == "scop":

                if format_mode == "html":
                    output_target.write("<td>")

                for scop_key in dic_scop:
                    if format_mode == "txt":
                        output_target.write("\tscopcf:%s\tscopsf:%s\tscopfa:%s" %(dic_scop[scop_key][0], dic_scop[scop_key][1], dic_scop[scop_key][2]) )
                    elif format_mode == "html":
                        output_target.write(" scop cf:%s scop sf:%s scop fa:%s " %(dic_scop[scop_key][0], dic_scop[scop_key][1], dic_scop[scop_key][2]) )

                if format_mode == "html":
                    output_target.write("</td>")

                    
            if info== "all" or info == "cath":
                if format_mode == "html":
                    output_target.write("<td>")
                    
                for cath in dic_cath:
                    if format_mode == "txt":
                        output_target.write("\tcath:%s" %(cath))
                    elif format_mode == "html":
                        output_target.write(" cath:%s " %(cath))
                if format_mode == "html":
                    output_target.write("</td>")
            
            if format_mode == "txt":
                output_target.write("\n")
            elif format_mode == "html":
                output_target.write("</tr>\n")
                
        
        # END OF for protein in distance_group:
        
        if format_mode == "html":
            output_target.write("\n</table></body></html>\n")



    def get_unified_edge(self,edge_key):
        """
        Get the edge node and edge attribute for this edge_key
        """

        if self.memory_mode=="memory":
            return self.edge_key_unified_edge_correspondence[edge_key]
        else:
            try:
                current_edge_id = self.edge_key_unified_edge_correspondence[edge_key][0]
                current_edge = self.get_edge( identifier1 = current_edge_id, get_mode= "error")
                current_edge_attribute = current_edge.get_edge_attribute_object()
                current_edge_attribute._set_associated_interactionPiana(self.edge_key_unified_edge_correspondence[edge_key][1:])
                current_edge.set_attribute(attribute_object=current_edge_attribute)

                return current_edge
        
            except KeyError:
                raise ValueError("Trying to get an unexisting unified edge!")



                                                                    
                                                                    
                                                                  

                                     
