"""
 File       : PianaDBaccess.py
 Author     : R. Aragues & J. Planas
 Creation   : 23.01.2004
 Contents   : class used as an interface with database piana
 Called from: all classes/programs that select/insert information from/into piana

=======================================================================================================

This class is the piana user interface to piana database. There are
lower levels that can be used to select/insert information from/into
database piana (e.g. PianaDB class) but they shouldn't be used unless
the programmer has a complete understanding of the system.

This class implements functions such as insert_interaction() that,
given all data about a particular interaction will make consistency
checks and all required insertions to piana table. This is done this
way to insure that a user doesn't forget to update all tables that are
related to a new interaction, making trasparent for him the internal
structure of the database.

Summarizing, this class should be the only means of interaction with
the database used by piana
"""

# PianaDBaccess.py: class used as an interface with database piana
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues


import re, sys, os
import Bio.SeqUtils.ProtParam
import time
from sets import *

from GraphNodeAttribute import *
from GraphEdgeAttribute import *
from PianaDB import *
import PianaGlobals

import utilities

verbose = 0
verbose_insert_interaction = 0
verbose_species= 0


class PianaDBaccess(object):
    """
    Class used as an interface with database piana
    """
    def __init__(self, dbname=None, dbhost=None, dbuser=None, dbpassword= None, dbport= None, buffer=1 ):
        """
        "dbname" is the database name to which you want to connect to (required)
 
        "dbhost" is the machine with the mysql server that holds the piana database (required)

        "dbuser" is the mysql user (not required in most systems)

        "dbpassword" is the mysql password (not required in most systems)

        "dbport" is the mysql port (not required in most systems)

        "buffer" indicates if the user wants to use the intern Piana Buffer or not. It can be "None" (not use buffer) or "1" (use buffer) The buffer is used by default

        """

        if verbose:
            sys.stderr.write("Arguments received in PianaDBaccess() are: %s, %s, %s, %s\n" %( dbname, dbhost, dbuser, dbpassword))

        if dbname is None:
            raise ValueError("A piana database name is needed to initialiate a connection")
            
        if dbhost is None:
            raise ValueError("A host name for the piana database is needed to initialiate a connection")

        self.dbname= dbname
        self.dbhost= dbhost
        self.dbuser= dbuser
        self.dbpassword= dbpassword
        self.dbport= dbport

        # opening connection to database piana using class PianaDB
        self.db = PianaDB(dbname=self.dbname, dbhost=self.dbhost, dbuser=self.dbuser, dbpassword= self.dbpassword, dbport= self.dbport)

        # Check into PIANA database which external protein Types are available
        self.validProteinIds = self._load_valid_protein_ids()

        # Check into PIANA database which database sources are available
        self.validSources = {}
        self.validSources = self._load_valid_source_dbs()

        # javi: modified. When there are many databases, it is computationally costly to generate all combinations
        #self.int_dbs_mask = self._load_ints_dbs_mask() # this is a dictionary that contains for each combination of interaction databases,
                                                       # which would be the integer representing them
                                                       # The format is the following:
                                                       #   { integer: ['db1', 'db2', 'db3'],
                                                       #     integer: ['db1'],
                                                       #     integer:
                                                       #     .......................

        # alternative: combinations will only be created when they are found. When a new combination is found, its correspondent source_dbs names will be associated.
        self.int_dbs_mask = {}


        # Check into PianaGlobals which interaction methods are available (key: external identifier for the method)
        # It is created as an empty dictionary, and only it is filled when it is needed for the first time
        self.validInteractionMethods = {}

        
        # Variables used when introducing data into database in order to make it faster
        self.lock_frequency = 20000
        self.current_lock_num = 0


        # Obtain which is the maximum packet size accepted by server, used in buffer
        # Size of the packet in bytes
        self.dbmaxpacket = self._get_max_allowed_packet_()
        
        # Create the Object Buffer. If the buffer size cannot be known (self.dbmaxpacket=None), the buffer will not be used
        # So, if the size is known, this maximum packet size will be used by default (in MySQL Server by default, 1Mb)
        if self.dbmaxpacket is None:
            self.insert_buffer = None

        self.db_temp_buffer = Buffer(self.dbmaxpacket,self) # Buffer to save temporal data in a table in the database

        if buffer:
            self.insert_buffer = Buffer(self.dbmaxpacket,self)
            # Create a temporal dictionary to save data that may be consulted by some methods (and if the data is in the buffer, it would not be found in the database)
            self.insert_protein_buffer = {}
            self.insert_sequence_buffer = {}
            self.insert_interaction_buffer = {}
            self.newproteinPiana = self.db.select_db_content( PianaGlobals.SelectNewProteinPiana.get_sqlquery(),
                                                              answer_mode="single" )
            self.newSequenceID = self.db.select_db_content( PianaGlobals.SelectNewSequenceID.get_sqlquery(),
                                                            answer_mode="single" )
            self.newInteractionPiana = self.db.select_db_content( PianaGlobals.SelectNewInteractionPiana.get_sqlquery(),
                                                                  answer_mode="single")
        
    # ----
    # methods required for using pickle with piana objects
    # ----
    def __getstate__(self):
        odict = self.__dict__.copy() # copy the dict since we change it
        del odict['db']              # remove filehandle entry
        return odict

    def __setstate__(self, dict):
        self.__dict__ = dict
        self.__class__.__init__(self, self.dbname, self.dbhost, self.dbuser, self.dbpassword)
        
    def __getnewargs__(self):
        return (self.dbname, self.dbhost, self.dbuser, self.dbpassword)


    # --
    # not sure it is being used... 
    # --
    def get_piana_db(self):
        """
        Returns the PianaDB object used for this PianaDBaccess object
        
        """
        return self.db


    def close(self):
        """
        Close the connection with database
        """

        # Insert buffer
        if self.insert_buffer:
            self.execute_insert_buffer()

        # Update temporal data to its owner tables
	sys.stderr.write("Emptying temporal table...")
        self.empty_temporal_table()

        # Unlocking tables
        self.unlock_tables()

        self.db.close()


    def check_lock_frequency(self,table_list=[]):

        if self.current_lock_num == self.lock_frequency:
            # Unlocking tables
            #self.execute_insert_buffer("temptable")
            #self.empty_temporal_table()
	    #raise ValueError("Control")
            self.unlock_tables()
            self.current_lock_num = 0
            #raise ValueError("control")
            self.lock_tables(table_list)
        else:
            self.current_lock_num += 1
            
    def set_lock_frequency(self,frequency_value):
        """
        Method to change the lock/unlock frequency (used only in parsers, to speed up insertions and deletions
        """
        self.lock_frequency = frequency_value

    # --------------------------------------------------------------------------
    # Methods used to load available protein types and databases in the database
    # --------------------------------------------------------------------------

    def _load_valid_protein_ids(self):
        """
        Returns a dictionary with valid protein types stored in database

        Each key in the dictionary is the identifier type, and each record contains a dictionary
        with the following format:

                   "table": table where this protein type is stored
                   "column": column name where the external code is stored
                   "description": description for this external identifier

        It is supposed that proteinPiana column name will be always "proteinPiana"
        and that sourceDBID field will be alwarys "sourceDBID"
        """

        valid_types = self.db.select_db_content( PianaGlobals.SelectValidProteinIds.get_sqlquery(validProteinIds_table=PianaGlobals.validProteinIds_table)
                                                 , answer_mode="raw", remove_duplicates="no" )

        # Dictionary where results will be stored
        valid_ids = {}
        for type in valid_types:
            valid_ids[type[0]] = {"table":type[1],
                                  "column":type[2],
                                  "description":type[3]}

        return valid_ids

    def _add_valid_protein_id(self, proteinTypeId, table, column, description):
        """
        Adds an external database as current external databases stored in memory
        """

        self.validProteinIds[proteinTypeId] = {"table": table,
                                               "column": column,
                                               "description": description}

    def get_valid_protein_ids(self):

        return self.validProteinIds


    def get_table_of_id_type(self, id_type):
	"""
	Method that returns the table in which "id_type" can be found (ie the pianaDB table where this protein identifier is located)	
	"""

	try:
		return self.validProteinIds[id_type]["table"]
            
	except KeyError:	
		raise ValueError("%s is not a valid protein type identifier in current database" %(id_type))

    def get_column_of_id_type(self, id_type):
	"""
	Method that returns the column in which "id_type" can be found
	"""
        
	try:
            return self.validProteinIds[id_type]["column"]
        
	except KeyError:
            raise ValueError("%s is not a valid protein type identifier in current database" %(id_type))
            
##     def get_table_from_column(self, sqlcolumn):
## 	"""
## 	Method that returns the column in which "id_type" can be found
## 	"""

## 	try:
##             for one_id_type in self.validProteinIds:
##                 if self.validProteinIds[one_id_type][1] == sqlcolumn:
##                     return self.validProteinIds[one_id_type][0]
            
## 	except KeyError:
## 		raise ValueError("%s is not a valid protein type identifier in current database" %(id_type))
            


        
    def _load_valid_source_dbs(self):
        """
        Returns a dictionary with current external databases stored in PIANA database

        and integer identifier assigned to this database
        """

        current_sources = self.db.select_db_content( PianaGlobals.SelectValidSourceDBs.get_sqlquery()
                                                     , answer_mode="raw", remove_duplicates="no" )

        return dict([(actual_source[0], actual_source[5]) for actual_source in current_sources])

        # return dictionary where results will be stored
	# Each database identifier has an integer identifier

        # OLD METHOD:
        # The database identifier integer will be defined by its alphabetical order
        #current_sources.sort()
	#return dict( [(current_sources[i],2**i) for i in xrange(len(current_sources))] )

    def _get_sourceDBID_int(self, sourceDBID):
        """
        Returns the decimal value assigned to the sourceDB (sourceDBID will be converted to lower_case)
        """
        return self.validSources[sourceDBID.lower()]

    def _add_valid_source_db(self, databaseName):
        """
        Adds an external database as current external databases stored in memory
        """
        current_sources = self.get_list_valid_sources()

        current_sources.append(databaseName)
        current_sources.sort()
        self.validSources = dict( [(current_sources[i],2**i) for i in xrange(len(current_sources))] )

    def add_external_database_information_type(self,externalDatabase,information_type):
        """
        Adds to an external database which kind of information is used from it

        ATTENTION! Information_type must be one element of the valid types: (it has to be checket previously!)

        "protein sequences"
        "protein attributes"
        "protein-protein interactions"
        """

        if not self.get_valid_sources().has_key(externalDatabase):
            raise ValueError("Not valid externalDatabase (%s) identifier. This externalDatabase must exist in the database" %(externalDatabase) )
        
        self.db.insert_db_content( PianaGlobals.InsertExternalDatabaseInformationType.get_sqlquery(databaseName=externalDatabase,
                                                                                                   information_type=information_type),
                                   answer_mode=None, buffer=None )

                                                   
    #def _load_complete_description_valid_source_dbs(self, databaseType="all"):
    
    def get_source_dbs_dict(self, database_type="all"):
        """
        Returns a dictionary with complete information of current external databases stored in PIANA database

        "databaseType" is used to restrict those databases for which information must be retrieved. It can be:
        
                  "all"
                  "protein sequences"
                  "protein attributes"
                  "protein-protein interactions"
                  "identifiers cross-references"

        The dictionary has the following format:
                key: databaseName (the database identifier in PIANA database)
                Each record in the dictinoary contains a dictionary with the keys (and their correspondent contents):
                        databaseVersion
                        parsed_file
                        date (date of the parsing)
                        database_description
                        database_type: List of which kind of data the database inserts
        """

        if database_type=="all":
            database_type=None

        current_sources = self.db.select_db_content( PianaGlobals.SelectValidSourceDBs.get_sqlquery(database_type=database_type)
                                                     , answer_mode="raw", remove_duplicates="no" )

        # return a dictionary with all the results in the table of databases
        # the key is the database identifier. Each record has a dictionary with the information
        current_sources_dict = {}

        for actual_source in current_sources:
            current_database_type = self.db.select_db_content( PianaGlobals.SelectDatabaseType.get_sqlquery(database_name=actual_source[0])
                                                               , answer_mode="list", remove_duplicates="no" )
            current_sources_dict[actual_source[0]] = {"databaseVersion": actual_source[1],
                                                      "parsed_file": actual_source[2],
                                                      "date": actual_source[3],
                                                      "databaseDescription": actual_source[4],
                                                      "database_type": current_database_type,
                                                      "databaseInteger": int(actual_source[5])}
            

        return current_sources_dict


        # OLD CODE:
        #return dict([(i[0],[j for j in i[1:]]) for i in current_sources])


      
##     def _load_ints_dbs_mask(self):
##         """
##         returns dictionary int_dbs_mask (see above for description)
##         """

##         # get databases with interactions
##         #  keys are database names, content is a dictionary with keys databaseVersion, parsed_file, date, databaseDescription, database_type, databaseInteger

##         interaction_dbs_dic = self.get_source_dbs_dict(database_type="protein-protein interactions")

##         #print len(interaction_dbs_dic)+1

##         interaction_dbs_dic = {}
        
##         # generate the dictionary with all combinations of databases
##         list_combinations = []
##         for one_size in range(1, len(interaction_dbs_dic)+1):
##             for c in utilities.xcombinations(interaction_dbs_dic.keys(), one_size):
##                 #print "appending for %s" %(one_size)
##                 list_combinations.append(c)

##         dict_to_return = {}

##         for one_combination in list_combinations:
##             this_int = 0
##             for one_db in one_combination:
##                 this_int += interaction_dbs_dic[one_db]["databaseInteger"]
            
##             dict_to_return[this_int] = one_combination

        
##         return dict_to_return
      

    # --------------------------------------------------------------------------
    # Methods used to update PIANA database
    # --------------------------------------------------------------------------

    def insert_new_id_type(self, proteinTypeId, proteinTable=PianaGlobals.proteinExternalId_table, externalIdColumn=PianaGlobals.proteinExternalId_col,
                           externalIdDescription=None):
        """
        Inserts into PIANA database a new external Id identifier

        First, it will check if this external Type Id was previously in the database: if it wasn't, insert it. Otherwise, error

        "proteinTable" is the SQL table where protein ids of "proteinTypeId" are kept
        "externalIdColumn" is the SQL column of "proteinTable" where protein ids of "proteinTypeId" are kept

        """

        # First: check if this external Id was previously in the database
        # If it is found, check if the parameters are the same as in the method parameters: if yes, don't worry
        # If the parameters are different, raise ValueError
        if proteinTypeId is None:
            raise ValueError("proteinTypeId cannot be None!")
        if self.validProteinIds.has_key(proteinTypeId):
            sys.stderr.write("This type of external identifier existed previously in the database! All parameters are going to be rewritten\n")
        
        self.db.insert_db_content( PianaGlobals.InsertNewProteinIdType.get_sqlquery(proteinTypeId=proteinTypeId,
                                                                                    proteinTable=proteinTable,
                                                                                    externalIdColumn=externalIdColumn,
                                                                                    externalIdDescription=externalIdDescription),
                                   answer_mode=None, buffer=None )

        # Add this proteinType in the current available protein types id stored in memory:
        self._add_valid_protein_id(proteinTypeId=proteinTypeId,
                                   table=proteinTable,
                                   column = externalIdColumn,
                                   description = externalIdDescription)
        
        # Update externalId table
        self.db.insert_db_content( PianaGlobals.UpdateENUMcolumn.get_sqlquery(table=PianaGlobals.proteinExternalId_table,
                                                                             column= PianaGlobals.externalIdType_col,
                                                                             list_enum= self.get_list_valid_protein_ids()),
                                   answer_mode=None, buffer=None )
        

    def insert_new_external_database(self,databaseName=None, databaseVersion=None, parsedFile=None, databaseDescription=None, databaseInformation=None):
        """
        Inserts into database the information of a new external database that is being integrated into PIANA database

        "databaseName" will be the internal identifier for this database, so, it must be unique.

        "pianaDB" is the name of the PIANA database where the new protein id type is inserted. It is used to generate automatic documentation on current database

        "databaseInformation" is a list of the type of information that the database inserts into PIANA database

        ATTENTION! Information_type must be one element of the valid types:

                 "protein sequences"
                 "protein attributes"
                 "protein-protein interactions"
                 "identifiers cross-references"
        """

        valid_information_types = {"protein sequences": None,
                                   "protein attributes": None,
                                   "identifiers cross-references": None,
                                   "protein-protein interactions": None}

        # Check the type of information given to the database

        if databaseInformation is not None:
            databaseInformation = databaseInformation.split(',')
            for actual_db_inf in databaseInformation:
                if not valid_information_types.has_key(actual_db_inf.strip().strip('"').strip("'")):
                    raise ValueError("Not valid database information. %s not recognized: It must be an element of the following list: %s" %(actual_db_inf,
                                                                                                                                            valid_information_types.keys()) )


        #self.lock_tables()

        # First: check if this databaseName is already in the database
        if not self.validSources.has_key(databaseName):
            date = time.localtime()
            actual_date = "%s-%s-%s" %(date[0],date[1],date[2])


            # Obtain the integer identifier that will be assigned to the database
            # First, obtain which identifiers are being used at the moment
            dict_current_integers={}
            for actual_key in self.validSources.keys():
                dict_current_integers[self.validSources[actual_key]]=None

            new_database_identifier = -1
            # Search the integer that will be assigned to this database
            for i in xrange(65):
                if not dict_current_integers.has_key(2**i):
                    new_database_identifier = 2**i
                    break

            sys.stderr.write("Database identifier: %s\n" %(new_database_identifier))
                    
            if new_database_identifier==-1:
                raise ValueError("Impossible to get an identifier for the database. Maybe are you putting more than 64 databases?")
            
            self.db.insert_db_content( PianaGlobals.InsertNewExternalDatabase.get_sqlquery(databaseName=databaseName.lower(),
                                                                                           databaseVersion=databaseVersion,
                                                                                           databaseFile=parsedFile,
                                                                                           databaseDate=actual_date,
                                                                                           databaseDescription=databaseDescription,
                                                                                           databaseInteger=new_database_identifier),
                                       answer_mode=None, buffer=None )

            # Add this external database in the current external databases on memory
            #self._add_valid_source_db(databaseName=databaseName)
            self.validSources[databaseName]=new_database_identifier


            # Delete from the SET datatype the "unknown" entry, if it is found
            if self.validSources.has_key("unknown"):
                del self.validSources["unknown"]


            # NOT NECESSARY NOW (TOO SLOW BECAUSE IT HAD TO UPDATE ALL TABLES... Now, it is processed in piana source code)
            # UPDATE ALL TABLES WHICH HAVE THE FIELD PianaGlobals.sourceDBID_col
            #sys.stderr.write("Updating SET datatypes...\n")
            #self._update_sourceDB_set()
            #sys.stderr.write("Done!\n")

            #self.unlock_tables()

            # Insert the type of information this database gives to PIANA database (previously, at the top of this method, it has been checked if
            # all information_type were correct
            if databaseInformation is not None:
                for actual_db_inf in databaseInformation:
                    self.add_external_database_information_type(externalDatabase=databaseName,information_type=actual_db_inf.strip().strip('"').strip("'"))

        else:
            raise ValueError("Database Name already existed in the database: try to change it")

    def _get_table_names(self):

        table_list =  self.db.select_db_content( PianaGlobals.GetTableNames.get_sqlquery(), answer_mode="list" )

        return table_list

    def _get_table_columns(self,table=None):

        if table is None:
            raise ValueError("table cannot be None! You must specify a database table")
        else:
            column_list = self.db.select_db_content( PianaGlobals.GetTableDescription.get_sqlquery(tableName=table), answer_mode="list" )

            return column_list

    def get_list_valid_protein_ids(self):
        """
        returns a sorted list with valid protein identifiers
        """

        valid_protein_ids = self.validProteinIds.keys()
        valid_protein_ids.sort()

        return valid_protein_ids

    def _update_sourceDB_set(self, sourceDBInteger):
        """
        Method used to update all available sourceDB's in piana Database, as this field is a SET datatype in the database
        
        """

        # Obtain all tables in the database
        tables_list = self._get_table_names()

        # Search those tables that have the PianaGlobals.sourceDBID_col field and update this field
        for current_table in tables_list:
            column_list = self._get_table_columns(current_table)

            for current_column in column_list:
                if current_column == PianaGlobals.sourceDBID_col:
                    self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery( table= current_table,
                                                                                           field= current_column,
                                                                                           value= sourceDBInteger,
                                                                                           type="remove"),
                                               answer_mode=None)
                    #self.db.insert_db_content( PianaGlobals.UpdateSourceDBSet.get_sqlquery( table= current_table,
                    #                                                                        field= current_column,
                    #                                                                        values = self.get_list_valid_sources()),
                    #                           answer_mode=None)


    def drop_multiple_external_databases(self,list_databases,pianaDB):
        """
        Method used to delete from the database all records from multiple external database
        """

       
        valid_sources = self.get_valid_sources()

        # Check databases and obtain the integer associated to them
        joined_identifier=0
        for actual_database in list_databases:
            if not valid_sources.has_key(actual_database):
                raise ValueError("Database identifier %s does not exist in the database" %(actual_database))
            else:
                joined_identifier += self._get_sourceDBID_int(actual_database)
                # delete those databases from current databases stored on memory
                del valid_sources[actual_database]

        
        self.lock_tables()


        # Obtain all tables in the databas
        tables_list = self._get_table_names()
        
        # Search those tables that have the PianaGlobals.sourceDBID_col field and update this field
        # to delete those records that have no sourceDBID
        for current_table in tables_list:
            sys.stderr.write("updating table %s" %(current_table))
            column_list = self._get_table_columns(current_table)
            
            for current_column in column_list:
                if current_column == PianaGlobals.sourceDBID_col:
                    # Update SourceDBID
                    sys.stderr.write("updating...")
                    self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery( table= current_table,
                                                                                           field= current_column,
                                                                                           value = joined_identifier,
                                                                                           type="remove" ),
                                               answer_mode=None)
                    
                    # Delete those records that have any sourceDBID
                    sys.stderr.write("deleting...")
                    self.db.insert_db_content( PianaGlobals.DeleteRecords.get_sqlquery( table= current_table,
                                                                                        field= current_column,
                                                                                        value=0),
                                               answer_mode=None)

        self.unlock_tables()




        # Updates proteinPianaCounter (to the maximum value of proteinPiana)
        maximum_proteinPiana = self.db.select_db_content( PianaGlobals.SelectMaxProteinPiana.get_sqlquery(), answer_mode="single")
        self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(int(maximum_proteinPiana)+1,PianaGlobals.proteinPianaCounter_col),
                                   answer_mode=None )

        # Delete from databases_table the reference to those databases
        for actual_database in list_databases:
            sys.stderr.write("deleting database reference %s" %(actual_database))
            self.db.insert_db_content( PianaGlobals.DeleteExternalDatabase.get_sqlquery(databaseName=actual_database),
                                       answer_mode=None)


    def get_max_proteinPiana(self):
        """
        Method that returns the maximum value of proteinPiana in the database
        """
        
        return self.db.select_db_content( PianaGlobals.SelectMaxProteinPiana.get_sqlquery(), answer_mode="single")
    

    def drop_external_database(self,database_name,pianaDB):
        """
        Method used to delete from the database all records from an external database
        """

       
        valid_sources = self.get_valid_sources()

        
        if not valid_sources.has_key(database_name):
            raise ValueError("This database identifier does not exist in the database")
        

        self.lock_tables()

        
        # NOT NECESSARY NOW
        # Updates SET datatypes
        #self._update_sourceDB_set(sourceDBInteger=self._get_sourceDBID_int(database_name))
        
        # Delete from databases table the reference to this database
        sys.stderr.write("deleting database")
        self.db.insert_db_content( PianaGlobals.DeleteExternalDatabase.get_sqlquery(databaseName=database_name),
                                   answer_mode=None)

        
        # Deletes from all tables in the database those records that does not have any sourceDB
        # Obtain all tables in the databas
        tables_list = self._get_table_names()
        
        # Search those tables that have the PianaGlobals.sourceDBID_col field and update this field
        # to delete those records that have no sourceDBID
        for current_table in tables_list:
            sys.stderr.write("updating table %s" %(current_table))
            column_list = self._get_table_columns(current_table)
            
            for current_column in column_list:
                if current_column == PianaGlobals.sourceDBID_col:
                    # Update SourceDBID
                    sys.stderr.write("updating...")
                    self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery( table= current_table,
                                                                                           field= current_column,
                                                                                           value = self._get_sourceDBID_int(database_name),
                                                                                           type="remove" ),
                                               answer_mode=None)
                    
                    # Delete those records that have any sourceDBID
                    sys.stderr.write("deleting...")
                    self.db.insert_db_content( PianaGlobals.DeleteRecords.get_sqlquery( table= current_table,
                                                                                        field= current_column,
                                                                                        value=0),
                                               answer_mode=None)

        self.unlock_tables()


        # Updates current databases file
        # Save current protein id types in the file
        #current_databases_fd = file(PianaGlobals.database_log_path+"README."+pianaDB+PianaGlobals.databases_file_suffix,'w')
        #current_databases_fd.write("%s" %(self.get_current_databases_string()) )
        #current_databases_fd.close()

        # delete this database from current databases stored on memory
        del valid_sources[database_name]

        # Updates proteinPianaCounter (to the maximum value of proteinPiana)

        #maximum_proteinPiana = self.db.select_db_content( PianaGlobals.SelectMaxProteinPiana.get_sqlquery(), answer_mode="single")
        
        #self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(int(maximum_proteinPiana)+1,PianaGlobals.proteinPianaCounter_col),
        #                           answer_mode=None )
        
        


    #javi:
    def get_temporalSimilarity_data(self):
        """
        Get all protein-protein similarities from database

        TO CHANGE: It now reads from temporalProteinSimilarity, but it must read it directly from search...
        """

        return self.db.select_db_content( "SELECT * FROM temporalProteinSimilarity", answer_mode="raw" )

    #javi:
    def get_all_proteinSimilarity(self,sourceDB_value):
        """
        Returns a tuple of tuples of similar proteins (proteinPianaA,proteinPianaB) for a given database
        """
        return self.db.select_db_content( PianaGlobals.SelectAllSimilarProteins.get_sqlquery( sourceDB = self._get_sourceDBID_int(sourceDB_value))
                                          , answer_mode="raw", remove_duplicates="no") 
                                                                                              

    # General methods:
    def get_complete_table(self,table,attributes=[]):
        """
        Get the complete content of table "table" for "attributes" values, without removing duplicates
        """

        return self.db.select_db_content( PianaGlobals.SelectCompleteTable.get_sqlquery( table = table,
                                                                                         attributes = attributes ),
                                          answer_mode="raw", remove_duplicates="no")

    
                   
    # ---------------------------------------
    # General internal methods
    # ---------------------------------------

    def _check_table_values(self,tableName=None,columns=[],values=[]):
        """
        General method to search in the database if a given table as an entry with determined values

        returns the number of rows that have this combination of values

        Not used for the moment

        """

        if len(columns)!=len(values) or len(columns)==0:
            raise ValueError("The length of columns and values must be the same and cannot be 0!")
        elif tableName is None:
            raise ValueError("tableName cannot be None!")
        else:
            number_rows = self.db.select_db_content( PianaGlobals.SelectColumnValues.get_sqlquery(tableName=tableName,
                                                                                                  tableColumns=columns,
                                                                                                  values=values),
                                                     answer_mode = "single")

            return number_rows
                                                                                                  
    # ----------------------------------------
    # Access methods to PianaGlobals variables
    # ----------------------------------------

    def get_from_dict(self, dict_name, description_value):
        """
        returns ID for a given description of a dictionarized entity to find its "description_value"
        description_value must exist in the dictionary PianaGlobals.dict_name, otherwise nothin will be returned
        """
        
        for entity_id in dict_name.keys():

            normalized_description = description_value.lower()
            
            if normalized_description in dict_name[entity_id]:

                if verbose:
                    sys.stderr.write(" <%s> normalized is equal to <%s>\n " %(normalized_description, entity_id))

                return entity_id


        return "unknown"

    def get_list_valid_sources(self):
        """
        returns a sorted list with valid database sources
        """

        current_sources = self.validSources.keys()
        current_sources.sort()

        return current_sources

    def get_valid_sources(self):
        """
        returns a dictionary with valid external database sources
        """

        return self.validSources


    def get_methodID(self, methodDescription_value):
        """
        returns methodID for a given description of a method to find interactions "methodDescription_value"

        methodDescription_value must exist in dictionary PianaGlobals.method_names, otherwise nothing will be returned

        
        """
        if not self.validInteractionMethods:
            # Check first if the dictionary is empty
            for entity_id in PianaGlobals.method_names.keys():
                for external_id in PianaGlobals.method_names[entity_id]:                    
                    self.validInteractionMethods[external_id.lower()] = entity_id
                    
        try:
            return self.validInteractionMethods[methodDescription_value.lower()]
        except KeyError:
            return "unknown"
            

        #methodID = self.get_from_dict(dict_name= PianaGlobals.method_names, description_value= methodDescription_value)
        #return methodID


#    def get_sourceDBID(self, sourceDBDescription_value):
#        """
#        return sourceDBID for a given description of an external source ("sourceDBDescription_value")
#        sourceDBDescription_value must exist in dictionary PianaGlobals.source_databases, otherwise nothing will be returned
#        """
#        #sourceDBID = self.get_from_dict(dict_name= PianaGlobals.source_databases, description_value= sourceDBDescription_value)
#        sourceDBID = self.get_from_dict(dict_name= self.validSources, description_value= sourceDBDescription_value)
#        
#        return sourceDBID
    
    def get_interaction_source_database_color(self, database_name):
        """
        returns the color code established in PianaGlobals for database_name
        """
        if PianaGlobals.interaction_source_databases_colors.has_key(database_name):
            return PianaGlobals.interaction_source_databases_colors[database_name]
        else:
            return PianaGlobals.interaction_source_databases_colors[""]
    
    def get_interaction_line_style(self, interaction_type):
        """
        returns the line style established in PianaGlobals for interaction_type

        interaction_type can be:
             - normal
             - expanded
        """
        return PianaGlobals.interaction_line_styles[interaction_type]
    
    def get_node_fill_color(self, node_type):
        """
        returns the color code established in PianaGlobals for node_type

        node_type can be:
             - root
             - normal
        """
        return PianaGlobals.node_fill_colors[node_type]
    
    def get_node_border_color(self, node_origin):
        """
        returns the color code established in PianaGlobals for node_origin

        node_type can be:
             - expanded
             - normal
        """
        return PianaGlobals.node_border_colors[node_origin]
   
    # ------------------------------------
    # Access (insertions and retrievals) methods to protein tables
    # ------------------------------------

    def update_table_column(self, proteinPiana= None, table= None, column= None, new_value= None):
        """
        updates column "column" of table "table" where proteinPiana="proteinPiana" with value "new_value"

        Used to update values in protein tables where the unique identifier in proteinPiana. it will change the current value
        in the column with the new value provided (see its use in update_sequence_ip.py)
        
        """

        # we use insert_d_content althought we are updating: the way of executing/retrieving the sql query is identical
        # 
        number_rows_updated = self.db.insert_db_content( PianaGlobals.UpdateTableColumn.get_sqlquery(proteinPiana= proteinPiana,
                                                                                                     table= table, column= column,
                                                                                                     new_value= new_value),
                                                         answer_mode="num_updated" )

        return number_rows_updated

    def get_all_proteinPiana(self, tax_id_value=0):
        """
        returns all proteinPianas in pianaDB of species tax_id_value (if 0, returns all proteinPianas in database)
        
        """
        return self.db.select_db_content( PianaGlobals.SelectAllProteinPiana.get_sqlquery(tax_id_value=tax_id_value)
                                          , answer_mode="list", remove_duplicates="yes" )


    def get_list_proteinPiana(self, protein_id= None, id_type= None, tax_id_value=0, source_db_info="no", only_primary="default" ):
        """
        Method used to retrieve a list of proteinPiana identifiers from a given protein "protein_id" of type "id_type"

        For example, one geneName "mot1" could correspond to many proteinPianas, since different databases give different sequences to mot1
        Internally, we have to work with all proteinPianas, and then give the answer to the user in the type of code he chose
        
        There can be many proteinPianas [id, id, id] or just one [id] or cero []. 

        valid "id_type" are those valid protein identifiers in current database.
                        For types "geneName" and "uniprotAcc", only will be returned

        "tax_id_value" fixes the species for the proteinPiana that will be returned. If tax_id_value is 0, all proteinPianas associated to
                       protein_id will be returned.
                       In case you don't know the species of your code, set tax_id_value to 0. If your code implicitly points to a single
                       species, this will be OK.

                       In which cases tax_id_value has an effect on the proteinPianas that are returned?
                       For example, if a geneName is associated to several proteinPianas, only those that are of tax_id_value will be returned.

        "source_db_info" determines if information about the sourceDB that inserted the proteinPiana is returned or not
           - "no" will simply return a list of proteinPianas
           - "yes" will return a list of tuples (proteinPiana, sourceDBID)

        if protein_code being passed is a pdb code, the format of the code must be pdb_code.chain_id
        -> If the chain_id is None, write pdb_code. (leaving the dot to mark that chain is unknown)

        "only_primary" determines, for those cases in which external codes can be primary or not (geneName and uniprotAcc), if only
                       primary correspondences are taken or not:
                       - "yes" will return only those proteinPianas that correspond to the code when this is primary.
                       - "yes-no" will return those proteinPianas that correspond to the code when this is primary. If any code is found, then search all for this
                       - "no" will return all proteinPianas, regarless if external code is primary or not
                       - "default": will be:
                                    "yes-no" for geneName and uniprotAccession (uniacc)
                                    "no" for the rest

        """


        if protein_id is None or id_type is None:
            raise ValueError("trying to get list of protein pianas for code (%s) and/or type (%s) None" %(protein_id, id_type ))

	id_sqlcolumn = self.get_column_of_id_type(id_type)  # this method gets the column from the id type
	id_sqltable = self.get_table_of_id_type(id_type)    # this method gets the table from the id type
        
        if id_sqlcolumn == PianaGlobals.sequence_col :
            
            # if the type passed is a sequence, fasten up things by setting the code to md5. To do so, 
            # get the md5 code for the fasta (using package utilities from /piana/code/utilities/utilities.py)
            #remove unwanted characters from sequence
            cleaned_proteinSequence_value = utilities.get_clean_sequence(input_sequence=protein_id)
            protein_id = utilities.sequence2md5(cleaned_proteinSequence_value)
            id_sqlcolumn = PianaGlobals.sequenceMD5_col
            
        elif id_sqlcolumn == PianaGlobals.proteinPiana_col:
            # the type passed is a proteinPiana... a little strange since the user is asking to get a proteinPiana from a proteinPiana
            # but maybe it makes sense for automated methods... this line makes this trasparent by returning the same value as long() in a list
            if source_db_info == "yes":
                return [long(protein_id), "protein"]
            else:
                return [long(protein_id)]
        # END OF elif id_type == PianaGlobals.proteinPiana_col: (if id_type == PianaGlobals.proteinSequence_col : )
            
        if verbose:
            sys.stderr.write("Calling SelectProteinPiana with protein_id %s and id_type %s\n" %(protein_id,
                                                                                                id_type))

        # set how many items are being selected, which depends on whethers source db info is requested or not
        if source_db_info == "no":   selected_items= 1
        else:                        selected_items= 2

        # determines what kind of search will be done in function of external primary value:
        if only_primary == "default":
            if id_type == "uniacc" or id_type == "geneName":
                isPrimary = 1
                only_primary = "yes-no"
            else:
                isPrimary = 0
        elif only_primary == "yes":
            isPrimary = 1
        elif only_primary == "no":
            isPrimary = 0
        elif only_primary == "yes-no":
            isPrimary = 1
        else:
            raise ValueError("only_primary can only be \"yes\",\"no\", or \"yes-no\"")

        
        # Specifies if id_type is generalized or not (i.e. if the generalized table is used or not)
        # because of it is used, the type_id is necessary to select the correct correspondences
        if id_sqltable != PianaGlobals.proteinExternalId_table:
            id_type = None
        
        # if tax_id_value is 0 , call to standard SelectProteinPiana. Otherwise, fix a taxonomy by calling SelectProteinPianaTax
        if tax_id_value == 0:
            list_proteinPiana_value = self.db.select_db_content( PianaGlobals.SelectProteinPiana.get_sqlquery(protein_code= protein_id,
                                                                                                              protein_code_column= id_sqlcolumn,
													      protein_code_table = id_sqltable,
                                                                                                              source_db_info= source_db_info,
                                                                                                              isPrimary= isPrimary,
                                                                                                              id_type = id_type)
                                                                 , answer_mode="list", remove_duplicates="no",
                                                                 number_of_selected_elems= selected_items )
            
            #sys.stderr.write("List of proteinPianas: %s\n" %(list_proteinPiana_value) )
            
        else:
            list_proteinPiana_value = self.db.select_db_content( PianaGlobals.SelectProteinPianaTax.get_sqlquery(protein_code= protein_id,
														 protein_code_column=id_sqlcolumn,
														 protein_code_table=id_sqltable,
                                                                                                                 taxonomy_value= tax_id_value,
                                                                                                                 source_db_info= source_db_info,
                                                                                                                 isPrimary= isPrimary,
                                                                                                                 id_type= id_type)
                                                                 , answer_mode="list", remove_duplicates="no",
                                                                 number_of_selected_elems= selected_items  )

        # END OF else: (if tax_id_value == 0)

        # Check if any proteinPiana has been found for this external code, in case the user wants to get first primary, and if not found, all
        if only_primary == "yes-no":
            # No proteinPianas found:
            if len(list_proteinPiana_value)==0:
                # Search with no restriction of primaryValue:
                # if tax_id_value is 0 , call to standard SelectProteinPiana. Otherwise, fix a taxonomy by calling SelectProteinPianaTax
                if tax_id_value == 0:
                    list_proteinPiana_value = self.db.select_db_content( PianaGlobals.SelectProteinPiana.get_sqlquery(protein_code= protein_id,
                                                                                                                      protein_code_column= id_sqlcolumn,
                                                                                                                      protein_code_table = id_sqltable,
                                                                                                                      source_db_info= source_db_info,
                                                                                                                      isPrimary=0,
                                                                                                                      id_type= id_type)
                                                                         , answer_mode="list", remove_duplicates="no",
                                                                         number_of_selected_elems= selected_items )
            
            
            
                else:
                    list_proteinPiana_value = self.db.select_db_content( PianaGlobals.SelectProteinPianaTax.get_sqlquery(protein_code= protein_id,
                                                                                                                         protein_code_column=id_sqlcolumn,
                                                                                                                         protein_code_table=id_sqltable,
                                                                                                                         taxonomy_value= tax_id_value,
                                                                                                                         source_db_info= source_db_info,
                                                                                                                         isPrimary=0,
                                                                                                                         id_type = id_type)
                                                                         , answer_mode="list", remove_duplicates="no",
                                                                         number_of_selected_elems= selected_items  )
                    
                # END OF else: (if tax_id_value == 0)

        return list_proteinPiana_value


    def get_proteinPianas_with_value(self, value, proteinCode_table, proteinCode_col):
        """
        returns proteinPianas whose value for column "proteinCode_col" in table "proteinCode_table" is "value"
        """
        return self.db.select_db_content( PianaGlobals.SelectProteinPianaGeneralized.get_sqlquery(proteinCode_table=proteinCode_table,
                                                                                                  proteinCode_col= proteinCode_col,
                                                                                                  code_value= value)
                                          , answer_mode="list", remove_duplicates="no",
                                          number_of_selected_elems= 1 )
    
    def get_new_protein_piana(self):
        """
        Method used to retrieve a new proteinPiana identifier from table proteinPianaCounter

        Called every time we want to insert a new (sequence,taxID) in the database
        
        It takes care of increasing by 1 the counter table

        proteinPiana is just an internal identifier used to uniquely identify proteins: each sequence is a different proteinPiana
        """

        if self.insert_buffer:
            self.newproteinPiana += 1
            return self.newproteinPiana
        else:
            # obtaining proteinPiana value from counter table
            new_proteinPiana_value = self.db.select_db_content( PianaGlobals.SelectNewProteinPiana.get_sqlquery(),
                                                                answer_mode="single" )
        
            # updating counter table with next proteinPiana value to be returned
            updated_proteinPiana_value = new_proteinPiana_value + 1
        
            self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(updated_proteinPiana_value,PianaGlobals.proteinPianaCounter_col),
                                       answer_mode=None )
        
            return new_proteinPiana_value

    def get_new_interaction_piana(self):
        """
        Method used to retrieve a new interactionPiana identifier from table proteinPianaCounter

        Called every time we want to insert a new interaction in the database
        
        It takes care of increasing by 1 the counter table

        proteinPiana is just an internal identifier used to uniquely identify proteins: each sequence is a different proteinPiana
        """

        if self.insert_buffer:
            self.newInteractionPiana += 1
            return self.newInteractionPiana
        else:
            # obtaining proteinPiana value from counter table
            new_interactionPiana_value = self.db.select_db_content( PianaGlobals.SelectNewInteractionPiana.get_sqlquery(),
                                                                    answer_mode="single" )
        
            # updating counter table with next proteinPiana value to be returned
            updated_interactionPiana_value = new_interactionPiana_value + 1
        
            self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(updated_proteinPiana_value,PianaGlobals.interactionCounter_col),
                                       answer_mode=None )
        
            return new_interactionPiana_value

    
    def get_new_sequenceID(self):
        """
        Method used to retrieve a new sequenceID identifier from table proteinPianaCounter

        Called every time we want to insert a new sequence in the database
        
        It takes care of increasing by 1 the counter table

        proteinPiana is just an internal identifier used to uniquely identify proteins: each sequence is a different proteinPiana
        """

        if self.insert_buffer:
            self.newSequenceID += 1
            return self.newSequenceID
        else:
            # obtaining sequenceID value from counter table
            new_sequenceID_value = self.db.select_db_content( PianaGlobals.SelectNewSequenceID.get_sqlquery(),
                                                              answer_mode="single" )
        
            # updating counter table with next proteinPiana value to be returned
            updated_sequenceID_value = new_sequenceID_value + 1
        
            self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(updated_sequenceID_value, PianaGlobals.sequenceIDCounter_col),
                                       answer_mode=None )
        
            return new_sequenceID_value


    def check_protein(self, proteinSequence_value, tax_id_value= None):
        """
        Method used to check if a protein is already in the database

        Returns a proteinPiana if the protein exists, or None if not exists

        Since pianaDB is a sequence-based DB, introducing proteinSequence_value is mandatory.

        There is one proteinPiana for each (sequence, tax_id): therefore, tax_id is mandatory. However, when
        the user doesn't know the tax id for the protein, we use a dummy tax_id (ie. 0) to allow searching
        that sequence in the database. Therefore, if the tax_id is unknown, leave it to None.
        """
        
        if verbose:
            sys.stderr.write("converting sequence to md5\n")
            
        #remove unwanted characters from sequence
        cleaned_proteinSequence_value = utilities.get_clean_sequence(input_sequence= proteinSequence_value)

        # get the md5 code for the fasta (using package utilities from /piana/code/utilities/utilities.py)
        proteinMD5_value = utilities.sequence2md5(cleaned_proteinSequence_value)

        list_proteinPiana_value = self.db.select_db_content(
            PianaGlobals.SelectProteinPianaFromTaxID_and_MD5.get_sqlquery(sequenceMD5_value = proteinMD5_value,
                                                                          tax_id_value = tax_id_value),
            answer_mode="list", remove_duplicates="yes")

        if not list_proteinPiana_value:
            proteinPiana_value = None
            
        elif len(list_proteinPiana_value) == 1:
            proteinPiana_value= list_proteinPiana_value[0]
            
        else:
            raise ValueError("how can we have more than one proteinPiana for the same (sequence, tax id)? Error!")

        return proteinPiana_value
        


    def insert_protein(self, proteinSequence_value, sourceDBID, tax_id_value= None):
        """
        Method used to insert new proteins into pianaDB in `protein` table.
        
        Returns a proteinPiana: it is the code corresponding to the  (sequence, tax_id)
                                  --> if (sequence, tax) didn't exist already, creates a new proteinPiana
                                  --> if the sequence is already present in pianaDB, returns the previous proteinPiana
        --------------------------------------------------------------------------------------------------------------

        Since pianaDB is a sequence-based DB, introducing proteinSequence_value is mandatory.

        There is one proteinPiana for each (sequence, tax_id): therefore, tax_id is mandatory. However, when
        the user doesn't know the tax id for the protein, we use a dummy tax_id (ie. 0) to allow inserting
        that sequence. Therefore, if the tax_id is unknown, leave it to None.
        
        We use MD5 codes (calculated here) instead of sequences to fasten up the process of comparing sequences

        MW and IP values for the protein are calculated using BioPython methods

        This method takes care of handling proteinPiana codes:
        
          1. It first looks for existence of the (sequence, tax_id) in table protein,
          which is the registry of correspondences between (protein sequence, tax id) and proteinPiana identifiers. We need to keep this
          registry in order to make sure that proteinPiana identifiers do not change when updating the database, or building a
          new one from scratch.

          2. If the (sequence, tax_id) does not exist in the database, then obtain a new proteinPiana identifier. This is done with a method
          that looks into a counter table, returns its value and increases the counter for the next proteinPiana identifier.

          3. Once proteinPiana is known (newly generated, or  assigned from an old insertion) insert the protein into the database

          

        """
        # TO DO!!!!!! How do I keep the correspondences between proteinPianas of different piana dbs?
        #             With a registry file (/data/proteinCorrespondence/....) updated??? Writing to the file whenever I insert a
        #             new protein? Doing a database dump (select proteinPiana, proteinMD5) each time I want to create a new database?
        #             --> Do I need to keep the correspondences between different piana dbs... maybe not...

        if verbose:
            sys.stderr.write("converting sequence to md5\n")

        #remove unwanted characters from sequence
        cleaned_proteinSequence_value = utilities.get_clean_sequence(input_sequence= proteinSequence_value)
        
        # get the md5 code for the fasta (using package utilities from /piana/code/utilities/utilities.py)
        proteinMD5_value = utilities.sequence2md5(cleaned_proteinSequence_value)

        # Once we have the MD5 for the sequence, search if this sequence existed before in the database
        sequenceID = self.db.select_db_content( PianaGlobals.SelectSequenceIDFromMD5.get_sqlquery(sequenceMD5_value = proteinMD5_value),
                                                answer_mode="single", remove_duplicates="no")

        # Check if the sequence is in the buffer in case it is not found in the database
        if sequenceID is None and self.insert_buffer:
            try:
                sequenceID = self.insert_sequence_buffer[proteinMD5_value]
            except KeyError:
                sequenceID = None


        # If this sequence did not exist previously, calculate MW and IP values and insert it into the database
                
        if sequenceID is None:
            # calculate the length
            proteinSequenceLength_value = len(cleaned_proteinSequence_value)
        
            if verbose:
                sys.stderr.write("creating biopython protein analyzer\n")
            
            # creating a BioPython protein analyzer
            analyzed_protein = Bio.SeqUtils.ProtParam.ProteinAnalysis(cleaned_proteinSequence_value)

            if verbose:
                sys.stderr.write("calculating molecular weight\n")
                
            try:
                # calculate the molecular weight
                proteinMW_value = analyzed_protein.molecular_weight()
            except:
                # any error in the function sets weight to 0
                proteinMW_value = 0

            if verbose:
                sys.stderr.write("calculating isoelectric point\n")
            
            try:
                # this is a modified version of isoelectric_point
                #
                # I have done two main things to the program:
                #
                #    - introduce an attribute correction_step in SeqUtils.ProtParam.isoelectric_point()
                #         def isoelectric_point(self, correction_step= 0.001):
                #
                #    - rewrite SeqUtils.IsoelectricPoint by modifying the step-wise correction of the charge so
                #      it never gets blocked. The modification I've made has to be done on file
                #      python2.X/site-packages/Bio/SeqUtils/IsoelectricPoint.py
                #
                #      I keep a copy of the modified code under piana/code/utilities/modified_IsoelectricPoint.py
                
                # TO CHECK!!! Does the latest version of biopython need these modifications? as of version 2.3, it does...
                proteinIP_value= analyzed_protein.isoelectric_point(correction_step = 0.001)
            except:
                # any error in the function sets isoelectric point to 0
                proteinIP_value = 0

            # Get the new sequenceID value:
            sequenceID = self.get_new_sequenceID()
        
            
            # Insert the sequence in the database
            self.db.insert_db_content( PianaGlobals.InsertSequence.get_sqlquery( sequenceID_value = sequenceID,
                                                                                 sequence_value = cleaned_proteinSequence_value,
                                                                                 sequenceMD5_value = proteinMD5_value,
                                                                                 sequenceLength_value = proteinSequenceLength_value,
                                                                                 sequenceMW_value = proteinMW_value,
                                                                                 sequenceIP_value = proteinIP_value,
                                                                                 sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                 buffer =  self.insert_buffer),
                                       answer_mode="None", buffer=  self.insert_buffer)

            # Add the relation proteinMD5_value:sequenceID to internal buffer
            if self.insert_buffer:
                self.insert_sequence_buffer[proteinMD5_value] = sequenceID

            # END OF if sequenceID is None
        else:
            # If sequenceID identifier existed previously, update sourceDB for this sequence
            self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.sequence_table,
                                                                                     column_values = [(PianaGlobals.sequenceID_col,sequenceID)],
                                                                                     sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                     db_temp_buffer = self.db_temp_buffer,
                                                                                     conditions = [PianaGlobals.sequenceID_col],
                                                                                     buffer = self.insert_buffer ),
                                       answer_mode=None, buffer= self.insert_buffer)

        # END OF if sequenceID is None




        

        if verbose:
            sys.stderr.write("initiating insertion of information\n")

        if not tax_id_value:
            # if no tax_id given, set a dummy value to represent that sequence under any species
            tax_id_value = 0

        # 1. look for existence of the (sequenceID, tax_id) in table protein
        #    to check for errors, try to get a list of proteinPiana... then take only one element

        # If we are using buffer, we only have to check in the temporal dictionary
        # Check if the protein is in the temporal dictionary if it is not in the database(due to it may be in the internal buffer)


        # Search the proteinPiana that corresponds to (sequenceID,taxID) combination in the database
        list_proteinPiana_value = self.db.select_db_content(
            PianaGlobals.SelectProteinPianaFromTaxID_and_sequenceID.get_sqlquery(sequenceID_value = sequenceID,
                                                                                 tax_id_value = tax_id_value),
            answer_mode="list", remove_duplicates="yes")

        if not list_proteinPiana_value:
            proteinPiana_value = None
            
        elif len(list_proteinPiana_value) == 1:
            proteinPiana_value= list_proteinPiana_value[0]
            
        else:
            raise ValueError("how can we have more than one proteinPiana for the same (sequence, tax id)? Error!")

        # If (sequnceID,taxID) is not found and insert buffer is being used, search it
        if proteinPiana_value is None and self.insert_buffer:
            try:
                proteinPiana_value = self.insert_protein_buffer[(sequenceID,tax_id_value)]
            except KeyError:
                proteinPiana_value = None

        # If proteinPiana identifier existed previously, update sourceDB for this protein
        if proteinPiana_value is not None:
            self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.protein_table,
                                                                                     column_values = [(PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                     sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                     db_temp_buffer = self.db_temp_buffer,
                                                                                     conditions = [PianaGlobals.proteinPiana_col],
                                                                                     buffer = self.insert_buffer ),
                                       answer_mode=None, buffer= self.insert_buffer)

        #  If the (sequence, tax_id) does not exist in the database (nor in the buffer), then obtain a new proteinPiana identifier.
        #  and insert it into the database.
        else:

            if verbose:
                sys.stderr.write("calculating length\n")
            
            # Get the new proteinPiana Identifer
            proteinPiana_value = self.get_new_protein_piana()
        
            # If it is a newly generated proteinPiana insert it into the database
            self.db.insert_db_content(
                PianaGlobals.InsertProtein.get_sqlquery(proteinPiana_value = proteinPiana_value,
                                                        tax_id_value = tax_id_value,
                                                        sequenceID_value = sequenceID,
                                                        sourceDBID=self._get_sourceDBID_int(sourceDBID),
                                                        buffer = self.insert_buffer)
                , answer_mode=None, buffer=self.insert_buffer )

            # If the buffer is being used, insert the proteinPiana into the temporal dictionary
            if self.insert_buffer:
                self.insert_protein_buffer[(sequenceID,tax_id_value)] = proteinPiana_value

            # END OF if self.insert_buffer
            

        # END OF if proteinPiana_value is None:

        return proteinPiana_value


    def insert_protein_similarity(self, proteinPiana_a_value, proteinPiana_b_value, sourceDB_value ):
        """
        Insert a pair of proteinPianas that are in fact the "same" protein (they are sufficiently similar to be considered the same
        for some situations)

        This is used to avoid comparing two proteins that are in fact the same.

        InsertProteinSimilarity makes sure that the order proteinPianaA < proteinPianaB is respected
        """

        # make sure proteinPianaA < proteinPianaB
        if proteinPiana_a_value > proteinPiana_b_value:
            temp = proteinPiana_a_value
            proteinPiana_a_value = proteinPiana_b_value
            proteinPiana_b_value = temp
                  
        #self.db.insert_db_content( PianaGlobals.InsertProteinSimilarity.get_sqlquery(proteinPianaA_value=proteinPiana_a_value,
        #                                                                             proteinPianaB_value=proteinPiana_b_value,
        #                                                                             source_db_value = self._get_sourceDBID_int(sourceDB_value),
        #                                                                             buffer = self.insert_buffer)
        #                           , answer_mode=None, buffer = self.insert_buffer )

        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinSimilarity_table,
                                                                                 column_values = [(PianaGlobals.proteinPianaA_col,proteinPiana_a_value),
                                                                                                  (PianaGlobals.proteinPianaB_col,proteinPiana_b_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(sourceDB_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.proteinPianaA_col,PianaGlobals.proteinPianaB_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer )

        

    def check_proteins_similarity(self, proteinPiana_a_value, proteinPiana_b_value ):
        """
        Returns 1 if the proteins are the same (ie. there is an entry proteinPiana_a, proteinPiana_b in table proteinSimilarity)
        """
        similar_proteins = self.db.select_db_content(PianaGlobals.SelectProteinSimilarity.get_sqlquery(proteinPianaA_value=proteinPiana_a_value,
                                                                                                       proteinPianaB_value=proteinPiana_b_value
                                                                                                       ),
                                                     answer_mode ="single", number_of_selected_elems=2)

        if similar_proteins is not None:
            return 1
        else:
            return 0

    def get_similar_proteins_dic(self, proteinPiana_value):
        """
        Returns a dictionary with keys those proteins that are similar to proteinPiana_value

        """
        similar_proteins_dic = {}

        similar_proteins_higher = self.db.select_db_content(PianaGlobals.SelectSimilarProteins.get_sqlquery(proteinPiana_value=proteinPiana_value,
                                                                                                            mode="higher"),
                                                     answer_mode ="list")
        
        similar_proteins_lower = self.db.select_db_content(PianaGlobals.SelectSimilarProteins.get_sqlquery(proteinPiana_value=proteinPiana_value,
                                                                                                            mode="lower"),
                                                     answer_mode ="list")

        for protein in similar_proteins_higher:
            similar_proteins_dic[protein] = None
            
        for protein in similar_proteins_lower:
            similar_proteins_dic[protein] = None

        return similar_proteins_dic
                  
    # ------------------------------------
    # Access (insertions and retrievals) methods to codes tables
    # ------------------------------------

    # --
    # insert methods (Access methods to codes tables)
    # --
            
    def insert_uniprotEntry(self, uniprotEntry, proteinPiana_value, sourceDBID ):
        """
        Insert correspondence between  "uniprotEntry" and "proteinPiana_value" in table uniprotEntry

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.uniprotEntry_table,
                                                                                 column_values = [(PianaGlobals.uniprotEntry_col,uniprotEntry),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.uniprotEntry_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        
                                                                                 
        
        #num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.uniprotEntry_table,
        #                                                                                    field= PianaGlobals.sourceDBID_col,
        #                                                                                    value= self._get_sourceDBID_int(sourceDBID),
        #                                                                                    conditions=[(PianaGlobals.uniprotEntry_col,uniprotEntry),
        #                                                                                                (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
        #                                         answer_mode="num_updated" )
        # If any entry has been updated, insert it into the database
        #if num_updated is None:
        #    self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.uniprotEntry_table,
        #                                                                                 code_value= uniprotEntry,
        #                                                                                 proteinPiana_value= proteinPiana_value,
        #                                                                                 #codeSource_value= self.get_sourceDBID(sourceDBID),
        #                                                                                 codeSource_value = sourceDBID,
        #                                                                                 isPrimary_value = None,
        #                                                                                 buffer = self.insert_buffer )
        #                              , answer_mode=None, buffer=self.insert_buffer )
        
    def insert_uniprotAcc(self, uniprotAcc_value, proteinPiana_value,sourceDBID , isPrimary_value ):
        """
        Insert correspondence between  "uniprotAcc_value" and "proteinPiana_value" in table uniprotAcc

        Insert as well source database (to keep origin database of that code) 
        
        isPrimary_value indicates whether it is the primary accession code or not
        
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.uniprotAcc_table,
                                                                                 column_values = [(PianaGlobals.uniprotAcc_col,uniprotAcc_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value),
                                                                                                  (PianaGlobals.isPrimary_col,isPrimary_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.uniprotAcc_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        
        #num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.uniprotAcc_table,
        #                                                                                    field= PianaGlobals.sourceDBID_col,
        #                                                                                    value= self._get_sourceDBID_int(sourceDBID),
        #                                                                                    conditions =[ (PianaGlobals.uniprotAcc_col,uniprotAcc_value),
        #                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
        #                                         answer_mode="num_updated" )

        # If any entry has been updated, insert it into the database
        #if num_updated is None:
        #    self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.uniprotAcc_table,
        #                                                                                 code_value= uniprotAcc_value,
        #                                                                                 proteinPiana_value= proteinPiana_value,
        #                                                                                 #codeSource_value= self.get_sourceDBID(sourceDBID),
        #                                                                                 codeSource_value = sourceDBID,
        #                                                                                 isPrimary_value= isPrimary_value,
        #                                                                                 buffer = self.insert_buffer )
        #                               , answer_mode=None, buffer= self.insert_buffer )
        
        
    def insert_geneName_code(self, geneName_code_value, proteinPiana_value, geneName_source_value, isPrimary_value ):
        """
        Insert correspondence between  geneName "geneName_code_value" and "proteinPiana_value" in table geneName

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.geneName_table,
                                                                                 column_values = [(PianaGlobals.geneName_col,str(geneName_code_value).upper()),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value),
                                                                                                  (PianaGlobals.isPrimary_col,isPrimary_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(geneName_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.geneName_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        ## num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.geneName_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(geneName_source_value),
##                                                                                             conditions =[ (PianaGlobals.geneName_col,geneName_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.geneName_table,
##                                                                                          code_value= geneName_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(geneName_source_value),
##                                                                                          codeSource_value = geneName_source_value,
##                                                                                          isPrimary_value= isPrimary_value,
##                                                                                          buffer = self.insert_buffer)
##                                        , answer_mode=None, buffer = self.insert_buffer )
        

    def insert_pdb_code(self, pdb_code_value, proteinPiana_value, chain_value, range_value, pdb_source_value ):
        """
        Insert correspondence between pdbs and "proteinPiana_value" in table pdb

        Internally, the pdb code is formed by "pdb_code_value" + "." + "chain_value"

        Insert as well source database (to keep origin database of that code) 
        
        """

        # column pdb_chain will always follow format pdb_code.chain, where chain will never be None
        #   if the input chain is None, we will write an A instead (convention)
        combined_pdb_chain = pdb_code_value

        if chain_value != "_" and chain_value != "" and chain_value is not None:
            combined_pdb_chain = combined_pdb_chain + "." + chain_value
        else:
            combined_pdb_chain = combined_pdb_chain + ".A"

        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.pdb_table,
                                                                                 column_values = [(PianaGlobals.pdb_chain_col,combined_pdb_chain),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(pdb_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.pdb_chain_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
       ##  num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.pdb_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(pdb_source_value),
##                                                                                             conditions =[ (PianaGlobals.pdb_chain_col,combined_pdb_chain),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinPdbCode.get_sqlquery(table_name= PianaGlobals.pdb_table,
##                                                                                       code_value= pdb_code_value,
##                                                                                       proteinPiana_value= proteinPiana_value,
##                                                                                       chain_value= chain_value,
##                                                                                       #codeSource_value= self.get_sourceDBID(pdb_source_value),
##                                                                                       codeSource_value = pdb_source_value,
##                                                                                       combined_pdb_chain = combined_pdb_chain,
##                                                                                       range_value = range_value,
##                                                                                       buffer= self.insert_buffer)
##                                        , answer_mode=None,buffer= self.insert_buffer )


    def insert_gi_code(self, gi_code_value, proteinPiana_value, gi_source_value):
        """
        Insert correspondence between  gi  "gi_code_value" and "proteinPiana_value" in table gi

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.gi_table,
                                                                                 column_values = [(PianaGlobals.giID_col,gi_code_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(gi_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.giID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        
       ##  num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.gi_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(gi_source_value),
##                                                                                             conditions =[ (PianaGlobals.giID_col,gi_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.gi_table,
##                                                                                          code_value= gi_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(gi_source_value),
##                                                                                          codeSource_value = gi_source_value,
##                                                                                          buffer= self.insert_buffer )
##                                        , answer_mode=None, buffer= self.insert_buffer )

    def insert_accessionNumber(self, accessionNumber_value, proteinPiana_value, accessionNumber_source_value ):
        """
        Insert correspondence between  accessionNumber  "accessionNumber_value" and "proteinPiana_value" in table accessionNumber

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.accessionNumber_table,
                                                                                 column_values = [(PianaGlobals.accessionNumber_col,accessionNumber_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(accessionNumber_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.accessionNumber_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
       ##  num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.accessionNumber_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(accessionNumber_source_value),
##                                                                                             conditions =[ (PianaGlobals.accessionNumber_col,accessionNumber_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.accessionNumber_table,
##                                                                                          code_value= accessionNumber_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(accessionNumber_source_value),
##                                                                                          codeSource_value = accessionNumber_source_value,
##                                                                                          buffer= self.insert_buffer)
##                                        , answer_mode=None, buffer= self.insert_buffer )

    def insert_geneID_code(self, geneID_code_value, proteinPiana_value, geneID_source_value):
        """
        Insert correspondence between  geneID  "geneID_code_value" and "proteinPiana_value" in table geneID

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.geneID_table,
                                                                                 column_values = [(PianaGlobals.geneID_col,geneID_code_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(geneID_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.geneID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        ## num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.geneID_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(geneID_source_value),
##                                                                                             conditions =[ (PianaGlobals.geneID_col,geneID_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.geneID_table,
##                                                                                          code_value= geneID_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(geneID_source_value),
##                                                                                          codeSource_value = geneID_source_value,
##                                                                                          buffer= self.insert_buffer)
##                                        , answer_mode=None, buffer= self.insert_buffer )


    def insert_unigene_code(self, unigene_code_value, proteinPiana_value, unigene_source_value ):
        """
        Insert correspondence between  unigene  "unigene_code_value" and "proteinPiana_value" in table unigene

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.unigene_table,
                                                                                 column_values = [(PianaGlobals.unigene_col,unigene_code_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(unigene_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.unigene_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
       ##  num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.unigene_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(unigene_source_value),
##                                                                                             conditions =[ (PianaGlobals.unigene_col,unigene_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.unigene_table,
##                                                                                          code_value= unigene_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(unigene_source_value),
##                                                                                          codeSource_value = unigene_source_value,
##                                                                                          buffer= self.insert_buffer)
##                                        , answer_mode=None, buffer= self.insert_buffer )


    def insert_refseq_code(self, refseq_code_value, proteinPiana_value, refseq_source_value):
        """
        Insert correspondence between  refseq  "refseq_code_value" and "proteinPiana_value" in table refseq

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.refseq_table,
                                                                                 column_values = [(PianaGlobals.refseq_col,refseq_code_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(refseq_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.refseq_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
       ##  num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.refseq_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(refseq_source_value),
##                                                                                             conditions =[ (PianaGlobals.refseq_col,refseq_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.refseq_table,
##                                                                                          code_value= refseq_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(refseq_source_value),
##                                                                                          codeSource_value = refseq_source_value,
##                                                                                          buffer=self.insert_buffer  )
##                                        , answer_mode=None, buffer=self.insert_buffer )

    def insert_external_id_code(self, external_id_type, external_id_value, proteinPiana_value, sourceDBID):
        """
        Insert correspondence between external id "external_id_value" of type "external_id_type"
        """

        if not self.get_valid_protein_ids().has_key(external_id_type):
            raise ValueError("Trying to insert a protein identifier type into database that was not previously defined. Please, define it before")
        else:
            self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = self.get_table_of_id_type(external_id_type),
                                                                                     column_values = [(self.get_column_of_id_type(external_id_type),external_id_value),
                                                                                                      (PianaGlobals.proteinPiana_col,proteinPiana_value),
                                                                                                      (PianaGlobals.externalIdType_col,external_id_type)],
                                                                                     sourceDBID = self._get_sourceDBID_int(sourceDBID),
                                                                                     db_temp_buffer = self.db_temp_buffer,
                                                                                     conditions = [self.get_column_of_id_type(external_id_type),
                                                                                                   PianaGlobals.proteinPiana_col,
                                                                                                   PianaGlobals.externalIdType_col],
                                                                                     buffer = self.insert_buffer ),
                                       
                                       answer_mode=None, buffer=self.insert_buffer)
            ## num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= self.get_table_of_id_type(external_id_type),
##                                                                                                 field= PianaGlobals.sourceDBID_col,
##                                                                                                 value= self._get_sourceDBID_int(sourceDBID),
##                                                                                                 conditions =[ (PianaGlobals.proteinExternalId_col,external_id_value),
##                                                                                                               (PianaGlobals.proteinPiana_col,proteinPiana_value),
##                                                                                                               (PianaGlobals.externalIdType_col,external_id_type)]),
##                                                      answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertGeneralizedProteinExternalID.get_sqlquery(table_name= self.get_table_of_id_type(external_id_type),
##                                                                                                     external_id_type = external_id_type,
##                                                                                                     externalId_column = self.get_column_of_id_type(external_id_type),
##                                                                                                     code_value = external_id_value,
##                                                                                                     proteinPiana_value = proteinPiana_value,
##                                                                                                     codeSource_value = sourceDBID,
##                                                                                                     buffer=self.insert_buffer),
##                                        answer_mode=None, buffer=self.insert_buffer )



    def insert_ensembl_code(self, ensembl_code_value, proteinPiana_value, ensembl_source_value ):
        """
        Insert correspondence between  ensembl  "ensembl_code_value" and "proteinPiana_value" in table ensembl

        Insert as well source database (to keep origin database of that code) 
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.ensembl_table,
                                                                                 column_values = [(PianaGlobals.ensembl_col,ensembl_code_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(ensembl_source_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.ensembl_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   
                                   answer_mode=None, buffer=self.insert_buffer)
        ## num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.ensembl_table,
##                                                                                             field= PianaGlobals.sourceDBID_col,
##                                                                                             value= self._get_sourceDBID_int(ensembl_source_value),
##                                                                                             conditions =[ (PianaGlobals.ensembl_col,ensembl_code_value),
##                                                                                                           (PianaGlobals.proteinPiana_col,proteinPiana_value)]),
##                                                  answer_mode="num_updated" )
##         # If any entry has been updated, insert it into the database
##         if num_updated is None:
##             self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.ensembl_table,
##                                                                                          code_value= ensembl_code_value,
##                                                                                          proteinPiana_value= proteinPiana_value,
##                                                                                          codeSource_value = ensembl_source_value,
##                                                                                          #codeSource_value= self.get_sourceDBID(ensembl_source_value),
##                                                                                          buffer=self.insert_buffer)
##                                        , answer_mode=None, buffer=self.insert_buffer )
            
#    def insert_protein_id_intDB_code(self, protein_id_intDB_value, proteinPiana_value, intDB_source_value):
#        """
#        Insert correspondence between xxxxx and "proteinPiana_value" in table xxxxx

#        Insert as well source database (to keep origin database of that code) 
#        """
#        self.db.insert_db_content( PianaGlobals.InsertProteinExternalID.get_sqlquery(table_name= PianaGlobals.protein_id_intDB_table,
#                                                                                     code_value= protein_id_intDB_value,
#                                                                                     proteinPiana_value= proteinPiana_value,
#                                                                                     #codeSource_value= self.get_sourceDBID(intDB_source_value),
#                                                                                     codeSource_value = intDB_source_value,
#                                                                                     buffer = self.insert_buffer)
#                                   , answer_mode=None, buffer=self.insert_buffer )


    # --
    # get methods (Access methods to codes tables)
    # --

    def get_protein_mw(self, proteinPiana=None):
        """
        returns molecular weight of protein proteinPiana
        """
        # TO DO!!!! If not found in db, or == 0, it could be calculated here!!!! using biopython...

        #mw = self.db.select_db_content(PianaGlobals.SelectProteinTableColumn.get_sqlquery(proteinTable_column_value=PianaGlobals.proteinMW_col,
        #                                                                                    proteinPiana_value=proteinPiana
        #                                                                                    ),
        #                                 answer_mode ="single")
        mw = self.db.select_db_content(PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(sequenceTable_column_value=PianaGlobals.sequenceMW_col,
                                                                                                         proteinPiana_value=proteinPiana),
                                         answer_mode ="single")


        return mw
        

    def get_protein_ip(self, proteinPiana=None):
        """
        returns isoelectric point of protein proteinPiana
        """
        # TO DO!!!! If not found in db, or == 0, it could be calculated here!!!! using biopython...

        #ip = self.db.select_db_content(PianaGlobals.SelectProteinTableColumn.get_sqlquery(proteinTable_column_value=PianaGlobals.proteinIP_col,
        #                                                                                    proteinPiana_value=proteinPiana
        #                                                                                    ),
        #                                 answer_mode ="single")
        
        ip = self.db.select_db_content(PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(sequenceTable_column_value=PianaGlobals.sequenceIP_col,
                                                                                                         proteinPiana_value=proteinPiana
                                                                                                         ),
                                         answer_mode ="single")
        return ip
 
    def get_protein_sequence(self, proteinPiana= None):
        """
        returns  the sequence of protein proteinPiana
        """

        #sequence = self.db.select_db_content(PianaGlobals.SelectProteinTableColumn.get_sqlquery(proteinTable_column_value=PianaGlobals.proteinSequence_col,
        #                                                                                        proteinPiana_value=proteinPiana
        #                                                                                        ),
        #                                     answer_mode ="single")

        sequence = self.db.select_db_content(PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(sequenceTable_column_value=PianaGlobals.sequence_col,
                                                                                                               proteinPiana_value=proteinPiana
                                                                                                               ),
                                             answer_mode ="single")
        
        return sequence


    def get_MD5_and_sequence_from_sequenceID(self, sequenceID= None):
        """
        returns  the sequence that corresponds to sequenceID
        """

        sequence = self.db.select_db_content(PianaGlobals.SelectSequenceSequenceMD5FromSequenceID.get_sqlquery(sequenceID_value = sequenceID ),
                                             answer_mode ="raw")
        
        return sequence

    def get_sequence_from_sequenceID(self, sequenceID= None):
        """
        returns  the sequence that corresponds to sequenceID
        """

        sequence = self.db.select_db_content(PianaGlobals.SelectSequenceFromSequenceID.get_sqlquery(sequenceID_value = sequenceID ),
                                             answer_mode ="single")
        
        return sequence
    
   
    def get_protein_sequenceLength(self, proteinPiana_value = None):
        """
        returns the length of the sequence of protein proteinPiana_value
        """
        #length = self.db.select_db_content(PianaGlobals.SelectProteinTableColumn.get_sqlquery(proteinTable_column_value=PianaGlobals.proteinSequenceLength_col,
        #                                                                                      proteinPiana_value=proteinPiana_value
        #                                                                                      ),
        #                                   answer_mode ="single")

        length = self.db.select_db_content(PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(sequenceTable_column_value=PianaGlobals.sequenceLength_col,
                                                                                                             proteinPiana_value=proteinPiana_value
                                                                                                             ),
                                           answer_mode ="single")
        return length


    def get_list_proteinMD5(self):
        """
        returns a list with all distinct proteinMD5
        """
        return self.db.select_db_content(PianaGlobals.SelectAllDistinctMD5.get_sqlquery(),
                                         answer_mode ="list", remove_duplicates="no")


    def get_list_sequenceID(self,taxID=0):
        """
        returns a list with all distinct sequenceID

        "taxID" Tax ID of the specie from we want to obtain the list of sequenceID. It it is "0", it is not take into account
        """
        return self.db.select_db_content(PianaGlobals.SelectAllDistinctSequenceID.get_sqlquery(taxID=taxID),
                                         answer_mode ="list", remove_duplicates="no")


    def get_list_proteinPiana_from_MD5(self, md5_value = None):
        """
        returns a list of proteinPianas that correspond to a sequence (represented by its md5 code)
        """
        proteinPiana_list = []
        if md5_value is None:
            raise ValueError("md5 cannot be none")
        else:
            proteinPiana_list = self.db.select_db_content(PianaGlobals.SelectProteinPianaFromMD5.get_sqlquery(proteinMD5_value=md5_value),
                                             answer_mode = "list", remove_duplicates="no")

        return proteinPiana_list

    
        
    def get_list_protein_external_codes(self, proteinPiana=None,
                                        id_type=None, alternative_id_types=[], answer_mode= "list", source_db_info="no" ):
        """

        returns the list of external protein codes in type "id_type" that correspond to protein proteinPiana.
        
        If no code is found for "id_type" , it succesively tries with types listed in alternative_id_types
           -> if an alternative type name code is found, the list returned will contain strings of the form "id_type:external_code"
           -> it will only return the list of codes for the first alternative type found
           -> eg, alternative_id_types could be ["uniacc", "geneName", "md5"]

        "proteinPiana" is the internal piana code for the protein you want to get the external codes

        "id_type" is an easy-to-remember protein code type
           -> They are stored in valid_protein_ids in MySQL database, and can be consulted using a piana command
           -> valid id_type values are those in the PIANA reference card (read README.piana_tutorial)
           -> it can be "all", meaning that the list returned will contain all codes in all types (all codes will be preceded by type name)
               -> alternative_id_types are ignored if id_type is all

        "alternative_id_types" is a list of easy-to-remember protein code type
           for example, alternative_id_types could be ["uniacc", "gi", "proteinPiana"]
           I recomend writing always md5 at the end to make sure you don't have Nones in your output
              -> alternative_id_types are ignored if id_type is all

        "answer_mode" can be:
               - "single" (if you just want one external identifier)
               - "list" (if you want all external identifiers for that proteinPiana)
               - "list_primary" (if you want all external identifiers for that proteinPiana, but identifiers should be primary, unless none is found)
        
        "source_db_info" determines if information about the sourceDB that inserted the proteinPiana is returned or not
           - "no" will simply return a list of external_codes
           - "yes" will return a list of tuples (external_code, sourceDBID) (this mode does not allow alternative_id_types nor id_type=='all')

        Attention!!! This method does not work if you set "id_type" to 'all' and "answer_mode" to 'single'!!!!
        Attention!!! This method does not work if you ask for all external codes (id_type="all") with their source db infos (source_db_info="yes")!!!!
        """
           
        # TO DO!!! Not very intuitive to call it get_list_* and then allow the method to return only one code as an element


        if source_db_info == "yes" and id_type == "all":
            raise ValueError("this method does not allow asking for all external codes with their source db infos (can be changed, but has not been implemented")
        
        if answer_mode == "single" and id_type == "all":
            raise ValueError("it doesn't make sense to ask for all id types and single answer at the same time")

        if id_type != "all":
            # set column and table for the protein type name that will be used for output
	    proteinCode_table = self.get_table_of_id_type(id_type)
            proteinCode_col = self.get_column_of_id_type(id_type)

            # Check if the external code is sequence, sequenceMD5... and obtain the list with a different method
            if proteinCode_table == PianaGlobals.sequence_table:
                list_protein_external_codes = self.db.select_db_content( PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(
                    sequenceTable_column_value = proteinCode_col,
                    proteinPiana_value = proteinPiana),
                                                                         answer_mode = "list",
                                                                         remove_duplicates="no",
                                                                         )
            else:
                # Specifies if id_type can be generalized or not (i.e. if the generalized table is used or not)
                # When accessing the proteinExternalId table, we need to restrict the codes to be of this type)
                # because this table holds identifiers coming from many different sources
                if proteinCode_table != PianaGlobals.proteinExternalId_table:
                    id_type_restriction = None
                else:
                    id_type_restriction = id_type

                # obtain proteinPiana in the code desired by the user (stated as argument id_type to this method)
                list_protein_external_codes = self.get_code_value(proteinPiana_value = proteinPiana,
                                                                  proteinCode_table = proteinCode_table,
                                                                  proteinCode_col = proteinCode_col,
                                                                  answer_mode = answer_mode,
                                                                  source_db_info= source_db_info,
                                                                  id_type=id_type_restriction)
        else:
            # id_type is "all": set alternative_id_types to all possible type names (except for fasta, which is an alias to sequence)
            alternative_id_types = self.get_valid_protein_ids().keys()
            alternative_id_types.remove("fasta")
            list_protein_external_codes = []


        # if no external codes found for id_type (or it is equal to "all"), find other codes (from the list of alternative types)
        #                                       ---> the type of code will precede the code (eg: "proteinPiana:111111")
        if not list_protein_external_codes:

            # transform easy-to-remember type names to  pairs [table, column] for alternative types
            list_pair_table_column_alternative_id_types = []
            for alternative_id_type in alternative_id_types:
		list_pair_table_column_alternative_id_types.append( (alternative_id_type, [self.get_table_of_id_type(alternative_id_type),
                                                                                           self.get_column_of_id_type(alternative_id_type)]) ) 

            # find alternative codes
            for alternative_id_type in list_pair_table_column_alternative_id_types:
                # alternative_id_type is a tuple  ( id type, [proteincode_table, proteincode_col ])

                if alternative_id_type[1][0] == PianaGlobals.sequence_table:
                    # if the external Code is sequence, sequenceMD5... obtain the list with a different method

                    if answer_mode == "list" or answer_mode == "list_primary":
                        this_answer_mode= "list"
                    else:
                        this_answer_mode= "single"
                    
                    list_temp_external_code = self.db.select_db_content( PianaGlobals.SelectSequencePropertyFromProteinPiana.get_sqlquery(
                        sequenceTable_column_value = alternative_id_type[1][1],
                        proteinPiana_value = proteinPiana),
                                                                         answer_mode = this_answer_mode,
                                                                         remove_duplicates="no",
                                                                         )
                else:

                    # Specifies if id_type is generalized or not (i.e. if the generalized table is used or not)
                    # because of it is used, the type_id is necessary to select the correct correspondences
                    if alternative_id_type[1][0] != PianaGlobals.proteinExternalId_table:
                        id_type_restriction = None
                    else:
                        id_type_restriction = alternative_id_type[0]
                
                    list_temp_external_code= self.get_code_value(proteinPiana_value = proteinPiana,
                                                                 proteinCode_table = alternative_id_type[1][0],
                                                                 proteinCode_col = alternative_id_type[1][1],
                                                                 answer_mode = answer_mode,
                                                                 id_type=id_type_restriction)
                # END OF else: (if alternative_id_type[1][0] == PianaGlobals.sequence_table:)

                if list_temp_external_code:

                    if answer_mode == "list" or answer_mode == "list_primary":
                        # if something was found, append the protein codes (describing the type being used)
                        for temp_external_code in list_temp_external_code:
                            list_protein_external_codes.append( "%s:%s" %(alternative_id_type[0], temp_external_code))
                            
                    elif answer_mode == "single":
                        list_protein_external_codes = "%s:%s" %(alternative_id_type[0], list_temp_external_code)
                    else:
                        raise ValueError("unknown answer mode used in get_list_protein_external_codes: %s" %(answer_mode))

                    if id_type != "all":
                        # if mode not all, we only want the protein codes for the first alternative type found... exit loop
                        break
                # END OF if list_temp_external_code:
                
            # END OF for alternative_id_type in list_alternative_id_types:
        # END OF if not list_protein_external_codes:

        # althougth it says list_*, in answer_mode "single" it is a single element
        return list_protein_external_codes

       

    def get_code_value(self, proteinCode_table, proteinCode_col, proteinPiana_value= None, answer_mode = None, source_db_info="no", id_type=None):
        """
        Method used to retrieve a protein external code from column "proteinCode_col" of table "proteinCode_table"
        for a given  proteinPiana (it is a generalization to obtain external code from a given table that corresponds to protein proteinPiana)
        
        "answer_mode" can be
            - 'single' (if you just want one external code)
            - 'list' (if you want all external codes from that table)
            - 'list_primary' (if you want all external codes from that table, but the list should contain only primary codes, unless there were none)
        
        "source_db_info" determines if information about the sourceDB that inserted the proteinPiana is returned or not
           - "no" will simply return a list of proteinPianas
           - "yes" will return a list of tuples (proteinPiana, sourceDBID)

        in those cases where there is only one external code for "proteinPiana", "answer_mode" 'single' will return it as a single element and "list"
        as the first (and only) element of a list

        If nothing is found, the method will return None (in "answer_mode" "single") and an empty list (in "answer_mode" "list")

        "id_type" is only used when proteinCode_table is the generalized table

        """
        # TO CHECK! I don't remember the difference between this method and get_list_external_codes
        #           (apart from allowing the use of alternative_id_types)
        
        if proteinPiana_value is None or answer_mode is None:
            raise ValueError("proteinPiana (%s) and answer_mode (%s) cannot be None when trying to retrieve a external code" %(proteinPiana_value,
                                                                                                                               answer_mode))

        if source_db_info == "yes":      selected_items = 2
        else:                            selected_items = 1
            
        if proteinCode_col == PianaGlobals.proteinPiana_col:
            # in case the type of value asked is proteinPiana, just return same value
            # looks strange to return a value that has been passed by the user, but it has to be here to
            # be trasparent to the type of values being used
            if answer_mode == "single":
                if source_db_info == "yes":     code_value = (proteinPiana_value, "piana")
                else:                           code_value = proteinPiana_value
                    
            elif answer_mode == "list" or answer_mode == "list_primary":

                if source_db_info == "yes":     code_value = [ (proteinPiana_value, "piana") ]
                else:                           code_value = [proteinPiana_value]

        else:
            # all other identifiers are returned with the same SQL query

            if (answer_mode=="single" or answer_mode=="list_primary") and (proteinCode_table == PianaGlobals.uniprotAcc_table or proteinCode_table == PianaGlobals.geneName_table):
                return_primary = 1
            else:
                return_primary = 0

            if answer_mode=="list_primary":
                # in this case, we want a list of primary codes, but that was already fixed above in the previous if
                answer_mode = "list"
            
            code_value = self.db.select_db_content( PianaGlobals.SelectProteinID.get_sqlquery(proteinCode_table=proteinCode_table,
                                                                                              proteinCode_col=proteinCode_col,
                                                                                              proteinPiana_value=proteinPiana_value,
                                                                                              source_db_info= source_db_info,
                                                                                              return_primary=return_primary,
                                                                                              id_type=id_type),
                                                    answer_mode = answer_mode,
                                                    remove_duplicates="yes",
                                                    number_of_selected_elems= selected_items)

#            print "this is the list for %s returned first: %s" %(proteinCode_col, code_value)
            
            if return_primary and not code_value:
                # nothing was returned, probably because we demanded a primary id and there wasn't one
                code_value = self.db.select_db_content( PianaGlobals.SelectProteinID.get_sqlquery(proteinCode_table=proteinCode_table,
                                                                                                  proteinCode_col=proteinCode_col,
                                                                                                  proteinPiana_value=proteinPiana_value,
                                                                                                  source_db_info= source_db_info,
                                                                                                  return_primary=0,
                                                                                                  id_type=id_type),
                                                        answer_mode = answer_mode,
                                                        remove_duplicates="yes",
                                                        number_of_selected_elems= selected_items)

#                print "this is the list for %s returned second: %s" %(proteinCode_col, code_value)
                
        # END OF else: (if proteinCode_col == PianaGlobals.proteinPiana_col:)

        return code_value
    
    
    # ----------------------------------------------------------
    # Access (insertions and retrievals) methods to protein attributes relationship tables
    # ----------------------------------------------------------


    # --
    # insert methods
    # --

    def insert_protein_reactome(self, reactome_id_value, proteinPiana_value, reactome_pathwayname_value, proteinReactomeSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its associated Reactome ID "reactome_id"

        "reactome_id" is the reactome ID from Reactome

        "proteinPiana_value" is the internal piana identifier for the protein

        "reactome_pathwayname_value" is the Reactome Pathway name

        "proteinReactomeSource_value" is the external database that has set this correspondence (must appear in PianaGlobals.source_databases)
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinReactome_table,
                                                                                 column_values = [(PianaGlobals.reactomeID_col,reactome_id_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(proteinReactomeSource_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.reactomeID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer)

##         self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table= PianaGlobals.proteinReactome_table,
##                                                                           attribute_value= reactome_id_value,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           attributeSource_value= self.get_sourceDBID(proteinReactomeSource_value),
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinReactomeSource_value),
##                                                                           buffer=self.insert_buffer)
##             , answer_mode=None, buffer=self.insert_buffer)


        exists = self.db.select_db_content( PianaGlobals.SelectPathWayDescriptionFromReactomeID.get_sqlquery( reactomeID_value = reactome_id_value ), answer_mode="single" )

        if not exists:
            self.db.insert_db_content(
            PianaGlobals.InsertReactome.get_sqlquery( reactome_id = reactome_id_value,
                                                      reactomeDescription_value = reactome_pathwayname_value,
                                                      reactomeSource_value = self._get_sourceDBID_int(proteinReactomeSource_value)))
        else:
            # Check if the Description is the same... (it should be!)
            if exists != reactome_pathwayname_value:
                raise ValueError(" %s contains a different description [%s] from the inserted before in the database [%s]" %(reactome_id_value,
                                                                                                                             reactome_pathwayname_value,
                                                                                                                             exists) )
                    

##     def insert_protein_species(self, tax_id, proteinPiana_value, proteinSpeciesSource_value ):
##         """
##         Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its Species "tax_id"

##         "tax_id" is the taxonomy id from ncbi

##         "proteinPiana_value" is the internal piana identifier for the protein

##         "proteinSpeciesSource_value" is the external database that has set this correspondence (must appear in PianaGlobals.source_databases)
##         """

##         self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table= PianaGlobals.proteinSpecies_table,
##                                                                           attribute_value= tax_id,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           #attributeSource_value= self.get_sourceDBID(proteinSpeciesSource_value),
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinSpeciesSource_value),
##                                                                           buffer = self.insert_buffer)
##             , answer_mode=None, buffer = self.insert_buffer)

    def insert_protein_go(self, go_id, proteinPiana_value, proteinGoSource_value):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its GO (Gene Ontology) term id "go_id"

        "go_id" is the GO term id

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinGoSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinGo_table,
                                                                                 column_values = [(PianaGlobals.goID_col,go_id),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(proteinGoSource_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.goID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer)
        
##         self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinGo_table,
##                                                                           attribute_value= go_id,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           #attributeSource_value= self.get_sourceDBID(proteinGoSource_value),
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinGoSource_value),
##                                                                           buffer = self.insert_buffer )
##             , answer_mode=None, buffer = self.insert_buffer )
        
  
    def insert_protein_cog(self, cog_id, proteinPiana_value, proteinCogSource_value):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its COG (Cluster of Orthologous Genes) "cog_id"

        "cog_id" is the COG term id

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinCogSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinCog_table,
                                                                                 column_values = [(PianaGlobals.cogID_col,cog_id),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(proteinCogSource_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.cogID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer)
##         self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinCog_table,
##                                                                           attribute_value= cog_id,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinCogSource_value),
##                                                                           #attributeSource_value= self.get_sourceDBID(proteinCogSource_value),
##                                                                           buffer = self.insert_buffer)
##             , answer_mode=None, buffer=self.insert_buffer)
  
    def insert_protein_dbali_cluster(self, dbali_cluster_id_value , proteinPiana_value, clustering_method_value, patch_residues_value,
                                     protein_dbali_cluster_source):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its DBAli cluster "dbali_cluster_id_value"

        "dbali_cluster_id_value" is the cluster id given to DBAli to that protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "clustering_method_value" is the method followed by DBAli to establish the correspondence
          -> the method must be listed in PianaGlobals.pibase_dbali_methods

        "patch_residues_value" is the list of residues (string comma-separated) in the protein that correspond to that DBAli cluster 

        "protein_dbali_cluster_source" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinDBAliRelationship.get_sqlquery( dbali_cluster_id_value= dbali_cluster_id_value,
                                                                      proteinPiana_value= proteinPiana_value,
                                                                      clustering_method_value= clustering_method_value ,
                                                                      patch_residues_value=patch_residues_value,
                                                                      protein_dbali_cluster_source=  self._get_sourceDBID_int(protein_dbali_cluster_source))
        , answer_mode=None)
  
    def insert_protein_scop(self, cf, sf, fa, proteinPiana_value, proteinScopSource_value):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its SCOP "cf", "sf", "fa" values

        "cf" is the fold id 
        "sf" is the superfamily id 
        "fa" is the family id
        
        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinScopSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinScopRelationship.get_sqlquery( cf_value= cf, sf_value= sf, fa_value= fa, 
                                                                     proteinPiana_value= proteinPiana_value,
                                                                     #source_value= self.get_sourceDBID(proteinScopSource_value),
                                                                     source_value = self._get_sourceDBID_int(proteinScopSource_value),
                                                                     buffer = self.insert_buffer)
            , answer_mode=None, buffer = self.insert_buffer)
  

    def insert_protein_cath(self, cath_id_value, res_start_value, res_end_value, segmentID_value,  proteinPiana_value, proteinCathSource_value):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its CATH "cath_id_value"

        "cath_id_value" is the CATH id

        "res_start_value" is the residue where the CATH domain starts
        "res_end_value" is the residue where the CATH domain ends
        "segmentID_value" indicates which segment of the domain we are inserting (there can be several separate segments for one domain)

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinCathSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinCathRelationship.get_sqlquery( cathID_value= cath_id_value,
                                                                     res_start_value=res_start_value,
                                                                     res_end_value=res_end_value,
                                                                     segmentID_value=segmentID_value, 
                                                                     proteinPiana_value= proteinPiana_value,
                                                                     #source_value= self.get_sourceDBID(proteinCathSource_value) )
                                                                     source_value = self._get_sourceDBID_int(proteinCathSource_value) )
            , answer_mode=None)
    
    def insert_interPro_code(self, interProID_code_value, proteinPiana_value, interProDescription_value, interPro_source_value ):
        """
        Insert interPro information into table interPro (interProID and interProDescription)
        
        Insert as well source (to keep origin database of that code) and proteinPiana (to establish the relationship)
        
        """
        self.db.insert_db_content( PianaGlobals.InsertProteinCode2.get_sqlquery(table_name= PianaGlobals.interPro_table,
                                                                                code_value= interProID_code_value,
                                                                                proteinPiana_value= proteinPiana_value,
                                                                                description_value= interProDescription_value,
                                                                                codeSource_value= self._get_sourceDBID_int(interPro_source_value),
                                                                                buffer = self.insert_buffer)
                                   , answer_mode=None, buffer = self.insert_buffer )
    
    def insert_pfam_code(self, pfamID_code_value, proteinPiana_value, pfamDescription_value, pfam_source_value ):
        """
        Insert pfam information into table pfam (pfamID and pfamDescription)
        
        Insert as well source (to keep origin database of that code) and proteinPiana (to establish the relationship)
        
        """
        self.db.insert_db_content( PianaGlobals.InsertProteinCode2.get_sqlquery(table_name= PianaGlobals.pfam_table,
                                                                                code_value= pfamID_code_value,
                                                                                proteinPiana_value= proteinPiana_value,
                                                                                description_value= pfamDescription_value,
                                                                                codeSource_value= self._get_sourceDBID_int(pfam_source_value),
                                                                                buffer = self.insert_buffer)
                                   , answer_mode=None, buffer = self.insert_buffer )

    def insert_protein_ec(self, ec_id, proteinPiana_value, proteinECSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its EC code "ec_id"

        "ec_id" is the EC id

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinECSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinEC_table,
                                                                                 column_values = [(PianaGlobals.ecID_col,ec_id),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(proteinECSource_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.ecID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer)
##         self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinEC_table,
##                                                                           attribute_value= ec_id,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           #attributeSource_value= self.get_sourceDBID(proteinECSource_value),
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinECSource_value),
##                                                                           buffer = self.insert_buffer)
##             , answer_mode=None, buffer = self.insert_buffer)
        
    def insert_protein_description(self, description, proteinPiana_value, proteinDescriptionSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its description (text string)

        "description" is the text string describing the protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinDescriptionSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinDescription_table,
                                                                          attribute_value= description.rstrip("."),
                                                                          proteinPiana_value= proteinPiana_value,
                                                                          #attributeSource_value= self.get_sourceDBID(proteinDescriptionSource_value),
                                                                          attributeSource_value = self._get_sourceDBID_int(proteinDescriptionSource_value),
                                                                          buffer=self.insert_buffer)
            , answer_mode=None, buffer=self.insert_buffer)


    def insert_protein_mim(self, mimID_value, proteinPiana_value, proteinMIMSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its associated MIM ID

        "mimID_value" is the integer Identification Number in MIM database

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinMIMSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.proteinMIM_table,
                                                                                 column_values = [(PianaGlobals.mimID_col,mimID_value),
                                                                                                  (PianaGlobals.proteinPiana_col,proteinPiana_value)],
                                                                                 sourceDBID = self._get_sourceDBID_int(proteinMIMSource_value),
                                                                                 db_temp_buffer = self.db_temp_buffer,
                                                                                 conditions = [PianaGlobals.mimID_col,PianaGlobals.proteinPiana_col],
                                                                                 buffer = self.insert_buffer ),
                                   answer_mode=None, buffer= self.insert_buffer)
        
        ## self.db.insert_db_content(
##             PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinMIM_table,
##                                                                           attribute_value= mimID_value,
##                                                                           proteinPiana_value= proteinPiana_value,
##                                                                           #attributeSource_value= self.get_sourceDBID(proteinMIMSource_value),
##                                                                           attributeSource_value = self._get_sourceDBID_int(proteinMIMSource_value),
##                                                                           buffer=self.insert_buffer)
##             , answer_mode=None, buffer=self.insert_buffer)

        
    def insert_protein_function(self, function, proteinPiana_value, proteinFunctionSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its function (text string)

        "function" is the text string describing the function of the protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinFunctionSource_value" is the external database that has set this correspondence
        """

        self.db.insert_db_content(
            PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinFunction_table,
                                                                          attribute_value= function.rstrip("."),
                                                                          proteinPiana_value= proteinPiana_value,
                                                                          #attributeSource_value= self.get_sourceDBID(proteinFunctionSource_value),
                                                                          attributeSource_value = self._get_sourceDBID_int(proteinFunctionSource_value),
                                                                          buffer = self.insert_buffer)
            , answer_mode=None, buffer=self.insert_buffer)

    def insert_protein_disease(self, disease, proteinPiana_value, proteinDiseaseSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its associated disease (text string)

        "disease" is the text string describing the associated pathologies/disease with the protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinDiseaseSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinDisease_table,
                                                                          attribute_value= disease.rstrip("."),
                                                                          proteinPiana_value= proteinPiana_value,
                                                                          #attributeSource_value= self.get_sourceDBID(proteinDiseaseSource_value),
                                                                          attributeSource_value = self._get_sourceDBID_int(proteinDiseaseSource_value),
                                                                          buffer=self.insert_buffer)
            , answer_mode=None, buffer=self.insert_buffer )
        
    def insert_protein_keyword(self, keyword, proteinPiana_value, proteinKeywordSource_value ):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and an associated keyword

        "keyword" is the text string with a keyword associated to the protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinKeywordSource_value" is the external database that has set this correspondence
        """
        self.db.insert_db_content(
            PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinKeyword_table,
                                                                          attribute_value= keyword,
                                                                          proteinPiana_value= proteinPiana_value,
                                                                          #attributeSource_value= self.get_sourceDBID(proteinKeywordSource_value),
                                                                          attributeSource_value = self._get_sourceDBID_int(proteinKeywordSource_value),
                                                                          buffer= self.insert_buffer)
            , answer_mode=None, buffer= self.insert_buffer)
        
        
    def insert_protein_subcellularLocation(self, subcellularLocation, proteinPiana_value, proteinSubcellularLocationSource_value):
        """
        Method that inserts the correspondence between proteinPiana "proteinPiana_value" and its cellular location

        "subcellularLocation" is the text string with the cellular location of the protein

        "proteinPiana_value" is the internal piana identifier for the protein

        "proteinSubcellularLocationSource_value" is the external database that has set this correspondence
        """

        self.db.insert_db_content(
            PianaGlobals.InsertProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinSubcellularLocation_table,
                                                                          attribute_value= subcellularLocation.rstrip("."),
                                                                          proteinPiana_value= proteinPiana_value,
                                                                          #attributeSource_value= self.get_sourceDBID(proteinSubcellularLocationSource_value),
                                                                          attributeSource_value = self._get_sourceDBID_int(proteinSubcellularLocationSource_value),
                                                                          buffer=self.insert_buffer)
            , answer_mode=None, buffer=self.insert_buffer)

    # --
    # select methods
    # --

    def get_protein_taxonomy_ids(self, proteinPiana_value):
        """
        Returns a list with taxonomy ids (as defined by ncbi) related to "proteinPiana_value"
        """
        
##         list_protein_taxonomy_id = self.db.select_db_content(
##             PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table= PianaGlobals.proteinSpecies_table,
##                                                                           attribute_relationship_col= PianaGlobals.taxID_col,
##                                                                           proteinPiana_value= proteinPiana_value)
##         , answer_mode="list", remove_duplicates="yes")
        list_protein_taxonomy_id = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table= PianaGlobals.protein_table,
                                                                          attribute_relationship_col= PianaGlobals.taxID_col,
                                                                          proteinPiana_value= proteinPiana_value)
        , answer_mode="list", remove_duplicates="yes")

        return list_protein_taxonomy_id

    def get_protein_kingdoms(self, proteinPiana_value):
        """
        Returns a list with kingdoms (as defined by ncbi) related to "proteinPiana_value"
        """
        
        list_protein_kingdoms = self.db.select_db_content(
            PianaGlobals.SelectProteinKingdom.get_sqlquery( proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_protein_kingdoms


    def get_protein_species_names(self, proteinPiana_value):
        """
        Returns a list with species names (as defined by ncbi) related to "proteinPiana_value"
        """
        list_taxonomy_ids = self.get_protein_taxonomy_ids(proteinPiana_value=proteinPiana_value )
        
        protein_species_names= self.get_species_names_from_taxonomies(self, list_taxonomy_ids = list_taxonomy_ids)
        
        return protein_species_names

    
    def get_species_names_from_taxonomies(self, list_taxonomy_ids = [], only_scientific="no"):
        """
        Returns a list with species names that are related to the taxonomy ids in "list_taxonomy_ids"

        if only_scientific is 'yes', then the scienfific name is returned
        """
        list_species_names = []
        for taxonomy_id in list_taxonomy_ids:

            list_species_names.extend(self.db.select_db_content(PianaGlobals.SelectSpeciesNamefromTaxID.get_sqlquery(taxID_value = taxonomy_id,
                                                                                                                     only_scientific=only_scientific)
                                                                , answer_mode="list", remove_duplicates="yes") )


        # remove duplicates

        return list(Set(list_species_names))


    def get_taxonomies_from_species_name(self, species_name_value = None):
        """
        Returns a list with taxonomy ids for a given speciesName species_name_value
        """
        list_taxID = self.db.select_db_content(
            PianaGlobals.SelectTaxIDfromSpeciesName.get_sqlquery(speciesName_value = species_name_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_taxID
    
    def get_protein_go_accession(self, proteinPiana_value):
        """
        returns a list with all GO accessions (codes that look like "GO:234324") that are assigned to protein "proteinPiana_value"

        In case no GO term is found, returns an empty list.
        """
        # get go ids using self.get_protein_go_term_id
        list_term_id= self.get_protein_go_term_id(proteinPiana_value)
        list_go_accessions = []
        for term_id in list_term_id:
            list_go_accessions.extend(
                PianaGlobals.SelectXFromGoTermId.get_sqlquery(x_col= PianaGlobals.go_acc_col,
                                                              term_id_value= term_id))
                # TO DO!!! check that there are no repetitions in list_go_accessions
        # END OF for term_id in list_term_id:


        # remove duplicates
        dic_gos= {}
        for go in list_go_accessions:
            dic_gos[go] = None
        
        return dic_gos.keys()

    def get_protein_go_name(self, go_term_id_value):
        """
        returns  the GO name associated to "go_term_id_value"

        In case no GO term is found, returns None
        """
   
        return self.db.select_db_content(PianaGlobals.SelectXFromGoTermId.get_sqlquery(x_col= PianaGlobals.go_name_col,
                                                                                       term_id_value= go_term_id_value),
                                         answer_mode="single", remove_duplicates="yes")
       
    def get_protein_go_term_id(self, proteinPiana_value, term_type_value=None):
        """
        returns a list with all GO terms ids (internal go identifiers) that are assigned to protein "proteinPiana_value"

        In case no GO term is found, returns an empty list.

        "term_type_value" can be None or one of the following:

          - "molecular_function"
          - "biological_process"
          - "cellular_component"

        if term_type_value is None, then returns all go terms independently of their term type category
        """
        # TO CHECK!!!  One proteinPiana will return one or more term_id values?
        #             If yes (and the answer is clearly yes) get_proteins_sharing_go has to be modified

        if term_type_value is None:
            list_protein_go = self.db.select_db_content(
                PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table= PianaGlobals.proteinGo_table,
                                                                              attribute_relationship_col= PianaGlobals.goID_col,
                                                                              proteinPiana_value= proteinPiana_value)
                , answer_mode="list", remove_duplicates="yes")

        else:
            list_protein_go = self.db.select_db_content(PianaGlobals.SelectGoTermIdUsingTermType.get_sqlquery( proteinPiana=proteinPiana_value,
                                                                                                               term_type=term_type_value)
                                                        , answer_mode="list", remove_duplicates="yes")
            
        return list_protein_go

    def get_protein_go_term_name(self, proteinPiana_value, term_type_value=None):
        """
        returns a list with all GO terms names (e.g. Apoptosis) that are assigned to protein "proteinPiana_value"
        
        In case no GO term is found, returns an empty list.
        
        "term_type_value" can be None or one of the following:
        
        - "molecular_function"
        - "biological_process"
        - "cellular_component"
        
        if term_type_value is None, then returns all go terms independently of their term type category
        """

        protein_go_ids = self.get_protein_go_term_id(proteinPiana_value=proteinPiana_value, term_type_value=term_type_value)

        dic_go_names = {}
        for one_go_id in protein_go_ids:
            dic_go_names[self.get_protein_go_name(go_term_id_value=one_go_id)] = None

        return dic_go_names.keys()
        
        
    
    def get_proteins_sharing_go(self, proteinPiana_value):
        """
        Returns a list with proteinsPianas that have the same go_term_id as "proteinPiana_value"
        """
        # TO CHECK!!!  One proteinPiana will return one or more term_id values?
        #             If yes (and the answer is clearly yes) get_proteins_sharing_go has to be modified
        #             I think this has been solved, but I am not sure...

        # Getting go_term_id assigned to a proteinPiana
        list_term_id=self.get_protein_go_term_id(proteinPiana_value)
        list_protein_piana = []
        # List of proteinsPiana that has the same go_term_id as proteinPiana_value
        for go_term_id in list_term_id:
            
            list_protein_piana.extend(
                PianaGlobals.SelectXFromGoTermId.get_sqlquery(go_table=PianaGlobals.proteinGo_table,
                                                              x_col=PianaGlobals.goID_col,
                                                              term_id_value=go_term_id), answer_mode="list", remove_duplicates="yes")

        # remove duplicates
        dic_protein = {}
        for protein in list_protein_piana:
            dic_protein[protein] = None

        return dic_protein.keys()

    def get_go_depth(self, term_id_value):
        """
        Returns depth value in GO tree for "term_id_value"
        """

        if term_id_value is None:
            # if no term_id given (proteinPiana didn't have a go assigned), depth is considered huge
            # TO DO!!! create a constant for huge distances and huge depths
            return PianaGlobals.huge_depth

        return  self.db.select_db_content(PianaGlobals.SelectXFromGoTermId.get_sqlquery(x_col= PianaGlobals.go_distance2root,
                                                                                        term_id_value=term_id_value)
                                          , answer_mode="single", remove_duplicates="yes")


    def get_term2term_distance(self, term1, term2):
        """
        Returns distance in the GO tree between "term1" and "term2"
        """
        if term1 == term2:
            return 0
        
        distance= self.db.select_db_content(PianaGlobals.Select_go_term2term_distance.get_sqlquery(term1_id= term1,
                                                                                                   term2_id= term2)
                                            , answer_mode="single", remove_duplicates="yes")
        if distance:
            return distance
        else:
            return PianaGlobals.huge_distance

    
    def get_protein_cog(self, proteinPiana_value):
        """
        returns a list with all COG identifiers that are assigned to protein "proteinPiana_value"

        In case no COG identifier is found, returns an empty list.
        """

        protein_cog = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinCog_table,
                                                                          attribute_relationship_col= PianaGlobals.cogID_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")
        
        return protein_cog

         
    def get_all_protein_dbali_cluster(self, clustering_method):
        """
        Returns a list with all tuples (proteinPiana, dbali cluster ID) for a specific clustering method
        """

        protein_dbali_clusters = self.db.select_db_content(
            PianaGlobals.SelectAllDBAliClusterRelationship.get_sqlquery(clustering_method_value= clustering_method)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=2)
        
        return protein_dbali_clusters
       
    def get_protein_dbali_cluster(self, proteinPiana_value, clustering_method, source_db):
        """
        Returns a list with all dbali cluster IDs of a proteinPiana for a specific clustering method

        "source_db" can be 'dbali', 'blast_transfer' or 'all'. It can be used to limit the dbali_clusters returned to those found through a specific technique

        "clustering_method" refers to the parameters used for the clustering. Valid values are shown in PianaGlobals.pibase_dbali_methods
        """
        protein_dbali_clusters = self.db.select_db_content(
            PianaGlobals.SelectDBAliClusterRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value,
                                                                     clustering_method_value= clustering_method,
                                                                     source_db_value= source_db)
            , answer_mode="list", remove_duplicates="yes")
        
        return protein_dbali_clusters
       
    def get_pair_method_protein_dbali_cluster(self, proteinPiana_value):
        """
        Returns a list of tuples (dbali cluster ID, clustering_method) for protein "proteinPiana_value"
        """
        protein_dbali_clusters = self.db.select_db_content(
            PianaGlobals.SelectPairMethodDBAliClusterRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=2)
        
        return protein_dbali_clusters
        
    def get_proteins_with_scop(self, cf=None, sf=None, fa=None):
        """
        Returns a list of proteinPianas that have a given SCOP code

        if any of the three categories is None, that category is ignored
        if all categories are set, then this method impose that the proteins returned have those three categories
        """

        list_proteinPianas = self.db.select_db_content(
            PianaGlobals.SelectProteinsWithScop.get_sqlquery(cf= cf, sf=sf, fa=fa)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=1)
        
        return list_proteinPianas
        
    def get_protein_scop_cf_sf_fa(self, proteinPiana_value):
        """
        Returns a list of tuples (cf, sf, fa) for protein "proteinPiana_value"
        """
        
        list_protein_scop = self.db.select_db_content(
            PianaGlobals.SelectProteinScopRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
            
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=3)
        
        return list_protein_scop
    
    def get_protein_scop_cf(self, proteinPiana_value):
        """
        Returns a list of cf codes for protein "proteinPiana_value"
        """
        list_protein_scop_cf = self.db.select_db_content(
            PianaGlobals.SelectProteinScopCfRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=1)
        
        return list_protein_scop_cf
    
    def get_protein_scop_sf(self, proteinPiana_value):
        """
        Returns a list of sf codes for protein "proteinPiana_value"
        """
        list_protein_scop_sf = self.db.select_db_content(
            PianaGlobals.SelectProteinScopSfRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=1)
        
        return list_protein_scop_sf
    
    def get_protein_scop_fa(self, proteinPiana_value):
        """
        Returns a list of fa codes for protein "proteinPiana_value"
        """
        list_protein_scop_fa = self.db.select_db_content(
            PianaGlobals.SelectProteinScopFaRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=1)
        
        return list_protein_scop_fa
    
    def get_protein_cath(self, proteinPiana_value, residue_value=None):
        """
        Returns a list of cathIDs for protein "proteinPiana_value"
        
        If "residue_value" is not None, then it will return only caths that are defined around that residue
        If "residue_value" is None, returns all cath values for the protein
        """
        if residue_value is None:
            list_protein_cath = self.db.select_db_content(
                PianaGlobals.SelectProteinCathRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
                , answer_mode="list", remove_duplicates="yes")
            
            return list_protein_cath
        
        else:
            dict_protein_cath = {}

            list_protein_triplet_cath_resstart_resend = self.db.select_db_content(
                PianaGlobals.SelectProteinResiduesCathRelationship.get_sqlquery(proteinPiana_value= proteinPiana_value)
                , answer_mode="list", remove_duplicates="yes", number_of_selected_elems=3)

            for protein_triplet_cath_resstart_resend in list_protein_triplet_cath_resstart_resend:
                # Here, list_protein_triplet_cath_resstart_resend has
                #   [0] --> cath_id
                #   [1] --> res start
                #   [2] --> res end
                
                if residue_value > int(protein_triplet_cath_resstart_resend[1]) and residue_value < int(protein_triplet_cath_resstart_resend[2]):
                    if not dict_protein_cath.has_key(protein_triplet_cath_resstart_resend[0]):
                        dict_protein_cath[protein_triplet_cath_resstart_resend[0]] = None


            return dict_protein_cath.keys()
            
    def get_protein_ec(self, proteinPiana_value):
        """
        returns a list with all EC identifiers that are assigned to protein "proteinPiana_value"

        In case no EC identifier is found, returns an empty list.
        """

        protein_ec =  self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinEC_table,
                                                                          attribute_relationship_col= PianaGlobals.ecID_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return protein_ec
        
    def get_protein_description(self, proteinPiana_value):
        """
        returns a list with all descriptions (text string) that are assigned to protein "proteinPiana_value"

        In case no description is found, returns an empty list.
        """

        protein_description = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinDescription_table,
                                                                          attribute_relationship_col= PianaGlobals.proteinDescription_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return protein_description
        
    def get_protein_function(self, proteinPiana_value):
        """
        returns a list with all functions (text string) that are assigned to protein "proteinPiana_value"

        In case no function is found, returns an empty list.
        """

        protein_function = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinFunction_table,
                                                                          attribute_relationship_col= PianaGlobals.proteinFunction_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return protein_function
        
    def get_protein_keyword(self, proteinPiana_value):
        """
        returns a list with all keywords (text string) that are assigned to protein "proteinPiana_value"

        In case no keyword is found, returns an empty list.
        """

        protein_keyword = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinKeyword_table,
                                                                          attribute_relationship_col= PianaGlobals.proteinKeyword_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return protein_keyword

    def get_protein_reactome_ids(self, proteinPiana_value):
        """
        returns a list with all REACTOME ids (e.g. REACT_1698.1) that are assigned to protein "proteinPiana_value"

        In case no REACTOME is found, returns an empty list
        """
        protein_reactome = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinReactome_table,
                                                                          attribute_relationship_col= PianaGlobals.reactomeID_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return protein_reactome

    def get_protein_reactome_names(self, proteinPiana_value):
        """
        returns a list with all REACTOME names (e.g. Gene Expression) that are assigned to protein "proteinPiana_value"

        In case no REACTOME is found, returns an empty list
        """
        list_reactome_ids = self.get_protein_reactome_ids(proteinPiana_value=proteinPiana_value)

        list_reactome_names = self.get_reactome_names_from_reactome_ids(list_reactome_ids = list_reactome_ids)

        return list_reactome_names

    def get_reactome_names_from_reactome_ids(self, list_reactome_ids=[]):
        """
        returns the reactome names that are associated to reactome ids in list 
        """
        list_reactome_names = []
        for reactome_id in list_reactome_ids:

            list_reactome_names.extend(self.db.select_db_content(PianaGlobals.SelectReactomeNamefromReactomeID.get_sqlquery(reactome_id_value = reactome_id)
                                                                 , answer_mode="list", remove_duplicates="yes") )


        # remove duplicates

        return list(Set(list_reactome_names))


        
    def get_protein_subcellularLocation(self, proteinPiana_value):
        """
        returns a list with all cellular locations (text string) that are assigned to protein "proteinPiana_value"

        In case no keyword is found, returns an empty list.
        """
 
        protein_subcellularLocation = self.db.select_db_content(
            PianaGlobals.SelectProteinAttributeRelationship.get_sqlquery( attribute_relationship_table=PianaGlobals.proteinSubcellularLocation_table,
                                                                          attribute_relationship_col= PianaGlobals.proteinSubcellularLocation_col,
                                                                          proteinPiana_value= proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")
        
        return protein_subcellularLocation


    def check_keywords_in_protein(self, list_proteinPiana, protein_ext_code=None, id_type=None, keywords=[], search_in=None):
        """
        Checks all protein info [ie. description, function, keyword list] for keywords given by user

        Returns the list of words in "keywords" that appear somewhere in the protein. (empty if nothing found)

        "list_proteinPiana" is the list of proteins that you want to check (if it is just one protein, write [your_proteinPiana]

        You can also use external code types: in that case, set list_proteinPiana to None and set protein_code and id_type to something:

             "protein_code" is the protein external code
             "id_type" is the type of code used for protein_code
                --> it has to be one of the types listed in the PIANA reference card (read README.piana_tutorial)
                --> attention: not fixing tax id here, therefore, I recommend not using geneNames...
                
          
        "keywords" is a list of strings
           --> strings of keywords must be in lower case.
           --> This method will check that the keywords appear for the protein, regardless of the case of the words in the protein info
           --> for example, keywords could be ['cancer', 'onco', 'apoptosis']

        "search_in" is a list of texts where the keywords should be searched
                    if it is not used, description, function and keyword will be queried to database
        """

        # TO DO!!! allow the user to set the tax_id of protein_code
        #          or use only proteinPianas for checking keywords...
             
        if list_proteinPiana is None:
            list_proteinPiana = self.get_list_proteinPiana(protein_id= protein_ext_code,
                                                           id_type= id_type,
                                                           tax_id_value=0 )

        list_all_words = []

        if search_in is None:
            for proteinPiana in list_proteinPiana:
                list_all_words.extend(self.get_protein_description(proteinPiana_value=proteinPiana))
                list_all_words.extend(self.get_protein_function(proteinPiana_value=proteinPiana))
                list_all_words.extend(self.get_protein_keyword(proteinPiana_value=proteinPiana))
            # END OF for proteinPiana in list_proteinPiana:
        else:
            list_all_words.extend(search_in)

        # Now, check in all text fields obtained for the protein, which are the keywords that appear
        # and remove duplicates
        words_matched = {}
        for text_protein in list_all_words:
            lowered_text_protein = text_protein.lower()
            for keyword in keywords:
                if lowered_text_protein.find(keyword) != -1:
                    words_matched[keyword] = None

        return words_matched.keys()
        
    # -------------------------------------------
    # Access methods that retrieve proteins with a common characteristic to a given proteinPiana
    #
    # These methods are used by PianaGraphExpansion classes
    # -------------------------------------------

    def get_proteins_sharing_ec(self, proteinPiana_value ):
        """
        Returns a list of proteinPiana proteins that share an EC identifier with "proteinPiana_value"

        Empty list returned if nothing is found
        """

        if verbose:
            sys.stderr.write("Inside get_proteins_sharing_ec with proteinPiana= %s\n" %proteinPiana_value)

        list_proteins_sharing_ec = self.db.select_db_content(
            PianaGlobals.SelectCommonCodeProteins.get_sqlquery( proteinCode_table = PianaGlobals.proteinEC_table,
                                                                proteinCodeID_col = PianaGlobals.ecID_col,
                                                                proteinPiana_value = proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_proteins_sharing_ec

    def get_proteins_sharing_cog(self, proteinPiana_value ):
        """
        Returns a list of proteinPiana proteins that share a COG identifier with "proteinPiana_value"

        Empty list returned if nothing is found
        """

        if verbose:
            sys.stderr.write("Inside get_proteins_sharing_cog with proteinPiana= %s\n" %proteinPiana_value)

        list_proteins_sharing_cog = self.db.select_db_content(
            PianaGlobals.SelectCommonCodeProteins.get_sqlquery( proteinCode_table = PianaGlobals.proteinCog_table,
                                                                proteinCodeID_col = PianaGlobals.cogID_col,
                                                                proteinPiana_value = proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_proteins_sharing_cog
    

    def get_proteins_sharing_scop(self, proteinPiana_value ):
        """
        Returns a list of proteinPiana proteins that share a SCOP family with "proteinPiana_value"

        Empty list returned if nothing is found
        """

        if verbose:
            sys.stderr.write("Inside get_proteins_sharing_scop with proteinPiana= %s\n" %proteinPiana_value)

        list_proteins_sharing_scop = self.db.select_db_content(
            PianaGlobals.SelectCommonCodeProteins.get_sqlquery( proteinCode_table = PianaGlobals.proteinScop_table,
                                                                proteinCodeID_col = PianaGlobals.fa_col,
                                                                proteinPiana_value = proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_proteins_sharing_scop
    

    def get_proteins_sharing_interpro(self, proteinPiana_value ):
        """
        Returns a list of proteinPiana proteins that share a interpro identifier with "proteinPiana_value"

        Empty list returned if nothing is found
        """

        if verbose:
            sys.stderr.write("Inside get_proteins_sharing_interpro with proteinPiana= %s\n" %proteinPiana_value)

        list_proteins_sharing_interpro = self.db.select_db_content(
            PianaGlobals.SelectCommonCodeProteins.get_sqlquery( proteinCode_table = PianaGlobals.interPro_table,
                                                                proteinCodeID_col = PianaGlobals.interProID_col,
                                                                proteinPiana_value = proteinPiana_value)
            , answer_mode="list", remove_duplicates="yes")

        return list_proteins_sharing_interpro
    
    
    # -------------------------------------------
    # Access methods that retrieve all proteins with a common given characteristic (not used by expansion methods)
    # -------------------------------------------

    def get_proteins_sharing_species(self, species_name_value= None, taxonomy_value= None ):
        """
        Returns a list of proteinPiana proteins that have species taxonomy_value (or species_name_value)

        The user can input "taxonomy_value" (9606 for human, etc) or a "speciesName" string ('human', 'yeast', ...)

        in case both species_name_value and taxonomy_value are set to something different from None, taxonomy_value is used

        Empty list returned if nothing is found
        
        """
        if taxonomy_value is None:
            if species_name_value is None:
                raise ValueError("a taxonomy or a species name is needed to get the proteins that have it")
            else:
                list_taxonomy_value = self.get_taxonomies_from_species_name(species_name_value = species_name_value)
        # END OF if taxonomy_value is None:
        else:
            list_taxonomy_value = [taxonomy_value]

        list_proteins_given_species=[]
        
        for taxonomy_value in list_taxonomy_value:

##             list_proteins_given_species.extend(self.db.select_db_content(
##                 PianaGlobals.SelectCommonCharacteristicProteins.get_sqlquery( proteinCode_table = PianaGlobals.proteinSpecies_table,
##                                                                               proteinCodeID_col = PianaGlobals.taxID_col,
##                                                                               proteinCodeID_value = taxonomy_value)
                
##                 , answer_mode="list", remove_duplicates="yes"))

            list_proteins_given_species.extend(self.db.select_db_content(
                PianaGlobals.SelectCommonCharacteristicProteins.get_sqlquery( proteinCode_table = PianaGlobals.protein_table,
                                                                              proteinCodeID_col = PianaGlobals.taxID_col,
                                                                              proteinCodeID_value = taxonomy_value)
                
                , answer_mode="list", remove_duplicates="yes"))
        # END OF for taxonomy_value in list_taxonomy_value:

        # remove duplicates:
        dic_proteins = {}
        for protein in list_proteins_given_species:
            dic_proteins[protein] = None

        return dic_proteins.keys()
        
    # -------------------------------------------
    # Access (insertions and retrievals) methods to protein attributes tables
    # -------------------------------------------

    def insert_species(self, tax_id, tax_name, tax_comment, tax_kingdom=None, source_db=None):
        """
        Inserts a species into Piana

        "tax_id" is the ncbi taxonomy identifier (eg. 1)

        "tax_name" is the name given to the species by ncbi (eg. 'human')

        "tax_comment" is the associated comment to the species (eg 'nothing')

        "tax_kingdom" is the kingdom of the species (eg. 'Eukaryota')

        "source_db" is the external database that is giving this data (eg. 'uniprot')
        """
        if tax_kingdom is None:
            tax_kingdom = "unknown"

        self.db.insert_db_content( PianaGlobals.InsertSpecies.get_sqlquery(taxID_value= tax_id,
                                                                           speciesName_value= tax_name,
                                                                           speciesDescription_value= tax_comment,
                                                                           speciesKingdom_value= tax_kingdom,
                                                                           #speciesSource_value= self.get_sourceDBID(source_db))
                                                                           speciesSource_value = self._get_sourceDBID_int(source_db),
                                                                           )
                                   , answer_mode=None)

        
    def insert_species_kingdom(self, tax_id, tax_name, tax_kingdom, source_db):
        """
        If tax_id exists in pianaDB, inserts the kingdom for tax id "tax_id" (tax name will be ignored)
        
        If tax_id doesn't exist, inserts the "tax_id", the "tax_name", the "tax_kingdom" and "source_db" as in insert_species()
        """

        if self.get_species_names_from_taxonomies(list_taxonomy_ids = [tax_id]):
            # if there is already an entry for this tax id, update kingdom
            self.db.insert_db_content( PianaGlobals.UpdateSpeciesKingdom.get_sqlquery(taxID_value= tax_id,
                                                                                      speciesKingdom_value= tax_kingdom)
                                       , answer_mode=None)
        else:
            # if there was no entry for tax id, insert all the information
            self.insert_species(tax_id=tax_id, tax_name=tax_name, tax_comment="", tax_kingdom=tax_kingdom, source_db=source_db)
        

    def insert_go(self, go_id, go_name, acc, term_type, distance2root, source_db):
        """
        Inserts a GO (Gene Ontology) entry into Piana

        "go_id" is the go term id

        "go_name" is the text string associated to the term id

        "acc" is the go accession (code that looks like 'GO:234234')

        "term_type" is the type of GO term
         -> can be one of the following (or None)
          - "molecular_function"
          - "biological_process"
          - "cellular_component"

        "distance2root" is the distance between this term id and the root of the GO hierarchy

        "source_db" is the external database that is giving this information
        """
        self.db.insert_db_content( PianaGlobals.InsertGo.get_sqlquery(goID_value= go_id,
                                                                      go_name_value= go_name,
                                                                      go_acc_value= acc,
                                                                      go_term_type_value=term_type,
                                                                      go_distance2root_value=distance2root,
                                                                      #goSource_value= self.get_sourceDBID(source_db),
                                                                      goSource_value = self._get_sourceDBID_int(source_db),
                                                                      buffer = self.insert_buffer)
                                   , answer_mode=None, buffer = self.insert_buffer)
        
    def insert_go_term2term_distance(self,term1_id,term2_id,distance):
        """
        Inserts distance "distance" between GO terms "term1" and "term2"

        This is the distance between those terms in the GO hierarchy
        """
        self.db.insert_db_content(PianaGlobals.InsertGo_term2term_distance.get_sqlquery(term1_id_value=term1_id,
                                                                                        term2_id_value=term2_id,
                                                                                        distance_value=distance,
                                                                                        buffer = self.insert_buffer), buffer= self.insert_buffer)
        
    def insert_cog(self, cog_id, cog_description, cog_function, source_db):
        """
        Inserts a COG (Cluster of Orthologous Genes) entry into Piana

        "cog_id" is the COG identifier

        "cog_description" is the text string describing the COG

        "cog_function" is the text string describing the function of genes in this cluster

        "source_db" is the database that is giving this information
        """
        self.db.insert_db_content( PianaGlobals.InsertCog.get_sqlquery(cogID_value= cog_id,
                                                                       cogDescription_value= cog_description,
                                                                       cogFunction_value= cog_function,
                                                                       #cogSource_value= self.get_sourceDBID(source_db),
                                                                       cogSource_value = self._get_sourceDBID_int(source_db),
                                                                       buffer = self.insert_buffer)
                                   , answer_mode=None, buffer=self.insert_buffer)
        
    # -----------------------------------------------------
    # Access (insertions and retrievals) methods to protein external databases tables
    # -----------------------------------------------------
    
    def insert_uniprotInfo(self, proteinPiana_value, uniprotEntry_value, uniprotAcc_value, data_class_value, description_value, geneName_value,
                           organism_value, organelle_value, proteinSequenceLength_value, proteinMW_value):
        """
        Inserts a uniprot entry into piana (all info found in uniprot... this is independent from uniprot entries and uniprot accession numbers)

        (this is not very used... just have it here for being able to query piana about uniprot info)

        For a description of these fields, please refer to the uniprot manual
        
        """

        self.db.insert_db_content( PianaGlobals.InsertUniprotInfo.get_sqlquery( proteinPiana_value= proteinPiana_value,
                                                                                uniprotEntry_value= uniprotEntry_value,
                                                                                uniprotAcc_value= uniprotAcc_value,
                                                                                data_class_value= data_class_value,
                                                                                description_value= description_value,
                                                                                geneName_value= geneName_value,
                                                                                organism_value= organism_value,
                                                                                organelle_value= organelle_value,
                                                                                proteinSequenceLength_value= proteinSequenceLength_value,
                                                                                proteinMW_value= proteinMW_value)
                                   , answer_mode=None)
   
    # ------------------------------------
    # Access  methods (insertions, deletions, retrievals) to interaction tables
    # ------------------------------------

    def delete_interaction(self, interactionPiana_value):
        """
        Deletes entry of table interaction with interactionPiana = interactionPiana_value

        Attention! You should not delete an interactionPiana before making sure that other tables of interactions "interactionPiana" are
        deleted as well: this will just delete the row in table interaction
        """
        self.db.insert_db_content(PianaGlobals.DeleteInteraction.get_sqlquery(interactionPiana_value= interactionPiana_value),
                                                     answer_mode = None)


    def delete_interaction_sourceDB_for_sourceDBID(self, sourceDBID_value):
        """
        Deletes entries of table interactionSourceDB with sourceDBID = "sourceDBID_value"

        To do it, before the deletion performs a select to keep a list of those entries that are going to be deleted.

        Returns a list of interactionPiana that have been deleted from table interactionSourceDB
        """
        raise ValueError("Attention: this method has to be updated in order to reflect the new way of managing sourceDBs")
    
        deleted_interactionPiana = self.db.select_db_content(PianaGlobals.SelectInteractionSourceDB.get_sqlquery(sourceDBID_value= sourceDBID_value),
                                                     answer_mode = "list")

        # insert_db_content can also be used for deleting... since it follows same format as insertions
        self.db.insert_db_content(PianaGlobals.DeleteInteractionSourceDB.get_sqlquery(sourceDBID_value= sourceDBID_value),
                                                     answer_mode = None)

        return deleted_interactionPiana

        
    def delete_interaction_method_for_sourceDBID(self, sourceDBID_value):
        """
        Deletes entries of table interactionMethod with sourceDBID = "sourceDBID_value"
        """
        self.db.insert_db_content(PianaGlobals.DeleteInteractionMethod.get_sqlquery(sourceDBID_value= sourceDBID_value),
                                  answer_mode = None)
        
    def delete_interaction_scores_for_sourceDBID(self, sourceDBID_value):
        """
        Deletes entries of table interactionScores with sourceDBID = "sourceDBID_value"
        """
        self.db.insert_db_content(PianaGlobals.DeleteInteractionScores.get_sqlquery(sourceDBID_value= sourceDBID_value),
                                  answer_mode = None)
        
    def delete_interaction_protein_source_for_sourceDBID(self, sourceDBID_value):
        """
        Deletes entries of table interactionProteinSource with sourceDBID = "sourceDBID_value"
        """
        self.db.insert_db_content(PianaGlobals.DeleteInteractionProteinSource.get_sqlquery(sourceDBID_value= sourceDBID_value),
                                  answer_mode = None)

 
    def insert_interaction(self, proteinPianaA_value, isSourceA_value, proteinPianaB_value, isSourceB_value, interactionConfidence_value,
                           methodDescription_value, sourceDBDescription_value, confidenceAssignedSourceDB_value, pubmed_id_value="unknown"  ):

        """
        Inserts a new interaction into piana database.

        "proteinPianaA_value" is the proteinPiana for one side of the interaction
        "isSourceA_value" sets if the interaction goes from A to B (1) or not (0)
        
        "proteinPianaB_value" is the proteinPiana for the other side of the interaction
        "isSourceB_value" sets if the interaction goes from B to A (1) or not (0)

        if "isSourceA_value" and "isSourceB_value" are 1, the interaction is bi-directional

        "interactionConfidence_value" is the relyability of the interaction (not being used currently)

        "methodDescription_value" is the method name that detected this interaction (eg. 'yeast two hybrids')
                                  It can be a single value or a list of values. The user has not to worry about it, the method takes it into account
          --> must appear in PianaGlobals.method_names
          
        "sourceDBDescription_value" is the source database that contains this interaction (eg. 'DIP')
          --> The database must exist in the database
       
        "confidenceAssignedSourceDB_value" is the relyability assigned to the interaction by the source database

        "pubmed_id_value" is the pubmed identifier for the article where this interaction was described
                          It can be a single value or a list of values. The user has not to worry about it, the methods takes it into account

        Things this method does are:

           - makes sure the order proteinPianaA < proteinPianaB is respected
           - searches the methodID corresponding to the method description
           - searches the sourceDBID corresponding to the sourceDB description

            1. check if interaction exists already, retrieve interactionPiana in case it does
            2. if interactionPiana is  None:
                   insert information of table interaction and retrieve new interactionPiana
            3. insert interactionSourceDB with interactionPiana 
            4. insert interactionMethod according to this sourceDB with interactionPiana 


        Attention! I don't allow the user to limit the insertions to those from an specific database or method. If he wants to
        filter the database, do it afterwards... the pianaDB will contains all interactions regardless of origin
         (if one day somebody wants to change this, he will have to receive arguments list_source_dbs and list_source_methods
          and use them below to limit the insertion)
        
        """
        # TO DO!!!! interactionConfidence_value should not be given by user, but calculated internally from the other confidences
        #           availables... therefore, it shouldn't be an argument of this method

        if isinstance(proteinPianaA_value, str) or isinstance(proteinPianaB_value, str) :
            raise TypeError("insert_interaction needs numeric values for proteinPianaA_value and proteinPianaB_value")
        
            
        # making sure that order proteinPianaA < proteinPianaB is respected
        if proteinPianaB_value < proteinPianaA_value:
            temp                = (proteinPianaA_value, isSourceA_value)
            (proteinPianaA_value, isSourceA_value) = (proteinPianaB_value, isSourceB_value)
            (proteinPianaB_value, isSourceB_value) = temp




        
        # obtaining the sourceDBID corresponding to sourceDBDescription_value
        #sourceDBID_value = self.get_sourceDBID(sourceDBDescription_value = sourceDBDescription_value)
	sourceDBID_value = sourceDBDescription_value

        if verbose_insert_interaction:
            sys.stderr.write("Inserting interaction from PianaDBaccess with values: \n")
            sys.stderr.write("proteinPianaA_value: %s\n" %proteinPianaA_value)
            sys.stderr.write("isSourceA_value: %s\n" %isSourceA_value)
            sys.stderr.write("proteinPianaB_value: %s\n" %proteinPianaB_value)
            sys.stderr.write("isSourceB_value: %s\n" %isSourceB_value)
            sys.stderr.write("interactionConfidence_value: %s\n" %interactionConfidence_value)
            sys.stderr.write("methodID_value: %s\n" %methodID_value)
            sys.stderr.write("sourceDBID_value: %s\n" %sourceDBID_value)
            sys.stderr.write("pubmed_id_value: %s\n" %pubmed_id_value)


        # TO DO!!!!!!! I am not taking into account if interactions are bi-directional or not!!!!

        # 1. check if interaction exists already, retrieve interactionPiana in case it does (ignoring source dbs and methods)
        last_interaction_id = self.get_interactionPiana(proteinPianaA_value=proteinPianaA_value,
                                                        proteinPianaB_value=proteinPianaB_value,
                                                        list_source_dbs= "all",
                                                        inverse_dbs= "no",
                                                        list_source_methods= "all",
                                                        inverse_methods= "no")


        # If interaction wasn't in DB, check if it is in the internal buffer (only if insert_buffer is being used)
        # because it may be that the interaction was inserted previously but it is not in the DB yet (due it may be in the internal buffer)
        if last_interaction_id is None and self.insert_buffer:
            try:
                last_interaction_id = self.insert_interaction_buffer[(proteinPianaA_value,proteinPianaB_value)]
            except KeyError:
                last_interaction_id = None
        
        # 2. if interaction wasn't in DB: insert new information into table interaction_table
        #    and if the insert buffer is being used, insert it into the insert_interaction_dictionary
        if last_interaction_id is None:
            # Get the new interactionPiana
            last_interaction_id = self.get_new_interaction_piana()
            
            self.db.insert_db_content(
                PianaGlobals.InsertInteraction.get_sqlquery(interactionPiana_value = last_interaction_id,
                                                            proteinPianaA_value=proteinPianaA_value,
                                                            isSourceA_value=isSourceA_value,
                                                            proteinPianaB_value=proteinPianaB_value,
							    sourceDBID_value=self._get_sourceDBID_int(sourceDBDescription_value),
                                                            isSourceB_value=isSourceB_value,
                                                            interactionConfidence_value=interactionConfidence_value,
                                                            buffer = self.insert_buffer)
                , answer_mode = None, buffer = self.insert_buffer)

            if self.insert_buffer:
                self.insert_interaction_buffer[(proteinPianaA_value,proteinPianaB_value)] = last_interaction_id
	else:
            # Update the sourceDBID for this interaction
            self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.interaction_table,
                                                                                     column_values = [(PianaGlobals.interactionPiana_col,last_interaction_id)],
                                                                                     sourceDBID = self._get_sourceDBID_int(sourceDBDescription_value),
                                                                                     db_temp_buffer = self.db_temp_buffer,
                                                                                     conditions = [PianaGlobals.interactionPiana_col],
                                                                                     buffer = self.insert_buffer ),
                                       answer_mode=None, buffer= self.insert_buffer)
            #num_updated = self.db.insert_db_content( PianaGlobals.UpdateSourceDBID.get_sqlquery(table= PianaGlobals.interaction_table,
            #                                                                                    field= PianaGlobals.sourceDBID_col,
            #                                                                                    value= self._get_sourceDBID_int(sourceDBID_value),
            #                                                                                    conditions =[(PianaGlobals.interactionPiana_col,last_interaction_id)]),
            #                                         answer_mode="num_updated" )



            #if self.insert_buffer:
            #    self.insert_interaction_buffer[(proteinPianaA_value,proteinPianaB_value)] = last_interaction_id
            # END OF if self.insert_buffer

        # END OF if last_interaction_id is not None:
        
        # 3. inserting information into table interactionSourceDB_table
        #    in case the row (interactionPiana, sourceDBID) already exists, the insertion is ignored (INSERT IGNORE)
        #self.db.insert_db_content(
        #    PianaGlobals.InsertInteractionSourceDB.get_sqlquery(sourceDBID_value=sourceDBID_value,
        #                                                        confidenceAssignedSourceDB_value=confidenceAssignedSourceDB_value,
        #                                                        interactionPiana_value=last_interaction_id)
        #    , answer_mode = None)
        
        # 4. inserting information into table interactionMethod_table
        #    in case the row (interactionPiana, methodID, sourceDBID) already exists, the insertion is ignored (INSERT IGNORE)

        # Take into account if it is a single value or it is a list
        if not isinstance(methodDescription_value,list):

            # obtaining the methodID corresponding to methodDescription_value
            methodID_value = self.get_methodID(methodDescription_value = methodDescription_value)

	    if methodID_value=="unknown":
		sys.stdout.write("method %s not known\n" %(methodDescription_value) )

            self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.interactionMethod_table,
                                                                                     column_values = [(PianaGlobals.methodID_col,methodID_value),
                                                                                                      (PianaGlobals.interactionPiana_col,last_interaction_id)],
                                                                                     sourceDBID = self._get_sourceDBID_int(sourceDBDescription_value),
                                                                                     db_temp_buffer = self.db_temp_buffer,
                                                                                     conditions = [PianaGlobals.methodID_col,PianaGlobals.interactionPiana_col],
                                                                                     buffer = self.insert_buffer ),
                                       answer_mode=None, buffer= self.insert_buffer)             
            #self.db.insert_db_content( PianaGlobals.InsertInteractionMethod.get_sqlquery(methodID_value=methodID_value,
            #                                                                             interactionPiana_value=last_interaction_id,
            #                                                                             sourceDBID_value=sourceDBID_value)
            #                           , answer_mode = None)
        else:
            for actual_method in methodDescription_value:
                # obtaining the methodID corresponding to methodDescription_value
                methodID_value = self.get_methodID(methodDescription_value = actual_method)
		if methodID_value=="unknown":
                	sys.stdout.write("method %s not known\n" %(actual_method) )

                
                self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.interactionMethod_table,
                                                                                         column_values = [(PianaGlobals.methodID_col,methodID_value),
                                                                                                          (PianaGlobals.interactionPiana_col,last_interaction_id)],
                                                                                         sourceDBID = self._get_sourceDBID_int(sourceDBDescription_value),
                                                                                         db_temp_buffer = self.db_temp_buffer,
                                                                                         conditions = [PianaGlobals.methodID_col,PianaGlobals.interactionPiana_col],
                                                                                         buffer = self.insert_buffer ),
                                           answer_mode=None, buffer= self.insert_buffer)             
                
                ## self.db.insert_db_content( PianaGlobals.InsertInteractionMethod.get_sqlquery(methodID_value=methodID_value,
##                                                                                              interactionPiana_value=last_interaction_id,
##                                                                                              sourceDBID_value=sourceDBID_value)
##                                            , answer_mode = None)
                

            
        
        # 5. inserting information into table interactionFeatures_table
        #    in case the row (interactionPiana, pubmedID, sourceDBID) already exists, the insertion is ignored (INSERT IGNORE)

        # Take into account if it is a single value or it is a list
        if pubmed_id_value != "unknown":
            if not isinstance(pubmed_id_value,list):
                
                #self.db.insert_db_content( PianaGlobals.InsertInteractionFeatures.get_sqlquery(pubmedID_value=pubmed_id_value,
                #                                                                               interactionPiana_value=last_interaction_id,
                #                                                                               sourceDBID_value=sourceDBID_value)
                #                           , answer_mode = None)
                self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.interactionFeatures_table,
                                                                                         column_values = [(PianaGlobals.pubmedID_col,pubmed_id_value),
                                                                                                          (PianaGlobals.interactionPiana_col,last_interaction_id)],
                                                                                         sourceDBID = self._get_sourceDBID_int(sourceDBDescription_value),
                                                                                         db_temp_buffer = self.db_temp_buffer,
                                                                                         conditions = [PianaGlobals.pubmedID_col,PianaGlobals.interactionPiana_col],
                                                                                         buffer = self.insert_buffer ),
                                           answer_mode=None, buffer= self.insert_buffer)             
                
            else:
                for actual_pubmed in pubmed_id_value:
                    self.db.insert_db_content( PianaGlobals.InsertTemporalTable.get_sqlquery(table_name = PianaGlobals.interactionFeatures_table,
                                                                                             column_values = [(PianaGlobals.pubmedID_col,actual_pubmed),
                                                                                                              (PianaGlobals.interactionPiana_col,last_interaction_id)],
                                                                                             sourceDBID = self._get_sourceDBID_int(sourceDBDescription_value),
                                                                                             db_temp_buffer = self.db_temp_buffer,
                                                                                             conditions = [PianaGlobals.pubmedID_col,PianaGlobals.interactionPiana_col],
                                                                                             buffer = self.insert_buffer ),
                                               answer_mode=None, buffer= self.insert_buffer)             
                
                #self.db.insert_db_content( PianaGlobals.InsertInteractionFeatures.get_sqlquery(pubmedID_value=actual_pubmed,
                #                                                                               interactionPiana_value=last_interaction_id,
                #                                                                               sourceDBID_value=sourceDBID_value)
                #                           , answer_mode = None)
                
        return last_interaction_id
    
    def insert_interaction_scores(self, interactionPiana_value, sourceDBDescription_value,
                                  equiv_nscore_value, equiv_nscore_transferred_value,
                                  equiv_fscore_value,
                                  equiv_pscore_value, equiv_hscore_value, array_score_value, array_score_transferred_value,
                                  experimental_score_value, experimental_score_transferred_value, database_score_value, database_score_transferred_value,
                                  textmining_score_value, textmining_score_transferred_value, combined_score_value  ):
        """
        Inserts interaction scores into table interactionScores_table

        This table only holds information for interactions contained in STRING 

        Refer to the string manual for description of the different arguments
        """
        # obtaining the sourceDBID corresponding to sourceDBDescription_value
        #sourceDBID_value = self.get_sourceDBID(sourceDBDescription_value = sourceDBDescription_value)

        sourceDBID_value = sourceDBDescription_value

        # inserting information into table interactionSourceDB_table
        self.db.insert_db_content( PianaGlobals.InsertInteractionScores.get_sqlquery(interactionPiana_value = interactionPiana_value,
                                                                                     sourceDBID_value = self._get_sourceDBID_int(sourceDBID_value),
                                                                                     equiv_nscore_value = equiv_nscore_value,
                                                                                     equiv_nscore_transferred_value = equiv_nscore_transferred_value,
                                                                                     equiv_fscore_value = equiv_fscore_value,
                                                                                     equiv_pscore_value = equiv_pscore_value,
                                                                                     equiv_hscore_value = equiv_hscore_value,
                                                                                     array_score_value = array_score_value,
                                                                                     array_score_transferred_value = array_score_transferred_value,
                                                                                     experimental_score_value = experimental_score_value,
                                                                                     experimental_score_transferred_value = experimental_score_transferred_value,
                                                                                     database_score_value = database_score_value,
                                                                                     database_score_transferred_value = database_score_transferred_value,
                                                                                     textmining_score_value = textmining_score_value,
                                                                                     textmining_score_transferred_value = textmining_score_transferred_value,
                                                                                     combined_score_value = combined_score_value)
                                   , answer_mode = None)
        

    def insert_interaction_protein_source(self, interactionPiana_value = None, proteinPianaSource_value = None, sourceDBDescription_value= None):
        """
        Inserts the interaction protein source into table interactionProteinSource_table

        "proteinPianaSource_value" is the protein that originated the predicted interaction "interactionPiana_value"

        This table only has information for interactions that were obtained by a expansion prediction
        
        """
        # obtaining the sourceDBID corresponding to sourceDBDescription_value
        #sourceDBID_value = self.get_sourceDBID(sourceDBDescription_value = sourceDBDescription_value)
        sourceDBID_value = sourceDBDescription_value

        self.db.insert_db_content( PianaGlobals.InsertInteractionProteinSource.get_sqlquery(interactionPiana_value = interactionPiana_value,
                                                                                            proteinPianaSource_value = proteinPianaSource_value,
                                                                                            sourceDBID_value=self._get_sourceDBID_int(sourceDBID_value))
                                   , answer_mode = None)
        

    def get_interactionPiana(self, proteinPianaA_value, proteinPianaB_value,
                             list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no"):
        """
        Returns the interacionPiana for the interaction between proteins "proteinPianaA_value" and "proteinPianaB_value", considering only those
        interaction source databases listed in "list_source_dbs" ('all' returns all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' returns all interactions from all methods)

        No need to respect order of proteinPianaA and proteinPianaB: this method will take care of it, place them as you want.

        If there is not such interaction, returns None
        """
        interactionPiana = self.db.select_db_content(PianaGlobals.SelectInteractionPiana.get_sqlquery(proteinPianaA_value=proteinPianaA_value,
                                                                                                      proteinPianaB_value=proteinPianaB_value,
                                                                                                      list_source_dbs=list_source_dbs,
                                                                                                      inverse_dbs= inverse_dbs,
                                                                                                      list_source_methods=list_source_methods,
                                                                                                      inverse_methods= inverse_methods,
												      valid_source_dbs=self.get_valid_sources()),
                                                         answer_mode = "single")
        return interactionPiana


    def get_all_partners(self, proteinPiana_value= None, use_self_ints="yes",
                         list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                         threshold = 0):
        """
        Returns a list with all partners (proteinPianas) of "proteinPiana_value", considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods)

        If use_self_ints == "yes", it won't return itself as a partner .Set to 'no' if you want to get it in case it exists.

        "threshold" is used to limit the number of partners that can be returned:
           - if "threshold" == 0, then always return all partners
           - if threshold!=0 and number of partners > threshold, then return empty list (no partners)

        Returns empty list if no partner is found
        """
        if proteinPiana_value is None:
            raise ValueError("arguments to get_all_partners cannot be None")

        
        # get partners in table for both possible cases: proteinPiana can be the lowest value (A) or the highest (B)
        list_partners = self.db.select_db_content(PianaGlobals.SelectPartners.get_sqlquery( proteinPiana_value= proteinPiana_value,
                                                                                            partner_side= "A",
                                                                                            list_source_dbs=list_source_dbs,
                                                                                            inverse_dbs= inverse_dbs,
                                                                                            list_source_methods=list_source_methods,
                                                                                            inverse_methods= inverse_methods,
                                                                                            valid_source_dbs=self.get_valid_sources()) ,
                                                   answer_mode="list", remove_duplicates="no")
        

        list_partners.extend( self.db.select_db_content(PianaGlobals.SelectPartners.get_sqlquery( proteinPiana_value= proteinPiana_value,
                                                                                                  partner_side= "B",
                                                                                                  list_source_dbs=list_source_dbs,
                                                                                                  inverse_dbs= inverse_dbs,
                                                                                                  list_source_methods=list_source_methods,
                                                                                                  inverse_methods= inverse_methods,
                                                                                                  valid_source_dbs=self.get_valid_sources()) ,
                                                   answer_mode="list", remove_duplicates="no"))

        

        # OLD CODE, previous to optimization
        #dic_partners = {}
        #for partner in list_partnersA:
        #    dic_partners[partner] = None
        #for partner in list_partnersB:
        #    dic_partners[partner] = None

        if use_self_ints == "no":

            # join lists so that we speed up afterwards when removing self interactions if asked
            dic_partners = dict( map(None, list_partners, []) )

            # if user asked to remove self interactions, do it...
            try:
                del(dic_partners[proteinPiana_value])
            except:
                # in case itself wasn't there... just continue...
                pass

            list_partners = dic_partners.keys()
        # END OF if use_self_ints == "no":

        num_partners = len(list_partners)

        if threshold == 0 or num_partners <= threshold:
            # use argument threshold to decide whether the partners are returned or not
            return list_partners
        else:
            return []



    def get_all_partners_with_interactionPiana(self, proteinPiana_value= None, use_self_ints="yes",
                                               list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                                               threshold = 0):
        """
        Returns a list with all partners (proteinPiana,interactionPiana) of "proteinPiana_value" and their associated interactionPiana, considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods)

        If use_self_ints == "yes", it won't return itself as a partner .Set to 'no' if you want to get it in case it exists.

        "threshold" is used to limit the number of partners that can be returned:
           - if "threshold" == 0, then always return all partners
           - if threshold!=0 and number of partners > threshold, then return empty list (no partners)

        Returns empty list if no partner is found
        """
        if proteinPiana_value is None:
            raise ValueError("arguments to get_all_partners cannot be None")

        
        # get partners in table for both possible cases: proteinPiana can be the lowest value (A) or the highest (B)
        list_partners = self.db.select_db_content(PianaGlobals.SelectPartners.get_sqlquery( proteinPiana_value= proteinPiana_value,
                                                                                            partner_side= "A",
                                                                                            list_source_dbs=list_source_dbs,
                                                                                            inverse_dbs= inverse_dbs,
                                                                                            list_source_methods=list_source_methods,
                                                                                            inverse_methods= inverse_methods,
                                                                                            valid_source_dbs=self.get_valid_sources(),
                                                                                            get_interactionPiana="yes") ,
                                                  answer_mode="list", remove_duplicates="no", number_of_selected_elems = 2)
        

        list_partners.extend( self.db.select_db_content(PianaGlobals.SelectPartners.get_sqlquery( proteinPiana_value= proteinPiana_value,
                                                                                                  partner_side= "B",
                                                                                                  list_source_dbs=list_source_dbs,
                                                                                                  inverse_dbs= inverse_dbs,
                                                                                                  list_source_methods=list_source_methods,
                                                                                                  inverse_methods= inverse_methods,
                                                                                                  valid_source_dbs=self.get_valid_sources(),
                                                                                                  get_interactionPiana="yes") ,
                                                        answer_mode="list", remove_duplicates="no", number_of_selected_elems = 2))

        # OLD CODE, previous to optimization
        #dic_partners = {}
        #for partner in list_partnersA:
        #    dic_partners[partner] = None
        #for partner in list_partnersB:
        #    dic_partners[partner] = None

        if use_self_ints == "no":
            # if user asked to remove self interactions, do it...
            
            # join lists removing duplicates and removing self interactions if asked
            dic_partners = dict( map(None, list_partners, []) )

            for actual_partner in dic_partners:
                # actual_partner is a tuple (partner_proteinPiana, interactionPiana)
                if actual_partner[0] == proteinPiana_value:
                    del(dic_partners[actual_partner])
                    break

            list_partners = dic_partners.keys()


            
            # Old code:
            #try:
            #    del(dic_partners[proteinPiana_value])
            #except:
            #    # in case itself wasn't there... just continue...
            #    pas
        # END OF if use_self_ints == "no":

        num_partners = len(list_partners)

        if threshold == 0 or num_partners <= threshold:
            # use argument threshold to decide whether the partners are returned or not
            return list_partners
        else:
            return []

            
    def get_all_g2_partners(self, proteinPiana_value= None, use_self_ints="yes",
                            list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                            threshold = 0):
        """
        Returns a list with all partners (proteinPianas) at distance 2 of "proteinPiana_value", considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods)

             --> If A interacts with B and C, and B interacts with D and C with no protein, then partners of A at distance 2 are [D]


        "threshold" is used to limit the number of partners that each of the proteins at distance 1 can have
           - if "threshold" == 0, then always return all partners for all proteins
           - if threshold!=0 and number of partners > threshold, then return empty list (no partners) for that particular protein

           --> ATTENTION!!!! threshold does not apply to the partners being returned by this method, but to the individual calls
                             that this method makes to get_all_partners. Therefore, this method can return whichever number of
                             partners, guaranteeing that no single protein has added to the list more than "threshold" partners
           

        Returns empty list if no partner at distance 2 is found
        """
        if proteinPiana_value is None:
            raise ValueError("arguments to get_all_g2_partners cannot be None")


        list_partners_g1 = self.get_all_partners( proteinPiana_value= proteinPiana_value,
                                                  use_self_ints= use_self_ints,
                                                  list_source_dbs= list_source_dbs, inverse_dbs=inverse_dbs, 
                                                  list_source_methods= list_source_methods, inverse_methods=inverse_methods,
                                                  threshold = threshold)

        dic_partners_g2 = {}
        for partner_g1 in list_partners_g1:

            for temp_partner_g2 in self.get_all_partners( proteinPiana_value= partner_g1,
                                                          use_self_ints= use_self_ints,
                                                          list_source_dbs= list_source_dbs, inverse_dbs=inverse_dbs, 
                                                          list_source_methods= list_source_methods, inverse_methods=inverse_methods,
                                                          threshold = threshold):
                
                dic_partners_g2[temp_partner_g2] = None
            # END OF for temp_partner_g2 in self.get_all_partners(....)
        # END OF for partner_g1 in list_partners_g1:
            
        return dic_partners_g2.keys()

    def get_all_protein_protein_interactions(self, use_self_ints="yes",
                                             list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no",
                                             only_id = "no", tax_id_value="all"):
        """
        Returns a list of triplets (proteinPianaA, proteinPianaB, interactionPiana), considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods).

        If "inverse_dbs"/"inverse_methods" is 'yes', then it returns everything except for those dbs/methods
        
        If use_self_ints == "no", it won't return interactions between proteinPiana_value and itself. Set to 'yes' if you want to get them

        if only_id is 'yes', then, instead of triplets, a list of interactionPianas is returned

        
        if no interaction is found, returns empty list
        """
        if only_id == "yes":
            number_of_selected_elems = 1
        else:
            number_of_selected_elems = 3

        if tax_id_value==0:
            tax_id_value="all"
            
            
        listProteinProteinInteraction = self.db.select_db_content(PianaGlobals.SelectAllProteinProteinInteractions.get_sqlquery(
            use_self_ints = use_self_ints,
            list_source_dbs=list_source_dbs,
            inverse_dbs= inverse_dbs,
            list_source_methods=list_source_methods,
            inverse_methods= inverse_methods,
            only_id = only_id,
            valid_source_dbs=self.get_valid_sources(),
            tax_id_value=tax_id_value),
                                                                  answer_mode = "list", remove_duplicates="no",
                                                                  number_of_selected_elems=number_of_selected_elems)


        return listProteinProteinInteraction


    def get_dic_proteins_with_known_ints(self, use_self_ints="yes",
                                         list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no"):
        """
        Returns a dictionary with keys those proteinPianas for which there are known interactions, considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods)

        Contents of the dictionary are None for all proteinPianas. 
        
        If use_self_ints == "no", it won't return proteinPianas that only interact with themselves itself. Set to 'yes' if you want to get them
        
        """

        all_interaction_triplets = self.get_all_protein_protein_interactions(use_self_ints=use_self_ints,
                                                                             list_source_dbs=list_source_dbs ,
                                                                             inverse_dbs=inverse_dbs,
                                                                             list_source_methods=list_source_methods,
                                                                             inverse_methods=inverse_methods)

        dic_proteinPianas_with_known_ints = {}

        for one_triplet in all_interaction_triplets:
            # one_triplet[0] is proteinPianaA
            # one_triplet[1] is proteinPianaB
            # one_triplet[2] is interactionPiana
            
            dic_proteinPianas_with_known_ints[one_triplet[0]] = None
            dic_proteinPianas_with_known_ints[one_triplet[1]] = None


        return dic_proteinPianas_with_known_ints
        
         
    def get_partners_of_proteins_sharing_cog(self, proteinPiana_value= None,
                                             list_source_dbs= "all", inverse_dbs="no", list_source_methods= "all", inverse_methods="no"):
        """
        Returns interaction partners (proteinPianas) of those proteins that share the cogID with "proteinPiana_value", considering only those
        interaction source databases listed in "list_source_dbs" ('all' gets all interactions from all databases) and those interaction
        source methods listed in "list_source_methods" ('all' gets all interactions from all methods)
        """
        if proteinPiana_value is None:
            raise ValueError("arguments to get_partners_of_proteins_sharing_cog cannot be None")

        list_partnersA = self.db.select_db_content(PianaGlobals.SelectPartnersProteinSharingCog.get_sqlquery( proteinPiana_value, "A",
                                                                                                              list_source_dbs=list_source_dbs,
                                                                                                              inverse_dbs= inverse_dbs,
                                                                                                              list_source_methods=list_source_methods,
                                                                                                              inverse_methods= inverse_methods,
                                                                                                              valid_source_dbs=self.get_valid_sources()) ,
                                                   answer_mode="list", remove_duplicates="yes")
        
        list_partnersB = self.db.select_db_content(PianaGlobals.SelectPartnersProteinSharingCog.get_sqlquery( proteinPiana_value, "B",
                                                                                                              list_source_dbs=list_source_dbs,
                                                                                                              inverse_dbs= inverse_dbs,
                                                                                                              list_source_methods=list_source_methods,
                                                                                                              inverse_methods= inverse_methods,
                                                                                                              valid_source_dbs=self.get_valid_sources()) ,
                                              answer_mode="list", remove_duplicates="yes")

        # if list_partnersA is empty then return list_partnersB. Otherwise, extend it with list_partnersB
        if list_partnersA:
            if list_partnersB:
                list_partnersA.extend(list_partnersB)
        else:
            list_partnersA = list_partnersB


        # remove duplicates
        dic_partners = {}
        for partner in list_partnersA:
            dic_partners[partner] = None


        return dic_partners.keys()


    def get_interactors(self, interaction_id= None):
        """
        Returns the tuple of interactors interactors (proteinPianaA, proteinPianaB) for a given interaction_id

        No need to check if the interaction belongs to a specific method: the user is only calling this method when he already
        retrieved an interactionPiana: and he will only retrieve interactionPianas that respect the source db and method restrictions
        """

        if interaction_id is None:
            raise ValueError("argument interaction_id cannot be None")

        return self.db.select_db_content(PianaGlobals.SelectInteractors.get_sqlquery( interactionPiana_value = interaction_id ),
                                                answer_mode="raw")
    
        
    

    def get_partner(self, interaction_id= None, proteinPiana_value= None):
        """
        Returns the partner (proteinPiana) of "proteinPiana_value" in interaction "interaction_id"

        No need to check if the interaction belongs to a specific method: the user is only calling this method when he already
        retrieved an interactionPiana: and he will only retrieve interactionPianas that respect the source db and method restrictions
        """

        # TO DO!! change argument name interaction_id to interactionPiana_value

        if interaction_id is None or proteinPiana_value is None:
            raise ValueError("arguments to get_partner cannot be None")
        
        partnerA = self.db.select_db_content(PianaGlobals.SelectInteractionPartner.get_sqlquery( proteinPiana_value ,interaction_id, "A"),
                                             answer_mode="single")
        
        if partnerA is not None:
            return partnerA
       
        else:
            partnerB = self.db.select_db_content(PianaGlobals.SelectInteractionPartner.get_sqlquery( proteinPiana_value ,interaction_id, "B"),
                                                 answer_mode="single")
          
            if partnerB is not None:
                return partnerB
            else:
                raise Exception("Strange error: not finding partner of proteinPiana_value %s which was supposed to have an interaction %s" \
                                %(proteinPiana_value, interaction_id ))

    def get_list_sourceDB_names_from_sourceDB_id(self, sourceDB_id):
        """
        returns a list of strings (ie. names of databases) from the integer associated to the record (i.e. sourceDB_id)
        """
        if sourceDB_id is None:
            return []

        try:
            #If it has been previously been calculated, get it
            return self.int_dbs_mask[sourceDB_id]
        except KeyError:
            #If it is not found, obtain which databases generate this combination
            self.int_dbs_mask[sourceDB_id] = []
            for actual_source in self.validSources:
                if sourceDB_id & self.validSources[actual_source]:
                    self.int_dbs_mask[sourceDB_id].append(actual_source)

            #if len(self.int_dbs_mask[sourceDB_id])==0:
            #    self.int_dbs_mask[sourceDB_id].append("Unknown")
            
            if len(self.int_dbs_mask[sourceDB_id])==0:
                #raise ValueError("Impossible combination of databases when obtaining database names for combination %s" %(sourceDB_id))
                self.int_dbs_mask[sourceDB_id].append("Unknown")
            #else:
            return self.int_dbs_mask[sourceDB_id]


    def get_sourceDB_id_from_interactionPiana(self, interactionPiana_value):
        """
        Returns the database integer sourceDBID for a particular "interactionPiana_value"

           --> this method returns an integer, not a list of sourceDB names!
           --> this should only be used internally by PianaGraphEdgeAttribute, to minimize the amount of memory used

        Returns None if nothing is found

        Attention! There is no argument list_source_dbs or list_source_methods: I let the user check if the interaction
        is in other databases or methods, even if he limited the network interactions to those in a list of source interaction databases
          -> I do it this way because this has no effect of the interactions added to the network: it will just affect the edge attribute, in the
          sense that it will have  a complete list of sourceDBs and methods, regardless of restrictions imposed by user.
        """
        
        sourceDB_id = self.db.select_db_content(PianaGlobals.SelectInteractionSourceDB.get_sqlquery(interactionPiana_value),
                                                  answer_mode="single", remove_duplicates="no")

        return sourceDB_id

    # the following method used to be called: get_interaction_sourceDB_list
    def get_list_sourceDB_names_from_interactionPiana(self, interactionPiana_value):
        """
        Returns a list of sourceDB for a particular "interactionPiana_value"

        Returns empty list if nothing is found

        Attention! There is no argument list_source_dbs or list_source_methods: I let the user check if the interaction
        is in other databases or methods, even if he limited the network interactions to those in a list of source interaction databases
          -> I do it this way because this has no effect of the interactions added to the network: it will just affect the edge attribute, in the
          sense that it will have  a complete list of sourceDBs and methods, regardless of restrictions imposed by user.
        """
        
        sourceDB_id = self.db.select_db_content(PianaGlobals.SelectInteractionSourceDB.get_sqlquery(interactionPiana_value),
                                                  answer_mode="single", remove_duplicates="no")

        # version when there was one record for each sourceDB
        # sourceDB_list = self.db.select_db_content(PianaGlobals.SelectInteractionSourceDB.get_sqlquery(interactionPiana_value),
        #                                          answer_mode="list", remove_duplicates="yes")
        # return sourceDB_list
                                                  
        
        return self.get_list_sourceDB_names_from_sourceDB_id(sourceDB_id=sourceDB_id)
    

    def get_interaction_methodID_list(self, interactionPiana_value):
        """
        Returns a list of methodID for a particular "interactionPiana_value"

        Returns empty list if nothing is found

        Attention! There is no argument list_source_dbs or list_source_methods: I let the user check if the interaction
        is in other databases or methods, even if he limited the network interactions to those in a list of source interaction databases
          -> I do it this way because this has no effect of the interactions added to the network: it will just affect the edge attribute, in the
          sense that it will have  a complete list of sourceDBs and methods, regardless of restrictions imposed by user.
          -> Right now this method is only used to check if an interaction can be deleted (./dbModification/delete_interactions_from_db.py)
             and to calculate piana stats


        """
        # TO DO!!! Right now, the edge attribute does not contain a list of methodIDs

        methodID_list = self.db.select_db_content(PianaGlobals.SelectInteractionMethod.get_sqlquery(interactionPiana_value= interactionPiana_value,
                                                                                                    sourceDBID_value = None),
                                                  answer_mode="list", remove_duplicates="no")
        
        return methodID_list


    def get_interaction_pubmedID_list(self, interactionPiana_value):
        """
        Returns a list of pubmedID for a particular "interactionPiana_value"

        Returns empty list if nothing is found

        Attention! There is no argument list_source_dbs or list_source_methods: I let the user check if the interaction
        is in other databases or methods, even if he limited the network interactions to those in a list of source interaction databases
          -> I do it this way because this has no effect of the interactions added to the network: it will just affect the edge attribute, in the
          sense that it will have  a complete list of sourceDBs and methods, regardless of restrictions imposed by user.
        """


        pubmedID_list = self.db.select_db_content(PianaGlobals.SelectInteractionPubmed.get_sqlquery(interactionPiana_value= interactionPiana_value,
                                                                                                    sourceDBID_value = None),
                                                  answer_mode="list", remove_duplicates="no")
        
        return pubmedID_list
  
    def get_interaction_methodID_list_for_sourceDB(self, interactionPiana_value, sourceDBID_value):
        """
        Returns a list of methodID for a particular interaction "interactionPiana_value" for a given "sourceDBID"
          (here, soucerDBID is the name of the database, not the integer associated to it)

        Returns empty list if nothing is found

        Attention! There is no argument list_source_dbs: since arguments of this method fix a sourceDB, I let the user be responsible for
        not asking to get methodIDs for a database that is not the one he wants...
        """

        methodID_list = self.db.select_db_content(PianaGlobals.SelectInteractionMethod.get_sqlquery(interactionPiana_value= interactionPiana_value,
                                                                                                    sourceDBID_value = self._get_sourceDBID_int(sourceDBID_value)),
                                                  answer_mode="list", remove_duplicates="yes")
        
        return methodID_list


    # --------------------------------------------
    #  BLAST/Protein Similarities related methods
    # --------------------------------------------
    def insert_blast_results(self, sequenceID_A_value, sequenceID_B_value, score_value, bit_score_value, start_A_value, end_A_value, start_B_value, end_B_value, identities_value, similarity_value, gaps_value, program_value, filter_value):
        """
        Inserts the blast result of two sequences in the database

        "program_value" is the program used to obtain the blast result. It can be "bl2seq" or "blastall"

        "filter_value" is used to check if filter is used when running blast. It can be "T" if the filter is used or "F" is it not used
        
        """

        if program_value != "bl2seq" and program_value != "blastall":
            raise ValueError("Error: %s. program_value parameter must be \"bl2seq\" or \"blastall\"" %(program_value))
        if filter_value != "F" and filter_value != "T":
            raise ValueError("Error: %s. filter_value parameter must be \"T\" or \"F\"" %(filter_value))
        if gaps_value is None:
            gaps_value = 0
        #if similarity_value is None:
        #    similarity_value = 0

        self.db.insert_db_content(PianaGlobals.InsertBlastResult.get_sqlquery(sequenceID_A = sequenceID_A_value,
                                                                              sequenceID_B = sequenceID_B_value,
                                                                              score = score_value,
                                                                              bit_score = bit_score_value,
                                                                              start_A = start_A_value,
                                                                              end_A = end_A_value,
                                                                              start_B = start_B_value,
                                                                              end_B = end_B_value,
                                                                              identities =identities_value,
                                                                              similarity = similarity_value,
                                                                              gaps = gaps_value,
                                                                              program = program_value,
                                                                              filter = filter_value,
                                                                              buffer = self.insert_buffer ),
                                                                              #buffer = None ),
                                  answer_mode=None, buffer= self.insert_buffer)

        

   
    # ------------------------------------
    # Parsers related methods
    # PIANA database updates
    # Lock and Buffer Related Methods
    # ------------------------------------



    def lock_tables(self, table_list=None):
        """
        Locks mysql tables indicated in "table_list"
        
        "table_list": list of tables to lock. If it is not defined or ff list is [], all tables will be locked
        """
        if table_list is None:
            table_list = self._get_table_names()
        else:
            if len(table_list)==0:
                table_list = self._get_table_names()

        if len(table_list)==0:
            raise ValueError("Error: there are no tables in the database.")

        self.db.select_db_content(PianaGlobals.LockTables.get_sqlquery(table_list),
                                      answer_mode="single")

    def unlock_tables(self):
        """
        Unlocks tables previously locked with method lock_tables()
        """
        self.db.select_db_content(PianaGlobals.UnlockTables.get_sqlquery(),
                                  answer_mode="single")

    def execute_insert_buffer(self, key_buffer= None):
        """
        Executes all inserts in an insert buffer and empties it

        If key_buffer is None, it will execute all buffe. Otherwise, it will insert only the key specified
        """


        if self.insert_buffer is None:
            raise ValueError("To execute insert buffer into database buffer object cannot be \"None\". First it is ncessary to create a buffer Object")
        else:
            if key_buffer is None:
                # Empties all buffer
                list_of_buffer_keys = self.insert_buffer.get_buffer().keys()
            else:
                # Empties only the key_buffer requested
                list_of_buffer_keys = [key_buffer]

            for actual_key in list_of_buffer_keys:

                bufferElement = self.insert_buffer.get_buffer()[actual_key]
                                    
                self.db.insert_db_content( PianaGlobals.MultipleInsert.get_sqlquery( table=bufferElement.getTable(),
                                                                                     columns=bufferElement.getColumns(),
                                                                                     values=bufferElement.getValues()),
                                           answer_mode=None, buffer=None)
                #self.db.insert_db_content(PianaGlobals.InsertBuffer.get_sqlquery(query_data = self.insert_buffer.get_buffer()[actual_key]),
                #                          answer_mode= None, buffer=None)
                # If the key_buffer to empty is "InsertProtein", empties the insert_protein_ffer dictionary and update proeinPianaCounter
                if actual_key=="InsertProtein":
                    self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(self.newproteinPiana+1,PianaGlobals.proteinPianaCounter_col),
                                               answer_mode=None )
                    self.insert_protein_buffer = {}
                    # If the key_buffer to empty is "InsertInteraction", empties the self.insert_interaction_buffer
                elif actual_key=="InsertInteraction":
                    self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(self.newInteractionPiana+1,PianaGlobals.interactionCounter_col),
                                               answer_mode=None )
                    self.insert_interaction_buffer = {}
                elif actual_key=="InsertSequence":
                    self.db.insert_db_content( PianaGlobals.InsertProteinPianaCounter.get_sqlquery(self.newSequenceID+1,PianaGlobals.sequenceIDCounter_col),
                                               answer_mode=None )
                    self.insert_sequence_buffer = {}

    
    def empty_temporal_table(self):
        """
        Executes all commands used to empty temporal table and update tables

        All table will be emptied
        """

        # Create indices in temporal table
        self.db.insert_db_content( PianaGlobals.CreateIndice.get_sqlquery(table=PianaGlobals.temporalData_table,
                                                                          indice_name="temp_index",
                                                                          #columns=("identifier","attr1","attr2","attr3")),
                                                                          columns=("identifier",)),
                                   answer_mode=None, buffer=None)
        

        list_of_buffer_keys = self.db_temp_buffer.get_buffer().keys()

        for actual_key in list_of_buffer_keys:
            temporalTableBuffer = self.db_temp_buffer.get_buffer_element(actual_key)

            # updates the table with previously existing values
	    itime = time.time()
            #sys.stderr.write("%s\n" %(PianaGlobals.UpdateSourceDBIDfromTempTable.get_sqlquery( identifier = actual_key,
            #                 		                                              table = temporalTableBuffer.getTable(),
            #                            	                                      conditions = temporalTableBuffer.get_conditions_list() ) ))
            self.db.insert_db_content( PianaGlobals.UpdateSourceDBIDfromTempTable.get_sqlquery( identifier = actual_key,
                                                                                                table = temporalTableBuffer.getTable(),
                                                                                                conditions = temporalTableBuffer.get_conditions_list() ),
                                       answer_mode=None, buffer=None)
	    sys.stderr.write("time: %s\n" %(time.time()-itime) )
	    itime = time.time()
            #sys.stderr.write("%s\n" %(PianaGlobals.DeleteElementsFromTempTable.get_sqlquery( identifier = actual_key,
            #                 		                                            table = temporalTableBuffer.getTable(),
            #                            	                                    conditions = temporalTableBuffer.get_conditions_list() ) ))
            # delete those records in temporal table
            #self.db.insert_db_content( PianaGlobals.DeleteElementsFromTempTable.get_sqlquery( identifier = actual_key,
            #                                                                                  table = temporalTableBuffer.getTable(),
            #                                                                                  conditions = temporalTableBuffer.get_conditions_list() ),
            #                           answer_mode=None, buffer=None)
	    #sys.stderr.write("time: %s\n" %(time.time()-itime) )
	    #itime = time.time()

            
	    #sys.stderr.write("%s\n" %(PianaGlobals.InsertElementsFromTempTable.get_sqlquery( table= temporalTableBuffer.getTable(),
	    #                                                                         columns= temporalTableBuffer.get_columns_tuple(),
            #                                                                         identifier = actual_key) ) )
            # inserts those records in table
            self.db.insert_db_content( PianaGlobals.InsertElementsFromTempTable.get_sqlquery( table= temporalTableBuffer.getTable(),
                                                                                              columns= temporalTableBuffer.get_columns_tuple(),
                                                                                              identifier = actual_key),
                                       answer_mode=None, buffer=None)
	    sys.stderr.write("time: %s\n" %(time.time()-itime ))
	    #itime=time.time()
            #sys.stderr.write("%s\n" %( PianaGlobals.DeleteTempTableIdentifier.get_sqlquery( identifier = actual_key ) ))
            #deletes all inserted records
            #self.db.insert_db_content( PianaGlobals.DeleteTempTableIdentifier.get_sqlquery( identifier = actual_key ),
            #                           answer_mode=None, buffer=None)
	    #sys.stderr.write("time: %s\n" %(time.time()-itime))

        # Delete data from temporal table
        self.db.insert_db_content( PianaGlobals.DeleteTempTableIdentifier.get_sqlquery(),
                                   answer_mode=None, buffer= None)

        #Delete indices on temporal table
        self.db.insert_db_content( PianaGlobals.DropIndice.get_sqlquery(table=PianaGlobals.temporalData_table,
                                                                        indice_name="temp_index"),
                                   answer_mode=None, buffer= None)
                                   

    def disable_buffer(self):
        """
        Disables the use of the insert buffer. If the buffer contains anything, it empties it first
        """

        if self.insert_buffer is None:
            raise ValueError("Cannot disable buffer: It was disabled before\n")
        else:
            self.execute_insert_buffer()
            self.insert_buffer = None


    def _get_max_allowed_packet_(self):
        """
        Get the size of the maximum allowed packet in database server
        """

        maxpacket = self.db.select_db_content(PianaGlobals.GetMaxPacket.get_sqlquery(), answer_mode="raw")

        if maxpacket:
            return maxpacket[0][1]
        else:
            return None


class Buffer(object):
    """
    Class used as an buffer for piana Inserts
    """
    def __init__(self, max_size, dbaccessObject):

        self.buffer = {}

        # Maxim size for each buffer... I take a margin of 10k aprox (it works)
        self.maxsize = int(max_size) - 10000
        
        # To connect with the PianaDBaccess Object that is using this buffer
        self.piana_access = dbaccessObject

        # Store the number of bytes (caracters) saved into the buffer
        self.bytes = 0

    def get_buffer(self):
        """
        Returns the buffer dictionary
        """
        return self.buffer

    def get_buffer_element(self,key):

        if self.buffer.has_key(key):
            return self.buffer[key]
        else:
            raise ValueError("Buffer has not Buffer Element with key %s\n" %(key) )

    def insert2buffer(self, key, table, columns, values):
        """
        Inserts a query into the insert buffer

        Alert! Only used for insert buffers!!!
        """

        # Check if exists a buffer element with key "key"
        if self.buffer.has_key(key):
            # Insert if the values can be inserted in Buffer Element or restart it
            if (self.buffer[key]).insert_values(values) is None:
                self.piana_access.execute_insert_buffer(key_buffer=key)
                self.buffer[key].restart_bufferElement(values)
        else:
            # Create a new buffer element
            self.buffer[key] = BufferElement(self.maxsize,table,columns,values)

            
    def insert2tempbuffer(self, key, table, columns, conditions):
        """
        Inserts a query into the temptable buffer
        """

        # Check if exists a buffer element with key "key"
        if self.buffer.has_key(key):
            pass
        else:
            self.buffer[key] = TemporalTableBuffer(self.maxsize, table, columns, conditions)


class BufferElement(object):
    """
    Class used in Buffer object to store a buffer element
    """

    def __init__(self,max_size,table,columns,values=None):
        """
        Initializes a new buffer element

        "max_size" is the maximum size of the buffer element

        "table" is the table name in the database

        "columns" is a tuple with the names of the columns

        "values" is a tuple with the values to insert

        """
        self.max_size = max_size
        self.table = table
        self.columns = columns
        self.values = {}
        self.initial_size = len(table)+4

        self.size = self.initial_size

        # Sum the length of the column plus 1 (the comma)
        for i in self.columns:
            self.size += len(i)+1

        # Insert the values to the buffer
        if values:
            self.insert_values(values=values)

        

    def getTable(self):

        return self.table

    def getColumns(self):

        return self.columns

    def getValues(self):

        return self.values

    def getSize(self):

        return self.size


    def insert_values(self,values):
        """
        Insert into buffer element a new tuple of values

        "values" must be a tuple of values to insert

        If the values cannot be inserted because the size has been exceeded, return None
        
        """

        # First, calculate the bytes size of this values
        # Sum the length of the value plus the comma plus the 2 parenthesis plus the 2"
        # plus the comma between two inserts
        actual_size=0
        for i in values:
            actual_size += len(i)+5+2

        if self.size+actual_size >= self.max_size:
            # values cannot be inserted: buffer size is exceeded
            return None
        else:
            #insert the new tuple of values and increment the size
            self.values[values]=None
            self.size += actual_size

            return 1

    def restart_bufferElement(self,values):
        """
        Restarts a Buffer Element to its initial values and insert the new tuple of values
        """

        # Empties the dictionary of tuples of values to insert and reset the size
                
        self.size = self.initial_size
        self.values={}
        self.values[values]=None

        # Calculate the bytes size of this values
        # Sum the length of the value plus the comma plus the 2 parenthesis plus the 2"
        # plus the comma between two inserts
        for i in values:
            self.size += len(i)+5+2



class TemporalTableBuffer(BufferElement):
    """
    Class to store the information about the temporal table

    The difference with BufferElement is a new attribute and that it will not contain values to insert (they will be inserted in a table of the database)
    """

    def __init__(self,max_size,table,columns,conditions):

        # List of condition column names will be saved. It is important to be a list as the order is important (because of the indices in database)
        # There is a maxim of ? attributes

        self.attributes_col = columns
        self.conditions_col = conditions

        BufferElement.__init__(self,max_size,table,tuple(columns),values=None)


    def get_conditions_list(self):

        return self.conditions_col
    
    def get_columns_tuple(self):

        return self.attributes_col
