"""
File        : gene_info2accession2piana.py
Author      : Javier Garcia and Ramon Aragues
Creation    : May 2007
Contents    : script that fills up tables in database piana related to  external DB "Gene"
Called from : 

=======================================================================================================

This file parsers gene_info NCBI file and inserts information into PianaDb

Command line option '--help' describes usage of this program

For more details on how to use it, read piana/README.populate_piana_db

"""


#=======================================================================================================
# This file parsers a Gene NCBI text file and inserts information into PianaDb
# Command line option '--help' describes usage of this program
# For more details on how to use it, read piana/README.populate_piana_db
#
# gene2accession2piana.py: script that fills up tables in database piana related to external DB "Gene" from NCBI
#
# Copyright (C) 2005  Ramon Aragues
# author email: ramon.aragues@upf.edu and boliva@imim.es
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#    http://www.gnu.org/copyleft/gpl.html
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# University Pompeu Fabra, hereby disclaims all copyright
# interest in the program 'PIANA'
# (software for working with protein-protein interaction networks) written 
# by Ramon Aragues

## STEP 1: IMPORT NECESSARY MODULES

import sys
import getopt

import re
import readline

import MySQLdb

import time

from PianaDBaccess import *
from parserUtilities import *


# ---------------------------------------------------------------
# Set here the default values for command line arguments
# ---------------------------------------------------------------
#
## STEP 2: DEFINE PARSER UTILITY AND READ PARSER ARGUMENTS

# Define parser parameters
# There must exist two lists:
#           - compulsory parameters
#           - optional parameters
# Arguments are defined as a list of tuples with the following format: (parameter, default_value, parameter_description)
# If parameter ends with "=" it means that the parameters must contain a value. Otherwise, the parameter does not accept values


compulsory_arguments_list = [("input-file-name=",None,"file name of input file containing genpept data (gene_info)"),
                             ("piana-dbname=",None,"name of database piana to be used (required)"),
                             ("piana-dbhost=",None,"name of host where database piana to be used is placed (required)"),
                             ("database-name=","ncbi_gene","internal identifier name to this database (it must be unique in the database)"),
                             ("database-version=",None,"version of the database to be inserted"),
                             ("database-description=",None,"Description of the database to be inserted")]
                             

                          
optional_arguments_list = [("piana-dbuser=",None,"username accessing the database (not required in most systems)"),
                           ("piana-dbpass=",None,"password of username accessing the database (not required in most systems)"),
                           ("help",None,"prints this message and exits"),
                           ("verbose",0,"prints process info to stdout"),
                           ("time-control",None,"prints to stderr a control of the timing of the parser"),
                           ("log-file=",None,"Prints a log file of the parsing result (number of inserted proteins, references...)"),
                           ("database-information=",None,"Type of information this database inserts to database. It can be one or more elements of the following list (separated by commas): \"protein sequences\",\"protein attributes\",\"identifiers cross-references\",\"protein-protein interaction\"")]



# Define the parser information to print in the help:

parser = ParserObject( script_name="gene_info2accession_parser.py",
                       script_description="This program fills up tables in database piana related to external DB 'Gene' (Inserts relation geneID-geneName",
                       compulsory_arguments=compulsory_arguments_list,
                       optional_arguments=optional_arguments_list )


# Read parser parameters (parameters are returned in the same order as defined in parameters list, in order: compulsory,optional)

arguments_dict = parser.parseArguments()

gene_info_file = arguments_dict["input-file-name"]
piana_dbname = arguments_dict["piana-dbname"]
piana_dbhost = arguments_dict["piana-dbhost"]
source_db = arguments_dict["database-name"]
database_version = arguments_dict["database-version"]
database_description = arguments_dict["database-description"]
piana_dbuser = arguments_dict["piana-dbuser"]
piana_dbpass = arguments_dict["piana-dbpass"]
help = arguments_dict["help"]
verbose = arguments_dict["verbose"]
time_control = arguments_dict["time-control"]
log_file= arguments_dict["log-file"]
database_information = arguments_dict["database-information"]

parser.parseArguments()




# Start time counter
if time_control:
    initial_time = time.time()

# Open log file
if log_file:
    log_file_fd = file(log_file, 'w')


# List of tables to lock. It is used to improve speed inserts, as the indices are not updated in each insert
tables_to_lock = [PianaGlobals.geneName_table,
                  PianaGlobals.geneID_table,
                  PianaGlobals.temporalData_table,
                  PianaGlobals.protein_table,
                  PianaGlobals.proteinDescription_table]

# --------
# --------
#  Main()               
# --------                               
# --------


# Initialisating connection to piana
piana_access = PianaDBaccess(dbname=piana_dbname, dbhost=piana_dbhost, dbuser=piana_dbuser, dbpassword= piana_dbpass)

# Introduce database info into PIANA database
piana_access.insert_new_external_database(databaseName = source_db,
                                          databaseVersion = database_version,
                                          parsedFile = gene_info_file.split("/")[-1],
                                          databaseDescription = database_description,
                                          databaseInformation = database_information)


# Locking tables to be updated
piana_access.lock_tables(table_list = tables_to_lock )

gene_fd = open(gene_info_file,"r")

#
# Reading external DB "Gene" and inserting its data into piana DB
# 

if verbose:
    sys.stderr.write( "Reading data from gene_info file...\n")

line_counter=0


# This dictionary will contain all associations between geneID and proteinPiana inserted (because there are some entries repeated in gene2accession file)
# For example: different geneID entries with status provisional... (Maybe it would be better not taking into account this entries...)
# It is used in order to not duplicate inserts
inserted_geneID_dict = {}

for line in gene_fd:

    line_counter += 1

    # Check the frequency of lock/unlocking tables
    piana_access.check_lock_frequency(tables_to_lock)

    if time_control:
        if not line_counter%20000:
            sys.stderr.write("%s lines in %s seconds\n" %(line_counter,time.time()-initial_time))

    if line[0] != '#':

        # Parsing the line using the regular expression followed by taxonomy data
        line_fields = line.split("\t")


        #      line_fields[0] = tax_id
        #      line_fields[1] = GeneID
        #      line_fields[2] = Symbol
        #      line_fields[3] = LocusTag
        #      line_fields[4] = Synonyms
        #      line_fields[5] = dbXrefs
        #      line_fields[6] = chromosome
        #      line_fields[7] = map_location
        #      line_fields[8] = description
        #      line_fields[9] = type_of_gene
        #      line_fields[10]= Symbol_from_nomenclature_authority
        #      line_fields[11]= Full_name_from_nomenclature_authority
        #      line_fields[12]= Nomenclature_status
        #      line_fields[13]= Other_designations



        # Example: 10090   19229   Ptk2b   -       A6|CADTK|CAKB|CAK[B]|CAKbeta|E430023O05Rik|FAK2|PKB|PTK|PYK2|Raftk      MGI:104908      14      14 D1|14 28.0 cM        PTK2 protein tyrosine kinase 2 beta  protein-coding  Ptk2b   PTK2 protein tyrosine kinase 2 beta     O       calcium-dependent tyrosine kinase|cellular adhesion kinase beta|proline-rich tyrosine kinase 2|protein tyrosine kinase 2 beta|related adhesion focal tyrosine kinase


        # We are obtaining the proteinPianas associated to these geneIDs and assigning those geneNames (as primaries!!!) to these proteinPianas
        if line_fields:
            
            #Get the proteinPiana list that correspond to this geneID Code
            proteinPiana_list = piana_access.get_list_proteinPiana(protein_id=line_fields[1],
                                                                   id_type="geneID",
                                                                   tax_id_value=line_fields[0], source_db_info="no" )

            if line_fields[2] != "-":
                if proteinPiana_list:
                    # Insert Symbol as primary geneName
                    for actual_proteinPiana in proteinPiana_list:
                        piana_access.insert_geneName_code( geneName_code_value = line_fields[2].strip(),
                                                           proteinPiana_value = actual_proteinPiana,
                                                           geneName_source_value = source_db,
                                                           isPrimary_value = 1 )
                        parser.add_to_log("Number of geneName inserted")


                        # Insert Synonyms as secondary geneName
                        if line_fields[4] != "-":
                            secondary_geneNames = line_fields[4].split("|")
                            for actual_secondary_geneName in secondary_geneNames:
                                piana_access.insert_geneName_code( geneName_code_value = actual_secondary_geneName.strip(),
                                                                   proteinPiana_value = actual_proteinPiana,
                                                                   geneName_source_value = source_db,
                                                                   isPrimary_value = 0 )
                                parser.add_to_log("Number of secondary geneName inserted")

                        # Insert description
                        line_fields[8] = line_fields[8].replace('"'," ").replace("\n", " ").replace("'"," ").replace('\\', " ").strip()
                        if line_fields[8].strip() != "-":
                            piana_access.insert_protein_description(description = line_fields[8],
                                                                    proteinPiana_value = actual_proteinPiana,
                                                                    proteinDescriptionSource_value = source_db)
                            parser.add_to_log("Number of descriptions inserted")

                        # Insert Other Designations as descriptions of the protein
                        line_fields[13] = line_fields[13].replace('"'," ").replace("\n", " ").replace("'"," ").replace('\\', " ").strip()
                        if line_fields[13] != "-":
                            designations = line_fields[13].split("|")
                            for actual_designation in designations:
                                if actual_designation != "-":
                                    piana_access.insert_protein_description(description=actual_designation,
                                                                            proteinPiana_value = actual_proteinPiana,
                                                                            proteinDescriptionSource_value = source_db)
                                    parser.add_to_log("Number of descriptions inserted")

                    
                else:
                    parser.add_to_log("Number of geneIDs not found")
                    # If any proteinPiana is found (it means geneID not exists), try to search the proteinPianas using geneName isPrimary
                    proteinPiana_list = piana_access.get_list_proteinPiana( protein_id = line_fields[1],
                                                                            id_type = "geneName",
                                                                            tax_id_value = line_fields[0],
                                                                            only_primary="yes" )

                    if proteinPiana_list:
                        for actual_proteinPiana in proteinPiana_list:
                            piana_access.insert_geneID_code( geneID_code_value = line_fields[1],
                                                             proteinPiana_value = actual_proteinPiana,
                                                             geneID_source_value = source_db)
                            parser.add_to_log("Number of geneIDs inserted")
                                                             

        # END OF if line_fields:
    # END OF line[0]!='#'
# END OF for line in gene_fd:
                    
                              

## LAST STEP: CLOSE DATABASE CONNECTION    IMPORTANT !!!!
## As PianaDBaccess uses an internal buffer, it is necessary to close the connection to sure that all inserts are correctly done, as well as unlock tables

piana_access.close()

if time_control:
    sys.stderr.write("\n Total time: %s \n" %(time.time()-initial_time) )

if verbose:
    sys.stdour.write("%s\n" %(parser.get_log_string()) )

if log_file:
    log_file_fd.write(parser.get_log_string())
    log_file_fd.close()

